/* $Id: Meme.java,v 1.2 1999/11/06 19:50:59 borg Exp $ */
/* Copyright  1998-1999 George Reese, All Rights Reserved */
package com.imaginary.meme;

import java.io.Serializable;
import java.util.Locale;
import com.sun.java.util.collections.HashMap;
import com.sun.java.util.collections.Iterator;

/**
 * A representation of a single concept that has many translations.  It is
 * generally something as simple as a word, like "red".  It can, however,
 * represent more complex memes like "a barrel full of monkeys" or
 * even full sentences or paragraphs.  An example of the use of this
 * class is:
 * <CODE>
 * Meme meme = getSomeMeme();
 * System.out.println(meme.getTranslation(Locale.getDefault));
 * </CODE>
 * <BR>
 * Last modified $Date: 1999/11/06 19:50:59 $
 * @version $Revision: 1.2 $
 * @author George Reese (borg@imaginary.com)
 */
public class Meme implements Cloneable, Serializable {
    /**
     * The serial version ID.
     */
    static final long serialVersionUID = -6690906953844033820L;

    /**
     * The default translation used for unresolvable locales.
     * @serial
     */
    private String    defaultTranslation = null;
    /**
     * A key used to identifier this meme.
     * @serial
     */
    private String    key                = null;
    /* A map of locales to translations.
     * @serial
     */
    private HashMap   translations       = new HashMap();

    /**
     * Constructs an empty meme object.
     */
    public Meme() {
        super();
    }

    /**
     * Constructs a meme object having the specified key.
     * @param the value to use as the meme key
     */
    public Meme(String k) {
        super();
        key = k;
    }

    /**
     * Constructs a meme object having the specified key and default
     * translation.
     * @param k the meme key
     * @param def the default translation
     */
    public Meme(String k, String def) {
        super();
        key = k;
        defaultTranslation = def;
    }
    
    /**
     * Adds a translation for this meme for the specified Locale.
     * @param l the Locale for which you wish to add a translation
     * @param meme the translated meme
     */
    public void addTranslation(Locale l, String ph) {
        String language = l.getLanguage();
        String ctry = l.getCountry();
        HashMap trans;

        if( ctry == null || ctry.length() < 2 ) {
            ctry = "default";
        }
        if( !translations.containsKey(language) ) {
            trans = new HashMap();
            translations.put(language, trans);
        }
        else {
            trans = (HashMap)translations.get(language);
        }
        trans.put(ctry, ph);
    }

    /**
     * The default translation is the translation for this ph that
     * is provided whenever a translation is requested for an unsupported
     * Locale.
     * @return the default translation
     */
    public String getDefaultTranslation() {
        return defaultTranslation;
    }

    /**
     * The ph key is an identifier that can be used to uniquely
     * identify this ph.  For example, "RED" might be a good
     * ph key for the color red.
     * @return the ph key
     */
    public String getKey() {
        return key;
    }

    /**
     * Provides a translation of this Meme for the specified Locale.
     * @param loc the Locale for which a translation is desired
     * @return the translated meme
     */
    public synchronized String getTranslation(Locale loc) {
        String language = loc.getLanguage();
        String ctry = loc.getCountry();
        HashMap trans;

        // See if this language has a translation
        if( !translations.containsKey(language) ) {
            // No translation exists, use the server's default Locale
            loc = Locale.getDefault();
            language = loc.getLanguage();
            // See if there is a translation for the server default
            if( !translations.containsKey(language) ) {
                return defaultTranslation;
            }
        }
        // Get the list of translations for this language
        trans = (HashMap)translations.get(language);
        // Check if there is a special translation for this country
        if( ctry == null || !trans.containsKey(ctry) ) {
            ctry = "default";
            if( !trans.containsKey(ctry) ) {
                Iterator ctrys = trans.keySet().iterator();

                if( !ctrys.hasNext() ) {
                    return defaultTranslation;
                }
                else {
                    ctry = (String)ctrys.next();
                }
            }
        }
        return (String)trans.get(ctry);
    }

    /**
     * Sets the default translation.
     * @param trans the default translation
     */
    public synchronized void setDefaultTranslation(String trans) {
        defaultTranslation = trans;
    }

    /**
     * Sets the meme key.
     * @param key the meme key
     */
    public synchronized void setKey(String key) {
        key = key;
    }
}
