/* $Id: ResultSetIterator.java,v 1.1.1.1 1999/11/06 18:38:04 borg Exp $ */
/* Copyright  1999 George Reese, All Rights Reserved */
package com.imaginary.util;

import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import com.sun.java.util.collections.HashMap;
import com.sun.java.util.collections.Iterator;
import com.sun.java.util.collections.NoSuchElementException;
import com.sun.java.util.collections.UnsupportedOperationException;

/**
 * Provides an <CODE>Iterator</CODE> interface into a JDBC
 * <CODE>ResultSet</CODE>. Using this iterator, you can move through a
 * result set sequentially and pull each row out as a <CODE>HashMap</CODE>
 * whose keys are column names and whose values are the column values.
 * This class will currently function poorly for result sets with
 * two columns of the same name.
 * <BR>
 * Last modified $Date: 1999/11/06 18:38:04 $
 * @version $Revision: 1.1.1.1 $
 * @author George Reese (borg@imaginary.com)
 */
public class ResultSetIterator implements Iterator {
    private boolean           complete    = false;
    private int               columnCount = -1;
    private ResultSetMetaData metaData    = null;
    private Object            nextRow     = null;
    private ResultSet         resultSet   = null;
    private Stack             rows        = new FifoStack();

    /**
     * Constructs a new <CODE>ResultSetIterator</CODE> using the
     * specified <CODE>ResultSet</CODE> as its source of data.
     * @param rs the source of data for this iterator
     * @throws java.sql.SQLException could not get the meta data
     */
    public ResultSetIterator(ResultSet rs) throws SQLException {
        super();
        resultSet = rs;
        metaData = rs.getMetaData();
        columnCount = metaData.getColumnCount();
        Thread t = new Thread() {
            public void run() {
                load();
            }
        };

        t.setPriority(Thread.MIN_PRIORITY);
        t.start();
    }

    /**
     * @return true if there are more rows to be processed
     */
    public synchronized boolean hasNext() {
        if( nextRow != null ) {
            return true;
        }
        while( rows.isEmpty() && !complete ) {
            try { wait(1000); }
            catch( InterruptedException e ) { }
        }
        if( rows.isEmpty() ) {
            return false;
        }
        nextRow = rows.pop();
        return true;
    }

    /**
     * Loads the result set data into the iterator in a background thread.
     */
    private void load() {
        try {
            while( resultSet.next() ) {
                HashMap row = new HashMap();
                
                for(int i=0; i<columnCount; i++) {
                    row.put(metaData.getColumnName(i), resultSet.getObject(i));
                }
                synchronized( this ) {
                    rows.push(row);
                    notifyAll();
                }
            }
            synchronized( this ) {
                complete = true;
                notifyAll();
            }
        }
        catch( SQLException e ) {
            synchronized( this ) {
                complete = true;
                notifyAll();
            }
            e.printStackTrace();
        }
    }

    /**
     * Provides the next row in the result set as a <CODE>HashMap</CODE>
     * where the keys are the column names and the values are the column
     * values.
     * @return the next row as a <CODE>HashMap</CODE>
     */
    public synchronized Object next() {
        Object ob;
        if( nextRow == null && !hasNext() ) {
            throw new NoSuchElementException("No more rows in result set.");
        }
        ob = nextRow;
        nextRow = null;
        return ob;
    }

    /**
     * This method is currently unsupported.
     * @throws java.lang.UnsupportedOperationException always thrown
     */
    public void remove() {
        throw new UnsupportedOperationException("Not yet supported.");
    }
}
