/**
 * $Id$
 * 
 * Copyright (c) 1998-1999 Lotus Corporation, Inc. All Rights Reserved.
 *				This software is provided without a warranty of any kind.
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 * 
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 * 
 * $State: Exp $
 */
package com.lotus.xsl;

import java.io.Writer;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Stack;
import org.xml.sax.*;

/**
 * FormatterToXML formats SAX-style events into XML.
 */
public class FormatterToXML implements FormatterListener 
{

  /** 
   * The writer where the XML will be written.
   */
  protected  Writer m_writer      =   null;
  
  /**
   * If true, cdata sections are simply stripped of their 
   * CDATA brackets, without escaping.
   */
  public boolean m_stripCData = false;

  /**
   * If true, characters in cdata sections are 
   * escaped, instead of being writted out as 
   * cdata sections.
   */
  public boolean m_escapeCData = false;

  /**
   * The character encoding.  Not currently used.
   */
  protected  String m_encoding    =   null;
  
  public boolean m_shouldWriteXMLHeader = false;
  
  /**
   * A stack of Boolean objects that tell if the given element 
   * has children.
   */
  protected  Stack m_elemStack = new Stack();
  
  protected boolean m_ispreserve    =   false;
  protected Stack   m_preserves     =   new Stack();
  protected boolean m_isprevtext    =   false;
  protected boolean m_doIndent =   false;
  protected int     m_currentIndent =   0;
  public int     indent        =   2;    
  protected   int     level   =   0;
  protected   String  doctype =   null;
  boolean m_startNewLine;
  
  /**
   * These are characters that will be escaped in the output.
   */
  public String m_attrSpecialChars = "<>&\"\'\r\n";

  /**
   * Constructor for customized encoding.
   * @param writer        The character output stream to use.
   * @param encoding      Java character encoding in use by <VAR>writer</VAR>.
   */
  public FormatterToXML(Writer writer, String encoding) 
  {
    this(writer, encoding, null, false, 0);
  }

  /**
   * Constructor for default encoding.
   * @param writer  The character output stream to use.
   */
  public FormatterToXML(Writer writer) 
  {
    this(writer, null, null, false, 0);
  }
  
  /**
   * Constructor for customized encoding and doctype.
   * @param writer        The character output stream to use.
   * @param encoding      Java character encoding in use by <VAR>writer</VAR>.
   * @param doctype       String to be printed at the top of the document.
   * @param indent        Number of spaces to indent at each nesting level.
   */
  public FormatterToXML(Writer writer, String encoding, 
                                String doctype, boolean doIndent, 
                                int indent) 
  {
    this.m_writer   =   writer;
    this.m_encoding =   encoding;
    this.doctype = doctype;
    this.m_doIndent = doIndent;
    this.indent = indent;
  }
  
  /**
   * Receive an object for locating the origin of SAX document events.
   *
   * @param locator An object that can return the location of
   *                any SAX document event.
   * @see org.xml.sax.Locator
   */
  public void setDocumentLocator (Locator locator)
  {
    // I don't do anything with this yet.
  }

  /**
    * Receive notification of the beginning of a document.
    *
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    */
  public void startDocument ()
    throws SAXException
  {
    try
    {
      m_startNewLine = false;
      
      if(m_shouldWriteXMLHeader)
      {
        this.m_writer.write("<?xml version=\"1.0\"?>\n\n");
      }
        
      if (this.doctype != null)
        this.m_writer.write(doctype);
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe);
    }
  }

  /**
   * Receive notification of the end of a document.
   *
   * @exception org.xml.sax.SAXException Any SAX exception, possibly
   *            wrapping another exception.
   */
  public void endDocument ()
    throws SAXException
  {
    try
    {
      this.m_writer.flush();
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe);
    }
  }

  /**
    * Receive notification of the beginning of an element.
    *
    * @param name The element type name.
    * @param atts The attributes attached to the element, if any.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see #endElement
    * @see org.xml.sax.AttributeList 
    */
  public void startElement (String name, AttributeList atts)
    throws SAXException
  {
    try
    {
      writeParentTagEnd();
      m_ispreserve = false;
      //  System.out.println(name+": m_doIndent = "+m_doIndent+", m_ispreserve = "+m_ispreserve+", m_isprevtext = "+m_isprevtext);
      if (shouldIndent() && !m_elemStack.isEmpty()) 
      {
        m_startNewLine = true;
        indent(this.m_writer, m_currentIndent);
      }

      this.m_writer.write('<');
      this.m_writer.write(name);
      
      int nAttrs = atts.getLength();
      for (int i = 0;  i < nAttrs ;  i++)
      {
        processAttribute(atts.getName(i), atts.getValue(i));
      }
      // Flag the current element as not yet having any children.
      openElementForChildren();
      
      m_currentIndent += this.indent;
      
      m_isprevtext = false;
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe);
    }
  }
  
  /**
   * Check to see if a parent's ">" has been written, and, if 
   * it has not, write it.
   */
  protected void writeParentTagEnd()
    throws SAXException
  {
    try
    {
      if(!m_elemStack.empty())
      {
        // See if the parent element has already been flagged as having children.
        if(false == ((Boolean)m_elemStack.peek()).booleanValue())
        {
          this.m_writer.write(">");
          m_elemStack.pop();
          m_elemStack.push(new Boolean(true));
          m_preserves.push(new Boolean(m_ispreserve));
        }
      }
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe);
    }
  }
  
  /**
   * Flag the current element as not yet having any 
   * children.
   */
  protected void openElementForChildren()
  {
    // Flag the current element as not yet having any children.
    m_elemStack.push(new Boolean(false));
  }
  
  /**
   * Tell if child nodes have been added to the current
   * element.  Must be called in balance with openElementForChildren().
   */
  protected boolean childNodesWereAdded()
  { 
       return m_elemStack.isEmpty() ? 
              false : ((Boolean)m_elemStack.pop()).booleanValue();
  }

  /**
    * Receive notification of the end of an element.
    *
    * @param name The element type name
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    */
  public void endElement (String name)
    throws SAXException
  {
    try
    {
      m_currentIndent -= this.indent;
      boolean hasChildNodes = childNodesWereAdded();
      if (hasChildNodes) 
      {
        if (shouldIndent()) 
          indent(this.m_writer, m_currentIndent);
        this.m_writer.write("</" + name + ">");
      }
      else
      {
        this.m_writer.write("/>");
      }
      if (hasChildNodes) 
      {
        m_ispreserve = m_preserves.isEmpty() 
                       ? false : ((Boolean)m_preserves.pop()).booleanValue();
      }
      m_isprevtext = false;
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe);
    }
  }

  /**
   * Process an attribute.
   * @param   name   The name of the attribute.
   * @param   value   The value of the attribute.
   */
  protected void processAttribute(String name, String value) 
    throws SAXException
  {
    try
    {
      this.m_writer.write(" ");
      this.m_writer.write(name);
      this.m_writer.write("=\"");
      this.m_writer.write(prepAttrString(value, m_attrSpecialChars, this.m_encoding));
      this.m_writer.write("\"");
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe);
    }
  }

  /**
    * Receive notification of a processing instruction.
    *
    * @param target The processing instruction target.
    * @param data The processing instruction data, or null if
    *        none was supplied.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    */
  public void processingInstruction (String target, String data)
    throws SAXException
  {
    try
    {
      writeParentTagEnd();
      if (shouldIndent())  
        indent(this.m_writer, m_currentIndent);
      this.m_writer.write("<?" + target);
      if (data.length() > 0 && !Character.isSpaceChar(data.charAt(0)))
        this.m_writer.write(" ");
      this.m_writer.write(data + "?>");
      m_startNewLine = true;
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe);
    }
  }
  

  /**
   * Called when a Comment is to be constructed.
   * @param   data  The comment data.
   * @exception org.xml.sax.SAXException Any SAX exception, possibly
   *            wrapping another exception.
   */
  public void comment(String data) throws SAXException
  {
    try
    {
      writeParentTagEnd();
      if (shouldIndent())  
        indent(this.m_writer, m_currentIndent);
      this.m_writer.write("<!--" + data + "-->");
      m_startNewLine = true;
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe);
    }
  }

  /**
    * Receive notification of cdata.
    *
    * <p>The Parser will call this method to report each chunk of
    * character data.  SAX parsers may return all contiguous character
    * data in a single chunk, or they may split it into several
    * chunks; however, all of the characters in any single event
    * must come from the same external entity, so that the Locator
    * provides useful information.</p>
    *
    * <p>The application must not attempt to read from the array
    * outside of the specified range.</p>
    *
    * <p>Note that some parsers will report whitespace using the
    * ignorableWhitespace() method rather than this one (validating
    * parsers must do so).</p>
    *
    * @param ch The characters from the XML document.
    * @param start The start position in the array.
    * @param length The number of characters to read from the array.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see #ignorableWhitespace 
    * @see org.xml.sax.Locator
    */
  public void cdata (char ch[], int start, int length)
    throws SAXException
  {
    try
    {
      if(m_escapeCData)
      {
        characters(ch, start, length);
      }
      else
      {
        writeParentTagEnd();
        m_ispreserve = true;
        if (shouldIndent())
          indent(this.m_writer, m_currentIndent);
        if(!m_stripCData)
          this.m_writer.write("<![CDATA[");
        this.m_writer.write(ch, start, length);
        if(!m_stripCData)
          this.m_writer.write("]]>");
      }
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe);
    }
  }
  
  static final int MAXCHARBUF = 4096;
  static final int MAXSAFECHARBUF = MAXCHARBUF-256;
  static char[] m_charBuf = new char[MAXCHARBUF];
  
  /**
   * Receive notification of character data.
   *
   * <p>The Parser will call this method to report each chunk of
   * character data.  SAX parsers may return all contiguous character
   * data in a single chunk, or they may split it into several
   * chunks; however, all of the characters in any single event
   * must come from the same external entity, so that the Locator
   * provides useful information.</p>
   *
   * <p>The application must not attempt to read from the array
   * outside of the specified range.</p>
   *
   * <p>Note that some parsers will report whitespace using the
   * ignorableWhitespace() method rather than this one (validating
   * parsers must do so).</p>
   *
   * @param chars The characters from the XML document.
   * @param start The start position in the array.
   * @param length The number of characters to read from the array.
   * @exception org.xml.sax.SAXException Any SAX exception, possibly
   *            wrapping another exception.
   * @see #ignorableWhitespace 
   * @see org.xml.sax.Locator
   */
  public void characters (char chars[], int start, int length)
    throws SAXException
  {
    if(0 == length)
      return;
    
    try
    {
      
      writeParentTagEnd();
      m_ispreserve = true;
      int pos = 0;
      for (int i = start;  i < length;  i ++) 
      {
        char ch = chars[i];
        int chNum = (int)ch;
        if ('<' == ch) 
        {
          m_charBuf[pos++] = '&';
          m_charBuf[pos++] = 'l';
          m_charBuf[pos++] = 't';
          m_charBuf[pos++] = ';';
        }
        else if ('>' == ch) 
        {
          m_charBuf[pos++] = '&';
          m_charBuf[pos++] = 'g';
          m_charBuf[pos++] = 't';
          m_charBuf[pos++] = ';';
        }
        else if ('&' == ch) 
        {
          m_charBuf[pos++] = '&';
          m_charBuf[pos++] = 'a';
          m_charBuf[pos++] = 'm';
          m_charBuf[pos++] = 'p';
          m_charBuf[pos++] = ';';
          /**
          } else if ('"' == ch) {
          sb.append("&quot;");
          } else if ('\'' == ch) {
          sb.append("&apos;");
          */
        }
        else if (0xd800 <= chNum && chNum < 0xdc00) 
        {
          // UTF-16 surrogate
          int next;
          if (i+1 >= length) 
          {
            throw new SAXException("Invalid UTF-16 surrogate detected: "
              +Integer.toHexString(ch)+ " ?");
          }
          else 
          {
            next = chars[++i];
            if (!(0xdc00 <= next && next < 0xe000))
              throw new SAXException("Invalid UTF-16 surrogate detected: "
                +Integer.toHexString(ch)+" "+Integer.toHexString(next));
            next = ((ch-0xd800)<<10)+next-0xdc00+0x00010000;
          }
          m_charBuf[pos++] = '&';
          m_charBuf[pos++] = '#';
          m_charBuf[pos++] = 'x';
          String intStr = Integer.toHexString(next);
          int nIntStr = intStr.length();
          for(int k = 0; k < nIntStr; k++)
          {
            m_charBuf[pos++] = intStr.charAt(k);
          }
          m_charBuf[pos++] = ';';
          /*} else if (null != ctbc && !ctbc.canConvert(ch)) {
          sb.append("&#x");
          sb.append(Integer.toString((int)ch, 16));
          sb.append(";");*/
        }
        else if(((chNum >= 20) && (chNum <= 126)) 
                || (chNum == 10)
                || (chNum == 13)
                || (chNum == 9))
        {
          m_charBuf[pos++] = ch;
        }
        else
        {
          m_charBuf[pos++] = '&';
          m_charBuf[pos++] = '#';
          String intStr = Integer.toString(chNum);
          int nIntStr = intStr.length();
          for(int k = 0; k < nIntStr; k++)
          {
            m_charBuf[pos++] = intStr.charAt(k);
          }
          m_charBuf[pos++] = ';';
        }
        // Use 80 as a best guess safe buffer
        if(pos > MAXSAFECHARBUF)
        {
          this.m_writer.write(m_charBuf, 0, pos);
          pos = 0;
        }
      }
      this.m_writer.write(m_charBuf, 0, pos);
      m_isprevtext = true;
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe);
    }
  }
  
  /**
   * Receive notification of ignorable whitespace in element content.
   *
   * Not sure how to get this invoked quite yet.
   *
   * @param ch The characters from the XML document.
   * @param start The start position in the array.
   * @param length The number of characters to read from the array.
   * @exception org.xml.sax.SAXException Any SAX exception, possibly
   *            wrapping another exception.
   * @see #characters
   */
  public void ignorableWhitespace (char ch[], int start, int length)
    throws SAXException
  {
    if(0 == length)
      return;
    characters (ch, start, length);
  }
  
  /**
   * Receive notivication of a entityReference.
   */
  public void entityReference(String name)
    throws SAXException
  {
    try
    {
      writeParentTagEnd();
      if (shouldIndent())  
        indent(this.m_writer, m_currentIndent);
      this.m_writer.write("&");
      this.m_writer.write(name);
      this.m_writer.write(";");
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe);
    }
  }

  /**
   * Returns the specified <var>string</var> after substituting <VAR>specials</VAR>,
   * and UTF-16 surrogates for chracter references <CODE>&amp;#xnn</CODE>.
   *
   * @param   string      String to convert to XML format.
   * @param   specials    Chracters, should be represeted in chracter referenfces.
   * @param   encoding    CURRENTLY NOT IMPLEMENTED.
   * @return              XML-formatted string.
   * @see #backReference
   */
  public static String prepAttrString(String string, String specials, String encoding)
    throws SAXException
  {
    StringBuffer sb = new StringBuffer(string.length()*12/10);

    for (int i = 0;  i < string.length();  i ++) 
    {
      char ch = string.charAt(i);
      int chNum = (int)ch;
      int index = specials.indexOf(chNum);
      if (index >= 0) 
      {
        sb.append("&#");
        sb.append(Integer.toString(ch));
        sb.append(";");
      }
      else if (0xd800 <= chNum && chNum < 0xdc00) 
      {
        // UTF-16 surrogate
        int next;
        if (i+1 >= string.length()) 
        {
          throw new SAXException("Invalid UTF-16 surrogate detected: "
            +Integer.toHexString(ch)+ " ?");
        }
        else 
        {
          next = string.charAt(++i);
          if (!(0xdc00 <= next && next < 0xe000))
            throw new SAXException("Invalid UTF-16 surrogate detected: "
              +Integer.toHexString(ch)+" "+Integer.toHexString(next));
          next = ((ch-0xd800)<<10)+next-0xdc00+0x00010000;
        }
        sb.append("&#x");
        sb.append(Integer.toHexString(next));
        sb.append(";");
        /*} else if (null != ctbc && !ctbc.canConvert(ch)) {
        sb.append("&#x");
        sb.append(Integer.toString((int)ch, 16));
        sb.append(";");*/
      }
      else 
      {
        sb.append(ch);
      }
    }
    return sb.toString();
  }
  
  protected boolean shouldIndent()
  {
    return m_doIndent && (!m_ispreserve && !m_isprevtext);
  }
  
  /**
   * Prints <var>n</var> spaces.
   * @param pw        The character output stream to use.
   * @param n         Number of spaces to print.
   * @exception IOException   Thrown if <var>pw</var> is invalid.
   */
  public static void printSpace(java.io.Writer pw, int n) throws SAXException 
  {
    try
    {
      for (int i = 0;  i < n;  i ++)
        pw.write(" ");
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe);
    }
  }

  /**
   * Prints a newline character and <var>n</var> spaces.
   * @param pw        The character output stream to use.
   * @param n         Number of spaces to print.
   * @exception IOException   Thrown if <var>pw</var> is invalid.
   */
  public void indent(java.io.Writer pw, int n) throws SAXException 
  {
    try
    {
      if(m_startNewLine)
        pw.write("\n");
      
      if(m_doIndent)
      {
        printSpace(pw, n);
      }
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe);
    }
  }

}  //ToXMLStringVisitor
