/**
 * $Id: NodeSorter.java,v 1.1 1998/12/18 18:57:09 sboag Exp $
 * 
 * Copyright (c) 1998-1999 Lotus Corporation, Inc. All Rights Reserved.
 *				This software is provided without a warranty of any kind.
 * 
 * $State: Exp $
 * 
 * @author Scott Boag (scott_boag@lotus.com)
 */
package com.lotus.xsl;

import org.w3c.dom.*;
import java.util.*;
import java.text.*;

/**
 * This class can sort vectors of DOM nodes according to a select pattern.
 * TODO: Optimize this so it can reuse queries for each of the nodes.
 */
class NodeSorter
{
  XSLProcessor m_xslp;
  
  Vector m_keys; // vector of NodeSortKeys
  
  /**
   * TODO: Adjust this for locale.
   */
  NumberFormat m_formatter = NumberFormat.getNumberInstance();
  
  /**
   * Construct a NodeSorter, passing in the XSL Processor 
   * so it can know how to get the node data according to 
   * the proper whitespace rules.
   */
  public NodeSorter(XSLProcessor p)
  {
    m_xslp = p;
  }
  
  /**
   * Given a vector of nodes, sort each node according to 
   * the criteria in the keys.
   * @param v an vector of Nodes.
   * @param keys a vector of NodeSortKeys.
   */
  public void sort(Vector v, Vector keys) 
     throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
 {    
    m_keys = keys;
    // QuickSort2(v, 0, v.size() - 1 );
    int n = v.size();
    Vector scratchVector = new Vector(n);
    scratchVector.setSize(n);
    mergesort(v, scratchVector, 0, n - 1);
  }
  
  /**
   * Return the results of a compare of two nodes.
   * TODO: Optimize compare -- cache the getStringExpr results, key by m_selectPat + hash of node.
   */
  int compare(Node n1, Node n2, int kIndex)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    int result = 0;
    NodeSortKey k = (NodeSortKey)m_keys.elementAt(kIndex);
    String n1String = m_xslp.getQueryEngine().getStringExpr(k.m_selectPat, n1, k.m_namespaceContext);
    String n2String = m_xslp.getQueryEngine().getStringExpr(k.m_selectPat, n2, k.m_namespaceContext);
    
    if(k.m_treatAsNumbers)
    {
      double n1Num, n2Num;
      try
      {
        n1Num = m_formatter.parse(n1String).doubleValue();
      }
      catch(ParseException e)
      {
        // System.out.println("Warning: value could not be treated as number in sort: "+n1String);
        n1Num = 0;
      }
      try
      {
        n2Num = m_formatter.parse(n2String).doubleValue();
      }
      catch(ParseException e)
      {
        // System.out.println("Warning: value could not be treated as number in sort: "+n2String);
        n2Num = 0;
      }
      if((n1Num == n2Num) && ((kIndex+1) < m_keys.size()))
      {
        result = compare(n1, n2, kIndex+1);
      }
      else
      {
        double diff = n1Num - n2Num;
        result = (int)((diff < 0.0) 
                       ? (k.m_descending ? 1 : -1) 
                         : (diff > 0.0) ? (k.m_descending ? -1 : 1)
                                          : 0);
      }
    }
    else
    {
      result = k.m_col.compare(n1String, n2String);
      if(k.m_descending)
      {
        result = -result;
      }
    }
    if(0 == result)
    {
      if((kIndex+1) < m_keys.size())
      {
        result = compare(n1, n2, kIndex+1);
      }
    }
    if(0 == result)
    {
      // I shouldn't have to do this except that there seems to 
      // be a glitch in the mergesort
      result = m_xslp.getQueryEngine().isNodeAfter(n1, n2) ? -1 : 1;
    }
    return result;
  }
  
  /**
   * This implements a standard Mergesort, as described in 
   * Robert Sedgewick's Algorithms book.  This is a better 
   * sort for our purpose than the Quicksort because it 
   * maintains the original document order of the input if 
   * the order isn't changed by the sort.
   */
  void mergesort(Vector a, Vector b, int l, int r)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    if((r-l)>0)
    {
      int m = (r+l) / 2;
      mergesort(a, b, l, m);
      mergesort(a, b, m+1, r);
      int i, j, k;
      for(i = m; i >= l; i--)
      {
        // b[i] = a[i];
        b.setElementAt(a.elementAt(i), i);
      }
      i=l;
      for(j = (m + 1); j <= r; j++)
      {
        // b[r+m+1-j] = a[j];
        b.setElementAt(a.elementAt(j), r+m+1-j);
      }
      j=r;
      for(k = l; k <= r; k++)
      {
        // if(b[i] < b[j])
        if(compare((Node)b.elementAt(i), (Node)b.elementAt(j), 0) <= 0)
        {
          // a[k]=b[i];
          a.setElementAt(b.elementAt(i), k);
          i++;
        }
        else
        {
          // a[k]=b[j]; 
          a.setElementAt(b.elementAt(j), k);
          j--;
        }
      }
    }
  }
  
  /** This is a generic version of C.A.R Hoare's Quick Sort 
   * algorithm.  This will handle arrays that are already
   * sorted, and arrays with duplicate keys.<BR>
   *
   * If you think of a one dimensional array as going from
   * the lowest index on the left to the highest index on the right
   * then the parameters to this function are lowest index or
   * left and highest index or right.  The first time you call
   * this function it will be with the parameters 0, a.length - 1.
   *
   * @param a       an integer array
   * @param lo0     left boundary of array partition
   * @param hi0     right boundary of array partition
   */
  private void QuickSort2(Vector v, int lo0, int hi0)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    int lo = lo0;
    int hi = hi0;

    if ( hi0 > lo0)
    {
      // Arbitrarily establishing partition element as the midpoint of
      // the array.
      Node midNode = (Node)v.elementAt( ( lo0 + hi0 ) / 2 );

      // loop through the array until indices cross
      while( lo <= hi )
      {
        // find the first element that is greater than or equal to 
        // the partition element starting from the left Index.
        while( (lo < hi0) && (compare((Node)v.elementAt(lo), midNode, 0) < 0) )
        {
          ++lo;
        } // end while

        // find an element that is smaller than or equal to 
        // the partition element starting from the right Index.
        while( (hi > lo0) && (compare((Node)v.elementAt(hi), midNode, 0) > 0) )
        {
          --hi;
        }

        // if the indexes have not crossed, swap
        if( lo <= hi ) 
        {
          swap(v, lo, hi);
          ++lo;
          --hi;
        }
      }

      // If the right index has not reached the left side of array
      // must now sort the left partition.
      if( lo0 < hi )
      {
        QuickSort2( v, lo0, hi );
      }

      // If the left index has not reached the right side of array
      // must now sort the right partition.
      if( lo < hi0 )
      {
        QuickSort2( v, lo, hi0 );
      }
    }
  } // end QuickSort2

  /**
   * Simple function to swap two elements in 
   * a vector.
   */
  private void swap(Vector v, int i, int j)
  {
    Node node = (Node)v.elementAt(i); 
    v.setElementAt(v.elementAt(j), i);
    v.setElementAt(node, j);
  }

}

/**
 * $Log: NodeSorter.java,v $
 * Revision 1.1  1998/12/18 18:57:09  sboag
 * New
 *
 */
