package com.lotus.xsl;

import java.net.*;
import java.io.*;
import java.util.*;
import java.lang.*;
import java.lang.reflect.*;
import com.ibm.xml.parser.*;
import com.ibm.xml.parser.util.*;
import org.w3c.dom.*;
import com.lotus.xsl.*;

/**
 * This class handles the command line interface for LotusXSL.  It ends up 
 * creating one of the XML4J liaisons, and calling the XSLProcessor class.
 */
public class Process
{
  /**
   * Print argument options.
   */ 
  protected static void printArgOptions()
  {
    System.out.println("xslproc options: ");
    System.out.println("    -IN inputXMLURL");
    System.out.println("   [-XSL XSLTransformationURL]");
    System.out.println("   [-OUT outputFileName]");
    System.out.println("   [-PARSER fully qualified class name of parser liaison]");
    System.out.println("   [-E (Do not expand entity refs)]");
    System.out.println("   [-V (Version info)]");
    System.out.println("   [-QC (Quiet Pattern Conflicts Warnings)]");
    System.out.println("   [-Q  (Quiet Mode)]");
    System.out.println("   [-LF (Use linefeeds only on output {default is CR/LF})]");
    System.out.println("   [-CR (Use carriage returns only on output {default is CR/LF})]");
    System.out.println("   [-ESCAPE (Which characters to escape {default is <>&\"\'\\r\\n}]");
    System.out.println("   [-INDENT (Control how many spaces to indent {default is 0})]");
    System.out.println("   [-TT (Trace the templates as they are being called.)]");
    System.out.println("   [-TS (Trace each select.)]");
    System.out.println("   [-TTC (Trace the template children as they are being processed.)]");
    System.out.println("   [-VALIDATE (Set whether validation occurs.  Validation is off by default.)]");
    System.out.println("   [-EDUMP (Do stackdump on error.)]");
    System.out.println("   [-XML (Use XML formatter and add XML header.)]");
    System.out.println("   [-TEXT (Use simple Text formatter.)]");
    System.out.println("   [-HTML (Use HTML formatter.)]");
    System.out.println("   [-STRIPCDATA (Simply strip CDATA sections. Must have -XML or -HTML)]");
    System.out.println("   [-ESCAPECDATA (Escape characters in CDATA.  Must have -XML)]");
    System.out.println("   [-PARAM name expression (Set a stylesheet parameter)]");
  }
  
  /**
   * Command line interfact to transform the XML according to 
   * the instructions found in the XSL document.
   * <pre>
   *    -in inputXMLURL
   *    -xsl XSLTransformationURL
   *    -out outputFileName
   *    -F (Format output pretty-printed)
   * </pre>
   */
  public static void main( String argv[] )    
  {    
    boolean doStackDumpOnError = false;
    boolean setQuietMode = false;

    // Runtime.getRuntime().traceMethodCalls(false);  
    // Runtime.getRuntime().traceInstructions(false);
    /**
    * The default diagnostic writer...
    */
    java.io.PrintWriter diagnosticsWriter = new PrintWriter(System.err, true);
    
    if(argv.length < 1)
    {
      printArgOptions();
    }
    else
    {
      XMLParserLiaison xmlProcessorLiaison;
      try
      {
        String parserLiaisonClassName = "com.lotus.xsl.xml4j2dom.XML4JLiaison4dom";
        boolean usingDefault = true;
        for (int i = 0;  i < argv.length;  i ++) 
        {
          if("-PARSER".equalsIgnoreCase(argv[i]))
          {
            i++;
            parserLiaisonClassName = argv[i];
            usingDefault = false;
          }
        }
        
        Class parserLiaisonClass = Class.forName(parserLiaisonClassName);
        
        Constructor parserLiaisonCtor = parserLiaisonClass.getConstructor(null);
        xmlProcessorLiaison 
          = (XMLParserLiaison)parserLiaisonCtor.newInstance(null);
      }
      catch(Exception e)
      {
        System.err.println("Could not create XML Processor ");
        return;
      }
      
      XSLProcessor processor = new XSLProcessor(xmlProcessorLiaison);
      boolean formatOutput = false;
      
      String mode = null;
      String inFileName = null;
      String outFileName = null;
      String xslFileName = null;
      String treedumpFileName = null;
      boolean didSetCR = false;
      boolean didSetLF = false;
      boolean stripCData = false;
      boolean escapeCData = false;
      
      int outputType = -1; // else 1=XML, 2=TEXT, or 3=HTML

      for (int i = 0;  i < argv.length;  i ++) 
      {
        if("-TT".equalsIgnoreCase(argv[i]))
        {
          processor.setTraceTemplates(true);
        }
        else if("-TS".equalsIgnoreCase(argv[i]))
        {
          processor.setTraceSelect(true);
        }
        else if("-TTC".equalsIgnoreCase(argv[i]))
        {
          processor.setTraceTemplateChildren(true);
        }
        else if ("-escape".equalsIgnoreCase(argv[i])) 
        {
          xmlProcessorLiaison.setSpecialCharacters(argv[++i]);
        }
        else if ("-indent".equalsIgnoreCase(argv[i])) 
        {
          xmlProcessorLiaison.setIndent(Integer.parseInt( argv[++i] ));
        }
        else if ("-mode".equalsIgnoreCase(argv[i])) 
        {
          mode = argv[++i];
        }
        else if ("-in".equalsIgnoreCase(argv[i])) 
        {
          inFileName = argv[++i];
        }
        else if ("-out".equalsIgnoreCase(argv[i])) 
        {
          outFileName = argv[++i];
        }
        else if ("-xsl".equalsIgnoreCase(argv[i])) 
        {
          xslFileName = argv[++i];
        }
        else if ("-PARAM".equalsIgnoreCase(argv[i])) 
        {
          String name = argv[++i];
          String expression = argv[++i];
          processor.setStylesheetParam(name, expression);
        }
        else if ("-treedump".equalsIgnoreCase(argv[i])) 
        {
          treedumpFileName = argv[++i];
        }
        else if("-S".equalsIgnoreCase(argv[i]))
        {
          processor.setStripWhiteSpace(true);
        }
        else if("-F".equalsIgnoreCase(argv[i]))
        {
          formatOutput = true;
        }
        else if("-E".equalsIgnoreCase(argv[i]))
        {
          xmlProcessorLiaison.SetShouldExpandEntityRefs(false);
        }
        else if("-V".equalsIgnoreCase(argv[i]))
        {
          diagnosticsWriter.println(">>>>>>> LotusXSL Version "
                                    +XSLProcessorVersion.S_VERSION+", "+
                                    xmlProcessorLiaison.getParserDescription()+
                                    " <<<<<<<");
        }
        else if("-QC".equalsIgnoreCase(argv[i]))
        {
          processor.setQuietConflictWarnings(true);
        }
        else if("-Q".equalsIgnoreCase(argv[i]))
        {
          setQuietMode = true;
        }
        else if("-LF".equalsIgnoreCase(argv[i]))
        {
          if(!didSetCR)
            processor.setOutputCarriageReturns(false);
          processor.setOutputLinefeeds(true);
          didSetLF = true;
        }
        else if("-CR".equalsIgnoreCase(argv[i]))
        {
          if(!didSetLF)
            processor.setOutputLinefeeds(false);
          processor.setOutputCarriageReturns(true);
          didSetCR = true;
        }
        else if("-SSATV".equalsIgnoreCase(argv[i]))
        {
          processor.m_useATVsInSelects = true;
        }
        else if("-VALIDATE".equalsIgnoreCase(argv[i]))
        {
          xmlProcessorLiaison.setUseValidation(true);
        }
        else if("-PARSER".equalsIgnoreCase(argv[i]))
        {
          i++;
          // Handled above
        }
        else if("-XML".equalsIgnoreCase(argv[i]))
        {
          outputType = 1;
        }
        else if("-TEXT".equalsIgnoreCase(argv[i]))
        {
          outputType = 2;
        }
        else if("-HTML".equalsIgnoreCase(argv[i]))
        {
          outputType = 3;
        }
        else if("-STRIPCDATA".equalsIgnoreCase(argv[i]))
        {
          stripCData = true;
        }
        else if("-ESCAPECDATA".equalsIgnoreCase(argv[i]))
        {
          escapeCData = true;
        }
        else if("-EDUMP".equalsIgnoreCase(argv[i]))
        {
          doStackDumpOnError = true;
        }
      }
      
      if(null == inFileName)
      {
        diagnosticsWriter.println("-in argument is required");
      }
      else
      {
        // The main XSL transformation occurs here!
        try
        {
          processor.setDiagnosticsOutput( setQuietMode ? null : diagnosticsWriter );
          processor.pushTime(processor);
          PrintWriter resultWriter;
          
          if(null != outFileName)
          {
            
            resultWriter = new PrintWriter( new BufferedWriter(new FileWriter(outFileName)) );
          }
          else
          {
            resultWriter = new PrintWriter( System.out );
          }
          
          if(1 == outputType)
          {
            FormatterToXML fToXML = new FormatterToXML(resultWriter, null, 
                                                       null, true, 
                                                       xmlProcessorLiaison.getIndent());
            fToXML.m_shouldWriteXMLHeader = true;
            fToXML.m_attrSpecialChars = xmlProcessorLiaison.getSpecialCharacters();
            fToXML.m_stripCData = stripCData;
            fToXML.m_escapeCData = escapeCData;
            processor.m_formatter.setFormatterListener(fToXML);
          }
          else if(2 == outputType)
          {
            FormatterToText fToText = new FormatterToText(resultWriter);
            processor.m_formatter.setFormatterListener(fToText);
          }
          else if(3 == outputType)
          {
            FormatterToHTML fToHTML
              = new FormatterToHTML(resultWriter, null, 
                                    "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n",
                                    true, xmlProcessorLiaison.getIndent());
            fToHTML.m_attrSpecialChars = xmlProcessorLiaison.getSpecialCharacters();
            fToHTML.m_stripCData = stripCData;
            processor.m_formatter.setFormatterListener(fToHTML);
          }
          
          // Runtime.getRuntime().traceMethodCalls(true);  
          // Runtime.getRuntime().traceInstructions(false);
          processor.process(inFileName, xslFileName, 
                            (null != outFileName) ? outFileName : "System.out", 
                            resultWriter);
          // Runtime.getRuntime().traceMethodCalls(false);  
          // Runtime.getRuntime().traceInstructions(false);
          
          // Document xmldoc = processor.parseXML(inFileName, 1);
          // Document stylesheet = processor.parseXML(xslFileName, 1);
          // TXDocument out = (TXDocument)processor.process(xmldoc, stylesheet, "");
          // out.print(new PrintWriter(new FileWriter(outFileName)));
        }
        catch(QueryEngineException qe)
        {
          if(doStackDumpOnError)
            qe.printStackTrace(diagnosticsWriter);
          else
            System.out.println("Error! "+qe.getMessage());
          // User will have already been informed of error.
          diagnosticsWriter.println("XSL Process was not successful.");
        }
        catch(XSLProcessorException pe)
        {
          if(doStackDumpOnError)
            pe.printStackTrace(diagnosticsWriter);
          else
            System.out.println("Error! "+pe.getMessage());
          // User will have already been informed of error.
          diagnosticsWriter.println("XSL Process was not successful.");
        }
        catch(Exception e)
        {
          if(doStackDumpOnError)
            e.printStackTrace(diagnosticsWriter);
          else
            System.out.println("Error! "+e.getMessage());
          diagnosticsWriter.println("XSL Process was not successful.");
        }
        catch(java.lang.NoClassDefFoundError ncdfe)
        {
          String msg = ncdfe.getMessage();

          if(msg.indexOf("com/ibm/xml/parser/Parser") >= 0)
          {
            diagnosticsWriter.println("==========================================");
            diagnosticsWriter.println("XSL Process was not successful.");
            diagnosticsWriter.println("** Could not find parser **");
            diagnosticsWriter.println("Please check your classpath.");
            diagnosticsWriter.println("If you don't have IBM's XML Parser for Java, you can download it from");
            diagnosticsWriter.println("IBM's AlphaWorks: http://www.alphaworks.ibm.com/formula/xml");
            diagnosticsWriter.println("==========================================");
          }
          else
          {
            if(doStackDumpOnError)
              ncdfe.printStackTrace(diagnosticsWriter);
            else
              System.out.println("Error! "+ncdfe.getMessage());
            diagnosticsWriter.println("XSL Process was not successful.");
          }
        }
      }
      
      if(null != diagnosticsWriter)
      {
        diagnosticsWriter.close();
      }
      
      diagnosticsWriter.println("XSLProcessor: done");
    }
  }
}
