/**
 * $Id: QueryEngine.java,v 1.5 1998/12/21 05:55:22 sboag Exp $
 * 
 * Copyright (c) 1998-1999 Lotus Corporation, Inc. All Rights Reserved.
 *				This software is provided without a warranty of any kind.
 * 
 * $State: Exp $
 * 
 * @author Scott Boag (scott_boag@lotus.com)
 * @author Dean Burson
 */
package com.lotus.xsl;

import java.util.*;
 
import org.w3c.dom.*;

/**
 * This class queries a DOM tree, and performs 
 * other services relating to XSL Expressions such as 
 * pattern matching.  The engine primarily does a a level-order 
 * traversal of the node tree   from the given context (as 
 * opposed to a preorder traversal... the position(), 
 * count(), and last() functions along with multiple 
 * predicates make this hard),  
 * based upon the given XSL pattern, collecting the nodes 
 * into a vector.  It uses the token structure in raw 
 * form, sliding up and down the token structure as it 
 * traverses and recursis into the nodes.
 * 
 * The design hypothesis used is that the optimization 
 * gained by pre-tokenizing into a bunch of intelligent 
 * objects, as can be seen in some other XSL designs 
 * (including some previous incarnations of this engine), 
 * can actually be counter-productive in web-sized 
 * documents. In other words, I -tried- to keep it very 
 * simple and not allocate a bunch of objects along the way.  
 * From the experience so far, I'm not happy with the 
 * performance, so maybe my hypothesis is somewhat flawed.
 * It looks like I will probably need to create some sort 
 * of smart token queue somewhat akin to compiled code, 
 * which I can then return to the caller.
 */
public class QueryEngine
{
  
  // ================= CONSTRUCTORS =====================
  
  /**
   * Construct a pattern analyzer.
   */
  QueryEngine(XSLProcessor processor)
  {
    m_xslProcessor = processor;
  }

  // ============= PUBLIC API FUNCTIONS =================
  
  /**
   * Given an string, make an Expression object.
   */
  Expression makeExpression(String expression, Element namespaceContext)
  {
    m_namespaceContext = namespaceContext;
    
    tokenize( expression );
    
    return new Expression(expression, m_tokenQueue, m_tokenQueueSize,
                          m_patternMap, m_patternMapSize);
  }
  
  /**
   * Given a select pattern, query a DOM for elements 
   * and/or attributes. This works exactly the opposite 
   * of matchAgainstPattern, which starts from the 
   * target node and marches inward and outward, qualifying
   * the element matches. Instead, this function starts 
   * from the anchor and marches left to right, traversing
   * the tree until it finds the target node (which it then 
   * qualifies the same as matchAgainstPattern).
   * @param expression The select expression.
   * @param contextNode The node that "." expresses.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public NodeList query(Expression expression, Node contextNode, Element namespaceContext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("^^^^^^^^^^^^ query("+expression+", "
                      +getAncestorString(contextNode, "")+")");
    m_namespaceContext = namespaceContext;
    
    restoreTokenQueue(expression);

    return query(contextNode, namespaceContext);
  }

  
  /**
   * Given a select pattern, query a DOM for elements 
   * and/or attributes. This works exactly the opposite 
   * of matchAgainstPattern, which starts from the 
   * target node and marches inward and outward, qualifying
   * the element matches. Instead, this function starts 
   * from the anchor and marches left to right, traversing
   * the tree until it finds the target node (which it then 
   * qualifies the same as matchAgainstPattern).
   * @param expression The select expression.
   * @param contextNode The node that "." expresses.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public NodeList query(String expression, Node contextNode, Element namespaceContext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("^^^^^^^^^^^^ query("+expression+", "
                      +getAncestorString(contextNode, "")+")");
    m_namespaceContext = namespaceContext;
    m_throwFoundIndex = false;
    
    tokenize( expression );
        
    return query(contextNode, namespaceContext);
  }
  
  /**
   * Given a select pattern, query a DOM for elements 
   * and/or attributes. This works exactly the opposite 
   * of matchAgainstPattern, which starts from the 
   * target node and marches inward and outward, qualifying
   * the element matches. Instead, this function starts 
   * from the anchor and marches left to right, traversing
   * the tree until it finds the target node (which it then 
   * qualifies the same as matchAgainstPattern).
   * @param expression The select expression.
   * @param contextNode The node that "." expresses.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public NodeList query(Node contextNode, Element namespaceContext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {    
    if(m_trace) trace("^^^^^^^^^^^^ query("+getAncestorString(contextNode, "")+")");
    
    Vector queryResults;

    nextToken();
    
    if((m_tokenQueueSize == 1) 
       && (Character.isLetter(m_tokenChar) || ('*' == m_tokenChar)
           || ('.' == m_tokenChar)))
    {
      queryResults = new NodeListImpl();
      if('.' == m_tokenChar)
      {
        if(m_token.equals("."))
        {
          queryResults.addElement(contextNode);
        }
        else
        {
          queryResults.addElement(m_xslProcessor.m_parserLiaison.getParentOfNode(contextNode));
        }
      }
      else
      {
        NodeList nl = contextNode.getChildNodes();
        if(null != nl)
        {
          int nNodes = nl.getLength();
          for(int i = 0; i < nNodes; i++)
          {
            Node child = nl.item(i);
            if(Node.ELEMENT_NODE == child.getNodeType())
            {
              if('*' == m_tokenChar)
              {
                queryResults.addElement(child);
              }
              else 
              {
                String contextNS = m_xslProcessor.m_parserLiaison.getNamespaceOfNode(child);
                if((null == contextNS) && child.getNodeName().equals(m_token))
                {
                  queryResults.addElement(child);
                }
              }
            }
          }
        }
      }
    }
    else
    {
      m_testForDocOrder = false;
      m_collectTargetStrings = false;
      
      queryResults = NodeSetExpr(contextNode);
      
      // HACK: Fake error detection.
      // Figuring out if we have an error is a bit hard because 
      // the parse can terminate just about anywhere when a non-match 
      // is detected, so patterns
      // like "foo/baz@attr" are hard to detect.  This will 
      // still let you get away with quite a bit, like "foo/baz=".
      // So this is pretty much a total hack. Need to think about 
      // a comprehensive way to deal with this.
      
      if(null != m_token)
      {
        switch(m_tokenChar)
        {
        case '(':
        case '@':  
        case ':': 
        case '$':
        case '\'': 
        case '\"': 
        case '\\': 
        case '<': 
        case '>': 
          error(contextNode, "Expected end of pattern but found '"+m_token+"'");
        }
      }
    }
    
    return (NodeList)queryResults;
  }

  /**
   * Returns a vector of target element strings.
   *  If the input is "foo", return {"foo"}.
   *  If the input is "foo | doo", return {"foo", "doo"}.
   *  If the input is "x/foo", return {"foo"}.
   *  If the input is "x/foo | z/foo", return {"foo"} -- NOTE the two element names are combined!
   *  If the input is "id(xx)/x//z[@aName]//foo[zz]", return {"foo"}.
   *  If the input is "id(xx)/x//z[@aName]//foo[zz] | roo/foo", return {"foo"}.
   *  If the input is "id(xx)/x//z[@aName]//foo[zz] | roo/doo", return {"foo", "doo"}.
   *  If the input is "id(xx)/x//z[@aName]//*[zz]", return {"*"}.
   * @param   pattern  Valid XSL pattern expression.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Vector getTargetElementStrings(String expression, Element namespaceContext,
                                        Expression[] expressionHolder)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    m_namespaceContext = namespaceContext;
    
    Vector results = new Vector();
    tokenize( expression, results );
    
    expressionHolder[0] 
      = new Expression(expression, m_tokenQueue,  m_tokenQueueSize, m_patternMap, m_patternMapSize);
    
    if(m_trace)
    {
      trace("Found "+results.size()+" target strings");
      for(int i = 0; i < results.size(); i++)
      {
        trace("Target string: "+(String)results.elementAt(i));
      }
    }
    
    // results = UnionExpr(null, results, true);
    return results;
  }

  /**
   * Returns a vector of target element strings.
   *  If the input is "foo", return {"foo"}.
   *  If the input is "foo | doo", return {"foo", "doo"}.
   *  If the input is "x/foo", return {"foo"}.
   *  If the input is "x/foo | z/foo", return {"foo"} -- NOTE the two element names are combined!
   *  If the input is "id(xx)/x//z[@aName]//foo[zz]", return {"foo"}.
   *  If the input is "id(xx)/x//z[@aName]//foo[zz] | roo/foo", return {"foo"}.
   *  If the input is "id(xx)/x//z[@aName]//foo[zz] | roo/doo", return {"foo", "doo"}.
   *  If the input is "id(xx)/x//z[@aName]//*[zz]", return {"*"}.
   * @param   pattern  Valid XSL pattern expression.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Vector getTargetElementStrings(String expression, Element namespaceContext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    m_namespaceContext = namespaceContext;
    
    Vector results = new Vector();
    tokenize( expression, results );
    
    if(m_trace)
    {
      trace("Found "+results.size()+" target strings");
      for(int i = 0; i < results.size(); i++)
      {
        trace("Target string: "+(String)results.elementAt(i));
      }
    }
    
    // results = UnionExpr(null, results, true);
    return results;
  }
  
  /**
   * Match a node against a list of ancestryPatterns.
   * @param expression The select expression.
   * @param node The node that is being tested.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public double matchAgainstPatterns(Expression expression, Node node, Element namespaceContext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("+++++ matchAgainstPatterns("+expression+", "+node.getNodeName()+")");
    m_namespaceContext = namespaceContext;
    
    restoreTokenQueue(expression);
    
    m_testForDocOrder = false;
    m_collectTargetStrings = false;
    
    nextToken();
    double score = Pattern(node);	
    if(m_trace) 
    {
      String ancestors = getAncestorString(node, "");
      if(m_trace)
        trace("----- "+ancestors+ " against '"+expression+"', score = "+score);
    }
    return score;
  }
  
  /**
   * Match a node against a list of ancestryPatterns.
   * @param expression The select expression.
   * @param node The node that is being tested.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public double matchAgainstPatterns(String expression, Node node, Element namespaceContext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("+++++ matchAgainstPatterns("+expression+", "+node.getNodeName()+")");
    m_namespaceContext = namespaceContext;
    
    tokenize( expression );
    
    m_testForDocOrder = false;
    m_collectTargetStrings = false;
    m_throwFoundIndex = false;
    
    nextToken();
    double score = Pattern(node);	
    if(m_trace) 
    {
      String ancestors = getAncestorString(node, "");
      if(m_trace)
        trace("----- "+ancestors+ " against '"+expression+"', score = "+score);
    }
    return score;
  }
  
  /**
   * For diagnostic purposes.
   */
  private String getAncestorString(Node node, String s)
  {
    Node parent = m_xslProcessor.m_parserLiaison.getParentOfNode(node);
    if(null != parent)
    {
      s = getAncestorString(parent, s);
    }
    s+="/"+node.getNodeName();
    return s;
  }
  
  /**
   * Given a match pattern and a context, find the first ancestor 
   * that matches the pattern (including the context handed in).
   * @param matchPatternString The match pattern.
   * @param node The node that "." expresses.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Node findAncestor(String matchPatternString, Node context, Element namespaceContext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    return findAncestor(null, matchPatternString, context, namespaceContext);
  }
  
  /**
   * Given a 'from' pattern (ala xsl:number), a match pattern 
   * and a context, find the first ancestor that matches the 
   * pattern (including the context handed in).
   * @param matchPatternString The match pattern.
   * @param node The node that "." expresses.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   */
  Node findAncestor(String fromMatchPattern, String countMatchPattern,
                    Node context, 
                    Element namespaceContext)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {      
    m_namespaceContext = namespaceContext;

    if(null != countMatchPattern)
    {
      tokenize( countMatchPattern );
    }
    
    QueryEngine q2 = null;
    if(null != fromMatchPattern)
    {
      q2 = new QueryEngine(m_xslProcessor);
      q2.tokenize(fromMatchPattern);
      q2.m_namespaceContext = namespaceContext;
    }
    
    while(null != context)
    {
      if(null != fromMatchPattern)
      {
        q2.m_testForDocOrder = false;
        q2.m_collectTargetStrings = false;
        q2.resetTokenMark(0);
        if(q2.Pattern(context) != MATCH_SCORE_NONE)
        {
          context = null;
          break;
        }
      }
      
      if(null != countMatchPattern)
      {
        m_testForDocOrder = false;
        m_collectTargetStrings = false;
        resetTokenMark(0);
        if(Pattern(context) != MATCH_SCORE_NONE)
        {
          break;
        }
      }
      
      context = m_xslProcessor.m_parserLiaison.getParentOfNode(context);
    }
    return context;
  }

  
  /**
   * Given an expression and a context, return the results 
   * in a string.
   * @param expression The string expression.
   * @param node The node that "." expresses.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public String getStringExpr(String expression, Node contextNode, Element namespaceContext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    m_namespaceContext = namespaceContext;
    
    tokenize( expression );
    
    m_testForDocOrder = false;
    m_collectTargetStrings = false;
    m_throwFoundIndex = false;
    
    nextToken();
    String resultString = (String)Expr(contextNode, CLASS_STRING);		
    return resultString;
  }
  
  /**
   * Given an expression and a context, return the result, without 
   * casting to any specific type.
   * @param expression The expression.
   * @param node The node that "." expresses.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Object getExpr(String expression, Node contextNode, 
                        Element namespaceContext, NodeList contextNodeList)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    m_namespaceContext = namespaceContext;
    m_contextNodeList = contextNodeList;
    
    tokenize( expression );
    
    m_testForDocOrder = false;
    m_collectTargetStrings = false;
    m_throwFoundIndex = false;
    
    nextToken();
    return Expr(contextNode, CLASS_DONTCARE);
  }
  
  /**
   * Given an expression and a context, return the result, without 
   * casting to any specific type.
   * @param expression The expression.
   * @param node The node that "." expresses.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Object getExpr(Expression expression, Node contextNode, 
                        Element namespaceContext, NodeList contextNodeList)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    Object result;
    m_namespaceContext = namespaceContext;
    m_contextNodeList = contextNodeList;
    
    restoreTokenQueue(expression);
    nextToken();

    if((m_tokenQueueSize == 1) 
       && (Character.isLetter(m_tokenChar) || ('*' == m_tokenChar)
            || ('.' == m_tokenChar)))
    {
      NodeListImpl queryResults = new NodeListImpl();
      result = queryResults;
      if('.' == m_tokenChar)
      {
        if(m_token.equals("."))
        {
          queryResults.addElement(contextNode);
        }
        else
        {
          queryResults.addElement(m_xslProcessor.m_parserLiaison.getParentOfNode(contextNode));
        }
      }
      else
      {
        NodeList nl = contextNode.getChildNodes();
        if(null != nl)
        {
          int nNodes = nl.getLength();
          for(int i = 0; i < nNodes; i++)
          {
            Node child = nl.item(i);
            if(Node.ELEMENT_NODE == child.getNodeType())
            {
              if('*' == m_tokenChar)
              {
                queryResults.addElement(child);
              }
              else 
              {
                String contextNS = m_xslProcessor.m_parserLiaison.getNamespaceOfNode(child);
                if((null == contextNS) && child.getNodeName().equals(m_token))
                {
                  queryResults.addElement(child);
                }
              }
            }
          }
        }
      }
    }
    else
    {

      m_testForDocOrder = false;
      m_collectTargetStrings = false;
      m_throwFoundIndex = false;
      
      result = Expr(contextNode, CLASS_DONTCARE);
    }
    return result;
  }


   // ============= TOKENIZATION FUNCTIONS =================
  private void ____TOKENIZATION_FUNCTIONS____(){}

  void restoreTokenQueue(Expression exp)
  {
    m_throwFoundIndex = false;
    m_tokenQueue = exp.m_tokenQueue;
    m_tokenQueueSize = exp.m_tokenQueueSize;
    m_patternMap = exp.m_patternMap;
    m_patternMapSize = exp.m_patternMapSize;
    m_currentPattern = exp.m_currentPattern;
    m_queueMark = 0;
  }
  
  /**
   * Walk through the expression and build a token queue, and a map of the top-level 
   * elements.
   * @param pat XSLT Expression.
   */
  private void tokenize(String pat)
  {
    tokenize(pat, null);
  }
  
  /**
   * Walk through the expression and build a token queue, and a map of the top-level 
   * elements.
   * @param pat XSLT Expression.
   * @param targetStrings Vector to hold Strings, may be null.
   */
  private void tokenize(String pat, Vector targetStrings)
  {
    m_tokenQueue = new String[MAXTOKENQUEUESIZE];
    m_tokenQueueSize = 0;
    m_patternMap = new int[100];
    m_currentPattern = pat;
    m_patternMapSize = 0;
    int nChars = pat.length();
    int startSubstring = -1;
    int posOfNSSep = -1;
    boolean isStartOfPat = true;
    boolean isAttrName = false;
    boolean isNum = false;
    
    // Nesting of '[' so we can know if the given element should be 
    // counted inside the m_patternMap.
    int nesting = 0;
    
    // char[] chars = pat.toCharArray();
    for(int i = 0; i < nChars; i++)
    {
      char c = pat.charAt(i);
      switch(c)
      {
      case '\"': 
        {
          if(startSubstring != -1)
          {
            isNum = false;
            isStartOfPat = mapPatternElemPos(nesting, isStartOfPat, isAttrName);
            isAttrName = false;
            if(-1 != posOfNSSep)
            {      
              posOfNSSep = mapNSTokens(pat, startSubstring, posOfNSSep, i);
            }
            else
            {
              addToTokenQueue(pat.substring(startSubstring, i));
            }
          }
          startSubstring = i;
          for(i++; (i < nChars) && ((c = pat.charAt(i)) != '\"'); i++);
          if(c == '\"')
          {
            addToTokenQueue(pat.substring(startSubstring, i+1));
            startSubstring = -1;
          }
          else
          {
            error("misquoted literal... expected double quote!");
          }
        }
        break;
        
      case '\'':
        if(startSubstring != -1)
        {
          isNum = false;
          isStartOfPat = mapPatternElemPos(nesting, isStartOfPat, isAttrName);
          isAttrName = false;
          if(-1 != posOfNSSep)
          {      
            posOfNSSep = mapNSTokens(pat, startSubstring, posOfNSSep, i);
          }
          else
          {
            addToTokenQueue(pat.substring(startSubstring, i));
          }
        }
        startSubstring = i;
        for(i++; (i < nChars) && ((c = pat.charAt(i)) != '\''); i++);
        if(c == '\'')
        {
          addToTokenQueue(pat.substring(startSubstring, i+1));
          startSubstring = -1;
        }
        else
        {
          error("misquoted literal... expected single quote!");
        }
        break;
        
      case 0x0A:
      case 0x0D:
      case ' ':
      case '\t':
        if(startSubstring != -1)
        {
          isNum = false;
          isStartOfPat = mapPatternElemPos(nesting, isStartOfPat, isAttrName);
          isAttrName = false;
          if(-1 != posOfNSSep)
          {      
            posOfNSSep = mapNSTokens(pat, startSubstring, posOfNSSep, i);
          }
          else
          {
            addToTokenQueue(pat.substring(startSubstring, i));
          }
          startSubstring = -1;
        }
        break;
        
      case '@':
        isAttrName = true;
        // fall-through on purpose
        
      case '-':
        if('-' == c)
        {
          if(!(isNum || (startSubstring == -1)))
          {
            break;
          }
          isNum = false;
        }
        // fall-through on purpose

      case '(':
      case '[':
      case ')':
      case ']':
      case '|':
      case '/':
      case '*':
      case '+':
      case '=':
      case ',':
      case '\\': // Unused at the moment
      case '^': // Unused at the moment
      case '!': // Unused at the moment
      case '$':
      case '<':
      case '>':
        if(startSubstring != -1)
        {
          isNum = false;
          isStartOfPat = mapPatternElemPos(nesting, isStartOfPat, isAttrName);
          isAttrName = false;
          if(-1 != posOfNSSep)
          {      
            posOfNSSep = mapNSTokens(pat, startSubstring, posOfNSSep, i);
          }
          else
          {
            addToTokenQueue(pat.substring(startSubstring, i));
          }
          startSubstring = -1;
        }
        else if(('/' == c) && isStartOfPat)
        {
          isStartOfPat = mapPatternElemPos(nesting, isStartOfPat, isAttrName);
        }
        else if('*' == c)
        {
          isStartOfPat = mapPatternElemPos(nesting, isStartOfPat, isAttrName);
          isAttrName = false;
        }
        
        if(0 == nesting)
        {
          if('|' == c)
          {
            if(null != targetStrings)
            {
              recordTokenString(targetStrings);
            }
            isStartOfPat = true;
          }
        }
        if((')' == c) || (']' == c))
        {
          nesting--;
        }
        else if(('(' == c) || ('[' == c))
        {
          nesting++;
        }
        addToTokenQueue(pat.substring(i, i+1));
        
       break;
                
      case ':':
        posOfNSSep = i;
        // fall through on purpose
        
      default:
        if(-1 == startSubstring)
        {
          startSubstring = i;
          isNum = Character.isDigit(c);
        }
        else if(isNum)
        {
          isNum = Character.isDigit(c);
        }
      }
    }
    if(startSubstring != -1)
    {
      isNum = false;
      isStartOfPat = mapPatternElemPos(nesting, isStartOfPat, isAttrName);
      if(-1 != posOfNSSep)
      {      
        posOfNSSep = mapNSTokens(pat, startSubstring, posOfNSSep, nChars);
      }
      else
      {
        addToTokenQueue(pat.substring(startSubstring, nChars));
      }
    }
    
    if(0 == m_tokenQueueSize)
    {
      error("Empty expression!");
    }
    else if(null != targetStrings)
    {
      recordTokenString(targetStrings);
    }
    m_queueMark = 0;
  }

  
  /**
   * Record the current position on the token queue as long as 
   * this is a top-level element.  Must be called before the 
   * next token is added to the m_tokenQueue.
   */
  private boolean mapPatternElemPos(int nesting, boolean isStart, boolean isAttrName)
  {
    if(0 == nesting)
    {
      if(!isStart)
      {
        m_patternMap[m_patternMapSize-1] -= TARGETEXTRA;
      }
      m_patternMap[m_patternMapSize] 
        = (m_tokenQueueSize - (isAttrName ? 1 : 0)) + TARGETEXTRA;
      m_patternMapSize++;
      isStart = false;
    }
    return isStart;
  }
  
  /**
   * Record the correct token string in the passed vector.
   */
  private void recordTokenString(Vector targetStrings)
  {
    int tokPos = getTokenQueuePosFromMap(m_patternMapSize-1);
    resetTokenMark(tokPos+1);

    if(lookahead('(', 1))
    {
      int tok = getKeywordToken(m_token);
      switch(tok)
      {
      case NODETYPE_COMMENT:
        targetStrings.addElement(Constants.PSEUDONAME_COMMENT);
        break;
      case NODETYPE_TEXT:
        targetStrings.addElement(Constants.PSEUDONAME_TEXT);
        break;
      case NODETYPE_NODE:
        targetStrings.addElement(Constants.PSEUDONAME_ANY);
        break;
      case NODETYPE_ROOT:
        targetStrings.addElement(Constants.PSEUDONAME_ROOT);
        break;
      case NODETYPE_ANYELEMENT:
        targetStrings.addElement(Constants.PSEUDONAME_ANY);
        break;
      case NODETYPE_PI:
        targetStrings.addElement(Constants.PSEUDONAME_ANY);
        break;
      default:
        targetStrings.addElement(Constants.PSEUDONAME_ANY);
      }
    }
    else
    {
      if(tokenIs('@'))
      {
        tokPos++;
        resetTokenMark(tokPos+1);
      }
      if(lookahead(':', 1))
      {
        tokPos += 2;
      }
      targetStrings.addElement(m_tokenQueue[tokPos]);
    }
  }

  // ============= TOKEN ACCESS & MANIPULATION FUNCTIONS =================
  private void ____TOKEN_ACCESS_AND_MANIPULATION____(){}

  /**
   * When a seperator token is found, see if there's a element name or 
   * the like to map.
   */
  private int mapNSTokens(String pat, int startSubstring, int posOfNSSep, int posOfScan)
  {
    String prefix = pat.substring(startSubstring, posOfNSSep);
    String uName = m_xslProcessor.m_parserLiaison.getNamespaceForPrefix(prefix, m_namespaceContext);
    if((null != uName) && (uName.length() > 0))
    {
      addToTokenQueue(uName);
      addToTokenQueue(":");
      String s = pat.substring(posOfNSSep+1, posOfScan);
      if(s.length() > 0)
        addToTokenQueue(s);
    }
    else
    {
      error("Could not locate namespace for prefix: "+prefix);
    }
    return -1;
  }

  /**
   * Given a map pos, return the corresponding token queue pos.
   */
  int getTokenQueuePosFromMap(int i)
  {
    int pos = m_patternMap[i];
    return (pos >= TARGETEXTRA) ? (pos - TARGETEXTRA) : pos;
  }
  
  /**
   * This will return the index above the passed index that 
   * is the target element, i.e. it holds a value >= TARGETEXTRA.
   * If there is no next target, it will return -1.
   * Pass -1 in to start testing from zero.
   */
  int getNextTargetIndexInMap(int i)
  {
    int next = -1;

    for(int k = i+1; k < m_patternMapSize; k++)
    {
      int pos = m_patternMap[k];
      if(pos >= TARGETEXTRA)
      {
        next = k;
        break;
      }
    }
    return next;
  }
  
  /**
   * This will return the normalized index into the pattern 
   * map above the passed index, or -1 if it is the last pattern.
   */
  int getNextIndexInMap(int i)
  {
    int next = (m_patternMap[i] >= TARGETEXTRA) 
               ? -1 : (((i+1) < m_patternMapSize) 
                       ? (i+1) : -1);    
    return next;
  }
  
  /**
   * This will return the index above the passed index that 
   * is the start if the next subpattern, or -1 if there is none.
   * If there is no next target, it will return -1.
   * Pass -1 in to start testing from zero.
   */
  int getNextSubpatternStartIndexInMap(int i)
  {
    int next = -1;

    if(-1 == i)
    {
      next = 0;
    }
    else
    {
      // loop to find the next element that is marked 
      // as a target element, then go one passed it to 
      // return the next subpattern start index in the map.
      for(int k = i; k < m_patternMapSize; k++)
      {
        int pos = m_patternMap[k];
        if(pos >= TARGETEXTRA)
        {
          if((k+1) < m_patternMapSize)
          {
            next = (k+1);
          }
          break;
        }
      }
    }
    return next;
  }

  /**
   * This will return the next index from the passed index,
   * or -1 if it the passed index is the last index of the 
   * subpattern.
   */
  int getNextPatternPos(int i)
  {
    int next = -1;
    int k = (i+1);
    int nElems = m_patternMapSize;

    if(k < nElems)
    {
      int prevPos = (i >= 0) ? m_patternMap[i] : 0;
      if(prevPos != TARGETEXTRA)
      {
        next = m_patternMap[k];
      }
    }
    return next;
  }


  /**
   * This will return the previous index from the passed index,
   * or -1 if it the passed index is the first index of the 
   * subpattern.
   */
  int getPrevMapIndex(int i)
  {
    int prev = -1;
    int k = (i-1);

    if(k >= 0)
    {
      int pos = m_patternMap[k];
      if(pos < TARGETEXTRA)
      {
        prev = k;
      }
    }
    return prev;
  }


  /**
   * Check if m_token==s. If m_token is null, this won't throw
   * an exception, instead it just returns false (or true
   * if s is also null).
   */
  private final boolean tokenIs( String s )
  {
    return ( m_token!=null ) ? (m_token.equals(s)) : ( s==null );
  }

  /**
   * Check if m_token==s. If m_token is null, this won't throw
   * an exception, instead it just returns false (or true
   * if s is also null).
   */
  private final boolean tokenIs( char c )
  {
    return ( m_token!=null ) ? (m_tokenChar == c) : false;
  }
  
  /**
   * Lookahead of the current token in order to 
   * make a branching decision.
   * @param s the string to compare it to.
   * @param n number of tokens to lookahead.  Must be 
   * greater than 1.
   */
  private final boolean lookahead( char c, int n )
  {
    int pos = (m_queueMark+n);
    return ( (pos <= m_tokenQueueSize) && (pos > 0) && (m_tokenQueueSize != 0)
            && (m_tokenQueue[pos-1].length() == 1)) ?
           (m_tokenQueue[pos-1].charAt(0) == c) :
           false;
  }

  /**
   * Lookbehind the first character of the current token in order to 
   * make a branching decision.
   * @param c the character to compare it to.
   * @param n number of tokens to lookbehind.  Must be 
   * greater than 1.  Note that the lookbehind terminates 
   * at either the beginning of the string or on a '|' 
   * character.  Because of this, this method should only
   * be used for pattern matching.
   */
  private final boolean lookbehind( char c, int n )
  {
    boolean isToken;
    int lookBehindPos = m_queueMark-(n+1);
    if( lookBehindPos >= 0 )
    {
      String lookbehind = m_tokenQueue[lookBehindPos];
      if(lookbehind.length() == 1)
      {
        char c0 = ( lookbehind == null ) ? '|' : lookbehind.charAt(0);
        isToken = ( c0 == '|' ) ? false : (c0 == c);
      }
      else
      {
        isToken = false;
      }
    }
    else
    {
      isToken = false;
    }
    return isToken;
  }

  /**
   * look behind the current token in order to 
   * see if there is a useable token.
   * @param n number of tokens to lookahead.  Must be 
   * greater than 1.  Note that the lookbehind terminates 
   * at either the beginning of the string or on a '|' 
   * character.  Because of this, this method should only
   * be used for pattern matching.
   * @return true if lookbehind has a token, false otherwise.
   */
  private final boolean lookbehindHasToken( int n )
  {
    boolean hasToken;
    if( (m_queueMark-n) > 0 )
    {
      String lookbehind = m_tokenQueue[m_queueMark - (n-1)];
      char c0 = ( lookbehind == null ) ? '|' : lookbehind.charAt(0);
      hasToken = ( c0 == '|' ) ? false : true;
    }
    else
    {
      hasToken = false;
    }
    return hasToken;
  }

  /**
   * Lookahead of the current token in order to 
   * make a branching decision.
   * @param s the string to compare it to.
   * @param n number of tokens to lookahead.  Must be 
   * greater than 1.
   */
  private final boolean lookahead( String s, int n )
  {
    boolean isToken;
    if( (m_queueMark+n) <= m_tokenQueueSize )
    {
      String lookahead = m_tokenQueue[m_queueMark + (n-1)];
      isToken = ( lookahead!=null ) ? lookahead.equals(s) : ( s==null );
    }
    else
    {
      isToken = (null == s);
    }
    return isToken;
  }
  
  /**
   * Retrieve the next token from the command and
   * store it in m_token string.
   */
  private final void nextToken()
  {
    if( m_queueMark < m_tokenQueueSize )
    {
      m_token = m_tokenQueue[m_queueMark++];
      m_tokenChar = m_token.charAt(0);
    }
    else
    {
      m_token = null;
      m_tokenChar = 0;
    }
  }

  /**
   * Retrieve the next token from the command and
   * store it in m_token string.
   */
  private final String getTokenRelative(int i)
  {
    String tok;
    int relative = m_queueMark+i;
    if( (relative > 0) && (relative < m_tokenQueueSize) )
    {
      tok = m_tokenQueue[relative];
    }
    else
    {
      tok = null;
    }
    return tok;
  }

  /**
   * Retrieve the next token from the command and
   * store it in m_token string.
   */
  private final void prevToken()
  {
    if( m_queueMark > 0 )
    {
      m_queueMark--;
      m_token = m_tokenQueue[m_queueMark];
      m_tokenChar = m_token.charAt(0);
    }
    else
    {
      m_token = null;
      m_tokenChar = 0;
    }
  }

  /**
   * Reset token queue mark and m_token to a 
   * given position.
   */
  private final void resetTokenMark(int mark)
  {
    int qsz = m_tokenQueueSize;
    m_queueMark = (mark > 0) ? ((mark <= qsz) ? mark -1 : mark) : 0;
    if( m_queueMark < qsz )
    {
      m_token = m_tokenQueue[m_queueMark++];
      m_tokenChar = m_token.charAt(0);
    }
    else
    {
      m_token = null;
      m_tokenChar = 0;
    }
  }
  
  /**
   * Consume an expected token, throwing an exception if it 
   * isn't there.
   */
  private final void consumeExpected(String expected)
  {
    if(tokenIs(expected))
    {
      nextToken();
    }
    else
    {
      error("Expected "+expected+", but found: "+m_token);
    }
  }

  /**
   * Consume an expected token, throwing an exception if it 
   * isn't there.
   */
  private final void consumeExpected(char expected)
  {
    if(tokenIs(expected))
    {
      nextToken();
    }
    else
    {
      error("Expected "+expected+", but found: "+m_token);
    }
  }
 
  // ============= EXPRESSIONS FUNCTIONS =================
  private void ____EXPRESSIONS____(){}

  /**
   * 
   * --------------------------------------------------------------------------------
   Expr  ::=  OrExpr   
   * --------------------------------------------------------------------------------
   */
  protected Object Expr(Node context, int requestType) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("Expr("+context+", "+requestType+")");
    Object result = OrExpr(context, requestType);
    return castToType(result, requestType);
  }

  /**
   * 
   * --------------------------------------------------------------------------------
   NodeSetExpr  ::=  Expr   
   * --------------------------------------------------------------------------------
   */
  protected NodeListImpl NodeSetExpr(Node context) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("NodeSetExpr("+context+")");
    NodeListImpl result = (NodeListImpl)Expr(context, CLASS_NODESET);
    
    return result;
  }

  /**
   * 
   * --------------------------------------------------------------------------------
   OrExpr  ::=  AndExpr   
   | OrExpr 'or' AndExpr   
   * --------------------------------------------------------------------------------
   */
  protected Object OrExpr(Node context, int requestType) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("OrExpr("+context+", "+requestType+")");
    Object result = AndExpr(context, requestType);
    if((null != m_token) && tokenIs("or"))
    {
      nextToken();
      if(booleanCast(result))
      {
        // just parse past
        OrExpr(context, requestType);
      }
      else
      {
        result = OrExpr(context, requestType);
      }
    }
    return result;
  }

  /**
   * 
   * --------------------------------------------------------------------------------
   AndExpr  ::=  EqualityExpr   
   | AndExpr 'and' EqualityExpr   
   * --------------------------------------------------------------------------------
   */
  protected Object AndExpr(Node context, int requestType) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("AndExpr("+context+", "+requestType+")");
    Object result = EqualityExpr(context, requestType);
    if((null != m_token) && tokenIs("and"))
    {
      nextToken();
      if((null != result) && booleanCast(result))
      {
        Object result2 = AndExpr(context, requestType);
        if((null != result2) && booleanCast(result2))
        {
          result = new Boolean(true);
        }
        else
        {
          result = new Boolean(false);
        }
      }
      else
      {
        AndExpr(context, requestType);
      }
    }
    return result;
  }
  
  /**
   * XXXX.
   * @returns an Object which is either a String, a Number, a Boolean, or a vector 
   * of nodes.
   * --------------------------------------------------------------------------------
   EqualityExpr  ::=  RelationalExpr   
   | EqualityExpr '=' RelationalExpr   
   * --------------------------------------------------------------------------------
   */
  protected Object EqualityExpr(Node context, int requestType) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) 
      trace("EqualityExpr("+context+", "+requestType+")");
    Object result = RelationalExpr(context, requestType);
    if((null != m_token) && tokenIs('='))
    {
      nextToken();
      Object result2 = EqualityExpr(context, requestType);
      int t1 = getClassType(result);
      int t2 = getClassType(result2);
      if(t1 != t2)
      {
        if(CLASS_NUMBER == t1)
        {
          result = castToType(result2, CLASS_NUMBER);
        }
        else if(CLASS_NUMBER == t2)
        {
          result = castToType(result, CLASS_NUMBER);
        }
        else if(CLASS_BOOLEAN == t1)
        {
          result2 = castToType(result2, CLASS_BOOLEAN);
        }
        else if(CLASS_BOOLEAN == t2)
        {
          result = castToType(result, CLASS_BOOLEAN);
        }
        else
        {
          result = castToType(result, CLASS_STRING);
          result2 = castToType(result2, CLASS_STRING);
        }
      }
      else if(CLASS_NODESET == t1)
      {
        result = castToType(result, CLASS_STRING);
        result2 = castToType(result2, CLASS_STRING);
      }
      // System.out.println(result2+" = "+result+" is "+result2.equals(result));
      result = new Boolean(result2.equals(result));
    }
    return result;
  }

  /**
   * XXXX.
   * @returns an Object which is either a String, a Number, a Boolean, or a vector 
   * of nodes.
   * --------------------------------------------------------------------------------
   RelationalExpr  ::=  AdditiveExpr   
   | RelationalExpr '<' AdditiveExpr   
   | RelationalExpr '>' AdditiveExpr   
   | RelationalExpr '<=' AdditiveExpr   
   | RelationalExpr '>=' AdditiveExpr   
   * --------------------------------------------------------------------------------
   */
  protected Object RelationalExpr(Node context, int requestType) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("RelationalExpr("+context+", "+requestType+")");
    Object result = AdditiveExpr(context, requestType);
    if(null != m_token)
    {
      if(tokenIs('<'))
      {
        nextToken();
        boolean isOrEqual = tokenIs('=');
        if(isOrEqual)
          nextToken();
        Object result2 = AdditiveExpr(context, requestType);
        // TODO: Handle types other than integer
        double num1 = numberCast(result);
        double num2 = numberCast(result2);
        result = isOrEqual ? new Boolean(num1 <= num2) : new Boolean(num1 < num2);
      }
      else if(tokenIs('>'))
      {
        nextToken();
        boolean isOrEqual = tokenIs('=');
        if(isOrEqual)
          nextToken();
        Object result2 = AdditiveExpr(context, requestType);
        // TODO: Handle types other than integer
        double num1 = numberCast(result);
        double num2 = numberCast(result2);
        result = isOrEqual ? new Boolean(num1 >= num2) : new Boolean(num1 > num2);
      }
    }
    return result;
  }

  /**
   * XXXX.
   * @returns an Object which is either a String, a Number, a Boolean, or a vector 
   * of nodes.
   * --------------------------------------------------------------------------------
   AdditiveExpr  ::=  MultiplicativeExpr   
   | AdditiveExpr '+' MultiplicativeExpr   
   | AdditiveExpr '-' MultiplicativeExpr   
   * --------------------------------------------------------------------------------
   */
  protected Object AdditiveExpr(Node context, int requestType) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    Object result = MultiplicativeExpr(context, requestType);
    if(null != m_token)
    {
      if(tokenIs('+'))
      {
        nextToken();
        Object result2 = AdditiveExpr(context, requestType);
        // TODO: Handle types other than integer
        double num1 = numberCast(result);
        double num2 = numberCast(result2);
        result = new Double(num1 + num2);
      }
      else if(tokenIs('-'))
      {
        nextToken();
        Object result2 = AdditiveExpr(context, requestType);
        // TODO: Handle types other than integer
        double num1 = numberCast(result);
        double num2 = numberCast(result2);
        result = new Double(num1 - num2);
      }
    }
    return result;
  }

  /**
   * XXXX.
   * @returns an Object which is either a String, a Number, a Boolean, or a vector 
   * of nodes.
   * --------------------------------------------------------------------------------
   MultiplicativeExpr  ::=  UnaryExpr   
   | MultiplicativeExpr MultiplyOperator UnaryExpr   
   | MultiplicativeExpr 'div' UnaryExpr   
   | MultiplicativeExpr 'mod' UnaryExpr   
   | MultiplicativeExpr 'quo' UnaryExpr   
   * --------------------------------------------------------------------------------
   */
  protected Object MultiplicativeExpr(Node context, int requestType) 
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    Object result = UnaryExpr(context, requestType);
    if(null != m_token)
    {
      if(tokenIs('*'))
      {
        nextToken();
        Object result2 = MultiplicativeExpr(context, requestType);

        double num1 = numberCast(result);
        double num2 = numberCast(result2);
        result = new Double(num1 * num2);
      }
      else if(tokenIs("div"))
      {
        nextToken();
        Object result2 = MultiplicativeExpr(context, requestType);

        double num1 = numberCast(result);
        double num2 = numberCast(result2);
        result = new Double(num1 / num2);
      }
      else if(tokenIs("mod"))
      {
        nextToken();
        Object result2 = MultiplicativeExpr(context, requestType);

        double num1 = numberCast(result);
        double num2 = numberCast(result2);
        result = new Double(num1 % num2);
      }
      else if(tokenIs("quo"))
      {
        nextToken();
        Object result2 = MultiplicativeExpr(context, requestType);

        double num1 = numberCast(result);
        double num2 = numberCast(result2);
        
        result = new Double((int)(num1 / num2));
      }
    }
    return result;
  }

  /**
   * XXXX.
   * @returns an Object which is either a String, a Number, a Boolean, or a vector 
   * of nodes.
   * --------------------------------------------------------------------------------
   UnaryExpr  ::=  UnionExpr   
   | '-' UnaryExpr   
   * --------------------------------------------------------------------------------
   */
  protected Object UnaryExpr(Node context, int requestType) 
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    boolean isNeg;
    if(m_tokenChar == '-')
    {
      nextToken();
      isNeg = true;
    }
    else
    {
      isNeg = false;
    }
    Object result = UnionExpr(context, requestType);
    if(isNeg)
    {
      result = new Double(-numberCast(result));
    }
    return result;
  }
  
  /**
   * 
   * --------------------------------------------------------------------------------
   StringExpr  ::=  Expr   
   * --------------------------------------------------------------------------------
   */
  protected String StringExpr(Node context) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("StringExpr("+context+")");
    return (String)Expr(context, CLASS_STRING);
  }

  /**
   * 
   * --------------------------------------------------------------------------------
   StringExpr  ::=  Expr   
   * --------------------------------------------------------------------------------
   */
  protected Boolean BooleanExpr(Node context) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("BooleanExpr("+context+")");
    return (Boolean)Expr(context, CLASS_BOOLEAN);
  }

  /**
   * 
   * --------------------------------------------------------------------------------
   NumberExpr  ::=  Expr   
   * --------------------------------------------------------------------------------
   */
  protected Double NumberExpr(Node context) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("NumberExpr("+context+")");
    return (Double)Expr(context, CLASS_NUMBER);
  }
  
  /**
   * The context of the right hand side expressions is the context of the 
   * left hand side expression. The results of the right hand side expressions 
   * are node sets. The result of the left hand side UnionExpr is the union 
   * of the results of the right hand side expressions.
   * 
   * --------------------------------------------------------------------------------
   UnionExpr  ::=  PathExpr   
   | UnionExpr '|' PathExpr   
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected Object UnionExpr(Node context, int requestType) 
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    if(m_trace) trace("UnionExpr("+context+", "+requestType+")");
    boolean continueOrLoop = true;
    
    Object result = null;
    NodeListImpl resultNodeSet = null;
    do 
    {
      result = PathExpr(context, requestType, result);
            
      if(null != resultNodeSet)
      {
        if(null != result)
        {
          NodeList nl = (NodeList)result;
          int nChildren = nl.getLength();
          for(int i = 0; i < nChildren; i++)
          {
            addNodeInDocOrder(resultNodeSet, nl.item(i), true);
          }
        }
        result = resultNodeSet;
      }
      
      if(tokenIs('|'))
      {
        if(null == resultNodeSet)
        {
          resultNodeSet = (NodeListImpl)result;
        }
        result = null;
        nextToken();
      }
      else
      {
        break;
      }

      // this.m_testForDocOrder = true;
    }
      while(continueOrLoop);
        
    return result;
  }

  /**
   * 
   * --------------------------------------------------------------------------------
   PathExpr  ::=  LocationPath   
   | FilterExpr   
   | FilterExpr '/' RelativeLocationPath   
   | FilterExpr '//' RelativeLocationPath   
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected Object PathExpr(Node context, int requestType, Object queryResults) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("PathExpr("+context+", "+queryResults+")");
    Object result = null;
        
    boolean foundLocationPath;

    // LocationPath called from PrimaryExpr.
    result = FilterExpr(context, requestType, queryResults);

    if(tokenIs('/'))
    {
      // I may be in big trouble here regarding the position() support...
      NodeList nodeListResult = (NodeList)result;
      NodeListImpl newResults = new NodeListImpl();
      int nNodesInResults = nodeListResult.getLength();
      if(0 == nNodesInResults)
      {
        continueLocationPath(null, null);
      }
      else
      {
        int mark = this.m_queueMark;
        for(int i = 0; i < nNodesInResults; i++)
        {
          resetTokenMark(mark);
          Node subContext = nodeListResult.item(i);
          newResults = continueLocationPath(subContext, newResults);
        }
      }
      result = newResults;
    }

    return result;
  }

  /**
   * 
   * --------------------------------------------------------------------------------
   FilterExpr  ::=  PrimaryExpr   
   | FilterExpr Predicate   
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected Object FilterExpr(Node context, int requestType, Object queryResults) 
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    if(m_trace) trace("FilterExpr("+context+", "+queryResults+")");
    Object results = PrimaryExpr(context, requestType, queryResults);
    
    if(tokenIs('['))
    {
      NodeListImpl nodeListResult = (NodeListImpl)results;
      if(null == queryResults) queryResults = new NodeListImpl();
      results = continueFromPredicate(nodeListResult, (NodeListImpl)queryResults,
                                     this.m_queueMark, false, false);
    }    
    
    return results;
  }
  
  /**
   * XXXX.
   * @returns an Object which is either a String, a Number, a Boolean, or a vector 
   * of nodes.
   * --------------------------------------------------------------------------------
   PrimaryExpr  ::=  VariableReference   
   | '(' Expr ')'   
   | Literal   
   | Number   
   | FunctionCall   
   * --------------------------------------------------------------------------------
   */
  protected Object PrimaryExpr(Node context, int requestType, Object queryResults) 
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    Object result = null;
    if((m_tokenChar == '\'') || (m_tokenChar == '"'))
    {
      result = Literal();
    }
    else if(m_tokenChar == '$')
    {
      nextToken(); // consume '$'
      
      String varName = NCName();
      try
      {
        result = m_xslProcessor.getVariable(varName);
      }
      catch(Exception e)
      {
        error("Could not get variable named "+varName);
      }

      if(null == result)
      {
        warn(context, "VariableReference given for variable out "+
                      "of context or without definition!  Name = " + varName);
      }
    }
    else if(m_tokenChar == '(')
    {
      nextToken();
      result = Expr(context, CLASS_DONTCARE);
      consumeExpected(')');
    }
    else if((null != m_token) && Character.isDigit( m_tokenChar ))
    {
      result = Number();
    }
    else if(lookahead('(', 1) || (lookahead(':', 1) && lookahead('(', 3)))
    {
      result = FunctionCall(context, requestType);
    }
    else
    {
      if(null == queryResults)
        queryResults = new NodeListImpl();
      result = LocationPath(context, (NodeListImpl)queryResults);
    }
    return result;
  }
  
  /**
   * HandleDocExpr
   */
  Document handleDocExpr(Node context, String uri)
  {
    Document newDoc = (Document)m_xslProcessor.getSourceDocsTable().get(uri);
    if(null == newDoc)
    {
      Document owner = (Node.DOCUMENT_NODE == context.getNodeType()) ? 
                       (Document)context : context.getOwnerDocument();
      
      String base = m_xslProcessor.findURIFromDoc(owner);
      try
      {
        newDoc = m_xslProcessor.parseXML(uri, null, base);
      }
      catch(Exception e)
      {
      }
      if(null == newDoc)
      {
        owner = (Node.DOCUMENT_NODE == m_namespaceContext.getNodeType()) ? 
                         (Document)m_namespaceContext : m_namespaceContext.getOwnerDocument();
        
        XSLProcessor.Stylesheet ss = (XSLProcessor.Stylesheet)m_xslProcessor.m_stylesheets.get(owner.getDocumentElement());
        if(null == ss)
          ss = m_xslProcessor.m_stylesheetRoot;
        base = (null == ss) ? null : ss.m_baseIdent;
        try
        {
          newDoc = m_xslProcessor.parseXML(uri, null, base);
        }
        catch(Exception e)
        {
        }
        if(null == newDoc)
        {
          warn("Can not load requested doc: "+uri);
        }
      }
      if(null == newDoc)
      {
        warn("Can not load requested doc: "+uri);
      }
      else
      {
        m_xslProcessor.getSourceDocsTable().put(uri, newDoc);
      }
    }
    return newDoc;
  }
  
  /**
   * XXXX.
   * @returns an Object which is either a String, a Number, a Boolean, or a vector 
   * of nodes.
   * --------------------------------------------------------------------------------
   [15]  FunctionCall  ::=  NodeSetFunctionCall   
   | BooleanFunctionCall   
   | StringFunctionCall   
   | NumberFunctionCall   
   | SystemFunctionCall   
   | ExtensionFunctionCall   

   [20]  NodeSetFunctionCall  ::=  'last' '(' ')'   
   | 'position' '(' ')'   
   | 'count' '(' NodeSetExpr ')'   
   | 'id' '(' StringExpr ')'   
   | 'idref' '(' NodeSetExpr ')'   
   | 'key' '(' StringExpr ',' StringExpr ')'   
   | 'keyref' '(' StringExpr ',' NodeSetExpr ')'   
   | 'doc' '(' StringExpr ')'   
   | 'docref' '(' NodeSetExpr ')'   
   | 'local-part' '(' NodeSetExpr? ')'   
   | 'namespace' '(' NodeSetExpr? ')'   
   | 'qname' '(' NodeSetExpr? ')'   
   | 'generate-id' '(' NodeSetExpr? ')'  

   [25]  BooleanFunctionCall  ::=  'not' '(' BooleanExpr ')'   
   | 'true' '(' ')'   
   | 'false' '(' ')'   
   | 'boolean' '(' Expr? ')'   
   | 'lang' '(' StringExpr ')'  

   [31]  NumberFunctionCall  ::=  'number '(' Expr? ')'   
   | 'floor' '(' NumberExpr ')'   
   | 'ceiling' '(' NumberExpr ')'   
   | 'round' '(' NumberExpr ')'   
   | 'sum' '(' NodeSetExpr ')'  

   [33]  StringFunction  ::=  'string' '(' Expr? ')'   
   | 'starts-with' '(' StringExpr ',' StringExpr ')'   
   | 'contains' '(' StringExpr ',' StringExpr ')'   
   | 'substring-before '(' StringExpr ',' StringExpr ')'   
   | 'substring-after '(' StringExpr ',' StringExpr ')'   
   | 'normalize' '(' StringExpr? ')'   
   | 'translate' '(' StringExpr ',' StringExpr ',' StringExpr ')'   
   | 'concat' '(' StringExpr (',' StringExpr)+ ')'   
   | 'format-number' '(' NumberExpr ',' StringExpr (',' StringExpr)? ')' 
   
   [34]  ExtensionFunctionCall  ::=  CName '(' ( Expr ( ',' Expr)*)? ')' 
   
   [35]  SystemFunctionCall  ::=  'system-property' '(' StringExpr ')'   
   | 'function-available' '(' StringExpr ')'   
   * --------------------------------------------------------------------------------
   */
  protected Object FunctionCall(Node context, int requestType) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    Object result = null;
    String funcName;
    int tok;
    boolean isNamespaced;
    if(lookahead(':',1))
    {
      tok = 0;
      funcName = getTokenRelative(2);
      isNamespaced = true;
    }
    else
    {
      isNamespaced = false;
      funcName = m_token;
      tok = getFunctionToken(funcName);
    }
    
    if(tok > 0)
    {
      nextToken(); nextToken();
      
      switch(tok)
      {
      case FUNC_COUNT:
        {
          NodeList nl = NodeSetExpr(context);
          if(null != nl)
          {
            result = new Double(nl.getLength());
          }
          else
          {
            result = new Double(0.0);
          } 
        }
        break;
        
      case FUNC_LAST:
        result = new Double((double)getCountOfContextNodeList());
        break;
        
      case FUNC_POSITION:
        result = new Double((double)getPositionInContextNodeList(context));
        break;
        
      case FUNC_ID:
        {
          String ref = StringExpr(context);
          Document ownerDoc;
          if(null != context)
          {
            ownerDoc = context.getOwnerDocument();
            if(null == ownerDoc)
            {
              if(Node.DOCUMENT_NODE == context.getNodeType())
              {
                ownerDoc = (Document)context;
              }
              else
              {
                error(context, "context does not have an owner document!");
              }
            }
            Node node = m_xslProcessor.m_parserLiaison.getElementByID(ref, ownerDoc);
            NodeListImpl nl = new NodeListImpl();
            if(null != node)
              nl.addElement(node);
            result = nl;
          }
          else
          {
            result = null;
          }
        }
        break;

      case FUNC_DOC:
        {
          String uri = StringExpr(context);
          if(null == uri)
          {
            warn("Can resolve expression to uri: "+m_currentPattern);
          }
          else
          {
            Node node = handleDocExpr(context, uri);
            NodeListImpl nl = new NodeListImpl();
            if(null != node)
              nl.addElement(node);
            result = nl;
          }
        }
        break;

      case FUNC_KEY:
        try
        {
          String keyname = StringExpr(context);
          consumeExpected(',');
          String ref = StringExpr(context);
          if(null != context)
          {
            Document docContext = (Node.DOCUMENT_NODE == context.getNodeType()) 
                                  ? (Document)context : context.getOwnerDocument();
            result = m_xslProcessor.getNodeSetByKey(docContext, keyname, ref, m_namespaceContext);
          }
        }
        // Have to catch generic exception because I don't want to 
        // propagate, or know about, SAX exceptions.
        catch(Exception e)
        {
          e.printStackTrace();
          error("keyref() threw an exception "+e.getMessage());
        }
        break;
        
      case FUNC_DOCREF:
      case FUNC_IDREF:
      case FUNC_KEYREF:
        try
        {
          String keyname;
          if(FUNC_KEYREF == tok)
          {
            keyname = StringExpr(context);
            consumeExpected(',');
          }
          else
          {
            keyname = null;
          }
          NodeList refs = NodeSetExpr(context);
          if(null != context)
          {
            Document docContext = (Node.DOCUMENT_NODE == context.getNodeType()) 
                                  ? (Document)context : context.getOwnerDocument();
            int nRefs = refs.getLength();
            NodeListImpl nodes = null;
            Hashtable usedrefs = new Hashtable();
            for(int i = 0; i < nRefs; i++)
            {
              String refval = m_xslProcessor.getNodeData((Node)refs.item(i));
              StringTokenizer tokenizer = new StringTokenizer(refval);
              while(tokenizer.hasMoreTokens())
              {
                String ref = tokenizer.nextToken();
                if(null == ref)
                  continue;
                if(usedrefs.get(ref) != null)
                {
                  continue;
                }
                else
                {
                  // m_currentPattern being used as a dummy value.
                  usedrefs.put(ref, m_currentPattern);
                }
                switch(tok)
                {
                case FUNC_DOCREF:
                  {
                    String uri = StringExpr(context);
                    Document newDoc = handleDocExpr(context, uri);
                    if(null != newDoc)
                      nodes.addElement(newDoc);
                  }
                  break;
                  
                case FUNC_IDREF:
                  {
                    Document ownerDoc;
                    ownerDoc = context.getOwnerDocument();
                    if(null == ownerDoc)
                    {
                      if(Node.DOCUMENT_NODE == context.getNodeType())
                      {
                        ownerDoc = (Document)context;
                      }
                      else
                      {
                        error(context, "context does not have an owner document!");
                      }
                    }
                    Node node = m_xslProcessor.m_parserLiaison.getElementByID(ref, ownerDoc);
                    if(null == nodes)
                    {
                      nodes = new NodeListImpl();
                    }
                    if(null != node)
                      nodes.addElement(node);
                  }
                  break;
                  
                case FUNC_KEYREF:
                  {
                    NodeList nl = m_xslProcessor.getNodeSetByKey(docContext, 
                                                                 keyname, ref, 
                                                                 m_namespaceContext);
                    if(null == nodes)
                    {
                      nodes = (NodeListImpl)nl;
                    }
                    else
                    {
                      int nNodes = nl.getLength();
                      for(int k = 0; k < nNodes; k++)
                      {
                        addNodeInDocOrder(nodes, nl.item(k));
                      }
                    }
                  }
                  break;
                }
              }
            }
            result = nodes;
          }
        }
        // Have to catch generic exception because I don't want to 
        // propagate, or know about, SAX exceptions.
        catch(Exception e)
        {
          error("keyref() threw an exception "+e.getMessage());
        }
        break;
                
      case FUNC_LOCAL_PART:
      case FUNC_NAMESPACE:
      case FUNC_QNAME:
        {
          Node qnameNode;
          if(tokenIs(')'))
          {
            qnameNode = context;
          }
          else
          {
            NodeListImpl nodes = NodeSetExpr(context);
            int nNodes = nodes.getLength();
            if(nNodes > 0)
            {
              qnameNode = nodes.item(0);
            }
            else
            {
              qnameNode = null;
            }
          }
          if(null != qnameNode)
          {
            switch(tok)
            {
            case FUNC_LOCAL_PART:
              result = m_xslProcessor.m_parserLiaison.getLocalNameOfNode(qnameNode);
              break;
            case FUNC_NAMESPACE:
              result = m_xslProcessor.m_parserLiaison.getNamespaceOfNode(qnameNode);
              break;
            case FUNC_QNAME:
              result = (Node.ATTRIBUTE_NODE == qnameNode.getNodeType()) ?
                       ((Attr)qnameNode).getName() :
                       qnameNode.getNodeName();
              break;
            }
          }
          else
          {
            result = new String();
          }
        }
        break;
        
      case FUNC_GENERATE_ID:
        Object n;
        if(tokenIs(')'))
        {
          n = context;
        }
        else
        {
          Vector v = NodeSetExpr(context);
          n = (v.size() > 0) ? v.elementAt(0) : null;
        }
        result = (null != n) ? "N"+Integer.toString(n.hashCode()) : "";
        break;
        
      case FUNC_NOT:
        {
          Boolean b = (Boolean)Expr(context, CLASS_BOOLEAN);
          result = new Boolean( !b.booleanValue() );
        }
        break;
        
      case FUNC_TRUE:
        result = new Boolean( true );
        break;
        
      case FUNC_FALSE:
        result = new Boolean( false );
        break;
        
      case FUNC_BOOLEAN:
        result = BooleanExpr(context);
        break;
        
      case FUNC_LANG:
        {
          String lang = StringExpr(context);
          Node parent = context;
          boolean isLang = false;
          while(null != parent)
          {
            if(Node.ELEMENT_NODE == parent.getNodeType())
            {
              String langVal = ((Element)parent).getAttribute("xml:lang");
              if(null != langVal)
              {
                if(langVal.toLowerCase().startsWith(lang.toLowerCase()))
                {
                  int valLen = lang.length();
                  if((langVal.length() == valLen) || (langVal.charAt(valLen) == '-'))
                  {
                    isLang = true;
                    break;
                  }
                }
              }
            }
            parent = m_xslProcessor.m_parserLiaison.getParentOfNode(parent);
          }
          result = new Boolean(isLang);
        }
        break;
        
      case FUNC_NUMBER:
        result = NumberExpr(context);
        break;
        
      case FUNC_FLOOR:
        {
          Double d = NumberExpr(context);
          result = new Double( java.lang.Math.floor(d.doubleValue()) );
        }
        break;
        
      case FUNC_CEILING:
        {
          Double d = NumberExpr(context);
          result = new Double( java.lang.Math.ceil(d.doubleValue()) );
        }
        break;
        
      case FUNC_ROUND:
        {
          Double d = NumberExpr(context);
          result = new Double( java.lang.Math.round(d.doubleValue()) );
        }
        break;
        
      case FUNC_SUM:
        {
          Vector nodes = NodeSetExpr(context);
          int nNodes = nodes.size();
          double sum = 0.0;
          for(int i = 0; i < nNodes; i++)
          {
            String s = m_xslProcessor.getNodeData((Node)nodes.elementAt(i));
            if(null != s) 
            {
              sum += Double.valueOf(s).doubleValue();
            }
          }
          result = new Double(sum);
        }
        break;
        
      case FUNC_STRING:
        result = StringExpr(context);
        break;

      case FUNC_SUBSTRING:
        {
          String str = StringExpr(context);
          consumeExpected(',');
          Double start = NumberExpr(context);
          // Convert one-based to zero-based.
          if(start.intValue() > 0) start = new Double(start.intValue() - 1);
          consumeExpected(',');
          Double len = NumberExpr(context);
          if(null != str)
          {
            result = str.substring(start.intValue(), len.intValue());
          }
          else
          {
            result = "";
          }
        }
        break;

      case FUNC_STRING_LENGTH:
        {
          String str = StringExpr(context);
          result = new Double((null != str) ? str.length() : 0);
        }
        break;

      case FUNC_STARTS_WITH:
        {
          String s1 = StringExpr(context);
          consumeExpected(',');
          String s2 = StringExpr(context);
          boolean startswith = s1.startsWith(s2);
          result = new Boolean(startswith);
        }
        break;
        
      case FUNC_CONTAINS:
        {
          String s1 = StringExpr(context);
          consumeExpected(',');
          String s2 = StringExpr(context);
          int index = s1.indexOf(s2);
          result = new Boolean(index > -1);
        }
        break;
        
      case FUNC_SUBSTRING_BEFORE:
        {
          String s1 = StringExpr(context);
          consumeExpected(',');
          String s2 = StringExpr(context);
          int index = s1.indexOf(s2);
          result = (-1 == index) ? "" : s1.substring(0,  index);
        }
        break;
        
      case FUNC_SUBSTRING_AFTER:
        {
          String s1 = StringExpr(context);
          consumeExpected(',');
          String s2 = StringExpr(context);
          int index = s1.indexOf(s2);
          result = (-1 == index) ? "" : s1.substring(index+s2.length());
        }
        break;
        
      case FUNC_NORMALIZE:
        {
          String s1;
          if(tokenIs(')'))
          {
            s1 = m_xslProcessor.getNodeData(context);
          }
          else
          {
            s1 = StringExpr(context);
          }
          result = m_xslProcessor.fixWhiteSpace(s1, true, true, false);
        }
        break;
        
      case FUNC_TRANSLATE:
        {
          String s1 = StringExpr(context);
          consumeExpected(',');
          String s2 = StringExpr(context);
          consumeExpected(',');
          String s3 = StringExpr(context);
          int nS2Chars = s2.length();
          int nS3Chars = s3.length();
          if(nS2Chars != nS3Chars)
          {
            error("The length of the second arg must match the length of the third arg in the translate function");
          }
          char[] s2chars = s2.toCharArray();
          char[] s3chars = s3.toCharArray();
          for(int i = 0; i < nS2Chars; i++)
          {
            s1 = s1.replace(s2chars[i], s3chars[i]);
          }
         
          result = s1;
        }
        break;
        
      case FUNC_CONCAT:
        {
          String s1 = StringExpr(context);
          consumeExpected(',');
          while(true)
          {
            String s2 = StringExpr(context);
            s1 = s1.concat(s2);
            if(tokenIs(','))
            {
              nextToken();
            }
            else
            {
              break;
            }
          }
          result = s1;
        }
        break;
        
      case FUNC_FORMAT_NUMBER:
        Double num = NumberExpr(context);
        consumeExpected(',');
        String formatStr = StringExpr(context);
        if(tokenIs(','))
        {
          nextToken();
          // String nameOfLocale = StringExpr(context);
          warn("don't yet handle the locale name in the format-number function");
        }

        java.text.DecimalFormat formatter = new java.text.DecimalFormat(formatStr);
        result = formatter.format(num.doubleValue());
        break;
        
      case FUNC_SYSTEM_PROPERTY:
        {
          String fullName = StringExpr(context);
          int indexOfNSSep = fullName.indexOf(':');
          if(indexOfNSSep > 0)
          {
            String prefix = (indexOfNSSep >= 0) ? fullName.substring(0, indexOfNSSep) : "";
            String namespace 
              = m_xslProcessor.m_parserLiaison.getNamespaceForPrefix(prefix,m_namespaceContext);
            String propName = (indexOfNSSep < 0) ? fullName 
                                                   : fullName.substring(indexOfNSSep+1);
            
            if(namespace.equals(m_xslProcessor.m_XSLNameSpaceURL))
            {
              if(propName.equals("version"))
              {
                result = m_xslProcessor.m_XSLTVerSupported;
              }
              else
              {
                warn("XSL Property not supported: "+fullName);
                result = "";
              }
            }
            else
            {
              warn("Don't currently do anything with namespace in property: "+fullName);
              try
              {
                result = System.getProperty(propName);
                if(null == result)
                {
                  // result = System.getenv(propName);
                  result = "";
                }
              }
              catch(SecurityException se)
              {
                warn("SecurityException when trying to access XSL system property: "+fullName);
                result = "";
              }
           }
          }
          else
          {
            try
            {
              result = System.getProperty(fullName);
              if(null == result)
              {
                // result = System.getenv(fullName);
                result = "";
              }
            }
            catch(SecurityException se)
            {
              warn("SecurityException when trying to access XSL system property: "+fullName);
              result = "";
            }
          }
        }
        break;
        
      case FUNC_FUNCTION_AVAILABLE:
        // The function-available() function returns true if an 
        // implementation of the named extension function is available. 
        {
          String fullName = StringExpr(context);
          boolean isAvailable = false;
          int indexOfNSSep = fullName.indexOf(':');
          String prefix = (indexOfNSSep >= 0) ? fullName.substring(0, indexOfNSSep) : "";
          String namespace 
            = m_xslProcessor.m_parserLiaison.getNamespaceForPrefix(prefix,m_namespaceContext);
          if(null != namespace)
          {
            Dispatcher dispatcher 
              = (Dispatcher)m_xslProcessor.m_extensionNamespaces.get(namespace);
            if(null != dispatcher)
            {
              String methName = (indexOfNSSep < 0) ? fullName 
                                                     : fullName.substring(indexOfNSSep+1);
              isAvailable = dispatcher.isFunctionAvailable(methName);
            }
          }
          result = new Boolean(isAvailable);
        }
        break;

      default:
        error(funcName+"(...) not implemented!");
        break;
      } // end switch
      
      consumeExpected(')');
    }
    else
    {
      if(!isNamespaced)
        tok = getKeywordToken(m_token);
      if(tok > 0)
      {
        NodeListImpl queryResults = new NodeListImpl();
        result = LocationPath(context, queryResults);
        // Check to make sure the location path was properly consumed, and 
        // that there isn't extra path-like characters left around.
        if(tokenIs('['))
        {
          error("Found extra '['!");
        }
        else if(tokenIs('/'))
        {
          error("Found extra '/'!");
        }
      }
      else if(lookahead(':', 1) || lookahead('(', 1))
      {
        String namespace;
        if(isNamespaced)
        {
          namespace = m_token;
          nextToken(); nextToken();
        }
        else
        {
          namespace = "";
        }
        
        String extensionName = m_token;
        // System.out.println("Found func name: "+extensionName);
        
        nextToken(); nextToken();
        
        Dispatcher dispatcher 
          = (Dispatcher)m_xslProcessor.m_extensionNamespaces.get(namespace);
        if(null != dispatcher)
        {
          Vector argVec = new Vector();
          while(!tokenIs(')'))
          {
            Object arg = Expr(context, CLASS_DONTCARE);
            argVec.addElement(arg);
            if(tokenIs(','))
            {
              nextToken();
            }
            else
              break;
          }
          
          int nNodes = argVec.size();
          Object[] args = null;
          if(nNodes > 0)
          {
            args = new Object[nNodes];
            for(int i = 0; i < nNodes; i++)
            {
              args[i] = argVec.elementAt(i);
            }
          }

          // Object obj = null;
          try
          {
            result = dispatcher.call(extensionName, args);
          }
          catch(Exception e)
          {
            warn(context, "Extension function, exception thrown: "+e);
          }
        }
        else
        {
          warn(context, "Extension not found: "+funcName);
        }
      }
      else
      {
        error(funcName+"(...) not known!");
      }
    }
    
    return result;
  }

  // ============= GRAMMAR FUNCTIONS =================
  private void ____LOCATION_PATHS____(){}

  /**
   * 
   * --------------------------------------------------------------------------------
   LocationPath ::= RelativeLocationPath 
   | AbsoluteLocationPath 
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected NodeListImpl LocationPath(Node context, NodeListImpl queryResults) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("LocationPath("+context+")");
    // If queryResults is null, then we're doing a MatchExpr
    
    NodeListImpl results; 
    if(tokenIs('/'))
    {
      results = AbsoluteLocationPath(context, queryResults, false);
    }
    else
    {
      results = RelativeLocationPath(context, queryResults);
    }
    return results;
  }
  
  /**
   * 
   * --------------------------------------------------------------------------------
   AbsoluteLocationPath ::= '/' RelativeLocationPath? 
   | AbbreviatedAbsoluteLocationPath 
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public NodeListImpl AbsoluteLocationPath(Node context, NodeListImpl queryResults, boolean isURI) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("AbsoluteLocationPath("+context+")");
    NodeListImpl results = null;

    if((null != context) && (context.getNodeType() != Node.DOCUMENT_NODE))
    {
      Document doc = context.getOwnerDocument();
      if(null == doc)
      {
        error(context, "Child doesn't have an owner document!");
      }
      context = doc;
    }
    
    nextToken(); // consume '/'
    
    if(tokenIs(',') 
       || tokenIs(')') 
       || tokenIs(null)  
       || tokenIs('|')) // I'm not happy with this test, but...
    {
      if(null != context)
      {
        if(null == queryResults)
        {
          queryResults = new NodeListImpl();
        }
        queryResults.addElement(context);
        results = queryResults;
      }
    }
    else
    {
      results = RelativeLocationPath( context, queryResults );
    }
    
    return results;
  }

  /**
   * 
   * --------------------------------------------------------------------------------
   RelativeLocationPath ::= Step 
   | RelativeLocationPath '/' Step 
   | AbbreviatedRelativeLocationPath 
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected NodeListImpl RelativeLocationPath(Node context, NodeListImpl queryResults) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("RelativeLocationPath("+context+")");
    // The rest of the processing after Step will be handled down the stack, 
    // rather than in a loop.
    NodeList savedContextNodeList = m_contextNodeList;
    m_contextNodeList = queryResults;
    Object resultObj = Step(context, queryResults, false);
    m_contextNodeList = savedContextNodeList;
    
    NodeListImpl results;
    try
    {
      results = (null != resultObj) ? (NodeListImpl)resultObj : null;
    }
    catch(Exception e)
    {
      results = null; // to shutup compiler
      error("Step returned something other than a NodeListImpl!");
    }
   
    return results;
  }
  
  
  /**
   * 
   * --------------------------------------------------------------------------------
   Step ::= AxisIdentifier '(' NodeTest Predicate* ')' 
   | AbbreviatedStep 

   AxisIdentifier ::= 'from-ancestors' 
   | 'from-ancestors-or-self' 
   | 'from-attributes' 
   | 'from-children' 
   | 'from-descendants' 
   | 'from-descendants-or-self' 
   | 'from-following' 
   | 'from-following-siblings' 
   | 'from-parent' 
   | 'from-preceding' 
   | 'from-preceding-siblings' 
   | 'from-self' 

   AbbreviatedStep ::= AbbreviatedNodeTestStep 
   | '.' 
   | '..' 
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected Object Step(Node context, NodeListImpl queryResults, boolean stepOnly) 
    throws XSLProcessorException, java.net.MalformedURLException, java.io.FileNotFoundException, java.io.IOException
  {
    if(m_trace) trace("Step("+context+")");
    NodeList nl = null;
    
    int savedStepPos = m_stepMark;
    m_stepMark = this.m_queueMark;
    
    Node savedStepContext = m_stepContext;
    m_stepContext = context;
        
    boolean isNotAbbrev = lookahead('(', 1);
    
    String keywordToken = m_token;
    NodeListImpl subQueryResults = new NodeListImpl();
    //  = ((null != queryResults) && (queryResults.size() == 0)) ? queryResults :
    //                                                             new NodeListImpl();

    int tok = getKeywordToken(keywordToken);
    boolean shouldReorder = false;
    
    switch(tok)
    {
    case NODETYPE_COMMENT:
    case NODETYPE_TEXT:
    case NODETYPE_NODE:
    case NODETYPE_PI:
      isNotAbbrev = false;
      break;
    case FROM_DESCENDANTS_OR_SELF:
    case FROM_PARENT:
    case FROM_SELF:
    case FROM_ATTRIBUTE:
      if(isNotAbbrev)
        nextToken();
      nextToken();
      break;
    default:
      if(isNotAbbrev)
      {
        nextToken(); nextToken();
        if(tok == FROM_DESCENDANTS)
        {
          tok = FROM_DESCENDANTS;
        }
      }
      else if((m_tokenChar == '*') || Character.isLetter(m_tokenChar))
      {
        tok = FROM_CHILDREN;
      }
      else
      {
        // Quick bail-out, sorry.
        return queryResults;
      }
      break;
    }
    boolean needsNodeTest = true;
        
    // finishedWalk will be set to true if we have finished 
    // the parse.
    int mark = this.m_queueMark;
    
    if(null != context)
    {
      switch(tok)
      {
      case NODETYPE_COMMENT:
      case NODETYPE_TEXT:
      case NODETYPE_NODE:
      case NODETYPE_PI:
        {
          int contextNodeType = context.getNodeType();
          if((Node.ELEMENT_NODE == contextNodeType) || (Node.DOCUMENT_NODE == contextNodeType))
          {
            nl = context.getChildNodes();
          }
          if( (nl != null) ||  (Node.DOCUMENT_NODE == contextNodeType))
          {          
            parseNodeTest(false, m_ntinfo);
            // look at child of each node in the context
            int nChildNodes = (null != nl) ? nl.getLength() : 1;
            for( int j=0; j < nChildNodes; j++ )
            {
              Node c = (null != nl) ? nl.item(j) : ((Document)context).getDocumentElement();
              if(NodeTest(c, m_ntinfo))
              {
                if(null != subQueryResults)
                  subQueryResults.addElement(c);
              }
            }
          }
        }
        break;
        
      case FROM_PARENT:
        context = m_xslProcessor.m_parserLiaison.getParentOfNode(context);
      case FROM_SELF:
       if(!isNotAbbrev)
          needsNodeTest = false;
        if(null != context)
        {
          if(isNotAbbrev)
          {
            if(NodeTest(context, false))
            {
              if(null != subQueryResults)
                subQueryResults.addElement(context);
            }
          }
          else
          {
           if(null != subQueryResults)
              subQueryResults.addElement(context);
          }
        }
        break;

      case FROM_ANCESTORS:
        context = m_xslProcessor.m_parserLiaison.getParentOfNode(context);
      case FROM_ANCESTORS_OR_SELF:
        shouldReorder = true;
        parseNodeTest(false, m_ntinfo);
        while(null != context)
        {
          if(NodeTest(context, m_ntinfo))
          {
            if(null != subQueryResults)
              addNodeInDocOrder(subQueryResults, context);
          }
          context = m_xslProcessor.m_parserLiaison.getParentOfNode(context);
        }
        break;
        
      case FROM_ATTRIBUTE:
      case FROM_ATTRIBUTES:
        if( (null != context) && (context.getNodeType()==Node.ELEMENT_NODE) )
        {
          Element e = (Element)context;
          NamedNodeMap attributeList = e.getAttributes();
          if( attributeList != null ) 
          {
            int nAttrs = attributeList.getLength();
            parseNodeTest(true, m_ntinfo);
            for( int j=0; j < nAttrs; j++ )
            {
              Attr attr = (Attr)attributeList.item(j);
              if(NodeTest(attr, m_ntinfo))
              {
                if(null != subQueryResults)
                  subQueryResults.addElement(attr);
              }
            }
          }
        }
        break;
        
      case FROM_CHILDREN:
        {
          int contextNodeType = context.getNodeType();
          if((Node.ELEMENT_NODE == contextNodeType) || (Node.DOCUMENT_NODE == contextNodeType))
          {
            nl = context.getChildNodes();
          }
          if( (nl != null) ||  (Node.DOCUMENT_NODE == contextNodeType))
          {          
            // look at child of each node in the context
            int nChildNodes = (null != nl) ? nl.getLength() : 1;
            parseNodeTest(false, m_ntinfo);
            for( int j=0; j < nChildNodes; j++ )
            {
              Node c = (null != nl) ? nl.item(j) : ((Document)context).getDocumentElement();
              if(NodeTest(c, m_ntinfo))
              {
                if(null != subQueryResults)
                  subQueryResults.addElement(c);
              }
            }
          }
        }
        break;
        
      case FROM_DESCENDANTS_OR_SELF:
      case FROM_DESCENDANTS:
        {
          // Perform a pre-order traversal of descendents.
          // Note that I would like to be able to do optimization here 
          // where if I have a simple tag name node test, I would 
          // like to be able to call 
          // ((Element)context).getElementsByTagName(m_token).
          // One problem is that it would return a NodeList, not a 
          // NodeListImpl.
          /*
          if(lookahead('[', 1) || lookahead('(', 1) 
          || (Node.ELEMENT_NODE != context.getNodeType()))
          {
          */
          if(isNotAbbrev)
            parseNodeTest(false, m_ntinfo);
          Node pos = context;
          
          while(null != pos)
          {                   
            if((tok == FROM_DESCENDANTS_OR_SELF) || (context != pos))
            {
              if(!isNotAbbrev)
              {
                resetTokenMark(mark);
                subQueryResults = RelativeLocationPath(pos, subQueryResults);
                /*
                NodeListImpl v = new NodeListImpl();
                v = RelativeLocationPath(pos, v);
                if(null != v)
                {
                  int n = v.size();
                  for(int k = 0; k < n; k++)
                  {
                    addNodeInDocOrder(subQueryResults, (Node)v.elementAt(k), true);
                  }
                }
                */
              }
              else
              {
                if(NodeTest(pos, m_ntinfo))
                {
                  if(null != subQueryResults)
                    subQueryResults.addElement(pos);
                }
              }
            }
            Node nextNode = pos.getFirstChild();
            while(null == nextNode)
            {
              if(context == pos)
                break;
              nextNode = pos.getNextSibling();
              if(null == nextNode)
              {
                pos = pos.getParentNode();
                if(context == pos)
                {
                  nextNode = null;
                  break;
                }
              }
            }
            pos = nextNode;
          }
          /*
          if(!isNotAbbrev)
          {
            resetTokenMark(mark);
            subQueryResults = RelativeLocationPath(null, null);
          }
          */

          /*
          }
          else
          {
          if(tok == FROM_DESCENDANTS_OR_SELF)
          {
          if(context.getNodeName().equals(m_token))
          {
          subQueryResults.addElement(context);
          }
          }
          NodeList descendants = ((Element)context).getElementsByTagName(m_token);
          nextToken();
          int n = descendants.getLength();
          for(int k = 0; k < n; k++)
          {
          subQueryResults.addElement(descendants.item(k));
          }
          }
          */
        }
        break;
        
      case FROM_FOLLOWING_SIBLINGS:
        {
          Node pos = context.getNextSibling();
          parseNodeTest(false, m_ntinfo);
          while(null != pos)
          {                   
            if(NodeTest(pos, m_ntinfo))
            {
              if(null != subQueryResults)
                subQueryResults.addElement(pos);
            }
            pos = pos.getNextSibling();
          }
        }
        break;

      case FROM_PRECEDING_SIBLINGS:
        {
          shouldReorder = true;
          Node pos = context.getPreviousSibling();
          parseNodeTest(false, m_ntinfo);
          while(null != pos)
          {                   
            if(NodeTest(pos, m_ntinfo))
            {
              if(null != subQueryResults)
                subQueryResults.addElement(pos);
            }
            pos = pos.getPreviousSibling();
          }
        }
        break;
        
      case FROM_FOLLOWING:
        {
          // What fun...
          Document doc = context.getOwnerDocument();
          Node pos = context;
          parseNodeTest(false, m_ntinfo);
          while(null != pos)
          {                   
            if(NodeTest(pos, m_ntinfo))
            {
              if((null != subQueryResults) && (pos != context))
                addNodeInDocOrder(subQueryResults, pos);
            }
            Node nextNode = pos.getFirstChild();
            while(null == nextNode)
            {
              nextNode = pos.getNextSibling();
              if(null == nextNode)
              {
                pos = pos.getParentNode();
                if(doc == pos)
                {
                  nextNode = null;
                  break;
                }
              }
            }
            pos = nextNode;
          }
        }
        break;
        
      case FROM_PRECEDING:
        {
          shouldReorder = true;
          // Ugh.  Reverse document order, no parents, I guess.
          Document doc = context.getOwnerDocument();
          Node pos = doc;
          parseNodeTest(false, m_ntinfo);
          while(null != pos)
          {       
            if(context == pos)
              break;

            if(NodeTest(pos, m_ntinfo))
            {
              // Ugh. If I could think a little better tonight, I'm
              // sure there's a better way to check for the parent.
              boolean isParent = false;
              Node parent = m_xslProcessor.m_parserLiaison.getParentOfNode(context);
              while(null != parent)
              {
                if(parent == pos)
                {
                  isParent = true;
                  break;
                }
                parent = m_xslProcessor.m_parserLiaison.getParentOfNode(parent);
              }
              
              if(!isParent)
              {
              if(null != subQueryResults)
                subQueryResults.insertElementAt(pos, 0);
              }
            }
            Node nextNode = pos.getFirstChild();
            while(null == nextNode)
            {
              nextNode = pos.getNextSibling();
              if(null == nextNode)
              {
                pos = pos.getParentNode();
                if(doc == pos)
                {
                  nextNode = null;
                  break;
                }
              }
            }
            pos = nextNode;
          }
        }
        break;
        
      default:
        error("?? Unknown AxisIdentifier: "+keywordToken);
        break;
        
      }
    }
    
    boolean nextIsPredicate = tokenIs('[');
            
    if((mark == this.m_queueMark) && needsNodeTest && !nextIsPredicate && !tokenIs('/'))
    {
      NodeTest(null, false);
    }
        
    if((null != subQueryResults) && (subQueryResults.getLength() > 0))
    {
      if(nextIsPredicate)
      {
        subQueryResults = continueFromPredicate(subQueryResults, queryResults,
                                                this.m_queueMark, 
                                                isNotAbbrev, stepOnly);
      }
      else
      {
        if(isNotAbbrev)
          consumeExpected(')');
        if(tokenIs('/'))
        {
          int nContexts = subQueryResults.getLength();
          mark = m_queueMark;
          for(int i = 0; i < nContexts; i++)
          {
            if(i >0)
              resetTokenMark(mark);
            queryResults = continueLocationPath(subQueryResults.item(i), queryResults);
          }
        }
        else
        {
          int nResults = subQueryResults.getLength();

          for(int i = 0; i < nResults; i++)
          {
            addNodeInDocOrder(queryResults, subQueryResults.item(i));
          }
        }
      }
    }
    else
    {
      while(tokenIs('['))
      {
        Predicate(null);  // just consume
      }
      if(isNotAbbrev)
        consumeExpected(')');
      if(!stepOnly)
        continueLocationPath(null, null);
    }
    
    m_stepContext = savedStepContext;
    m_stepMark = savedStepPos;
    
    if(shouldReorder && (null != queryResults))
    {
      NodeListImpl ordered = new NodeListImpl();
      int n = queryResults.size();
      for(int i = 0; i < n; i++)
      {
        addNodeInDocOrder(ordered, (Node)queryResults.elementAt(i), i != 0);
      }
      queryResults = ordered;
    }

    return queryResults;
  }
  
  /**
   * This function takes a set of nodes, and tests the predicates on them 
   * from a given mark.  It should always complete the 
   */
  private NodeListImpl continueFromPredicate(NodeListImpl subQueryResults, 
                                             NodeListImpl queryResults, 
                                             int nextPredStartMark,
                                             boolean isNotAbbrev, boolean stepOnly)
    throws XSLProcessorException, java.net.MalformedURLException, java.io.FileNotFoundException, java.io.IOException
  {
    int endMark = -2;
    boolean didConsumeEndParen = false;
    
    if((nextPredStartMark > 0) && (null != queryResults))
    {
      do
      {
        m_contextNodeList = subQueryResults;
        NodeListImpl newQueryResults = new NodeListImpl();
        int nContextNodes = subQueryResults.getLength();
        if(0 == nContextNodes)
        {
          break;
        }
        int predEndMark = -2;
        for(int i = 0; i < nContextNodes; i++)
        {
          resetTokenMark(nextPredStartMark);
          didConsumeEndParen = false;

          Node subContext = subQueryResults.item(i);
          if(Predicate(subContext))
          {
            if(isNotAbbrev && tokenIs(')'))
            {
              didConsumeEndParen = true;
              nextToken();
              if(tokenIs('/'))
              {
                if(!stepOnly)
                {
                  nextToken();
                  queryResults = RelativeLocationPath(subContext, queryResults);
                }
                else
                {
                  addNodeInDocOrder(queryResults, subContext);
                }
                endMark = this.m_queueMark;
              }
              else if(!tokenIs('['))
              {
                endMark = this.m_queueMark;
                addNodeInDocOrder(queryResults, subContext);
              }
              else
              {
                addNodeInDocOrder(newQueryResults, subContext);
              }
            }
            else if(tokenIs('/'))
            {
              if(!stepOnly)
              {
                nextToken();
                queryResults = RelativeLocationPath(subContext, queryResults);
              }
              else
              {
                addNodeInDocOrder(queryResults, subContext);
              }
              endMark = this.m_queueMark;
            }
            else if(!tokenIs('['))
            {
              endMark = this.m_queueMark;
              addNodeInDocOrder(queryResults, subContext);
            }
            else
            {
              addNodeInDocOrder(newQueryResults, subContext);
            }
          }
          predEndMark = this.m_queueMark;
        }
        subQueryResults = newQueryResults;
        nextPredStartMark = predEndMark;
        m_contextNodeList = queryResults;
      }
        while(tokenIs('['));
      
      if(endMark > -2)
      {
        // resetTokenMark(endMark);
      }
    }
    
    if(null != m_token)
    {
      if(endMark > -2)
      {
        // resetTokenMark(endMark+1);
        if(null != m_token)
        {
          if(isNotAbbrev && !didConsumeEndParen)
            consumeExpected(')');
          if(m_queueMark < m_tokenQueueSize)
            continueLocationPath(null, null);
        }
      }
      else
      {
        while(tokenIs('['))
        {
          Predicate(null);  // just consume
        }
        if(isNotAbbrev)
          consumeExpected(')');
        continueLocationPath(null, null);
      }
    }
    return queryResults;
  }
  
  /**
   * This continues the RelativeLocationPath...
   */
  private NodeListImpl continueLocationPath(Node context, NodeListImpl queryResults)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(tokenIs('/'))
    {
      nextToken();
      queryResults = RelativeLocationPath(context, queryResults);
    }
    else if(null != context)
    {
      addNodeInDocOrder(queryResults, context);
    }
    
    return queryResults;
  }
  
/**
 * Record of info needed to do a fast node test, without the
 * overhead of reparsing each time.
 */
  private class NodeTestInfo
  {
    public NodeTestInfo()
    {
    }
    
    public void reset()
    {
      m_isAttrTest = false;
      m_tok = -1;
      m_ns = null;
      m_name = null;
      m_isWildcard = false;
      m_isXMLNSDecl = false;
    }
    
    boolean m_isAttrTest = false;
    int m_tok = -1;
    String m_ns = null;
    String m_name = null;
    boolean m_isWildcard = false;
    boolean m_isXMLNSDecl = false;
  }
  
  private NodeTestInfo m_ntinfo = new NodeTestInfo();
  
  /**
   * Parse the NodeTest only, don't perform the actual tests.
   * --------------------------------------------------------------------------------
   NodeTest ::= WildcardName 
   | NodeType '(' ')' 
   | 'pi' '(' Literal ')' 
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void parseNodeTest(boolean isAttrTest, 
                                  NodeTestInfo ntinfo) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    ntinfo.reset();
    if(tokenIs('@')) // don't think this is kosher, but it's hard to see how it can hurt.
    {
      nextToken();
      isAttrTest = true;
    }
    ntinfo.m_isAttrTest = isAttrTest;
    if( lookahead('(', 1) )
    {
      int tok = getKeywordToken(m_token);
      ntinfo.m_tok = tok;
      switch(tok)
      {
      case NODETYPE_COMMENT:
      case NODETYPE_TEXT:
      case NODETYPE_NODE:
        NodeType(null, tok);
        break;
        
      case NODETYPE_PI:
        if( lookahead(')', 2) )
        {
          NodeType(null, tok);
        }
        else
        {
          nextToken(); nextToken();
          ntinfo.m_name = Literal();
          consumeExpected(')');
        }
        break;
        
      default:
        {
          error("Expected 'text' or 'comment' or 'id' or 'pi' or 'ancestor', but found: "+m_token);
        }
      }
    }
    else
    {
      if(lookahead(':', 1))
      {
        ntinfo.m_ns = WildcardName();
        nextToken(); // consume ':'
      }
      ntinfo.m_name = WildcardName();
      if(ntinfo.m_name.equals("*"))
      {
        ntinfo.m_isWildcard = true;
      }
      if(ntinfo.m_isWildcard && ntinfo.m_isAttrTest)
      {
        ntinfo.m_isXMLNSDecl = ((ntinfo.m_name.startsWith("xmlns:") 
                                 || ntinfo.m_name.equals("xmlns")));
      }
    }
  }
  
  /**
   * Process the NodeTest without any parsing.
   * --------------------------------------------------------------------------------
   NodeTest ::= WildcardName 
   | NodeType '(' ')' 
   | 'pi' '(' Literal ')' 
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected boolean NodeTest(Node context, NodeTestInfo ntinfo) 
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    if(m_trace) trace("NodeTest("+context+"), "+m_token);
    
    boolean test;
    switch(ntinfo.m_tok)
    {
    case NODETYPE_COMMENT:
    case NODETYPE_TEXT:
    case NODETYPE_NODE:
      test = NodeTypeNoParse(context, ntinfo.m_tok);
      break;
      
    case NODETYPE_PI:
      test = NodeTypeNoParse(context, ntinfo.m_tok);
      if( test && (null != ntinfo.m_name) )
      {
        test = ((ProcessingInstruction)context).getNodeName().equals(ntinfo.m_name);
        if(test && (MATCH_SCORE_NONE == m_matchScore))
          m_matchScore = MATCH_SCORE_QNAME;
      }
      break;
      
    default:
      {
        if(null != context)
        {
          String contextNS = m_xslProcessor.m_parserLiaison.getNamespaceOfNode(context);
          if((null != ntinfo.m_ns) && (null != contextNS))
          {
            test = contextNS.equals(ntinfo.m_ns);
          }
          else
          {
            test = ntinfo.m_isWildcard || 
                   ((null == contextNS) || (contextNS.length() == 0));
          }

          int nodeType = context.getNodeType();
          switch(nodeType)
          {
          case Node.ATTRIBUTE_NODE:
            if(ntinfo.m_isAttrTest)
            {            
              if(test)
              {
                if(ntinfo.m_isWildcard)
                {
                  test = !ntinfo.m_isXMLNSDecl;
                }
                else
                {
                  String localAttrName 
                    = m_xslProcessor.m_parserLiaison.getLocalNameOfNode(context);
                  test = localAttrName.equals(ntinfo.m_name);
                  if(test && (MATCH_SCORE_NONE == m_matchScore))
                    m_matchScore = MATCH_SCORE_QNAME;

                }
              }
            }
            else
            {
              test = false;
            }
            break;

          case Node.ELEMENT_NODE:
            if(test)
            {
              if(!ntinfo.m_isAttrTest)
              {
                if(ntinfo.m_isWildcard)
                {
                  test = true;
                }
                else
                {
                  
                  test = (m_xslProcessor.m_parserLiaison.getLocalNameOfNode((Element)context).equals(ntinfo.m_name));
                  if(test && (MATCH_SCORE_NONE == m_matchScore))
                    m_matchScore = MATCH_SCORE_QNAME;
                }
              }
              else
              {
                test = false;
              }
            }
            break;
            
            // case Node.DOCUMENT_TYPE_NODE:
            // case Node.ENTITY_REFERENCE_NODE:
            // case Node.ENTITY_NODE:
            // case Node.DOCUMENT_FRAGMENT_NODE:
            // case Node.NOTATION_NODE:
          default:
            // Trying to match on anything else causes nasty bugs.
            test = false;
            break;

          }
        }
        else
        {
          test = false;
        }
      }
    }
    return test;    
  }


  /**
   * 
   * --------------------------------------------------------------------------------
   NodeTest ::= WildcardName 
   | NodeType '(' ')' 
   | 'pi' '(' Literal ')' 
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected boolean NodeTest(Node context, boolean isAttrTest) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("NodeTest("+context+"), "+m_token);
    
    if(tokenIs('@')) // don't think this is kosher, but it's hard to see how it can hurt.
    {
      nextToken();
      isAttrTest = true;
    }
    boolean test;
    if( lookahead('(', 1) )
    {
      int tok = getKeywordToken(m_token);
      switch(tok)
      {
      case NODETYPE_COMMENT:
      case NODETYPE_TEXT:
      case NODETYPE_NODE:
        test = NodeType(context, tok);
        break;
        
      case NODETYPE_PI:
        if( lookahead(')', 2) )
        {
          test = NodeType(context, tok);
        }
        else if((null != context) &&
                Node.PROCESSING_INSTRUCTION_NODE == context.getNodeType())
        {
          nextToken(); nextToken();
          String lit = Literal();
          test = ((ProcessingInstruction)context).getNodeName().equals(lit);
          if(test && (MATCH_SCORE_NONE == m_matchScore))
            m_matchScore = MATCH_SCORE_QNAME;
          consumeExpected(')');
        }
        else
        {
          test = false;
        }
        break;
        
      default:
        {
          test = false; // to make compiler happy
          error(context, "Expected 'text' or 'comment' or 'id' or 'pi' or 'ancestor', but found: "+m_token);
        }
      }
    }
    else
    {
      String ns;
      if(lookahead(':', 1))
      {
        ns = WildcardName();
        nextToken(); // consume ':'
      }
      else
      {
        ns = null;
      }
      String wildCardName = WildcardName();
      if(null != context)
      {
        String contextNS = m_xslProcessor.m_parserLiaison.getNamespaceOfNode(context);
        if((null != ns) && (null != contextNS))
        {
          test = contextNS.equals(ns);
        }
        else
        {
          test = wildCardName.equals("*") || 
                 ((null == contextNS) || (contextNS.length() == 0));
        }

        int nodeType = context.getNodeType();
        switch(nodeType)
        {
        case Node.ATTRIBUTE_NODE:
          if(isAttrTest)
          {            
            if(test)
            {
              if(wildCardName.equals("*"))
              {
                test = (!(wildCardName.startsWith("xmlns:") || wildCardName.equals("xmlns")));
              }
              else
              {
                String localAttrName 
                  = m_xslProcessor.m_parserLiaison.getLocalNameOfNode(context);
                test = localAttrName.equals(wildCardName);
                if(test && (MATCH_SCORE_NONE == m_matchScore))
                  m_matchScore = MATCH_SCORE_QNAME;

              }
            }
          }
          else
          {
            test = false;
          }
          break;

        case Node.ELEMENT_NODE:
          if(test)
          {
            if(!isAttrTest)
            {
              if(wildCardName.equals("*"))
              {
                test = true;
              }
              else
              {
                
                test = (m_xslProcessor.m_parserLiaison.getLocalNameOfNode((Element)context).equals(wildCardName));
                if(test && (MATCH_SCORE_NONE == m_matchScore))
                  m_matchScore = MATCH_SCORE_QNAME;
              }
            }
            else
            {
              test = false;
            }
          }
          break;
          
          // case Node.DOCUMENT_TYPE_NODE:
          // case Node.ENTITY_REFERENCE_NODE:
          // case Node.ENTITY_NODE:
          // case Node.DOCUMENT_FRAGMENT_NODE:
          // case Node.NOTATION_NODE:
        default:
          // Trying to match on anything else causes nasty bugs.
          test = false;
          break;

        }
      }
      else
      {
        test = false;
      }
    }
    return test;    
  }
  
  /**
   * 
   * --------------------------------------------------------------------------------
   NodeType ::= 'comment' 
   | 'text' 
   | 'pi' 
   | 'node' 
   * --------------------------------------------------------------------------------
   */
  protected boolean NodeTypeNoParse(Node context, int type) 
  {
    if(m_trace) trace("NodeType("+context+")");
    boolean isOfType;
    if(null != context)
    {
      int nodeType = context.getNodeType();
      switch(type)
      {
      case NODETYPE_COMMENT:
        isOfType = (Node.COMMENT_NODE == nodeType);
        break;
      case NODETYPE_TEXT:
        isOfType = (Node.CDATA_SECTION_NODE == nodeType) 
                   || (Node.TEXT_NODE == nodeType);
        break;
      case NODETYPE_PI:
        isOfType = (Node.PROCESSING_INSTRUCTION_NODE == nodeType);
        break;
      case NODETYPE_NODE:
        isOfType = true;
        break;
      default:
        isOfType = false; // to make compiler happy
        error(context, "Expected 'text' or 'comment' or 'pi' or 'node'!");
      }
    }
    else
    {
      isOfType = false;
    }
    return isOfType;
  }

  /**
   * 
   * --------------------------------------------------------------------------------
   NodeType ::= 'comment' 
   | 'text' 
   | 'pi' 
   | 'node' 
   * --------------------------------------------------------------------------------
   */
  protected boolean NodeType(Node context, int type) 
  {
    if(m_trace) trace("NodeType("+context+")");
    boolean isOfType;
    if(null != context)
    {
      int nodeType = context.getNodeType();
      switch(type)
      {
      case NODETYPE_COMMENT:
        isOfType = (Node.COMMENT_NODE == nodeType);
        break;
      case NODETYPE_TEXT:
        isOfType = (Node.CDATA_SECTION_NODE == nodeType) 
                   || (Node.TEXT_NODE == nodeType);
        break;
      case NODETYPE_PI:
        isOfType = (Node.PROCESSING_INSTRUCTION_NODE == nodeType);
        break;
      case NODETYPE_NODE:
        isOfType = true;
        break;
      default:
        isOfType = false; // to make compiler happy
        error(context, "Expected 'text' or 'comment' or 'pi' or 'node'!");
      }
    }
    else
    {
      isOfType = false;
    }
    nextToken(); nextToken(); nextToken();
    return isOfType;
  }
  
  /**
   * 
   * --------------------------------------------------------------------------------
   Predicate ::= '[' PredicateExpr ']' 
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected boolean Predicate(Node context) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("Predicate("+context+")");
    boolean nodeIsQualified = true;
    if(tokenIs('['))
    {
      nextToken();
      nodeIsQualified = PredicateExpr( context );      
      consumeExpected(']');
    }
    return nodeIsQualified;
  }

  /**
   * (see BooleanPrimaryExpr).
   *
   *--------------------------------------------------------------------------------
   PredicateExpr ::= Expr 
   *--------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected boolean PredicateExpr(Node context)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("PredicateExpr("+context+")");
    Object pred = Expr(context, CLASS_DONTCARE);
    int type = getClassType(pred);
    boolean result;
    if(CLASS_NUMBER == type)
    {
      int assertedPos = ((Double)pred).intValue();
      int contextPos = getPositionInContextNodeList(context);
      result = (assertedPos == contextPos);
    }
    else
    {
      Boolean b = (Boolean)castToType(pred, CLASS_BOOLEAN);
      result = b.booleanValue();
    }
    
    return result;
  }

  
  /**
   * 
   * --------------------------------------------------------------------------------
   AbbreviatedAbsoluteLocationPath ::= '//' RelativeLocationPath 
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected NodeListImpl AbbreviatedAbsoluteLocationPath(Node context, NodeListImpl queryResults) 
    throws XSLProcessorException, java.net.MalformedURLException, java.io.FileNotFoundException, java.io.IOException
  {
    if(m_trace) trace("AbbreviatedAbsoluteLocationPath("+context+")");
    return RelativeLocationPath(context, queryResults);
  }

  /**
   * 
   * --------------------------------------------------------------------------------
   AbbreviatedRelativeLocationPath  ::=  RelativeLocationPath '//' Step   
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected NodeListImpl AbbreviatedRelativeLocationPath(Node context, 
                                                         NodeListImpl queryResults,
                                                         boolean bRecursiveDescent, 
                                                         boolean isMatchExpr) 
    throws XSLProcessorException, java.net.MalformedURLException, 
           java.io.FileNotFoundException, java.io.IOException
  {
    if(m_trace) trace("AbbreviatedRelativeLocationPath("+context+")");
    queryResults = RelativeLocationPath(context, queryResults);

    return queryResults;
  }
  
  /**
   * Class to let us know when it's time to do 
   * a search from the parent because of indexing.
   */
  private class FoundIndex extends RuntimeException
  {
  }
  
  /**
   * Tells if FoundIndex should be thrown if index is found.
   */
  private boolean m_throwFoundIndex = false;

  /**
   * Only for use in patterns... 
   * --------------------------------------------------------------------------------
   AbbreviatedNodeTestStep  ::=  '@'? NodeTest Predicate* 
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected boolean AbbreviatedNodeTestStep(Node context) 
    throws XSLProcessorException, java.net.MalformedURLException, java.io.FileNotFoundException, java.io.IOException
  {
    // System.out.println(m_currentPattern);
    if(m_trace) trace("AbbreviatedNodeTestStep("+context+")");
    boolean isAttrTest;
    int savedStepContext = m_queueMark;
    if(tokenIs('@'))
    {
      isAttrTest = true;
      nextToken();
    }
    else
    {
      isAttrTest = false;
    }
    boolean test = NodeTest(context, isAttrTest);

    if(test)
    {
      if(tokenIs('['))
      {
        try
        {
          m_throwFoundIndex = true;
          do
          {
            test = Predicate(context);
          }
            while(tokenIs('[') && test);
          
          m_throwFoundIndex = false;
        }
        catch(FoundIndex fi)
        {
          // We have to get real ugly here because of the position()
          // and last() functions in patterns.  Major area for optimization.
          // Do a complete query step from the start of the AbbreviatedNodeTestStep,
          // using the parent of the current context as the context, 
          // then test to see if the context to be tested is in the results.
          m_throwFoundIndex = false;
          Node parentContext = m_xslProcessor.m_parserLiaison.getParentOfNode(context);
          resetTokenMark(savedStepContext);
          NodeListImpl queryResults = new NodeListImpl();
          queryResults = (NodeListImpl)Step(parentContext, queryResults, true);
          int nNodes = queryResults.getLength();
          test = false;
          for(int i = 0; i < nNodes; i++)
          {
            if(queryResults.item(i) == context)
            {
              test = true;
              break;
            }
          }
        }
        if(test)
          m_matchScore = MATCH_SCORE_OTHER;
      }
      else
      {
        if(MATCH_SCORE_NONE == m_matchScore)
          m_matchScore = MATCH_SCORE_NODETEST;
      }
    }
    return test;
  }
  
  // ============= PATTERN FUNCTIONS =================
  private void ____PATTERNS____(){}

  /**
   * 
   * --------------------------------------------------------------------------------
   Pattern  ::=  LocationPathPattern   
   | Pattern '|' LocationPathPattern   
   * --------------------------------------------------------------------------------
   */
  protected double Pattern(Node context)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("Pattern("+context+")");
    
    if((m_tokenQueueSize == 1) 
       && (Character.isLetter(m_tokenChar) || ('*' == m_tokenChar)))
    {
      if(Node.ELEMENT_NODE == context.getNodeType())
      {
        if('*' == m_tokenChar)
          return MATCH_SCORE_NODETEST;
        else 
        {
          String contextNS = m_xslProcessor.m_parserLiaison.getNamespaceOfNode(context);
          if((null == contextNS) && context.getNodeName().equals(m_token))
          {
            return MATCH_SCORE_QNAME;
          }
          else
          {
            return MATCH_SCORE_NONE;
          }
        }
      }
      else
      {
        return MATCH_SCORE_NONE;
      }
    }
    
    // Start out with the parent list as the context node list.  Not sure if 
    // this is correct.
    Node parent = m_xslProcessor.m_parserLiaison.getParentOfNode(context);
    m_contextNodeList = (null != parent) ? parent.getChildNodes() : null;

    boolean test = false;
    int subPatternStart = getNextSubpatternStartIndexInMap(-1);
    m_matchScore = MATCH_SCORE_NONE;
    int highestScore = MATCH_SCORE_NONE;
    while(!test)
    {
      test= LocationPathPattern( context, false, subPatternStart);
      
      if(!test)
        m_matchScore = MATCH_SCORE_NONE;
      
      if(m_matchScore > highestScore)
      {
        highestScore = m_matchScore;
      }
      if(test)
        break;

      subPatternStart = getNextSubpatternStartIndexInMap(subPatternStart);

      // if(tokenIs('|'))
      if(subPatternStart > -1)
      {
        // subPatternStart = getNextSubpatternStartIndexInMap(subPatternStart);
        resetTokenMark(subPatternStart);
        m_matchScore = MATCH_SCORE_NONE;
      }
      else
      {
        break;
      }
    }
    return (double)highestScore;
  }
  
  /**
   * 
   * --------------------------------------------------------------------------------
   LocationPathPattern  ::=  '/' RelativePathPattern?   
   | IdKeyPattern (('/' | '//') RelativePathPattern)?   
   | '//'? RelativePathPattern   
   * --------------------------------------------------------------------------------
   */
  protected boolean LocationPathPattern(Node context, boolean isFromAncestorsOrSelf, 
                                        int patMapIndex) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("LocationPathPattern("+context+")");
    boolean test = true;
    boolean testRoot = false;
    // boolean foundEnd = false;
    
    if(tokenIs('/'))
    {      
      nextToken();
      if(tokenIs('/'))
      {
        nextToken();
      }
      testRoot = true;
    }
    
    do
    {
      context = RelativePathPattern(context, patMapIndex, testRoot);
      
      if(null == context)
      {
        test = false;
      }          
    }
      while(isFromAncestorsOrSelf);
    
    return test;
  }

  /**
   * 
   * --------------------------------------------------------------------------------
   IdKeyPattern  ::=  'id' '(' Literal ')'   
   | 'key' '(' Literal ',' Literal ')'   
   * (Also handle doc())
   * --------------------------------------------------------------------------------
   */
  protected boolean IdKeyPattern(Node context) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("IdKeyPattern("+context+")");

    NodeListImpl nl = (NodeListImpl)FunctionCall(context, CLASS_NODESET);
    boolean foundMatch = false;
    int nNodes = nl.getLength();
    for(int i = 0; i < nNodes; i++)
    {
      if(context == nl.item(i))
      {
        foundMatch = true;
        break;
      }
    }
    
    return foundMatch;
  }

  /**
   * 
   * --------------------------------------------------------------------------------
   RelativePathPattern  ::=  StepPattern   
   | RelativePathPattern '/' StepPattern   
   | RelativePathPattern '//' StepPattern   
   * --------------------------------------------------------------------------------
   */
  protected Node RelativePathPattern(Node context, int patMapIndex, boolean testRoot) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("RelativePathPattern("+context+")");
    
    boolean isFromAncestorsOrSelf = false;

    // recurse to next pattern elem
    int nextPatMapIndex = getNextIndexInMap(patMapIndex);
    
    if(nextPatMapIndex > -1)
    {
      m_matchScore = MATCH_SCORE_OTHER;

      // Set the token mark to the current position in the queue
      resetTokenMark(getTokenQueuePosFromMap(nextPatMapIndex)+1);
            
      if(lookbehind( '/', 1 ) && lookbehind( '/', 2 ))
      {        
        isFromAncestorsOrSelf = true;
      }

      context = RelativePathPattern(context, nextPatMapIndex, false);
    }
    
    boolean test = false;
    if(null != context)
    {            
      do
      {
        // Set the token mark to the current position in the queue
        resetTokenMark(getTokenQueuePosFromMap(patMapIndex)+1);

        if(Node.DOCUMENT_NODE == context.getNodeType())
        {
          if(testRoot)
          {
            test = true;
          }
          else
          {
            test = StepPattern( context );
          }
          isFromAncestorsOrSelf = false;
          if(test)
            m_matchScore = MATCH_SCORE_OTHER;
        }
        else
        {
          test = StepPattern( context );
          context = m_xslProcessor.m_parserLiaison.getParentOfNode(context);
        }
        
        // If we had a '//' and a false test, then keep looking up 
        // the the ancestors.
        if(true == isFromAncestorsOrSelf)
        {
          if(!test)
          {
            if(null == context)
              isFromAncestorsOrSelf = false; // end search
          }
          else
          {
            isFromAncestorsOrSelf = false; // end search
          }
        }
      }
        while(isFromAncestorsOrSelf);
    }
    
    if(false == test)
    {
      context = null;
    }
    
    return context;
  }

  /**
   * 
   * --------------------------------------------------------------------------------
   StepPattern  ::=  AbbreviatedNodeTestStep  
   * --------------------------------------------------------------------------------
   */
  protected boolean StepPattern(Node context) 
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("StepPattern("+context+")");
    
    // Set up the state so that a Step(...) search can be 
    // performed in order to determine the position() and count().
    int savedStepMark = m_stepMark;
    m_stepMark = m_queueMark;
    Node savedStepContext = m_stepContext;
    m_stepContext = m_xslProcessor.m_parserLiaison.getParentOfNode(context);
    
    boolean test;
    
    if(lookahead('(', 1) && (tokenIs("id") || tokenIs("key") || tokenIs("doc")))
    {
      // TODO: Need to check to make sure this is the start 
      // of the subpattern.
      test = IdKeyPattern(context);
      if(test)
        m_matchScore = MATCH_SCORE_OTHER;
    }
    else
    {
      // Test the node, including the predicate test.
      test = AbbreviatedNodeTestStep( context );
    }
    
    // ...and restore the Step state.
    m_stepMark = savedStepMark;
    m_stepContext = savedStepContext;
    
    return test;
  }
  
  /**
   WildcardName  ::=  '*'   
   | NCName ':' '*'   
   | QName   
   */
  protected String WildcardName()
  {
    return QName();
  }

  
  /**
   * QName ::=  (Prefix ':')? LocalPart 
   * Prefix ::=  NCName 
   * LocalPart ::=  NCName 
   */
  protected String QName()
  {
    String name = this.m_token;
    if(m_trace) trace("QName = "+name);
    nextToken();
    return name;
  }

  /**
   * NCName ::=  (Letter | '_') (NCNameChar)*
   * NCNameChar ::=  Letter | Digit | '.' | '-' | '_' | CombiningChar | Extender 
   */
  protected String NCName()
  {
    String ncname = this.m_token;
    if(m_trace) trace("NCName = "+ncname);
    nextToken();
    return  ncname;
  }
  
  /**
   * The value of the Literal is the sequence of characters inside 
   * the " or ' characters>.
   * --------------------------------------------------------------------------------
   Literal  ::=  '"' [^"]* '"'   
   | "'" [^']* "'"   
   * --------------------------------------------------------------------------------
   */
  protected String Literal()
  {
    String lit;
    int last = m_token.length()-1;
    char c0 = m_tokenChar;
    char cX = m_token.charAt(last);
    if(((c0 == '\"') && (cX == '\"')) ||
       ((c0 == '\'') && (cX == '\'')))
    {
      lit = m_token.substring(1, last);
      nextToken();
    }
    else
    {
      lit = null; // to suppress compile error
      error("Pattern literal ("+m_token+") needs to be quoted!");
    }
    if(m_trace) trace("Literal = "+lit);
    return lit;
  }
  
  /**
   * --------------------------------------------------------------------------------
   * Number ::= [0-9]+('.'[0-9]+)? | '.'[0-9]+
   * --------------------------------------------------------------------------------
   */
  protected Double Number()
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(m_trace) trace("Number()");
    Double result = null;
    if(null != m_token)
    {
      try
      {
        result = Double.valueOf(m_token);
        nextToken();
      }
      catch(NumberFormatException nfe)
      {
        // I don't think this should be flagged as an error, 
        // though it could be depending on the context.
      }
    }
    if(m_trace) trace("Number = "+result);
    return result;
  }

  // ============= INDEXING FUNCTIONS ================= 
  private void ____INDEXING_FUNCTIONS____(){}
    
  /**
   * Get the position in the current context node list.
   */
  int getPositionInContextNodeList(Node context)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    // assert(null != m_contextNodeList, "m_contextNodeList must be non-null");
    
    if(m_throwFoundIndex)
      throw new FoundIndex();
    
    int pos = -1;

    if(null != m_contextNodeList)
    {
      int nNodes = m_contextNodeList.getLength();
      
      for(int i = 0; i < nNodes; i++)
      {
        if(m_contextNodeList.item(i) == context)
        {
          pos = i+1; // for 1-based XSL count.
          break;
        }
      }
    }
    return pos;
  }
  
  /**
   * Get the position in the current context node list.
   */
  int getCountOfContextNodeList()
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    assert(null != m_contextNodeList, "m_contextNodeList must be non-null");
    
    if(m_throwFoundIndex)
      throw new FoundIndex();
    
    return m_contextNodeList.getLength();
  }
  

  // ============= PRIVATE SUPPORT FUNCTIONS ================= 
  private void ____PRIVATE_SUPPORT_FUNCTIONS____(){}
  
  /**
   * Figure out if child2 is after child1 in document order.
   * @param parent Must be the parent of child1 and child2.
   * @param child1 Must be the child of parent and not equal to child2.
   * @param child2 Must be the child of parent and not equal to child1.
   * @returns true if child 2 is after child1 in document order.
   */
  private boolean isNodeAfterSibling(
                                     Node parent, Node child1, Node child2)
  {
    boolean isNodeAfterSibling = false;
    int child1type = child1.getNodeType();
    int child2type = child2.getNodeType();
    if((Node.ATTRIBUTE_NODE != child1type) && (Node.ATTRIBUTE_NODE == child2type))
    {
      // always sort attributes before non-attributes.
      isNodeAfterSibling = false;
    }
    else if((Node.ATTRIBUTE_NODE == child1type) && (Node.ATTRIBUTE_NODE != child2type))
    {
      // always sort attributes before non-attributes.
      isNodeAfterSibling = true;
    }
    else if(Node.ATTRIBUTE_NODE == child1type)
    {
      NamedNodeMap children = parent.getAttributes();
      int nNodes = children.getLength();
      boolean found1 = false, found2 = false;
      for(int i = 0; i < nNodes; i++)
      {
        Node child = children.item(i);
        if(child1 == child)
        {
          if(found2)
          {
            isNodeAfterSibling = false;
            break;
          }
          found1 = true;
        }
        else if(child2 == child)
        {
          if(found1)
          {
            isNodeAfterSibling = true;
            break;
          }
          found2 = true;
        }
      }
    }
    else
    {
      NodeList children = parent.getChildNodes();
      int nNodes = children.getLength();
      boolean found1 = false, found2 = false;
      for(int i = 0; i < nNodes; i++)
      {
        Node child = children.item(i);
        if(child1 == child)
        {
          if(found2)
          {
            isNodeAfterSibling = false;
            break;
          }
          found1 = true;
        }
        else if(child2 == child)
        {
          if(found1)
          {
            isNodeAfterSibling = true;
            break;
          }
          found2 = true;
        }
      }
    }
    return isNodeAfterSibling;
  } // end isNodeAfterSibling(Node parent, Node child1, Node child2)

  
  /**
   * Figure out if node2 should be placed after node1 when 
   * placing nodes in a list that is to be sorted in 
   * document order.  Assumes that node1 and node2 are not
   * equal.
   * NOTE: Make sure this does the right thing with attribute nodes!!!
   * @return true if node2 should be placed 
   * after node1, and false if node2 should be placed 
   * before node1.
   */
  public boolean isNodeAfter(Node node1, Node node2)
  {
    boolean isNodeAfter = false; // return value.
    
    Node parent1 = m_xslProcessor.m_parserLiaison.getParentOfNode(node1);
    Node parent2 = m_xslProcessor.m_parserLiaison.getParentOfNode(node2);
    
    // Optimize for most common case
    if(parent1 == parent2) // then we know they are siblings
    {
      isNodeAfter = isNodeAfterSibling(
                                       parent1, node1, node2);
    }
    else
    {
      // General strategy: Figure out the lengths of the two 
      // ancestor chains, and walk up them looking for the 
      // first common ancestor, at which point we can do a 
      // sibling compare.  Edge condition where one is the 
      // ancestor of the other.
      
      // Count parents, so we can see if one of the chains 
      // needs to be equalized.
      int nParents1 = 2, nParents2 = 2; // count node & parent obtained above
      while(parent1 != null)
      {
        nParents1++;
        parent1 = m_xslProcessor.m_parserLiaison.getParentOfNode(parent1);
      }
      
      while(parent2 != null)
      {
        nParents2++;
        parent2 = m_xslProcessor.m_parserLiaison.getParentOfNode(parent2);
      }
      
      Node startNode1 = node1, startNode2 = node2; // adjustable starting points
      
      // Do I have to adjust the start point in one of 
      // the ancesor chains?
      if(nParents1 < nParents2)
      {
        // adjust startNode2
        int adjust = nParents2 - nParents1;
        for(int i = 0; i < adjust; i++)
        {
          startNode2 = m_xslProcessor.m_parserLiaison.getParentOfNode(startNode2);
        }
      }
      else if(nParents1 > nParents2)
      {
        // adjust startNode1
        int adjust = nParents1 - nParents2;
        for(int i = 0; i < adjust; i++)
        {
          startNode1 = m_xslProcessor.m_parserLiaison.getParentOfNode(startNode1);
        }
      }
      
      Node prevChild1 = null, prevChild2 = null; // so we can "back up"
      
      // Loop up the ancestor chain looking for common parent.
      while(null != startNode1)
      {
        if(startNode1 == startNode2) // common parent?
        {
          if(null == prevChild1) // first time in loop?
          {
            // Edge condition: one is the ancestor of the other.
            isNodeAfter = (nParents1 < nParents2) ? true : false;
            break; // from while loop
          }
          else
          {
            isNodeAfter = isNodeAfterSibling(
                                             startNode1, prevChild1, prevChild2);
            break; // from while loop
          }
        } // end if(startNode1 == startNode2)
        prevChild1 = startNode1;
        startNode1 = m_xslProcessor.m_parserLiaison.getParentOfNode(startNode1);
        prevChild2 = startNode2;
        startNode2 = m_xslProcessor.m_parserLiaison.getParentOfNode(startNode2);
      } // end while
    } // end big else
    
    /* -- please do not remove... very useful for diagnostics --
    System.out.println("node1 = "+node1.getNodeName()+"("+node1.getNodeType()+")"+
    ", node2 = "+node2.getNodeName()
    +"("+node2.getNodeType()+")"+
    ", isNodeAfter = "+isNodeAfter); */
    
    return isNodeAfter;
  } // end isNodeAfter(Node node1, Node node2)
  
  /**
   * Only add the node if it is qualified.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  void addNodeInDocOrderIfQualified(Vector v, Node node)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(Predicate(node))
    {
      addNodeInDocOrder(v, node);
    }
  }

  void addNodeIfQualified(Vector v, Node node)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException
  {
    if(Predicate(node))
    {
      v.addElement(node);
    }
  }
  
  /**
   * Add the node into a vector of nodes where it should occur in 
   * document order.  Since isNodeAfter is pretty expensive, 
   * I'll probably make this function smarter down the line.
   * @param v Vector of nodes, presumably containing Nodes
   * @param obj Node object.
   */
  void addNodeInDocOrder(Vector v, Node node, boolean test)
  {
    if(test)
    {
      int size = v.size(), i;
      for(i = 0; i < size; i++)
      {
        Node child = (Node)v.elementAt(i);
        if(child == node)
        {
          i = -1; // Duplicate, suppress insert
          break; 
        }
        if(isNodeAfter(node, child))
        {
          break;
        }
      }
      if(i >= 0)
      {
        v.insertElementAt(node, i);
      }
    }
    else
    {
      v.addElement(node);
    }
  } // end addNodeInDocOrder(Vector v, Object obj)


  /**
   * Add the node into a vector of nodes where it should occur in 
   * document order.  Since isNodeAfter is pretty expensive, 
   * I'll probably make this function smarter down the line.
   * @param v Vector of nodes, presumably containing Nodes
   * @param obj Node object.
   */
  void addNodeInDocOrder(Vector v, Node node)
  {
    addNodeInDocOrder(v, node, this.m_testForDocOrder);
  } // end addNodeInDocOrder(Vector v, Object obj)
  
  /**
   * Get a document identified by a URL, parsing it if necessary.
   */
  public Document getDocByURL(Node context, String url)
    throws XSLProcessorException, java.net.MalformedURLException, 
    java.io.FileNotFoundException, java.io.IOException
  {
    Document doc;
    if(url.trim().length() == 0)
    {
      doc = context.getOwnerDocument();
    }
    else
    {
      doc = (Document)m_xslProcessor.getSourceDocsTable().get(url);
      if(null == doc)
      {
        doc = m_xslProcessor.parseXML(url, 0);
        if(null == doc)
        {
          warn("Can not load uri: "+url);
        }
        else
        {
          m_xslProcessor.getSourceDocsTable().put(url, doc);
        }
      }
    }
    return doc;
  }
  
  /**
   * Get a document identified by a URL, parsing it if necessary.
   */
  public Element getElementByID(Node context, String id)
    throws XSLProcessorException, java.net.MalformedURLException, 
    java.io.FileNotFoundException, java.io.IOException
  {
    Document ownerDoc = context.getOwnerDocument();
    if(null == ownerDoc)
    {
      if(Node.DOCUMENT_NODE == context.getNodeType())
      {
        ownerDoc = (Document)context;
      }
      else
      {
        error(context, "Child does not have an owner document!");
      }
    }
    return m_xslProcessor.getElementByID(id, ownerDoc);
  }
  
  /**
   * 
   * --------------------------------------------------------------------------------
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected Vector idExpr(int exprType, Node context, Vector queryResults) 
    throws XSLProcessorException, java.net.MalformedURLException, java.io.FileNotFoundException, java.io.IOException
  {
    String name;
    if(FROM_ID == exprType)
    {
      // String ids = StringExpr(context);
      String ids = "";
      consumeExpected(')'); // We'll have to do the same thing as with AncestorExpr
      int n = ids.length();
      int start = -1;
      for(int k = 0; k <= n; k++ )
      {
        if((k == n) || Character.isWhitespace(ids.charAt(k)))
        {
          if(start > -1)
          {
            String id = ((start == 0) && (k==n)) ? ids : ids.substring(start, k);
            context = getElementByID(context, id);
            start = -1;
          }
        }
        else
        {
          if(-1 == start)
            start = k;
        }
      }
    }
    else
    {
      Vector idNodes = LocationPath(context, new NodeListImpl());
      consumeExpected(')'); // We'll have to do the same thing as with AncestorExpr
      int nIDNodes = idNodes.size();
      for(int i = 0; i < nIDNodes; i++)
      {
        Node node = (Node)idNodes.elementAt(i);
        String ids = m_xslProcessor.getNodeData(node);
        int n = ids.length();
        int start = -1;
        for(int k = 0; k <= n; k++ )
        {
          if((k == n) || Character.isWhitespace(ids.charAt(k)))
          {
            if(start > -1)
            {
              String id = ((start == 0) && (k==n)) ? ids : ids.substring(start, k);
              context = getElementByID(context, id);
              start = -1;
            }
          }
          else
          {
            if(-1 == start)
              start = k;
          }
        }
      }
    }
    return queryResults;
  }
  

  // ============= TYPE HANDLINING =================
  private void ____TYPE_HANDLING____(){}

  /**
   * Given an expression object, return what type of class it is...
   */
  public int getClassType(Object obj)
  {
    int classType;
    if(null != obj)
    {
      if(obj instanceof Boolean)
      {
        classType = CLASS_BOOLEAN;
      }
      else if(obj instanceof DocumentFragment)
      {
        // Assume CLASS_RTREEFRAG for the sake of speed.
        classType = CLASS_RTREEFRAG;
      }
      else if(obj instanceof NodeList)
      {
        classType = CLASS_NODESET;
      }
      else if(obj instanceof String)
      {
        classType = CLASS_STRING;
      }
      else if(obj instanceof Double)
      {
        classType = CLASS_NUMBER;
      }
      else // if(obj instanceof DocumentFragment)
      {
        // Assume CLASS_RTREEFRAG for the sake of speed.
        classType = CLASS_RTREEFRAG;
      }
    }
    else
    {
      classType = CLASS_NULL;
    }
    return classType;
  }

  
  /**
   * Cast result object to a number.
   */
  public double numberCast(Object obj)
  {
    int t = getClassType(obj);
    double result;
    if(null != obj)
    {
      switch(t)
      {
      case CLASS_NODESET:
        {
          Vector v = ((Vector)obj);
          int nNodes = v.size();
          if(nNodes > 0)
          {
            String s = m_xslProcessor.getNodeData((Node)v.elementAt(0));
            if(null == s) 
              result = 0.0;
            else
              result = Double.valueOf(s).doubleValue();
          }
          else
          {
            result = 0.0;
          }
        }
        break;
      case CLASS_BOOLEAN:
        {
          result = ((Boolean)obj).booleanValue() ? 1.0 : 0.0;
        }
        break;
      case CLASS_NUMBER:
        {
          result = ((Double)obj).doubleValue();
        }
        break;
      case CLASS_STRING:
        {
          String s = (String)obj;
          if(null != s)
          {
            try
            {
              result = Double.valueOf(s).doubleValue();
            }
            catch(NumberFormatException nfe)
            {
              result = 0.0;
            }
          }
          else
          {
            result = 0.0;
          }
        }
        break;
      case CLASS_RTREEFRAG:
        {
          String s = m_xslProcessor.getNodeData((DocumentFragment)obj);
          if(null != s)
          {
            try
            {
              result = Double.valueOf(s).doubleValue();
            }
            catch(NumberFormatException nfe)
            {
              result = 0.0;
            }
          }
          else
          {
            result = 0.0;
          }
        }
       break;
      default:
        error("Can not convert "+obj.getClass()+" to a number");
        result = 0.0;
      }
    }
    else
    {
      result = 0.0;
    }

    return result;
  }

  /**
   * Cast result object to a boolean.
   */
  public boolean booleanCast(Object obj)
  {
    int t = getClassType(obj);
    boolean result;
    if(null != obj)
    {
      switch(t)
      {
      case CLASS_NODESET:
        {
          Vector v = ((Vector)obj);
          int nNodes = v.size();
          result = (nNodes > 0);
        }
        break;
      case CLASS_BOOLEAN:
        {
          result = ((Boolean)obj).booleanValue();
        }
        break;
      case CLASS_NUMBER:
        {
          result = ((Double)obj).doubleValue() != 0.0;
        }
        break;
      case CLASS_STRING:
        {
          // What about "true" and "false"?
          result = ((String)obj).length() > 0;
        }
        break;
      case CLASS_RTREEFRAG:
        {
          NodeList nl = ((DocumentFragment)obj).getChildNodes();
          int nChildren = nl.getLength();
          result = false;
          for(int i = 0; i < nChildren; i++)
          {
            Node n = nl.item(i);
            if((Node.TEXT_NODE == n.getNodeType()) &&
               (m_xslProcessor.m_parserLiaison.isIgnorableWhitespace((Text)n) ||
                (((Text)n).getData().trim().length() == 0)))
            {
              continue;
            }
            
            result = true;
            break;
          }
        }
        break;
      default:
        error("Can not convert "+obj.getClass()+" to a boolean");
        result = false;
      }
    }
    else
    {
      result = false;
    }

    return result;
  }

  /**
   * Cast result object to a string.
   */
  public String stringCast(Object obj)
  {
    int t = getClassType(obj);
    String result;
    if(null != obj)
    {
      switch(t)
      {
      case CLASS_NODESET:
        {
          Vector v = ((Vector)obj);
          int nNodes = v.size();
          if(nNodes > 0)
          {
            result = m_xslProcessor.getNodeData((Node)v.elementAt(0));
            if(null == result) 
              result = "";
          }
          else
          {
            result = "";
          }
        }
        break;
      case CLASS_BOOLEAN:
        {
          result = ((Boolean)obj).booleanValue() ? "true" : "false";
        }
        break;
      case CLASS_NUMBER:
        {
          // result = String.valueOf((Double)result);
          
          java.text.NumberFormat m_formatter 
            = java.text.NumberFormat.getNumberInstance();
          result = m_formatter.format((Double)obj);
        }
        break;
      case CLASS_STRING:
        {
          result = (String)obj;
        }
        break;
      case CLASS_RTREEFRAG:
        {
          result = m_xslProcessor.getNodeData((DocumentFragment)obj);
        }
        break;
      default:
        error("Can not convert "+obj.getClass()+" to a string");
        result = "";
      }
    }
    else
    {
      result = "";
    }
    return result;
  }
  
  /**
   * Cast result object to a result tree fragment.
   */
  public DocumentFragment rtreefragCast(Object obj)
  {
    int t = getClassType(obj);
    DocumentFragment result;
    switch(t)
    {
    case CLASS_NODESET:
      {
        warn("Not sure if you can cast a node-set to a result tree fragment");
        DocumentFragment df = m_xslProcessor.createDocFrag();
        Vector v = (Vector)obj;
        int nNodes = v.size();
        for(int i = 0; i < nNodes; i++)
        {
          df.appendChild(((Node)v.elementAt(i)).cloneNode(true));
        }
        result = df;
      }
      break;
    case CLASS_BOOLEAN:
      {
        DocumentFragment df = m_xslProcessor.createDocFrag();
        // java.text.NumberFormat m_formatter 
        //  = java.text.NumberFormat.getNumberInstance();
        String s = ((Boolean)obj).booleanValue() ? "true" : "false";
        Text textNode = m_xslProcessor.m_resultTreeFactory.createTextNode(s);
        df.appendChild(textNode);
        result = df;
      }
      break;
    case CLASS_NUMBER:
      {
        DocumentFragment df = m_xslProcessor.createDocFrag();
        java.text.NumberFormat m_formatter 
          = java.text.NumberFormat.getNumberInstance();
        String s = m_formatter.format((Double)obj);
        Text textNode = m_xslProcessor.m_resultTreeFactory.createTextNode(s);
        df.appendChild(textNode);
        result = df;
      }
      break;
    case CLASS_STRING:
      {
        DocumentFragment df = m_xslProcessor.createDocFrag();
        // java.text.NumberFormat m_formatter 
        //  = java.text.NumberFormat.getNumberInstance();
        Text textNode = m_xslProcessor.m_resultTreeFactory.createTextNode((String)obj);
        df.appendChild(textNode);
        result = df;
      }
      break;
    case CLASS_RTREEFRAG:
      {
        result = (DocumentFragment)obj;
      }
      break;
    default:
      error("Can not convert "+obj.getClass()+" to a result tree fragment.");
      result = null;
    }
    return result;
  }



  /**
   * Cast result object to a nodelist.
   */
  public Vector nodelistCast(Object obj)
  {
    int t = getClassType(obj);
    if(CLASS_NODESET != t)
    {
      error("Can not convert "+obj.getClass()+" to a vector!");
    }
    return (Vector)obj;
  }
  
  /**
   * Cast object to type t.
   */
  public Object castToType(Object obj, int t)
  {
    Object result;
    switch(t)
    {
    case CLASS_STRING:
      result = stringCast(obj);
      break;
    case CLASS_NUMBER:
      result = new Double(numberCast(obj));
      break;
    case CLASS_NODESET:
      result = nodelistCast(obj);
      break;
    case CLASS_BOOLEAN:
      result = new Boolean( booleanCast(obj) );
      break;
    case CLASS_DONTCARE:
      result = obj;
      break;
    case CLASS_RTREEFRAG:
      result = rtreefragCast(obj);
      break;
    default:
      error("Can not convert "+obj.getClass()+" to a type#"+t);
      result = null;
    }
    return result;
  }
  

  /**
   * Given a request type, return the equivalent string. 
   * For diagnostic purposes.
   */
  private String getTypeString(int t)
  {
    String result;
    switch(t)
    {
    case CLASS_STRING:
      result = "String";
      break;
    case CLASS_NUMBER:
      result = "Number";
      break;
    case CLASS_NODESET:
      result = "Node-set";
      break;
    case CLASS_BOOLEAN:
      result = "Boolean";
      break;
    case CLASS_RTREEFRAG:
      result = "Result-tree-fragment";
      break;
    default:
      result = "Unknown";
    }
    return result;
  }
  
  
  /**
   * Given an empty type, return an appropriate 
   * value for a predicate failure.
   */
  private Object emptyType(int t)
  {
    Object result;
    switch(t)
    {
    case CLASS_STRING:
      result = "";
      break;
    case CLASS_NUMBER:
      result = new Double(0.0);
      break;
    case CLASS_NODESET:
      result = new NodeListImpl();
      break;
    case CLASS_BOOLEAN:
      result = new Boolean(false);
      break;
    case CLASS_RTREEFRAG:
      result = m_xslProcessor.createDocFrag();
      break;
    default:
      result = null;
    }
    return result;
  }

  // ============= DIAGNOSTIC & ERROR HANDLINING =================
  private void ____DIAGNOSTICS_AND_ERRORS____(){}

  /**
   * If this is true, extra programmer error checks will be made.
   */
  static public final boolean m_debug = false;
  
  /**
   * If m_trace is set to true, trace strings will be written 
   * out to System.out.
   */
  static public final boolean m_trace = false;
  
  private final void trace(String s)
  {
    System.out.println(s);
  }

  /**
   * Warn the user of an problem.
   */
  public void warn(String msg)
    throws QueryEngineException
  {
    warn(null, msg);
  }

  /**
   * Warn the user of an problem.
   */
  public void warn(Node sourceNode, String msg)
    throws QueryEngineException
  {
    boolean shouldThrow = m_xslProcessor.getProblemListener().problem(ProblemListener.QUERYENGINE, 
                                                                      ProblemListener.WARNING,
                                                                      m_namespaceContext, 
                                                                      sourceNode, msg, 0, 0);
    if(shouldThrow)
    {
      throw new QueryEngineException(msg);
    }
  }

  /**
   * Tell the user of an assertion error, and probably throw an 
   * exception.
   */
  private void assert(boolean b, String msg)
    throws QueryEngineException
  {
    if(!b)
      error(null, "Programmer assertion is incorrect! - "+msg);
  }

  /**
   * Tell the user of an error, and probably throw an 
   * exception.
   */
  public void error(String msg)
    throws QueryEngineException
  {
    error(null, msg);
  }

  /**
   * Tell the user of an error, and probably throw an 
   * exception.
   */
  public void error(Node sourceNode, String msg)
    throws QueryEngineException
  {
    String emsg = ((null != m_currentPattern) 
                   ? ("pattern = '"+m_currentPattern+"'\n") : "") +
                  msg + dumpRemainingTokenQueue();

    boolean shouldThrow = m_xslProcessor.getProblemListener().problem(ProblemListener.QUERYENGINE, 
                                                                      ProblemListener.ERROR,
                                                                      m_namespaceContext, 
                                                                      sourceNode, emsg, 0, 0);
    if(shouldThrow)
    {
      throw new QueryEngineException(msg);
    }
  }
  
  /**
   * Dump the remaining token queue.
   * Thanks be to Craig for this.
   */
  protected String dumpRemainingTokenQueue()
  {
    int q = m_queueMark;
    String returnMsg;
    if(q < m_tokenQueueSize)
    {
      String msg = "\n Remaining tokens: (";
      while (q < m_tokenQueueSize ) 
      {
        String t = m_tokenQueue[q++];
        msg += (" '" + t + "'");
      }
      returnMsg = msg + ")";
    }
    else
    {
      returnMsg = "";
    }
    return returnMsg;
  }

  
  // ============= STATE VARIABLES =================
  private void ____STATE_VARIABLES____(){}
  
  /**
   * The context in which namespaces in the queries are supposed 
   * to be expanded.
   */
  Element m_namespaceContext = null; 
  
  /**
  * Save the token queues so the tokenization doesn't have to 
  * be redone each time.  This helps performance some, though 
  * not a lot.  If memory footprint is more important than 
  * performance, this is a candidate for removal.
  */
  // Hashtable m_tokenQueues = new Hashtable();
  
  /**
   *  TokenStack is the queue of used tokens. The current token is the token at the 
   * end of the m_tokenQueue. The idea is that the queue can be marked and a sequence
   * of tokens can be reused.
   */
  private String[] m_tokenQueue;
  
  private int m_tokenQueueSize;
  
  private static final int MAXTOKENQUEUESIZE = 200;
  
  private final void addToTokenQueue(String s)
  {
    m_tokenQueue[m_tokenQueueSize++] = s;
  }
  
  /**
   * This holds a map to the m_tokenQueue that tells where the top-level elements are.
   * It is used for pattern matching so the m_tokenQueue can be walked backwards.
   * Each element that is a 'target', (right-most top level element name) has 
   * TARGETEXTRA added to it.
   */
  private int m_patternMap[];
  
  /**
   * The number of elements in m_patternMap;
   */
  private int m_patternMapSize = 0;
  
  /**
   * This value is added to each element name in the TARGETEXTRA 
   * that is a 'target', (right-most top level element name).
   */
  static final int TARGETEXTRA = 10000;

  /**
   * The next token in the pattern.
   */
  private String m_token;
  
  /**
   * The first char in m_token, the theory being that this 
   * is an optimization because we won't have to do charAt(0) as
   * often.
   */
  private char m_tokenChar = 0;

  /**
   * The current pattern string, for diagnostics purposes
   */
  String m_currentPattern = null;

  /**
   * The position in the token queue is tracked by m_queueMark.
   */
  private int m_queueMark = 0;
  
  /**
   * The position of the current step.
   */
  private int m_stepMark = -1;
  
  /**
   * The current step node.  This and m_stepMark can be 
   * used to call Step(...) in order to arrive at the 
   * current context list.
   */
  private Node m_stepContext = null;

  /**
   * The current step node.  This and m_stepMark can be 
   * used to call Step(...) in order to arrive at the 
   * current context list.
   */
  private NodeList m_contextNodeList = null;

  /**
   * This tells whether the query needs to use a method 
   * that will add elements in document order.
   */
  private boolean m_testForDocOrder = false;
  
  /**
   * This tells us if we are in a state where we 
   * only want to collect target strings.
   */
  private boolean m_collectTargetStrings = false;
  
  /**
   * The match score if no match is made.
   */
  public static final int MATCH_SCORE_NONE = -1000;

  /**
   * The match score if the pattern has the form 
   * of a QName optionally preceded by an @ character.
   */
  public static final int MATCH_SCORE_QNAME = 0;

  /**
   * The match score if the pattern consists of just a NodeTest.
   */
  public static final int MATCH_SCORE_NODETEST = -1;

  /**
   * The match score if the pattern consists of something 
   * other than just a NodeTest or just a qname.
   */
  public static final int MATCH_SCORE_OTHER = 1;

  /**
   * A value to keep track of the specificity of a match pattern.
   */
  private int m_matchScore = MATCH_SCORE_NONE;

  /**
   * At least for now, we need access to a few 
   * XSLProcessor functions, like getNodeData, and 
   * getIDOfElement.
   */
  protected XSLProcessor m_xslProcessor;

  // ============= CONSTANTS =================
  private void ____CONSTANTS____(){}
  
  public static final int CLASS_NULL = -1;
  public static final int CLASS_DONTCARE = 0;
  public static final int CLASS_BOOLEAN = 1;
  public static final int CLASS_NUMBER = 2;
  public static final int CLASS_STRING = 3;
  public static final int CLASS_NODESET = 4;
  public static final int CLASS_RTREEFRAG = 5;

  static final int QUALIFIER_FIRST_OF_TYPE = 1;
  static final int QUALIFIER_NOT_FIRST_OF_TYPE = 2;
  static final int QUALIFIER_FIRST_OF_ANY = 3;
  static final int QUALIFIER_NOT_FIRST_OF_ANY = 4;
  static final int QUALIFIER_LAST_OF_TYPE = 5;
  static final int QUALIFIER_NOT_LAST_OF_TYPE = 6;
  static final int QUALIFIER_LAST_OF_ANY = 7;
  static final int QUALIFIER_NOT_LAST_OF_ANY = 8;
  static final int QUALIFIER_ONLY_OF_TYPE = 9;
  static final int QUALIFIER_NOT_ONLY_OF_TYPE = 10;
  static final int QUALIFIER_ONLY_OF_ANY = 11;
  static final int QUALIFIER_NOT_ONLY_OF_ANY = 12;
  
  static final void _______KEYWORDS________(){};
  
  static Hashtable m_keywords = new Hashtable();
  static Hashtable m_functions = new Hashtable();
  
  static final int FROM_ANCESTORS = 1;
  static final String FROM_ANCESTORS_STRING = "from-ancestors";
  
  static final int FROM_ANCESTORS_OR_SELF = 2;
  static final String FROM_ANCESTORS_OR_SELF_STRING = "from-ancestors-or-self";
  
  static final int FROM_ATTRIBUTES = 3;
  static final String FROM_ATTRIBUTES_STRING = "from-attributes";
  
  static final int FROM_CHILDREN = 4;
  static final String FROM_CHILDREN_STRING = "from-children";
  
  static final int FROM_DESCENDANTS = 5;
  static final String FROM_DESCENDANTS_STRING = "from-descendants";
  
  static final int NODETYPE_ROOT = 6;
  static final int FROM_DESCENDANTS_OR_SELF = 6;
  static final String FROM_DESCENDANTS_OR_SELF_STRING = "from-descendants-or-self";
  static final String FROM_DESCENDANTS_OR_SELF_ABBREVIATED_STRING = "/";
  
  static final int FROM_FOLLOWING = 7;
  static final String FROM_FOLLOWING_STRING = "from-following";
  
  static final int FROM_FOLLOWING_SIBLINGS = 8; 
  static final String FROM_FOLLOWING_SIBLINGS_STRING = "from-following-siblings";
  
  static final int FROM_PARENT = 9;
  static final String FROM_PARENT_STRING = "from-parent";
  static final String FROM_PARENT_ABBREVIATED_STRING = "..";
  
  static final int FROM_PRECEDING = 10;
  static final String FROM_PRECEDING_STRING = "from-preceding";
  
  static final int FROM_PRECEDING_SIBLINGS = 11;
  static final String FROM_PRECEDING_SIBLINGS_STRING = "from-preceding-siblings";
  
  static final int FROM_SELF = 12;
  static final String FROM_SELF_STRING = "from-self";
  static final String FROM_SELF_ABBREVIATED_STRING = ".";

  static final int NODETYPE_COMMENT = 13;
  static final String NODETYPE_COMMENT_STRING = "comment";

  static final int NODETYPE_TEXT = 14;
  static final String NODETYPE_TEXT_STRING = "text";

  static final int NODETYPE_PI = 15;
  static final String NODETYPE_PI_STRING = "pi";

  static final int NODETYPE_NODE = 16;
  static final String NODETYPE_NODE_STRING = "node";
  
  static final int FROM_ATTRIBUTE = 17;
  static final String FROM_ATTRIBUTE_STRING = "@";

  static final int FROM_DOC = 18;
  static final String FROM_DOC_STRING = "doc";

  static final int FROM_DOCREF = 19;
  static final String FROM_DOCREF_STRING = "docref";

  static final int FROM_ID = 20;
  static final String FROM_ID_STRING = "id";

  static final int FROM_IDREF = 21;
  static final String FROM_IDREF_STRING = "idref";

  static final int NODETYPE_ANYELEMENT = 23;
  static final String NODETYPE_ANYELEMENT_STRING = "*";
  
  static final int FUNC_LAST = 24;
  static final String FUNC_LAST_STRING = "last";

  static final int FUNC_POSITION = 25;
  static final String FUNC_POSITION_STRING = "position";

  static final int FUNC_COUNT = 26;
  static final String FUNC_COUNT_STRING = "count";

  static final int FUNC_ID = 27;
  static final String FUNC_ID_STRING = "id";

  static final int FUNC_IDREF = 28;
  static final String FUNC_IDREF_STRING = "idref";

  static final int FUNC_KEY = 29;
  static final String FUNC_KEY_STRING = "key";

  static final int FUNC_KEYREF = 30;
  static final String FUNC_KEYREF_STRING = "keyref";

  static final int FUNC_DOC = 31;
  static final String FUNC_DOC_STRING = "doc";

  static final int FUNC_DOCREF = 32;
  static final String FUNC_DOCREF_STRING = "docref";

  static final int FUNC_LOCAL_PART = 33;
  static final String FUNC_LOCAL_PART_STRING = "local-part";

  static final int FUNC_NAMESPACE = 34;
  static final String FUNC_NAMESPACE_STRING = "namespace";

  static final int FUNC_QNAME = 35;
  static final String FUNC_QNAME_STRING = "qname";

  static final int FUNC_GENERATE_ID = 36;
  static final String FUNC_GENERATE_ID_STRING = "generate-id";

  static final int FUNC_NOT = 37;
  static final String FUNC_NOT_STRING = "not";

  static final int FUNC_TRUE = 38;
  static final String FUNC_TRUE_STRING = "true";

  static final int FUNC_FALSE = 39;
  static final String FUNC_FALSE_STRING = "false";

  static final int FUNC_BOOLEAN = 40;
  static final String FUNC_BOOLEAN_STRING = "boolean";

  static final int FUNC_LANG = 41;
  static final String FUNC_LANG_STRING = "lang";

  static final int FUNC_NUMBER = 42;
  static final String FUNC_NUMBER_STRING = "number";

  static final int FUNC_FLOOR = 43;
  static final String FUNC_FLOOR_STRING = "floor";

  static final int FUNC_CEILING = 44;
  static final String FUNC_CEILING_STRING = "ceiling";

  static final int FUNC_ROUND = 45;
  static final String FUNC_ROUND_STRING = "round";

  static final int FUNC_SUM = 46;
  static final String FUNC_SUM_STRING = "sum";

  static final int FUNC_STRING = 47;
  static final String FUNC_STRING_STRING = "string";

  static final int FUNC_STARTS_WITH = 48;
  static final String FUNC_STARTS_WITH_STRING = "starts-with";

  static final int FUNC_CONTAINS = 49;
  static final String FUNC_CONTAINS_STRING = "contains";

  static final int FUNC_SUBSTRING_BEFORE = 50;
  static final String FUNC_SUBSTRING_BEFORE_STRING = "substring-before";

  static final int FUNC_SUBSTRING_AFTER = 51;
  static final String FUNC_SUBSTRING_AFTER_STRING = "substring-after";

  static final int FUNC_NORMALIZE = 52;
  static final String FUNC_NORMALIZE_STRING = "normalize";

  static final int FUNC_TRANSLATE = 53;
  static final String FUNC_TRANSLATE_STRING = "translate";

  static final int FUNC_CONCAT = 54;
  static final String FUNC_CONCAT_STRING = "concat";

  static final int FUNC_FORMAT_NUMBER = 55;
  static final String FUNC_FORMAT_NUMBER_STRING = "format-number";

  static final int FUNC_SYSTEM_PROPERTY = 56;
  static final String FUNC_SYSTEM_PROPERTY_STRING = "system-property";

  static final int FUNC_FUNCTION_AVAILABLE = 57;
  static final String FUNC_FUNCTION_AVAILABLE_STRING = "function-available";

  static final int FUNC_SUBSTRING = 58;
  static final String FUNC_SUBSTRING_STRING = "substring";

  static final int FUNC_STRING_LENGTH = 59;
  static final String FUNC_STRING_LENGTH_STRING = "string-length";

  static 
  {
    m_keywords.put((FROM_ANCESTORS_STRING).intern(), new Integer(FROM_ANCESTORS));
    m_keywords.put((FROM_ANCESTORS_OR_SELF_STRING).intern(), new Integer(FROM_ANCESTORS_OR_SELF));
    m_keywords.put((FROM_ATTRIBUTES_STRING).intern(), new Integer(FROM_ATTRIBUTES));
    m_keywords.put((FROM_CHILDREN_STRING).intern(), new Integer(FROM_CHILDREN));
    m_keywords.put((FROM_DESCENDANTS_STRING).intern(), new Integer(FROM_DESCENDANTS));
    m_keywords.put((FROM_DESCENDANTS_OR_SELF_STRING).intern(), new Integer(FROM_DESCENDANTS_OR_SELF));
    m_keywords.put((FROM_DESCENDANTS_OR_SELF_ABBREVIATED_STRING).intern(), new Integer(FROM_DESCENDANTS_OR_SELF));
    m_keywords.put((FROM_FOLLOWING_STRING).intern(), new Integer(FROM_FOLLOWING));
    m_keywords.put((FROM_FOLLOWING_SIBLINGS_STRING).intern(), new Integer(FROM_FOLLOWING_SIBLINGS));
    m_keywords.put((FROM_PARENT_STRING).intern(), new Integer(FROM_PARENT));
    m_keywords.put((FROM_PRECEDING_STRING).intern(), new Integer(FROM_PRECEDING));
    m_keywords.put((FROM_PRECEDING_SIBLINGS_STRING).intern(), new Integer(FROM_PRECEDING_SIBLINGS));
    m_keywords.put((FROM_SELF_STRING).intern(), new Integer(FROM_SELF));
    m_keywords.put((NODETYPE_COMMENT_STRING).intern(), new Integer(NODETYPE_COMMENT));
    m_keywords.put((NODETYPE_TEXT_STRING).intern(), new Integer(NODETYPE_TEXT));
    m_keywords.put((NODETYPE_PI_STRING).intern(), new Integer(NODETYPE_PI));
    m_keywords.put((NODETYPE_NODE_STRING).intern(), new Integer(NODETYPE_NODE));
    m_keywords.put((FROM_PARENT_ABBREVIATED_STRING).intern(), new Integer(FROM_PARENT));
    m_keywords.put((FROM_SELF_ABBREVIATED_STRING).intern(), new Integer(FROM_SELF));
    m_keywords.put((FROM_ATTRIBUTE_STRING).intern(), new Integer(FROM_ATTRIBUTE));
    m_keywords.put((FROM_DOC_STRING).intern(), new Integer(FROM_DOC));
    m_keywords.put((FROM_DOCREF_STRING).intern(), new Integer(FROM_DOCREF));
    m_keywords.put((FROM_ID_STRING).intern(), new Integer(FROM_ID));
    m_keywords.put((FROM_IDREF_STRING).intern(), new Integer(FROM_IDREF));
    m_keywords.put((NODETYPE_ANYELEMENT_STRING).intern(), new Integer(NODETYPE_ANYELEMENT));

    m_keywords.put((FUNC_ID_STRING).intern(), new Integer(FUNC_ID));
    m_keywords.put((FUNC_KEY_STRING).intern(), new Integer(FUNC_KEY));
    m_keywords.put((FUNC_DOC_STRING).intern(), new Integer(FUNC_DOC));

    m_functions.put((FUNC_LAST_STRING).intern(), new Integer(FUNC_LAST));
    m_functions.put((FUNC_POSITION_STRING).intern(), new Integer(FUNC_POSITION));
    m_functions.put((FUNC_COUNT_STRING).intern(), new Integer(FUNC_COUNT));
    m_functions.put((FUNC_ID_STRING).intern(), new Integer(FUNC_ID));
    m_functions.put((FUNC_IDREF_STRING).intern(), new Integer(FUNC_IDREF));
    m_functions.put((FUNC_KEY_STRING).intern(), new Integer(FUNC_KEY));
    m_functions.put((FUNC_KEYREF_STRING).intern(), new Integer(FUNC_KEYREF));
    m_functions.put((FUNC_DOC_STRING).intern(), new Integer(FUNC_DOC));
    m_functions.put((FUNC_DOCREF_STRING).intern(), new Integer(FUNC_DOCREF));
    m_functions.put((FUNC_LOCAL_PART_STRING).intern(), new Integer(FUNC_LOCAL_PART));
    m_functions.put((FUNC_NAMESPACE_STRING).intern(), new Integer(FUNC_NAMESPACE));
    m_functions.put((FUNC_QNAME_STRING).intern(), new Integer(FUNC_QNAME));
    m_functions.put((FUNC_GENERATE_ID_STRING).intern(), new Integer(FUNC_GENERATE_ID));
    m_functions.put((FUNC_NOT_STRING).intern(), new Integer(FUNC_NOT));
    m_functions.put((FUNC_TRUE_STRING).intern(), new Integer(FUNC_TRUE));
    m_functions.put((FUNC_FALSE_STRING).intern(), new Integer(FUNC_FALSE));
    m_functions.put((FUNC_BOOLEAN_STRING).intern(), new Integer(FUNC_BOOLEAN));
    m_functions.put((FUNC_LANG_STRING).intern(), new Integer(FUNC_LANG));
    m_functions.put((FUNC_NUMBER_STRING).intern(), new Integer(FUNC_NUMBER));
    m_functions.put((FUNC_FLOOR_STRING).intern(), new Integer(FUNC_FLOOR));
    m_functions.put((FUNC_CEILING_STRING).intern(), new Integer(FUNC_CEILING));
    m_functions.put((FUNC_ROUND_STRING).intern(), new Integer(FUNC_ROUND));
    m_functions.put((FUNC_SUM_STRING).intern(), new Integer(FUNC_SUM));
    m_functions.put((FUNC_STRING_STRING).intern(), new Integer(FUNC_STRING));
    m_functions.put((FUNC_STARTS_WITH_STRING).intern(), new Integer(FUNC_STARTS_WITH));
    m_functions.put((FUNC_CONTAINS_STRING).intern(), new Integer(FUNC_CONTAINS));
    m_functions.put((FUNC_SUBSTRING_BEFORE_STRING).intern(), new Integer(FUNC_SUBSTRING_BEFORE));
    m_functions.put((FUNC_SUBSTRING_AFTER_STRING).intern(), new Integer(FUNC_SUBSTRING_AFTER));
    m_functions.put((FUNC_NORMALIZE_STRING).intern(), new Integer(FUNC_NORMALIZE));
    m_functions.put((FUNC_TRANSLATE_STRING).intern(), new Integer(FUNC_TRANSLATE));
    m_functions.put((FUNC_CONCAT_STRING).intern(), new Integer(FUNC_CONCAT));
    m_functions.put((FUNC_FORMAT_NUMBER_STRING).intern(), new Integer(FUNC_FORMAT_NUMBER));
    m_functions.put((FUNC_SYSTEM_PROPERTY_STRING).intern(), new Integer(FUNC_SYSTEM_PROPERTY));
    m_functions.put((FUNC_FUNCTION_AVAILABLE_STRING).intern(), new Integer(FUNC_FUNCTION_AVAILABLE));
    m_functions.put((FUNC_SUBSTRING_STRING).intern(), new Integer(FUNC_SUBSTRING));
    m_functions.put((FUNC_STRING_LENGTH_STRING).intern(), new Integer(FUNC_STRING_LENGTH));
  }
  
  /**
   * Given a string, return the corrisponding token.
   */
  final int getKeywordToken(String key)
  {
    int tok;
    try
    {
      Integer itok = (Integer)m_keywords.get(key);
      tok = (null != itok) ? itok.intValue() : 0;
    }
    catch(NullPointerException npe)
    {
      tok = 0;
    }
    catch(ClassCastException cce)
    {
      tok = 0;
    }
    return tok;
  }

  /**
   * Given a string, return the corrisponding token.
   */
  final int getFunctionToken(String key)
  {
    int tok;
    try
    {
      tok = ((Integer)(m_functions.get(key))).intValue();
    }
    catch(NullPointerException npe)
    {
      tok = 0;
    }
    catch(ClassCastException cce)
    {
      tok = 0;
    }
    return tok;
  }
  
}

/**
 * $Log: QueryEngine.java,v $
 * Revision 1.5  1998/12/21 05:55:22  sboag
 * For "..", don't continue query if getParent() returns null (i.e. for the root).
 *
 * Revision 1.4  1998/12/18 18:22:24  sboag
 * Fixed bug with select qualifiers on text(), attributes, pi's, doc nodes, and comments.  Fixed some bugs with the attribute() function.
 *
 * Revision 1.3  1998/12/11 08:46:04  sboag
 * Don't think any changes were made.
 *
 * Revision 1.2  1998/12/10 03:11:56  sboag
 * LotusXSL Version 13a
 *
 */

