package com.lotus.xsl;

import java.util.*;
import org.w3c.dom.*;

/**
 * Classes that implement the QueryEngineService interface are responsible 
 * for querying nodes in XML documents or DOM trees, matching node patterns, 
 * and evaluating string expressions.
 * 
 * DESIGN ISSUE: Should the Match pattern calls and the expression evaluation 
 * calls have their own interfaces?
 */
public interface QueryEngineService
{
  /**
   * Given a select pattern, query a DOM for elements 
   * and/or attributes. This works exactly the opposite 
   * of matchAgainstPattern, which starts from the 
   * target node and marches inward and outward, qualifying
   * the element matches. Instead, this function starts 
   * from the anchor and marches left to right, traversing
   * the tree until it finds the target node (which it then 
   * qualifies the same as matchAgainstPattern).
   * @param expression The select expression.
   * @param contextNode The node that "." expresses.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   * @return A NodeList of nodes.
   */
  public NodeList query(String expression, Node contextNode, Element namespaceContext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException;

  /**
   * Given a select pattern, query a filesystem for documents that 
   * qualify according to the given pattern. This works exactly the opposite 
   * of matchAgainstPattern, which starts from the 
   * target node and marches inward and outward, qualifying
   * the element matches. Instead, this function starts 
   * from the anchor and marches left to right, traversing
   * the tree until it finds the target node (which it then 
   * qualifies the same as matchAgainstPattern).
   * @param expression The select expression.
   * @param contextNode The node that "." expresses.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   * @return A vector of URIs.
   */
  public Vector query(String expression, java.net.URL location)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException;

  /**
   * Returns a vector of target element strings.
   *  If the input is "foo", return {"foo"}.
   *  If the input is "foo | doo", return {"foo", "doo"}.
   *  If the input is "x/foo", return {"foo"}.
   *  If the input is "x/foo | z/foo", return {"foo"} -- NOTE the two element names are combined!
   *  If the input is "id(xx)/x//z[@aName]//foo[zz]", return {"foo"}.
   *  If the input is "id(xx)/x//z[@aName]//foo[zz] | roo/foo", return {"foo"}.
   *  If the input is "id(xx)/x//z[@aName]//foo[zz] | roo/doo", return {"foo", "doo"}.
   *  If the input is "id(xx)/x//z[@aName]//*[zz]", return {"*"}.
   * @param   pattern  Valid XSL pattern expression.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Vector getTargetElementStrings(String expression, Element namespaceContext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException;
  
  /**
   * Match a node against a list of ancestryPatterns.
   * @param expression The select expression.
   * @param node The node that is being tested.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   * @return Zero if no match, otherwise a specificity value.
   */
  public int matchAgainstPatterns(String expression, Node node, Element namespaceContext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException;

  /**
   * Given a match pattern and a context, find the first ancestor 
   * that matches the pattern (including the context handed in).
   * @param matchPatternString The match pattern.
   * @param node The node that "." expresses.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Node findAncestor(String matchPatternString, Node context, Element namespaceContext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException;

  /**
   * Given a 'from' pattern (ala xsl:number), a match pattern 
   * and a context, find the first ancestor that matches the 
   * pattern (including the context handed in).
   * @param matchPatternString The match pattern.
   * @param node The node that "." expresses.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   */
  Node findAncestor(String fromMatchPattern, String countMatchPattern,
                    Node context, 
                    Element namespaceContext);
  
  /**
   * Given an expression and a context, return the results 
   * in a string.
   * @param expression The string expression.
   * @param node The node that "." expresses.
   * @param namespaceContext The context in which namespaces in the 
   * queries are supposed to be expanded.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public String getStringExpr(String expression, Node contextNode, Element namespaceContext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException;
}
