/**
 * $Id: XMLParserLiaison.java,v 1.1 1999/01/15 23:20:07 sboag Exp $
 * 
 * Copyright (c) 1998-1999 Lotus Corporation, Inc. All Rights Reserved.
 *				This software is provided without a warranty of any kind.
 * 
 * $State: Exp $
 * 
 * @author Scott Boag (scott_boag@lotus.com)
 */
package com.lotus.xsl;

import java.io.*;
import org.w3c.dom.*;

/**
 * An implementation of this interface acts as a liaison 
 * between the XSL processor and the XML parser. It is 
 * needed in order to support features like included files, 
 * and to cover for deficiencies in the DOM. An implementation 
 * of this interface is a requirement for the XSL processor.
 */
public interface XMLParserLiaison
{
  /**
   * Set which processor this instance is associated with. 
   * This must be set, but should normally by called by the 
   * XSLProcessor class.
   */
  public void setProcessorOwner(XSLProcessor processor);

  /**
   * Parse the text pointed at by the reader as XML, and return 
   * a DOM Document interface.  May return null if not 
   * supported.
   * @param reader A stream that should hold valid XML.
   * @exception IOException   thrown if unable to open the XML stream.
   * @exception XSLProcessorException thrown for all other error conditions.
   */
  public Document parseXMLStream(Reader reader)
        throws IOException, XSLProcessorException;

  /**
   * Parse the text pointed at by the reader as XML, and return 
   * a DOM Document interface.  May return null if not 
   * supported.  It is recommended that you pass in some sort 
   * of recognizable name, such as the filename or URI, with 
   * which the reader can be recognized if the parse fails.
   * @param reader A stream that should hold valid XML.
   * @param identifier Used for diagnostic purposes only, 
   * some sort of identification for error reporting, may be 
   * null or empty string.
   * @exception IOException   thrown if unable to open the XML stream.
   * @exception XSLProcessorException thrown for all other error conditions.
   */
  public Document parseXMLStream(Reader reader, String identifier)
        throws IOException, XSLProcessorException;

  /**
   * Parse the text pointed at by the URL as XML, and return 
   * a DOM Document interface.  May return null if not 
   * supported.
   * @param url A URI to a valid XML document.
   * @exception IOException   thrown if unable to open the XML stream.
   * @exception XSLProcessorException thrown for all other error conditions.
   */
  public Document parseXMLStream(java.net.URL url)
        throws IOException, XSLProcessorException;

  /**
   * Create an empty DOM Document.  Mainly used for creating an 
   * output document.
   */
  public Document createDocument();
  
  /**
   * Return the expanded element name.
   */
  public String getExpandedElementName(Element elem);

  /**
   * Returns the attribute name with the namespace expanded.
   */
  public String getExpandedAttributeName(Attr attr);

  /**
   * Given a prefix and a namespace context, return the expanded namespace.
   */
  public String getNamespaceForPrefix(String prefix, Element namespaceContext);

  /**
   * Returns the namespace of the given node.
   */
  public String getNamespaceOfNode(Node n);

  /**
   * Returns the local name of the given node.
   */
  public String getLocalNameOfNode(Node n);
  
  /**
   * Tell if the node is ignorable whitespace.
   * This should be in the DOM.  Return false if the 
   * parser doesn't handle this.
   */
  public boolean isIgnorableWhitespace(Text node);
  
  /**
   * This function has to be implemented, 
   * because the DOM WG decided that attributes don't 
   * have parents.  If LotusXSL is used with a DOM implementation
   * that reuses attribute nodes, this will not work correctly.
   */
  public Node getParentOfNode(Node node);

  /**
   * Given an ID, return the element.
   */
  public Element getElementByID(String id, Document doc);
  
  /**
   * Set special characters for attributes that will be escaped.
   */
  public void setSpecialCharacters(String str);

  /**
   * Get special characters for attributes that will be escaped.
   */
  public String getSpecialCharacters();
  
  /**
   * Get the amount to indent when indent-result="yes".
   */
  public int getIndent();
  
  /**
   * Set the amount to indent when indent-result="yes".
   */
  public void setIndent(int i);

  /**
   * Get whether or not to expand all entity references in the 
   * source and style trees.
   */
  public boolean getShouldExpandEntityRefs();
  
  /**
   * Set whether or not to expand all entity references in the 
   * source and style trees.
   */
  public void SetShouldExpandEntityRefs(boolean b);

  /**
   * Get whether or not validation will be performed.  Validation is off by default.
   */
  public boolean getUseValidation();
  
  /**
   * If set to true, validation will be performed.  Validation is off by default.
   */
  public void setUseValidation(boolean b);
  
  /**
   * Return a string suitible for telling the user what parser is being used.
   */
  public String getParserDescription();
}

/**
 * $Log: XMLParserLiaison.java,v $
 * Revision 1.1  1999/01/15 23:20:07  sboag
 * no message
 *
 * Revision 1.3  1998/12/11 08:45:21  sboag
 * Added isIgnorableWhitespace API.
 *
 * Revision 1.2  1998/12/10 03:11:54  sboag
 * LotusXSL Version 13a
 *
 */
