/**
 * $Id$
 * 
 * Copyright (c) 1998-1999 Lotus Corporation, Inc. All Rights Reserved.
 *				This software is provided without a warranty of any kind.
 * 
 * $State$
 * 
 * @author Scott Boag <scott_boag@lotus.com>
 */
package com.lotus.xsl;

import java.net.*;
import java.io.*;
import java.util.*;
import org.w3c.dom.*;

/**
 * This class is the default XMLParserLiaison for the XSL Processor.
 * It can be used alone, for programmatically built DOMs (I expect 
 * this use will be very rare), or it can be derived from, in order 
 * to make parser specific implementations.
 */
public class XMLParserLiaisonDefault implements XMLParserLiaison, Formatter
{  
  /**
   * Return a string suitible for telling the user what parser is being used.
   */
  public String getParserDescription()
  {
    return "(No parser - generic DOM)";
  }

  /**
   * A pointer back to the owning XSL processor, mainly for use for 
   * warning and error handling.
   */
  public XSLProcessor m_processor = null;
  
  /**
   * The amount to indent when indent-result="yes".
   */
  protected int m_indent = 0;
  
  /**
   * Get the amount to indent when indent-result="yes".
   */
  public int getIndent()
  {
    return m_indent;
  }
  
  /**
   * Set the amount to indent when indent-result="yes".
   */
  public void setIndent(int i)
  {
    m_indent = i;
  }
  
  /**
   * By default expand all entity references in the 
   * source and style trees.
   */
  protected boolean m_shouldExpandEntityRefs = true;
  
  /**
   * get whether or not to expand all entity references in the 
   * source and style trees.
   */
  public boolean getShouldExpandEntityRefs()
  {
    return m_shouldExpandEntityRefs;
  }
  
  /**
   * Set whether or not to expand all entity references in the 
   * source and style trees.
   */
  public void SetShouldExpandEntityRefs(boolean b)
  {
    m_shouldExpandEntityRefs = b;
  }

  /**
   * If set to true, validation will be performed.  Validation is off by default.
   */
  public boolean m_use_validation = false;
 
  /**
   * If set to true, validation will be performed.  Validation is off by default.
   */
  public boolean getUseValidation()
  {
    return m_use_validation;
  }
  
  /**
   * If set to true, validation will be performed.  Validation is off by default.
   */
  public void setUseValidation(boolean b)
  {
    m_use_validation = b;
  }

  
  /**
   * These are characters that will be escaped in the output.
   */
  public String m_attrSpecialChars = "<>&\"\'\r\n";
  
  /**
   * Set special characters for attributes that will be escaped.
   */
  public void setSpecialCharacters(String str)
  {
    m_attrSpecialChars = str;
  }
  
  /**
   * Get special characters for attributes that will be escaped.
   */
  public String getSpecialCharacters()
  {
    return m_attrSpecialChars;
  }
  
  public static final String S_XMLNAMESPACEURI = "http://www.w3.org/XML/1998/namespace";
  
  public FormatterListener m_formatterListener = null;
  
  /**
   * Construct an instance.
   */
  public XMLParserLiaisonDefault()
  {
  }

  /**
   * Set which processor this instance is associated with, mainly for 
   * use for error and warning handling. This must be set, but should 
   * normally by set by the XSLProcessor class.
   */
  public void setProcessorOwner(XSLProcessor processor)
  {
    m_processor = processor;
  }

  /**
   * Parse the text pointed at by the reader as XML, and return 
   * a DOM Document interface.  May return null if not 
   * supported.
   * @param reader A stream that should hold valid XML.
   * @exception IOException   thrown if unable to open the XML stream.
   * @exception XSLProcessorException thrown for all other error conditions.
   */
  public Document parseXMLStream(Reader reader)
        throws IOException, XSLProcessorException
  {
    return parseXMLStream(reader, "Anonymous Document");
  }

  /**
   * Parse the text pointed at by the reader as XML, and return 
   * a DOM Document interface.  Implementation of XMLParserLiaison
   * interface method.
   * Default handling: Not supported - need parser-specific implementation.
   * @exception IOException   thrown if unable to open the XML stream.
   * @exception XSLProcessorException thrown for all other error conditions.
   */
  public Document parseXMLStream(Reader reader, String filename)
    throws IOException, XSLProcessorException
  {
    throw new RuntimeException("parseXMLStream(Reader) not supported in XMLParserLiaisonDefault! "+
      "Can not open "+filename);
  }
  
  /**
   * Parse the text pointed at by the URL, and return 
   * a DOM Document interface.  Implementation of XMLParserLiaison
   * interface method.
   * Default handling: Not supported - need parser-specific implementation.
   * @exception IOException   thrown if unable to open the XML stream.
   * @exception XSLProcessorException thrown for all other error conditions.
   */
  public Document parseXMLStream(URL url)
    throws IOException, XSLProcessorException
  {
    Document doc = null;
    InputStream instream = null;
    String whichCall = "";
    
    whichCall = "openConnection";
    URLConnection connect = url.openConnection();
    
    whichCall = "getInputStream";
    instream = connect.getInputStream();
    
    Reader reader = new java.io.BufferedReader(
                                               new java.io.InputStreamReader(instream), 1024 * 16);
    doc = parseXMLStream(reader, url.toExternalForm());
    return doc;
  }
  
  /**
   * Create an empty DOM Document.  Mainly used for creating an 
   * output document.  Implementation of XMLParserLiaison
   * interface method.
   * Default handling: Not supported - need parser-specific implementation.
   */
  public Document createDocument()
  {
    throw new RuntimeException("createDocument() not supported in XMLParserLiaisonDefault!");
    // return null;
  }
  
  
  /**
   * Given a prefix and a namespace context, return the expanded namespace.
   * Default handling: 
   */
  public String getNamespaceForPrefix(String prefix, Element namespaceContext)
  {
    int type;
    Node parent = namespaceContext;
    String namespace = null;
    if(prefix.equals("xml"))
    {
      namespace = S_XMLNAMESPACEURI;
    }
    else
    {
      while ((null != parent) && (null == namespace)
             && (((type = parent.getNodeType()) == Node.ELEMENT_NODE)
                 || (type == Node.ENTITY_REFERENCE_NODE))) 
      {
        if (type == Node.ELEMENT_NODE) 
        {
          NamedNodeMap nnm = parent.getAttributes();
          for (int i = 0;  i < nnm.getLength();  i ++) 
          {
            Node attr = nnm.item(i);
            String aname = attr.getNodeName();
            boolean isPrefix = aname.startsWith("xmlns:");
            if (aname.equals("xmlns") || isPrefix) 
            {
              int index = aname.indexOf(':');
              String p = isPrefix ? aname.substring(index+1) : "";
              if (p.equals(prefix)) 
              {
                namespace = attr.getNodeValue();
                break;
              }
            }
          }
        }
        parent = getParentOfNode(parent);
      }
    }
    return namespace;
  }
  
  /**
   * An experiment for the moment.
   */
  protected Hashtable m_NSInfos = new Hashtable();
  
  protected static final NSInfo m_NSInfoUnProcWithXMLNS = new NSInfo(false, true);
  protected static final NSInfo m_NSInfoUnProcWithoutXMLNS = new NSInfo(false, false);
  protected static final NSInfo m_NSInfoUnProcNoAncestorXMLNS = new NSInfo(false, false, NSInfo.ANCESTORNOXMLNS);
  protected static final NSInfo m_NSInfoNullWithXMLNS = new NSInfo(true, true);
  protected static final NSInfo m_NSInfoNullWithoutXMLNS = new NSInfo(true, false);
  protected static final NSInfo m_NSInfoNullNoAncestorXMLNS = new NSInfo(true, false, NSInfo.ANCESTORNOXMLNS);
    
  protected Vector m_candidateNoAncestorXMLNS =  new Vector();
  
  /**
   * Returns the namespace of the given node.
   */
  public String getNamespaceOfNode(Node n)
  {
    String namespaceOfPrefix;
    boolean hasProcessedNS;
    NSInfo nsInfo;
    int ntype = n.getNodeType();
    if(Node.ATTRIBUTE_NODE != ntype)
    {
      Object nsObj = m_NSInfos.get(n); // return value
      nsInfo = (nsObj == null) ? null : (NSInfo)nsObj;
      hasProcessedNS = (nsInfo == null) ? false : nsInfo.m_hasProcessedNS;
    }
    else
    {
      hasProcessedNS = false;
      nsInfo = null;
    }
    if(hasProcessedNS)
    {
      namespaceOfPrefix = nsInfo.m_namespace;
    }
    else
    {
      namespaceOfPrefix = null;
      String nodeName = n.getNodeName();
      int indexOfNSSep = nodeName.indexOf(':');
      String prefix;
      // If we have an attribute node without a prefix, then 
      // we should use the prefix of the element parent.
      if(Node.ATTRIBUTE_NODE == ntype)
      {
        if(indexOfNSSep >= 0)
        {
          prefix = nodeName.substring(0, indexOfNSSep);
        }
        else
        {
          Node parentOfAttr = getParentOfNode(n);
          if(null == parentOfAttr)
          {
            System.out.println("Warning, could not find parent for attribute node: "+n);
            nodeName = parentOfAttr.getNodeName();
            return namespaceOfPrefix;
          }
          else
          {
            n = parentOfAttr;
            nodeName = n.getNodeName();
            indexOfNSSep = nodeName.indexOf(':');
            prefix = (indexOfNSSep >= 0) ? nodeName.substring(0, indexOfNSSep) : "";
          }
        }
      }
      else
      {
        prefix = (indexOfNSSep >= 0) ? nodeName.substring(0, indexOfNSSep) : "";
      }
      boolean ancestorsHaveXMLNS = false;
      boolean nHasXMLNS = false;
      if(prefix.equals("xml"))
      {
        namespaceOfPrefix = S_XMLNAMESPACEURI;
      }
      else
      {
        int parentType;
        Node parent = n;
        while ((null != parent) && (null == namespaceOfPrefix)) 
        {
          if((null != nsInfo) 
             && (nsInfo.m_ancestorHasXMLNSAttrs == nsInfo.ANCESTORNOXMLNS))
          {
            break;
          }
          parentType = parent.getNodeType();
          if((null == nsInfo) || nsInfo.m_hasXMLNSAttrs)
          {
            boolean elementHasXMLNS = false;
            if (parentType == Node.ELEMENT_NODE) 
            {
              NamedNodeMap nnm = parent.getAttributes();
              for (int i = 0;  i < nnm.getLength();  i ++) 
              {
                Node attr = nnm.item(i);
                String aname = attr.getNodeName();
                if(aname.charAt(0) == 'x')
                {
                  boolean isPrefix = aname.startsWith("xmlns:");
                  if (aname.equals("xmlns") || isPrefix) 
                  {
                    if(n == parent)
                      nHasXMLNS = true;
                    elementHasXMLNS = true;
                    ancestorsHaveXMLNS = true;
                    String p = isPrefix ? aname.substring(6) : "";
                    if (p.equals(prefix)) 
                    {
                      namespaceOfPrefix = attr.getNodeValue();
                      break;
                    }
                  }
                }
              }
            }
            if((Node.ATTRIBUTE_NODE != parentType) && (null == nsInfo) && (n != parent))
            {
              nsInfo = elementHasXMLNS ? m_NSInfoUnProcWithXMLNS : m_NSInfoUnProcWithoutXMLNS;
              m_NSInfos.put(parent, nsInfo);
            }
          }
          if(Node.ATTRIBUTE_NODE == parentType)
          {
            parent = getParentOfNode(parent);
          }
          else
          {
            m_candidateNoAncestorXMLNS.addElement(parent);
            m_candidateNoAncestorXMLNS.addElement(nsInfo);
            parent = parent.getParentNode();
          }
          if(null != parent)
          {
            Object nsObj = m_NSInfos.get(parent); // return value
            nsInfo = (nsObj == null) ? null : (NSInfo)nsObj;
          }
       }
        int nCandidates = m_candidateNoAncestorXMLNS.size();
        if(nCandidates > 0)
        {
          if((false == ancestorsHaveXMLNS) && (null == parent))
          {
            for(int i = 0; i < nCandidates; i+=2)
            {
              Object candidateInfo = m_candidateNoAncestorXMLNS.elementAt(i+1);
              if(candidateInfo == m_NSInfoUnProcWithoutXMLNS)
              {
                m_NSInfos.put(m_candidateNoAncestorXMLNS.elementAt(i), 
                              m_NSInfoUnProcNoAncestorXMLNS );
              }
              else if(candidateInfo == m_NSInfoNullWithoutXMLNS)
              {
                m_NSInfos.put(m_candidateNoAncestorXMLNS.elementAt(i), 
                              m_NSInfoNullNoAncestorXMLNS );
              }
            }
          }
          m_candidateNoAncestorXMLNS.removeAllElements();
        }
      }
      
      if(Node.ATTRIBUTE_NODE != ntype)
      {
        if(null == namespaceOfPrefix)
        {
          if(ancestorsHaveXMLNS)
          {
            if(nHasXMLNS)
              m_NSInfos.put(n, m_NSInfoNullWithXMLNS );
            else
              m_NSInfos.put(n, m_NSInfoNullWithoutXMLNS );
          }
          else
          {
            m_NSInfos.put(n, m_NSInfoNullNoAncestorXMLNS );
          }
        }
        else
        {
          m_NSInfos.put(n,  new NSInfo(namespaceOfPrefix, nHasXMLNS));
        }
      }
    }
    return namespaceOfPrefix;
  }

  /**
   * Returns the local name of the given node.
   */
  public String getLocalNameOfNode(Node n)
  {
    String qname = n.getNodeName();
    int index = qname.indexOf(':');
    return (index < 0) ? qname : qname.substring(index+1);
  }

  /**
   * Returns the element name with the namespace expanded.
   */
  public String getExpandedElementName(Element elem)
  {
    String namespace = getNamespaceOfNode(elem);
    return (null != namespace) ? namespace+":"+ getLocalNameOfNode(elem) 
                                 : getLocalNameOfNode(elem);
  }

  /**
   * Returns the attribute name with the namespace expanded.
   */
  public String getExpandedAttributeName(Attr attr)
  {
    String namespace = getNamespaceOfNode(attr);
    return (null != namespace) ? namespace+":"+ getLocalNameOfNode(attr) 
                                 : getLocalNameOfNode(attr);
  }

  /** 
   * Tell if the node is ignorable whitespace.
   */
  public boolean isIgnorableWhitespace(Text node)
  {
    boolean isIgnorable = false; // return value
    // TODO: I can probably do something to figure out if this 
    // space is ignorable from just the information in
    // the DOM tree.
    return isIgnorable;
  }
  
  /**
   * I have to write this silly, and expensive function, 
   * because the DOM WG decided that attributes don't 
   * have parents.  If LotusXSL is used with a DOM implementation
   * that reuses attribute nodes, this will not work correctly.
   */
  public Node getParentOfNode(Node node)
    throws RuntimeException
  {
    Node parent;
    int nodeType = node.getNodeType();
    if(Node.ATTRIBUTE_NODE == nodeType)
    {
      Document doc = node.getOwnerDocument();
      
      if(null == doc)
      {
        throw new RuntimeException("Attribute child does not have an owner document!");
      }
      
      Element rootElem = doc.getDocumentElement();
      
      if(null == rootElem)
      {
        throw new RuntimeException("Attribute child does not have an owner document element!");
      }
      
      parent = locateAttrParent(rootElem, node);
    }
    else
    {
      parent = node.getParentNode();
    }
    return parent;
  }
  
  /**
   * Given an ID, return the element.
   */
  public Element getElementByID(String id, Document doc)
  {
    return null;
  }
  
  /**
   * Support for getParentOfNode.
   */
  private Node locateAttrParent(Element elem, Node attr)
  {
    Node parent = null;
    NamedNodeMap attrs = elem.getAttributes();
    if(null != attrs)
    {
      int nAttrs = attrs.getLength();
      for(int i = 0; i < nAttrs; i++)
      {
        if(attr == attrs.item(i))
        {
          parent = elem;
          break;
        }
      }
    }
    if(null == parent)
    {
      NodeList children = elem.getChildNodes();
      int nChildren = children.getLength();
      for(int i = 0; i < nChildren; i++)
      {
        Node node = children.item(i);
        if(Node.ELEMENT_NODE == node.getNodeType())
        {
          parent = locateAttrParent((Element)node, attr);
          if(null != parent) break;
        }
      }
    }
    return parent;
  }
  
  /*=================== Formatter Interface Methods ===================*/

  /**
   * Set an instance of the formatter listener that is associated 
   * with this formatter.
   */
  public void setFormatterListener(PrintWriter pw, String resultns, boolean format)
  {
    if((null != resultns) && resultns.trim().equals("http://www.w3.org/TR/REC-html40"))
    {
      FormatterToHTML htmlFormatVisitor
        = new FormatterToHTML(pw, null, 
                                     "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n",
                                     format, m_indent);
      m_formatterListener =  htmlFormatVisitor;
      htmlFormatVisitor.indent = m_indent;
      htmlFormatVisitor.m_attrSpecialChars = m_attrSpecialChars;
    }
    else if(null == m_formatterListener)
    {
      FormatterToXML fToXML = new FormatterToXML(pw, null, null, format, m_indent);
      fToXML.m_attrSpecialChars = m_attrSpecialChars;
      m_formatterListener = fToXML;
    }
  }
  

  /**
   * Set an instance of the formatter listener that is associated 
   * with this formatter.
   */
  public void setFormatterListener(FormatterListener fl)
  {
    m_formatterListener = fl;
  }
  
  /**
   * Get an instance of the formatter listener that is associated 
   * with this formatter.  For now this just returns null.
   */
  public FormatterListener getFormatterListener()
  {
    return m_formatterListener;
  }
  
  /**
   * Print a DOM tree.
   * @param doc The result tree.
   * @param pw The PrintWriter to print the contents to.
   * @param resultns the value of the resultns attribute.
   * @param format Tells if it should be pretty-printed.
   */
  public void toMarkup(Document doc, PrintWriter pw, String resultns, boolean format)
    throws Exception
  {
    FormatterListener visitor = null;
    if((null != resultns) && resultns.trim().equals("http://www.w3.org/TR/REC-html40"))
    {
      FormatterToHTML htmlFormatVisitor
        = new FormatterToHTML(pw, null, 
                                     "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n",
                                     format, m_indent);
      visitor =  htmlFormatVisitor;
      htmlFormatVisitor.indent = m_indent;
      htmlFormatVisitor.m_attrSpecialChars = m_attrSpecialChars;
    }
    else
    {
      FormatterToXML fToXML = new FormatterToXML(pw, null, null, format, m_indent);
      fToXML.m_attrSpecialChars = m_attrSpecialChars;
      visitor = fToXML;
    }
    if(null != visitor)
    {
      TreeWalker treeTraversal 
        = new TreeWalker(visitor);
      treeTraversal.traverse(doc);
    }
  } 
    
}

/**
 * $Log$
 */
