/**
 * $Id: JavaDispatcher.java,v 1.2 1998/12/10 03:17:36 sboag Exp $
 * 
 * Copyright (c) 1998-1999 Lotus Corporation, Inc. All Rights Reserved.
 *				This software is provided without a warranty of any kind.
 * 
 * $State: Exp $
 * 
 * @author Scott Boag (scott_boag@lotus.com)
 */
package com.lotus.xsl.dispatchers;

import org.w3c.dom.*;

import java.lang.reflect.*;
import java.util.*;
import com.lotus.xsl.*;

/**
 * A Java extension.  This uses Java reflection to find 
 * the right method to call.
 * @author "Sanjiva Weerawarana" <sanjiva@watson.ibm.com>
 * @author "Joseph Kesselman" <keshlam@us.ibm.com>
 * @author "Scott Boag" <scott_boag@lotus.com>
 */
public class JavaDispatcher implements Dispatcher
{    
  String m_functions;
  String m_classid;
  String m_archive;
  String m_codetype;
  String m_codebase;
  
  Element m_fallbackElement;
  
  public static DispatcherFactory register(String type) 
  {
    return new JavaDispatcherFactory();
  }
  
  /**
   * Construct an extension object.
   */
  public JavaDispatcher(String functions, 
                       String classid, 
                       String archive, 
                       String codetype, 
                       String codebase,
                       Element fallbackElement)
  {
    m_functions = functions;
    m_classid = classid;
    m_archive = archive;
    m_codetype = codetype;
    m_codebase = codebase;
    m_fallbackElement = fallbackElement;
  }
  
  /**
   * Get the fallback element associated with this dispatcher.
   */
  public Element getFallbackElement()
  {
    return m_fallbackElement;
  }
  
  /** 
   * Internal subroutine for getEntryPoint(): Format arguments as a
   * string describing the function being searched for. Used in
   * verbose exceptions. 
   */
  public static String callToString(Class targetClass,String methodName,
                             Class[] argTypes,boolean isStaticReference)
  {
    StringBuffer buf = new StringBuffer();
    if(isStaticReference)
    {
      buf.append("static ");
    }
    buf.append(targetClass.getName());
    if(methodName!=null)
    {
      buf.append(".").append(methodName);
    }
    buf.append("(");
    if (argTypes != null && argTypes.length>0) 
    {
      if(false)
      {
        // ????? Sanjiva has an ArrayToString method. Using it would save
        // a few bytes, at cost of giving up some reusability.
      }
      else
      {
        buf.append(argTypes[0]);
        for (int i = 1; i < argTypes.length; i++) 
        {
          buf.append(",").append(argTypes[i]);
        }
      }
    }
    else
    {
      buf.append("[none]");
    }
    buf.append(")");
    return buf.toString();
  }
  
  /** 
   * Determine whether a given method parameter type can accept
   * arguments of another type.
   *
   * parm: The type given in the method's signature.
   * arg: The type we want to pass in.
   * 
   * Legal METHOD CONVERSIONS (5.3) are Identity, Widening Primitive
   * Conversion, or Widening Reference Conversion. NOTE that this is a
   * subset of the legal ASSIGNMENT CONVERSIONS (5.2) -- in particular,
   * we can't implicitly narrow int to byte, short or char.  
   */
  private static boolean isMethodConvertable(Class parm, Class arg)
  {
    if(parm.equals(arg))	// If same class, short-circuit now!
      return true;
    
    // Arrays are convertable if their elements are convertable
    // ????? Does this have to be done before isAssignableFrom, or
    // does it successfully handle arrays of primatives?
    while(parm.isArray())
    {
      if(!arg.isArray())
      {
        return false;		// Unequal array depth
      }
      else
      {
        parm=parm.getComponentType();
        arg=arg.getComponentType();
      }
    }
    if(arg.isArray())
      return false;		// Unequal array depth
    
    // Despite its name, the 1.1.6 docs say that this function does
    // NOT return true for all legal ASSIGNMENT CONVERSIONS
    // (5.2):
    //	 "Specifically, this method tests whether the type
    //	 represented by the specified class can be converted
    //	 to the type represented by this Class object via
    //	 an identity conversion or via a widening reference
    //	 conversion."
    if(parm.isAssignableFrom(arg))
      return true;

    // That leaves us the Widening Primitives case. Four possibilities:
    // void (can only convert to void), boolean (can only convert to boolean),
    // numeric (which are sequenced) and char (which inserts itself into the
    // numerics by promoting to int or larger)

    if(parm.equals(Void.TYPE) || parm.equals(Boolean.TYPE) ||
       arg.equals(Void.TYPE) || arg.equals(Boolean.TYPE))
      return false;
    
    Class[] primTypes=
    {
      Character.TYPE, Byte.TYPE, Short.TYPE, Integer.TYPE,
      Long.TYPE, Float.TYPE, Double.TYPE 
    };
    int parmscore,argscore;
    
    for(parmscore=0;parmscore<primTypes.length;++parmscore)
    {
      if (parm.equals(primTypes[parmscore]))
        break;
    }
    if(parmscore>=primTypes.length)
      return false;		// Off the end
    
    for(argscore=0;argscore<primTypes.length;++argscore)
    {
      if (arg.equals(primTypes[argscore]))
        break;
    }
    if(argscore>=primTypes.length)
      return false;		// Off the end
    
    // OK if ordered AND NOT char-to-smaller-than-int
    return (argscore<parmscore && (argscore!=0 || parmscore>2) );
  }
  
  /** Utility function: obtain common data from either Method or
   Constructor. (In lieu of an EntryPoint interface.) */
  static int entryGetModifiers(Object entry)
  {
    return (entry instanceof Method)
           ? ((Method)entry).getModifiers()
             : ((Constructor)entry).getModifiers();
  }
  
  /** Utility function: obtain common data from either Method or
   Constructor. (In lieu of an EntryPoint interface.) */
  static String entryGetName(Object entry)
  {
    return (entry instanceof Method)
           ? ((Method)entry).getName()
             : ((Constructor)entry).getName();
  }
  
  /** 
   * Utility function: obtain common data from either Method or
   * Constructor. (In lieu of an EntryPoint interface.) 
   */
  static Class[] entryGetParameterTypes(Object entry)
  {
    return (entry instanceof Method)
           ? ((Method)entry).getParameterTypes()
             : ((Constructor)entry).getParameterTypes();
  }
  
  /** 
   * Convenience method: Test an entire parameter-list/argument-list pair
   * for isMethodConvertable(), qv.
   */
  static boolean areMethodConvertable(Class[] parms,Class[] args)
  {
    if(parms.length!=args.length)
    {
      return false;
    }
    
    for(int i=0;i<parms.length;++i)
    {
      if(!isMethodConvertable(parms[i],args[i]))
      {
        return false;
      }
    }
    
    return true;
  }
  
  /** 
   * Search for entry point, per  Java Language Spec 1.0
   * as amended, verified by comparison against compiler behavior.
   * 
   * targetClass: Class object for the class to be queried.
   * 
   * methodName: Name of method to invoke, or null for constructor.
   * Only Public methods will be accepted.
   * 
   * argTypes: Classes of intended arguments.  Note that primitives
   * must be specified via their TYPE equivalents, rather than as
   * their wrapper classes -- Integer.TYPE rather than Integer.
   * 
   * isStaticReference: If true, and if the target is a Class object,
   * only static methods will be accepted as valid matches.
   */
  private static Object getEntryPoint(Class targetClass,
                               String methodName,
                               Class[] argTypes,
                               boolean isStaticReference) 
    throws SecurityException, NoSuchMethodException
  {
    // 15.11.1: OBTAIN STARTING CLASS FOR SEARCH
    Object m=null;
    
    // 15.11.2 DETERMINE ARGUMENT SIGNATURE
    // (Passed in as argTypes array.)
    
    // Shortcut: If an exact match exists, return it.
    try 
    {
      if(methodName!=null)
      {
        m=targetClass.getMethod (methodName, argTypes);
        if(isStaticReference &&
           !Modifier.isStatic(entryGetModifiers(m)) )
        {
          throw new NoSuchMethodException();
        }
        return m;
      }
      else
      {
        return targetClass.getConstructor (argTypes);
      }
      
    }
    catch (NoSuchMethodException e) 
    {
      if(argTypes==null || argTypes.length==0)
      {
        throw e;		// no-args has no alternatives!
      }
      // Else fall through.
    }
    
    // Well, _that_ didn't work. Time to search for the Most Specific
    // matching function. NOTE that conflicts are possible!
    
    // 15.11.2.1 ACCESSIBLE: We apparently need to gather from two
    // sources to be sure we have both instance and static methods.
    Object[] local, inheritedAndStatic;
    if(methodName!=null)
    {
      local=targetClass.getDeclaredMethods();
      inheritedAndStatic=targetClass.getMethods();
    }
    else
    {
      local=targetClass.getDeclaredConstructors();
      inheritedAndStatic=targetClass.getConstructors();
    }
    if(0==local.length+inheritedAndStatic.length)
    {
      throw new NoSuchMethodException();
    }
    Object[] methods=new Object[local.length+inheritedAndStatic.length];
    System.arraycopy(local,0,methods,0,local.length);
    System.arraycopy(inheritedAndStatic,0,
                     methods,local.length,
                     inheritedAndStatic.length);
    
    MoreSpecific best=new MoreSpecific();
    for(int i=0;i<methods.length;++i)
    {
      Object mi=methods[i];
      if (
          // 15.11.2.1 ACCESSIBLE: Method is public.
          Modifier.isPublic(entryGetModifiers(mi))
          &&
          // 15.11.2.1 APPLICABLE: Right method name (or c'tor)
          (methodName==null || entryGetName(mi).equals(methodName) )
          &&
          // 15.11.2.1 APPLICABLE: Parameters match arguments
          areMethodConvertable(entryGetParameterTypes(mi),argTypes)
          )
        // 15.11.2.2 MORE SPECIFIC displace less specific.
        best.add(mi);
    }

    // May throw NoSuchMethodException; we pass in info needed to
    // create a useful VERBOSE message.
    m=best.getMostSpecific(targetClass,methodName,argTypes,isStaticReference);
    
    // 15.11.3 APPROPRIATE: Class invocation can call only static
    // methods. Note that the defined order of evaluation permits a
    // call to be resolved to an inappropriate method and then
    // rejected, rather than finding the best of the appropriate
    // methods.
    //
    // Constructors are never static, so we don't test them.
    if( methodName!=null &&
        isStaticReference &&
        !Modifier.isStatic(entryGetModifiers(m)) )
    {
      throw new NoSuchMethodException();
    }

    if(m==null)
    {
      throw new NoSuchMethodException();
    }
    return m;
  }

  /**
   * Class.getMethod() finds only the entry point (if any) _exactly_
   * matching the specified argument types. Our implmentation can
   * decide between several imperfect matches, using the same search
   * algorithm as the Java compiler.
   * 
   * This version emulates the compiler behavior by allowing lookup to
   * be performed against either a class or an instance --
   * classname.foo() must be a static method call, instance.foo() can
   * invoke either static or instance methods 
   */
  public static Method getMethod(Object target, String methodName,
                          Class[] argTypes)
  {
    boolean staticRef=target instanceof Class;
    return getMethod( staticRef ? (Class)target : target.getClass(),
                      methodName,argTypes,staticRef);
  }
  
  /**
   *  Class.getMethod() finds only the entry point (if any) _exactly_
   *  matching the specified argument types. Our implmentation can
   *  decide between several imperfect matches, using the same search
   *  algorithm as the Java compiler.
   * 
   *  This version more closely resembles Class.getMethod() -- we always
   *  ask the Class for the method. It differs in testing for
   *  appropriateness before returning the method; if the query is
   *  being made via a static reference, only static methods will be
   *  found and returned. 
   */
  public static Method getMethod(Class target, String methodName,
                          Class[] argTypes, boolean isStaticReference)
  {
    try 
    {
      return (Method) getEntryPoint(target, methodName, argTypes, isStaticReference);
    }
    catch (Exception e) 
    {
      // System.out.println(
      //                   "Exception while getting method '" +
      //                   callToString(target,methodName,argTypes,isStaticReference) + "': " + e);
    }
    return null;
  }
  
  /**
   * Tells if a given function is available.
   */
  public boolean isFunctionAvailable(String funcName)
  {
    boolean isFuncAvailable = false;
    try
    {
      Class classObj = Class.forName(m_classid);
      Method[] meths = classObj.getMethods();
      int nMeths = meths.length;
      for(int i = 0; i < nMeths; i++)
      {
        Method m = meths[i];
        if(m.getName().equals(funcName))
        {
          isFuncAvailable = true;
          break;
        }
      }
    }
    catch(ClassNotFoundException classNotFound)
    {
      System.out.println("Error when trying to call extension: "+classNotFound);
    }
    return isFuncAvailable;
  }

  /**
   * Return a String object from an extension.
   */
  public Object call(String method, Object[] args)
  {
    Object theReturnValue = null;
    
    Object instance = null;
    
    // System.out.println("JavaDispatcher.call("+m_classid+"."+m_name+")");
    Class[] staticArgTypes = null;
    Class[] nonStaticArgTypes = null;
    int nArgs;
    if(null != args)
    {
      nArgs = args.length;
      if(nArgs > 0)
      {
        staticArgTypes = new Class[nArgs];
        if(nArgs > 1) 
        {
          nonStaticArgTypes = new Class[nArgs-1];
        }
        instance = args[0];
        for(int i = 0; i < nArgs; i++)
        {
          staticArgTypes[i] = args[i].getClass();
          if((i >= 1) && (null != nonStaticArgTypes))
          {
            nonStaticArgTypes[i-1] = args[i].getClass();
          }
        }
      }
    }
    else
    {
      nArgs = 0;
    }
    
    try
    {
      Class classObj = Class.forName(m_classid);
      try
      {
        if(method.equals("new"))
        {
          Constructor constructor = classObj.getConstructor(staticArgTypes);
          theReturnValue = constructor.newInstance(args);
        }
        else
        {
          Method m = null;
          if(null != instance)
          {
            m = getMethod (instance, method, nonStaticArgTypes);
            if(null != m)
            {
              Object[] nonStaticArgs;
              if(nArgs > 1)
              {
                nArgs--;
                nonStaticArgs = new Object[nArgs];
                for(int k = 0; k < nArgs; k++)
                {
                  nonStaticArgs[k] = args[k+1];
                }
              }
              else
              {
                nonStaticArgs = null;
              }
              theReturnValue = m.invoke( instance, nonStaticArgs );
            }
          }
          if(null == m)
          {
            m = getMethod (classObj, method, staticArgTypes);
            if(null != m)
            {
              theReturnValue = m.invoke (null, args);
            }
            else
            {
              System.out.println("Can not find method when trying to call extension: "+m_classid+"."+method);
            }
          }
        }
      }
      catch(NoSuchMethodException nsme)
      {
        System.out.println("Error when trying to call extension: "+nsme);
        // ie.printStackTrace();
      }
      catch(InstantiationException ie)
      {
        System.out.println("Error when trying to call extension: "+ie);
        // ie.printStackTrace();
      }
      catch(InvocationTargetException invocationTarget)
      {
        System.out.println("Error when trying to call extension: "+invocationTarget);
        // invocationTarget.printStackTrace();
      }
      catch(IllegalAccessException illegalAccess)
      {
        System.out.println("Error when trying to call extension: "+illegalAccess);
      }
    }
    catch(ClassNotFoundException classNotFound)
    {
      System.out.println("Error when trying to call extension: "+classNotFound);
    }
    
    return theReturnValue;
  }
}

/** 
 * Internal Class for getEntryPoint(). Implements 15.11.2.2 MORE
 * SPECIFIC rules.
 * 
 * Retains a list of methods (already known to match the
 * arguments). As each method is added, we check against past entries
 * to determine which if any is "more specific" -- defined as having
 * _all_ its arguments (not just a preponderance) be
 * method-convertable into those of another. If such a relationship
 * is found, the more-specific method is retained and the
 * less-specific method is discarded. At the end, if this has yielded
 * a single winner it is considered the Most Specific Method and
 * hence the one that should be invoked.  Otherwise, a
 * NoSuchMethodException is thrown.
 * 
 * PERFORMANCE VERSUS ARCHITECTURE: Arguably, this should "have-a"
 * Vector. But the code is 6% smaller, and possibly faster, if we
 * code it as "is-a" Vector. Since it's an inner class, nobody's
 * likely to abuse the privilage.
 * 
 * Note: "Static" in the case of an inner class means "Does not
 * reference instance data in the outer class", and is required since
 * our caller is a static method. 
 */
class MoreSpecific
  extends Vector
{    
  MoreSpecific()
  {
  }
  
  /** 
   * Submit an entry-point to the list. May be discarded if a past
   * entry is more specific, or may cause others to be discarded it
   * if is more specific.
   * 
   * newEntry: Method or Constructor under consideration.
   */
  public boolean add(Object newEntry)
  {
    boolean didAdd = true;
    if(size()==0)
    {
      addElement(newEntry);
    }
    else
    {
      Class[] newargs=JavaDispatcher.entryGetParameterTypes(newEntry);
      boolean keep=true;
      for (Enumeration e = elements();
           keep & e.hasMoreElements() ;
           )
      {
        Object oldEntry=e.nextElement();
        // CAVEAT: Implicit references to enclosing class!
        Class[] oldargs=JavaDispatcher.entryGetParameterTypes(oldEntry);
        if(JavaDispatcher.areMethodConvertable(oldargs,newargs))
          removeElement(oldEntry); // New more specific; discard old
        else if(JavaDispatcher.areMethodConvertable(newargs,oldargs))
          keep=false;	// Old more specific; discard new
        // Else they're tied. Keep both and hope someone beats both.
      }
      if(keep)
      {
        addElement(newEntry);
      }
    }
    return didAdd;
  }

  /** 
   * Obtain the single Most Specific entry-point. If there is no clear
   * winner, or if the list is empty, throw NoSuchMethodException.
   * 
   * Arguments describe the call we were hoping to resolve. They are
   * actually not used unless a VERBOSE exception is being thrown, in
   * which case we use entryToString() to tell the user what we were
   * looking for. 
   */
  Object getMostSpecific(Class targetClass,String methodName,
                         Class[] argTypes,boolean isStaticReference)
    throws NoSuchMethodException
  {
    if(size()==1)
    {
      return firstElement();
    }
    else if(size()>1)
    {
      throw new NoSuchMethodException();
    }
    return null;
  }
}


class JavaDispatcherFactory implements DispatcherFactory
{
  /**
   * Create an extension object.
   */
  public Dispatcher create(Element extensionElem, Element fallbackElem)
  {
    String classid = extensionElem.getAttribute(Constants.ATTRNAME_CLASSID);
    String archive = extensionElem.getAttribute(Constants.ATTRNAME_ARCHIVE);
    String codetype = extensionElem.getAttribute(Constants.ATTRNAME_CODETYPE);
    String codebase = extensionElem.getAttribute(Constants.ATTRNAME_CODEBASE);
    String functions = extensionElem.getAttribute(Constants.ATTRNAME_METHOD);

    return new JavaDispatcher(functions, 
                              classid, 
                              archive, 
                              codetype, 
                              codebase, 
                              fallbackElem);
  }
}

/**
 * $Log: JavaDispatcher.java,v $
 * Revision 1.2  1998/12/10 03:17:36  sboag
 * LotusXSL Version 13a
 *
 */
