/**
 * $Id: JavaScriptDispatcher.java,v 1.1 1998/12/18 21:34:27 sboag Exp $
 * 
 * Copyright (c) 1998-1999 Lotus Corporation, Inc. All Rights Reserved.
 *				This software is provided without a warranty of any kind.
 * 
 * $State: Exp $
 * 
 * @author "Adam Peller" <peller@lotus.com>
 */
package com.lotus.xsl.dispatchers;

import java.util.*;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import com.lotus.xsl.*;
import com.netscape.javascript.Context;
import com.netscape.javascript.EvaluatorException;
import com.netscape.javascript.JavaScriptException;
import com.netscape.javascript.Scriptable;
import com.netscape.javascript.ScriptRuntime;
import com.netscape.javascript.ThreadLinkException;
import com.netscape.javascript.WrappedException;
import com.netscape.javascript.Wrapper;


/**
 * A JavaScript dispatcher.  This evaluates and executes 
 * JavaScript.
 * @author "Adam Peller" <peller@lotus.com>
 */
public class JavaScriptDispatcher implements Dispatcher
{    
  /**
   * The lotusxsl:fallback element associated 
   * with this extension -- just save it and return it 
   * when asked.
   */
  Element m_fallbackElement;
  
  /**
   * The global script object, where all embedded functions are defined,
   * as well as the standard ECMA "core" objects.
   *
   * Need to decide whether this is a Dispatcher instance variable, or
   * a static shared by all Dispatchers
   *
   * This may eventually tie in static objects, such as the root element
   * of the DOM
   */
  private static Scriptable global;

  /**
   * Create an object to act as a scope for scripting.
   * Sets the value <code>global</code>
   */
  private static void initGlobal() 
  {
    // Initialize global scope object
    Context cx = new Context();
    try {
      cx.enter();
      global = cx.initStandardObjects(null);
    }
    catch (ThreadLinkException tae) {
      cx = null;
      tae.printStackTrace();
    }
    catch (InterruptedException ie) {
      cx = null;
      ie.printStackTrace();
    }
    finally {
      if (cx != null) {
        try {
          cx.exit();
        }
        catch (ThreadLinkException tae) {
          tae.printStackTrace();
        }
      }
    }      
  }

  /**
   * The LotusXSL processor will find this method via 
   * reflection and will call it when it sees the 
   * lotusxsl:code-dispatcher function.
   */
  public static DispatcherFactory register(String type) 
  {
    initGlobal();
    return new JavaScriptDispatcherFactory();
  }
  
  /**
   * Construct an extension object.
   * @param extensionElem The lotusxsl:extension element.
   * @param fallbackElement The lotusxsl:fallback element associated 
   * with this extension -- just save it and return it when asked.
   */
  public JavaScriptDispatcher(Element extensionElem, Element fallbackElement)
  {
    m_fallbackElement = fallbackElement; 
    evaluateScript(extensionElem);
  }
  
  private Context makeContext()
  {
    Context cx = Context.getCurrentContext();

    if (cx == null) 
    {
      // the current thread does not have a context so make one
      // and attach it.
      cx = new Context();
    }
    return cx;
  }
  
  /**
   * Get the text from the extension node, and 
   * evaluate it.
   * @param extensionElem The lotusxsl:extension element.
   */
  public void evaluateScript(Element extensionElem)
  {
    // Collect the script from the extension element's contents
    String script = null;
    NodeList children = extensionElem.getChildNodes();
    int nChildren = children.getLength();
    for(int i = 0; i < nChildren; i++)
    {
      Node child = children.item(i);
      int nodeType = child.getNodeType();
      switch(nodeType)
      {
      case Node.CDATA_SECTION_NODE:
      case Node.TEXT_NODE:
        if(null == script)
        {
          script = ((Text)child).getData();
        }
        else
        {
          script += ((Text)child).getData();
        }
      }
    }
    
    // evaluate code here.
    Context cx = makeContext();
    try {
      cx.enter();
      cx.evaluateString(global, script, /*"xsl"*/extensionElem.toString(), 0, null);
    }
    catch (ThreadLinkException tae) {
      cx = null;
      tae.printStackTrace();
    }
    catch (InterruptedException ie) {
      cx = null;
      ie.printStackTrace();
    }
    catch (Throwable t) { // includes JavaScriptException, rethrows Errors
      handleError(t);
    }
    finally {
      if (cx != null) {
        try {
          cx.exit();
        }
        catch (ThreadLinkException tae) {
          tae.printStackTrace();
        }
      }
    }
  }
  
  /**
   * Get the fallback element associated with this dispatcher.
   */
  public Element getFallbackElement()
  {
    return m_fallbackElement;
  }
  
  /**
   * Tells if a given function is available.
   */
  public boolean isFunctionAvailable(String method)
  {
      Object fun = global.get(method, global);
      return !(fun == Scriptable.NOT_FOUND);
  }

  /**
   * Return an object from an extension.
   * @param method The name of the method to call.
   * @param args an array of arguments to be 
   * passed to the extension, which may be either 
   * Vectors of Nodes, or Strings.
   */
  public Object call(String method, Object[] args)
  {
    // copy and translate arguments
    args = translateArgList(args);

    // Make the function call here
    Object theReturnValue = null;
    Context cx = makeContext();
    try {
      cx.enter();

      Object fun = global.get(method, global);

      if (fun == Scriptable.NOT_FOUND) {
        throw new JavaScriptException("function " + method + " not found.");
      }

      theReturnValue = ScriptRuntime.call(cx, fun, global, args);
      if (theReturnValue instanceof Wrapper)
      {
        theReturnValue = ((Wrapper)theReturnValue).unwrap();  
      }
    }
    catch (ThreadLinkException tae) {
      cx = null;
      tae.printStackTrace();
    }
    catch (InterruptedException ie) {
      cx = null;

      ie.printStackTrace();
    }
    catch (Throwable t) {
      handleError(t);
    }
    finally {
      if (cx != null) {
        try {
          cx.exit();
        }
        catch (ThreadLinkException tae) {
          tae.printStackTrace();
        }
      }
    }

    return theReturnValue;
  }

  private void handleError(Throwable t)
  {
    if (t instanceof WrappedException) 
    {
      t = (Throwable)((WrappedException)t).unwrap();
    }

    String message = null;
    Throwable dump = null;

    if (t instanceof JavaScriptException) 
    {
      message = t.getLocalizedMessage();

      // Is it an exception wrapped in a JavaScriptException?
      Object value = ((JavaScriptException)t).getValue();
      if (value instanceof Throwable) 
      {
        // likely a wrapped exception from a LiveConnect call.
        // Display its stack trace as a diagnostic
        dump = (Throwable)value;
      }
    }
    else if (t instanceof EvaluatorException
             || t instanceof SecurityException) 
    {
      message = t.getLocalizedMessage();
    }
    else if (t instanceof RuntimeException) 
    {
      message = "Internal Error: " + t.toString();
      dump = t;
    }
    else if (t instanceof StackOverflowError) 
    {
      message = "Stack Overflow";
    }
    else 
    {
      dump = t;
    }

    if (message == null) 
    {
      message = t.toString();
    }

    //REMIND: can we recover the line number here?  I think
    // Rhino does this by looking up the stack for bytecode 
    // see Context.getSourcePositionFromStack()
    // but I don't think this would work in interpreted mode

    //REMIND: pass this off to some sort of GUI or interface?  Just
    // dump to console, for now.
    System.out.println("LotusXSL JavaScript Error: " + message);

    // Print stack trace for diagnostics, where appropriate
    if (dump != null) 
    {
      dump.printStackTrace();
    }
    // t.printStackTrace();

    if (t instanceof Error && !(t instanceof StackOverflowError)) 
    {
      // Re-throw Errors because we're supposed to let the JVM see it
      // Don't re-throw StackOverflows, because we know we've
      // corrected the situation by aborting the loop and
      // a long stacktrace would end up on the user's console
      throw (Error)t;
    }
  }
  
  /**
   * Copies and translates any Vector top-level elements of an Object array
   * into Object arrays.
   *
   * @param args incoming Object array; not modified
   * @return Object array with Vectors translated to arrays
   */
  private Object[] translateArgList(Object[] args)
  {
    //REMIND: construct associative arrays instead of simple arrays,
    // using node name as index as well as integer

    Object[] argList = null;

    if (args != null) {
      argList = new Object[args.length];
      for (int i = 0; i < args.length; i++)
      {
        Object arg = args[i];
        if (arg instanceof Vector)
        {
          // Copy the Vector argument into an Object Array
          Vector vector = (Vector)arg;
          Object[] array = new Object[vector.size()];

          for (int j = 0; j < vector.size(); j++)
          {
            array[j] = vector.elementAt(j);
          }

          arg = array;
        }

        argList[i] = arg;
      }
    }

    return argList;
  }
}

/**
 * This factory object, which is pretty much a singleton, 
 * is created in response to JavaScriptDispatcher.register,
 * and is responsible for creating dispatcher instances.
 */
class JavaScriptDispatcherFactory implements DispatcherFactory
{
  // I'll have to come up with something better than a static 
  // for this... somehow it needs to be keyed to the processor 
  // context.  Maybe I'll pass the LotusXSL instance in, though 
  // then I would need a destruct method.  For now, use the static.
  static JavaScriptDispatcher m_dispatcher = null;
  
  /**
   * Create an extension object.
   */
  public Dispatcher create(Element extensionElem, Element fallbackElem)
  {
    return new JavaScriptDispatcher(extensionElem, fallbackElem);
  }
}

/**
 * $Log: JavaScriptDispatcher.java,v $
 * Revision 1.1  1998/12/18 21:34:27  sboag
 * Adam Peller's version.
 *
 */

