/*****************************************************************************************************
 * $Id: DefaultApplyXSL.java,v 1.1 1999/06/08 23:20:26 trowe Exp $
 * 
 * Copyright (c) 1998-1999 Lotus Corporation, Inc. All Rights Reserved.
 *				This software is provided without a warranty of any kind.
 * 
 * $State: Exp $
 *****************************************************************************************************/

package com.lotus.xsl.server;

import org.w3c.dom.*;
import org.xml.sax.*;
import java.io.*;
import java.net.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import com.lotus.xsl.*;
import com.lotus.xsl.xml4j2dom.XML4JLiaison4dom;

/*****************************************************************************************************
 *
 * ApplyXSL is an abstract class that can be extended to supply the basic 
 * functions for transforming XML data using XSL stylesheets.  
 *
 * @author Spencer Shepard (sshepard@us.ibm.com)
 * @author R. Adam King (raking@us.ibm.com)
 * @author Tom Rowe (trowe@us.ibm.com)
 *
 *****************************************************************************************************/

public abstract class ApplyXSL extends HttpServlet {

        /**
          * Operational parameters for this class.
          * <p>Request-time values override init-time values which override class defaults.</p>
          * @see #init
	  * @serial
          */
        protected ApplyXSLProperties ourDefaultParameters = null;

        /**
          * Mapping of stylesheet result-ns values to MIME types.
          * <p>This mapping is defined by a file pointed to by the operational parameter "mimeURL" which can 
	  *  either contain a full URL or a path relative to the System's server.root /servlets directory.
	  *  If this parameter is not specified, the default mime properties file (located in the System's 
	  *  server.root /servlets directory) will be used</p>
	  * @see #setMimeProps
          * @see #getContentType  
	  * @serial
          */
        protected Properties ourResultToMimeProps = null;

	/**
	  *  String representing the end of line characters for the System
	  */
	protected final static String EOL = System.getProperty("line.separator");

	/**
	  * Initialize operational parameters from the configuration.
	  * @param config Configuration
	  * @exception ServletException Never thrown
	  */
	public void init(ServletConfig config)
	throws ServletException
	{
	    super.init(config);
	    ourDefaultParameters = new ApplyXSLProperties(config);
	    setMimeProps(config.getInitParameter("mimeURL"));
	}

	/**
	  * Loads the result-ns to mime properties file specified by the given string.
	  * The result-ns for HTML will be included in the properties no matter what the result. 
	  * @param mimeURLstring Location for the mime properties file.  Can be either a full URL 
	  * or a path realative to the System's server.root /servlets directory.  If this parameter is
	  * null, the default properties file "mime.properties" will be used out of this directory.
	  */
	protected void setMimeProps(String mimeURLstring)
	{
	    String currentDir = System.getProperty("server.root") + System.getProperty("file.separator") + 
		"servlets" + System.getProperty("file.separator");
	    ourResultToMimeProps = new Properties();
	    // Ensure that the HTML mapping is present even if mimeURL is not available.
	    ourResultToMimeProps.put("http://www.w3.org/TR/REC-html40", "text/html");
	    if (mimeURLstring != null) { 
		URL url = null;
		try {
		    url = new URL(mimeURLstring);
		} catch (MalformedURLException mue1) {
		    try {
			url = new URL("file", "", currentDir + mimeURLstring);
		    } catch (MalformedURLException mue2) {
			writeLog("Unable to find the mime properties file location based on the parameter 'mimeURL' = "
				 + currentDir + mimeURLstring, HttpServletResponse.SC_ACCEPTED, mue2);
			url = null;
		    }  
		}
		if (url != null) {
		    try {		    
			ourResultToMimeProps.load(url.openStream());
		    } catch (IOException ioe1) {
			writeLog("Exception occurred while loading mime properties file: " + mimeURLstring +
				 ".  Mime table may be invalid.", HttpServletResponse.SC_ACCEPTED, ioe1);
		    }
		}
	    } else {
		String defaultProp = currentDir + "mime.properties";
		try {
		    ourResultToMimeProps.load(new FileInputStream(defaultProp));
		} catch (IOException ioe2) {
		    writeLog("Default mime properties file " + defaultProp + " was to be opened, but is not found.",
			 HttpServletResponse.SC_ACCEPTED, ioe2);
		}		
	    }
	}
	
	/**
	  * Implementers of this abstract method must return an XML document DOM.
	  * @param request  May contain or point to the XML document
          * @param listener To record detailed parsing messages for possible return to requestor
	  * @return XML document DOM, or null if no XML document can be created, found, or parsed	  
	  * @see #process
	  * @exception ApplyXSLException Thrown if exception occurs while handling request
	  */
	protected abstract Document getDocument(HttpServletRequest request, ApplyXSLListener listener)
	throws ApplyXSLException;

	/**
	  * Implementers of this abstract method must return an XSL document DOM.
	  * @param request May contain or point to the XSL document
	  * @param xmlDoc  XML document to be transformed
          * @param listener To record detailed parsing messages for possible return to requestor
	  * @return XSL document DOM, or null if no stylesheet can be created, found, or parsed	  
          * @see #process
	  * @exception ApplyXSLException Thrown if exception occurs while handling request
	  */
	protected abstract Document getStylesheet(HttpServletRequest request, Document xmlDoc, 
						  ApplyXSLListener listener)
	throws ApplyXSLException;

	/**
          * Returns the response MIME type using the result-ns (result namespace) attribute of
          * the stylesheet to be applied and result-ns to MIME mapping table.  If a mapping can
          * not be performed, the response MIME type will be set to text/xml.
	  * @param xslDoc XSL document DOM which hopefully contains a result-ns attribute on
          *               an xsl-stylesheet element.
	  * @return The response MIME type to be set
	  * @see #ourResultToMimeProps
	  * @see #process	 
	  */
	public String getContentType(Document xslDoc)
	{	  
	    if (xslDoc != null) { 
		NodeList nl = xslDoc.getElementsByTagName("xsl:stylesheet");
		if ((nl != null) && (nl.getLength() > 0)) {
		    Node styleNode = nl.item(0); // should be only one
		    NamedNodeMap nnm = styleNode.getAttributes();
		    Node n = nnm.getNamedItem("result-ns");
		    if (n != null) {		
			String resultStr = n.getNodeValue();
			if (resultStr.equals("")) 
			    n = nnm.getNamedItem("xmlns");
			else
			    n = nnm.getNamedItem("xmlns:" + resultStr);
			if (n != null)
			    resultStr = n.getNodeValue();		
			if (resultStr != null) {
			    String type = ourResultToMimeProps.getProperty(resultStr);
			    if (type != null)
				return type;   
			}
		    }
		}
	    }
	    return "text/xml"; // default content type
	}
	
	// doPost removed for security reasons

	/**
	  * HTTP Get method passed on to process.
	  * @param request The request
	  * @param response The response
          * @see #process	 
          * @exception ServletException Never thrown
          * @exception IOException Never thrown
	  */
	public void doGet (HttpServletRequest request, 
			   HttpServletResponse response)
	throws ServletException, IOException
	{
		process(request, response);
	} 

	/**
          * Coordinates applying an XSL stylesheet to XML data using operational parameters.
          * <p>If successfully applied, the result tree will be streamed to the response object
          * and the content type set according to the XSL stylesheet's result-ns attribute.</p>
          * <p>If there is a problem in parsing the XML/XSL or if there is a problem in applying
          * the XSL to the XML, an exception will be streamed to the response object.  The detail
          * of the information returned in the response object will depend on whether we're
          * running in debug mode or not.</p>
	  * @param request  May contain information relevant to creating XML and XSL document DOMs
	  * @param response Where to write the transformation result
          * @see #getDocument
          * @see #getStylesheet
          * @see #getContentType
          * @see #displayException
          * @exception ServletException Never thrown
          * @exception IOException Never thrown
	  */
	public void process(HttpServletRequest request, 
			    HttpServletResponse response)
	throws ServletException, IOException
	{
	    //String mode = ourDefaultParameters.getMode(request); not used yet
	    boolean debug = ourDefaultParameters.getDebug(request);

	    long time = 0;
	    if (debug)
		time = System.currentTimeMillis();

	    // Listener has to be used for all reporting
	    ApplyXSLListener listener = new ApplyXSLListener();

	    Document xmlDoc = null, xslDoc = null;
	    // creating XML Document
	    try {	      
		if ((xmlDoc = getDocument(request, listener)) == null)
		    throw new ApplyXSLException("getDocument() returned null" + EOL, 
						new NullPointerException(), 
						response.SC_NOT_FOUND); 
	    } catch (ApplyXSLException axe) {
		axe.appendMessage(EOL + "getDocument() resulted in ApplyXSLException" + EOL 
				  + listener.getMessage());
		if (debug) writeLog(axe);  
		displayException(response, axe, debug);
		xmlDoc = null;
	    }
	    // creating XSL Stylesheet
	    if (xmlDoc != null) {
		try {  
		    if ((xslDoc = getStylesheet(request, xmlDoc, listener)) == null)
			throw new ApplyXSLException("getStylesheet() returned null" + EOL, 
						    new NullPointerException(), 
						    response.SC_NOT_FOUND);  
		} catch (ApplyXSLException axe) {
		    axe.appendMessage(EOL + "getStylesheet() resulted in ApplyXSLException" + EOL 
				      + listener.getMessage());
		    if (debug) writeLog(axe);  
		    displayException(response, axe, debug);
		    xslDoc = null;
		}
	    }
	    // perform Transformation
	    if ((xmlDoc != null) && (xslDoc != null)) {
		response.setContentType(getContentType(xslDoc));
		PrintWriter out = response.getWriter();
		try {
		    // ** Note that LotusXSL specification says a new processor 
		    //  must be created for each transformation **
		    XSLProcessor xslprocessor = new XSLProcessor(new XML4JLiaison4dom());
		    xslprocessor.setStripWhiteSpace(ourDefaultParameters.getStripWS(request));
		    xslprocessor.setQuietConflictWarnings(ourDefaultParameters.getNoCW(request));
		    xslprocessor.setProblemListener(listener);
		    xslprocessor.process(xmlDoc, xslDoc, 
					 ourDefaultParameters.getXSLurl(request), out); // need work here
		    if (debug) writeLog(listener.getMessage(), response.SC_OK);  
		} catch (Exception exc) {
		    ApplyXSLException axe = new ApplyXSLException("Exception occurred during Transformation:" + EOL
								  + listener.getMessage() + EOL
								  + exc.getMessage(), exc, 
								  response.SC_INTERNAL_SERVER_ERROR);
		    if (debug) writeLog(axe);
		    displayException(response, axe, debug);
		} finally {
		    out.close();
		}
	    }
	    if (debug) {
		time = System.currentTimeMillis() - time;
		writeLog("Mode = " + ourDefaultParameters.getMode(request) +
			 "  Strip White Space = " + ourDefaultParameters.getStripWS(request) +
			 "  No Conflict Warnings = " + ourDefaultParameters.getNoCW(request) +
			 "  Transformation time: " + time + " ms", response.SC_OK);
	    }
	}
	
	/**
	  * Writes the following information to the servlet log:
          * <ol>
          * <li>HTTP status code</li>
          * <li>Message</li>
          * <li>Stack trace</li>
          * </ol>
	  * @param axe Contains valid HTTP status code, message,and stack trace (optional)
	  */
	protected void writeLog(ApplyXSLException axe)
	{
	    writeLog(axe.getMessage(), axe.getStatusCode(), axe.getException());
	}

	/**
	  * Writes the following information to the servlet log:
          * <ol>
          * <li>HTTP status code</li>
          * <li>Message</li>
          * <li>Stack trace</li>
          * </ol>
	  * @param msg Message to be logged
	  * @param statusCode Valid status code from javax.servlet.http.HttpServletResponse
	  * @param t Used to generate stack trace (may be =null to suppress stack trace)
	  */
	protected void writeLog(String msg, int statusCode, Throwable t)
	{
	    if (t == null)
		writeLog(msg, statusCode);  
	    ByteArrayOutputStream bytes = new ByteArrayOutputStream();
	    PrintWriter writer = new PrintWriter(bytes, true);
	    t.printStackTrace(writer);	    
	    log("HTTP Status Code: " + statusCode + " - " + msg + EOL + bytes.toString());
	}

	/**
	  * Writes the following information to the servlet log:
          * <ol>
          * <li>HTTP status code</li>
          * <li>Message</li>
          * </ol>
	  * @param msg Message to be logged
	  * @param statusCode Valid status code from javax.servlet.http.HttpServletResponse
	  */
	protected void writeLog(String msg, int statusCode)
	{
	    log("HTTP Status Code: " + statusCode + " - " + msg);
	}

	/**
	  * Invokes response.sendError setting an HTTP status code and optionally an error message.
          * <p>If running in debug mode, also try to return a stack trace of the exception and
          * and xml/xsl processor messages.</p> 
	  * @param response Where to stream the exception to
	  * @param xse The wrapper which contains the exception and its HTTP status code
	  * @param debug Indicates whether to include stack trace, etc.
	  */
	protected void displayException(HttpServletResponse response, ApplyXSLException xse, boolean debug)
	{	    
	    String mesg = xse.getMessage();
	    if (mesg == null)
		mesg = "";
	    else mesg = "<B>" + mesg + "</B>";
	    StringTokenizer tokens = new StringTokenizer(mesg, EOL);
	    StringBuffer strBuf = new StringBuffer();
	    while (tokens.hasMoreTokens())
		strBuf.append(tokens.nextToken() + EOL + "<BR>");
	    mesg = strBuf.toString();
	    if (debug) {
		ByteArrayOutputStream bytes = new ByteArrayOutputStream();
		PrintWriter writer = new PrintWriter(bytes, true);
		xse.getException().printStackTrace(writer);		
		mesg += " <PRE> " + bytes.toString() + " </PRE> ";
	    }
	    response.setContentType("text/html");
	    try {
		response.sendError(xse.getStatusCode(), mesg);
	    } 
	    catch (IOException ioe) {} 
	}
}