/*****************************************************************************************************
 * $Id: DefaultApplyXSL.java,v 1.1 1999/06/08 23:20:26 trowe Exp $
 * 
 * Copyright (c) 1998-1999 Lotus Corporation, Inc. All Rights Reserved.
 *				This software is provided without a warranty of any kind.
 * 
 * $State: Exp $
 *****************************************************************************************************/

package com.lotus.xsl.server;

import javax.servlet.*;
import javax.servlet.http.*;

/*****************************************************************************************************
 * 
 * ApplyXSLProperties contains operational parameters for ApplyXSL based 
 * on program defaults and configuration.  
 * <p>This class is also used to return values for request-time parameters.</p>
 *
 * @author Spencer Shepard (sshepard@us.ibm.com)
 * @author R. Adam King (raking@us.ibm.com)
 * @author Tom Rowe (trowe@us.ibm.com)
 *
 *****************************************************************************************************/

public class ApplyXSLProperties {

    /**
      * Program default for parameter "mode"
      */
    private final String DEFAULT_mode;

    /**
      * Program default for parameter "xslURL"
      */
    private final String DEFAULT_xslURL;
    
    /**
      * Program default for parameter "debug"
      */
    private final boolean DEFAULT_debug;

    /**
      * Program default for parameter "stripWhiteSpace"
      */
    private final boolean DEFAULT_stripWS;

    /**
      * Program default for parameter "noConflictWarnings"
      */
    private final boolean DEFAULT_noCW;
    
    /**
      * Constuctor to use program defaults.
      */
    public ApplyXSLProperties() 
    {
	DEFAULT_mode = null;
	DEFAULT_xslURL = null;
	DEFAULT_debug = false;
	DEFAULT_stripWS = false;
	DEFAULT_noCW = false;
    }

    /**
      * Constructor to use configuration to override program defaults.
      * @param config Servlet configuration
      * @see #setDefaults
      */
    public ApplyXSLProperties(ServletConfig config)
    {
	String md = config.getInitParameter("mode"),
	       xu = config.getInitParameter("xslURL"),
	       db = config.getInitParameter("debug"),
	       sw = config.getInitParameter("stripWhiteSpace"),
	       cw = config.getInitParameter("noConflictWarnings");
	       

	if (md != null) DEFAULT_mode = md;
	else DEFAULT_mode = null;
	if (xu != null) DEFAULT_xslURL = xu;
	else DEFAULT_xslURL = null;
	if (db != null) DEFAULT_debug = new Boolean(db).booleanValue();
	else DEFAULT_debug = false;
	if (sw != null) DEFAULT_stripWS = new Boolean(sw).booleanValue();
	else DEFAULT_stripWS = false;
	if (cw != null) DEFAULT_noCW = new Boolean(cw).booleanValue();
	else DEFAULT_noCW = false;
    }
   
    /**
      * Given a parameter name, returns the HTTP request's String value; if not present in request,
      * returns default String value.
      * @param request Request to check for default override
      * @param param Name of the parameter
      * @return String value of named parameter
      * @see #setDefaults
      */
    public String getRequestParmString(HttpServletRequest request, String param)
    {
	if (request != null) { 
	    String[] paramVals = request.getParameterValues(param); 
	    if (paramVals != null) 
		return paramVals[0];
	}
	return null;
    }

    /**
      * Returns the current setting for "URL".
      * @param request Request to check for parameter value
      * @return String value for "URL"
      */
    public String getXMLurl(HttpServletRequest request)
    {
	return getRequestParmString(request, "URL");
    } 
    
    /**
      * Returns the current setting for "mode".
      * @param request Request to check for parameter value
      * @return String value for "mode"
      */
    public String getMode(HttpServletRequest request)
    {
	String temp = getRequestParmString(request, "mode");
	if (temp != null)
	    return temp;
	return DEFAULT_mode;
    }

    /**
      * Returns the current setting for "xslURL".
      * @param request Request to check for parameter value
      * @return String value for "xslURL"
      */
    public String getXSLurl(HttpServletRequest request)
    {  
	String temp = getRequestParmString(request, "xslURL");
	if (temp != null)
	    return temp;
	return DEFAULT_xslURL;
    }
    
    /**
      * Returns the current setting for "debug".
      * @param request Request to check for parameter value
      * @return Boolean value for "debug"
      */
    public boolean getDebug(HttpServletRequest request)
    {
	String temp = getRequestParmString(request, "debug");
	if (temp != null)
	    return new Boolean(temp).booleanValue();
	return DEFAULT_debug;
    }

    /**
      * Returns the current setting for "stripWhiteSpace".
      * @param request Request to check for parameter value
      * @return Boolean value for "stripWhiteSpace"
      */
    public boolean getStripWS(HttpServletRequest request)
    {
	String temp = getRequestParmString(request, "stripWhiteSpace");
	if (temp != null)
	    return new Boolean(temp).booleanValue();
	return DEFAULT_stripWS;
    }

    /**
      * Returns the current setting for "noConflictWarnings".
      * @param request Request to check for parameter value
      * @return Boolean value for "noConflictWarnings"
      */
    boolean getNoCW(HttpServletRequest request)
    {
	String temp = getRequestParmString(request, "noConflictWarnings");
	if (temp != null)
	    return new Boolean(temp).booleanValue();
	return DEFAULT_noCW;
    }    
}