/**
 * $Id: ProcessXSL.java,v 1.4 1998/12/18 21:35:50 sboag Exp $
 * 
 * Copyright (c) 1998-1999 Lotus Corporation, Inc. All Rights Reserved.
 *				This software is provided without a warranty of any kind.
 * 
 * $State: Exp $
 * 
 * @author Scott Boag (scott_boag@lotus.com)
 */
package com.lotus.xsl.xml4j;

import java.net.*;
import java.io.*;
import java.util.*;
import com.ibm.xml.parser.*;
import com.ibm.xml.parser.util.*;
import org.w3c.dom.*;
import com.lotus.xsl.*;

/**
 * A class that handles interfacing IBM's XML4J to the 
 * Lotus XSL Processor.
 */
public class ProcessXSL extends XMLParserLiaisonDefault
{
  /**
   * Return a string suitible for telling the user what parser is being used.
   */
  public String getParserDescription()
  {
    return "XML4J Version "+ getXML4JVersionString();
  }

  public static final String XML4J_VERSION = "1.1.16";    
  
  private static boolean m_alreadyTriedVersion2 = false;

  /**
   * By default expand all entity references in the 
   * source and style trees.
   */
  public boolean m_shouldExpandEntityRefs = true;
  
  /**
   * A table keyed by namespace URI, with values of 
   * XML4J print visitors.
   */
  private Hashtable m_visitors = new Hashtable();
  
  /**
   * Flag to make sure we only do the version check once.
   */
  private static boolean m_didVersionCheck = false;
  
  public boolean m_supports_userData = false;
  
  /**
   * If this flag is set to false, namespaces won't 
   * be expanded, which may be a performance improvement.
   */
  // I have to be able to distinguish between source and stylesheet to use this.
  // public boolean m_sourceIsNamespaced = true;
  
  /**
   * This tells if it's all right to use the userdata field of 
   * nodes to store the expanded namespace in.
   */
  public boolean m_canUseUserData = false;
  
  /**
   * Use XML4J's PrintVisitor class instead of LotusXSL's Formatter classes.
   */
  public boolean m_xml4jprint = false;
  
  /**
   * Callback object that will be informed when XML-related 
   * warnings or errors occur
   */    
  private com.ibm.xml.parser.ErrorListener m_xmlErrorListener = null;
  
  /**
  * Constructor that takes XML4J ErrorListener as an argument. The error listener
  * is registered with the XML Parser. Any XML-related errors will be reported
  * to the calling application using this error listener.
  *
  * @param	listener XML4J ErrorListener instance.
  */
  public ProcessXSL(ErrorListener listener)
  {
    m_xmlErrorListener = listener;
  }

  /**
   * Construct an ProcessXSL instance, which implements the XMLProcessorLiaison 
   * interface.
   */
  public ProcessXSL()
  {
  }

  /**
   * Install a print visitor that will be visited when the 
   * print function is called, given the result namespace.
   * If the namespace is already installed, it will be 
   * replaced by the passed value.  If visitor is null, 
   * the namespace will be removed.
   */
  void installPrintVisitor(String resultns, com.ibm.xml.parser.Visitor visitor)
  {
    if(null != visitor)
    {
      m_visitors.put(resultns, visitor);
    }
    else
    {
      m_visitors.remove(resultns);
    }
  }
  
  /**
   * Get XML4J's version field... we have to do this 'cause 
   * the "public static final" fields seem to get bound
   * at compile time otherwise.
   */
  private static String getXML4JVersionString()
  {
    String version = "";
    try
    {
      java.lang.reflect.Field versionField = Version.class.getField("S_VERSION");
      version = (String)versionField.get(null);
    }
    catch(Exception e)
    {
    }
    return version;
  }

  /**
   * Get one of the XML4J's version numbers... 
   * we have to do this 'cause the "public static final" 
   * fields seem to get bound at compile time otherwise.
   */
  private static int getXML4JVersionNum(String fieldName)
  {
    int versionNum = 0;
    try
    {
      java.lang.reflect.Field versionField = Version.class.getField(fieldName);
      versionNum = versionField.getInt(null);
    }
    catch(Exception e)
    {
    }
    return versionNum;
  }

  /**
   * Check the XML4J version number
   */
  public static void checkXML4JVersion()
    throws WrongXML4JVersionException
  {

    // Only do this once
    if(!m_didVersionCheck)
    {
      if(!getXML4JVersionString().equals(XML4J_VERSION))
      {
        if(((getXML4JVersionNum("MAJOR") == 1) 
            && (getXML4JVersionNum("MINOR") == 1) 
            && (getXML4JVersionNum("SUBMINOR") >= 9))
           ||
           ((getXML4JVersionNum("MAJOR") == 1) 
            && (getXML4JVersionNum("MINOR") > 1))
           ||
           (getXML4JVersionNum("MAJOR") > 1))
        {
          System.out.println("Found version "+getXML4JVersionString()+" of XML4J. \n"+
                             "ProcessXSL is only tested with version "+
                             XML4J_VERSION+" of XML4J! \n");
        }
        else
        {
            throw new WrongXML4JVersionException("Found version "+getXML4JVersionString()+" of XML4J. \n"+
                              "ProcessXSL doesn't work with versions less than "+
                              XML4J_VERSION+" of XML4J! \n");
        }
      }
      else
      {
        // m_supports_userData = true;
      }
      m_didVersionCheck = true;
    }
  }
  
  /**
   * Parse the text pointed at by the reader as XML, and return 
   * a DOM Document interface.  May return null if not 
   * supported.  It is recommended that you pass in some sort 
   * of recognizable name, such as the filename or URI, with 
   * which the reader can be recognized if the parse fails.
   * Implementation of XMLParserLiaison interface method.
   * 
   * @param reader A stream that should hold valid XML.
   * @param identifier Used for diagnostic purposes only, 
   * some sort of identification for error reporting, may be 
   * null or empty string.
   * @exception IOException   thrown if unable to open the XML stream.
   * @exception XSLProcessorException thrown for all other error conditions.
   */
  public Document parseXMLStream(Reader reader, String filename)
    throws IOException, XSLProcessorException
  {
    // Runtime.getRuntime().traceMethodCalls(false);  
    // Runtime.getRuntime().traceInstructions(false);
    checkXML4JVersion();
    
    Parser xmlParser = null;
    
    if (m_xmlErrorListener == null)
    {
      xmlParser = new Parser(filename);
    }
    else 
    {
      // register error listener with the XML parser
      xmlParser = new Parser (filename, m_xmlErrorListener, null);
    }
            
    // xmlParser.setAllowJavaEncodingName(isallowj);
    xmlParser.setKeepComment(true);
    xmlParser.setWarningNoDoctypeDecl(false);
    xmlParser.setProcessNamespace(true);
    xmlParser.setWarningNoXMLDecl(true);
    xmlParser.setExpandEntityReferences(m_shouldExpandEntityRefs);
    xmlParser.setProcessExternalDTD(m_use_validation);
    // xmlParser.setPreserveSpace(true);
    Document doc = xmlParser.readStream(reader);
    // Runtime.getRuntime().traceMethodCalls(true);  
    // Runtime.getRuntime().traceInstructions(false);
    return doc;
  }
  
  /**
   * Parse the text pointed at by the URL, and return 
   * a DOM Document interface.  Implementation of XMLParserLiaison
   * interface method.
   * @exception IOException   thrown if unable to open the XML stream.
   * @exception XSLProcessorException thrown for all other error conditions.
   */
  public Document parseXMLStream(URL url)
    throws IOException, XSLProcessorException
  {
    Document doc = null;
    InputStream instream = null;
    String whichCall = "";
    
    whichCall = "openConnection";
    URLConnection connect = url.openConnection();
    
    whichCall = "getInputStream";
    instream = connect.getInputStream();
    
    Reader reader = new java.io.BufferedReader(
                                               new java.io.InputStreamReader(instream));
    doc = parseXMLStream(reader, url.toExternalForm());
    return doc;
  }
  
  /**
   * Create an empty DOM Document.  Mainly used for creating an 
   * output document.  Implementation of XMLParserLiaison
   * interface method.
   */
  public Document createDocument()
  {
    return new TXDocument();
  }
  
  /**
   * Given a prefix and a namespace context, return the expanded namespace.
   */
  public String getNamespaceForPrefix(String prefix, Element namespaceContext)
  {
    TXElement txelem = (TXElement)namespaceContext;
    
    return txelem.getNamespaceForPrefix(prefix);
  }
  
  /**
   * Returns the namespace of the given node.
   */
  // public String getNamespaceOfNode(Node n)
  // {
  //   return (n instanceof Namespace) ? 
  //               ((Namespace)n).getNSName() : null;
  // }

  /**
   * Returns the local name of the given node.
   */
  public String getLocalNameOfNode(Node n)
  {
    String qname = n.getNodeName();
    int index = qname.indexOf(':');
    return (index < 0) ? qname : qname.substring(index+1);
  }

  /**
   * Returns the element name with the namespace expanded.
   */
  public String getExpandedElementName(Element elem)
  {
    String ename;
    // if(m_sourceIsNamespaced)
    // {
    TXElement txelem = (TXElement)elem;
    if(m_canUseUserData && m_supports_userData)
    {
      // ename = (String)txelem.getUserData();
      ename = null;
    }
    else
      ename = null;
    if(null == ename)
    {
      ename = txelem.createExpandedName();
      if(m_canUseUserData && m_supports_userData)
      {
        // txelem.setUserData(ename);
      }
    }
    // }
    // else
    // {
    //  ename = elem.getNodeName();
    // }
    return ename;
  }

  /**
   * Returns the attribute name with the namespace expanded.
   */
  public String getExpandedAttributeName(Attr attr)
  {
    // For right now, use getUniversalName instead of 
    // createExpandedName, 'cause createExpandedName has 
    // a bug in it.
    
    // return ((TXAttribute)attr).createExpandedName();
    String ename;
    // if(m_sourceIsNamespaced)
    // {
    TXAttribute txattr = ((TXAttribute)attr);
    if(m_canUseUserData && m_supports_userData)
    {
      ename = null;
      // ename = (String)txattr.getUserData();
    }
    else
      ename = null;
    if(null == ename)
    {
      ename = txattr.getNSName() == null ? txattr.getNSLocalName() : 
                                           (txattr.getNSName()+":"+txattr.getNSLocalName());
      if(m_canUseUserData && m_supports_userData)
      {
        // txattr.setUserData(ename);
      }
    }
    // }
    // else
    // {
    //  ename = attr.getName();
    // }
    return ename;
  }

  /** 
   * Tell if the node is ignorable whitespace.
   */
  public boolean isIgnorableWhitespace(Text node)
  {
    return ((TXText)node).getIsIgnorableWhitespace();
  }
  
  /**
   * Get the parent of a node.
   */
  public Node getParentOfNode(Node node)
    throws RuntimeException
  {
    return (Node.ATTRIBUTE_NODE == node.getNodeType()) 
           ? ((TXAttribute)node).getOwnerElement() : node.getParentNode();
  }
    
  /**
   * Given an ID, return the element.
   */
  public Element getElementByID(String id, Document doc)
  {
    DTD dtd;
    if(null != (dtd = ((TXDocument)doc).getDTD()))
    {
      return dtd.checkID(id);
    }
    return null;
  }
  
  /**
   * Print the result tree.
   * @param doc The result tree.
   * @param pw The PrintWriter to print the contents to.
   * @param resultns the value of the resultns attribute.
   * @param format Tells if it should be pretty-printed.
   * @exception Exception thrown if the result tree is corrupted.
   */
  public void toMarkup(Document doc, PrintWriter pw, String resultns, boolean format)
    throws Exception
  {
    if(!m_xml4jprint)
    {
      super.toMarkup(doc, pw, resultns, format);
    }
    else
    {
      TXDocument txdoc = (TXDocument)doc;
      com.ibm.xml.parser.Visitor visitor = null;
      if(null != resultns)
      {
        Object obj = m_visitors.get(resultns.trim());
        if(null != obj)
        {
          visitor = (com.ibm.xml.parser.Visitor)obj;
        }
        else if(resultns.trim().equals("http://www.w3.org/TR/REC-html40"))
        {
          if(format)
          {
            HTMLFormatPrintVisitor htmlFormatVisitor
              = new HTMLFormatPrintVisitor(pw, null, 
                                           "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n");
            visitor =  htmlFormatVisitor;
            htmlFormatVisitor.indent = m_indent;
            htmlFormatVisitor.m_attrSpecialChars = m_attrSpecialChars;
          }
          else
          {
            visitor = new GoodHTMLPrintVisitor(pw, null, 
                                               "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n");
          }
        }
      }
      if(null != visitor)
      {
        TreeTraversal treeTraversal = new NonRecursivePreorderTreeTraversal(visitor);
        treeTraversal.traverse(txdoc);
      }
      else if(format)
      {
        txdoc.printWithFormat(pw, null, m_indent);
      }
      else
      {
        TreeTraversal treeTraversal 
          = new NonRecursivePreorderTreeTraversal(new GoodToXMLStringVisitor(pw, null));
        treeTraversal.traverse(txdoc);
      }
    }
  }
      
  /**
   * For support of printClassTree... get the last
   * name of a class.
   */
  protected String getLastName(String cl) 
  {
    String ret = cl;
    StringTokenizer st = new StringTokenizer(cl, " .");
    while (st.hasMoreTokens())
      ret = st.nextToken();
    return ret;
  }
  
  /**
   * Diagnostics function to recursivly print an element tree.
   */
  void printClassTree(Parent par, PrintWriter pw, int indent) 
  {
    NodeList ne = par.getChildNodes();    
    int nNodes = ne.getLength();
    for(int i = 0; i < nNodes; i++) 
    {
      Child ch = (Child)ne.item(i);
      // Format.printSpace(pw, indent);
      pw.println(getLastName(ch.getClass().toString())
        +" \""+TXText.makePrintable(ch.getText())+"\"");
      if (ch instanceof Parent) 
      {
        printClassTree((Parent)ch, pw, indent+2);
      }
    }
    pw.flush();
  }

  /**
   * Diagnostics function to print an entire document tree.
   */
  void printDocTree(TXDocument doc, java.io.PrintWriter pw) 
  {
    Element root = (Element)doc.getDocumentElement();
    if((null != root) && (root instanceof Child))
    {
      printClassTree((Parent)root, pw, 0);
    }
  }
  
  /**
   * Get a document Tree as a text string suiteable 
   * for display in a browser.
   */
  public String getTreeAsText(Document docTree)
    throws IOException
  {
    String sourceTreeText = null;
    com.ibm.xml.parser.TXDocument doc = (com.ibm.xml.parser.TXDocument)docTree;
    com.ibm.xml.parser.TXText textNode;
    {
      StringWriter osw = new StringWriter();
      PrintWriter pw = new PrintWriter(osw, false);
      // doc.print(pw);
      com.ibm.xml.parser.Visitor visitor 
        = new com.ibm.xml.parser.FormatPrintVisitor(pw);
      try
      {
        new com.ibm.xml.parser.NonRecursivePreorderTreeTraversal(visitor).traverse(doc);
      }
      catch(Exception e)
      {
        throw new RuntimeException(e.getMessage());
      }
      String xmlString = osw.toString();
      textNode = (com.ibm.xml.parser.TXText)doc.createTextNode(xmlString);
    }
    {
      StringWriter osw = new StringWriter();
      PrintWriter pw = new PrintWriter(osw, false);
      textNode.print(pw);
      sourceTreeText = osw.toString();
    }
    return sourceTreeText;
  }
   
  /**
   * Command line interfact to transform the XML according to 
   * the instructions found in the XSL document.
   */
  public static void main( String argv[] )    
  {    
    String[] argv2 = new String[argv.length+2];
    argv2[0] = "-PARSER";
    argv2[1] = "com.lotus.xsl.xml4j.ProcessXSL";
    int n = argv2.length;
    for(int i = 2; i < n; i++)
    {
      argv2[i] = argv[i-2];
    }
    com.lotus.xsl.Process.main(argv2);
  }
        
  class GoodHTMLPrintVisitor extends HTMLPrintVisitor
  {
    public GoodHTMLPrintVisitor(Writer writer, String encoding, String doctype) 
    {
      super(writer, encoding, doctype);
    }
    
    public void visitAttributePre(TXAttribute attribute) throws Exception 
    {
      String aname = attribute.getName().toLowerCase();
      if(!aname.equals("xmlns") && !aname.startsWith("xmlns:"))
      {
        // super.visitAttributePre(attribute);
        this.writer.write(" ");
        this.writer.write(attribute.getName());
        this.writer.write("=\"");
        for (Node child = attribute.getFirstChild();  child != null;
             child = child.getNextSibling()) 
        {
          int type = child.getNodeType();
          if (type == Node.TEXT_NODE) 
          {
            this.writer.write(Util.backReference(child.getNodeValue(), m_attrSpecialChars, this.encoding));
          }
          else if (type == Node.ENTITY_REFERENCE_NODE) 
          {
            this.writer.write("&");
            this.writer.write(child.getNodeName());
            this.writer.write(";");
          }
          else 
          {
            //???
          }
        }
        this.writer.write("\"");
        
        // throw new ToNextSiblingTraversalException();
      }
    }

  }
  
  class GoodToXMLStringVisitor extends ToXMLStringVisitor
  {
    public GoodToXMLStringVisitor(Writer writer, String encoding) 
    {
      super(writer, encoding);
    }
    
    public void visitAttributePre(TXAttribute attribute) throws Exception 
    {
      // super.visitAttributePre(attribute);
      this.writer.write(" ");
      this.writer.write(attribute.getName());
      this.writer.write("=\"");
      for (Node child = attribute.getFirstChild();  child != null;
           child = child.getNextSibling()) 
      {
        int type = child.getNodeType();
        if (type == Node.TEXT_NODE) 
        {
          this.writer.write(Util.backReference(child.getNodeValue(), m_attrSpecialChars, this.encoding));
        }
        else if (type == Node.ENTITY_REFERENCE_NODE) 
        {
          this.writer.write("&");
          this.writer.write(child.getNodeName());
          this.writer.write(";");
        }
        else 
        {
          //???
        }
        this.writer.write("\"");
        
        // throw new ToNextSiblingTraversalException();
      }
    }

  }

  
}


/**
 * $Log: ProcessXSL.java,v $
 * Revision 1.4  1998/12/18 21:35:50  sboag
 * Removed -T option -- use lotusxsl:css-style-conversion instead.
 *
 * Revision 1.3  1998/12/11 08:43:03  sboag
 * Added Process() APIs
 *
 * Revision 1.2  1998/12/10 03:18:55  sboag
 * LotusXSL Version 13a
 *
 */
