
/*******************************************************************
 * 
 *							DominoXMLXSLServlet
 * 
 *	This is a very basic servlet that illustrate some core
 *  techniques that can be used to dynamically create XML based
 *  from the contents of a Domino .nsf database, and then translate
 *  the XML to HTML via an XSL stylesheet, using the LotusXSL 
 *  Processor.
 * 
 *  The LotusXSL Processor can be found on IBM's AlphaWorks site:
 *      http://www.alphaworks.ibm.com/formula/LotusXSL
 *  Also required is IBM's XML4J, also on the AlphaWorks site:
 *      http://www.alphaworks.ibm.com/formula/xml
 * 
 * Copyright 1999: Lotus Development Corporation
 *				   All rights reserved.
 * This sample code is supplied without warranty of any kind.
 * 
 *******************************************************************/

import java.io.*;
import java.util.Vector;
import javax.servlet.*;
import javax.servlet.http.*;
import lotus.domino.*;
import com.lotus.xsl.*;
	  

public class DominoXMLXSLServlet extends HttpServlet
{

	//
	//	Variables needed to access the Domino database.
	//	See documentation on the Domino R5 Java API for 
	//	more information.
	//
	
	protected Session notesSession;
	protected Database notesDatabase;
	protected String notesDataBaseName;
	protected String defaultNotesDataBaseName = "RealEstate.nsf";
	protected View view;
	protected String notesViewName;
	protected String defaultNotesViewName = "byBedrooms";
	
	//
	//  The XSL stylesheet directive that we'll put in the output, if 
	//  any.
	//
	String styleSheetname;
	


	/*******************************************************************
	 * 
	 *							doGet
	 * 
	 *	Servlets get called here to respond to http "get" requests.
	 *  This is where we do all the work to retrieve the XML.
	 * 
	 *******************************************************************/
	
	public void doGet(HttpServletRequest request,HttpServletResponse response)
		throws ServletException, IOException        
	{	
		
		// -------------------------------------------------------
		//		 Re-establish defaults in case servlet gets reused
		//		 (To save overhead, web servers often keep servlets
		//       around and reuse them from one get request to the
		//       next.  Values from earlier runs can still be in
		//		 the variables.)
		// -------------------------------------------------------
		
		notesDataBaseName=defaultNotesDataBaseName;
		notesViewName = defaultNotesViewName;
		styleSheetname = null;

		// -------------------------------------------------------
		//		 Get query request from servlet URL
		//		 and prepare response stream
		// -------------------------------------------------------

		// Get the query string following the "?" in the URL
		String query = request.getQueryString();
		
		// Parse the query string to find out what database
		// and view to open.  This sets notesDatabaseName, etc.
		parseQuery(query);
		
		// Get the response stream: this is how XML is sent back to browser
		PrintWriter out=response.getWriter();

    // Create an output stream into which the XML will be written.
    ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
		PrintWriter xmlWriter = new PrintWriter(byteStream);

		// set content type and other response header fields first
		response.setContentType("text/html");  	

		// -------------------------------------------------------
		// Open the Notes db and write all of the XML for this request
		// -------------------------------------------------------
		
		try
		{
			//
			//	Initialize Notes: open the database and view
			//
			NotesThread.sinitThread();     // this call needed before any other Notes calls
			notesSession = NotesFactory.createSession();
			notesDatabase = notesSession.getDatabase(null, notesDataBaseName);
			view = notesDatabase.getView(notesViewName);
			
			//
			//	Print first few lines of XML
			//
			
			xmlWriter.println("<?xml version =\"1.0\"?>");
			
			//  If we put a stylesheet reference into the XML, 
			//  then some browsers will actually find the 
			//  style sheet and use XSL to format
			//  the XML.
			if (styleSheetname != null) 
				xmlWriter.println("<?xml:stylesheet href=\"http://" + 
							styleSheetname +
							"\" type=\"text/xsl\" ?>");
			xmlWriter.println();
			xmlWriter.println("<!-- This XML generated automatically from Domino database=\"" + 
						notesDataBaseName + 
						"\".  View name=\"" + notesViewName +
						"\".  -->");
			xmlWriter.println();
			
			//
			//	Print the XML for the data in the Notes database
			//
			xmlWriter.println("<HOUSES>");

			Document notesDocument;
			for (notesDocument = view.getFirstDocument();
				 notesDocument != null;
				 notesDocument = view.getNextDocument(notesDocument)) 
			{
				xmlWriter.println();
				xmlWriter.println("    <HOUSE>");
				printXMLforNotesField(out,"HOUSETYPE", notesDocument.getItemValueString("Type"));
				printXMLforNotesField(out,"BEDROOMS", new Integer(notesDocument.getItemValueInteger("Beds")));
				printXMLforNotesField(out,"BATHROOMS",new Integer(notesDocument.getItemValueInteger("Baths")));
				printXMLforNotesField(out,"PRICE","$" + notesDocument.getItemValueInteger("Price"));
				xmlWriter.println("    </HOUSE>");
			}
			
			xmlWriter.println();
			xmlWriter.println("</HOUSES>");
      
      // Create an input stream from an output stream.
      Reader xmlReader 
        = new java.io.InputStreamReader(new ByteArrayInputStream(byteStream.toByteArray()));
      
      try
      {
        // Create an LotusXSL processor.  The class that is handed in 
        // is the liaison from the IBM XML parser to the XSL processor.
        XSLProcessor xslprocessor 
          = new XSLProcessor(new com.lotus.xsl.xml4j.ProcessXSL());
        
        // Have the XSL processor read the XML, use the href pointed to 
        // by the stylesheet PI we inserted above to obtain a stylesheet, 
        // and pipe the resulting HTML to the output result writer.
        xslprocessor.process(xmlReader, out);
      }
      catch(Error err)
      {
        out.println("<H3>Server call to LotusXSL Failed: "
                               +err.getMessage()+"</H3>");
        java.util.Properties props = System.getProperties();
        props.list(out);
        out.println();
      }
      catch(Exception exc)
      {
        out.println("<H3>LotusXSL threw an exception on the server: "
                               +exc.getMessage()+"</H3>");
      }
		}
		
		// -------------------------------------------------------
		//  If anything goes wrong, print an error message as a comment in the XML
		// -------------------------------------------------------
		catch(Exception e) 
		{
			//TODO: http error here?
			out.println("<!-- DominoXMLServlet: Error in servlet,  JavaException: " + e.toString()
						+ " -->");
		}
		
		// -------------------------------------------------------
		//	Whether or not there's been an error, clean up
		//  and exit.
		// -------------------------------------------------------
		finally
		{
			NotesThread.stermThread();      //MUST balance sInitThread above
			out.close();
		}
	}

	/*******************************************************************
	 *				printXMLforNotesField
	 * 
	 *		This subroutine formats the output of a single XML tag for a single
	 *		Notes field.  The form of the tag written is:
	 * 
	 *				<XMLELEMENTTAG>
	 *					contents of Notes field
	 *				</XMLELEMENTTAG>
	 * 
	 *		We'll use fieldContents.toString() to get the contents to print
	 *		
	 * 
	 *******************************************************************/

	protected void printXMLforNotesField(PrintWriter out, 
										 String XMLElementTag, 
										 Object fieldContents)
		throws IOException, NotesException
	{
		//
		//	If we got a field, print it
		//
		if (fieldContents != null) 
		{
			debugMessage(out, "DEBUG: ready to print " + XMLElementTag);
			out.println("\t<" + XMLElementTag + ">");
			//
			//  If you want to reformat or convert values to change the
			//  representation, this is a good place to do it.
			//
			out.println("\t\t" + fieldContents.toString());
			out.println("\t</" + XMLElementTag + ">");
		} 
		else
		{
			out.println("\t<" + XMLElementTag + "NOTESDATAFOUND = FALSE/>");
		}
	}

	/*******************************************************************
	 *				parseQuery
	 * 
	 *		For demo purposes, we parse an URL that looks like:
	 * 
	 *		http://yourserver.yoursite.com/servlet/DominoXMLServlet?dbname,dbview
	 * 
	 *		Here we see and parse just the dbname, dbview
	 * 
	 *		To keep the demo code simple, no error checking is done.
	 *		We'll rely on the Domino API to reflect errors in the 
	 *		db or view names.
	 * 
	 *******************************************************************/
	
	// maximum number of positional parameters
	static final int maxArgs = 3;
	
	// helper function to find the commas in the argument list.
	// array looks like (-1, comma1pos, comma2pos, .... , string length, -1, -1)

	private int[] getCommaPositions(String q)
	{
		int i;
		int lastComma = -1;
		int retVal[] =  new int[maxArgs+2]; // make sure there's an extra -1 at beginning

		for (i=0; i<retVal.length; i++)
			retVal[i]=-1;
		for (i=1; i<retVal.length; i++) // starts at index 1
		{
			lastComma = q.indexOf(",",lastComma + 1);
			if (lastComma == (-1)) 
			{
				retVal[i] = q.length(); // pretend there's a comma at the end
				break;
			}
			retVal[i] = lastComma;
		};
		return retVal;
	}
	
	// helper function to return a positional parm
	String getParm(String q, int argno, int commaPositions[])
	{
		// the char following the previous comma, if any
		int start = commaPositions[argno] + 1;
		// the position of the next comma, if any
		int pastEnd = commaPositions[argno + 1];
		
		// end of string is always marked with positive length, but
		// after that are -1's
		if (pastEnd < 0)
			return null;
		// if ",," return null
		if ((pastEnd - start) == 1)
			return null;
		// there's a real parm, return it
		return q.substring(start,pastEnd);
				
	}
	
	protected void parseQuery(String q)
	{
		// if no string, all parms have default values
		if (q == null || q.length() == 0)
			return;

		int commas[] = getCommaPositions(q);
		
		notesDataBaseName = getParm(q, 0, commas);
		notesViewName = getParm(q, 1, commas);
		styleSheetname = getParm(q, 2, commas);

		return;
		
	}
	
	
	

	
	/*******************************************************************
	 * 
	 *				debugMessage
	 * 
	 *		Uncomment the body of this to see debugging messages as
	 *		HTML responses to a browser.
	 * 
	 *******************************************************************/

	void debugMessage(PrintWriter pw, String s) 
	{
		/*		  pw.println("<P>");
		pw.println(s);
		pw.println("</P1>");
		pw.flush(); */
	}

}

