/* $Id: MsqlResultSet.java,v 2.1.2.3 1999/07/09 23:02:56 borg Exp $ */
/* Copyright (c) 1997-1999 George Reese */
package com.imaginary.sql.msql;

import com.imaginary.util.Encoder;
import com.imaginary.util.NoSuchEncoderException;
import com.sun.java.util.collections.ArrayList;
import com.sun.java.util.collections.HashMap;
import com.sun.java.util.collections.Iterator;
import com.sun.java.util.collections.Map;
import java.io.ByteArrayOutputStream;
import java.io.CharArrayReader;
import java.io.InputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Reader;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.DatabaseMetaData;
import java.sql.Date;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.Statement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.Time;
import java.sql.Timestamp;
import java.sql.Types;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.util.Calendar;
import java.util.Locale;

/**
 * Acts as an abstract base class for all of the different kinds of result
 * sets supported by mSQL-JDBC. 
 * <BR>
 * Last modified $Date: 1999/07/09 23:02:56 $
 * @version $Revision: 2.1.2.3 $
 * @author George Reese (borg@imaginary.com)
 */
public abstract class MsqlResultSet implements ResultSet {
    // The logging object for this result set
    protected MsqlLog       log              = null;
    // The Statement that owns this ResultSet
    private   MsqlStatement statement        = null;

    MsqlResultSet(MsqlStatement stmt, int ll) throws SQLException {
	super();
	statement = stmt;
	log = new MsqlLog(ll, this);
    }
	
    /**
     * mSQL does not generate warnings, so this method is a NO-OP
     * @exception java.sql.SQLException this is never thrown
     */
    public void clearWarnings() throws SQLException {
	log.log("clearWarnings()", MsqlLog.JDBC, "Clear warnings.");
    }

    /**
     * A NO-OP for this base class.
     * @exception java.sql.SQLException this is never thrown
     */
    public void close() throws SQLException {
	log.log("close()", MsqlLog.JDBC, "Closing result set.");
	log.close();
    }

    /**
     * This method is called to mark a result set as complete.
     */
    public void complete() {
	log.log("complete()", MsqlLog.DRIVER, "Result set complete.");
	if( statement != null ) {
	    statement.completeLoad();
	}
    }
    
   /**
     * For performance reasons, you should get values by column
     * number when at all possible.
     * @param cname the name of the desired column
     * @return an ASCII input stream for the column
     * @exception java.sql.SQLException thrown when the column cannot be read
     */
    public InputStream getAsciiStream(String cname) throws SQLException {
	return getAsciiStream(findColumn(cname));
    }

    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the column being sought
     * @return the specified column as a BigDecimal
     * @exception java.sql.SQLException thrown when the column cannot be read
     */
    public BigDecimal getBigDecimal(String cname) throws SQLException {
	return getBigDecimal(findColumn(cname));
    }
    
    /**
     * @param column the column number of the desired column
     * @return the specified column as a BigDecimal
     * @exception java.sql.SQLException thrown when the column cannot be read
     */
    public BigDecimal getBigDecimal(int column) throws SQLException {
	String tmp = getString(column);

	if( tmp == null ) {
	    return new BigDecimal(new BigInteger("0"));
	}
	else {
	    return new BigDecimal(new BigInteger(tmp));
	}
    }

    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the column being sought
     * @param scale the scale with which the BigDecimal should be constructed
     * @return the specified column as a BigDecimal
     * @exception java.sql.SQLException thrown when the column cannot be read
     * @deprecated
     */
    public BigDecimal getBigDecimal(String cname, int scale)
    throws SQLException {
	return getBigDecimal(findColumn(cname), scale);
    }
    
    /**
     * @param column the column number of the desired column
     * @param scale the scale with which the BigDecimal should be constructed
     * @return the specified column as a BigDecimal
     * @exception java.sql.SQLException thrown when the column cannot be read
     * @deprecated
     */
    public BigDecimal getBigDecimal(int column, int scale)throws SQLException {
	String tmp = getString(column);

	if( tmp == null ) {
	    return new BigDecimal(new BigInteger("0"), scale);
	}
	else {
	    return new BigDecimal(new BigInteger(tmp), scale);
	}
    }

    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as an InputStream
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public InputStream getBinaryStream(String cname) throws SQLException {
	return getBinaryStream(findColumn(cname));
    }
    
    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as a boolean
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public boolean getBoolean(String cname) throws SQLException {
	return getBoolean(findColumn(cname));
    }
    
    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as a byte
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public byte getByte(String cname) throws SQLException {
	return getByte(findColumn(cname));
    }
    
    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as a byte array
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public byte[] getBytes(String cname) throws SQLException {
	return getBytes(findColumn(cname));
    }
    
    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as a Reader
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Reader getCharacterStream(String cname) throws SQLException {
	return getCharacterStream(findColumn(cname));
    }
    
    /**
     * @param column the number of the desired column
     * @return the column as a Reader
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Reader getCharacterStream(int column) throws SQLException {
	return new CharArrayReader(getString(column).toCharArray());
    }
  
    /**
     * @return the cursor name
     * @exception java.sql.SQLException this is always thrown as mSQL does
     * not support named cursors
     */
    public String getCursorName() throws SQLException {
	log.log("getCursorName()", MsqlLog.ERROR, "Cursors not supported.");
	throw new SQLException("mSQL does not support cursors.");
    }

    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as a Date
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Date getDate(String cname) throws SQLException {
	return getDate(findColumn(cname));
    }

    /**
     * mSQL 1.0 does not support a date type, but 2.0 does.  This method
     * allows any version of mSQL to store dates as a CHAR type.  It also
     * allows mSQL 2.0 to store them as TEXT or DATE types.  If you choose to
     * store your date as a CHAR or TEXT type for any reason, you must store
     * the date as a long converted into a CHAR or in DD-Mon-YYYY.
     * @param column the number of desired column
     * @return a Date representation of the specified column
     * @exception java.sql.SQLException a database error occurred
     */
    public Date getDate(int column) throws SQLException {
	String tmp = getString(column);  

	if( tmp == null ) {
	    return null;
	}
	try {
	    SimpleDateFormat fmt = new SimpleDateFormat("dd-MMM-yyyy",
							Locale.US);
	    java.util.Date date = fmt.parse(tmp);

	    return new Date(date.getTime());
	}
	catch( ParseException e ) {
	    try {
		return new Date(Long.parseLong(tmp));
	    }
	    catch( NumberFormatException e2 ) {
		log.log("getDate()", MsqlLog.ERROR,
			"Invalid date format: " + tmp);
		throw new SQLException("Invalid date format.");
	    }
	}
    }

    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @param cal the Calendar to use in constructing the Date
     * @return the column as a Date
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Date getDate(String cname, Calendar cal) throws SQLException {
	return getDate(findColumn(cname), cal);
    }

    /**
     * mSQL 1.0 does not support a date type, but 2.0 does.  This method
     * allows any version of mSQL to store dates as a CHAR type.  It also
     * allows mSQL 2.0 to store them as TEXT or DATE types.  If you choose to
     * store your date as a CHAR or TEXT type for any reason, you must store
     * the date as a long converted into a CHAR or in DD-Mon-YYYY.
     * @param column the number of desired column
     * @param cal the Calendar to use in constructing the date
     * @return a Date representation of the specified column
     * @exception java.sql.SQLException a database error occurred
     */
    public Date getDate(int column, Calendar cal) throws SQLException {
	String tmp = getString(column);  

	if( tmp == null ) {
	    return null;
	}
	try {
	    SimpleDateFormat fmt = new SimpleDateFormat("dd-MMM-yyyy",
							Locale.US);
	    java.util.Date date;

	    fmt.setCalendar(cal);
	    date = fmt.parse(tmp);
	    return new Date(date.getTime());
	}
	catch( ParseException e ) {
	    try {
		return new Date(Long.parseLong(tmp));
	    }
	    catch( NumberFormatException e2 ) {
		log.log("getDate()", MsqlLog.ERROR,
			"Invalid date format: " + tmp);
		throw new MsqlException("Invalid date format.");
	    }
	}
    }

    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as a double
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public double getDouble(String cname) throws SQLException {
	return getDouble(findColumn(cname));
    }
    
    /**
     * @param column the number of the desired column
     * @return the column as a double
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public double getDouble(int column) throws SQLException {
	String tmp = getString(column);

	if( tmp == null ) {
	    return 0;
	}
	try {
	    return Double.valueOf(tmp).doubleValue();
	}
	catch( NumberFormatException e ) {
	    log.log("getDouble()", MsqlLog.ERROR,
		    "Invalid number format: " + tmp + ".");
	    throw new MsqlException(e);
	}
    }  

    public String getEncoding() {
	if( statement != null ) {
	    return statement.getEncoding();
	}
	else {
	    return "8859_1";
	}
    }

    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as a float
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public float getFloat(String cname) throws SQLException {
	return getFloat(findColumn(cname));
    }
    
    /**
     * @param column the number of the desired column
     * @return the column as a float
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public float getFloat(int column) throws SQLException {
	String tmp = getString(column);

	if( tmp == null ) {
	    return 0;
	}
	try {
	    return Float.valueOf(tmp).floatValue();
	}
	catch( NumberFormatException e ) {
	    log.log("getFloat()", MsqlLog.ERROR,
		    "Invalid number format: " + tmp);
	    throw new MsqlException(e);
	}
    }  

    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as an int
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public int getInt(String cname) throws SQLException {
	return getInt(findColumn(cname));
    }
    
    /**
     * @param column the number of the desired column
     * @return the column as an int
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public int getInt(int column) throws SQLException {
	String tmp = getString(column);

	if( tmp == null ) {
	    return 0;
	}
	try {
	    return Integer.parseInt(tmp);
	}
	catch( NumberFormatException e ) {
	    log.log("getInt()", MsqlLog.ERROR,
		    "Invalid number format: " + tmp + ".");
	    throw new MsqlException(e);
	}
    }

    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as a long
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public long getLong(String cname) throws SQLException {
	return getLong(findColumn(cname));
    }
    
    /**
     * @param column the number of the desired column
     * @return the column as a long
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public long getLong(int column) throws SQLException {
	String tmp = getString(column);

	if( tmp == null ) {
	    return 0;
	}
	try {
	    return Long.parseLong(tmp);
	}
	catch( NumberFormatException e ) {
	    log.log("getLong()", MsqlLog.ERROR, 
		    "Invalid numver format: " + tmp + ".");
	    throw new MsqlException(e);
	}
    }  

    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as a Java object based on the standard
     * SQL -> Java mapping
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Object getObject(String cname) throws SQLException {
	return getObject(findColumn(cname));
    }
  
    /**
     * @param column the number of the desired column
     * @return the column as a Java object based on the standard
     * SQL -> Java mapping
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Object getObject(int column) throws SQLException {
	ResultSetMetaData meta = getMetaData();
	int type = meta.getColumnType(column);

	switch(type) {
	case Types.BIT:
	    return new Boolean(getBoolean(column));
	    
	case Types.TINYINT:
	    return new Character((char)getInt(column));
	    
	case Types.SMALLINT:
	    return new Integer(getShort(column));
	    
	case Types.INTEGER:
	    return new Integer(getInt(column));
	    
	case Types.BIGINT:
	    return new Long(getLong(column));
	    
	case Types.FLOAT:
	    return new Float(getFloat(column));
	    
	case Types.REAL:
	    return new Float(getFloat(column));
	    
	case Types.DOUBLE:
	    return new Double(getDouble(column));
	    
	case Types.NUMERIC:
	    return getBigDecimal(column, 0);

	case Types.DECIMAL:
	    return getBigDecimal(column, 0);

	case Types.CHAR:
	    return getString(column);
	    
	case Types.VARCHAR:
	    return getString(column);
	    
	case Types.LONGVARCHAR:
	    return getString(column);
	    
	case Types.DATE:
	    return getDate(column);
	    
	case Types.TIME:
	    return getTime(column);
	    
	case Types.TIMESTAMP:
	    return getTimestamp(column);
	    
	case Types.BINARY:
	    return getBytes(column);
	    
	case Types.VARBINARY:
	    return getBytes(column);
	    
	case Types.LONGVARBINARY:
	    return getBytes(column);
	}
	log.log("getObject()", MsqlLog.ERROR, "Unknown SQL type " + type +".");
	throw new MsqlException("Unknown SQL type: " + type);
    }

    public abstract int getRow() throws SQLException;
    
    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as a short
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public short getShort(String cname) throws SQLException {
	return getShort(findColumn(cname));
    }
    
    /**
     * @param column the number of the desired column
     * @return the column as a short
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public short getShort(int column) throws SQLException {
	String tmp = getString(column);

	if( tmp == null ) {
	    return 0;
	}
	try {
	    return (short)Short.parseShort(tmp);
	}
	catch( NumberFormatException e ) {
	    log.log("getShort()", MsqlLog.ERROR,
		    "Invalid number format: " + tmp + ".");
	    throw new MsqlException(e);
	}
    }

    /**
     * @return the Statement that generated this result set
     * @exception java.sql.SQLException this is never thrown
     */
    public Statement getStatement() throws SQLException {
	return statement;
    }
    
    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as a Java String
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public String getString(String cname) throws SQLException {
	return getString(findColumn(cname));
    }
    
    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as a Time object
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Time getTime(String cname) throws SQLException {
	return getTime(findColumn(cname));
    }
    
    /**
     * @param column the number of the desired column
     * @return the column as a Java String
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Time getTime(int column) throws SQLException {
	String tmp = getString(column);  

	if( tmp == null ) {
	    return null;
	}
	try {
	    SimpleDateFormat fmt = new SimpleDateFormat("HH:mm:ss",
							Locale.US);
	    java.util.Date time = fmt.parse(tmp);

	    return new Time(time.getTime());
	}
	catch( ParseException e ) {
	    try {
		return new Time(Long.parseLong(tmp));
	    }
	    catch( NumberFormatException e2 ) {
		log.log("getTime()", MsqlLog.ERROR,
			"Invalid time format: " + tmp + ".");
		throw new SQLException("Invalid time format.");
	    }
	}
    }

    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @param cal the Calendar to use in creating the Time object
     * @return the column as a Time object
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Time getTime(String cname, Calendar cal) throws SQLException {
	return getTime(findColumn(cname), cal);
    }
    
    /**
     * @param column the number of the desired column
     * @param cal the Calendar to use in creating the Time object
     * @return the column as a Java String
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Time getTime(int column, Calendar cal) throws SQLException {
	String tmp = getString(column);  

	if( tmp == null ) {
	    return null;
	}
	try {
	    SimpleDateFormat fmt = new SimpleDateFormat("HH:mm:ss",
							Locale.US);
	    java.util.Date time;

	    fmt.setCalendar(cal);
	    time = fmt.parse(tmp);
	    return new Time(time.getTime());
	}
	catch( ParseException e ) {
	    try {
		return new Time(Long.parseLong(tmp));
	    }
	    catch( NumberFormatException e2 ) {
		log.log("getTime()", MsqlLog.ERROR,
			"Invalid time format: " + tmp + ".");
		throw new SQLException("Invalid time format.");
	    }
	}
    }

    /**
     * <P>
     * Timestamps are not really supported in mSQL.  To make use of
     * them, store timestamps in char fields.
     * </P>
     * <P>
     * Avoid getting columns by name whenever possible.
     * </P>
     * @param cname the name of the desired column
     * @return the column as a Timestamp
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Timestamp getTimestamp(String cname) throws SQLException {
	return getTimestamp(findColumn(cname));
    }

    /**
     * <P>
     * Timestamps are not really supported in mSQL.  To make use of
     * them, store timestamps in char fields.
     * </P>
     * @param column the number of the desired column
     * @return the column as a Timestamp
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Timestamp getTimestamp(int column) throws SQLException {
	String tmp = getString(column);

	if( tmp == null ) {
	    return null;
	}
	else {
	    return new Timestamp(Long.parseLong(tmp));
	}
    }

    /**
     * <P>
     * Timestamps are not really supported in mSQL.  To make use of
     * them, store timestamps in char fields.
     * </P>
     * <P>
     * Avoid getting columns by name whenever possible.
     * </P>
     * @param cname the name of the desired column
     * @param cal this is currently ignored
     * @return the column as a Timestamp
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Timestamp getTimestamp(String cname, Calendar cal)
    throws SQLException {
	return getTimestamp(findColumn(cname), cal);
    }

    /**
     * <P>
     * Timestamps are not really supported in mSQL.  To make use of
     * them, store timestamps in char fields.
     * </P>
     * @param column the number of the desired column
     * @param c this is currently ignored
     * @return the column as a Timestamp
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     */
    public Timestamp getTimestamp(int column, Calendar c) throws SQLException {
	String tmp = getString(column);

	if( tmp == null ) {
	    return null;
	}
	else {
	    return new Timestamp(Long.parseLong(tmp));
	}
    }

    /**
     * Avoid getting columns by name whenever possible.
     * @param cname the name of the desired column
     * @return the column as an InputStream
     * @exception java.sql.SQLException thrown in the event of an error
     * reading the column
     * @deprecated use the getCharacterStream()
     */
    public InputStream getUnicodeStream(String cname) throws SQLException {
	return getUnicodeStream(findColumn(cname));
    }

    /**
     * mSQL does not generate warnings, so this method always returns null.
     * @return null
     * @exception java.sql.SQLException this is never thrown
     */
    public SQLWarning getWarnings() throws SQLException {
	return null;
    }
    
    /**
     * Determines if the driver is positioned after the last row.
     * @return true if the result set is positioned after the last row
     * @exception java.sql.SQLException this is never thrown
     */
    public boolean isAfterLast() throws SQLException {
	return (getRow() == -2);
    }
    
    /**
     * Because of the way the driver works, isAfterLast() == isBeforeFirst().
     * @return true if the result set is positioned before the first row
     * @exception java.sql.SQLException this is never thrown
     */
    public boolean isBeforeFirst() throws SQLException {
	return (getRow() == 0);
    }
}
