/*
 * (C) Copyright IBM Corp. 1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package SAXPrint; 

import XJParseHandler;

import java.io.File;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.net.URL;

import org.xml.sax.AttributeList;
import org.xml.sax.HandlerBase;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.Parser;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.ParserFactory;

/**
 * Parses an XML document using a specified SAX Parser and prints
 * out canonical XML format. The output format follows the James
 * Clark <tt>XMLTest</tt> program with the following exceptions:
 * <ul>
 * <li> XMLTest incorrectly outputs processing instructions. A
 *      processing instruction with a <i>null</i> data value will
 *      print as <em>&lt;?target null?&gt;</em> instead of as
 *      <em>&lt?target?&gt;</em>. This class outputs the latter.
 * </ul>
 *
 * @author  Andy Clark, IBM
 * @version 1.0
 */
public class SAXPrintHandlers
    extends HandlerBase implements XJParseHandler
    {

    public void registerHandlers(Parser parser) {
		parser.setDocumentHandler(this);
        parser.setErrorHandler(this);
    }
    /**
     *
     */
    public void reset() {
    }

	public void printResults(String label, long duration) {
	}

    //
    // Data
    //

    /** Output stream. */
    private PrintStream out;

    /** Compatability mode. */
    private boolean compatable = false;

    //
    // Constructors
    //

    /** Default constructor. */
    public SAXPrintHandlers() {
        this(System.out);
        }

    /** Constructs a tester with specified print stream. */
    public SAXPrintHandlers(OutputStream out) {
        if (!(out instanceof PrintStream)) {
            out = new PrintStream(out);
            }
        this.out = (PrintStream)out;
        }

    //
    // Public methods
    //

    /** Sets the compatability mode. */
    public void setCompatable(boolean compatable) {
        this.compatable = compatable;
        }

    /** Returns true if the tester is in compatability mode. */
    public boolean isCompatable() {
        return compatable;
        }

    //
    // DocumentHandler methods
    //

    /** Characters. */
    public void characters(char ch[], int start, int length) {

        for (int i = start; i < start + length; i++) {
            print(ch[i]);
            }

        } // characters(char[],int,int)

    /** Ignorable whitespace. */
    public void ignorableWhitespace(char ch[], int start, int length) {
        characters(ch, start, length);
        }

    /** Start element. */
    public void startElement(String name, AttributeList atts) {

        out.print('<');
        out.print(name);
        int len = atts.getLength();
        for (int i = 0; i < len; i++) {
            out.print(' ');
            out.print(atts.getName(i));
            out.print("=\"");
            String value = atts.getValue(i);
            int valueLength = value.length();
            for (int j = 0; j < valueLength; j++) {
                print(value.charAt(j));
                }
            out.print('"');
            }
        out.print('>');

        } // startElement(String,AttributeList)

    /** End element. */
    public void endElement(String name) {

        out.print("</");
        out.print(name);
        out.print('>');

        } // endElement(String)

    /** Processing instruction. */
    public void processingInstruction(String target, String data) {

        out.print("<?");
        out.print(target);
        if (compatable) {
            out.print(' ');
            out.print(data);
            }
        else {
            if (data != null) {
                out.print(' ');
                out.print(data);
                }
            }
        out.print("?>");

        } // processingInstruction(String,String)

    //
    // Private methods
    //

    /** Normalizes a character when printing to the print stream. */
    private void print(char ch) {

        switch (ch) {
            case '&': {
                out.print("&amp;");
                break;
                }
            case '<': {
                out.print("&lt;");
                break;
                }
            case '>': {
                out.print("&gt;");
                break;
                }
            case '"': {
                out.print("&quot;");
                break;
                }
            case '\t': {
                if (compatable) {
                    out.print("&#9;");
                    }
                else {
                    out.print(ch);
                    }
                break;
                }
            case '\n': {
                if (compatable) {
                    out.print("&#10;");
                    }
                else {
                    out.print(ch);
                    }
                break;
                }
            case '\r': {
                if (compatable) {
                    out.print("&#12;");
                    }
                else {
                    out.print(ch);
                    }
                break;
                }
            default: {
                out.print(ch);
                }
            }

        } // print(char)

    /** Expands a system id. */
    private static String expandSystemId(String systemId) {

        // see if it's already in URL form
        try {
            return new URL(systemId).toString();
            }
        catch (MalformedURLException e) {
            }

        // try to make it a file: URL
        try {
            URL base = new URL("file", "", new File(".").getAbsolutePath().replace(File.separatorChar, '/'));
            return new URL(base, systemId.replace(File.separatorChar, '/')).toString();
            }
        catch (MalformedURLException e) {
            }

        // I tried my best
        return systemId;

        } // expandSystemId(String):String

    /** Prints the usage information. */
    private static void printUsage() {
        System.err.println("usage: java "+
                           SAXPrintHandlers.class.getName()+
                           " [option ...] (file | uri) ...");
        System.err.println();
        System.err.println("options:");
        System.err.print("  -p parser_class  Sets the SAX parser to use.");
        Parser parser = null;
        try { parser = ParserFactory.makeParser(); }
        catch (Exception e) {}
        if (parser == null) {
            System.err.print(" No default parser.");
            }
        else {
            System.err.println();
            System.err.print("                   Default parser: "+parser.getClass().getName());
            }
        System.err.println();
        System.err.println("  -c               Compatability mode for James Clark's XMLTest.");
        System.err.println("  -h | -?          Usage information.");
        }

    //
    // Classes
    //

    /**
     * UTF-8 output stream. 
     *
     * @author  Andy Clark, IBM
     * @version 1.0
     */
    private static class UTF8Printer
        extends FilterOutputStream
        {

        //
        // Constants
        //
        
        /** High surrogate start. */
        private static final int HIGH_SURROGATE_START = 0x0D800;

        /** High surrogate end. */
        private static final int HIGH_SURROGATE_END = 0x0DBFF;

        /** Low surrogate start. */
        private static final int LOW_SURROGATE_START = 0x0D800;

        /** Low surrogate end. */
        private static final int LOW_SURROGATE_END = 0x0DBFF;

        //
        // Data
        //

        /** High surrogate. */
        private int surrogate = -1;

        //
        // Constructors
        //

        /** Default constructor. */
        public UTF8Printer(OutputStream out) {
            super(out);
            }

        //
        // OutputStream methods
        //

        /** Writes a character. */
        public void write(int ch) 
            throws IOException
            {

            if (surrogate != -1) {
                if (ch >= LOW_SURROGATE_START && ch <= LOW_SURROGATE_END) {
                    // high: [110110ww] [wwzzzzyy] (surrogate)
                    // low:  [110111yy] [yyxxxxxx] (ch)
                    int wwww   = (surrogate >> 6) & 0x003F;
                    int uuuuu  = wwww + 1;
                    int zzzz   = (surrogate >> 2) & 0x003F;
                    int yyyyyy = ((surrogate & 0x0003) << 4) |
                                 ((ch >> 6) & 0x003F);
                    int xxxxxx = ch & 0x003F;

                    // out: [11110uuu] [10uuzzzz] [10yyyyyy] [10xxxxxx]
                    super.out.write(0x00F0 | (uuuuu >> 2));
                    super.out.write(0x0080 | ((uuuuu & 0x0007) << 4) | zzzz);
                    super.out.write(0x0080 | yyyyyy);
                    super.out.write(0x0080 | xxxxxx);

                    // reset
                    surrogate = -1;
                    }
                else {
                    throw new IOException("invalid low surrogate value: 0x"+Integer.toString(ch, 16));
                    }
                }

            // 0x0000 -> 0x007F
            else if (ch < 0x0080) {

                // ch: [00000000] [0xxxxxxx]
                int xxxxxxx = ch;

                // out: [0xxxxxxx]
                super.out.write(xxxxxxx);
                }

            // 0x0080 -> 0x07FF
            else if (ch < 0x0800) {

                // ch: [00000yyy] [yyxxxxxx]
                int yyyyy  = ch >> 6;
                int xxxxxx = ch & 0x003F;

                // out: [110yyyyy] [10xxxxxx]
                super.out.write(0x00C0 | yyyyy);
                super.out.write(0x0080 | xxxxxx);
                }

            // surrogate blocks
            else if (ch >= HIGH_SURROGATE_START && ch <= HIGH_SURROGATE_END) {
                // save character for next surrogate character
                surrogate = ch;
                }

            // 0x0800 -> 0xFFFF -- minus surrogate blocks
            else if (ch < 0x10000) {

                // ch: [zzzzyyyy] [yyxxxxxx]
                int zzzz   = ch >> 12;
                int yyyyyy = (ch >> 6) & 0x003F;
                int xxxxxx = ch & 0x003F;

                // out: [1110zzzz] [10yyyyyy] [10xxxxxx]
                super.out.write(0x00E0 | zzzz);
                super.out.write(0x0080 | yyyyyy);
                super.out.write(0x0080 | xxxxxx);
                }

            else {
                throw new IOException("invalid Unicode character: 0x"+Integer.toString(ch, 16));
                }

            }

        } // class UTF8Printer

    //
    // MAIN
    //

    /** Main program entry point. */
    public static void main(String argv[]) {

        // check command line args
        if (argv.length == 0) {
            printUsage();
            System.exit(1);
            }

        int start = 0;
        String parserClass = null;
        boolean compatable = false;
        for ( ; start < argv.length; start++) {
            String arg = argv[start];
            if (arg.startsWith("-")) {
                if (arg.equals("-h") || arg.equals("-?")) {
                    printUsage();
                    System.exit(0);
                    }
                else if (arg.equals("-p")) {
                    if (++start < argv.length) {
                        parserClass = argv[start];
                        }
                    else {
                        System.err.println("error: missing parameter to -p");
                        System.exit(1);
                        }
                    }
                else if (arg.equals("-c")) {
                    compatable = true;
                    }
                else {
                    System.out.println("error: unknown option ("+arg+")");
                    System.exit(1);
                    }
                }
            else {
                break;
                }
            }

        // create handler
        SAXPrintHandlers handler = new SAXPrintHandlers();
        handler.setCompatable(compatable);

        // run through files
        for (int i = start; i < argv.length; i++) {
            
            // create parser
            Parser parser = null;
            try {
                if (parserClass != null) {
                    parser = ParserFactory.makeParser(parserClass);
                    }
                else {
                    parser = ParserFactory.makeParser();
                    }
                }
            catch (Exception e) {
                System.err.println("error: Unable to instantiate parser.");
                System.exit(1);
                }

            // set handlers
            parser.setDocumentHandler(handler);
            parser.setDTDHandler(handler);
            parser.setErrorHandler(handler);

            // parse file
            try {
                System.err.println("uri["+(i-start+1)+'/'+(argv.length-start)+"]="+argv[i]);
                parser.parse(expandSystemId(argv[i]));
                }
            catch (Exception e) {
                System.err.println("error: Error during parse.");
                System.exit(1);
                }
            }

        // all's well
        System.exit(0);

        } // main(String[])

    } // class SAXPrintHandlers
