/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.dom;

import org.w3c.dom.*;

/**
 * Text nodes hold the non-markup, non-Entity content of
 * an Element or Attribute.
 * <P>
 * When a document is first made available to the DOM, there is only
 * one Text object for each block of adjacent plain-text. Users (ie,
 * applications) may create multiple adjacent Texts during editing --
 * see {@link Element#normalize} for discussion.
 * <P>
 * Note that CDATASection is a subclass of Text. This is conceptually
 * valid, since they're really just two different ways of quoting
 * characters when they're written out as part of an XML stream.
 *
 * @version Revision: 41 1.2 src/com/ibm/xml/dom/DeferredTextImpl.java, parser, xml4j2, xml4j2_0_13 
 * @since  PR-DOM-Level-1-19980818.
 */
public class DeferredTextImpl 
    extends TextImpl 
    implements DeferredNode {

    //
    // Constants
    //

    /** Serialization version. */
    static final long serialVersionUID = 2310613872100393425L;
    
    //
    // Data
    //

    /** Node index. */
    protected transient int fNodeIndex;

    //
    // Constructors
    //

    /**
     * This is the deferred constructor. Only the fNodeIndex is given here.
     * All other data, can be requested from the ownerDocument via the index.
     */
    DeferredTextImpl(DeferredDocumentImpl ownerDocument, int nodeIndex) {
        super(ownerDocument, null);

        fNodeIndex = nodeIndex;
        syncData = true;

    } // <init>(DeferredDocumentImpl,int)
  
    //
    // DeferredNode methods
    //

    /** Returns the node index. */
    public int getNodeIndex() {
        return fNodeIndex;
    }

    //
    // Protected methods
    //

    /** Synchronizes the underlying data. */
    protected void synchronizeData() {

        // no need for future synchronizations
        syncData = false;

        // get initial text value
        DeferredDocumentImpl ownerDocument = (DeferredDocumentImpl)this.ownerDocument;
        value = ownerDocument.getNodeValueString(fNodeIndex);

        // revisit: we only normalize text nodes with Elements as parents.
        int parent = -1;
        if (getNodeType() == Node.TEXT_NODE &&
            (parent = ownerDocument.getParentNode(fNodeIndex)) != -1 && 
            ownerDocument.getNodeType(parent) == Node.ELEMENT_NODE) {

            int realNext = ownerDocument.getRealNextSibling(fNodeIndex);
            int type     = ownerDocument.getNodeType(realNext);
            if (realNext != -1 && type == Node.TEXT_NODE) {
                StringBuffer sb = new StringBuffer(value);
                while (realNext != -1 && type == Node.TEXT_NODE) {
                    sb.append(ownerDocument.getNodeValue(realNext));
                    realNext = ownerDocument.getRealNextSibling(realNext);
                    type = ownerDocument.getNodeType(realNext);
                }
                value = sb.toString();
            }
        }

        // ignorable whitespace
        ignorableWhitespace = ownerDocument.getFirstChild(fNodeIndex) == 1;
        
    } // synchronizeData()
    
} // class DeferredTextImpl
