/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.dom;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;

import org.w3c.dom.*;

/**
 * NamedNodeMaps represent collections of Nodes that can be accessed
 * by name. They're currently used in two modes. Attributes are
 * placed in a NamedNodeMap rather than being children of the node
 * they describe. On the other hand Entity and Notation appear both
 * in the NamedNodeMap _and_ as kids of the DocumentType, requiring
 * some additional logic so these are maintained as "live views" of
 * each other.
 * <P>
 * Only one Node may be stored per name; attempting to
 * store another will replace the previous value.
 * <P>
 * NOTE: The storage key is taken from the NodeName attribute of the
 * node. As a result, Node types that have fixed names can only have a
 * single instance stored in the NamedNodeMap. This is a DOM
 * restriction.
 * <P>
 * NOTE: item()'s integer index does _not_ imply that the named nodes
 * must be stored in an array; that's only an access method. Note too
 * that these indices are "live"; if someone changes the map's
 * contents, the indices associated with nodes may change.
 * <P>
 * As of REC-DOM-Level-1-19981001, Entities and Notations are no longer
 * shadowed as children of DocumentType.
 * <P>
 * Looked at basing this on Hashtable. Doing so would save about 250 bytes
 * of classfile, at the expense of wasting memory at runtime... and the
 * performance gains for such short sets are questionable. For
 * now I'm going to stick with sorted vectors and binary search.
 *
 * @version Revision: 65 1.6 src/com/ibm/xml/dom/NamedNodeMapImpl.java, parser, xml4j2, xml4j2_0_13 
 * @since  PR-DOM-Level-1-19980818.
 */
public class NamedNodeMapImpl
    implements NamedNodeMap, Serializable {

    //
    // Constants
    //

    /** Serialization version. */
    static final long serialVersionUID = -7039242451046758020L;

    //
    // Data
    //

    /** Nodes. */
	protected Vector nodes = new Vector();

    /** Owner document. */
	protected Document ownerDocument; // support for WRONG_DOCUMENT_ERR

    /**
     * Element. Only named node maps holding attributes for elements
     * have an element object. This value is here to support the
     * Attr#getElement method (DOM Level 2, 19 June 1999).
     */
    protected ElementImpl element;

    /** Default nodes. */
	protected NamedNodeMapImpl defaults;

    /** Changes. */
	protected int changes;	// Support for reconcileDefaults

    /** Last defaults changes. */
	protected int lastDefaultsChanges = -1;

    /** Read-only. */
	protected boolean readOnly = false;

    //
    // Constructors
    //

    /** Constructs a named node map. */
    protected NamedNodeMapImpl(Document ownerDoc, NamedNodeMapImpl defaults) {
	    ownerDocument = ownerDoc;
	    this.defaults = defaults;
    }

    /** Constructs a named node map. */
    protected NamedNodeMapImpl(ElementImpl element, NamedNodeMapImpl defaults) {
        this(element.getOwnerDocument(), defaults);
        this.element = element;
    }

    //
    // NamedNodeMap methods
    //

    /**
     * Report how many nodes are currently stored in this NamedNodeMap.
     * Caveat: This is a count rather than an index, so the
     * highest-numbered node at any time can be accessed via
     * item(getLength()-1).
     * <P>
     * COMPLICATION: In the case of attributes,
     * the count has to include any defaults that may be inherited, yet
     * not double-count when there is both a default and a specified
     * value. Convolving the two lists is expensive, and for GC reasons
     * we don't want to register with the DTD (it wouldn't know when to
     * release us).
     * <P>
     * One solution is to go into the change-counter domain, as we did for
     * DeepNodeList. Another is to accept the convolution, and count
     * on the fact that our implementation is a sorted list to keep the cost
     * managable... which works pretty well for getLength(), but makes
     * item() expensive.
     * <P>
     * The ideal fix would be to rearchitect to eliminate integer indexing,
     * but of course that wouldn't meet the spec!
     */
    public int getLength() {
    	reconcileDefaults();
    	return (nodes != null) ? nodes.size() : 0;
    }

    /**
     * Retrieve an item from the map by 0-based index.
     *
     * @param index Which item to retrieve. Note that indices are just an
     * enumeration of the current contents; they aren't guaranteed to be
     * stable, nor do they imply any promises about the order of the
     * NamedNodeMap's contents. In other words, DO NOT assume either that
     * index(i) will always refer to the same entry, or that there is any
     * stable ordering of entries... and be prepared for double-reporting
     * or skips as insertion and deletion occur.
     *
     * @returns the node which currenly has the specified index, or null
     * if index is greater than or equal to getLength().
     */
    public Node item(int index) {
    	reconcileDefaults();
    	return (nodes != null && index < nodes.size()) ?
                    (Node)(nodes.elementAt(index)) : null;
    }

    /**
     * Retrieve a node by name. If not explicitly defined, checks the
     * defaults before giving up.
     *
     * @param name Name of a node to look up.
     * @returns the Node (of unspecified sub-class) stored with that name,
     * or null if no value has been assigned to that name.
     */
    public Node getNamedItem(String name) {

    	int i = findNamePoint(name);
        return (i < 0) ? null : (Node)(nodes.elementAt(i));

    } // getNamedItem(String):Node

    /**
     * Internal routine: Only update this structure.
     * @see setNamedItem
     * @return org.w3c.dom.Node
     * @param arg org.w3c.dom.Node
     * @exception org.w3c.dom.DOMException The exception description.
     */
    public Node setNamedItem(Node arg)
        throws DOMException {

    	if(arg.getOwnerDocument() != ownerDocument) {
            throw new DOMExceptionImpl(DOMException.WRONG_DOCUMENT_ERR,null);
        }

    	if (arg instanceof AttrImpl && ((AttrImpl)arg).owned) {
            throw new DOMExceptionImpl(DOMException.INUSE_ATTRIBUTE_ERR,null);
        }

    	int i = findNamePoint(arg.getNodeName());
    	Node previous = null;
    	if (i >= 0) {
    		previous = (Node)nodes.elementAt(i);
    		nodes.setElementAt(arg,i);
    	}
    	else {
    		i = -1 - i; // Insert point (may be end of list)
    		if (null == nodes) {
    			nodes = new Vector();
            }
    		nodes.insertElementAt(arg, i);
        }			

        // change owning element
        if (element != null) {
            ((NodeImpl)arg).parentNode = element;
        }

    	++changes;
    	return previous;

    } // setNamedItem(Node):Node

    /**
     * Remove an entry from the NamedNodeMap. @see removeNamedItem.
     * @return org.w3c.dom.Node
     * @param name java.lang.String
     * @exception org.w3c.dom.DOMException The exception description.
     */
    public Node removeNamedItem(String name)
        throws DOMException {

    	int i = findNamePoint(name);
    	if (i < 0) {
    		throw new DOMExceptionImpl(DOMException.NOT_FOUND_ERR, null);
        }

        Node n = (Node)(nodes.elementAt(i));
        // If there's a default, add it instead
        Node d;
        if (defaults != null &&
            (d = defaults.getNamedItem(name)) != null) {
            nodes.setElementAt(d, i);
        }
        else {
            nodes.removeElementAt(i);
        }

        // remove owning element
        if (element != null) {
            ((NodeImpl)n).parentNode = null;
        }

        ++changes;
        return n;

    } // removeNamedItem(String):Node

    //
    // Public methods
    //

    /**
     * Cloning a NamedNodeMap is a DEEP OPERATION; it always clones
     * all the nodes contained in the map.
     * <P>
     * ????? Currently, this does _not_ clone the docType reference.
     * Should the new docType object (if any) be a parameter, rather
     * than being set manually later on?
     * <P>
     * ????? We _do_ clone the defaults reference, if any.
     */
    public NamedNodeMapImpl cloneMap() {

    	boolean deep = true;
    	NamedNodeMapImpl newmap = new NamedNodeMapImpl(ownerDocument, defaults);
    	if (nodes != null) {
    		newmap.nodes = new Vector(nodes.size());
    		for (int i = 0; i < nodes.size(); ++i) {
    			newmap.nodes.addElement(((NodeImpl) nodes.elementAt(i)).cloneNode(deep));
            }
        }
    	newmap.defaults = defaults;
    	return newmap;

    } // cloneMap():NamedNodeMapImpl

    //
    // Package methods
    //

    /**
     * Internal subroutine to allow read-only Nodes to make their contained
     * NamedNodeMaps readonly too. I expect that in fact the shallow
     * version of this operation will never be
     *
     * @param readOnly boolean true to make read-only, false to permit editing.
     * @param deep boolean true to pass this request along to the contained
     * nodes, false to only toggle the NamedNodeMap itself. I expect that
     * the shallow version of this operation will never be used, but I want
     * to design it in now, while I'm thinking about it.
     */
    void setReadOnly(boolean readOnly, boolean deep) {

    	this.readOnly = readOnly;
    	if(deep && nodes != null) {
    		Enumeration e=nodes.elements();
    		while(e.hasMoreElements()) {
    			((NodeImpl)e.nextElement()).setReadOnly(readOnly,deep);
            }
    	}

    } // setReadOnly(boolean,boolean)

    //
    // Protected methods
    //

    /**
     * Subroutine: If this NamedNodeMap is backed by a "defaults" map (eg,
     * if it's being used for Attributes of an XML file validated against
     * a DTD), we need to deal with the risk that those defaults might
     * have changed. Entries may have been added, changed, or removed, and
     * if so we need to update our version of that information
     * <P>
     * Luckily, this currently applies _only_ to Attributes, which have a
     * "specified" flag that allows us to distinguish which we set manually
     * versus which were defaults... assuming that the defaults list is being
     * maintained properly, of course.
     * <P>
     * Also luckily, I made the decision to maintain NamedNodeMaps as
     * sorted lists. This should keep the cost of convolving the two lists
     * managable... not wonderful, but at least more like 2N than N**2..
     * <P>
     * Finally, to avoid doing the convolution except when there are actually
     * changes to be absorbed, I've made the Map aware of whether or not
     * its defaults Map has changed. This is not 110% reliable, but it should
     * work under normal circumstances, especially since the DTD is usually
     * relatively static information.
     * <P>
     * Note: This is NON-DOM implementation, though used to support behavior
     * that the DOM requires.
     */
    protected void reconcileDefaults() {

    	if (defaults != null && lastDefaultsChanges != defaults.changes) {

    		int n = 0;
            int d = 0;
            int nsize = nodes.size();
            int dsize = defaults.nodes.size();

    		Attr nnode = (nsize == 0) ? null : (Attr) nodes.elementAt(0);
    		Attr dnode = (dsize == 0) ? null : (Attr) defaults.nodes.elementAt(0);

    		while (n < nsize && d < dsize) {
    			nnode = (Attr) nodes.elementAt(n);
    			dnode = (Attr) defaults.nodes.elementAt(d);
    			int test = nnode.getNodeName().compareTo(dnode.getNodeName());

    			// Same name and a default -- make sure same value
    			if (test == 0 && !nnode.getSpecified()) {
    				nodes.setElementAt(dnode, n);
    				// Advance over both, since names in sync
    				++n;
    				++d;
    			}

    			// Different name, new default in table; add it
    			else if (test > 0) {
    				nodes.insertElementAt(dnode, n);
    				// Now in sync, so advance over both
    				++n;
    				++d;
    			}

    			// Different name, old default here; remove it.
    			else if (!nnode.getSpecified()) {
    				nodes.removeElementAt(n);
    				// n didn't advance but represents a different element
    			}

    			// Different name, specified; accept it
                else {
    				++n;
                }
        	}

    		// If we ran out of local before default, pick up defaults
    		while (d < dsize) {
    			nodes.addElement(defaults.nodes.elementAt(d++));
    		}
    		lastDefaultsChanges = defaults.changes;
    	}

    } // reconcileDefaults()

    //
    // Private methods
    //

    /**
     * Subroutine: Locate the named item, or the point at which said item
     * should be added. Factoring this out saved very few bytes, but improved
     * code readability.
     *
     * SIDE EFFECT: Resynchronizes w/ Defaults. That code could
     * be moved directly in here to save a few bytes, but for now I like
     * the conceptual separation.
     *
     * @param name Name of a node to look up.
     *
     * @return If positive or zero, the index of the found item.
     * If negative, index of the appropriate point at which to insert
     * the item, encoded as -1-index and hence reconvertable by subtracting
     * it from -1. (Encoding because I don't want to recompare the strings
     * but don't want to burn bytes on a datatype to hold a flagged value.)
     */
    private int findNamePoint(String name) {

    	reconcileDefaults();

    	// Binary search
    	int i = 0;
    	if(nodes != null) {
    		int first = 0;
            int last  = nodes.size() - 1;

    		while (first <= last) {
    			i = (first + last) / 2;
    			int test = name.compareTo(((Node)(nodes.elementAt(i))).getNodeName());
    			if(test == 0) {
    				return i; // Name found
                }
    			else if (test < 0) {
    				last = i - 1;
                }
    			else {
    				first = i + 1;
                }
    		}

    		if (first > i) {
                i = first;
            }
    	}

    	return -1 - i; // not-found has to be encoded.

    } // findNamePoint(String):int

} // class NamedNodeMapImpl
