/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.dom;

import java.io.*;

import org.w3c.dom.*;

/**
 * Node provides the basic structure of a DOM tree. It is never used
 * directly, but instead is subclassed to add type and data
 * information, and additional methods, appropriate to each node of
 * the tree. Only its subclasses should be instantiated -- and those,
 * with the exception of Document itself, only through a specific
 * Document's factory methods.
 * <P>
 * Node itself implements shared behaviors such as siblings and
 * children, both for consistancy and so that the most common tree
 * operations may be performed without constantly having to downcast
 * to specific node types. When there is no obvious mapping for one of
 * these queries, it will respond with null.
 * <P>
 * NodeImpl also implements NodeList, so it can return itself in
 * response to the getChildNodes() query. This eliminiates the need
 * for a separate ChildNodeList object. Note that this is an
 * IMPLEMENTATION DETAIL; applications should _never_ assume that
 * this identity exists.
 * <P>
 * Note that the default behavior is that children are forbidden. To
 * permit them, the subclass's constructor sets isLeafNode false.
 * <P>
 * LEVEL 1 RESTRICTION: All nodes in a single document must originate
 * in that document. (Note that this is much tighter than "must be
 * same implementation") Nodes are all aware of their ownerDocument,
 * and attempts to mismatch will throw WRONG_DOCUMENT_ERR.
 *
 * @version 66 1.3 src/com/ibm/xml/dom/NodeImpl.java, parser, xml4j2 
 * @since  PR-DOM-Level-1-19980818.
 */
public abstract class NodeImpl 
    implements Node, NodeList, Cloneable, Serializable {

    //
    // Constants
    //

    /** Serialization version. */
    static final long serialVersionUID = 2815829867052120872L;
    
    // public
    
    /** Element definition node type. */
    public static final short ELEMENT_DEFINITION_NODE = -1;

    //
    // Data
    //

    // links

    /** Owner document. */
	protected DocumentImpl ownerDocument;

    /** Parent node. */
    protected NodeImpl parentNode;

    /** Previous sibling. */
	protected NodeImpl previousSibling;

    /** Next sibling. */
    protected NodeImpl nextSibling;

    // data

    /** Node name. */
	protected String name;

    /** Node value. */
	protected String value;

    /** Read-only property. */
	protected boolean readOnly;

	/** NON-DOM FEATURE; see setUserData/getUserData. **/
	protected Object userData;
	
    // children

    /** First child. */
	protected NodeImpl firstChild;

    /** Last child. */
    protected NodeImpl lastChild;

    // lazy-evaluation nifo

    /** Synchronization of child nodes needed. */
    protected transient boolean syncChildren;

    /** Synchronization of data needed. */
    protected transient boolean syncData;

    // internal data

	/**
     * Number of alterations made to this subtree since its creation.
	 * Serves as a "dirty bit" so NodeList can recognize when an
	 * alteration has been made and discard its cached state information.
	 * <p>
	 * Any method that alters the tree structure MUST cause or be
	 * accompanied by a call to changed(), to inform it and its
	 * parents that any outstanding NodeLists may have to be updated.
     * <p>
	 * (Required because NodeList is simultaneously "live" and integer-
	 * indexed -- a bad decision in the DOM's design.)
	 * <p>
	 * Note that changes which do not affect the tree's structure -- changing
	 * the node's name, for example -- do _not_ have to call changed().
	 * <p>
	 * Alternative implementation would be to use a cryptographic
	 * Digest value rather than a count. This would have the advantage that
	 * "harmless" changes (those producing equal() trees) would not force
	 * NodeList to resynchronize. Disadvantage is that it's slightly more prone
	 * to "false negatives", though that's the difference between "wildly
	 * unlikely" and "absurdly unlikely". IF we start maintaining digests,
	 * we should consider taking advantage of them.
     */
	int changes = 0;

    /** Table for quick check of child insertion. */
	protected static int[] kidOK;

    //
    // Static initialization
    //

	static {

		kidOK = new int[13];
		
		kidOK[DOCUMENT_NODE] = 
			1 << ELEMENT_NODE | 1 << PROCESSING_INSTRUCTION_NODE | 
			1 << COMMENT_NODE | 1 << DOCUMENT_TYPE_NODE;
			
		kidOK[DOCUMENT_FRAGMENT_NODE] = 
		kidOK[ENTITY_NODE] = 
		kidOK[ENTITY_REFERENCE_NODE] = 
		kidOK[ELEMENT_NODE] = 
			1 << ELEMENT_NODE | 1 << PROCESSING_INSTRUCTION_NODE | 
			1 << COMMENT_NODE | 1 << TEXT_NODE |
			1 << CDATA_SECTION_NODE | 1 << ENTITY_REFERENCE_NODE;
			
		kidOK[ATTRIBUTE_NODE] = 
			1 << TEXT_NODE | 1 << ENTITY_REFERENCE_NODE;
			
        kidOK[DOCUMENT_TYPE_NODE] =
		kidOK[PROCESSING_INSTRUCTION_NODE] = 
		kidOK[COMMENT_NODE] = 
		kidOK[TEXT_NODE] = 
		kidOK[CDATA_SECTION_NODE] = 
		kidOK[NOTATION_NODE] = 
			0;

    } // static

    //
    // Constructors
    //

    /**
     * No public constructor; only subclasses of Node should be
     * instantiated, and those normally via a Document's factory methods
     * <p>
     * Every Node knows what Document it belongs to.
     */
    protected NodeImpl(DocumentImpl ownerDocument, String name, String value) {

        // set information
        this.ownerDocument = 
            (ownerDocument != null) ? ownerDocument : (DocumentImpl)this;

        this.name  = name;
        this.value = value;

    } // <init>(DocumentImpl,String,short,boolean,String)

    /** Constructor for serialization. */
    public NodeImpl() {}

    //
    // Node methods
    //

    /** 
     * A short integer indicating what type of node this is. The named
     * constants for this value are defined in the org.w3c.dom.Node interface.
     */
    public abstract short getNodeType();

    /** 
     * For many nodes, nodeName is derived from their nodeType mnemonic
     * (lowercasified and preceeded by a '#' character). Others return the
     * "name of the data" -- tag name, attribute name, and so on.  
     */
    public String getNodeName() {

        if (syncData) {
            synchronizeData();
        }
        return name;

    } // getNodeName():String

    /**
     * Any node which can have a nodeValue (@see getNodeValue) will
     * also accept requests to set it to a string.  The exact response to
     * this varies from node to node -- Attribute, for example, stores
     * its values in its children and has to replace them with a new Text
     * holding the replacement value.
     * <p>
     * For most types of Node, value is null and attempting to set it
     * will throw DOMException(NO_MODIFICATION_ALLOWED_ERR). This will
     * also be thrown if the node is read-only.
     */
    public void setNodeValue(String value) {

    	if (readOnly)
    		throw new DOMExceptionImpl(
    			DOMException.NO_MODIFICATION_ALLOWED_ERR, null);
        // revisit: may want to set the value in ownerDocument.
    	// Default behavior, overridden in some subclasses
        if (syncData) {
            synchronizeData();
            }
    	this.value = value;

    } // setNodeValue(String)

    /** 
     * nodeValue is a string representation of the data contained in
     * this node. Each subclass of Node defines what information is
     * considered its value, or returns null if the query is not
     * appropriate. Entity references should be expanded before being
     * returned.
     * <p>
     * For most types of Node, value is null and may not be changed.
     * <p>
     * For Text, CDATASection, ProcessingInstruction, and Comment value
     * is a string representing the contents of the node.
     * <p>
     * For Attribute, value is the string returned by the attribute's
     * getValue method. This will be gathered from the Attribute's
     * children.
     *
     * @throws DOMException(DOMSTRING_SIZE_ERR) when it would return more
     * characters than fit in a DOMString variable on the implementation
     * platform. (Will never arise in this implementation)
     */
    public String getNodeValue() {

        if (syncData) {
            synchronizeData();
            }
    	return value;

    } // getNodeValue():String

    /** 
     * Adds a child node to the end of the list of children for this node.
     * Convenience shorthand for insertBefore(newChild,null).
     * @see insertBefore()
     *
     * @returns newChild, in its new state (relocated, or emptied in the
     * case of DocumentNode.)
     * 
     * @throws DOMException(HIERARCHY_REQUEST_ERR) if newChild is of a
     * type that shouldn't be a child of this node.
     * 
     * @throws DOMException(WRONG_DOCUMENT_ERR) if newChild has a
     * different owner document than we do.
     * 
     * @throws DOMException(NO_MODIFICATION_ALLOWED_ERR) if this node is
     * read-only.
     */
    public Node appendChild(Node newChild) throws DOMException {
    	return insertBefore(newChild, null);
    }

    /** 
     * Returns a duplicate of a given node. You can consider this a
     * generic "copy constructor" for nodes. The newly returned object should
     * be completely independent of the source object's subtree, so changes
     * in one after the clone has been made will not affect the other.
     * <p>
     * Example: Cloning a Text node will copy both the node and the text it
     * contains.
     * <p>
     * Example: Cloning something that has children -- Element or Attr, for
     * example -- will _not_ clone those children unless a "deep clone"
     * has been requested. A shallow clone of an Attr node will yield an
     * empty Attr of the same name.
     * <p>
     * NOTE: Clones will always be read/write, even if the node being cloned
     * is read-only, to permit applications using only the DOM API to obtain
     * editable copies of locked portions of the tree.
     */
    public Node cloneNode(boolean deep) {

    	
        if (syncData) {
            synchronizeData();
            }
        if (syncChildren) { 
            synchronizeChildren(); 
        }
    	
    	NodeImpl newnode;
    	try {
    		newnode = (NodeImpl)clone();
    	}
    	catch (CloneNotSupportedException e) {
//	        Revisit : don't fail silently - but don't want to tie to parser guts
//    		System.out.println("UNEXPECTED "+e);
    		return null;
    	}
    	
    	newnode.readOnly = false;

    	// Need to break the association w/ original kids
    	newnode.parentNode      = null;
    	newnode.previousSibling = null;
        newnode.nextSibling     = null;
    	newnode.firstChild      = null;
        newnode.lastChild       = null;

    	// Then, if deep, clone the kids too.
    	if (deep) {
    		for (NodeImpl child = (NodeImpl)getFirstChild(); 
    			child != null; 
    			child = (NodeImpl)child.getNextSibling()) {
    			newnode.appendChild(child.cloneNode(true));
            }
        }

        // REVISIT: Should the user data be cloned?

    	return newnode;

    } // cloneNode(boolean):Node

    /** 
     * Find the Document that this Node belongs to (the document in
     * whose context the Node was created). The Node may or may not
     * currently be part of that Document's actual contents.  
     */
    public Document getOwnerDocument() {
        return ownerDocument;
    }  

    /** 
     * Obtain the DOM-tree parent of this node, or null if it is not
     * currently active in the DOM tree (perhaps because it has just been
     * created or removed). Note that Document, DocumentFragment, and
     * Attribute will never have parents.  
     */
    public Node getParentNode() {
        return parentNode;
    }

    /** The next child of this node's parent, or null if none */
    public Node getNextSibling() {
        return nextSibling;
    }  

    /** The previous child of this node's parent, or null if none */
    public Node getPreviousSibling() {
        return previousSibling;
    }  

    /**
     * Return the collection of attributes associated with this node,
     * or null if none. At this writing, Element is the only type of node
     * which will ever have attributes.
     *
     * @see com.ibm.domimpl.Element 
     */
    public NamedNodeMap getAttributes() {
    	return null; // overridden in ElementImpl
    }

    /** 
     * Test whether this node has any children. Convenience shorthand 
     * for (Node.getFirstChild()!=null)
     */
    public boolean hasChildNodes() {
        if (syncChildren) { 
            synchronizeChildren(); 
        }
        return firstChild != null;
    }  

    /** 
     * Obtain a NodeList enumerating all children of this node. If there
     * are none, an (initially) empty NodeList is returned.
     * <p>
     * NodeLists are "live"; as children are added/removed the NodeList
     * will immediately reflect those changes. Also, the NodeList refers
     * to the actual nodes, so changes to those nodes made via the DOM tree
     * will be reflected in the NodeList and vice versa.
     * <p>
     * In this implementation, Nodes implement the NodeList interface and
     * provide their own getChildNodes() support. Other DOMs may solve this
     * differently.
     */
    public NodeList getChildNodes() {

        if (syncChildren) {
            synchronizeChildren();
        }
        return this;

    } // getChildNodes():NodeList

    /** The first child of this Node, or null if none. */
    public Node getFirstChild() {

        if (syncChildren) { 
            synchronizeChildren(); 
        }
    	return firstChild;

    }   // getFirstChild():Node

    /** The first child of this Node, or null if none. */
    public Node getLastChild() {

        if (syncChildren) { 
            synchronizeChildren(); 
        }
   	    return lastChild;

    } // getLastChild():Node

    /**
     * Move one or more node(s) to our list of children. Note that this
     * implicitly removes them from their previous parent.
     * 
     * @param newChild The Node to be moved to our subtree. As a
     * convenience feature, inserting a DocumentNode will instead insert
     * all its children.
     * 
     * @param refChild Current child which newChild should be placed
     * immediately before. If refChild is null, the insertion occurs
     * after all existing Nodes, like appendChild().
     * 
     * @returns newChild, in its new state (relocated, or emptied in the
     * case of DocumentNode.)
     * 
     * @throws DOMException(HIERARCHY_REQUEST_ERR) if newChild is of a
     * type that shouldn't be a child of this node, or if newChild is an
     * ancestor of this node.
     * 
     * @throws DOMException(WRONG_DOCUMENT_ERR) if newChild has a
     * different owner document than we do.
     * 
     * @throws DOMException(NOT_FOUND_ERR) if refChild is not a child of
     * this node.  
     * 
     * @throws DOMException(NO_MODIFICATION_ALLOWED_ERR) if this node is
     * read-only.
     */
    public Node insertBefore(Node newChild, Node refChild) 
        throws DOMException {

    	if (readOnly)
    		throw new DOMExceptionImpl(
    			DOMException.NO_MODIFICATION_ALLOWED_ERR, null);
     	  
    	if( !(newChild instanceof NodeImpl)
    		||
    		!(
    		  newChild.getOwnerDocument() == ownerDocument
    		  ||
    	 	 // SPECIAL-CASE: Document has no owner, but may be the owner.
    		  ( getNodeType() == Node.DOCUMENT_NODE &&
    		    newChild.getOwnerDocument() == (Document)this ) 
    		) ) {
    		throw new DOMExceptionImpl(DOMException.WRONG_DOCUMENT_ERR, null);
        }
        
        if (syncChildren) {
            synchronizeChildren();
        }

    	// Convert to internal type, to avoid repeated casting 
    	NodeImpl newInternal = (NodeImpl)newChild;

    	// Prevent cycles in the tree
    	boolean treeSafe = true;
    	for (NodeImpl a = parentNode; treeSafe && a != null; a = a.parentNode) {
    		treeSafe = newInternal != a;
        }
    	if(!treeSafe) {
    	  	throw new DOMExceptionImpl(DOMException.HIERARCHY_REQUEST_ERR, null);
        }

    	// refChild must in fact be a child of this node (or null)
    	if(refChild != null && refChild.getParentNode() != this) {
    		throw new DOMExceptionImpl(DOMException.NOT_FOUND_ERR,null);
        }
    	  
    	if (newInternal.getNodeType() == Node.DOCUMENT_FRAGMENT_NODE) {
    		// SLOW BUT SAFE: We could insert the whole subtree without
    		// juggling so many next/previous pointers. (Wipe out the
    		// parent's child-list, patch the parent pointers, set the
    		// ends of the list.) But we know some subclasses have special-
    		// case behavior they add to insertBefore(), so we don't risk it.
    		// This approch also takes fewer bytecodes.

    		// NOTE: If one of the children is not a legal child of this
    		// node, throw HIERARCHY_REQUEST_ERR before _any_ of the children
    		// have been transferred. (Alternative behaviors would be to
    		// reparent up to the first failure point or reparent all those
    		// which are acceptable to the target node, neither of which is
    		// as robust. PR-DOM-0818 isn't entirely clear on which it
    		// recommends?????

    		// No need to check kids for right-document; if they weren't,
    		// they wouldn't be kids of that DocFrag.
    		for (Node kid = newInternal.getFirstChild(); // Prescan
    			kid != null;
    			kid = kid.getNextSibling()) {

    		    if (!isKidOK(this, kid)) {
        		  	throw new DOMExceptionImpl(DOMException.HIERARCHY_REQUEST_ERR, null);
                }
    		}		  	

    		while (newInternal.hasChildNodes())	{
    		    insertBefore(newInternal.getFirstChild(), refChild);
            }
    	}
    	  
    	else if (!isKidOK(this, newInternal)) {
    		throw new DOMExceptionImpl(DOMException.HIERARCHY_REQUEST_ERR, null);
        }
    	else {
    		Node oldparent = newInternal.getParentNode();
    		if (oldparent != null) {
    			oldparent.removeChild(newInternal);
            }

    		NodeImpl prev;
    		// Find the node we're inserting after, if any (null if
    		// inserting to head of list)
    		prev = (refChild == null) 
                   ? lastChild : ((NodeImpl)refChild).previousSibling;

    		// Attach up
    		newInternal.parentNode = this;

    		// Attach after
    		newInternal.previousSibling = prev;
    		if (prev == null) {
    		    firstChild = newInternal;
            }
            else {
    		    prev.nextSibling = newInternal;
            }

    		// Attach before
    		newInternal.nextSibling = (NodeImpl)refChild;
    		if (refChild == null) {
    		    lastChild=newInternal;
            }
            else {
    		    ((NodeImpl)refChild).previousSibling = newInternal;
            }
    	}

    	changed();
    	
    	return newInternal;

    } // insertBefore(Node,Node):Node

    /**
     * Remove a child from this Node. The removed child's subtree
     * remains intact so it may be re-inserted elsewhere.
     * 
     * @return oldChild, in its new state (removed).
     * 
     * @throws DOMException(NOT_FOUND_ERR) if oldChild is not a child of
     * this node.  
     * 
     * @throws DOMException(NO_MODIFICATION_ALLOWED_ERR) if this node is
     * read-only.
     */
    public Node removeChild(Node oldChild) 
        throws DOMException {

    	if (readOnly) {
    		throw new DOMExceptionImpl(
    			DOMException.NO_MODIFICATION_ALLOWED_ERR, null);
        }
     	  
    	if (oldChild != null && oldChild.getParentNode() != this) {
    		throw new DOMExceptionImpl(DOMException.NOT_FOUND_ERR, null);
        }

    	// Patch tree past oldChild
    	NodeImpl oldInternal = (NodeImpl) oldChild;
    	NodeImpl prev = oldInternal.previousSibling;
    	NodeImpl next = oldInternal.nextSibling;

    	if (prev != null) {
    		prev.nextSibling = next;
        }
        else {
    		firstChild = next;
        }

    	if (next != null) {
    		next.previousSibling = prev;
        }
        else {
    		lastChild = prev;
        }

    	// Remove oldInternal's references to tree
    	oldInternal.parentNode      = null;
        oldInternal.nextSibling     = null;
        oldInternal.previousSibling = null;

    	changed();

    	return oldInternal;

    } // removeChild(Node):Node

    /**
     * Make newChild occupy the location that oldChild used to
     * have. Note that newChild will first be removed from its previous
     * parent, if any. Equivalent to inserting newChild before oldChild,
     * then removing oldChild.
     * 
     * @returns oldChild, in its new state (removed).
     * 
     * @throws DOMException(HIERARCHY_REQUEST_ERR) if newChild is of a
     * type that shouldn't be a child of this node, or if newChild is
     * one of our ancestors.
     * 
     * @throws DOMException(WRONG_DOCUMENT_ERR) if newChild has a
     * different owner document than we do.
     * 
     * @throws DOMException(NOT_FOUND_ERR) if oldChild is not a child of
     * this node.  
     * 
     * @throws DOMException(NO_MODIFICATION_ALLOWED_ERR) if this node is
     * read-only.
     */
    public Node replaceChild(Node newChild, Node oldChild) 
        throws DOMException {
    	insertBefore(newChild, oldChild);
        return removeChild(oldChild);
    }

    //
    // NodeList methods
    //

    /**
     * NodeList method: Count the immediate children of this node
     * @return int
     */
    public int getLength() {

        // It is assumed that the getChildNodes call synchronized
        // the children. Therefore, we can access the first child
        // reference directly.
    	int count = 0;
    	for (NodeImpl node = firstChild; node != null; node = node.nextSibling) {
    		count++;
    	}
    	return count;

    } // getLength():int
    
    /**
     * NodeList method: Return the Nth immediate child of this node, or
     * null if the index is out of bounds.
     * @return org.w3c.dom.Node
     * @param Index int
     */
    public Node item(int index) {

        // It is assumed that the getChildNodes call synchronized
        // the children. Therefore, we can access the first child
        // reference directly.
        NodeImpl node = firstChild;
        for (int i = 0; i < index && node != null; i++) {
            node = node.nextSibling;
        }
        return node;

    } // item(int):Node

    //
    // Public methods
    //

    /**
     * NON-DOM: PR-DOM-Level-1-19980818 mentions readonly nodes in conjunction
     * with Entities, but provides no API to support this.
     * <P>
     * Most DOM users should not touch this method. Its anticpated use
     * is during construction of EntityRefernces, where it will be used to
     * lock the contents replicated from Entity so they can't be casually 
     * altered. It _could_ be published as a DOM extension, if desired.
     * 
     * @param readOnly True or false as desired.
     * @param deep If true, children are also toggled. Note that this will
     *	not change the state of an EntityReference or its children,
     *  which are always read-only.
     */
    public void setReadOnly(boolean readOnly, boolean deep) {

        if (syncData) {
            synchronizeData();
        }

    	this.readOnly = readOnly;

    	if (deep) {

            if (syncChildren) {
                synchronizeChildren();
            }

    		// Recursively set kids
    		for (NodeImpl mykid = firstChild; 
    			mykid != null; 
    			mykid = mykid.nextSibling) {
    			if(!(mykid instanceof EntityReference)) {
    				mykid.setReadOnly(readOnly,true);
                }
            }
        }

    } // setReadOnly(boolean,boolean)

    /** 
     * NON-DOM: Returns true if this node is read-only. This is a
     * shallow check.
     */
    public boolean getReadOnly() {

        if (syncData) {
            synchronizeData();
        }

        return readOnly;

    } // getReadOnly():boolean

    /** 
     * NON-DOM: As an alternative to subclassing the DOM, this implementation
     * has been extended with the ability to attach an object to each node.
     * (If you need multiple objects, you can attach a collection such as a
     * vector or hashtable, then attach your application information to that.)
     *
     * @returns the previous user object, or null if none.
     */
    public void setUserData(Object data) {
        userData = data;
    }

    /** Returns the user data associated to this node. */
    public Object getUserData() {
        return userData;
    }  
    
    //
    // Protected methods
    //

    /**
     * Override this method in subclass to hook in efficient
     * internal data structure.
     */
    protected void synchronizeChildren() {}

    /**
     * Override this method in subclass to hook in efficient
     * internal data structure.
     */
    protected void synchronizeData() {}

    /** Denotes that this node has changed. */
    protected void changed() {
    	++changes;
    	if (parentNode != null) {
            parentNode.changed();
        }
    }  

    /** 
     * Uses the kidOK lookup table to check whether the proposed
     * tree structure is legal.
     */
    protected static boolean isKidOK(Node parent, Node child) {
    	return 0 != (kidOK[parent.getNodeType()] & 1 << child.getNodeType());
    }  

    //
    // Object methods
    //

    /** NON-DOM method for debugging convenience. */
    public String toString() {
        return "["+getNodeName()+": "+getNodeValue()+"]";
    }    

    //
    // Serialization methods
    //

    /** Serialize object. */
    private void writeObject(ObjectOutputStream out) throws IOException {

        // synchronize data and chilren
        if (syncData) {
            synchronizeData();
        }
        if (syncChildren) {
            synchronizeChildren();
        }

        // write object
        out.defaultWriteObject();

    } // writeObject(ObjectOutputStream)

} // class NodeImpl
