/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.framework;

import com.ibm.xml.internal.StringHasher;

/**
 *
 * @version Revision: 56 1.8 src/com/ibm/xml/framework/ChunkyCharArray.java, parser, xml4j2, xml4j2_0_13 
 */

public final class ChunkyCharArray {
    //
    // Chunk size constants
    //
    protected static final int INITIAL_CHUNK_SHIFT = 7;        // 2^7 = 128
    protected static final int INITIAL_CHUNK_SIZE = (1 << INITIAL_CHUNK_SHIFT);
    protected static final int CHUNK_SHIFT = 14;                // 2^14 = 16k
    protected static final int CHUNK_SIZE = (1 << CHUNK_SHIFT);
    protected static final int CHUNK_MASK = CHUNK_SIZE - 1;
    //
    // Instance variables
    //
    protected CharArrayChunk fCurrentChunk = null;
    protected int fLength = 0;
    /**
     * Constructor
     */
    public ChunkyCharArray(StringPool stringPool) {
        fCurrentChunk = new CharArrayChunk(stringPool, null);
    }
    /**
     *
     */
    public int length() {
        return fLength;
    }
    /**
     *
     */
    public void setLength(int length) {
        fCurrentChunk.setLength(length);
    }
    /**
     *
     */
    public void append(char ch) {
        fCurrentChunk.append(ch);
    }
    /**
     *
     */
    public void append(String s) {
        int slen = s.length();
        for (int i = 0; i < slen; i++)
            fCurrentChunk.append(s.charAt(i));
    }
    /**
     *
     */
    public void append(char[] ch, int offset, int length) {
        while (length-- > 0)
            fCurrentChunk.append(ch[offset++]);
    }
    /**
     *
     */
    public void append(ChunkyCharArray charArray, int offset, int length) {
        fCurrentChunk.append(charArray, offset, length);
    }
    /**
     *
     */
    public int addString(int offset, int length) {
        if (length == 0)
            return 0;
        return fCurrentChunk.addString(offset, length);
    }
    /**
     *
     */
    public int addSymbol(int offset, int length) {
        if (length == 0)
            return 0;
        return fCurrentChunk.addSymbol(offset, length);
    }
    //
    //
    //
    protected final class CharArrayChunk implements StringProducer {
        //
        //
        //
        public CharArrayChunk(StringPool stringPool, CharArrayChunk prev) {
            fStringPool = stringPool;
            fChunk = prev == null ? 0 : prev.fChunk + 1;
            fPreviousChunk = prev;
            if (prev != null)
                prev.fNextChunk = this;
        }
        //
        // Constructor for creating clumps.
        //
        protected CharArrayChunk(CharArrayChunk proto) {
            fStringPool = proto.fStringPool;
            fChunk = proto.fChunk;
            fData = proto.fData;
            fPreviousChunk = null;
            fNextChunk = null;
        }
        //
        //
        //
        protected CharArrayChunk chunkFor(int offset) {
            int firstChunk = offset >> CHUNK_SHIFT;
            CharArrayChunk dataChunk = this;
            while (firstChunk != dataChunk.fChunk)
                dataChunk = dataChunk.fPreviousChunk;
            return dataChunk;
        }
        /**
         *
         */
        public void setLength(int length) {
            fLength = length;
            fCurrentChunk = chunkFor(length);
            fCurrentIndex = length & CHUNK_MASK;
        }
        /**
         *
         */
        public void append(char ch) {
            try {
                fData[fCurrentIndex] = ch;
            } catch (ArrayIndexOutOfBoundsException ex) {
                if (fCurrentIndex == CHUNK_SIZE) {
                    fCurrentChunk = new CharArrayChunk(fStringPool, this);
                    fCurrentChunk.append(ch);
                    return;
                }
                char[] newData = new char[fCurrentIndex * 2];
                System.arraycopy(fData, 0, newData, 0, fCurrentIndex);
                fData = newData;
                fData[fCurrentIndex] = ch;
            } catch (NullPointerException ex) {
                fData = new char[INITIAL_CHUNK_SIZE];
                fData[fCurrentIndex] = ch;
            }
            fLength++;
            fCurrentIndex++;
        }
        /**
         *
         */
        public void append(ChunkyCharArray charArray, int offset, int length) {
            //
            // REVISIT - replace this with the better code from the reader data chunk classes.
            //
            CharArrayChunk dataChunk = chunkFor(offset);
            int index = offset & CHUNK_MASK;
            if (index + length <= CHUNK_SIZE) {
                //
                // All the chars are in the same chunk
                //
                charArray.append(dataChunk.fData, index, length);
                return;
            }
            //
            // The data is spread across chunks.
            //
            int nbytes = CHUNK_SIZE - index;
            charArray.append(dataChunk.fData, index, nbytes);
            length -= nbytes;
            //
            // Use each Chunk in turn until we are done.
            //
            do {
                dataChunk = dataChunk.fNextChunk;
                nbytes = length <= CHUNK_SIZE ? length : CHUNK_SIZE;
                charArray.append(dataChunk.fData, 0, nbytes);
                length -= nbytes;
            } while (length > 0);
        }
        //
        //
        //
        protected CharArrayChunk createClump(int lastChunk) {
            CharArrayChunk result = new CharArrayChunk(this);
            CharArrayChunk nextChunk = fNextChunk;
            CharArrayChunk aChunk = result;
            do {
                CharArrayChunk next = new CharArrayChunk(nextChunk);
                aChunk.fNextChunk = next;
                nextChunk = nextChunk.fNextChunk;
                aChunk = next;
            } while (aChunk.fChunk != lastChunk);
            return result;
        }
        /**
         *
         */
        public int addString(int offset, int length) {
            int chunk = offset >> CHUNK_SHIFT;
            if (chunk != fChunk) {
                if (fPreviousChunk != null)
                    return fPreviousChunk.addString(offset, length);
                System.out.println("fPreviousChunk == null");
                return -1;
            }
            int lastChunk = (offset + length - 1) >> CHUNK_SHIFT;
            if (chunk == lastChunk)
                return fStringPool.addString(this, offset & CHUNK_MASK, length);
            if (fNextChunk == null)
                System.out.println("fNextChunk == null");
            return fStringPool.addString(createClump(lastChunk), offset & CHUNK_MASK, length);
        }
        /**
         *
         */
        public int addSymbol(int offset, int length) {
            int chunk = offset >> CHUNK_SHIFT;
            if (chunk != fChunk) {
                if (fPreviousChunk != null)
                    return fPreviousChunk.addSymbol(offset, length);
                System.out.println("fPreviousChunk == null");
                return -1;
            }
            int lastChunk = (offset + length - 1) >> CHUNK_SHIFT;
            int index = offset & CHUNK_MASK;
            if (chunk == lastChunk) {
                int hashcode = 0;
                for (int i = 0; i < length; i++)
                    hashcode = StringHasher.hashChar(hashcode, i, fData[index++] & 0xFFFF);
                hashcode = StringHasher.finishHash(hashcode);
                return fStringPool.addSymbol(this, offset & CHUNK_MASK, length, hashcode);
            }
            int hashcode = 0;
            int position = 0;
            int hlength = length;
            int nbytes = CHUNK_SIZE - index;
            while (index < CHUNK_SIZE)
                hashcode = StringHasher.hashChar(hashcode, position++, fData[index++] & 0xFFFF);
            hlength -= nbytes;
            //
            // Check each chunk in turn until we are done.
            //
            CharArrayChunk aChunk = fNextChunk;
            do {
                index = 0;
                nbytes = hlength <= CHUNK_SIZE ? hlength : CHUNK_SIZE;
                while (index < nbytes)
                    hashcode = StringHasher.hashChar(hashcode, position++, aChunk.fData[index++]& 0xFFFF);
                hlength -= nbytes;
                aChunk = aChunk.fNextChunk;
            } while (hlength > 0);
            hashcode = StringHasher.finishHash(hashcode);
            return fStringPool.addSymbol(createClump(lastChunk), offset & CHUNK_MASK, length, hashcode);
        }
        //
        //
        //
        public String toString(int offset, int length) {
            if (offset + length <= CHUNK_SIZE) {
                //
                // All the chars are in the same chunk
                //
                return new String(fData, offset, length);
            }
            //
            // The data is spread across chunks, so we need to build it in pieces.
            //
            StringBuffer sb = new StringBuffer(length);
            //
            // Copy the partial data from the first chunk.
            //
            int nbytes = CHUNK_SIZE - offset;
            sb.append(fData, offset, nbytes);
            length -= nbytes;
            //
            // Use each chunk in turn until we are done.
            //
            CharArrayChunk aChunk = fNextChunk;
            do {
                nbytes = length <= CHUNK_SIZE ? length : CHUNK_SIZE;
                sb.append(aChunk.fData, 0, nbytes);
                length -= nbytes;
                aChunk = aChunk.fNextChunk;
            } while (length > 0);
            String retval = sb.toString();
            sb = null; // REVISIT - does this help gc ?
            return retval;
        }
        /**
         *
         */
        public boolean equalsString(int offset, int length, String str, int strLength) {
            //
            // REVISIT - replace this with the better code from the reader data chunk classes.
            //
            if (length != strLength)
                return false;
            if (offset + length <= CHUNK_SIZE) {
                //
                // All the chars are in the same chunk
                //
                for (int i = 0; i < length; i++) {
                    if (fData[offset++] != str.charAt(i))
                        return false;
                }
                return true;
            }
            //
            // Compare the partial data from the first chunk.
            //
            int nbytes = CHUNK_SIZE - offset;
            int index = 0;
            while (index < nbytes) {
                if (fData[offset++] != str.charAt(index++))
                    return false;
            }
            length -= nbytes;
            //
            // Check each chunk in turn until we are done.
            //
            CharArrayChunk aChunk = fNextChunk;
            do {
                offset = 0;
                nbytes = length <= CHUNK_SIZE ? length : CHUNK_SIZE;
                length -= nbytes;
                while (nbytes-- > 0) {
                    if (aChunk.fData[offset++] != str.charAt(index++))
                        return false;
                }
                aChunk = aChunk.fNextChunk;
            } while (length > 0);
            return true;
        }
        //
        //
        //
        protected StringPool fStringPool = null;
        protected int fChunk = -1;
        protected int fCurrentIndex = 0;
        protected char[] fData = null;
        protected CharArrayChunk fPreviousChunk = null;
        protected CharArrayChunk fNextChunk = null;

    } // class CharArrayChunk

} // class ChunkyCharArray
