/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.framework;


/**
 * ---------------------------------------------------------------------------
 *  InsertableElementsInfo is a simple 'data packet' class that is used to
 *  get information into and out of the validator APIs that allow you to ask
 *  what kind of elements can be inserted into a particular place in an
 *  element's content model.
 *
 *  The parent element is not explicitly stored here, since it is a separate
 *  parameter to the methods that do the query.
 *
 *  Since it exists purely to exchange data, it just uses simple public
 *  data members.
 * ---------------------------------------------------------------------------
 *
 * @version Revision: 54 1.4 src/com/ibm/xml/framework/InsertableElementsInfo.java, parser, xml4j2, xml4j2_0_13 
 */
public class InsertableElementsInfo
{
    // -----------------------------------------------------------------------
    //  Public data members
    //
    //  canHoldPCData
    //      One of the things that could be inserted here is a PCDATA node,
    //      in addition to the element type nodes reported.
    //
    //  childCount
    //      The count of elements in the curChildren array. The array can be
    //      larger than this (for reuse purposes), so this value indicates
    //      how many elements are valid.
    //
    //      Note that, since the curChildren array must have an empty slot at
    //      the insertion index, this value can never be zero.
    //
    //      Note also that this value can be changed during processing, though
    //      its value on return is meaningless to the caller.
    //
    //  curChildren
    //      The current list of children of the parent element. This may or
    //      may not be the *real* list of children, since the caller can lie
    //      but that's of no concern to the validator. These query APIs are
    //      intended to be for 'what if' kind of work, so any list of children
    //      could be passed in.
    //
    //      There must be an empy slot in the array at the requested insertion
    //      point. That slot does not have to have any particular value, but
    //      it will be used by the validator to do brute force validation in
    //      some cases when a 'fully valid' check is done for valid insertable
    //      elements.
    //
    //      Note that this array can be modified by the call, so do not expect
    //      its contents to remain the same as on input.
    //
    //  isValidEOC
    //      Indicates that one of the valid things after the insert point is
    //      'end of content', which means that the element being inserted
    //      after can legally be the last element.
    //
    //  insertAt
    //      The insertion point. The question is 'what can go here' and this
    //      indicates where 'where' is. It is an offset into curChildren.
    //
    //  possibleChildren
    //      This array is filled with flags that indicate what the possible
    //      insertable elements are (i.e. the list of unique elements that
    //      could possibly be inserted somewhere in this type of element.)
    //      Effectively this is the list of unique children in the content
    //      model of the parent element.
    //
    //      If this array is not big enough to hold the results, or is null,
    //      then it will be replaced with a new array of the correct size.
    //
    //  results
    //      This array must be at least as large as possibleChildren since
    //      a flag is set in the same indexes in this array to indicate that
    //      the possible child at that index in possibleChildren can be
    //      inserted at the requested insertion point.
    //
    //      If this array is not big enough to hold the results, or is null,
    //      then it will be replaced with a new array of the correct size.
    //
    //  resultsCount
    //      The number of elements that are valid in the possibleChildren
    //      and resultsCount arrays. They can be larger than that, so there
    //      must be a way to indicate how many elements are filled in with
    //      value results. If they were not at least this large on input,
    //      then they will be reallocated up to this size.
    // -----------------------------------------------------------------------
    public boolean      canHoldPCData;
    public int          childCount;
    public int[]        curChildren;
    public boolean      isValidEOC;
    public int          insertAt;
    public int[]        possibleChildren;
    public boolean[]    results;
    public int          resultsCount;
};
