/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.framework;

import org.xml.sax.InputSource;
import org.xml.sax.EntityResolver;

/**
 * This is the interface used for entity events.
 *
 * @author  Glenn Marcy
 * @author  Andy Clark, IBM
 * @version Revision: 47 1.8 src/com/ibm/xml/framework/XMLEntityHandler.java, parser, xml4j2, xml4j2_0_13 
 * @see XMLParser
 */
public interface XMLEntityHandler 
    {
    
    //
    // XMLEntityHandler methods
    //

    /** 
     * Resets the entity handler, usually in response to a new call to
     * <tt>XMLParser.parse</tt>. 
     *
     * @param parserState The parser state.
     */
    public void reset(ParserState parserState);

    /** 
     * Adds a recognizer. 
     *
     * @param recognizer The XML recognizer to add.
     */
    public void addRecognizer(XMLDeclRecognizer recognizer);

    /**
     * Sets the resolver used to resolve external entities. The EntityResolver
     * interface supports resolution of public and system identifiers. If the
     * resolver set is also of type com.ibm.xml.framework.Catalog, then 
     * catalogs can be loaded in a general manner by calling 
     * <tt>XMLParser.loadCatalog()</tt>.
     *
     * @param resolver The new entity resolver. Passing a null value will
     *                 uninstall the currently installed resolver.
     *
     * @see com.ibm.xml.framework.Catalog
     * @see com.ibm.xml.framework.XMLParser#loadCatalog
     */
    public void setEntityResolver(EntityResolver resolver);

    /**
     * Returns the resolver used to resolve external entities.
     *
     * @return The resolver.
     *
     * @see #setEntityResolver
     */
    public EntityResolver getEntityResolver();


    /**
     * Performs the entity resolution. The most common way to resolve
     * entities is to get the strings represented by the publicId and
     * systemId indexes and ask the installed EntityResolver to resolve
     * the entities.
     * 
     * @param publidId The index of the public identifier string in the
     *                 parser StringPool. If this value is -1, then a
     *                 system identifier alias is requested; else the
     *                 system identifier represents the default entity
     *                 resolution if no public identifier mapping is
     *                 possible.
     * @param systemId The index of the system identifier string in the
     *                 parser StringPool.
     *
     * @return Returns an InputSource of the resolved entity or null if
     *         no resolution is possible.
     *
     * @exception java.lang.Exception If entity resolution fails, this method can
     *                                throw an exception.
     *
     * @see #getEntityResolver
     * @see com.ibm.xml.framework.StringPool#toString
     */
    public InputSource resolveEntity(int publicId, int systemId) throws Exception;

    /**
     * Fully expands a given system identified. The system identifier string
     * to be resolved and the fully expanded system identifier string are
     * stored in the parser StringPool.
     * <p>
     * The following method expands a system identifier given as a
     * <tt>java.lang.String</tt>.
     * <pre>
     *     public String expandSystemId(XMLParser parser, String systemId) throws Exception {
     *         StringPool pool = parser.getParserState().getStringPool();
     *
     *         // expand entity
     *         int oldIndex = pool.addString(systemId);
     *         int newIndex = parser.getEntityHandler().expandSystemId(oldIndex);
     *         systemId = pool.toIndex(newIndex);
     *
     *         // release string pool resources
     *         pool.releaseString(oldIndex);
     *         pool.releaseString(newIndex);
     *
     *         // return expanded system id
     *         return systemId;
     *         }
     * </pre>
     *
     * @param systemId The index of the system identifier string in the
     *                 parser StringPool.
     *
     * @return Returns the index in the StringPool that holds the expanded
     *         system identifier. If the given system identifier is already
     *         fully expanded, this method can return the original systemId
     *         value.
     *
     * @see com.ibm.xml.framework.StringPool#toString
     */
    public int expandSystemId(int systemId);

    /**
     * Creates a parser reader object from the given input source.
     * <p>
     * This is the point in the default implementation where relative
     * system identifiers are expanded. For example, if the current
     * source URI is "http://host/path/file.xml" and the specified
     * input source system identifier is "dtd/file.dtd", then the
     * system identifier, if not already opened, will be set to
     * "http://host/path/dtd/file.dtd".
     *
     * @param source  The input source.
     * @param xmlDecl True if the source represents an XML decl.
     *
     * @return Returns a parser reader object.
     *
     * @exception java.lang.Exception Throws exception on error.
     */
    public XMLReader createReader(InputSource source, boolean xmlDecl) throws Exception;

    /** 
     * Start of an input source. 
     *
     * @param inputSource The input source that is starting.
     */
    public void startInputSource(InputSource inputSource);

    /** 
     * End of an input source. 
     *
     * @param inputSource The input source that is ended.
     */
    public void endInputSource(InputSource inputSource);

    } // interface XMLEntityHandler
