/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.framework;

import com.ibm.xml.internal.DefaultScanner;
import com.ibm.xml.internal.DefaultEntityHandler;
import com.ibm.xml.internal.DefaultValidationHandler;
import com.ibm.xml.internal.ErrorCode;
import org.xml.sax.InputSource;
import java.io.InputStream;
import java.io.IOException;
import java.io.Reader;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.ListResourceBundle;

/**
 * This is the base class of all parsers.
 *
 * This class implements the <code>Parser</code> interface.  The application can use
 * the <code>ParserFactory</code> class to construct a parser from a class name.
 *
 * This class implements the <code>XMLDocumentTypeHandler</code>, <code>XMLDocumentHandler</code>,
 * and <code>XMLErrorHandler</code> interfaces.  This is provided as a convenience to parser
 * subclasses so that they can implement these interfaces directly in the subclass instead of
 * needing a separate class.
 *
 * @author  Glenn Marcy   
 * @author  Andy Clark, IBM
 * @version Revision: 51 1.15 src/com/ibm/xml/framework/XMLParser.java, parser, xml4j2, xml4j2_0_13 
 * @see XMLParser
 * @see org.xml.sax.InputSource
 * @see org.xml.sax.Locator
 */
public abstract class XMLParser
    implements org.xml.sax.Parser, XMLDocumentTypeHandler, XMLDocumentHandler, XMLErrorHandler
{
    //
    // Instance variables
    //
    private ParserState fParserState = null;
    protected XMLScanner fScanner = null;
    private XMLEntityHandler fEntityHandler = null;
    private org.xml.sax.ErrorHandler fSAXErrorHandler = null;
    protected Locale fLocale = null;
    protected ResourceBundle fResourceBundle = null;
    private boolean fAllowJavaEncodingName = false;
    private boolean fWarningOnDuplicateAttDef = false;
    private boolean fCheckNamespace = false;
    private boolean fContinueAfterFatalError = false;
    protected boolean fParseInProgress = false;
    protected boolean fNeedReset = false;
    //
    // Handlers
    //
    private XMLDocumentHandler fDocHandler = null;
    private XMLDocumentTypeHandler fDocTypeHandler = null;
    private XMLValidationHandler fValidationHandler = null;
    /**
     * Constructor
     */
    protected XMLParser() {
        fParserState = new ParserState(this);
        fScanner = new DefaultScanner(fParserState);
        fEntityHandler = new DefaultEntityHandler(fParserState);
    }
    /**
     *
     */
    public void reset() {
        fParserState.reset();
        fScanner.reset(fParserState);
        fEntityHandler.reset(fParserState);
        fNeedReset = false;
    }
    /**
     *
     */
    protected void resetOrCopy() {
        fParserState = fParserState.resetOrCopy();
        fScanner.reset(fParserState);
        fEntityHandler.reset(fParserState);
        fNeedReset = false;
    }
    /**
     *
     */
    protected void checkHandlers() throws Exception {
    }
    /**
     *
     */
    public ParserState getParserState() {
        return fParserState;
    }
    /**
     *
     */
    public XMLScanner getScanner() {
        return fScanner;
    }
    /**
     * Return the <code>Locator</code> object.
     *
     * @return The locator.
     */
    protected org.xml.sax.Locator getLocator() {
        return fScanner.getLocator();
    }
    /**
     *
     */
    public void setAllowJavaEncodingName(boolean flag) {
        fAllowJavaEncodingName = flag;
    }
    /**
     *
     */
    public boolean getAllowJavaEncodingName() {
        return fAllowJavaEncodingName;
    }
    /**
     *
     */
    public void setWarningOnDuplicateAttDef(boolean flag) {
        fWarningOnDuplicateAttDef = flag;
    }
    /**
     *
     */
    public boolean getWarningOnDuplicateAttDef() {
        return fWarningOnDuplicateAttDef;
    }
    /**
     *
     */
    public void setCheckNamespace(boolean flag) {
        fCheckNamespace = flag;
    }
    /**
     *
     */
    public boolean getCheckNamespace() {
        return fCheckNamespace;
    }
    /**
     *
     */
    public void setContinueAfterFatalError(boolean flag) {
        fContinueAfterFatalError = flag;
    }
    /**
     *
     */
    public boolean getContinueAfterFatalError() {
        return fContinueAfterFatalError;
    }
    /**
     *
     */
    protected void setDocumentHandler(XMLDocumentHandler handler) {
        fDocHandler = handler;
    }
    /**
     *
     */
    public XMLDocumentHandler getDocumentHandler() {
        return fDocHandler;
    }
    /**
     *
     */
    protected void setDocumentTypeHandler(XMLDocumentTypeHandler handler) {
        fDocTypeHandler = handler;
    }
    /**
     *
     */
    public XMLDocumentTypeHandler getDocumentTypeHandler() {
        return fDocTypeHandler;
    }
    /**
     *
     */
    protected void setEntityHandler(XMLEntityHandler handler) {
        fEntityHandler = handler;
    }
    /**
     *
     */
    public XMLEntityHandler getEntityHandler() {
        return fEntityHandler;
    }
    /**
     *
     */
    public XMLErrorHandler getErrorHandler() {
        return this;
    }
    /**
     *
     */
    protected void setValidationHandler(XMLValidationHandler handler) {
        fValidationHandler = handler;
    }
    /**
     *
     */
    public XMLValidationHandler getValidationHandler() {
        return fValidationHandler;
    }
    /**
     *
     */
    protected void useDefaultValidationHandler() {
        fValidationHandler = new DefaultValidationHandler(fParserState);
    }

    /**
     * Loads the catalog stream specified by the given input source and
     * appends the contents to the catalog.
     *
     * @param source The catalog source.
     *
     * @exception java.lang.Exception Throws an exception if an error
     *                                occurs while loading the catalog source.
     */
    public void loadCatalog(InputSource source)
        throws Exception 
        {

        // TODO: Move error strings into a resource bundle. -Ac
        XMLEntityHandler handler = getEntityHandler();
        if (handler != null) {
            org.xml.sax.EntityResolver resolver = handler.getEntityResolver();
            if (resolver != null) {
                if (resolver instanceof Catalog) {
                    ((Catalog)resolver).loadCatalog(source);
                    }
                else {
                    throw new Exception("EntityResolver is not a Catalog");
                    }
                }
            else {
                throw new Exception("Catalog not installed");
                }
            }
        else {
            throw new Exception("XMLEntityHandler not installed");
            }

        } // loadCatalog(InputSource)

    /**
     * Set the locale to use for messages.
     *
     * This method is part of the <code>Parser</code> interface.
     *
     * @param locale The locale object to use for localization of messages.
     * @exception org.xml.sax.SAXException An exception thrown if the parser
     *                                     does not support the specified locale.
     * @see org.xml.sax.Parser
     */
    public final void setLocale(Locale locale) throws org.xml.sax.SAXException {
        if (fParseInProgress)
            throw new org.xml.sax.SAXException("setLocale may not be called while parsing"); // REVISIT - need to add new error message
        if (fLocale != locale) {
            fLocale = locale;
            fResourceBundle = null;
        }
    }
    /**
     *
     */
    public void setEntityResolver(org.xml.sax.EntityResolver resolver) {
        if (fEntityHandler != null)
            fEntityHandler.setEntityResolver(resolver);
    }
    /**
     *
     */
    public void setDTDHandler(org.xml.sax.DTDHandler handler) {
    }
    /**
     *
     */
    public void setDocumentHandler(org.xml.sax.DocumentHandler handler) {
    }
    /**
     *
     */
    public void setErrorHandler(org.xml.sax.ErrorHandler handler) {
        fSAXErrorHandler = handler;
    }

    /**
     * Parses the specified input source.
     *
     * @param source The input source.
     *
     * @exception org.xml.sax.SAXException Throws exception on SAX error.
     * @exception java.io.IOException Throws exception on i/o error.
     */
    public void parse(InputSource source) throws org.xml.sax.SAXException, IOException {
        if (fParseInProgress)
            throw new org.xml.sax.SAXException("parse may not be called while parsing"); // REVISIT - need to add new error message
        if (fNeedReset)
            resetOrCopy();
        try {
            checkHandlers();
            fParseInProgress = true;
            fNeedReset = true;
            fScanner.scanDocument(source);
            fParseInProgress = false;
        } catch (org.xml.sax.SAXException ex) {
            fParseInProgress = false;
            throw ex;
        } catch (IOException ex) {
            fParseInProgress = false;
            throw ex;
        } catch (Exception ex) {
            fParseInProgress = false;
            throw new org.xml.sax.SAXException(ex);
        }
    }

    /**
     * Parses the input source specified by the given system identifier.
     * <p>
     * This method is equivalent to the following:
     * <pre>
     *     parse(new InputSource(systemId));
     * </pre>
     *
     * @param source The input source.
     *
     * @exception org.xml.sax.SAXException Throws exception on SAX error.
     * @exception java.io.IOException Throws exception on i/o error.
     */
    public void parse(String systemId) throws org.xml.sax.SAXException, IOException
    {
        /***
        // NOTE: This isn't needed because the parse(InputSource) method
        //       checks the parse in progress flag. -Ac
        if (fParseInProgress)
            throw new org.xml.sax.SAXException("parse may not be called while parsing"); // REVISIT - need to add new error message
        // NOTE: This check isn't needed because the default implementation of
        //       XMLEntityHandler.createReader checks the system id. -Ac 
        try {
            URL u = new URL(systemId);
//            if (!systemId.equals(u.toString()))
//                throw new org.xml.sax.SAXException("URL must be fully resolved before calling parse"); // REVISIT - need to add new error message
        } catch (MalformedURLException ex) {
            throw new org.xml.sax.SAXException(ex);
        }
        /***/
        InputSource source = new InputSource(systemId);
        parse(source);
        try {
            Reader reader = source.getCharacterStream();
            if (reader != null) {
                reader.close();
            }
            else {
                InputStream is = source.getByteStream();
                if (is != null) {
                    is.close();
                }
            }
        }
        catch (IOException e) {
            // ignore
        }
    }

    /**
     * The default for XMLDocumentHandler is to use the StringPool to pass CharData, not char arrays.
     */
    public boolean sendCharDataAsCharArray() {
        return false;
    }

    //
    // Error handling support.
    //

    /**
     *
     */
    protected boolean errorHandlingEnabled() {
        return fSAXErrorHandler != null;
    }
    /**
     *
     */
    protected void handleError(String errorKey, String errorMsg) throws Exception {
        org.xml.sax.SAXParseException spe = new org.xml.sax.SAXParseException(errorMsg, getLocator());
        if (fSAXErrorHandler == null) {
            if (isFatal(errorKey))
                throw spe;
            else
                return;
        }
        if (isWarning(errorKey))
            fSAXErrorHandler.warning(spe);
        else if (isFatal(errorKey))
            fSAXErrorHandler.fatalError(spe);
        else
            fSAXErrorHandler.error(spe);
    }
    /**
     *
     */
    protected final String getErrorMsgString(String errorKey, Object[] errorArgs) {
        if (fResourceBundle == null) {
            if (fLocale == null)
                fResourceBundle = ListResourceBundle.getBundle("com.ibm.xml.internal.msg.Message");
            else
                fResourceBundle = ListResourceBundle.getBundle("com.ibm.xml.internal.msg.Message", fLocale);
        }
        String retval = fResourceBundle.getString(errorKey);
        if (errorArgs != null)
            try {
                retval = java.text.MessageFormat.format(retval, errorArgs);
            } catch (Exception e) {
                retval = fResourceBundle.getString(errorKey);
				String errorKey1 = ErrorCode.getErrorKey(ErrorCode.E_INT_MSGFMT);
                retval += fResourceBundle.getString(errorKey1);
            }

        return retval;
    }
    /**
     *
     */
    protected final boolean isWarning(String errorKey) {
        boolean warning = errorKey.startsWith("W_");
        return warning;
    }
    /**
     *
     */
    protected final boolean isFatal(String errorKey) {
        boolean fatal = errorKey.startsWith("E_");
        return fatal;
    }
    /**
     *
     */
    public final void error(int errorCode) throws Exception {
        if (!errorHandlingEnabled())
            return;
        String errorKey = ErrorCode.getErrorKey(errorCode);
        String errorMsg = getErrorMsgString(errorKey, null);
        handleError(errorKey, errorMsg);
    }
    /**
     *
     */
    public final void error1(int errorCode, int stringIndex1) throws Exception {
        String errorKey = ErrorCode.getErrorKey(errorCode);
        if (!errorHandlingEnabled() && !isFatal(errorKey)) {
            fParserState.getStringPool().releaseString(stringIndex1);
            return;
        }
        Object[] errorArgs = new String[1];
        errorArgs[0] = fParserState.getStringPool().orphanString(stringIndex1);
        String errorMsg = getErrorMsgString(errorKey, errorArgs);
        handleError(errorKey, errorMsg);
    }
    /**
     *
     */
    public final void error2(int errorCode, int stringIndex1, int stringIndex2) throws Exception {
        String errorKey = ErrorCode.getErrorKey(errorCode);
        if (!errorHandlingEnabled() && !isFatal(errorKey)) {
            fParserState.getStringPool().releaseString(stringIndex1);
            fParserState.getStringPool().releaseString(stringIndex2);
            return;
        }
        Object[] errorArgs = new String[2];
        errorArgs[0] = fParserState.getStringPool().orphanString(stringIndex1);
        errorArgs[1] = fParserState.getStringPool().orphanString(stringIndex2);
        String errorMsg = getErrorMsgString(errorKey, errorArgs);
        handleError(errorKey, errorMsg);
    }
    /**
     *
     */
    public final void error3(int errorCode, int stringIndex1, int stringIndex2, int stringIndex3) throws Exception {
        String errorKey = ErrorCode.getErrorKey(errorCode);
        if (!errorHandlingEnabled() && !isFatal(errorKey)) {
            fParserState.getStringPool().releaseString(stringIndex1);
            fParserState.getStringPool().releaseString(stringIndex2);
            fParserState.getStringPool().releaseString(stringIndex3);
            return;
        }
        Object[] errorArgs = new String[3];
        errorArgs[0] = fParserState.getStringPool().orphanString(stringIndex1);
        errorArgs[1] = fParserState.getStringPool().orphanString(stringIndex2);
        errorArgs[2] = fParserState.getStringPool().orphanString(stringIndex3);
        String errorMsg = getErrorMsgString(errorKey, errorArgs);
        handleError(errorKey, errorMsg);
    }
    /**
     *
     */
    public final void error4(int errorCode, int stringIndex1, int stringIndex2, int stringIndex3, int stringIndex4) throws Exception {
        String errorKey = ErrorCode.getErrorKey(errorCode);
        if (!errorHandlingEnabled() && (getContinueAfterFatalError() || !isFatal(errorKey))) {
            fParserState.getStringPool().releaseString(stringIndex1);
            fParserState.getStringPool().releaseString(stringIndex2);
            fParserState.getStringPool().releaseString(stringIndex3);
            fParserState.getStringPool().releaseString(stringIndex4);
            return;
        }
        Object[] errorArgs = new String[4];
        errorArgs[0] = fParserState.getStringPool().orphanString(stringIndex1);
        errorArgs[1] = fParserState.getStringPool().orphanString(stringIndex2);
        errorArgs[2] = fParserState.getStringPool().orphanString(stringIndex3);
        errorArgs[3] = fParserState.getStringPool().orphanString(stringIndex4);
        String errorMsg = getErrorMsgString(errorKey, errorArgs);
        handleError(errorKey, errorMsg);
    }
}
