/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.framework;

/**
 * This is the class used by the scanner to process the XML data.
 *
 * This class provides the scanner with location information through
 * the <code>Locator</code> interface.
 *
 * @author  Glenn Marcy
 * @version Revision: 59 1.7 src/com/ibm/xml/framework/XMLReader.java, parser, xml4j2, xml4j2_0_13 
 * @see XMLParser
 * @see org.xml.sax.Locator
 */
public abstract class XMLReader implements org.xml.sax.Locator {
    /*
     * Return values for skipPastCharData method.  This may change if we
     * pass control of the scanning thread from the scanner to the reader
     * during character data processing to reduce the overhead of passing
     * the state back and forth between the scanner and the reader as we
     * alternate between elements and character data.
     */
    public static final int
        CHARDATA_RESULT_MARKUP       =  1, // stopped at markup, i.e. '<'
        CHARDATA_RESULT_REFERENCE    =  2, // stopped at reference, i.e. '&'
        CHARDATA_RESULT_CDEND        =  3, // stopped at CDEnd, i.e. ']]>'
        CHARDATA_RESULT_INVALID_CHAR =  4, // stopped at invalid character
        CHARDATA_RESULT_STATE_MASK   =  7, // this is a mask for the previous states
        CHARDATA_RESULT_CHARDATA     =  8, // this is a flag indicating we found some CharData
        CHARDATA_RESULT_ALL_SPACE    = 16; // this is a flag indicating the CharData was all whitespace
    /*
     * Protected instance variables.
     */
    protected ParserState fParserState = null;
    protected String fPublicId = null;
    protected String fSystemId = null;
    protected int fCarriageReturnCounter = 1;
    protected int fLinefeedCounter = 1;
    protected int fCharacterCounter = 1;
    protected int fCurrentOffset = 0;
    /**
     * Default constructor
     *
     * @param parser The parser that we are working for.
     * @param publicID The public identifier of the <code>InputSource</code> that we are processing, or null if not provided.
     * @param systemID The system identifier of the <code>InputSource</code> that we are processing, or null if not provided.
     */
    protected XMLReader(ParserState parserState, String publicId, String systemId) {
        fParserState = parserState;
        fPublicId = publicId;
        fSystemId = systemId;
        initCharFlags(); // REVISIT - lazy-evaluate this call.
    }
    /**
     * Return the public identifier of the <code>InputSource</code> that we are processing.
     *
     * @return The public identifier, or null if not provided.
     * @see org.xml.sax.Locator
     */
    public String getPublicId() {
        return fPublicId;
    }
    /**
     * Return the system identifier of the <code>InputSource</code> that we are processing.
     *
     * @return The system identifier, or null if not provided.
     * @see org.xml.sax.Locator
     */
    public String getSystemId() {
        return fSystemId;
    }
    /**
     * Return the line number of the current position within the document that we are processing.
     *
     * @return The current line number.
     * @see org.xml.sax.Locator
     */
    public int getLineNumber() {
        if (fLinefeedCounter > 1)
            return fLinefeedCounter;
        else
            return fCarriageReturnCounter;
    }
    /**
     * Return the column number of the current position within the document that we are processing.
     *
     * @return The current column number.
     * @see org.xml.sax.Locator
     */
    public int getColumnNumber() {
        return fCharacterCounter;
    }
    /**
     * Add a string to the <code>StringPool</code> from the characters scanned using this
     * reader as described by <code>offset</code> and <code>length</code>.
     *
     * @param offset The offset within this reader where the characters start.
     * @param length The length within this reader where the characters end.
     * @return The index within the <code>StringPool</code> of the resulting string.
     */
    public abstract int addString(int offset, int length);
    /**
     * Add a symbol to the <code>StringPool</code> from the characters scanned using this
     * reader as described by <code>offset</code> and <code>length</code>.
     *
     * @param offset The offset within this reader where the characters start.
     * @param length The length within this reader where the characters end.
     * @return The index within the <code>StringPool</code> of the resulting symbol.
     */
    public abstract int addSymbol(int offset, int length);
    /**
     * Append the characters processed by this reader associated with <code>offset</code> and
     * <code>length</code> to the <code>ChunkyCharArray</code>.
     *
     * @param charArray The <code>ChunkyCharArray</code> to append the characters to.
     * @param offset The offset within this reader where the copy should start.
     * @param length The length within this reader where the copy should stop.
     */
    public abstract void append(ChunkyCharArray charArray, int offset, int length);
    /**
     * Skip one character.
     *
     * @return The offset within this reader where we will look next.
     */
    public abstract int skipOneChar() throws Exception;
    /**
     * Skip one ascii character.
     *
     * The caller will call this method after determining that the current character is
     * an ascii character.  This is useful if the reader can advance through the input
     * data more efficiently if it knows that the character is ascii.
     *
     * @return The offset within this reader where we will look next.
     */
    public abstract int skipAsciiChar() throws Exception;
    /**
     * Advance through the input data up to the next <code>ch</code> character.
     *
     * @return The offset within this reader where we will look next.
     */
    public abstract int skipToChar(char ch) throws Exception;
    /**
     * Advance through the input data beyond the next <code>ch</code> character.
     *
     * @return The offset within this reader where we will look next.
     */
    public abstract int skipPastChar(char ch) throws Exception;
    /**
     * Advance past one valid character.
     * 
     * @return <code>true</code> if the current character was valid and skipped;
     *         <code>false</code> otherwise.
     */
    public abstract boolean skippedValidChar() throws Exception;
    /**
     * Test that the current character is valid.
     *
     * @return <code>true</code> if the current character is valid;
     *         <code>false</code> otherwise.
     */
    public abstract boolean lookingAtValidChar() throws Exception;
    /**
     * Advance past an invalid character, reporting it as an <code>errorCode</code> error.
     *
     * @return The offset within this reader where we will look next.
     */
    public abstract int skipInvalidChar(int errorCode) throws Exception;
    /**
     * Advance past a <code>ch</code> character.
     *
     * @param ch The character to match against.
     * @return <code>true</code> if the current character was a <code>ch</code> character and skipped;
     *         <code>false</code> otherwise.
     */
    public abstract boolean skippedChar(char ch) throws Exception;
    /**
     * Test that the current character is a <code>ch</code> character.
     *
     * @param ch The character to match against.
     * @return <code>true</code> if the current character is a <code>ch</code> character;
     *         <code>false</code> otherwise.
     */
    public abstract boolean lookingAtChar(char ch) throws Exception;
    /**
     * Advance past one whitespace character.
     *
     * @return <code>true</code> if the current character was whitespace and skipped;
     *         <code>false</code> otherwise.
     */
    public abstract boolean skippedSpace() throws Exception;
    /**
     * Test that the current character is a whitespace character.
     *
     * @return <code>true</code> if the current character is whitespace;
     *         <code>false</code> otherwise.
     */
    public abstract boolean lookingAtSpace() throws Exception;
    /**
     * Skip past whitespace characters starting at the current position.
     *
     * @return The offset within this reader where we will look next.
     */
    public abstract int skipPastSpaces() throws Exception;
    /**
     * Skip a single decimal digit character.
     *
     * @return The value of the skipped character, if it was a decimal digit, as an integer
     *         between 0 and 9, or -1 if the current character is not a decimal digit.
     */
    public abstract int skipDecimalDigit() throws Exception;
    /**
     * Skip a single hexadecimal digit character.
     *
     * @return The value of the skipped character, if it was a hexadecimal digit, as an integer
     *         between 0 and 15, or -1 if the current character is not a hexadecimal digit.
     */
    public abstract int skipHexDigit() throws Exception;
    /**
     * Skip a single ascii alpha character.  An ascii alpha character is either in the ascii
     * range 'a' through 'z' or in the range 'A' through 'Z'.
     *
     * @return <code>true</code> if the current character was alpha and skipped;
     *         <code>false</code> otherwise.
     */
    public abstract boolean skippedAlpha() throws Exception;
    /**
     * Skip a single character defined by XML as a VersionNum character.
     *
     * @return <code>true</code> if the current character is a VersionNum character and skipped;
     *         <code>false</code> otherwise.
     */
    public abstract boolean skippedVersionNum() throws Exception;
    /**
     * Skip a single character defined by XML as a EncName character.
     *
     * @return <code>true</code> if the current character is a EncName character and skipped;
     *         <code>false</code> otherwise.
     */
    public abstract boolean skippedEncName() throws Exception;
    /**
     * Skip a single character defined by XML as a PubidChar character.
     *
     * @return <code>true</code> if the current character is a PubidChar character and skipped;
     *         <code>false</code> otherwise.
     */
    public abstract boolean skippedPubidChar() throws Exception;
    /**
     * Skip past a sequence of characters that matches the specified character array.
     *
     * @param s The characters to match.
     * @return <code>true</code> if the current character is valid;
     *         <code>false</code> otherwise.
     */
    public abstract boolean skippedString(char[] s) throws Exception;
    /**
     * Add a sequence of characters that match the XML definition of a Name to the <code>StringPool</code>.
     *
     * If we find a name at the current position we will add it to the <code>StringPool</code>
     * as a symbol and will return the index within the string pool to the caller.
     *
     * @param fastcheck A character that is not a legal name character that is provided as a
     *                  hint to the reader of a character likely to terminate the Name.
     * @param expectedName The index within the <code>StringPool</code> of the name that
     *                     must match the name at the current position, or -1 if there is no
     *                     specific name expected.
     * @return The index within the <code>StringPool</code> of the name that
     *         was scanned, or -1 if a name was not found at the current position
     *         within the input data.
     * @see StringPool
     */
    public abstract int scanName(char fastcheck, int expectedName) throws Exception;
    /**
     * Skip past a sequence of characters that match the XML definition of a Name.
     *
     * @return The offset within this reader where we will look next.
     */
    public abstract int skipPastName(char fastcheck) throws Exception;
    /**
     * Skip past a sequence of characters that match the XML definition of an Nmtoken.
     *
     * @return The offset within this reader where we will look next.
     */
    public abstract int skipPastNmtoken(char fastcheck) throws Exception;
    /**
     * Skip through the input data while we are looking at character data.
     *
     * The result codes are:
     *
     *   CHARDATA_RESULT_MARKUP         // stopped at markup, i.e. '<'
     *   CHARDATA_RESULT_REFERENCE      // stopped at reference, i.e. '&'
     *   CHARDATA_RESULT_CDEND          // stopped at CDEnd, i.e. ']]>'
     *   CHARDATA_RESULT_INVALID_CHAR   // stopped at invalid character
     *   CHARDATA_RESULT_CHARDATA       // a flag indicating we found some CharData
     *   CHARDATA_RESULT_ALL_SPACE      // a flag indicating the CharData was all whitespace
     *
     * @return The result code.
     * @see ScanContentState
     */
    public abstract int scanContent(ScanContentState scanState) throws Exception;
    /**
     * Return the current offset within this reader.
     *
     * @return The offset.
     */
    public final int currentOffset() {
        return fCurrentOffset;
    }
    /*
     * Here are tables used to build character properties.
     *
     * REVISIT - need to add scanVersionNum, scanEncodingName, scanPublicId, etc., instead
     *  of "one char at a time" access.
     */
    protected static final byte fgAsciiXDigitChar[] = {
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, // '0' - '9'
        0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, // 'A' - 'F'
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, // 'a' - 'f'
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    };
    protected static final byte fgAsciiAlphaChar[] = {
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 'A' - 'O'
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, // 'P' - 'Z'
        0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 'a' - 'o'
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0  // 'p' - 'z'
    };
    protected static final byte fgAsciiInitialNameChar[] = {
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, // ':' is 0x3A
        0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 'A' - 'O'
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, // 'P' - 'Z' and '_' is 0x5F
        0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 'a' - 'o'
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0  // 'p' - 'z'
    };
    protected static final byte fgAsciiNameChar[] = {
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, // '-' is 0x2D and '.' is 0x2E
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, // '0' - '9' and ':' is 0x3A
        0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 'A' - 'O'
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, // 'P' - 'Z' and '_' is 0x5F
        0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 'a' - 'o'
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0  // 'p' - 'z'
    };
    protected static final byte fgAsciiCharData[] = {
        4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 4, 4, 4, 4, 4, 4, // tab is 0x09
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, // '&' is 0x26
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, // '<' is 0x3C
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, // ']' is 0x5D
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    };
    protected static final byte fgAsciiWSCharData[] = {
        4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 4, 4, 5, 4, 4, // tab is 0x09,  LF is 0x0A,  CR is 0x0D
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        5, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, // ' ' is 0x20, '&' is 0x26
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, // '<' is 0x3C
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, // ']' is 0x5D
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    };
    protected static final byte E_VersionNumFlag = 1<<0;
    protected static final byte E_EncNameFlag = 1<<1;
    protected static final byte E_PubidCharFlag = 1<<2;
    protected static final byte E_CharDataFlag = 1<<3;
    protected static final byte E_InitialNameCharFlag = 1<<4;
    protected static final byte E_NameCharFlag = 1<<5;
    protected static byte[] fgCharFlags = null;
    private static synchronized void initCharFlags() {
        if (fgCharFlags == null) {
            fgCharFlags = new byte[0x10000];
            setFlagForRange(fgVersionNumRanges, E_VersionNumFlag);
            setFlagForRange(fgEncNameRanges, E_EncNameFlag);
            setFlagForRange(fgPubidCharRanges, E_PubidCharFlag);
            setFlagForRange(fgCharDataRanges, E_CharDataFlag);
            setFlagForRange(fgInitialNameCharRanges, (byte)(E_InitialNameCharFlag | E_NameCharFlag));
            setFlagForRange(fgNameCharRanges, E_NameCharFlag);
        }
    }
    private static void setFlagForRange(char[] ranges, byte flag)
    {
        int i;
        int ch;
        for (i = 0; (ch = ranges[i]) != 0; i += 2) {
            int endch = ranges[i+1];
            while (ch <= endch)
                fgCharFlags[ch++] |= flag;
        }
        for (i++; (ch = ranges[i]) != 0; i++)
            fgCharFlags[ch] |= flag;
    }
    /*
     * [26] VersionNum ::= ([a-zA-Z0-9_.:] | '-')+
     */
    private static final char fgVersionNumRanges[] = {
        0x002D, 0x002E, // '-' and '.'
        0x0030, 0x003A, // '0' - '9' and ':'
        0x0041, 0x0046, // 'A' - 'F'
        0x0061, 0x0066, // 'a' - 'f'
            0x0000,
        0x005F, // '_'
            0x0000
    };
    /*
     * [81] EncName ::= [A-Za-z] ([A-Za-z0-9._] | '-')*
     */
    private static final char fgEncNameRanges[] = {
        0x002D, 0x002E, // '-' and '.'
        0x0030, 0x0039, // '0' - '9'
        0x0041, 0x005A, // 'A' - 'Z'
        0x0061, 0x007A, // 'a' - 'z'
            0x0000,
        0x005F, // '_'
            0x0000
    };
    /*
     * [13] PubidChar ::= #x20 | #xD | #xA | [a-zA-Z0-9] | [-'()+,./:=?;!*#@$_%]
     */
    private static final char fgPubidCharRanges[] = {
        0x0020, 0x0021, // ' ' and '!'
        0x0023, 0x0025, // '#', '$' and '%'
        0x0027, 0x003B, // '\'', '(', ')', '*', '+', ',', '-', '.', '/', '0' - '9', ':' and ';'
        0x003F, 0x005A, // '?', '@' and 'A' - 'Z'
        0x0061, 0x007A, // 'a' - 'z'
            0x0000,
        0x003D, // '='
        0x005F, // '_'
            0x0000
    };
    /*
     *  [2] Char ::= #x9 | #xA | #xD | [#x20-#xD7FF]        // any Unicode character, excluding the
     *               | [#xE000-#xFFFD] | [#x10000-#x10FFFF] // surrogate blocks, FFFE, and FFFF.
     * [14] CharData ::= [^<&]* - ([^<&]* ']]>' [^<&]*)
     *
     * We will use Char - ( [^<&] | ']' | #xA | #xD ) and handle the special cases inline.
     */
    private static final char fgCharDataRanges[] = {
        0x0020, 0x0025, // '&' is 0x0026
        0x0027, 0x003B, // '<' is 0x003C
        0x003D, 0x005C, // ']' is 0x005D
        0x005E, 0xD7FF,
        0xE000, 0xFFFD,
            0x0000,
        0x0009,         // tab
            0x0000
    };
    /*
     *  [5] Name ::= (Letter | '_' | ':') (NameChar)*
     *  [4] NameChar ::= Letter | Digit | '.' | '-' | '_' | ':' | CombiningChar | Extender
     * [84] Letter ::= BaseChar | Ideographic
     * [85] BaseChar ::= <see standard>
     * [86] Ideographic ::= <see standard>
     * [87] CombiningChar ::= <see standard>
     * [88] Digit ::= <see standard>
     * [89] Extender ::= <see standard>
     */
    private static final char fgInitialNameCharRanges[] = {
        //
        // Ranges:
        //
        //  BaseChar ranges
        //
        0x0041, 0x005A, 0x0061, 0x007A, 0x00C0, 0x00D6, 0x00D8, 0x00F6,
        0x00F8, 0x0131, 0x0134, 0x013E, 0x0141, 0x0148, 0x014A, 0x017E,
        0x0180, 0x01C3, 0x01CD, 0x01F0, 0x01F4, 0x01F5, 0x01FA, 0x0217,
        0x0250, 0x02A8, 0x02BB, 0x02C1, 0x0388, 0x038A, 0x038E, 0x03A1,
        0x03A3, 0x03CE, 0x03D0, 0x03D6, 0x03E2, 0x03F3, 0x0401, 0x040C,
        0x040E, 0x044F, 0x0451, 0x045C, 0x045E, 0x0481, 0x0490, 0x04C4,
        0x04C7, 0x04C8, 0x04CB, 0x04CC, 0x04D0, 0x04EB, 0x04EE, 0x04F5,
        0x04F8, 0x04F9, 0x0531, 0x0556, 0x0561, 0x0586, 0x05D0, 0x05EA,
        0x05F0, 0x05F2, 0x0621, 0x063A, 0x0641, 0x064A, 0x0671, 0x06B7,
        0x06BA, 0x06BE, 0x06C0, 0x06CE, 0x06D0, 0x06D3, 0x06E5, 0x06E6,
        0x0905, 0x0939, 0x0958, 0x0961, 0x0985, 0x098C, 0x098F, 0x0990,
        0x0993, 0x09A8, 0x09AA, 0x09B0, 0x09B6, 0x09B9, 0x09DC, 0x09DD,
        0x09DF, 0x09E1, 0x09F0, 0x09F1, 0x0A05, 0x0A0A, 0x0A0F, 0x0A10,
        0x0A13, 0x0A28, 0x0A2A, 0x0A30, 0x0A32, 0x0A33, 0x0A35, 0x0A36,
        0x0A38, 0x0A39, 0x0A59, 0x0A5C, 0x0A72, 0x0A74, 0x0A85, 0x0A8B,
        0x0A8F, 0x0A91, 0x0A93, 0x0AA8, 0x0AAA, 0x0AB0, 0x0AB2, 0x0AB3,
        0x0AB5, 0x0AB9, 0x0B05, 0x0B0C, 0x0B0F, 0x0B10, 0x0B13, 0x0B28,
        0x0B2A, 0x0B30, 0x0B32, 0x0B33, 0x0B36, 0x0B39, 0x0B5C, 0x0B5D,
        0x0B5F, 0x0B61, 0x0B85, 0x0B8A, 0x0B8E, 0x0B90, 0x0B92, 0x0B95,
        0x0B99, 0x0B9A, 0x0B9E, 0x0B9F, 0x0BA3, 0x0BA4, 0x0BA8, 0x0BAA,
        0x0BAE, 0x0BB5, 0x0BB7, 0x0BB9, 0x0C05, 0x0C0C, 0x0C0E, 0x0C10,
        0x0C12, 0x0C28, 0x0C2A, 0x0C33, 0x0C35, 0x0C39, 0x0C60, 0x0C61,
        0x0C85, 0x0C8C, 0x0C8E, 0x0C90, 0x0C92, 0x0CA8, 0x0CAA, 0x0CB3,
        0x0CB5, 0x0CB9, 0x0CE0, 0x0CE1, 0x0D05, 0x0D0C, 0x0D0E, 0x0D10,
        0x0D12, 0x0D28, 0x0D2A, 0x0D39, 0x0D60, 0x0D61, 0x0E01, 0x0E2E,
        0x0E32, 0x0E33, 0x0E40, 0x0E45, 0x0E81, 0x0E82, 0x0E87, 0x0E88,
        0x0E94, 0x0E97, 0x0E99, 0x0E9F, 0x0EA1, 0x0EA3, 0x0EAA, 0x0EAB,
        0x0EAD, 0x0EAE, 0x0EB2, 0x0EB3, 0x0EC0, 0x0EC4, 0x0F40, 0x0F47,
        0x0F49, 0x0F69, 0x10A0, 0x10C5, 0x10D0, 0x10F6, 0x1102, 0x1103,
        0x1105, 0x1107, 0x110B, 0x110C, 0x110E, 0x1112, 0x1154, 0x1155,
        0x115F, 0x1161, 0x116D, 0x116E, 0x1172, 0x1173, 0x11AE, 0x11AF,
        0x11B7, 0x11B8, 0x11BC, 0x11C2, 0x1E00, 0x1E9B, 0x1EA0, 0x1EF9,
        0x1F00, 0x1F15, 0x1F18, 0x1F1D, 0x1F20, 0x1F45, 0x1F48, 0x1F4D,
        0x1F50, 0x1F57, 0x1F5F, 0x1F7D, 0x1F80, 0x1FB4, 0x1FB6, 0x1FBC,
        0x1FC2, 0x1FC4, 0x1FC6, 0x1FCC, 0x1FD0, 0x1FD3, 0x1FD6, 0x1FDB,
        0x1FE0, 0x1FEC, 0x1FF2, 0x1FF4, 0x1FF6, 0x1FFC, 0x212A, 0x212B,
        0x2180, 0x2182, 0x3041, 0x3094, 0x30A1, 0x30FA, 0x3105, 0x312C,
        0xAC00, 0xD7A3,
        //
        //  Ideographic ranges
        //
        0x3021, 0x3029, 0x4E00, 0x9FA5,
        //
        // Ranges end marker
        //
        0x0000,
        //
        // Single char values
        //
        0x003A, // ':'
        0x005F, // '_'
        //
        //  BaseChar singles
        //
        0x0386, 0x038C, 0x03DA, 0x03DC, 0x03DE, 0x03E0, 0x0559, 0x06D5,
        0x093D, 0x09B2, 0x0A5E, 0x0A8D, 0x0ABD, 0x0AE0, 0x0B3D, 0x0B9C,
        0x0CDE, 0x0E30, 0x0E84, 0x0E8A, 0x0E8D, 0x0EA5, 0x0EA7, 0x0EB0,
        0x0EBD, 0x1100, 0x1109, 0x113C, 0x113E, 0x1140, 0x114C, 0x114E,
        0x1150, 0x1159, 0x1163, 0x1165, 0x1167, 0x1169, 0x1175, 0x119E,
        0x11A8, 0x11AB, 0x11BA, 0x11EB, 0x11F0, 0x11F9, 0x1F59, 0x1F5B,
        0x1F5D, 0x1FBE, 0x2126, 0x212E,
        //
        //  Ideographic singles
        //
        0x3007,
        //
        // Singles end marker
        //
        0x0000
    };
    private static final char fgNameCharRanges[] = {
        //
        // Ranges:
        //
        0x002D, 0x002E, // '-' and '.'
        //
        //  CombiningChar ranges
        //
        0x0300, 0x0345, 0x0360, 0x0361, 0x0483, 0x0486, 0x0591, 0x05A1,
        0x05A3, 0x05B9, 0x05BB, 0x05BD, 0x05C1, 0x05C2, 0x064B, 0x0652,
        0x06D6, 0x06DC, 0x06DD, 0x06DF, 0x06E0, 0x06E4, 0x06E7, 0x06E8,
        0x06EA, 0x06ED, 0x0901, 0x0903, 0x093E, 0x094C, 0x0951, 0x0954,
        0x0962, 0x0963, 0x0981, 0x0983, 0x09C0, 0x09C4, 0x09C7, 0x09C8,
        0x09CB, 0x09CD, 0x09E2, 0x09E3, 0x0A40, 0x0A42, 0x0A47, 0x0A48,
        0x0A4B, 0x0A4D, 0x0A70, 0x0A71, 0x0A81, 0x0A83, 0x0ABE, 0x0AC5,
        0x0AC7, 0x0AC9, 0x0ACB, 0x0ACD, 0x0B01, 0x0B03, 0x0B3E, 0x0B43,
        0x0B47, 0x0B48, 0x0B4B, 0x0B4D, 0x0B56, 0x0B57, 0x0B82, 0x0B83,
        0x0BBE, 0x0BC2, 0x0BC6, 0x0BC8, 0x0BCA, 0x0BCD, 0x0C01, 0x0C03,
        0x0C3E, 0x0C44, 0x0C46, 0x0C48, 0x0C4A, 0x0C4D, 0x0C55, 0x0C56,
        0x0C82, 0x0C83, 0x0CBE, 0x0CC4, 0x0CC6, 0x0CC8, 0x0CCA, 0x0CCD,
        0x0CD5, 0x0CD6, 0x0D02, 0x0D03, 0x0D3E, 0x0D43, 0x0D46, 0x0D48,
        0x0D4A, 0x0D4D, 0x0E34, 0x0E3A, 0x0E47, 0x0E4E, 0x0EB4, 0x0EB9,
        0x0EBB, 0x0EBC, 0x0EC8, 0x0ECD, 0x0F18, 0x0F19, 0x0F71, 0x0F84,
        0x0F86, 0x0F8B, 0x0F90, 0x0F95, 0x0F99, 0x0FAD, 0x0FB1, 0x0FB7,
        0x20D0, 0x20DC, 0x302A, 0x302F,
        //
        //  Digit ranges
        //
        0x0030, 0x0039, 0x0660, 0x0669, 0x06F0, 0x06F9, 0x0966, 0x096F,
        0x09E6, 0x09EF, 0x0A66, 0x0A6F, 0x0AE6, 0x0AEF, 0x0B66, 0x0B6F,
        0x0BE7, 0x0BEF, 0x0C66, 0x0C6F, 0x0CE6, 0x0CEF, 0x0D66, 0x0D6F,
        0x0E50, 0x0E59, 0x0ED0, 0x0ED9, 0x0F20, 0x0F29,
        //
        //  Extender ranges
        //
        0x3031, 0x3035, 0x309D, 0x309E, 0x30FC, 0x30FE,
        //
        // Ranges end marker
        //
        0x0000,
        //
        // Single char values
        //
        //  CombiningChar singles
        //
        0x05BF, 0x05C4, 0x0670, 0x093C, 0x094D, 0x09BC, 0x09BE, 0x09BF,
        0x09D7, 0x0A02, 0x0A3C, 0x0A3E, 0x0A3F, 0x0ABC, 0x0B3C, 0x0BD7,
        0x0D57, 0x0E31, 0x0EB1, 0x0F35, 0x0F37, 0x0F39, 0x0F3E, 0x0F3F,
        0x0F97, 0x0FB9, 0x20E1, 0x3099, 0x309A,
        //
        //  Extender singles
        //
        0x00B7, 0x02D0, 0x02D1, 0x0387, 0x0640, 0x0E46, 0x0EC6, 0x3005,
        //
        // Singles end marker
        //
        0x0000
    };
}
