/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.framework;

/**
 * This class defines the interface between the scanner and a validator. As the
 * scanner scans the XML content, it will call these methods to validate the data
 * as it goes. It is assumed that the implementation of the interface will use the
 * provided XMLParser object to gain access to the element declaration and string
 * pools required to interpret the passed element and attribute indexes.
 *
 * @author  Glenn Marcy
 * @author  Dean Roddey
 * @version Revision: 52 1.8 src/com/ibm/xml/framework/XMLValidationHandler.java, parser, xml4j2, xml4j2_0_13 
 * @see XMLParser
 */
public interface XMLValidationHandler {

    /**
     * Reset the handler.
     *
     * @param parserState The parser state to use for the next parse call.
     */
    public void reset(ParserState parserState);

    /**
     * Check that the root element name is valid.
     * <p>
     * This method is called when the scanner sees the root element of the target
     * XML document. It will pass the element name index of the root element, which
     * the validator must confirm was the root name indicated in the DTD (or one of
     * the valid roots indicated by a Schema.)
     *
     * @param elementNameIndex The index within the <code>StringPool</code> of the
     *                         root element name.
     *
     * @exception Exception Thrown on error.
     */
    public void checkRootElementName(int elementNameIndex) throws Exception;

    /**
     * Check that the attributes for an element are valid.
     * <p>
     * This method is called as a convenience. The scanner will do any required
     * checks for well-formedness on the attributes, as well as fill out any
     * defaulted ones. However, if the validator has any other constraints or
     * semantics it must enforce, it can use this API to do so.
     * <p>
     * The scanner provides the element index (within the decl pool, i.e. not the
     * name index which is within the string pool), and the index of the first
     * attribute within the attribute pool that holds the attributes of the element.
     * By this time, all defaulted attributes are present and all fixed values have
     * been confirmed. For most validators, this will be a no-op.
     *
     * @param elementIndex The index within the <code>ElementDeclPool</code> of
     *                     this element.
     * @param firstAttrIndex The index within the <code>AttrPool</code> of the
     *                       first attribute of this element, or -1 if there are
     *                       no attributes.
     *
     * @exception Exception Thrown on error.
     */
    public void checkAttributes(int elementIndex, int firstAttrIndex) throws Exception;

    /**
     * Check that the content of an element is valid.
     * <p>
     * This is the method of primary concern to the validator. This method is called
     * upon the scanner reaching the end tag of an element. At that time, the
     * element's children must be structurally validated, so it calls this method.
     * The index of the element being checked (in the decl pool), is provided as
     * well as an array of element name indexes of the children. The validator must
     * confirm that this element can have these children in this order.
     * <p>
     * This can also be called to do 'what if' testing of content models just to see
     * if they would be valid.
     * <p>
     * Note that the element index is an index into the element decl pool, whereas
     * the children indexes are name indexes, i.e. into the string pool.
     * <p>
     * A value of -1 in the children array indicates a PCDATA node. All other 
     * indexes will be positive and represent child elements. The count can be
     * zero, since some elements have the EMPTY content model and that must be 
     * confirmed.
     *
     * @param elementIndex The index within the <code>ElementDeclPool</code> of this
     *                     element.
     * @param childCount The number of entries in the <code>children</code> array.
     * @param children The children of this element.  Each integer is an index within
     *                 the <code>StringPool</code> of the child element name.  An index
     *                 of -1 is used to indicate an occurrence of non-whitespace character
     *                 data.
     *
     * @return The value -1 if fully valid, else the 0 based index of the child
     *         that first failed. If the value returned is equal to the number
     *         of children, then the specified children are valid but additional
     *         content is required to reach a valid ending state.
     *
     * @exception Exception Thrown on error.
     */
    public int checkContent(int elementIndex, int childCount, int[] children) throws Exception;

    /**
     * Check that all ID references were to ID attributes present in the document.
     * <p>
     * This method is a convenience call that allows the validator to do any id ref
     * checks above and beyond those done by the scanner. The scanner does the checks
     * specificied in the XML spec, i.e. that ID refs refer to ids which were
     * eventually defined somewhere in the document.
     * <p>
     * If the validator is for a Schema perhaps, which defines id semantics beyond
     * those of the XML specificiation, this is where that extra checking would be
     * done. For most validators, this is a no-op.
     *
     * @exception Exception Thrown on error.
     */
    public void checkIDRefNames() throws Exception;

    /**
     * Returns information about which elements can be placed at a particular point
     * in the passed element's content model.
     * <p>
     * Note that the incoming content model to test must be valid at least up to
     * the insertion point. If not, then -1 will be returned and the info object
     * will not have been filled in.
     * <p>
     * If, on return, the info.isValidEOC flag is set, then the 'insert after'
     * elemement is a valid end of content, i.e. nothing needs to be inserted
     * after it to make the parent element's content model valid.
     *
     * @param elementIndex The index within the <code>ElementDeclPool</code> of the
     *                     element which is being querying.
     * @param fullyValid Only return elements that can be inserted and still
     *                   maintain the validity of subsequent elements past the
     *                   insertion point (if any).  If the insertion point is at
     *                   the end, and this is true, then only elements that can
     *                   be legal final states will be returned.
     * @param info An object that contains the required input data for the method,
     *             and which will contain the output information if successful.
     *
     * @return The value -1 if fully valid, else the 0 based index of the child
     *         that first failed before the insertion point. If the value 
     *         returned is equal to the number of children, then the specified
     *         children are valid but additional content is required to reach a
     *         valid ending state.
     *
     * @exception Exception Thrown on error.
     *
     * @see InsertableElementsInfo
     */
    public int whatCanGoHere(int elementIndex, boolean fullyValid, InsertableElementsInfo info) throws Exception;
}
