/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */
package com.ibm.xml.internal;

import com.ibm.xml.framework.*;

import java.io.Reader;

/**
 *
 * @version Revision: 73 1.9 src/com/ibm/xml/internal/CharReader.java, parser, xml4j2, xml4j2_0_13 
 */
final class CharReader extends XMLReader {
    //
    // Chunk size constants
    //
    protected static final int CHUNK_SHIFT = 14;           // 2^14 = 16k
    protected static final int CHUNK_SIZE = (1 << CHUNK_SHIFT);
    protected static final int CHUNK_MASK = CHUNK_SIZE - 1;
    //
    //
    //
    CharReader(ParserState parserState, String publicId, String systemId, Reader reader) throws Exception {
        super(parserState, publicId, systemId);
        fReader = reader;
        fScanner = parserState.getScanner();
        fStringPool = parserState.cacheStringPool();
        fDocumentHandler = parserState.getDocumentHandler();
        fErrorHandler = parserState.getErrorHandler();
        fCurrentChunk = new CharDataChunk(fStringPool, null);
        fillCurrentChunk();
    }
    /**
     *
     */
    public int addString(int offset, int length) {
        if (length == 0)
            return 0;
        return fCurrentChunk.addString(offset, length);
    }
    /**
     *
     */
    public int addSymbol(int offset, int length) {
        if (length == 0)
            return 0;
        return fCurrentChunk.addSymbol(offset, length, 0);
    }
    /**
     *
     */
    protected int addSymbol(int offset, int length, int hashcode) {
        if (length == 0)
            return 0;
        return fCurrentChunk.addSymbol(offset, length, hashcode);
    }
    /**
     *
     */
    public void append(ChunkyCharArray charArray, int offset, int length) {
        fCurrentChunk.append(charArray, offset, length);
    }
    //
    //
    //
    protected int slowLoadNextChar()  throws Exception {
        if (fCurrentChunk.nextChunk() != null) {
            fCurrentChunk = fCurrentChunk.nextChunk();
            fCurrentIndex = 0;
            fMostRecentData = fCurrentChunk.toCharArray();
            return (fMostRecentChar = fMostRecentData[fCurrentIndex] & 0xFFFF);
        } else {
            fCurrentChunk = new CharDataChunk(fStringPool, fCurrentChunk);
            return fillCurrentChunk();
        }
    }
    protected int loadNextChar() throws Exception {
        fCurrentOffset++;
        if (++fCurrentIndex == CHUNK_SIZE)
            return slowLoadNextChar();
        return (fMostRecentChar = fMostRecentData[fCurrentIndex] & 0xFFFF);
    }
    //
    //
    //
    protected void checkEOF(int offset) {
        if (offset > fLength)
            throw new ArrayIndexOutOfBoundsException();
    }
    //
    //
    //
    public int skipOneChar() throws Exception {
        int ch = fMostRecentChar;
        if (ch == 0) {
            //
            // REVISIT - shouldn't happen... the exception should get thrown
            //  when you look at the EOF before you ask to skip past it.
            //
            checkEOF(fCurrentOffset + 1);
            fCharacterCounter++;
        } else if (ch == 0x0A) {
            fLinefeedCounter++;
            fCharacterCounter = 1;
        } else if (ch >= 0xD800 && ch < 0xDC00) {
            fCharacterCounter++;
            ch = loadNextChar();
            if (ch < 0xDC00 || ch >= 0xE000)
                return fCurrentOffset;
        } else
            fCharacterCounter++;
        loadNextChar();
        return fCurrentOffset;
    }
    public int skipAsciiChar() throws Exception {
        fCharacterCounter++; // we never call skip_asciiChar for CR or LF.
        loadNextChar();
        return fCurrentOffset;
    }
    public int skipToChar(char chr) throws Exception {
        int ch = fMostRecentChar;
        while (true) {
            if (ch == chr)
                return fCurrentOffset;
            if (ch == 0) {
                //
                // REVISIT - shouldn't happen... the exception should get thrown
                //  when you look at the EOF before you ask to skip past it.
                //
                checkEOF(fCurrentOffset + 1);
                fCharacterCounter++;
            } else if (ch == 0x0A) {
                fLinefeedCounter++;
                fCharacterCounter = 1;
            } else if (ch >= 0xD800 && ch < 0xDC00) {
                fCharacterCounter++;
                ch = loadNextChar();
                if (ch < 0xDC00 || ch >= 0xE000)
                    continue;
            } else
                fCharacterCounter++;
            ch = loadNextChar();
        }
    }
    public int skipPastChar(char chr) throws Exception {
        int ch = fMostRecentChar;
        while (true) {
            if (ch == chr) {
                fCharacterCounter++; // chr will never be CR or LF.
                loadNextChar();
                return fCurrentOffset;
            }
            if (ch == 0) {
                //
                // REVISIT - shouldn't happen... the exception should get thrown
                //  when you look at the EOF before you ask to skip past it.
                //
                checkEOF(fCurrentOffset + 1);
                fCharacterCounter++;
            } else if (ch == 0x0A) {
                fLinefeedCounter++;
                fCharacterCounter = 1;
            } else if (ch >= 0xD800 && ch < 0xDC00) {
                fCharacterCounter++;
                ch = loadNextChar();
                if (ch < 0xDC00 || ch >= 0xE000)
                    continue;
            } else
                fCharacterCounter++;
            ch = loadNextChar();
        }
    }
    //
    //
    //
    public boolean skippedValidChar() throws Exception {
        int ch = fMostRecentChar;
        if (ch < 0xD800) {
            if (ch >= 0x20 || ch == 0x09) {
                fCharacterCounter++;
                loadNextChar();
                return true;
            }
            if (ch == 0x0A) {
                fLinefeedCounter++;
                fCharacterCounter = 1;
                loadNextChar();
                return true;
            }
            if (ch == 0)
                checkEOF(fCurrentOffset + 1);
            return false;
        }
        if (ch > 0xFFFD) {
            return false;
        }
        if (ch < 0xDC00) {
            ch = loadNextChar();
            if (ch < 0xDC00 || ch >= 0xE000)
                return false;
        } else if (ch < 0xE000) {
            return false;
        }
        fCharacterCounter++;
        loadNextChar();
        return true;
    }
    //
    //
    //
    public boolean lookingAtValidChar() throws Exception {
        int ch = fMostRecentChar;
        if (ch < 0x20) {
            if (ch == 0)
                checkEOF(fCurrentOffset + 1);
            return (ch == 0x09 || ch == 0x0A);
        }
        //
        // REVISIT - this does not handle surrogate pairs properly...
        //
        return (ch <= 0xFFFD);
    }
    //
    //
    //
    public int skipInvalidChar(int errorCode) throws Exception {
        int ch = fMostRecentChar;
        if (ch == 0x0A) {
            fLinefeedCounter++;
            fCharacterCounter = 1;
            loadNextChar();
        } else {
            fCharacterCounter++;
            if (ch >= 0xD800 && ch < 0xDC00) {
                int ch2 = loadNextChar();
                if (ch2 >= 0xDC00 && ch2 < 0xE000) {
                    ch = ((ch-0xD800)<<10)+(ch2-0xDC00)+0x10000;
                    loadNextChar();
                }
            } else
                loadNextChar();
        }
        switch (errorCode) {
        case ErrorCode.E_INVCHAR0:
        case ErrorCode.E_ENTITY3:
        {
            String arg = Integer.toHexString(ch);
            fErrorHandler.error1(errorCode, fStringPool.addString(arg));
            break;
        }
        case ErrorCode.E_REFER0:
        case ErrorCode.E_EXT2:
        case ErrorCode.E_EXT4:
        {
            String arg = (ch < 0x10000) ? (new Character((char)ch).toString()) : Integer.toHexString(ch);
            fErrorHandler.error1(errorCode, fStringPool.addString(arg));
            break;
        }
        case ErrorCode.E_DTD0:
        {
            String arg1 = (ch < 0x10000) ? (new Character((char)ch).toString()) : Integer.toHexString(ch);
            String arg2 = Integer.toHexString(ch);
            fErrorHandler.error2(errorCode, fStringPool.addString(arg1), fStringPool.addString(arg2));
            break;
        }
        default:
            break;
        }
        return fCurrentOffset;
    }
    //
    //
    //
    public boolean skippedChar(char chr) throws Exception {
        int ch = fMostRecentChar;
        if (ch != chr)
            return false;
        fCharacterCounter++;
        fCurrentOffset++;
        if (++fCurrentIndex == CHUNK_SIZE)
            slowLoadNextChar();
        else
            fMostRecentChar = fMostRecentData[fCurrentIndex] & 0xFFFF;
        return true;
    }
    public boolean lookingAtChar(char chr) throws Exception {
        int ch = fMostRecentChar;
        return ch == chr;
    }
    //
    //
    //
    public boolean skippedSpace() throws Exception {
        int ch = fMostRecentChar;
        if (ch > 0x20)
            return false;
        if (ch == 0x20 || ch == 0x09)
            fCharacterCounter++;
        else if (ch == 0x0A) {
            fLinefeedCounter++;
            fCharacterCounter = 1;
        } else
            return false;
        fCurrentOffset++;
        if (++fCurrentIndex == CHUNK_SIZE)
            slowLoadNextChar();
        else
            fMostRecentChar = fMostRecentData[fCurrentIndex] & 0xFFFF;
        return true;
    }
    public boolean lookingAtSpace() {
        int ch = fMostRecentChar;
        return (ch == 0x20 || ch == 0x09 || ch == 0x0A);
    }
    //
    //
    //
    public int skipPastSpaces() throws Exception {
        int ch = fMostRecentChar;
        while (true) {
            if (ch == 0x20 || ch == 0x09)
                fCharacterCounter++;
            else if (ch == 0x0A) {
                fLinefeedCounter++;
                fCharacterCounter = 1;
            } else
                return fCurrentOffset;
            ch = loadNextChar();
        }
    }
    //
    //
    //
    public int skipDecimalDigit() throws Exception {
        int ch = fMostRecentChar;
        if (ch < '0' || ch > '9')
            return -1;
        fCharacterCounter++;
        loadNextChar();
        return ch - '0';
    }
    public int skipHexDigit() throws Exception {
        int ch = fMostRecentChar;
        if (ch > 'f' || fgAsciiXDigitChar[ch] == 0)
            return -1;
        fCharacterCounter++;
        loadNextChar();
        return ch - (ch < 'A' ? '0' : (ch < 'a' ? 'A' : 'a') - 10);
    }
    public boolean skippedAlpha() throws Exception {
        int ch = fMostRecentChar;
        if (ch > 'z' || fgAsciiAlphaChar[ch] == 0)
            return false;
        fCharacterCounter++;
        loadNextChar();
        return true;
    }
    protected boolean skippedAsciiCharWithFlag(byte flag) throws Exception {
        int ch = fMostRecentChar;
        if (ch >= 0x80 || (fgCharFlags[ch] & flag) == 0)
            return false;
        fCharacterCounter++;
        loadNextChar();
        return true;
    }
    public boolean skippedVersionNum() throws Exception {
        return skippedAsciiCharWithFlag(E_VersionNumFlag);
    }
    public boolean skippedEncName() throws Exception {
        return skippedAsciiCharWithFlag(E_EncNameFlag);
    }
    public boolean skippedPubidChar() throws Exception {
        int ch = fMostRecentChar;
        if (ch >= 0x80)
            return false;
        if ((fgCharFlags[ch] & E_PubidCharFlag) != 0) {
            fCharacterCounter++;
            loadNextChar();
            return true;
        }
        if (ch == 0x0A) {
            fLinefeedCounter++;
            fCharacterCounter = 1;
            loadNextChar();
            return true;
        }
        return false;
    }
    public boolean skippedString(char[] s) throws Exception {
        int length = s.length;
        char[] data = fMostRecentData;
        int index = fCurrentIndex;
        if (index + length <= CHUNK_SIZE) {
            for (int i = 0; i < length; i++) {
                if (data[index++] != s[i])
                    return false;
            }
            fCharacterCounter += length;
            fCurrentOffset += length;
            fCurrentIndex = index;
            if (index == CHUNK_SIZE)
                slowLoadNextChar();
            else
                fMostRecentChar = data[index] & 0xFFFF;
            return true;
        }
        CharDataChunk dataChunk = fCurrentChunk;
        int offset = fCurrentOffset;
        int savedIndex = index;
        int i = 0;
        while (index < CHUNK_SIZE) {
            if (data[index++] != s[i++])
                return false;
        }
        slowLoadNextChar();
        data = fMostRecentData;
        index = 0;
        while (i < length) {
            if (data[index++] != s[i++]) {
                fCurrentChunk = dataChunk;
                fCurrentIndex = savedIndex;
                fCurrentOffset = offset;
                fMostRecentData = dataChunk.toCharArray();
                fMostRecentChar = fMostRecentData[savedIndex] & 0xFFFF;
                return false;
            }
        }
        fCharacterCounter += length;
        fCurrentOffset += length;
        fCurrentIndex = index;
        if (index == CHUNK_SIZE)
            slowLoadNextChar();
        else
            fMostRecentChar = data[index] & 0xFFFF;
        return true;
    }
    public int scanName(char fastcheck, int expectedName) throws Exception {
        int ch = fMostRecentChar;
        if (ch < 0x80) {
            if (fgAsciiInitialNameChar[ch] == 0)
                return -1;
        } else {
            if ((fgCharFlags[ch] & E_InitialNameCharFlag) == 0)
                return -1;
        }
        int offset = fCurrentOffset;
        int index = fCurrentIndex;
        char[] data = fMostRecentData;
        if (++index == CHUNK_SIZE) {
            fCurrentChunk = new CharDataChunk(fStringPool, fCurrentChunk);
            fillCurrentChunk();
            index = 0;
            data = fMostRecentData;
        }
        fCharacterCounter++;
        fCurrentOffset++;
        int hashcode = 0;
        int position = 0;
        while (true) {
            hashcode = StringHasher.hashChar(hashcode, position++, ch);
            ch = data[index] & 0xFFFF;
            if (fastcheck == ch)
                break;
            if (ch < 0x80) {
                if (fgAsciiNameChar[ch] == 0)
                    break;
            } else {
                if ((fgCharFlags[ch] & E_NameCharFlag) == 0)
                    break;
            }
            if (++index == CHUNK_SIZE) {
                fCurrentChunk = new CharDataChunk(fStringPool, fCurrentChunk);
                fillCurrentChunk();
                index = 0;
                data = fMostRecentData;
            }
            fCharacterCounter++;
            fCurrentOffset++;
        }
        fCurrentIndex = index;
        fMostRecentChar = ch;
        hashcode = StringHasher.finishHash(hashcode);
        int length = fCurrentOffset - offset;
        int nameIndex = addSymbol(offset, length, hashcode);
        return (expectedName == -1 || expectedName == nameIndex) ? nameIndex : -1;
    }
    //
    //
    //
    public int skipPastName(char fastcheck) throws Exception {
        int ch = fMostRecentChar;
        if (ch < 0x80) {
            if (fgAsciiInitialNameChar[ch] == 0)
                return fCurrentOffset;
        } else {
            if ((fgCharFlags[ch] & E_InitialNameCharFlag) == 0)
                return fCurrentOffset;
        }
        while (true) {
            fCharacterCounter++;
            ch = loadNextChar();
            if (fastcheck == ch)
                return fCurrentOffset;
            if (ch < 0x80) {
                if (fgAsciiNameChar[ch] == 0)
                    return fCurrentOffset;
            } else {
                if ((fgCharFlags[ch] & E_NameCharFlag) == 0)
                    return fCurrentOffset;
            }
        }
    }
    public int skipPastNmtoken(char fastcheck) throws Exception {
        int ch = fMostRecentChar;
        while (true) {
            if (fastcheck == ch)
                return fCurrentOffset;
            if (ch < 0x80) {
                if (fgAsciiNameChar[ch] == 0)
                    return fCurrentOffset;
            } else {
                if ((fgCharFlags[ch] & E_NameCharFlag) == 0)
                    return fCurrentOffset;
            }
            fCharacterCounter++;
            ch = loadNextChar();
        }
    }
    //
    //
    //
    public int scanContent(ScanContentState scanState) throws Exception {
        fCurrentChunk.clearPreviousChunk();
        int charDataOffset = fCurrentOffset;
        int ch = fMostRecentChar;
        if (ch < 0x80) {
            switch (fgAsciiWSCharData[ch]) {
            case 0:
                fCharacterCounter++;
                fCurrentOffset++;
                if (++fCurrentIndex == CHUNK_SIZE)
                    slowLoadNextChar();
                else
                    fMostRecentChar = fMostRecentData[fCurrentIndex] & 0xFFFF;
                break;
            case 1: // '<'
                fCharacterCounter++;
                fCurrentOffset++;
                if (++fCurrentIndex == CHUNK_SIZE)
                    slowLoadNextChar();
                else
                    fMostRecentChar = fMostRecentData[fCurrentIndex] & 0xFFFF;
                if (!scanState.inCDSect)
                    return CHARDATA_RESULT_MARKUP;
                break;
            case 2: // '&'
                fCharacterCounter++;
                fCurrentOffset++;
                if (++fCurrentIndex == CHUNK_SIZE)
                    slowLoadNextChar();
                else
                    fMostRecentChar = fMostRecentData[fCurrentIndex] & 0xFFFF;
                if (!scanState.inCDSect)
                    return CHARDATA_RESULT_REFERENCE;
                break;
            case 3: // ']'
                fCharacterCounter++;
                ch = loadNextChar();
                if (ch != ']')
                    break;
                if (fCurrentIndex + 1 == CHUNK_SIZE) {
                    CharDataChunk dataChunk = fCurrentChunk;
                    int index = fCurrentIndex;
                    int offset = fCurrentOffset;
                    if (loadNextChar() != '>') {
                        fCurrentChunk = dataChunk;
                        fCurrentIndex = index;
                        fCurrentOffset = offset;
                        fMostRecentData = dataChunk.toCharArray();
                        fMostRecentChar = ']';
                        break;
                    }
                } else {
                    if (fMostRecentData[fCurrentIndex + 1] != '>')
                        break;
                    fCurrentIndex++;
                    fCurrentOffset++;
                }
                loadNextChar();
                fCharacterCounter += 2;
                if (scanState.inCDSect) {
                    scanState.inCDSect = false;
                    return scanContent(scanState);
                }
                return CHARDATA_RESULT_CDEND;
            case 4: // invalid char
                return CHARDATA_RESULT_INVALID_CHAR;
            case 5:
                do {
                    if (ch == 0x0A) {
                        fLinefeedCounter++;
                        fCharacterCounter = 1;
                    } else
                        fCharacterCounter++;
                    fCurrentOffset++;
                    if (++fCurrentIndex == CHUNK_SIZE)
                        ch = slowLoadNextChar();
                    else
                        ch = fMostRecentChar = fMostRecentData[fCurrentIndex] & 0xFFFF;
                } while (ch == 0x20 || ch == 0x09 || ch == 0x0A);
                if (ch < 0x80) {
                    switch (fgAsciiCharData[ch]) {
                    case 0:
                        fCharacterCounter++;
                        fCurrentOffset++;
                        if (++fCurrentIndex == CHUNK_SIZE)
                            slowLoadNextChar();
                        else
                            fMostRecentChar = fMostRecentData[fCurrentIndex] & 0xFFFF;
                        break;
                    case 1: // '<'
                        if (!scanState.inCDSect) {
                            if (fDocumentHandler != null)
                                callWSCharDataHandler(charDataOffset, fCurrentOffset, false);
                            fCharacterCounter++;
                            loadNextChar();
                            return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_MARKUP);
                        }
                        fCharacterCounter++;
                        fCurrentOffset++;
                        if (++fCurrentIndex == CHUNK_SIZE)
                            slowLoadNextChar();
                        else
                            fMostRecentChar = fMostRecentData[fCurrentIndex] & 0xFFFF;
                        break;
                    case 2: // '&'
                        if (!scanState.inCDSect) {
                            if (fDocumentHandler != null)
                                callWSCharDataHandler(charDataOffset, fCurrentOffset, false);
                            fCharacterCounter++;
                            loadNextChar();
                            return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_REFERENCE);
                        }
                        fCharacterCounter++;
                        loadNextChar();
                        break;
                    case 3: // ']'
                        int endOffset = fCurrentOffset;
                        ch = loadNextChar();
                        if (ch != ']') {
                            fCharacterCounter++;
                            break;
                        }
                        if (fCurrentIndex + 1 == CHUNK_SIZE) {
                            CharDataChunk dataChunk = fCurrentChunk;
                            int index = fCurrentIndex;
                            int offset = fCurrentOffset;
                            if (loadNextChar() != '>') {
                                fCurrentChunk = dataChunk;
                                fCurrentIndex = index;
                                fCurrentOffset = offset;
                                fMostRecentData = dataChunk.toCharArray();
                                fMostRecentChar = ']';
                                fCharacterCounter++;
                                break;
                            }
                        } else {
                            if (fMostRecentData[fCurrentIndex + 1] != '>') {
                                fCharacterCounter++;
                                break;
                            }
                            fCurrentIndex++;
                            fCurrentOffset++;
                        }
                        loadNextChar();
                        if (fDocumentHandler != null)
                            callWSCharDataHandler(charDataOffset, endOffset, scanState.inCDSect);
                        fCharacterCounter += 3;
                        if (scanState.inCDSect) {
                            scanState.inCDSect = false;
                            return scanContent(scanState);
                        }
                        return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_CDEND);
                    case 4: // invalid char
                        if (fDocumentHandler != null)
                            callWSCharDataHandler(charDataOffset, fCurrentOffset, scanState.inCDSect);
                        return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                    }
                } else if (!skipMultiByteCharData(ch)) {
                    if (fDocumentHandler != null)
                        callWSCharDataHandler(charDataOffset, fCurrentOffset, scanState.inCDSect);
                    return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                }
                break;
            }
        } else if (!skipMultiByteCharData(ch)) {
            return CHARDATA_RESULT_INVALID_CHAR;
        }
        ch = skipAsciiCharData();
        while (true) { // REVISIT - EOF check ?
            if (ch < 0x80) {
                switch (fgAsciiCharData[ch]) {
                case 0:
                    fCharacterCounter++;
                    ch = loadNextChar();
                    break;
                case 1: // '<'
                    if (!scanState.inCDSect) {
                        if (fDocumentHandler != null)
                            callCharDataHandler(charDataOffset, fCurrentOffset, false);
                        fCharacterCounter++;
                        fCurrentOffset++;
                        if (++fCurrentIndex == CHUNK_SIZE)
                            slowLoadNextChar();
                        else
                            fMostRecentChar = fMostRecentData[fCurrentIndex] & 0xFFFF;
                        return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_MARKUP);
                    }
                    fCharacterCounter++;
                    ch = loadNextChar();
                    break;
                case 2: // '&'
                    if (!scanState.inCDSect) {
                        if (fDocumentHandler != null)
                            callCharDataHandler(charDataOffset, fCurrentOffset, false);
                        fCharacterCounter++;
                        loadNextChar();
                        return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_REFERENCE);
                    }
                    fCharacterCounter++;
                    ch = loadNextChar();
                    break;
                case 3: // ']'
                    int endOffset = fCurrentOffset;
                    ch = loadNextChar();
                    if (ch != ']') {
                        fCharacterCounter++;
                        break;
                    }
                    if (fCurrentIndex + 1 == CHUNK_SIZE) {
                        CharDataChunk dataChunk = fCurrentChunk;
                        int index = fCurrentIndex;
                        int offset = fCurrentOffset;
                        if (loadNextChar() != '>') {
                            fCurrentChunk = dataChunk;
                            fCurrentIndex = index;
                            fCurrentOffset = offset;
                            fMostRecentData = dataChunk.toCharArray();
                            fMostRecentChar = ']';
                            fCharacterCounter++;
                            break;
                        }
                    } else {
                        if (fMostRecentData[fCurrentIndex + 1] != '>') {
                            fCharacterCounter++;
                            break;
                        }
                        fCurrentIndex++;
                        fCurrentOffset++;
                    }
                    loadNextChar();
                    if (fDocumentHandler != null)
                        callCharDataHandler(charDataOffset, endOffset, scanState.inCDSect);
                    fCharacterCounter += 3;
                    if (scanState.inCDSect) {
                        scanState.inCDSect = false;
                        return scanContent(scanState);
                    }
                    return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_CDEND);
                case 4: // invalid char
                    if (ch == 0x0A) {
                        fLinefeedCounter++;
                        fCharacterCounter = 1;
                        ch = loadNextChar();
                        break;
                    }
                    if (fDocumentHandler != null)
                        callCharDataHandler(charDataOffset, fCurrentOffset, scanState.inCDSect);
                    return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                }
            } else {
                if (!skipMultiByteCharData(ch)) {
                    if (fDocumentHandler != null)
                        callCharDataHandler(charDataOffset, fCurrentOffset, scanState.inCDSect);
                    return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                }
                ch = fMostRecentChar;
            }
        }
    }
    protected boolean skipMultiByteCharData(int ch) throws Exception {
        if (ch > 0xFFFD)
            return false;
        if (ch >= 0xDC000 && ch < 0xE000)
            return false;
        if (ch >= 0xD8000 && ch < 0xDC00) {
            ch = loadNextChar();
            if (ch < 0xDC000 || ch >= 0xE000)
                return false;
        }
        loadNextChar();
        return true;
    }
    protected int skipAsciiCharData() throws Exception {
        int index = fCurrentIndex;
        int offset = fCurrentOffset - index;
        while (true) {
            char[] data = fMostRecentData;
            while (index < CHUNK_SIZE) {
                int ch = data[index] & 0xFFFF;
                if (ch >= 0x80) {
                    fCurrentOffset = offset + index;
                    fCurrentIndex = index;
                    fMostRecentChar = ch;
                    return ch;
                }
                if (fgAsciiCharData[ch] != 0) {
                    if (ch != 0x0A) {
                        fCurrentOffset = offset + index;
                        fCurrentIndex = index;
                        fMostRecentChar = ch;
                        return ch;
                    }
                    fLinefeedCounter++;
                    fCharacterCounter = 1;
                } else
                    fCharacterCounter++;
                index++;
            }
            offset += index;
            fCurrentChunk = new CharDataChunk(fStringPool, fCurrentChunk);
            fillCurrentChunk();
            index = 0;
        }
    }
    public void callCharDataHandler(int offset, int endOffset, boolean inCDSect) throws Exception {
        //
        // REVISIT - need more up front bounds checking code of params...
        //
        int length = endOffset - offset;
        if (!fDocumentHandler.sendCharDataAsCharArray()) {
            int stringIndex = addString(offset, length);
            fDocumentHandler.characters(stringIndex, inCDSect);
            return;
        }
        //
        //
        //
        CharDataChunk dataChunk = fCurrentChunk.chunkFor(offset);
        int index = offset & CHUNK_MASK;
        if (index + length <= CHUNK_SIZE) {
            //
            // All the chars are in the same chunk
            //
            if (length != 0)
                fDocumentHandler.characters(dataChunk.toCharArray(), index, length, inCDSect);
            return;
        }
        //
        // The data is spread across chunks.
        //
        int count = length;
        int nbytes = CHUNK_SIZE - index;
        fDocumentHandler.characters(dataChunk.toCharArray(), index, nbytes, inCDSect);
        count -= nbytes;
        //
        // Use each Chunk in turn until we are done.
        //
        do {
            dataChunk = dataChunk.nextChunk();
            if (dataChunk == null)
				fErrorHandler.error(ErrorCode.E_INT_DCN);
            nbytes = count <= CHUNK_SIZE ? count : CHUNK_SIZE;
            fDocumentHandler.characters(dataChunk.toCharArray(), 0, nbytes, inCDSect);
            count -= nbytes;
        } while (count > 0);
    }
    public void callWSCharDataHandler(int offset, int endOffset, boolean inCDSect) throws Exception {
        //
        // If the contentSpecType is CONTENTSPEC_CHILDREN we will call the
        // ignorableWhitespace() method otherwise we will call characters().
        //
        int contentSpecType = fScanner.getCurrentContentSpecType();
        if (contentSpecType != ElementDecl.CONTENTSPEC_CHILDREN) {
            callCharDataHandler(offset, endOffset, inCDSect);
            return;
        }
        //
        // REVISIT - need more up front bounds checking code of params...
        //
        int length = endOffset - offset;
        if (!fDocumentHandler.sendCharDataAsCharArray()) {
            int stringIndex = addString(offset, length);
            fDocumentHandler.ignorableWhitespace(stringIndex, inCDSect);
            return;
        }
        //
        //
        //
        CharDataChunk dataChunk = fCurrentChunk.chunkFor(offset);
        int index = offset & CHUNK_MASK;
        if (index + length <= CHUNK_SIZE) {
            //
            // All the chars are in the same chunk
            //
            if (length != 0)
                fDocumentHandler.ignorableWhitespace(dataChunk.toCharArray(), index, length, inCDSect);
            return;
        }
        //
        // The data is spread across chunks.
        //
        int count = length;
        int nbytes = CHUNK_SIZE - index;
        fDocumentHandler.ignorableWhitespace(dataChunk.toCharArray(), index, nbytes, inCDSect);
        count -= nbytes;
        //
        // Use each Chunk in turn until we are done.
        //
        do {
            dataChunk = dataChunk.nextChunk();
            nbytes = count <= CHUNK_SIZE ? count : CHUNK_SIZE;
            fDocumentHandler.ignorableWhitespace(dataChunk.toCharArray(), 0, nbytes, inCDSect);
            count -= nbytes;
        } while (count > 0);
    }
    //
    //
    //
    protected XMLScanner fScanner = null;
    protected XMLDocumentHandler fDocumentHandler = null;
    protected XMLErrorHandler fErrorHandler = null;
    protected StringPool fStringPool = null;
    protected Reader fReader = null;
    protected CharDataChunk fCurrentChunk = null;
    protected int fCurrentIndex = 0;
    protected char[] fMostRecentData = null;
    protected int fMostRecentChar = 0;
    protected int fLength = 0;
    //
    // When we fill a chunk there may be data that was read from the
    // input stream that has not been "processed".  We need to save
    // that data, and any in-progress state, between the calls to
    // UTF8CharDataChunk.fill() in these instance variables.  
    //
    protected boolean fCheckOverflow = false;
    protected char[] fOverflow = null;
    protected int fOverflowOffset = 0;
    protected int fOverflowEnd = 0;
    protected int fOutputOffset = 0;
    protected boolean fSkipLinefeed = false;
    //
    //
    //
    protected int fillCurrentChunk() throws Exception {
        //
        // If we have overflow from the last call, normalize from where
        // we left off, copying into the front of the output buffer.
        //
        fOutputOffset = 0;
        if (fCheckOverflow) {
            //
            // The fOverflowEnd should always be equal to CHUNK_SIZE, unless we hit
            // EOF during the previous call.  Copy the remaining data to the front
            // of the buffer and return it as the final chunk.
            //
            if (fOverflowEnd < CHUNK_SIZE) {
                if (fOverflowEnd > 0) {
                    fMostRecentData = new char[1 + fOverflowEnd - fOverflowOffset];
                    copyNormalize(fOverflow, fOverflowOffset, fMostRecentData, fOutputOffset);
                } else {
                    fMostRecentData = new char[1];
                }
                fMostRecentData[fOutputOffset] = 0;
                //
                // Update the outer class instance variables
                //
                fOverflow = null;
                fLength += fOutputOffset;
                fCurrentIndex = 0;
                fCurrentChunk.setCharArray(fMostRecentData);
                return (fMostRecentChar = fMostRecentData[0]);
            }
            fMostRecentData = new char[CHUNK_SIZE];
            copyNormalize(fOverflow, fOverflowOffset, fMostRecentData, fOutputOffset);
            fCheckOverflow = false;
        } else {
            if (fOverflow == null)
                fOverflow = new char[CHUNK_SIZE];
            fMostRecentData = null;
        }
        while (true) {
            fOverflowOffset = 0;
            fOverflowEnd = 0;
            int capacity = CHUNK_SIZE;
            int result = 0;
            do {
                result = fReader.read(fOverflow, fOverflowEnd, capacity);
                if (result == -1) {
                    //
                    // We have reached the end of the stream.
                    //
                    fReader.close();
                    fReader = null;
                    if (fMostRecentData == null) {
                        //
                        // There is no previous output data, so we know that all of the
                        // new input data will fit.
                        //
                        fMostRecentData = new char[1 + fOverflowEnd];
                        copyNormalize(fOverflow, fOverflowOffset, fMostRecentData, fOutputOffset);
                        fOverflow = null;
                        fMostRecentData[fOutputOffset] = 0;
                    } else {
                        //
                        // Copy the input data to the end of the output buffer.
                        //
                        boolean alldone = copyNormalize(fOverflow, fOverflowOffset, fMostRecentData, fOutputOffset);
                        if (alldone) {
                            if (fOverflowEnd == CHUNK_SIZE) {
                                //
                                // Special case - everything fit into the overflow buffer,
                                // except that there is no room for the nul char we use to
                                // indicate EOF.  Set the overflow buffer length to zero.
                                // On the next call to this method, we will detect this
                                // case and which we will handle above .
                                //
                                fCheckOverflow = true;
                                fOverflowOffset = 0;
                                fOverflowEnd = 0;
                            } else {
                                //
                                // It all fit into the output buffer.
                                //
                                fOverflow = null;
                                fMostRecentData[fOutputOffset] = 0;
                            }
                        } else {
                            //
                            // There is still input data left over, save the remaining data as
                            // the overflow buffer for the next call.
                            //
                            fCheckOverflow = true;
                        }
                    }
                    break;
                }
                if (result > 0) {
                    fOverflowEnd += result;
                    capacity -= result;
                }
            } while (capacity > 0);
            //
            //
            //
            if (result == -1)
                break;
            if (fMostRecentData != null) {
                boolean alldone = copyNormalize(fOverflow, fOverflowOffset, fMostRecentData, fOutputOffset);
                if (fOutputOffset == CHUNK_SIZE) {
                    //
                    // We filled the output buffer.
                    //
                    if (!alldone) {
                        //
                        // The input buffer will become the next overflow buffer.
                        //
                        fCheckOverflow = true;
                    }
                    break;
                }
            } else {
                //
                // Now normalize the end-of-line characters and see if we need to read more
                // chars to fill up the buffer.
                //
                fMostRecentData = new char[CHUNK_SIZE];
                copyNormalize(fOverflow, fOverflowOffset, fMostRecentData, fOutputOffset);
                if (fOutputOffset == CHUNK_SIZE) {
                    //
                    // The output buffer is full.  We can return now.
                    //
                    break;
                }
            }
            //
            // We will need to get another intput buffer to be able to fill the
            // overflow buffer completely.
            //
        }
        //
        // Update the outer class instance variables
        //
        fLength += fOutputOffset;
        fCurrentIndex = 0;
        fCurrentChunk.setCharArray(fMostRecentData);
        return (fMostRecentChar = fMostRecentData[0]);
    }
    //
    // Copy and normalize chars from the overflow buffer into chars in our data buffer.
    //
    protected boolean copyNormalize(char[] in, int inOffset, char[] out, int outOffset) throws Exception {
        //
        // Handle all edge cases before dropping into the inner loop.
        //
        int inEnd = fOverflowEnd;
        int outEnd = out.length;
        if (inOffset == inEnd)
            return true;
        char b = in[inOffset];
        if (fSkipLinefeed) {
            fSkipLinefeed = false;
            if (b == 0x0A) {
                if (++inOffset == inEnd)
                    return exitNormalize(inOffset, outOffset, true);
                b = in[inOffset];
            }
        }
        while (outOffset < outEnd) {
            //
            // Find the longest run that we can guarantee will not exceed the
            // bounds of the outer loop.
            //
            int inCount = inEnd - inOffset;
            int outCount = outEnd - outOffset;
            if (inCount > outCount)
                inCount = outCount;
            inOffset++;
            while (true) {
                while (b == 0x0D) {
                    out[outOffset++] = 0x0A;
                    if (inOffset == inEnd) {
                        fSkipLinefeed = true;
                        return exitNormalize(inOffset, outOffset, true);
                    }
                    b = in[inOffset];
                    if (b == 0x0A) {
                        if (++inOffset == inEnd)
                            return exitNormalize(inOffset, outOffset, true);
                        b = in[inOffset];
                    }
                    if (outOffset == outEnd)
                        return exitNormalize(inOffset, outOffset, false);
                    inCount = inEnd - inOffset;
                    outCount = outEnd - outOffset;
                    if (inCount > outCount)
                        inCount = outCount;
                    inOffset++;
                }
                while (true) {
                    out[outOffset++] = b;
                    if (--inCount == 0)
                        break;
                    b = in[inOffset++];
                    if (b == 0x0D)
                        break;
                }
                if (inCount == 0)
                    break;
            }
            if (inOffset == inEnd)
                break;
        }
        return exitNormalize(inOffset, outOffset, inOffset == inEnd);
    }
    //
    //
    //
    protected boolean exitNormalize(int inOffset, int outOffset, boolean result) {
        fOverflowOffset = inOffset;
        fOutputOffset = outOffset;
        return result;
    }
    //
    //
    //
    protected final class CharDataChunk implements StringProducer {
        //
        //
        //
        public CharDataChunk(StringPool stringPool, CharDataChunk prev) throws Exception {
            fInnerStringPool = stringPool;
            fChunk = prev == null ? 0 : prev.fChunk + 1;
            fPreviousChunk = prev;
            if (prev != null)
                prev.fNextChunk = this;
        }
        //
        // Constructor for creating clumps.
        //
        protected CharDataChunk(CharDataChunk proto) {
            fInnerStringPool = proto.fInnerStringPool;
            fChunk = proto.fChunk;
            fData = proto.fData;
            fPreviousChunk = null;
            fNextChunk = null;
        }
        //
        //
        //
        public CharDataChunk chunkFor(int offset) {
            int firstChunk = offset >> CHUNK_SHIFT;
            CharDataChunk dataChunk = this;
            while (firstChunk != dataChunk.fChunk)
                dataChunk = dataChunk.fPreviousChunk;
            return dataChunk;
        }
        //
        //
        //
        public char[] toCharArray() {
            return fData;
        }
        //
        //
        //
        public void setCharArray(char[] data) {
            fData = data;
        }
        //
        //
        //
        public CharDataChunk nextChunk() {
            return fNextChunk;
        }
        //
        //
        //
        public void clearPreviousChunk() {
            if (fPreviousChunk != null) {
//                System.out.println("clearPreviousChunk: clear chunk " + fPreviousChunk.fChunk);
                fPreviousChunk.fNextChunk = null;
                fPreviousChunk = null;
            }
        }
        /**
         *
         */
        public String toString(int offset, int length) {
            if (offset + length <= CHUNK_SIZE) {
                //
                // All the chars are in the same chunk
                //
                return new String(fData, offset, length);
            }
            //
            // The data is spread across chunks, so we need to build it in pieces.
            //
            StringBuffer sb = new StringBuffer(length);
            //
            // Copy the partial data from the first chunk.
            //
            int nbytes = CHUNK_SIZE - offset;
            sb.append(fData, offset, nbytes);
            length -= nbytes;
            //
            // Use each chunk in turn until we are done.
            //
            CharDataChunk aChunk = fNextChunk;
            do {
                nbytes = length <= CHUNK_SIZE ? length : CHUNK_SIZE;
                sb.append(aChunk.fData, 0, nbytes);
                length -= nbytes;
                aChunk = aChunk.fNextChunk;
            } while (length > 0);
            String retval = sb.toString();
            sb = null; // REVISIT - does this help gc ?
            return retval;
        }
        /**
         *
         */
        public boolean equalsString(int offset, int length, String str, int strLength) {
            if (length != strLength)
                return false;
            if (offset + length <= CHUNK_SIZE) {
                //
                // All the chars are in the same chunk
                //
                for (int i = 0; i < length; i++) {
                    if (fData[offset++] != str.charAt(i))
                        return false;
                }
                return true;
            }
            //
            // Compare the partial data from the first chunk.
            //
            int nbytes = CHUNK_SIZE - offset;
            int index = 0;
            while (index < nbytes) {
                if (fData[offset++] != str.charAt(index++))
                    return false;
            }
            length -= nbytes;
            //
            // Check each chunk in turn until we are done.
            //
            CharDataChunk aChunk = fNextChunk;
            do {
                offset = 0;
                nbytes = length <= CHUNK_SIZE ? length : CHUNK_SIZE;
                length -= nbytes;
                while (nbytes-- > 0) {
                    if (aChunk.fData[offset++] != str.charAt(index++))
                        return false;
                }
                aChunk = aChunk.fNextChunk;
            } while (length > 0);
            return true;
        }
        //
        //
        //
        protected CharDataChunk createClump(int lastChunk) {
            CharDataChunk result = new CharDataChunk(this);
            CharDataChunk nextChunk = fNextChunk;
            CharDataChunk aChunk = result;
            do {
                CharDataChunk next = new CharDataChunk(nextChunk);
                aChunk.fNextChunk = next;
                nextChunk = nextChunk.fNextChunk;
                aChunk = next;
            } while (aChunk.fChunk != lastChunk);
            return result;
        }
        //
        //
        //
        public int addString(int offset, int length) {
            int chunk = offset >> CHUNK_SHIFT;
            if (chunk != fChunk) {
                if (fPreviousChunk != null)
                    return fPreviousChunk.addString(offset, length);
				try {
					fErrorHandler.error(ErrorCode.E_INT_PCN);
				} catch (Exception e) {
					// revisit: don't fail silently
				}
                return -1;
            }
            int lastChunk = (offset + length - 1) >> CHUNK_SHIFT;
            if (chunk == lastChunk)
                return fInnerStringPool.addString(this, offset & CHUNK_MASK, length);
            return fInnerStringPool.addString(createClump(lastChunk), offset & CHUNK_MASK, length);
        }
        //
        //
        //
        public int addSymbol(int offset, int length, int hashcode) {
            int chunk = offset >> CHUNK_SHIFT;
            if (chunk != fChunk) {
                if (fPreviousChunk != null)
                    return fPreviousChunk.addSymbol(offset, length, hashcode);
				try {
					fErrorHandler.error(ErrorCode.E_INT_PCN);
				} catch (Exception e) {
					// revisit: don't fail silently
				}
                return -1;
            }
            int lastChunk = (offset + length - 1) >> CHUNK_SHIFT;
            int index = offset & CHUNK_MASK;
            if (chunk == lastChunk) {
                if (hashcode == 0) {
                    for (int i = 0; i < length; i++)
                        hashcode = StringHasher.hashChar(hashcode, i, fData[index++] & 0xFFFF);
                    hashcode = StringHasher.finishHash(hashcode);
                }
                return fInnerStringPool.addSymbol(this, offset & CHUNK_MASK, length, hashcode);
            }
            if (hashcode == 0) {
                int position = 0;
                int hlength = length;
                int nbytes = CHUNK_SIZE - index;
                while (index < CHUNK_SIZE)
                    hashcode = StringHasher.hashChar(hashcode, position++, fData[index++] & 0xFFFF);
                hlength -= nbytes;
                //
                // Check each chunk in turn until we are done.
                //
                CharDataChunk aChunk = fNextChunk;
                do {
                    index = 0;
                    nbytes = hlength <= CHUNK_SIZE ? hlength : CHUNK_SIZE;
                    while (index < nbytes)
                        hashcode = StringHasher.hashChar(hashcode, position++, aChunk.fData[index++] & 0xFFFF);
                    hlength -= nbytes;
                    aChunk = aChunk.fNextChunk;
                } while (hlength > 0);
                hashcode = StringHasher.finishHash(hashcode);
            }
            return fInnerStringPool.addSymbol(createClump(lastChunk), offset & CHUNK_MASK, length, hashcode);
        }
        /**
         *
         */
        public void append(ChunkyCharArray charArray, int offset, int length) {
            //
            // Setup for the operation.
            //
            CharDataChunk dataChunk = chunkFor(offset);
            int index = offset & CHUNK_MASK;
            int nbytes = (index + length <= CHUNK_SIZE) ? length : CHUNK_SIZE - index;
            //
            // Visit each Chunk in turn until we are done.
            //
            while (true) {
                charArray.append(dataChunk.fData, index, nbytes);
                length -= nbytes;
                if (length == 0)
                    break;
                dataChunk = dataChunk.fNextChunk;
                index = 0;
                nbytes = length <= CHUNK_SIZE ? length : CHUNK_SIZE;
            }
        }
        //
        //
        //
        protected StringPool fInnerStringPool = null;
        protected int fChunk;
        protected char[] fData = null;
        protected CharDataChunk fPreviousChunk = null;
        protected CharDataChunk fNextChunk = null;
    }
}
