/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */
package com.ibm.xml.internal;

import com.ibm.xml.framework.*;
import org.xml.sax.AttributeList;

/**
 *
 * @version Revision: 79 1.6 src/com/ibm/xml/internal/DefaultAttrPool.java, parser, xml4j2, xml4j2_0_13 
 */
public final class DefaultAttrPool implements AttrPool, AttributeList {
    //
    // Chunk size constants
    //
    static final int CHUNK_SHIFT = 5;           // 2^5 = 32
    static final int CHUNK_SIZE = (1 << CHUNK_SHIFT);
    static final int CHUNK_MASK = CHUNK_SIZE - 1;
    static final int INITIAL_CHUNK_COUNT = (1 << (10 - CHUNK_SHIFT));   // 2^10 = 1k
    //
    // Extra flags added to AttDef.attType
    //
    static final int ATTRFLAG_SPECIFIED = 64;
    static final int ATTRFLAG_LASTATTR = 128;
    //
    // Instance variables
    //
    private StringPool fStringPool = null;
    private XMLErrorHandler fErrorHandler = null;
    private int fAttlistIndex = -1;
    private int fAttlistLength = 0;
    private int fAttrCount = 0;
    private int[][] fAttName = new int[INITIAL_CHUNK_COUNT][];
    private int[][] fAttValue = new int[INITIAL_CHUNK_COUNT][];
    private byte[][] fAttTypeAndFlags = new byte[INITIAL_CHUNK_COUNT][];
    //
    //
    //
    public DefaultAttrPool(ParserState parserState) {
        fStringPool = parserState.cacheStringPool();
        fErrorHandler = parserState.getErrorHandler();
    }
    //
    //
    //
    public void reset(ParserState parserState) {
        fStringPool = parserState.cacheStringPool();
        fErrorHandler = parserState.getErrorHandler();
        fAttlistIndex = -1;
        fAttlistLength = 0;
        fAttrCount = 0;
    }
    //
    //
    //
    public AttrPool resetOrCopy(ParserState parserState) {
        return new DefaultAttrPool(parserState);
    }
    //
    //
    //
    private boolean ensureCapacity(int chunk, int index) {
        try {
            return fAttName[chunk][index] != 0;
        } catch (ArrayIndexOutOfBoundsException ex) {
            int[][] newIntArray = new int[chunk * 2][];
            System.arraycopy(fAttName, 0, newIntArray, 0, chunk);
            fAttName = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fAttValue, 0, newIntArray, 0, chunk);
            fAttValue = newIntArray;
            byte[][] newByteArray = new byte[chunk * 2][];
            System.arraycopy(fAttTypeAndFlags, 0, newByteArray, 0, chunk);
            fAttTypeAndFlags = newByteArray;
        } catch (NullPointerException ex) {
        }
        fAttName[chunk] = new int[CHUNK_SIZE];
        fAttValue[chunk] = new int[CHUNK_SIZE];
        fAttTypeAndFlags[chunk] = new byte[CHUNK_SIZE];
        return true;
    }
    public int addAttr(Attr attr, int firstAttrIndex) throws Exception
    {
        int chunk;
        int index;
        if (firstAttrIndex != -1) {
            chunk = firstAttrIndex >> CHUNK_SHIFT;
            index = firstAttrIndex & CHUNK_MASK;
            while (firstAttrIndex < fAttrCount) {
                if (fAttName[chunk][index] == attr.attName) {
                    fErrorHandler.error1(ErrorCode.E_ATTEQ1, attr.attName);
                    return -1;
                }
                firstAttrIndex++;
                if (++index == CHUNK_SIZE) {
                    chunk++;
                    index = 0;
                }
            }
        } else {
            chunk = fAttrCount >> CHUNK_SHIFT;
            index = fAttrCount & CHUNK_MASK;
        }
        ensureCapacity(chunk, index);
        fAttName[chunk][index] = attr.attName;
        fAttValue[chunk][index] = attr.attValue;
        fAttTypeAndFlags[chunk][index] = (byte)((attr.specified ? ATTRFLAG_SPECIFIED : 0) | attr.attType);
        int attrIndex = fAttrCount++;
        return attrIndex;
    }
    public void setIsLastAttr(int attrIndex) {
        if (attrIndex < 0 || attrIndex >= fAttrCount)
            return;
        int chunk = attrIndex >> CHUNK_SHIFT;
        int index = attrIndex & CHUNK_MASK;
        fAttTypeAndFlags[chunk][index] |= ATTRFLAG_LASTATTR;
    }
    //
    //
    //
    public int getAttrName(int attrIndex) {
        if (attrIndex < 0 || attrIndex >= fAttrCount)
            return -1;
        int chunk = attrIndex >> CHUNK_SHIFT;
        int index = attrIndex & CHUNK_MASK;
        return fAttName[chunk][index];
    }
    public int getAttType(int attrIndex) {
        if (attrIndex < 0 || attrIndex >= fAttrCount)
            return -1;
        int chunk = attrIndex >> CHUNK_SHIFT;
        int index = attrIndex & CHUNK_MASK;
        return fAttTypeAndFlags[chunk][index] & (ATTRFLAG_SPECIFIED - 1);
    }
    public int getAttValue(int attrIndex) {
        if (attrIndex < 0 || attrIndex >= fAttrCount)
            return -1;
        int chunk = attrIndex >> CHUNK_SHIFT;
        int index = attrIndex & CHUNK_MASK;
        return fAttValue[chunk][index];
    }
    public boolean isSpecified(int attrIndex) {
        if (attrIndex < 0 || attrIndex >= fAttrCount)
            return true;
        int chunk = attrIndex >> CHUNK_SHIFT;
        int index = attrIndex & CHUNK_MASK;
        return (fAttTypeAndFlags[chunk][index] & ATTRFLAG_SPECIFIED) != 0;
    }
    public boolean isLastAttr(int attrIndex) {
        if (attrIndex < 0 || attrIndex >= fAttrCount)
            return true;
        int chunk = attrIndex >> CHUNK_SHIFT;
        int index = attrIndex & CHUNK_MASK;
        return (fAttTypeAndFlags[chunk][index] & ATTRFLAG_LASTATTR) != 0;
    }
    //
    //
    //
    public void releaseAttrList(int attrListIndex) {
        if (attrListIndex == -1)
            return;
        int chunk = attrListIndex >> CHUNK_SHIFT;
        int index = attrListIndex & CHUNK_MASK;
//System.out.println("release AttrList " + attrListIndex);
        while (true) {
            boolean last = (fAttTypeAndFlags[chunk][index] & ATTRFLAG_LASTATTR) != 0;
//System.out.println("release AttName " + fStringPool.toString(fAttName[chunk][index]) +
//" = AttValue " + fStringPool.toString(fAttValue[chunk][index]) +
//" [" + fAttValue[chunk][index] + "] type/flags " + fAttTypeAndFlags[chunk][index]);
            fAttName[chunk][index] = -1;
            if ((fAttTypeAndFlags[chunk][index] & ATTRFLAG_SPECIFIED) != 0)
                fStringPool.releaseString(fAttValue[chunk][index]);
            fAttValue[chunk][index] = -1;
            if (++index == CHUNK_SIZE) {
                chunk++;
                index = 0;
            }
            if (last)
                break;
        }
        if (fAttrCount == ((chunk << CHUNK_SHIFT) + index))
            fAttrCount = attrListIndex;
        else
			try {
				fErrorHandler.error(ErrorCode.E_INT_ATR);
			} catch (Exception e) {
				// revisit: don't fail silently
			}
    }
    //
    // AttributeList support
    //
    public AttributeList getAttributeList(int attrListIndex) {
        fAttlistIndex = attrListIndex;
        if (fAttlistIndex == -1)
            fAttlistLength = 0;
        else {
            int chunk = fAttlistIndex >> CHUNK_SHIFT;
            int index = fAttlistIndex & CHUNK_MASK;
            fAttlistLength = 1;
            while ((fAttTypeAndFlags[chunk][index] & ATTRFLAG_LASTATTR) == 0) {
                if (++index == CHUNK_SIZE) {
                    chunk++;
                    index = 0;
                }
                fAttlistLength++;
            }
        }
        return this;
    }

  /**
    * Return the number of attributes in this list.
    *
    * <p>The SAX parser may provide attributes in any
    * arbitrary order, regardless of the order in which they were
    * declared or specified.  The number of attributes may be
    * zero.</p>
    *
    * @return The number of attributes in the list.
    */
    public int getLength() {
        return fAttlistLength;
    }


  /**
    * Return the name of an attribute in this list (by position).
    *
    * <p>The names must be unique: the SAX parser shall not include the
    * same attribute twice.  Attributes without values (those declared
    * #IMPLIED without a value specified in the start tag) will be
    * omitted from the list.</p>
    *
    * <p>If the attribute name has a namespace prefix, the prefix
    * will still be attached.</p>
    *
    * @param i The index of the attribute in the list (starting at 0).
    * @return The name of the indexed attribute, or null
    *         if the index is out of range.
    * @see #getLength
    */
    public String getName(int i) {
        if (i < 0 || i >= fAttlistLength)
            return null;
        int chunk = (fAttlistIndex + i) >> CHUNK_SHIFT;
        int index = (fAttlistIndex + i) & CHUNK_MASK;
        return fStringPool.toString(fAttName[chunk][index]);
    }


  /**
    * Return the type of an attribute in the list (by position).
    *
    * <p>The attribute type is one of the strings "CDATA", "ID",
    * "IDREF", "IDREFS", "NMTOKEN", "NMTOKENS", "ENTITY", "ENTITIES",
    * or "NOTATION" (always in upper case).</p>
    *
    * <p>If the parser has not read a declaration for the attribute,
    * or if the parser does not report attribute types, then it must
    * return the value "CDATA" as stated in the XML 1.0 Recommentation
    * (clause 3.3.3, "Attribute-Value Normalization").</p>
    *
    * <p>For an enumerated attribute that is not a notation, the
    * parser will report the type as "NMTOKEN".</p>
    *
    * @param i The index of the attribute in the list (starting at 0).
    * @return The attribute type as a string, or
    *         null if the index is out of range.
    * @see #getLength
    * @see #getType(java.lang.String)
    */
    public String getType(int i) {
        if (i < 0 || i >= fAttlistLength)
            return null;
        int chunk = (fAttlistIndex + i) >> CHUNK_SHIFT;
        int index = (fAttlistIndex + i) & CHUNK_MASK;
        int attType = (fAttTypeAndFlags[chunk][index] & (ATTRFLAG_SPECIFIED - 1));
        if (attType == AttDef.ENUMERATION)
            attType = AttDef.NMTOKEN;
        return AttDef.getAttTypeString(attType);
    }


  /**
    * Return the value of an attribute in the list (by position).
    *
    * <p>If the attribute value is a list of tokens (IDREFS,
    * ENTITIES, or NMTOKENS), the tokens will be concatenated
    * into a single string separated by whitespace.</p>
    *
    * @param i The index of the attribute in the list (starting at 0).
    * @return The attribute value as a string, or
    *         null if the index is out of range.
    * @see #getLength
    * @see #getValue(java.lang.String)
    */
    public String getValue(int i) {
        if (i < 0 || i >= fAttlistLength)
            return null;
        int chunk = (fAttlistIndex + i) >> CHUNK_SHIFT;
        int index = (fAttlistIndex + i) & CHUNK_MASK;
        return fStringPool.toString(fAttValue[chunk][index]);
    }


  /**
    * Return the type of an attribute in the list (by name).
    *
    * <p>The return value is the same as the return value for
    * getType(int).</p>
    *
    * <p>If the attribute name has a namespace prefix in the document,
    * the application must include the prefix here.</p>
    *
    * @param name The name of the attribute.
    * @return The attribute type as a string, or null if no
    *         such attribute exists.
    * @see #getType(int)
    */
    public String getType(String name) {
        int nameIndex = fStringPool.addSymbol(name.intern());
        if (nameIndex == -1)
            return null;
        int chunk = fAttlistIndex >> CHUNK_SHIFT;
        int index = fAttlistIndex & CHUNK_MASK;
        for (int i = 0; i < fAttlistLength; i++) {
            if (fAttName[chunk][index] == nameIndex) {
                int attType = (fAttTypeAndFlags[chunk][index] & (ATTRFLAG_SPECIFIED - 1));
                if (attType == AttDef.ENUMERATION)
                    attType = AttDef.NMTOKEN;
                return AttDef.getAttTypeString(attType);
            }
            if (++index == CHUNK_SIZE) {
                chunk++;
                index = 0;
            }
        }
        return null;
    }


  /**
    * Return the value of an attribute in the list (by name).
    *
    * <p>The return value is the same as the return value for
    * getValue(int).</p>
    *
    * <p>If the attribute name has a namespace prefix in the document,
    * the application must include the prefix here.</p>
    *
    * @param i The index of the attribute in the list.
    * @return The attribute value as a string, or null if
    *         no such attribute exists.
    * @see #getValue(int)
    */
    public String getValue(String name) {
        int nameIndex = fStringPool.addSymbol(name.intern());
        if (nameIndex == -1)
            return null;
        int chunk = fAttlistIndex >> CHUNK_SHIFT;
        int index = fAttlistIndex & CHUNK_MASK;
        for (int i = 0; i < fAttlistLength; i++) {
            if (fAttName[chunk][index] == nameIndex)
                return fStringPool.toString(fAttValue[chunk][index]);
            if (++index == CHUNK_SIZE) {
                chunk++;
                index = 0;
            }
        }
        return null;
    }
    /** 
     * Returns the index of the first attribute in the specified
     * attribute list or -1 if the index is invalid.
     *
     * @param attrListIndex The attribute list index.
     */
    public int getFirstAttr(int attrListIndex) {
        if (attrListIndex < 0 || attrListIndex >= fAttrCount) {
            return -1;
            }
        // the first attribute in a list is implemented as
        // the same index of the attribute list index
        return attrListIndex;
        }

    /**
     * Returns the next index of the specified attribute or -1 if
     * there is no next index.
     *
     * @param attrIndex The attribute index.
     *
     * @see #isLastAttr
     */
    public int getNextAttr(int attrIndex) {
        if (isLastAttr(attrIndex)) {
            return -1;
            }
        // attribute lists are implemented in the
        // chunks one after another with the last
        // attribute having a "last" flag set
        return attrIndex + 1;
        }
}
