/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.internal;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;

import com.ibm.xml.framework.Catalog;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * This default catalog implementation does not provide a method
 * for loading multiple catalogs from various input sources.
 * Instead, it is a convenient base class for other catalog
 * implementations.
 * <p>
 * To create a catalog implementation, simply extend this class
 * and override the <tt>loadCatalog</tt> method. Public and system
 * identifier mappings can be stored and accessed using the 
 * convenient public methods on this class.
 *
 * @author Andy Clark, IBM
 * @version Revision: 66 1.3 src/com/ibm/xml/internal/DefaultCatalog.java, parser, xml4j2, xml4j2_0_13 
 */ 
public class DefaultCatalog
    implements Catalog
    {

    //
    // Data
    //

    /** Public identifier mappings. */
    private Hashtable publicMap = new Hashtable();

    /** System identifier mappings (aliases). */
    private Hashtable systemMap = new Hashtable();

    //
    // Public methods
    //

    /**
     * Adds a public to system identifier mapping.
     *
     * @param publicId The public identifier, or "key".
     * @param systemId The system identifier, or "value".
     */
    public void addPublicMapping(String publicId, String systemId) {
        publicMap.put(publicId, systemId);
        }

    /**
     * Removes a public identifier mapping.
     *
     * @param publicId The public identifier to remove.
     */
    public void removePublicMapping(System publicId) {
        publicMap.remove(publicId);
        }

    /** Returns an enumeration of public identifier mapping keys. */
    public Enumeration getPublicMappingKeys() {
        return publicMap.keys();
        }

    /**
     * Returns a public identifier mapping.
     *
     * @param publicId The public identifier, or "key".
     *
     * @return Returns the system identifier value or null if there
     *         is no mapping defined.
     */
    public String getPublicMapping(String publicId) {
        return (String)publicMap.get(publicId);
        }

    /**
     * Adds a system identifier alias.
     *
     * @param publicId The system identifier "key".
     * @param systemId The system identifier "value".
     */
    public void addSystemMapping(String systemId1, String systemId2) {
        systemMap.put(systemId1, systemId2);
        }

    /**
     * Removes a system identifier alias.
     *
     * @param systemId The system identifier to remove.
     */
    public void removeSystemMapping(String systemId) {
        systemMap.remove(systemId);
        }

    /** Returns an enumeration of system identifier mapping keys. */
    public Enumeration getSystemMappingKeys() {
        return systemMap.keys();
        }

    /**
     * Returns a system identifier alias.
     *
     * @param systemId The system identifier "key".
     *
     * @return Returns the system identifier alias value or null if there
     *         is no alias defined.
     */
    public String getSystemMapping(String systemId) {
        return (String)systemMap.get(systemId);
        }

    //
    // EntityResolver methods
    //

    /**
     * Resolves external entities.
     *
     * @param publicId The public identifier used for entity resolution.
     * @param systemId If the publicId is not null, this systemId is
     *                 to be considered the default system identifier;
     *                 else a system identifier alias mapping is
     *                 requested.
     *
     * @return Returns the input source of the resolved entity or null
     *         if no resolution is possible.
     *
     * @exception org.xml.sax.SAXException Exception thrown on SAX error.
     * @exception java.io.IOException Exception thrown on i/o error. 
     */
    public InputSource resolveEntity(String publicId, String systemId)
        throws SAXException, IOException 
        {
        
        // public id -> system id
        if (publicId != null) {
            String value = getPublicMapping(publicId);
            if (value != null) {
                return new InputSource(value);
                }
            }

        // system id(1) -> system id(2)
        if (systemId != null) {
            String value = getSystemMapping(systemId);
            if (value == null) {
                value = systemId;
                }

            return new InputSource(value);
            }

        return null;

        } // resolveEntity(String,String):InputSource

    //
    // Catalog methods
    //

    /** 
     * To provide catalog loading function, override this method in
     * a subclass.
     *
     * @param source The input source.
     *
     * @exception java.lang.Exception Exception thrown on error.
     */
    public void loadCatalog(InputSource source) 
        throws Exception 
        {
        throw new Exception("loadCatalog not implemented");
        }

    } // class DefaultCatalog
