/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */
package com.ibm.xml.internal;

import com.ibm.xml.framework.*;

import java.util.Enumeration;
import java.util.Hashtable;

//
// REVISIT - add code to share the leaf node entries...
//
/**
 *
 * @version Revision: 74 1.9 src/com/ibm/xml/internal/DefaultElementDeclPool.java, parser, xml4j2, xml4j2_0_13 
 */
public final class DefaultElementDeclPool implements ElementDeclPool {

    //
    // Chunk size constants
    //
    private static final int CHUNK_SHIFT = 5;           // 2^5 = 32
    private static final int CHUNK_SIZE = (1 << CHUNK_SHIFT);
    private static final int CHUNK_MASK = CHUNK_SIZE - 1;
    private static final int INITIAL_CHUNK_COUNT = (1 << (10 - CHUNK_SHIFT));   // 2^10 = 1k
    //
    // Instance variables
    //
    private ParserState fParserState = null;
    private StringPool fStringPool = null;
    private AttrPool fAttrPool = null;
    private XMLErrorHandler fErrorHandler = null;
    private XMLValidationHandler fValidationHandler;
    private int fRootElement = -1;
    private Attr fAttr = new Attr();
    //
    // Element list
    //
    private int fElementCount = 0;
    private int[][] fElementName = new int[INITIAL_CHUNK_COUNT][];
    private byte[][] fContentSpecType = new byte[INITIAL_CHUNK_COUNT][];
    private int[][] fContentSpec = new int[INITIAL_CHUNK_COUNT][];
    private ContentModel[][] fContentModel = new ContentModel[INITIAL_CHUNK_COUNT][];
    private int[][] fAttlistHead = new int[INITIAL_CHUNK_COUNT][];
    private int[][] fAttlistTail = new int[INITIAL_CHUNK_COUNT][];
    //
    // ContentSpecNode list
    //
    private int fNodeCount = 0;
    private byte[][] fNodeType = new byte[INITIAL_CHUNK_COUNT][];
    private int[][] fNodeValue = new int[INITIAL_CHUNK_COUNT][];
    //
    // AttDef list
    //
    private int fAttDefCount = 0;
    private int[][] fAttName = new int[INITIAL_CHUNK_COUNT][];
    private byte[][] fAttType = new byte[INITIAL_CHUNK_COUNT][];
    private int[][] fEnumeration = new int[INITIAL_CHUNK_COUNT][];
    private byte[][] fAttDefaultType = new byte[INITIAL_CHUNK_COUNT][];
    private int[][] fAttValue = new int[INITIAL_CHUNK_COUNT][];
    private int[][] fNextAttDef = new int[INITIAL_CHUNK_COUNT][];
    //
    // Element Name Hashtable
    //
    private static final int INITIAL_BUCKET_SIZE = 4;
    private static final int HASHTABLE_SIZE = 128;
    private int[][] fElementNameHashtable = new int[HASHTABLE_SIZE][];
    //
    //
    //
    private int fXMLSpace = -1;
    private int fDefault = -1;
    private int fPreserve = -1;
    private Hashtable fIdDefs = null;
    private Hashtable fIdRefs = null;
    private Object fNullValue = null;
    private int[] localpartcache = new int[8];
    private int[] nsnamecache = new int[8];
    //
    //
    //
    public DefaultElementDeclPool(ParserState parserState) {
        fParserState = parserState;
        fStringPool = parserState.cacheStringPool();
        fAttrPool = parserState.cacheAttrPool();
        fErrorHandler = parserState.getErrorHandler();
        fValidationHandler = parserState.getValidationHandler();
    }
    //
    //
    //
    public void reset(ParserState parserState) {
        fParserState = parserState;
        fStringPool = parserState.cacheStringPool();
        fAttrPool = parserState.cacheAttrPool();
        fErrorHandler = parserState.getErrorHandler();
        fValidationHandler = parserState.getValidationHandler();
        int chunk = 0;
        int index = 0;
        for (int i = 0; i < fElementCount; i++) {
            fContentModel[chunk][index] = null;
            if (++index == CHUNK_SIZE) {
                chunk++;
                index = 0;
            }
        }
        for (int i = 0; i < HASHTABLE_SIZE; i++)
            fElementNameHashtable[i] = null;
        fRootElement = -1;
        fElementCount = 0;
        fNodeCount = 0;
        fAttDefCount = 0;
        fXMLSpace = -1;
        fDefault = -1;
        fPreserve = -1;
        if (fIdDefs != null)
            fIdDefs.clear();
        if (fIdRefs != null)
            fIdRefs.clear();
    }
    //
    //
    //
    public ElementDeclPool resetOrCopy(ParserState parserState) {
        return new DefaultElementDeclPool(parserState);
    }
    //
    //
    //
    public void setRootElement(int elementIndex) {
        fRootElement = elementIndex;
    }
    public int getRootElement() {
        return fRootElement;
    }
    //
    // Element entries
    //
    private boolean ensureElementCapacity(int chunk) {
        try {
            return fElementName[chunk][0] == 0;
        } catch (ArrayIndexOutOfBoundsException ex) {
            byte[][] newByteArray = new byte[chunk * 2][];
            System.arraycopy(fContentSpecType, 0, newByteArray, 0, chunk);
            fContentSpecType = newByteArray;
            int[][] newIntArray = new int[chunk * 2][];
            System.arraycopy(fElementName, 0, newIntArray, 0, chunk);
            fElementName = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fContentSpec, 0, newIntArray, 0, chunk);
            fContentSpec = newIntArray;
            ContentModel[][] newContentModel = new ContentModel[chunk * 2][];
            System.arraycopy(fContentModel, 0, newContentModel, 0, chunk);
            fContentModel = newContentModel;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fAttlistHead, 0, newIntArray, 0, chunk);
            fAttlistHead = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fAttlistTail, 0, newIntArray, 0, chunk);
            fAttlistTail = newIntArray;
        } catch (NullPointerException ex) {
        }
        fElementName[chunk] = new int[CHUNK_SIZE];
        fContentSpecType[chunk] = new byte[CHUNK_SIZE];
        fContentSpec[chunk] = new int[CHUNK_SIZE];
        fContentModel[chunk] = new ContentModel[CHUNK_SIZE];
        fAttlistHead[chunk] = new int[CHUNK_SIZE];
        fAttlistTail[chunk] = new int[CHUNK_SIZE];
        return true;
    }
    public int getElement(int elementNameIndex) {
        int hc = elementNameIndex % HASHTABLE_SIZE;
        int[] bucket = fElementNameHashtable[hc];
        if (bucket != null) {
            int j = 1;
            for (int i = 0; i < bucket[0]; i++) {
                if (bucket[j] == elementNameIndex)
                    return bucket[j+1];
                j += 2;
            }
        }
        return -1;
    }
    public int addElement(int elementNameIndex) {
        int hc = elementNameIndex % HASHTABLE_SIZE;
        int[] bucket = fElementNameHashtable[hc];
        if (bucket != null) {
            int j = 1;
            for (int i = 0; i < bucket[0]; i++) {
                if (bucket[j] == elementNameIndex)
                    return bucket[j+1];
                j += 2;
            }
        }
        int chunk = fElementCount >> CHUNK_SHIFT;
        int index = fElementCount & CHUNK_MASK;
        ensureElementCapacity(chunk);
        fElementName[chunk][index] = elementNameIndex;
        fContentSpecType[chunk][index] = (byte) ElementDecl.CONTENTSPEC_UNKNOWN;
        fContentSpec[chunk][index] = -1;
        fContentModel[chunk][index] = null;
        fAttlistHead[chunk][index] = -1;
        fAttlistTail[chunk][index] = -1;
        if (bucket == null) {
            bucket = new int[1 + (INITIAL_BUCKET_SIZE * 2)];
            bucket[0] = 1;
            bucket[1] = elementNameIndex;
            bucket[2] = fElementCount;
            fElementNameHashtable[hc] = bucket;
        } else {
            int count = bucket[0];
            int offset = 1 + (count * 2);
            if (offset == bucket.length) {
                int newSize = count + INITIAL_BUCKET_SIZE;
                int[] newBucket = new int[1 + (newSize * 2)];
                System.arraycopy(bucket, 0, newBucket, 0, offset);
                bucket = newBucket;
                fElementNameHashtable[hc] = bucket;
            }
            bucket[offset++] = elementNameIndex;
            bucket[offset++] = fElementCount;
            bucket[0] = ++count;
        }
        return fElementCount++;
    }
    public int addElementDecl(ElementDecl decl) {
        int elementNameIndex = decl.elementName;
        int hc = elementNameIndex % HASHTABLE_SIZE;
        int[] bucket = fElementNameHashtable[hc];
        if (bucket != null) {
            int j = 1;
            for (int i = 0; i < bucket[0]; i++) {
                if (bucket[j] == elementNameIndex) {
                    int elementIndex = bucket[j+1];
                    int chunk = elementIndex >> CHUNK_SHIFT;
                    int index = elementIndex & CHUNK_MASK;
                    if (fContentSpecType[chunk][index] != ElementDecl.CONTENTSPEC_UNKNOWN)
                        return -1;
                    fContentSpecType[chunk][index] = (byte)decl.contentSpecType;
                    fContentSpec[chunk][index] = decl.contentSpec;
                    fContentModel[chunk][index] = null;
                    return elementIndex;
                }
                j += 2;
            }
        }
        int chunk = fElementCount >> CHUNK_SHIFT;
        int index = fElementCount & CHUNK_MASK;
        ensureElementCapacity(chunk);
        fElementName[chunk][index] = decl.elementName;
        fContentSpecType[chunk][index] = (byte)decl.contentSpecType;
        fContentSpec[chunk][index] = decl.contentSpec;
        fContentModel[chunk][index] = null;
        fAttlistHead[chunk][index] = -1;
        fAttlistTail[chunk][index] = -1;
        if (bucket == null) {
            bucket = new int[1 + (INITIAL_BUCKET_SIZE * 2)];
            bucket[0] = 1;
            bucket[1] = elementNameIndex;
            bucket[2] = fElementCount;
            fElementNameHashtable[hc] = bucket;
        } else {
            int count = bucket[0];
            int offset = 1 + (count * 2);
            if (offset == bucket.length) {
                int newSize = count + INITIAL_BUCKET_SIZE;
                int[] newBucket = new int[1 + (newSize * 2)];
                System.arraycopy(bucket, 0, newBucket, 0, offset);
                bucket = newBucket;
                fElementNameHashtable[hc] = bucket;
            }
            bucket[offset++] = elementNameIndex;
            bucket[offset++] = fElementCount;
            bucket[0] = ++count;
        }
        return fElementCount++;
    }
    public int getElementName(int elementIndex) {
        if (elementIndex < 0 || elementIndex >= fElementCount)
            return -1;
        int chunk = elementIndex >> CHUNK_SHIFT;
        int index = elementIndex & CHUNK_MASK;
        return fElementName[chunk][index];
    }
    public int getContentSpecType(int elementIndex) {
        if (elementIndex < 0 || elementIndex >= fElementCount)
            return -1;
        int chunk = elementIndex >> CHUNK_SHIFT;
        int index = elementIndex & CHUNK_MASK;
        return fContentSpecType[chunk][index];
    }
    public int getContentSpec(int elementIndex) {
        if (elementIndex < 0 || elementIndex >= fElementCount)
            return -1;
        int chunk = elementIndex >> CHUNK_SHIFT;
        int index = elementIndex & CHUNK_MASK;
        return fContentSpec[chunk][index];
    }
    public String getContentSpecAsString(int elementIndex) {
        if (elementIndex < 0 || elementIndex >= fElementCount)
            return null;
        int chunk = elementIndex >> CHUNK_SHIFT;
        int index = elementIndex & CHUNK_MASK;
        switch (fContentSpecType[chunk][index]) {
        case ElementDecl.CONTENTSPEC_EMPTY:
            return "EMPTY";
        case ElementDecl.CONTENTSPEC_ANY:
            return "ANY";
        case ElementDecl.CONTENTSPEC_MIXED:
        case ElementDecl.CONTENTSPEC_CHILDREN:
            return getContentSpecNodeAsString(fContentSpec[chunk][index]);
        }
        return null;
    }
    public ContentModel getContentModel(int elementIndex) {
        if (elementIndex < 0 || elementIndex >= fElementCount)
            return null;
        int chunk = elementIndex >> CHUNK_SHIFT;
        int index = elementIndex & CHUNK_MASK;
        return fContentModel[chunk][index];
    }
    public void setContentModel(int elementIndex, ContentModel cm) {
        if (elementIndex < 0 || elementIndex >= fElementCount)
            return;
        int chunk = elementIndex >> CHUNK_SHIFT;
        int index = elementIndex & CHUNK_MASK;
        fContentModel[chunk][index] = cm;
    }
    //
    // contentspec entries
    //
    private boolean ensureNodeCapacity(int chunk) {
        try {
            return fNodeType[chunk][0] == 0;
        } catch (ArrayIndexOutOfBoundsException ex) {
            byte[][] newByteArray = new byte[chunk * 2][];
            System.arraycopy(fNodeType, 0, newByteArray, 0, chunk);
            fNodeType = newByteArray;
            int[][] newIntArray = new int[chunk * 2][];
            System.arraycopy(fNodeValue, 0, newIntArray, 0, chunk);
            fNodeValue = newIntArray;
        } catch (NullPointerException ex) {
        }
        fNodeType[chunk] = new byte[CHUNK_SIZE];
        fNodeValue[chunk] = new int[CHUNK_SIZE];
        return true;
    }
    //
    //
    //
    public int addContentSpecNode(ContentSpecNode csn) {
        int chunk = fNodeCount >> CHUNK_SHIFT;
        int index = fNodeCount & CHUNK_MASK;
        ensureNodeCapacity(chunk);
        switch (csn.type) {
        case ContentSpecNode.NODE_LEAF:
        case ContentSpecNode.NODE_ZERO_OR_ONE:
        case ContentSpecNode.NODE_ZERO_OR_MORE:
        case ContentSpecNode.NODE_ONE_OR_MORE:
            fNodeType[chunk][index] = (byte)csn.type;
            fNodeValue[chunk][index] = csn.value;
            return fNodeCount++;
        case ContentSpecNode.NODE_CHOICE:
        case ContentSpecNode.NODE_SEQ:
            fNodeType[chunk][index] = (byte)csn.type;
            fNodeValue[chunk][index] = csn.value;
            int nodeIndex = fNodeCount++;
            if (++index == CHUNK_SIZE) {
                chunk++;
                ensureNodeCapacity(chunk);
                index = 0;
            }
            fNodeType[chunk][index] = (byte)(csn.type | 64); // flag second entry for consistancy checking
            fNodeValue[chunk][index] = csn.otherValue;
            fNodeCount++;
            return nodeIndex;
        default:
            return -1;
        }
    }
    public void getContentSpecNode(int contentSpecIndex, ContentSpecNode csn) {
        int chunk = contentSpecIndex >> CHUNK_SHIFT;
        int index = contentSpecIndex & CHUNK_MASK;
        csn.type = fNodeType[chunk][index];
        csn.value = fNodeValue[chunk][index];
        if (csn.type == ContentSpecNode.NODE_CHOICE || csn.type == ContentSpecNode.NODE_SEQ) {
            if (++index == CHUNK_SIZE) {
                chunk++;
                index = 0;
            }
            csn.otherValue = fNodeValue[chunk][index];
        } else
            csn.otherValue = -1;
    }
    private void appendContentSpecNode(int contentSpecIndex, StringBuffer sb, boolean noParen) {
        int chunk = contentSpecIndex >> CHUNK_SHIFT;
        int index = contentSpecIndex & CHUNK_MASK;
        int type = fNodeType[chunk][index];
        int value = fNodeValue[chunk][index];
        switch (type) {
        case ContentSpecNode.NODE_LEAF:
            sb.append(value == -1 ? "#PCDATA" : fStringPool.toString(value));
            return;
        case ContentSpecNode.NODE_ZERO_OR_ONE:
            appendContentSpecNode(value, sb, false);
            sb.append('?');
            return;
        case ContentSpecNode.NODE_ZERO_OR_MORE:
            appendContentSpecNode(value, sb, false);
            sb.append('*');
            return;
        case ContentSpecNode.NODE_ONE_OR_MORE:
            appendContentSpecNode(value, sb, false);
            sb.append('+');
            return;
        case ContentSpecNode.NODE_CHOICE:
        case ContentSpecNode.NODE_SEQ:
            if (!noParen)
                sb.append('(');
            int leftChunk = value >> CHUNK_SHIFT;
            int leftIndex = value & CHUNK_MASK;
            int leftType = fNodeType[leftChunk][leftIndex];
            appendContentSpecNode(value, sb, leftType == type);
            sb.append(type == ContentSpecNode.NODE_CHOICE ? '|' : ',');
            if (++index == CHUNK_SIZE) {
                chunk++;
                index = 0;
            }
            appendContentSpecNode(fNodeValue[chunk][index], sb, false);
            if (!noParen)
                sb.append(')');
            return;
        default:
            return;
        }
    }
    public String getContentSpecNodeAsString(int contentSpecIndex) {
        int chunk = contentSpecIndex >> CHUNK_SHIFT;
        int index = contentSpecIndex & CHUNK_MASK;
        int type = fNodeType[chunk][index];
        int value = fNodeValue[chunk][index];
        StringBuffer sb = new StringBuffer();
        switch (type) {
        case ContentSpecNode.NODE_LEAF:
            sb.append("(" + (value == -1 ? "#PCDATA" : fStringPool.toString(value)) + ")");
            break;
        case ContentSpecNode.NODE_ZERO_OR_ONE:
            chunk = value >> CHUNK_SHIFT;
            index = value & CHUNK_MASK;
            if (fNodeType[chunk][index] == ContentSpecNode.NODE_LEAF) {
                value = fNodeValue[chunk][index];
                sb.append("(" + (value == -1 ? "#PCDATA" : fStringPool.toString(value)) + ")?");
            } else
                appendContentSpecNode(contentSpecIndex, sb, false);
            break;
        case ContentSpecNode.NODE_ZERO_OR_MORE:
            chunk = value >> CHUNK_SHIFT;
            index = value & CHUNK_MASK;
            if (fNodeType[chunk][index] == ContentSpecNode.NODE_LEAF) {
                value = fNodeValue[chunk][index];
                sb.append("(" + (value == -1 ? "#PCDATA" : fStringPool.toString(value)) + ")*");
            } else
                appendContentSpecNode(contentSpecIndex, sb, false);
            break;
        case ContentSpecNode.NODE_ONE_OR_MORE:
            chunk = value >> CHUNK_SHIFT;
            index = value & CHUNK_MASK;
            if (fNodeType[chunk][index] == ContentSpecNode.NODE_LEAF) {
                value = fNodeValue[chunk][index];
                sb.append("(" + (value == -1 ? "#PCDATA" : fStringPool.toString(value)) + ")+");
            } else
                appendContentSpecNode(contentSpecIndex, sb, false);
            break;
        case ContentSpecNode.NODE_CHOICE:
        case ContentSpecNode.NODE_SEQ:
            appendContentSpecNode(contentSpecIndex, sb, false);
            break;
        default:
            return null;
        }
        return sb.toString();
    }
    //
    // attribute list interfaces
    //
    private boolean ensureAttrCapacity(int chunk) {
        try {
            return fAttName[chunk][0] == 0;
        } catch (ArrayIndexOutOfBoundsException ex) {
            byte[][] newByteArray = new byte[chunk * 2][];
            System.arraycopy(fAttType, 0, newByteArray, 0, chunk);
            fAttType = newByteArray;
            newByteArray = new byte[chunk * 2][];
            System.arraycopy(fAttDefaultType, 0, newByteArray, 0, chunk);
            fAttDefaultType = newByteArray;
            int[][] newIntArray = new int[chunk * 2][];
            System.arraycopy(fAttName, 0, newIntArray, 0, chunk);
            fAttName = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fEnumeration, 0, newIntArray, 0, chunk);
            fEnumeration = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fAttValue, 0, newIntArray, 0, chunk);
            fAttValue = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fNextAttDef, 0, newIntArray, 0, chunk);
            fNextAttDef = newIntArray;
        } catch (NullPointerException ex) {
        }
        fAttType[chunk] = new byte[CHUNK_SIZE];
        fAttDefaultType[chunk] = new byte[CHUNK_SIZE];
        fAttName[chunk] = new int[CHUNK_SIZE];
        fEnumeration[chunk] = new int[CHUNK_SIZE];
        fAttValue[chunk] = new int[CHUNK_SIZE];
        fNextAttDef[chunk] = new int[CHUNK_SIZE];
        return true;
    }
    //
    public int addAttDef(int elementIndex, AttDef attDef) throws Exception {
        //
        // check fields
        //
        int elemChunk = elementIndex >> CHUNK_SHIFT;
        int elemIndex = elementIndex & CHUNK_MASK;
        int attlistIndex = fAttlistHead[elemChunk][elemIndex];
        while (attlistIndex != -1) {
            int attrChunk = attlistIndex >> CHUNK_SHIFT;
            int attrIndex = attlistIndex & CHUNK_MASK;
            if (fAttName[attrChunk][attrIndex] == attDef.attName) {
                if (fParserState.getWarningOnDuplicateAttDef())
                    fErrorHandler.error1(ErrorCode.W_ATTL2, attDef.attName);
                return -1;
            }
            if (attDef.attType == AttDef.ID && fAttType[attrChunk][attrIndex] == attDef.ID) {
                fErrorHandler.error1(ErrorCode.V_ATTL2, attDef.attName); // REVISIT - inaccurate error message
                return -1;
            }
            attlistIndex = fNextAttDef[attrChunk][attrIndex];
        }
        //
        if (fXMLSpace == -1) {
            fXMLSpace = fStringPool.addSymbol("xml:space".intern());
            fDefault = fStringPool.addSymbol("default".intern());
            fPreserve = fStringPool.addSymbol("preserve".intern());
        }
        if (attDef.attName == fXMLSpace) {
            boolean ok = false;
            if (attDef.attType == AttDef.ENUMERATION) {
                int index = attDef.enumeration;
                if (index != -1) {
                    ok = fStringPool.stringListLength(index) == 2 &&
                         fStringPool.stringInList(index, fDefault) &&
                         fStringPool.stringInList(index, fPreserve);
                }
            }
            if (!ok)
                fErrorHandler.error(ErrorCode.W_ATTL1);
        }
        //
        // save the fields
        //
        int chunk = fAttDefCount >> CHUNK_SHIFT;
        int index = fAttDefCount & CHUNK_MASK;
        ensureAttrCapacity(chunk);
        fAttName[chunk][index] = attDef.attName;
        fAttType[chunk][index] = (byte)attDef.attType;
        fEnumeration[chunk][index] = attDef.enumeration;
        fAttDefaultType[chunk][index] = (byte)attDef.attDefaultType;
        fAttValue[chunk][index] = attDef.attValue;
        //
        // add to the attr list for this element
        //
        int nextIndex = -1;
        if (attDef.attValue != -1) {
            nextIndex = fAttlistHead[elemChunk][elemIndex];
            fAttlistHead[elemChunk][elemIndex] = fAttDefCount;
            if (nextIndex == -1)
                fAttlistTail[elemChunk][elemIndex] = fAttDefCount;
        } else {
            nextIndex = fAttlistTail[elemChunk][elemIndex];
            fAttlistTail[elemChunk][elemIndex] = fAttDefCount;
            if (nextIndex == -1)
                fAttlistHead[elemChunk][elemIndex] = fAttDefCount;
            else {
                fNextAttDef[nextIndex >> CHUNK_SHIFT][nextIndex & CHUNK_MASK] = fAttDefCount;
                nextIndex = -1;
            }
        }
        fNextAttDef[chunk][index] = nextIndex;
//System.out.println("AttDef#" + fAttDefCount + ": <" + fStringPool.toString(getElementName(elementIndex)) +
//                        " " + fStringPool.toString(attDef.attName) +
//                        " " + AttDef.getAttTypeString(attDef.attType) + ">");
        return fAttDefCount++;
    }
    public int getAttDef(int elementNameIndex, int attrNameIndex) {
        int chunk = 0;
        int index = 0;
        int elementIndex;
        for (elementIndex = 0; elementIndex < fElementCount; elementIndex++) {
            if (fElementName[chunk][index] == elementNameIndex) {
                int attDefIndex = fAttlistHead[chunk][index];
                while (true) {
                    if (attDefIndex == -1)
                        return -1;
                    chunk = attDefIndex >> CHUNK_SHIFT;
                    index = attDefIndex & CHUNK_MASK;
                    if (fAttName[chunk][index] == attrNameIndex)
                        return attDefIndex;
                    attDefIndex = fNextAttDef[chunk][index];
                }
            }
            if (++index == CHUNK_SIZE) {
                chunk++;
                index = 0;
            }
        }
        return -1;
    }
    //added by RIP
    public int getAttName(int attDefIndex) {
        int chunk = attDefIndex >> CHUNK_SHIFT;
        int index = attDefIndex & CHUNK_MASK;
        return fAttName[chunk][index];
    }
    //added by RIP
    public int getAttValue(int attDefIndex) {
        int chunk = attDefIndex >> CHUNK_SHIFT;
        int index = attDefIndex & CHUNK_MASK;
        return fAttValue[chunk][index];
    }
    public int getAttType(int attDefIndex) {
        int chunk = attDefIndex >> CHUNK_SHIFT;
        int index = attDefIndex & CHUNK_MASK;
        return fAttType[chunk][index];
    }
    public int getAttDefaultType(int attDefIndex) {
        int chunk = attDefIndex >> CHUNK_SHIFT;
        int index = attDefIndex & CHUNK_MASK;
        return fAttDefaultType[chunk][index];
    }
    public int getEnumeration(int attDefIndex) {
        int chunk = attDefIndex >> CHUNK_SHIFT;
        int index = attDefIndex & CHUNK_MASK;
        return fEnumeration[chunk][index];
    }
    public int addDefaultAttributes(int elementIndex, AttrPool attrPool, int firstAttrIndex, int lastAttrIndex) throws Exception {
        //
        // Check after all specified attrs are scanned
        // (1) report error for REQUIRED attrs that are missing (V_TAGc)
        // (2) check that FIXED attrs have matching value (V_TAGd)
        // (3) add default attrs (FIXED and NOT_FIXED)
        //
        int elemChunk = elementIndex >> CHUNK_SHIFT;
        int elemIndex = elementIndex & CHUNK_MASK;
        int attlistIndex = fAttlistHead[elemChunk][elemIndex];
        int firstCheck = firstAttrIndex;
        int lastCheck = lastAttrIndex;
//System.out.println("addDefAttrs: Element " + fStringPool.toString(fElementName[elemChunk][elemIndex]));
        while (attlistIndex != -1) {
            int attrChunk = attlistIndex >> CHUNK_SHIFT;
            int attrIndex = attlistIndex & CHUNK_MASK;
            int attDefType = fAttDefaultType[attrChunk][attrIndex];
//System.out.println("  Attr " + fStringPool.toString(fAttName[attrChunk][attrIndex]) + " (" + fAttName[attrChunk][attrIndex] + ")");
            if (attDefType != AttDef.IMPLIED) {
                int attName = fAttName[attrChunk][attrIndex];
                boolean specified = false;
                if (firstCheck != -1) {
                    for (int i = firstCheck; i <= lastCheck; i++) {
                        int aName = attrPool.getAttrName(i);
//System.out.println("      Attr " + fStringPool.toString(aName) + " (" + aName + ")");
                        if (attrPool.getAttrName(i) == attName) {
                            if (attDefType == AttDef.FIXED && attrPool.getAttValue(i) != fAttValue[attrChunk][attrIndex])
                                fErrorHandler.error3(ErrorCode.V_TAGd, attName, fAttValue[attrChunk][attrIndex], attrPool.getAttValue(i));
                            specified = true;
                            break;
                        }
                    }
                }
                if (!specified) {
                    if (attDefType == AttDef.REQUIRED)
                        fErrorHandler.error1(ErrorCode.V_TAGc, attName);
                    else {
                        fAttr.attName = attName;
                        fAttr.attType = fAttType[attrChunk][attrIndex];
                        fAttr.attValue = fAttValue[attrChunk][attrIndex];
                        fAttr.specified = false;
                        lastAttrIndex = attrPool.addAttr(fAttr, -1);
                        if (firstAttrIndex == -1)
                            firstAttrIndex = lastAttrIndex;
                    }
                }
            }
            attlistIndex = fNextAttDef[attrChunk][attrIndex];
        }
        if (lastAttrIndex != -1)
            attrPool.setIsLastAttr(lastAttrIndex);
        return firstAttrIndex;
    }
    //
    //
    //
    public boolean addId(int idIndex, int elementNameIndex) {
        Integer key = new Integer(idIndex);
        if (fIdDefs == null)
            fIdDefs = new Hashtable();
        else if (fIdDefs.containsKey(key))
            return false;
        if (fNullValue == null)
            fNullValue = new Object();
        fIdDefs.put(key, fNullValue/*new Integer(elementNameIndex)*/);
        return true;
    }
    public void addIdRef(int idIndex, int elementNameIndex) {
        Integer key = new Integer(idIndex);
        if (fIdDefs != null && fIdDefs.containsKey(key))
            return;
        if (fIdRefs == null)
            fIdRefs = new Hashtable();
        else if (fIdRefs.containsKey(key))
            return;
        if (fNullValue == null)
            fNullValue = new Object();
        fIdRefs.put(key, fNullValue/*new Integer(elementNameIndex)*/);
    }
    public void checkIdRefs() throws Exception {
        if (fIdRefs == null)
            return;
        Enumeration en = fIdRefs.keys();
        while (en.hasMoreElements()) {
            Integer key = (Integer)en.nextElement();
            if (fIdDefs == null || !fIdDefs.containsKey(key))
                fErrorHandler.error1(ErrorCode.V_IDREF0, key.intValue());
        }
    }
    //
    // Check simple namespace constraints
    //
    public void checkNamespace(int elementIndex, int firstAttrIndex) throws Exception {
        if (firstAttrIndex != -1) {
            int i = 0;
            int attrIndex = firstAttrIndex;
			//            while (true) {
                int attrName = fAttrPool.getAttrName(attrIndex);
                String anString = fStringPool.toString(attrName);
                int errorCode = checkQName(elementIndex, anString);
                if (errorCode != -1)
                    fErrorHandler.error1(errorCode, attrName);
                int attValue = fAttrPool.getAttValue(attrIndex);
                String avString = fStringPool.toString(attValue);
                if (avString.length() == 0 && anString.startsWith("xmlns:"))
                    fErrorHandler.error1(ErrorCode.E_TAGl, attrName);
                //
                // The following code is O(n^2).
                //
                if (i == localpartcache.length) {
                    int[] newcache = new int[i * 2];
                    System.arraycopy(localpartcache, 0, newcache, 0, i);
                    localpartcache = newcache;
                    newcache = new int[i * 2];
                    System.arraycopy(nsnamecache, 0, newcache, 0, i);
                    nsnamecache = newcache;
                }
                localpartcache[i] = getNSLocalName(anString, attrName);
                nsnamecache[i] = getNSName(elementIndex, anString);
                if (nsnamecache[i] != -1) {
                    for (int j = 0;  j < i;  j ++) {
                        if (localpartcache[j] == localpartcache[i]) {
                            int ns = nsnamecache[j];
                            if (ns != -1 && ns == nsnamecache[i])
                                fErrorHandler.error4(ErrorCode.E_TAGm, attrName, nsnamecache[i],
                                             localpartcache[i], fAttrPool.getAttrName(firstAttrIndex +j));
                        }
                    }
                }
				//            }
        }
        int elementName = getElementName(elementIndex);
        String enString = fStringPool.toString(elementName);
        int errorCode = checkQName(elementIndex, enString);
        if (errorCode != -1)
            fErrorHandler.error1(errorCode, elementName);
    }

    /** 
     * Checks that all declared elements refer to declared elements
     * in their content models. This method calls out to the error
     * handler to indicate warnings.
     */
    public void checkDeclaredElements() throws Exception {

        // check to make sure that we should even be doing this
        if (fParserState.getValidationHandler() != null) {
            for (int i = 0; i < fElementCount; i++) {
                int type  = getContentSpecType(i);
                if (type == ElementDecl.CONTENTSPEC_MIXED ||
                    type == ElementDecl.CONTENTSPEC_CHILDREN) 
                    {
                    int chunk = i >> CHUNK_SHIFT;
                    int index = i &  CHUNK_MASK;
                    int contentSpecIndex = fContentSpec[chunk][index];
                    checkDeclaredElements(i, contentSpecIndex);
                }
            }
        }

    } // checkDeclaredElements()

    //
    // Private methods
    //

    /** 
     * Does a recursive (if necessary) check on the specified element's
     * content spec to make sure that all children refer to declared
     * elements.
     * <p>
     * This method assumes that it will only be called when there is
     * a validation handler.
     */
    private void checkDeclaredElements(int elementIndex, int contentSpecIndex)
        throws Exception {
        
        // get spec type and value
        int chunk = contentSpecIndex >> CHUNK_SHIFT;
        int index = contentSpecIndex &  CHUNK_MASK;
        int type  = fNodeType[chunk][index];
        int value = fNodeValue[chunk][index];

        // handle type
        switch (type) {
        
            // #PCDATA | element
            case ContentSpecNode.NODE_LEAF: {
                // perform check for declared element
                if (value != -1 && getElement(value) == -1) {
                    int elemChunk = elementIndex >> CHUNK_SHIFT;
                    int elemIndex = elementIndex &  CHUNK_MASK;
                    int elementNameIndex = fElementName[elemChunk][elemIndex];
                    fErrorHandler.error2(ErrorCode.W_DTD5, 
                                         elementNameIndex, value);
                }
                break;
            }

            // (...)? | (...)* | (...)+
            case ContentSpecNode.NODE_ZERO_OR_ONE:
            case ContentSpecNode.NODE_ZERO_OR_MORE: 
            case ContentSpecNode.NODE_ONE_OR_MORE: {
                checkDeclaredElements(elementIndex, value);
                break;
            }

            // (... , ...) | (... | ...)
            case ContentSpecNode.NODE_CHOICE:
            case ContentSpecNode.NODE_SEQ: {
                checkDeclaredElements(elementIndex, value);
                if (++index == CHUNK_SIZE) {
                    chunk++;
                    index = 0;
                }
                checkDeclaredElements(elementIndex, fNodeValue[chunk][index]);
                break;
            }

        } // switch

    } // checkDeclaredElements(int,int)

    private int checkQName(int elementIndex, String qname) {
        int index = qname.indexOf(':');
        if (index < 0)
            return -1;
        if (getNamespaceForPrefix(elementIndex, qname.substring(0, index)) == -1)
            return ErrorCode.E_TAGf;
        if (qname.indexOf(':', index+1) >= 0)
            return ErrorCode.E_TAGe;
        return -1;
    }
    private int getNSLocalName(String qname, int qnameIndex) {
        int index = qname.indexOf(':');
        if (index < 0)
            return qnameIndex;
        return fStringPool.addSymbol(qname.substring(index+1));
    }
    private int getNSName(int elementIndex, String qname) {
        int index = qname.indexOf(':');
        String prefix = index < 0 ? "" : qname.substring(0, index);
        return getNamespaceForPrefix(elementIndex, prefix);
    }
    private int getNamespaceForPrefix(int elementIndex, String prefix) {
        return -1;
    }
/*

    static final String S_XMLNS         = "xmlns";

    public String getNamespaceForPrefix(String prefix) {
        if (prefix.equals("xml") || prefix.equals(S_XMLNS))
            return prefix;                      // ???
        String attributeName = prefix.length() == 0 ? S_XMLNS : ("xmlns:"+prefix);
        Node element = this;
        do {
            String nsname = ((Element)element).getAttribute(attributeName);
            if (nsname != null)
                return nsname;
            while (true) {
                element = element.getParentNode();
                if (element == null)  return attributeName == S_XMLNS ? "" : null;
                int type = element.getNodeType();
                if (type == Node.ENTITY_REFERENCE_NODE) {
                } else if (type == Node.ELEMENT_NODE) {
                    break;
                } else
                    return attributeName == S_XMLNS ? "" : null;
            }
        } while (element != null);              // This condition has no sense.
        return null;                            // Not reached
    }

 */
}
