/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.internal;

import com.ibm.xml.framework.*;
import org.xml.sax.InputSource;
import org.xml.sax.EntityResolver;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Stack;

/**
 * Entity handler interface.
 *
 * @author  Glenn Marcy
 * @author  Andy Clark, IBM
 * @version Revision: 65 1.8 src/com/ibm/xml/internal/DefaultEntityHandler.java, parser, xml4j2, xml4j2_0_13 
 *
 * @see XMLParser
 */
public final class DefaultEntityHandler implements XMLEntityHandler {

    //
    // Constants
    //

    /** Set to true and recompile to include debugging information. */
    private static final boolean DEBUG = false;

    //
    // Data
    //

    /** Recognizers. */
    private Stack fRecognizers = null;

    /**
     * Stack of input sources which holds the "path" of input sources
     * to the current point in the parsing.
     */
    private Stack fInputSourceStack = null;

    /** Parser state. */
    private ParserState fParserState = null;

    /** The entity resolver, which can be a catalog. */
    private EntityResolver fResolver = null;

    //
    // Constructors
    //

    /** 
     * Constructs a default entity handler with a given parser state.
     *
     * @param parserState The parser state.
     */
    public DefaultEntityHandler(ParserState parserState) {
        fParserState = parserState;
        fInputSourceStack = new Stack();
        initializeRecognizers();
    }

    //
    // XMLEntityHandler methods
    //

    /** 
     * Resets the entity handler, usually in response to a new call to
     * <tt>XMLParser.parse</tt>. 
     *
     * @param parserState The parser state.
     */
    public void reset(ParserState parserState) {
        fParserState = parserState;
        fInputSourceStack.removeAllElements();
    }

    /** 
     * Adds a recognizer. 
     *
     * @param recognizer The XML recognizer to add.
     */
    public void addRecognizer(XMLDeclRecognizer recognizer) {
        fRecognizers.push(recognizer);
    }
    
    /**
     * Sets the resolver used to resolve external entities. The EntityResolver
     * interface supports resolution of public and system identifiers. If the
     * resolver set is also of type com.ibm.xml.framework.Catalog, then 
     * catalogs can be loaded in a general manner by calling 
     * <tt>XMLParser.loadCatalog()</tt>.
     *
     * @param resolver The new entity resolver. Passing a null value will
     *                 uninstall the currently installed resolver.
     *
     * @see com.ibm.xml.framework.Catalog
     * @see com.ibm.xml.framework.XMLParser#loadCatalog
     */
    public void setEntityResolver(EntityResolver resolver) {
        fResolver = resolver;
    }

    /**
     * Returns the resolver used to resolve external entities.
     *
     * @return The resolver.
     *
     * @see #setEntityResolver
     */
    public EntityResolver getEntityResolver() {
        return fResolver;
        }

    /**
     * Performs the entity resolution. The most common way to resolve
     * entities is to get the strings represented by the publicId and
     * systemId indexes and ask the installed EntityResolver to resolve
     * the entities.
     * 
     * @param publidId The index of the public identifier string in the
     *                 parser StringPool. If this value is -1, then a
     *                 system identifier alias is requested; else the
     *                 system identifier represents the default entity
     *                 resolution if no public identifier mapping is
     *                 possible.
     * @param systemId The index of the system identifier string in the
     *                 parser StringPool.
     *
     * @return Returns an InputSource of the resolved entity or null if
     *         no resolution is possible.
     *
     * @exception java.lang.Exception If entity resolution fails, this method can
     *                                throw an exception.
     *
     * @see #getEntityResolver
     * @see com.ibm.xml.framework.StringPool#toString
     */
    public InputSource resolveEntity(int publicId, int systemId) 
        throws Exception 
        {

        // use registered entity resolver
        EntityResolver resolver = getEntityResolver();
        if (resolver != null) {
            StringPool pool = fParserState.getStringPool();

            // get ids as strings
            String publicIdStr = pool.toString(publicId);
            String systemIdStr = pool.toString(systemId);

            // resolve entity
            return resolver.resolveEntity(publicIdStr, systemIdStr);
            }

        // no resolution possible
        return null;

        } // resolveEntity(int,int):InputSource

    /**
     * Fully expands a given system identified. The system identifier string
     * to be resolved and the fully expanded system identifier string are
     * stored in the parser StringPool.
     * <p>
     * The following method expands a system identifier given as a
     * <tt>java.lang.String</tt>.
     * <pre>
     *     public String expandSystemId(XMLParser parser, String systemId) throws Exception {
     *         StringPool pool = parser.getParserState().getStringPool();
     *
     *         // expand entity
     *         int oldIndex = pool.addString(systemId);
     *         int newIndex = parser.getEntityHandler().expandSystemId(oldIndex);
     *         systemId = pool.toIndex(newIndex);
     *
     *         // release string pool resources
     *         pool.releaseString(oldIndex);
     *         pool.releaseString(newIndex);
     *
     *         // return expanded system id
     *         return systemId;
     *         }
     * </pre>
     *
     * @param systemId The index of the system identifier string in the
     *                 parser StringPool.
     *
     * @return Returns the index in the StringPool that holds the expanded
     *         system identifier. If the given system identifier is already
     *         fully expanded, this method can return the original systemId
     *         value.
     *
     * @see com.ibm.xml.framework.StringPool#toString
     */
    public int expandSystemId(int systemId) {
        
        if (systemId != -1) {
            ParserState state  = fParserState;
            StringPool  string = state.getStringPool();
            try {
                String systemLiteral = string.toString(systemId);
                URL url = expandSystemId(systemLiteral);
                if (url != null) {
                    if (DEBUG) {
                        System.out.println("url="+url);
                        }
                    return string.addString(url.toString());
                    }

                // else, return system id passed in
                if (DEBUG) {
                    System.out.println("url == null");
                    }
                }
            catch (Exception e) {
                // return system id passed in
                if (DEBUG) {
                    System.out.println("error:"+e.getMessage());
                    e.printStackTrace(System.err);
                    }
                }
            }

        // return string pool index of system id
        return systemId;

        } // expandSystemId(int):int
    
    /**
     * Creates a parser reader object from the given input source.
     * <p>
     * This is the point in the default implementation where relative
     * system identifiers are expanded. For example, if the current
     * source URI is "http://host/path/file.xml" and the specified
     * input source system identifier is "dtd/file.dtd", then the
     * system identifier, if not already opened, will be set to
     * "http://host/path/dtd/file.dtd".
     *
     * @param source  The input source.
     * @param xmlDecl True if the source represents an XML decl.
     *
     * @return Returns a parser reader object.
     *
     * @exception java.lang.Exception Throws exception on error.
     */
    public XMLReader createReader(InputSource source, boolean xmlDecl) 
        throws Exception 
        {

        // create reader from source's character stream
        if (source.getCharacterStream() != null) {
            return new CharReader(fParserState, 
                                  source.getPublicId(), 
                                  source.getSystemId(),
                                  source.getCharacterStream()
                                  );
            } 

        // create reader from source's byte stream
        if (source.getEncoding() != null && source.getByteStream() != null) {
            return new CharReader(fParserState, 
                                  source.getPublicId(), 
                                  source.getSystemId(),
                                  new InputStreamReader(source.getByteStream(), 
                                                        source.getEncoding()
                                                        )
                                  );
            }

        // create new input stream
        InputStream is = source.getByteStream();
        if (is == null) {
            try {
                String id = source.getSystemId();
                if (id != null) {
                    try {
                        if (DEBUG) { 
                            System.out.println("BEFORE: systemId=\""+id+"\""); 
                            }
                        ParserState state = fParserState;
                        StringPool string = state.getStringPool();
                        int oldindex = string.addString(id);
                        int newindex = state.getEntityHandler().expandSystemId(oldindex);
                        if (oldindex != newindex) {
                            id = string.orphanString(newindex);
                            source.setSystemId(id);
                            }
                        if (DEBUG) { 
                            System.out.println("AFTER:  systemId=\""+id+"\"");
                            }
                        string.releaseString(oldindex);
                        }
                    catch (Exception e) { 
                        // let it go through
                        }
                    }

                // create url and open the stream
                URL url = new URL(id);
                is = url.openStream();
                } 
            catch (Exception e) {
//                System.err.println(e);
                throw e;
                }
            }

        // create array and find recognizer
        ChunkyByteArray data = new ChunkyByteArray(is);
        XMLReader reader = callRecognizers(source, xmlDecl, data);
        if (reader == null) {
            if (fParserState.getUseCharReaderForUTF8()) {
//                reader = new CharReader(fParserState, source.getPublicId(), source.getSystemId(),
//                                        new InputStreamReader(data, "UTF8"));
                reader = new UTF8CharReader(fParserState, source.getPublicId(), source.getSystemId(), data);
                }
            else {
                reader = new UTF8Reader(fParserState, source.getPublicId(), source.getSystemId(), data);
                }
            }

        return reader;

        } // createReader(InputSource,boolean):XMLReader

    /** 
     * Start of an input source. 
     * <p>
     * This method pushes the given input source onto the top of
     * the input source stack.
     *
     * @param inputSource The input source that is starting.
     */
    public void startInputSource(InputSource inputSource) {
        fInputSourceStack.push(inputSource);
    }

    /** 
     * End of an input source. 
     * <p>
     * This method pops the current input source off of the top of
     * the input source stack.
     *
     * @param inputSource The input source that is ended.
     */
    public void endInputSource(InputSource inputSource) {
        InputSource lastSource = (InputSource) fInputSourceStack.pop();
        if (lastSource != inputSource)
			try {
				fParserState.getErrorHandler().error(ErrorCode.E_INT_ISS);
			} catch (Exception e) {
				// revisit: don't fail silently
			}
    }

    //
    // Public methods
    //

    /**
     * Expands a system id and returns the system id as a URL, if
     * it can be expanded. A return value of null means that the 
     * identifier is already expanded. An exception thrown
     * indicates a failure to expand the id.
     *
     * @param systemId The systemId to be expanded. 
     *
     * @return Returns the URL object representing the expanded system
     *         identifier. A null value indicates that the given
     *         system identifier is already expanded.
     *
     * @exception java.lang.Exception Throws exception on error.
     */
    public URL expandSystemId(String systemId)
        throws Exception
        {

        // normalize id
        if (systemId == null) {
            systemId = "/";
            }

        // if id already expanded, return
        try {
            URL url = new URL(systemId);
            if (url != null) {
                return null;
                }
            }
        catch (MalformedURLException e) {
            // continue on...
            }

        // normalize id
        systemId = fixURI(systemId);

        // normalize base
        String currentSystemId;
        try {
            InputSource currentSource = (InputSource) fInputSourceStack.peek();
            currentSystemId = currentSource.getSystemId();
            }
        catch (Exception e) {
            currentSystemId = null;
            }
        URL base = null;
        if (currentSystemId == null) {
            String dir;
            try {
                dir = fixURI(System.getProperty("user.dir"));
                }
            catch (SecurityException se) {
                dir = "";
                }
            if (!dir.endsWith("/")) {
                dir = dir + "/";
                }
            base = new URL("file", "", dir);
            }
        else {
            base = new URL(currentSystemId);
            }

        // expand id
        return new URL(base, systemId);

        } // expandSystemId(URL,String):URL
    
    //
    // Private methods
    //

    /** Initializes the recognizers. */
    private void initializeRecognizers() {
        if (fRecognizers == null) {
            fRecognizers = new Stack();
            fRecognizers.push(new EBCDICRecognizer());
            fRecognizers.push(new UCSRecognizer());
            fRecognizers.push(new UTF8Recognizer());
        }
    }

    /** 
     * Finds a recognizer that can handle the given input source. 
     *
     * @param source  The input source to recognize.
     * @param xmlDecl True if XML decl.
     * @param data    The input data.
     *
     * @return Returns the recognizer that can handle the given input
     *         source.
     *
     * @exception java.lang.Exception Throws exception on error.
     */
    private XMLReader callRecognizers(InputSource source, boolean xmlDecl, ChunkyByteArray data) 
        throws Exception 
        {

        for (int i = fRecognizers.size() - 1; i >= 0; i--) {
            XMLDeclRecognizer recognizer = (XMLDeclRecognizer)fRecognizers.elementAt(i);
            XMLReader reader = recognizer.recognize(fParserState, source, data, xmlDecl);
            if (reader != null) {
                return reader;
                }
            }

        return null;

        } // callRecognizers(InputSource,boolean,ChunkyByteArray):XMLReader

    /**
     * Fixes a platform dependent filename to standard URI form.
     *
     * @param str The string to fix.
     *
     * @return Returns the fixed URI string.
     */
    private static String fixURI(String str) {

        // handle platform dependent strings
        str = str.replace(java.io.File.separatorChar, '/');

        // Windows fix
        if (str.length() >= 2) {
            char ch1 = str.charAt(1);
            if (ch1 == ':') {
                char ch0 = Character.toUpperCase(str.charAt(0));
                if (ch0 >= 'A' && ch0 <= 'Z') {
                    str = "/" + str;
                    }
                }
            }

        // done
        return str;

        } // fix(String):String

    } // class DefaultEntityHandler
