/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */
package com.ibm.xml.internal;

import com.ibm.xml.framework.*;

/**
 *
 * @version Revision: 83 1.6 src/com/ibm/xml/internal/DefaultEntityPool.java, parser, xml4j2, xml4j2_0_13 
 */
public final class DefaultEntityPool implements EntityPool {

    //
    // Constants
    //

    /** Defect #530. */
    private static final boolean DEBUG_530 = false;

    //
    // Chunk size constants
    //
    static final int CHUNK_SHIFT = 5;           // 2^5 = 32
    static final int CHUNK_SIZE = (1 << CHUNK_SHIFT);
    static final int CHUNK_MASK = CHUNK_SIZE - 1;
    static final int INITIAL_CHUNK_COUNT = (1 << (10 - CHUNK_SHIFT));   // 2^10 = 1k
    //
    // Instance variables
    //
    private ParserState fParserState = null;
    private StringPool fStringPool = null;
    //
    // We store both EntityDecl and NotationDecl instances in this pool.
    // A NotationDecl has an fEntityName field of -1.  The fNotationDeclHead
    // index is -1 if the NotationDecl list is empty, otherwise it contains
    // the index of the the last NotationDecl in the list and the fEntityValue
    // field contains the index of the previous NotationDecl, or -1 when at
    // the end of the list.
    //
    private int fEntityCount = 0;
    private int[][] fEntityName = new int[INITIAL_CHUNK_COUNT][];
    private int[][] fEntityValue = new int[INITIAL_CHUNK_COUNT][];
    private int[][] fPublicId = new int[INITIAL_CHUNK_COUNT][];
    private int[][] fSystemId = new int[INITIAL_CHUNK_COUNT][];
    private int[][] fNotationName = new int[INITIAL_CHUNK_COUNT][];
    private int fNotationListHead = -1;
    private EntityDecl fInternalEntityDecl = null;
    //
    // Constructor
    //
    public DefaultEntityPool(ParserState parserState, boolean createStandardEntities) {
        fParserState = parserState;
        fStringPool = parserState.cacheStringPool();
        if (createStandardEntities) {
            fInternalEntityDecl = new EntityDecl();
            createInternalEntity("lt", "&#60;");
            createInternalEntity("gt", ">");
            createInternalEntity("amp", "&#38;");
            createInternalEntity("apos", "\'");
            createInternalEntity("quot", "\"");
        }
    }
    //
    //
    //
    public void reset(ParserState parserState) {
        fParserState = parserState;
        fStringPool = parserState.cacheStringPool();
        fEntityCount = 0;
        fNotationListHead = -1;
        if (fInternalEntityDecl != null) {
            createInternalEntity("lt", "&#60;");
            createInternalEntity("gt", ">");
            createInternalEntity("amp", "&#38;");
            createInternalEntity("apos", "\'");
            createInternalEntity("quot", "\"");
        }
    }
    //
    //
    //
    public EntityPool resetOrCopy(ParserState parserState) {
        return new DefaultEntityPool(parserState, fInternalEntityDecl != null);
    }
    //
    //
    //
    private void createInternalEntity(String name, String value) {
        
        fInternalEntityDecl.entityName = fStringPool.addSymbol(name);
        fInternalEntityDecl.entityValue = fStringPool.addString(value);
        fInternalEntityDecl.publicId = -1;
        fInternalEntityDecl.systemId = -1;
        fInternalEntityDecl.notationName = -1;
        addEntityDecl(fInternalEntityDecl);
    }
    //
    //
    //
    private boolean ensureCapacity(int chunk) {
        try {
            return fEntityName[chunk][0] == 0;
        } catch (ArrayIndexOutOfBoundsException ex) {
            int[][] newIntArray = new int[chunk * 2][];
            System.arraycopy(fEntityName, 0, newIntArray, 0, chunk);
            fEntityName = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fEntityValue, 0, newIntArray, 0, chunk);
            fEntityValue = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fPublicId, 0, newIntArray, 0, chunk);
            fPublicId = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fSystemId, 0, newIntArray, 0, chunk);
            fSystemId = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fNotationName, 0, newIntArray, 0, chunk);
            fNotationName = newIntArray;
        } catch (NullPointerException ex) {
        }
        fEntityName[chunk] = new int[CHUNK_SIZE];
        fEntityValue[chunk] = new int[CHUNK_SIZE];
        fPublicId[chunk] = new int[CHUNK_SIZE];
        fSystemId[chunk] = new int[CHUNK_SIZE];
        fNotationName[chunk] = new int[CHUNK_SIZE];
        return true;
    }
    public int addEntityDecl(EntityDecl decl) {
        int chunk = fEntityCount >> CHUNK_SHIFT;
        int index = fEntityCount & CHUNK_MASK;
        ensureCapacity(chunk);
        fEntityName[chunk][index] = decl.entityName;
        fEntityValue[chunk][index] = decl.entityValue;
        fPublicId[chunk][index] = decl.publicId;
        fSystemId[chunk][index] = decl.systemId;
        fNotationName[chunk][index] = decl.notationName;
        int entityIndex = fEntityCount++;
//System.out.println("EntityDecl#" + entityIndex +
//                      ": Name " + fParser.getStringPool().toString(decl.entityName) +
//                      ": Value (" + (decl.entityValueData == null ? "" :
//                                     decl.entityValueData.toString(decl.entityValueOffset, decl.entityValueLength)) + ")");
        return entityIndex;
    }
    public int addNotationDecl(NotationDecl decl) {
        int nIndex = fNotationListHead;
        while (nIndex != -1) {
            int chunk = nIndex >> CHUNK_SHIFT;
            int index = nIndex & CHUNK_MASK;
            if (fNotationName[chunk][index] == decl.notationName)
                return -1;
            nIndex = fEntityValue[chunk][index];
        }
        int chunk = fEntityCount >> CHUNK_SHIFT;
        int index = fEntityCount & CHUNK_MASK;
        ensureCapacity(chunk);
        fEntityName[chunk][index] = -1;
        fEntityValue[chunk][index] = fNotationListHead;
        fPublicId[chunk][index] = decl.publicId;
        fSystemId[chunk][index] = decl.systemId;
        fNotationName[chunk][index] = decl.notationName;
        fNotationListHead = fEntityCount++;
        return fNotationListHead;
    }
    public int lookupEntity(int nameIndex) {
        int chunk = 0;
        int index = 0;
        for (int entityIndex = 0; entityIndex < fEntityCount; entityIndex++) {
            if (fEntityName[chunk][index] == nameIndex)
                return entityIndex;
            if (++index == CHUNK_SIZE) {
                chunk++;
                index = 0;
            }
        }
        return -1;
    }
    public boolean isExternal(int entityIndex) {
        int chunk = entityIndex >> CHUNK_SHIFT;
        int index = entityIndex & CHUNK_MASK;
        return (fEntityValue[chunk][index] == -1);
    }
    public boolean isUnparsed(int entityIndex) {
        int chunk = entityIndex >> CHUNK_SHIFT;
        int index = entityIndex & CHUNK_MASK;
        return (fNotationName[chunk][index] != -1);
    }
    public int getEntityName(int entityIndex) {
        int chunk = entityIndex >> CHUNK_SHIFT;
        int index = entityIndex & CHUNK_MASK;
        return fEntityName[chunk][index];
    }
    public int getEntityValue(int entityIndex) {
        int chunk = entityIndex >> CHUNK_SHIFT;
        int index = entityIndex & CHUNK_MASK;
        return fEntityValue[chunk][index];
    }
    public int getPublicId(int entityIndex) {
        int chunk = entityIndex >> CHUNK_SHIFT;
        int index = entityIndex & CHUNK_MASK;
        return fPublicId[chunk][index];
    }
    public int getSystemId(int entityIndex) {
        int chunk = entityIndex >> CHUNK_SHIFT;
        int index = entityIndex & CHUNK_MASK;
        return fSystemId[chunk][index];
    }
    public int lookupNotation(int nameIndex) {
        int nIndex = fNotationListHead;
        while (nIndex != -1) {
            int chunk = nIndex >> CHUNK_SHIFT;
            int index = nIndex & CHUNK_MASK;
            if (fNotationName[chunk][index] == nameIndex)
                return nIndex;
            nIndex = fEntityValue[chunk][index];
        }
        return -1;
    }
    public int getNotationName(int entityIndex) {
        int chunk = entityIndex >> CHUNK_SHIFT;
        int index = entityIndex & CHUNK_MASK;
        return fNotationName[chunk][index];
    }

    /**
     * Checks that all of the declared unparsed entites refer
     * to declared notations. This method emits an error for each
     * unparsed entity that fails the check.
     *
     * @exception java.lang.Exception Thrown if error handler
     *                                throws exception to signal
     *                                a fatal error.
     */
    public void checkUnparsedEntities() 
        throws Exception
        {

        // check names in list
        XMLErrorHandler handler = fParserState.getErrorHandler();
        for (int i = 0; i < fEntityCount; i++) {
            int chunk = i >> CHUNK_SHIFT;
            int index = i & CHUNK_MASK;
            // #628: Notation decls don't need to be checked. -Ac
            if (fEntityName[chunk][index] != -1) {
                int notationNameIndex = fNotationName[chunk][index];
                if (notationNameIndex != -1) {
                    if (DEBUG_530) {
                        System.out.println("#530: notation name: "+
                                           fParserState.getStringPool().
                                           toString(notationNameIndex));
                        }
                    if (lookupNotation(notationNameIndex) == -1) {
                        handler.error1(ErrorCode.V_ENTITY8, notationNameIndex);
                        }
                    }
                }
            }

        } // checkUnparsedEntities()

}
