/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.tr
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */
package com.ibm.xml.internal;

import com.ibm.xml.framework.*;

import org.xml.sax.InputSource;
import org.xml.sax.Locator;

//
// REVISIT - Study up on what standalone means...
//
// REVISIT - Need to check for the following errors:
//   E_PEREF7, V_PEREF9, V_ENTITY9
//
// REVISIT - We need to mark places where EOF should generate an error
//  because we were in the middle of processing something...  otherwise
//  we will not report half-parsed tokens hitting EOF as an error...
//
// REVISIT - rewrite early code, like XMLDecl, and derivatives, like TextDecl,
//  to handle error recovery better, i.e. like newer code...
//
/**
 * Default implementation of an XML scanner.
 * <p>
 * Added ability to stop scanning. A call to the <tt>XMLScanner.stop()</tt>
 * method will set a <i>stop</i> flag, causing the scanner to stop at a
 * reasonable point. Currently, the scanner polls the stop flag at the
 * beginning of <tt>scanElementDecl()</tt> and <tt>scanElement()</tt>.
 * These two "reasonable" points are arbitrary and could occur more by
 * adding the following code to the beginning of any <tt>scanXXX</tt>
 * method, where it is desired.
 * <pre>
 *     if (fStop) {
 *         throw new StopException();
 *         }
 * </pre>
 *
 * @version Revision: 68 1.44 src/com/ibm/xml/internal/DefaultScanner.java, parser, xml4j2, xml4j2_0_13 
 */
public final class DefaultScanner implements XMLScanner, Locator {

    //
    // Constants
    //

    /** Set to true and recompile to include stop() debugging output. */
    private static final boolean DEBUG_STOP = false;

    /** Set to true to debug NMTOKEN attribute values that contain entitie refs. */
    private static final boolean DEBUG_NMTOKEN_ENTITY = false;

    //
    // Scanner state
    //
    // REVISIT - refactor this so that there is no overhead for
    // work not needed by one state versus another.
    //
    private static final int
        ST_NORMAL = 1,          // normal document processing
        ST_INTERNAL_SUBSET = 2, // internal DTD subset
        ST_EXTERNAL_SUBSET = 3; // extSubset (also used for extPE)
    //
    //
    //
    private static final byte GENERAL_ENTITY_TYPE = 0;
    private static final byte PARAMETER_ENTITY_TYPE = 1;
    //
    // [19] CDStart ::= '<![CDATA['
    // [55] StringType ::= 'CDATA'
    //
    private static final char[] cdata_string = { 'C','D','A','T','A' };
    //
    // [23] XMLDecl ::= '<?xml' VersionInfo EncodingDecl? SDDecl? S? '?>'
    // [77] TextDecl ::= '<?xml' VersionInfo? EncodingDecl S? '?>'
    //
    private static final char[] xml_string = { 'x','m','l' };
    //
    // [24] VersionInfo ::= S 'version' Eq (' VersionNum ' | " VersionNum ")
    //
    private static final char[] version_string = { 'v','e','r','s','i','o','n' };
    //
    // [28] doctypedecl ::= '<!DOCTYPE' S Name (S ExternalID)? S?
    //                      ('[' (markupdecl | PEReference | S)* ']' S?)? '>'
    //
    private static final char[] doctype_string = { 'D','O','C','T','Y','P','E' };
    //
    // [32] SDDecl ::= S 'standalone' Eq (("'" ('yes' | 'no') "'")
    //                 | ('"' ('yes' | 'no') '"'))
    //
    private static final char[] standalone_string = { 's','t','a','n','d','a','l','o','n','e' };
    private static final char[] yes_string = { 'y','e','s' };
    private static final char[] no_string = { 'n','o' };
    //
    // [45] elementdecl ::= '<!ELEMENT' S Name S contentspec S? '>'
    //
    private static final char[] element_string = { 'E','L','E','M','E','N','T' };
    //
    // [46] contentspec ::= 'EMPTY' | 'ANY' | Mixed | children
    //
    private static final char[] empty_string = { 'E','M','P','T','Y' };
    private static final char[] any_string = { 'A','N','Y' };
    //
    // [51] Mixed ::= '(' S? '#PCDATA' (S? '|' S? Name)* S? ')*'
    //                | '(' S? '#PCDATA' S? ')'
    //
    private static final char[] pcdata_string = { '#','P','C','D','A','T','A' };
    //
    // [52] AttlistDecl ::= '<!ATTLIST' S Name AttDef* S? '>'
    //
    private static final char[] attlist_string = { 'A','T','T','L','I','S','T' };
    //
    // [56] TokenizedType ::= 'ID' | 'IDREF' | 'IDREFS' | 'ENTITY' | 'ENTITIES'
    //                        | 'NMTOKEN' | 'NMTOKENS'
    //
    // Note: We search for common substrings always trying to move forward
    //
    //  'ID'      - Common prefix of ID, IDREF and IDREFS
    //  'REF'     - Common substring of IDREF and IDREFS after matching ID prefix
    //  'ENTIT'   - Common prefix of ENTITY and ENTITIES
    //  'IES'     - Suffix of ENTITIES
    //  'NMTOKEN' - Common prefix of NMTOKEN and NMTOKENS
    //
    private static final char[] id_string = { 'I','D' };
    private static final char[] ref_string = { 'R','E','F' };
    private static final char[] entit_string = { 'E','N','T','I','T' };
    private static final char[] ies_string = { 'I','E','S' };
    private static final char[] nmtoken_string = { 'N','M','T','O','K','E','N' };
    //
    // [58] NotationType ::= 'NOTATION' S '(' S? Name (S? '|' S? Name)* S? ')'
    // [82] NotationDecl ::= '<!NOTATION' S Name S (ExternalID |  PublicID) S? '>'
    //
    private static final char[] notation_string = { 'N','O','T','A','T','I','O','N' };
    //
    // [60] DefaultDecl ::= '#REQUIRED' | '#IMPLIED' | (('#FIXED' S)? AttValue)
    //
    private static final char[] required_string = { '#','R','E','Q','U','I','R','E','D' };
    private static final char[] implied_string = { '#','I','M','P','L','I','E','D' };
    private static final char[] fixed_string = { '#','F','I','X','E','D' };
    //
    // [62] includeSect ::= '<![' S? 'INCLUDE' S? '[' extSubsetDecl ']]>'
    //
    private static final char[] include_string = { 'I','N','C','L','U','D','E' };
    //
    // [63] ignoreSect ::= '<![' S? 'IGNORE' S? '[' ignoreSectContents* ']]>'
    //
    private static final char[] ignore_string = { 'I','G','N','O','R','E' };
    //
    // [71] GEDecl ::= '<!ENTITY' S Name S EntityDef S? '>'
    // [72] PEDecl ::= '<!ENTITY' S '%' S Name S PEDef S? '>'
    //
    private static final char[] entity_string = { 'E','N','T','I','T','Y' };
    //
    // [75] ExternalID ::= 'SYSTEM' S SystemLiteral
    //                     | 'PUBLIC' S PubidLiteral S SystemLiteral
    // [83] PublicID ::= 'PUBLIC' S PubidLiteral
    //
    private static final char[] system_string = { 'S','Y','S','T','E','M' };
    private static final char[] public_string = { 'P','U','B','L','I','C' };
    //
    // [76] NDataDecl ::= S 'NDATA' S Name
    //
    private static final char[] ndata_string = { 'N','D','A','T','A' };
    //
    // [80] EncodingDecl ::= S 'encoding' Eq ('"' EncName '"' |  "'" EncName "'" )
    //
    private static final char[] encoding_string = { 'e','n','c','o','d','i','n','g' };
    //
    // Debugging options
    //
    private static final boolean PRINT_EXCEPTION_STACK_TRACE = false;
    //
    // Instance Variables
    //
    private ParserState fParserState = null;
    private int fScannerState = ST_NORMAL;
    private int fScannerMarkupDepth = 0;
    private int fScannerParenDepth = 0;
    private XMLReader fReader = null;
    private boolean fStandalone = false;
    private boolean fCalledStartDocument = false;
    //
    // Handlers
    //
    private XMLDocumentHandler fDocumentHandler = null;
    private XMLDocumentTypeHandler fDocumentTypeHandler = null;
    private XMLValidationHandler fValidationHandler = null;
    private XMLEntityHandler fEntityHandler = null;
    private XMLErrorHandler fErrorHandler = null;
    //
    // Pools
    //
    private StringPool fStringPool = null;
    private AttrPool fAttrPool = null;
    private EntityPool fEntityPool = null;
    private ElementDeclPool fElementDeclPool = null;
    //
    // Literal Strings - AttValue, EntityValue, SystemLiteral, PubidLiteral
    //
    private ChunkyCharArray fLiteralData = null;
    private EntityPool fParameterEntityPool = null;
    private int fXMLSymbolIndex = -1;
    private Attr fAttr = new Attr();
    private int fAttValueIndex = -1;
    private char[] fCharRefData = null;
    private int fSystemLiteral = -1;
    private int fPubidLiteral = -1;
    private ElementDecl fElementDecl = null;
    private ContentSpecNode fContentSpecNode = null;
    private AttDef fAttDef = null;
    private NotationDecl fNotationDecl = null;
    private EntityDecl fEntityDecl = null;
    private XMLReader[] fReaderStack = new XMLReader[8];
    private InputSource[] fReaderSource = new InputSource[8];
    private int[] fReaderMarkupDepth = new int[8];
    private int[] fReaderParenDepth = new int[8];
    private int fReaderStackDepth = 0;
    private int fActiveReaderLimit = 0;
    private int[] fElementNameStack = new int[8];
    private int[] fElementChildCount = new int[8];
    private int[][] fElementChildren = new int[8][];
    private int fElementDepth = -1;
    private byte[] fEntityTypeStack = null;
    private int[] fEntityNameStack = null;
    private int fEntityStackDepth = 0;
    private byte[] opStack = null;
    private int[] nodeIndexStack = null;
    private int[] prevNodeIndexStack = null;

    // other data

    /** Stop flag. */
    private boolean fStop = false;

    //
    // Constructors
    //
    public DefaultScanner(ParserState parserState) {
        fParserState = parserState;
    }
    //
    //
    //
    public void reset(ParserState parserState) {
        fParserState = parserState;
        fStop = false;
		fCalledStartDocument = false;
    }
    //
    // Locator interface
    //
    public Locator getLocator() {
        return this;
    }
    public String getPublicId() {
        if (fReader == null)
            return null;
        return fReader.getSystemId();
    }
    public String getSystemId() {
        if (fReader == null)
            return null;
        return fReader.getSystemId();
    }
    public int getLineNumber() {
        if (fReader == null)
            return -1;
        return fReader.getLineNumber();
    }
    public int getColumnNumber() {
        if (fReader == null)
            return -1;
        return fReader.getColumnNumber();
    }
    //
    // Wrappers for ErrorHandler
    //
    private void emitError(int errorCode) throws Exception {
        fErrorHandler.error(errorCode);
    }
    private void emitError(int errorCode, int stringIndex1) throws Exception {
        fErrorHandler.error1(errorCode, stringIndex1);
    }
    private void emitError(int errorCode, int stringIndex1, int stringIndex2) throws Exception {
        fErrorHandler.error2(errorCode, stringIndex1, stringIndex2);
    }

    /**
     * Stops the scanner in a reasonable place and returns.
     */
    public void stop() {

        if (DEBUG_STOP) {
            System.out.println(
                "!! called stop() at "+
                getLineNumber()+':'+getColumnNumber()+
                " in "+getSystemId()
                );
            }

        // set flag
        fStop = true;

        } // stop()

    //
    // From the standard:
    //
    // [1] document ::= prolog element Misc*
    //
    public void scanDocument(InputSource source) throws Exception
    {
        //
        // Get handlers
        //
        fDocumentHandler = fParserState.getDocumentHandler();
        fDocumentTypeHandler = fParserState.getDocumentTypeHandler();
        fEntityHandler = fParserState.getEntityHandler();
        fErrorHandler = fParserState.getErrorHandler();
        fValidationHandler = fParserState.getValidationHandler();
        //
        // Get pools
        //
        fStringPool = fParserState.getStringPool();
        fAttrPool = fParserState.getAttrPool();
        fEntityPool = fParserState.getEntityPool();
        fElementDeclPool = fParserState.getElementDeclPool();
        //
        // Set other scanner instance variables
        //
        fScannerState = ST_NORMAL;
        fLiteralData = new ChunkyCharArray(fStringPool);
        fParameterEntityPool = null;
        fXMLSymbolIndex = -1;
        //
        // Create the reader
        //
        boolean xmlDecl = true; // xmlDecl if true, textDecl if false
		try {
			fReader = fEntityHandler.createReader(source, xmlDecl);
		} catch (java.io.UnsupportedEncodingException uee) {
			emitError(ErrorCode.E_ENC0, fStringPool.addString(uee.getMessage()));
			return; // impossible to proceed from here anyway
		} catch (Exception ex) {
            if (PRINT_EXCEPTION_STACK_TRACE)
                ex.printStackTrace();
			throw ex;
		}
        fEntityHandler.startInputSource(source);
        //
        // Be prepared to handle EOF
        //
        try
        {
            //Moved startDocument to inside
            // scanXMLDecl, and scanProlog
            //if (fDocumentHandler != null)
            //    fDocumentHandler.startDocument();

            //
            // scan prolog
            //

            // scan
            scanProlog();

            //revisit: An effort has been made to call fDocumentHandler.startDocument
            //  within the scanProlog function above, and scanXMLDecl which is called
            //  by scanProlog.
            //
            //  The following checks to see if startDocument has indeed been called,
            //  and if not, it is called here.
            //
            if (fDocumentHandler != null && !fCalledStartDocument) {
                fDocumentHandler.startDocument(-1,-1,-1);
                fCalledStartDocument = true;
            }

            // #530: check unparsed entities
            fEntityPool.checkUnparsedEntities();

            //
            // scan element
            //
            int elementNameIndex = fReader.scanName('>', -1);
            if (elementNameIndex == -1) {
                emitError(ErrorCode.E_TAG0);
            } else {
                if (fValidationHandler != null)
                    fValidationHandler.checkRootElementName(elementNameIndex);
                if (scanElement(elementNameIndex)) {
                    if (elementStackDepth() >= 0) {
                        //
                        // we have content
                        //
                        ScanContentState scanState = new ScanContentState();
                        scanState.inCDSect = false;
                        scanState.extParsedEnt = false;
                        scanState.parseTextDecl = false;
                        scanState.elementDepth = 0;
                        scanContent(scanState);
                        if (fValidationHandler != null) {
                            //
                            // Check after document is fully parsed
                            // (1) check that there was an element with a matching id for every
                            //   IDREF and IDREFS attr (V_IDREF0)
                            //
                            fValidationHandler.checkIDRefNames();
                        }
                    }
                    //
                    // scan trailing Misc?
                    //
                    scanMisc();
                }
            }
        } catch (ArrayIndexOutOfBoundsException ex) {
            // walked off the end of the data buffer

            if (elementStackDepth() >= 0) {
                emitError(ErrorCode.E_TAG3, peekElementName());
            }
            if (fScannerMarkupDepth > 0) emitError(ErrorCode.E_STRUCT0);

            if (PRINT_EXCEPTION_STACK_TRACE)
                ex.printStackTrace();
            }
        catch (StopException e) {
            if (DEBUG_STOP) {
                System.out.println("!! scanDocument(): "+e.getMessage());
                System.out.println("   reader stack depth="+fReaderStackDepth);
                }
            // let scanning end
            }
        catch (Exception ex) {
            if (PRINT_EXCEPTION_STACK_TRACE)
                ex.printStackTrace();
            throw ex;
        }

        // scanning stopped before end
        if (fStop) {
            while (fReaderStackDepth > 0) {
                // NOTE: Do not call popReader() because that notifies
                //       the registered XMLEntityHandler and the default
                //       handler will note the input source stack out
                //       of sync. -Ac
                fReaderStack[--fReaderStackDepth] = null;
                fReaderSource[fReaderStackDepth]  = null;
                }
            }

        // scanning finished properly
        else {
            if (fDocumentHandler != null && fCalledStartDocument )
                fDocumentHandler.endDocument();
            fEntityHandler.endInputSource(source);
            }

        } // scanDocument(InputSource)

    //
    // Not XML standard. For TX backward-compatibility.
    //
    // dtd ::= extSubset
    //
    public void scanDTD(InputSource source) throws Exception
    {
        //
        // Get handlers
        //
        fDocumentHandler = fParserState.getDocumentHandler();
        fDocumentTypeHandler = fParserState.getDocumentTypeHandler();
        fEntityHandler = fParserState.getEntityHandler();
        fErrorHandler = fParserState.getErrorHandler();
        fValidationHandler = fParserState.getValidationHandler();
        //
        // Get pools
        //
        fStringPool = fParserState.getStringPool();
        fAttrPool = fParserState.getAttrPool();
        fEntityPool = fParserState.getEntityPool();
        fElementDeclPool = fParserState.getElementDeclPool();
        //
        // Set other scanner instance variables
        //
        fScannerState = ST_NORMAL;
        fLiteralData = new ChunkyCharArray(fStringPool);
        fParameterEntityPool = null;
        fXMLSymbolIndex = -1;

        //
        // Create the reader
        //

        int prevState = fScannerState;
        fScannerState = ST_INTERNAL_SUBSET;

        boolean textDecl = false; // xmlDecl if true, textDecl if false
        fReader = fEntityHandler.createReader(source, textDecl);
        fEntityHandler.startInputSource(source);
        //int depth = pushReader(reader, source);
        //int oldLimit = setActiveReaderLimit(depth);


        //
        // Invoke a startDocument callback for to setup the implementation
        //

        if (fDocumentHandler != null && !fCalledStartDocument) {
            fDocumentHandler.startDocument(-1,-1,-1);
            fCalledStartDocument = true;
        }

        // revist: we don't know what the root element name is.
        int rootIndex = fStringPool.addSymbol("");
        fElementDeclPool.setRootElement(rootIndex);

        //
        // Invoke a doctypeDecl callback for to setup the implementation
        //
        if (fDocumentTypeHandler != null)
            fDocumentTypeHandler.doctypeDecl(rootIndex);


        // We think of the the DTD info in this context as an internal subset.
        if (fDocumentTypeHandler != null) {
            fDocumentTypeHandler.startInternalSubset();
        }

        //
        // Be prepared to handle EOF
        //
        try
        {
            scanExtSubsetDecl(false);

        } catch (ArrayIndexOutOfBoundsException ex) {
            // walked off the end of the data buffer

            if (fScannerMarkupDepth > 0) emitError(ErrorCode.E_STRUCT0);

            if (PRINT_EXCEPTION_STACK_TRACE)
                ex.printStackTrace();
            }
        catch (StopException e) {
            if (DEBUG_STOP) {
                System.out.println("!! scanDocument(): "+e.getMessage());
                System.out.println("   reader stack depth="+fReaderStackDepth);
                }
            // let scanning end
            }
        catch (Exception ex) {
            if (PRINT_EXCEPTION_STACK_TRACE)
                ex.printStackTrace();
            throw ex;
        }

        if (fDocumentTypeHandler != null) {
            fDocumentTypeHandler.endInternalSubset();
        }

        // check declared elements
        if (fParserState.getValidationHandler() != null) {
            fElementDeclPool.checkDeclaredElements();
        }

        // #530: check unparsed entities
        fEntityPool.checkUnparsedEntities();

        // scanning stopped before end
        if (fStop) {
            while (fReaderStackDepth > 0) {
                // NOTE: Do not call popReader() because that notifies
                //       the registered XMLEntityHandler and the default
                //       handler will note the input source stack out
                //       of sync. -Ac
                fReaderStack[--fReaderStackDepth] = null;
                fReaderSource[fReaderStackDepth]  = null;
                }
            }

        // scanning finished properly
        else {
            if (fDocumentHandler != null && fCalledStartDocument)
                fDocumentHandler.endDocument();
            fEntityHandler.endInputSource(source);
            }

        } // scanDTD(InputSource, String root)

    //
    // From the standard:
    //
    // [25] Eq ::= S? '=' S?
    //
    private boolean scanEq() throws Exception
    {
        fReader.skipPastSpaces();
        if (fReader.skippedChar('=')) {
            fReader.skipPastSpaces();
            return true;
        }
        return false;
    }
    //
    // From the standard:
    //
    // [15] Comment ::= '<!--' ((Char - '-') | ('-' (Char - '-')))* '-->'
    //
    // Called after scanning past '<!--'
    //
    private boolean scanComment(boolean inDocument) throws Exception
    {
        int commentOffset = fReader.currentOffset();
        boolean sawDashDash = false;
        while (true) {
            boolean sawDash = false;
            while (true) {
                try {
                    sawDash = fReader.lookingAtChar('-');
                    break;
                } catch (ArrayIndexOutOfBoundsException ex) {
                    if (peekEntity() == 0)
                        emitError(ErrorCode.E_COM1);
                    throw ex;
                }
            }
            if (sawDash) {
                int nextEndOffset = fReader.currentOffset();
                int endOffset = 0;
                int offset = fReader.skipAsciiChar();
                int count = 1;
                while (fReader.skippedChar('-')) {
                    count++;
                    endOffset = nextEndOffset;
                    nextEndOffset = offset;
                    offset = fReader.currentOffset();
                }
                if (count > 1) {
                    if (fReader.skippedChar('>')) {
                        if (!sawDashDash && count > 2) {
                            emitError(ErrorCode.E_COM1);
                            sawDashDash = true;
                        }
                        decreaseMarkupDepth();
                        if (inDocument && fDocumentHandler != null)
                            fDocumentHandler.comment(fReader.addString(commentOffset, endOffset - commentOffset));
                        return !sawDashDash;
                    } else if (!sawDashDash) {
                        emitError(ErrorCode.E_COM2);
                        sawDashDash = true;
                    }
                }
            } else {
                while (true) {
                    try {
                        if (!fReader.skippedValidChar())
                            fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                        break;
                    } catch (ArrayIndexOutOfBoundsException ex) {
                        if (peekEntity() == 0)
                            emitError(ErrorCode.E_COM1);
                        throw ex;
                    }
                }
            }
        }
    }
    //
    // From the standard:
    //
    // [22] prolog ::= XMLDecl? Misc* (doctypedecl Misc*)?
    // [23] XMLDecl ::= '<?xml' VersionInfo EncodingDecl? SDDecl? S? '?>'
    // [24] VersionInfo ::= S 'version' Eq (' VersionNum ' | " VersionNum ")
    //
    // The beginning of XMLDecl simplifies to:
    //    '<?xml' S ...
    //
    // [27] Misc ::= Comment | PI |  S
    // [15] Comment ::= '<!--' ((Char - '-') | ('-' (Char - '-')))* '-->'
    // [16] PI ::= '<?' PITarget (S (Char* - (Char* '?>' Char*)))? '?>'
    // [17] PITarget ::= Name - (('X' | 'x') ('M' | 'm') ('L' | 'l'))
    //
    // [28] doctypedecl ::= '<!DOCTYPE' S Name (S ExternalID)? S?
    //                      ('[' (markupdecl | PEReference | S)* ']' S?)? '>'
    //
    // That should be enough to keep us busy here...
    //
    // Returns after scanning past the '<' before the root element
    //
    private void scanProlog() throws Exception
    {
        //revisit: If there is an XMLDecl and scanXMLDecl is called, then
        // fCalledStartDocument==true. If not then we test fCalledStartDocument
        // and call fDocumentHandler.startDocument *before* any other fDocumentHandler
        // callback method is called. So we have multiple conditional tests to
        // ensure startDocument is always the first callback.
        boolean acceptXMLDecl = true;
        while (true) {
            if (fReader.skippedChar('<')) {
                increaseMarkupDepth();
                if (fReader.skippedChar('?')) {
                    int piTarget = fReader.scanName(' ', -1);
                    if (piTarget == -1) {
                        emitError(ErrorCode.E_PI0);
                        fReader.skipPastChar('>');
                        decreaseMarkupDepth();
                    } else if ("xml".equals(fStringPool.toString(piTarget))) {
                        if (fReader.skippedSpace()) { // an XMLDecl looks like a PI with the target 'xml'
                            if (acceptXMLDecl) {
                                scanXMLDecl();
                            } else {
                                emitError(ErrorCode.E_STRUCT1);
                                fReader.skipPastChar('>');
                                decreaseMarkupDepth();
                            }
                        } else { // a PI target matching 'xml'
                            emitError(ErrorCode.E_PI5);
                            fReader.skipPastChar('>');
                            decreaseMarkupDepth();
                        }
                    } else {// PI
                        if (fDocumentHandler != null && !fCalledStartDocument) {
                            fDocumentHandler.startDocument(-1,-1,-1);
                            fCalledStartDocument = true;
                        }
                        scanPI(piTarget, true);
                    }
                } else if (fReader.skippedChar('!')) {
                    if (fDocumentHandler != null && !fCalledStartDocument) {
                        fDocumentHandler.startDocument(-1,-1,-1);
                        fCalledStartDocument = true;
                    }
                    if (fReader.skippedChar('-')) { // comment ?
                        if (fReader.skippedChar('-')) {
                            scanComment(true); // scan through the closing '-->'
                        } else {
                            emitError(ErrorCode.E_COM0);
                            fReader.skipPastChar('>');
                            decreaseMarkupDepth();
                        }
                    } else if (fReader.skippedString(doctype_string))
                        scanDoctypeDecl(); // scan through the closing '>'
                    else {
                        emitError(ErrorCode.E_DOCTYPE0);
                        fReader.skipPastChar('>');
                        decreaseMarkupDepth();
                    }
                } else {// root element
                    if (fDocumentHandler != null && !fCalledStartDocument) {
                        fDocumentHandler.startDocument(-1,-1,-1);
                        fCalledStartDocument = true;
                    }
                    return;
                }
            } else if (fReader.skippedSpace())
                fReader.skipPastSpaces();
            else {
                emitError(ErrorCode.E_STRUCT0);
                fReader.skipToChar('<');
            }
            if (fDocumentHandler != null && !fCalledStartDocument) {
                fDocumentHandler.startDocument(-1,-1,-1);
                fCalledStartDocument = true;
            }
            acceptXMLDecl = false;
        }
    }
    //
    // From the standard:
    //
    // [23] XMLDecl ::= '<?xml' VersionInfo EncodingDecl? SDDecl? S? '?>'
    // [24] VersionInfo ::= S 'version' Eq (' VersionNum ' | " VersionNum ")
    // [80] EncodingDecl ::= S 'encoding' Eq ('"' EncName '"' |  "'" EncName "'" )
    // [81] EncName ::= [A-Za-z] ([A-Za-z0-9._] | '-')*
    // [32] SDDecl ::= S 'standalone' Eq (("'" ('yes' | 'no') "'")
    //                 | ('"' ('yes' | 'no') '"'))
    //
    private void scanXMLDecl() throws Exception
    {
        int versionIndex = -1;
        int encodingIndex = -1;
        int standAloneIndex = -1;

        //
        int versOffset = fReader.skipPastSpaces();
        if (!fReader.skippedString(version_string)) {
            emitError(ErrorCode.E_XML1);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        int versLength = fReader.currentOffset() - versOffset;
        if (!scanEq()) {
            emitError(ErrorCode.E_ATTEQ0, fReader.addString(versOffset, versLength));
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        boolean single;
        if (!(single = fReader.skippedChar('\'')) && !fReader.skippedChar('\"')) {
            emitError(ErrorCode.E_ATTVAL0);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        int offset = fReader.currentOffset();
        char qchar = single ? '\'' : '\"';
        while (!fReader.lookingAtChar(qchar)) {
            if (!fReader.skippedVersionNum()) {
                while (!fReader.lookingAtChar(qchar) && !fReader.lookingAtChar('<'))
                    fReader.skipOneChar();
                if (fReader.lookingAtChar('<'))
                    emitError(ErrorCode.E_ATTVAL1);
                else
                    emitError(ErrorCode.E_XML6, fReader.addString(offset, fReader.currentOffset() - offset));
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
        }
        //String version = fStringPool.orphanString(fReader.addString(offset, fReader.currentOffset() - offset));
        versionIndex = fReader.addString(offset, fReader.currentOffset() - offset);
        String version = fStringPool.toString(versionIndex);
        if (!"1.0".equals(version)) {
            emitError(ErrorCode.E_XML4);
            // return fReader.skipPastChar('>'); // REVISIT - hope it is compatible...
        }
        fReader.skipAsciiChar();
        if (!fReader.skippedSpace()) {
            if (!fReader.skippedChar('?')) {
                emitError(ErrorCode.E_XML0);
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            if (!fReader.skippedChar('>')) {
                emitError(ErrorCode.E_XML0);
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            decreaseMarkupDepth();
            //
            // Now that we have hit '?>' we are done with XML decl. Call the
            // handler before returning.
            //
            if (fDocumentHandler != null) {
                fDocumentHandler.startDocument(versionIndex,
                                     encodingIndex, standAloneIndex);
                fCalledStartDocument = true;
            }
            return;
        }
        int encOffset = fReader.skipPastSpaces();
        if (fReader.skippedString(encoding_string))
        {
            int encLength = fReader.currentOffset() - encOffset;
            if (!scanEq()) {
                emitError(ErrorCode.E_ATTEQ0, fReader.addString(encOffset, encLength));
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            if (!(single = fReader.skippedChar('\'')) && !fReader.skippedChar('\"')) {
                emitError(ErrorCode.E_ATTVAL0);
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            offset = fReader.currentOffset();
            qchar = single ? '\'' : '\"';
            if (!fReader.skippedAlpha()) {
                while (!fReader.lookingAtChar(qchar) && !fReader.lookingAtChar('<'))
                    fReader.skipOneChar();
                if (fReader.lookingAtChar('<'))
                    emitError(ErrorCode.E_ATTVAL1);
                else
                    emitError(ErrorCode.E_PI4, fReader.addString(offset, fReader.currentOffset() - offset));
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            while (!fReader.lookingAtChar(qchar)) {
                if (!fReader.skippedEncName()) {
                    while (!fReader.lookingAtChar(qchar) && !fReader.lookingAtChar('<'))
                        fReader.skipOneChar();
                    if (fReader.lookingAtChar('<'))
                        emitError(ErrorCode.E_ATTVAL1);
                    else
                        emitError(ErrorCode.E_PI4, fReader.addString(offset, fReader.currentOffset() - offset));
                    fReader.skipPastChar('>');
                    decreaseMarkupDepth();
                    return;
                }
            }
            //
            // Add the encoding string to the string pool.
            //
            encodingIndex =
                fReader.addString(offset, fReader.currentOffset() - offset);
            fReader.skipAsciiChar();
            if (!fReader.skippedSpace()) {
                if (!fReader.skippedChar('?')) {
                    emitError(ErrorCode.E_XML0);
                    fReader.skipPastChar('>');
                    decreaseMarkupDepth();
                    return;
                }
                if (!fReader.skippedChar('>')) {
                    emitError(ErrorCode.E_XML0);
                    fReader.skipPastChar('>');
                    decreaseMarkupDepth();
                    return;
                }
                decreaseMarkupDepth();
                //
                // Now that we have hit '?>' we are done with XML decl. Call the
                // handler before returning.
                //
                if (fDocumentHandler != null) {
                    fDocumentHandler.startDocument(versionIndex,
                                         encodingIndex, standAloneIndex);
                    fCalledStartDocument = true;
                }
                return;
            }
        }
        int sddeclOffset = fReader.skipPastSpaces();
        if (fReader.skippedString(standalone_string))
        {
            int sddeclLength = fReader.currentOffset() - sddeclOffset;
            if (!scanEq()) {
                emitError(ErrorCode.E_ATTEQ0, fReader.addString(sddeclOffset, sddeclLength));
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            if (!(single = fReader.skippedChar('\'')) && !fReader.skippedChar('\"')) {
                emitError(ErrorCode.E_ATTVAL0);
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            offset = fReader.currentOffset();
            qchar = single ? '\'' : '\"';
            boolean yes;
            if ((!(yes = fReader.skippedString(yes_string)) && !fReader.skippedString(no_string)) || !fReader.lookingAtChar(qchar))
            {
                while (!fReader.lookingAtChar(qchar) && !fReader.lookingAtChar('<'))
                    fReader.skipOneChar();
                if (fReader.lookingAtChar('<'))
                    emitError(ErrorCode.E_ATTVAL1);
                else
                    emitError(ErrorCode.E_PI4, fReader.addString(offset, fReader.currentOffset() - offset));
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            fStandalone = yes;
            standAloneIndex = fReader.addString(offset, fReader.currentOffset() - offset);
            fReader.skipAsciiChar();
        }
        fReader.skipPastSpaces();
        if (!fReader.skippedChar('?')) {
            emitError(ErrorCode.E_XML0);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        if (!fReader.skippedChar('>')) {
            emitError(ErrorCode.E_XML0);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        decreaseMarkupDepth();
        //
        // Call the handler here to pass on XML Decl information just parsed.
        //
        if (fDocumentHandler != null) {
            fDocumentHandler.startDocument(versionIndex,
                                 encodingIndex, standAloneIndex);
            fCalledStartDocument = true;
        }
    }
    //
    // From the standard:
    //
    // [39] element ::= EmptyElemTag | STag content ETag
    // [44] EmptyElemTag ::= '<' Name (S Attribute)* S? '/>'
    // [40] STag ::= '<' Name (S Attribute)* S? '>'
    // [41] Attribute ::= Name Eq AttValue
    // [10] AttValue ::= '"' ([^<&"] | Reference)* '"'
    //                   | "'" ([^<&'] | Reference)* "'"
    // [67] Reference ::= EntityRef | CharRef
    // [68] EntityRef ::= '&' Name ';'
    // [66] CharRef ::= '&#' [0-9]+ ';' | '&#x' [0-9a-fA-F]+ ';'
    // [43] content ::= (element | CharData | Reference | CDSect | PI | Comment)*
    // [42] ETag ::= '</' Name S? '>'
    //
    // Note: We have already scanned Name.
    //
    private boolean scanElement(int elementNameIndex) throws Exception
    {
        // are we stopped?
        if (fStop) {
            throw new StopException();
            }

        //
        // Scan for attributes
        //
        boolean greater = false;
        int firstAttrIndex = -1;
        int lastAttrIndex = -1;
        if (fReader.skippedSpace()) {
            while (true) {
                fReader.skipPastSpaces();
                //
                // [41] Attribute ::= Name Eq AttValue
                //
                if ((greater = fReader.skippedChar('>')) || fReader.skippedChar('/'))
                    break;
                //
                // Name
                //
                int attrNameIndex = fReader.scanName('=', -1);
                if (attrNameIndex == -1) {
                    emitError(ErrorCode.E_TAG1);
                    fReader.skipPastChar('>');
                    decreaseMarkupDepth();
                    return false;
                }
                //
                // Eq
                //
                if (!scanEq()) {
                    emitError(ErrorCode.E_ATTEQ0, attrNameIndex);
                    fReader.skipPastChar('>');
                    decreaseMarkupDepth();
                    return false;
                }
                //
                // Check that attr has attr decl (V_TAG6)
                //
                int attDef = fElementDeclPool.getAttDef(elementNameIndex, attrNameIndex);
                int attType;
                boolean fixed = false;
                if (attDef != -1) {
                    attType = fElementDeclPool.getAttType(attDef);
                    fixed = fElementDeclPool.getAttDefaultType(attDef) == AttDef.FIXED;
                } else {
                    if (fValidationHandler != null) {
                        // REVISIT - cache the elem/attr tuple so that we only give
                        //  this error once for each unique occurrence
                        emitError(ErrorCode.V_TAG6, attrNameIndex, elementNameIndex);
                    }
                    attType = AttDef.CDATA;
                }
                //
                // REVISIT - determine when saveData can be false
                //
                if (!scanAttValue(elementNameIndex, attrNameIndex, attType, true, fixed, false)) {
                    fReader.skipPastChar('>');
                    decreaseMarkupDepth();
                    return false;
                }
                //
                //
                //
                fAttr.attName = attrNameIndex;
                fAttr.attType = attType;
                fAttr.attValue = fAttValueIndex;
                fAttr.specified = true;
                int attrIndex = fAttrPool.addAttr(fAttr, firstAttrIndex);
                if (attrIndex != -1) {
                    lastAttrIndex = attrIndex;
                    if (firstAttrIndex == -1)
                        firstAttrIndex = attrIndex;
                }
            }
        }
        else if (!(greater = fReader.skippedChar('>')) && !fReader.skippedChar('/'))
        {
            emitError(ErrorCode.E_TAG1);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return false;
        }
        //
        // Check after all specified attrs are scanned
        // (1) report error for REQUIRED attrs that are missing (V_TAGc)
        // (2) add default attrs (FIXED and NOT_FIXED)
        //
        int elementIndex = fElementDeclPool.addElement(elementNameIndex);
        firstAttrIndex = fElementDeclPool.addDefaultAttributes(elementIndex, fAttrPool, firstAttrIndex, lastAttrIndex);
        if (fValidationHandler != null)
            fValidationHandler.checkAttributes(elementIndex, firstAttrIndex);
        if (fDocumentHandler != null)
            fDocumentHandler.startElement(elementNameIndex, firstAttrIndex);
        else
            fAttrPool.releaseAttrList(firstAttrIndex);
        if (!greater) {
            // '/>'
            if (!fReader.skippedChar('>')) {
                emitError(ErrorCode.E_TAG1);
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return false;
            }
            decreaseMarkupDepth();
            pushElementName(elementNameIndex);
            if (fDocumentHandler != null)
                fDocumentHandler.endElement(elementNameIndex);
            if (fValidationHandler != null) {
                int childCount = peekChildCount();
                int result = fValidationHandler.checkContent(elementIndex,
                                                             childCount,
                                                             peekChildren());
                if (result != -1) {
                    String cs = fElementDeclPool.getContentSpecAsString(elementIndex);
                    int errorCode = result != childCount ?
                                    ErrorCode.V_CONT1 :
                                    ErrorCode.V_CONT2 ;
                    emitError(errorCode, elementNameIndex,
                              fStringPool.addString(cs));
                }
            }
            popElementName();
            return true;
        }
        decreaseMarkupDepth();
        pushElementName(elementNameIndex);
        return true;
    }
    //
    // From the standard:
    //
    // [43] content ::= (element | CharData | Reference | CDSect | PI | Comment)*
    // [78] extParsedEnt ::= TextDecl? content
    //
    // If extParsedEnt is true, then we are being called to parse an external entity.
    //
    private void scanContent(ScanContentState scanState) throws Exception
    {
        while (true) {
            int result = fReader.scanContent(scanState);
            if ((result & XMLReader.CHARDATA_RESULT_CHARDATA) != 0) {
                if (fValidationHandler != null && (result & XMLReader.CHARDATA_RESULT_ALL_SPACE) == 0)
                    charDataInContent();
                scanState.parseTextDecl = false;
            }
            switch (result & XMLReader.CHARDATA_RESULT_STATE_MASK) {
            case XMLReader.CHARDATA_RESULT_MARKUP:
            {
                //
                // Allowed to be (element | CDSect | PI | Comment)
                // If !extParsedEnt, may also be ETag
                //
                int oldDepth = increaseMarkupDepth();
                int elementNameIndex = fReader.scanName('>', -1);
                if (elementNameIndex != -1) {
                    //
                    // element
                    //
                    scanElement(elementNameIndex);
                } else if (fReader.skippedChar('/')) {
                    //
                    // [42] ETag ::= '</' Name S? '>'
                    //
                    if (scanState.extParsedEnt && elementStackDepth() == scanState.elementDepth)
                        emitError(ErrorCode.E_TAGg);
                    elementNameIndex = peekElementName();
                    int tagNameIndex = fReader.scanName('>', elementNameIndex);
                    if (tagNameIndex == -1) {
                        emitError(ErrorCode.E_TAG3, elementNameIndex);
                        fReader.skipPastChar('>');
                        decreaseMarkupDepth();
                    } else {
                        if (!fReader.skippedChar('>')) {
                            fReader.skipPastSpaces();
                            if (!fReader.skippedChar('>'))
                                emitError(ErrorCode.E_TAG4, tagNameIndex);
                        }
                        decreaseMarkupDepth();
                        if (fDocumentHandler != null)
                            fDocumentHandler.endElement(elementNameIndex);
                        if (fValidationHandler != null) {
                            int elementIndex = fElementDeclPool.getElement(elementNameIndex);
                            int childCount = peekChildCount();
                            result = fValidationHandler.checkContent(elementIndex,
                                                                     childCount,
                                                                     peekChildren());
                            if (result != -1) {
                                String cs = fElementDeclPool.getContentSpecAsString(elementIndex);
                                int errorCode = result != childCount ?
                                                ErrorCode.V_CONT1 :
                                                ErrorCode.V_CONT2 ;
                                emitError(errorCode, elementNameIndex,
                                          fStringPool.addString(cs));
                            }
                        }
                        if (!popElementName())
                            return;
                    }
                } else if (fReader.skippedChar('!')) {
                    if (fReader.skippedChar('-')) {
                        //
                        // [15] Comment ::= '<!--' ((Char - '-') | ('-' (Char - '-')))* '-->'
                        //
                        if (fReader.skippedChar('-')) {
                            scanComment(true); // scan through the closing '-->'
                        } else {
                            emitError(ErrorCode.E_COM0);
                            fReader.skipPastChar('>');
                            decreaseMarkupDepth();
                        }
                    } else if (fReader.skippedChar('[')) {
                        //
                        // [18] CDSect ::= CDStart CData CDEnd
                        // [19] CDStart ::= '<![CDATA['
                        // [20] CData ::= (Char* - (Char* ']]>' Char*))
                        // [21] CDEnd ::= ']]>'
                        //
                        if (fReader.skippedString(cdata_string)) {
                            if (fReader.skippedChar('[')) {
                                scanState.inCDSect = true;
                                decreaseMarkupDepth(); // REVISIT - this should be called by the reader scanContent method...
                            } else {
                                emitError(ErrorCode.E_CDATA0);
                                fReader.skipPastChar('>');
                                decreaseMarkupDepth();
                            }
                        } else {
                            emitError(ErrorCode.E_CDATA0);
                            fReader.skipPastChar('>');
                            decreaseMarkupDepth();
                        }
                    } else {
                        emitError(ErrorCode.E_EXCLAM0);
                        fReader.skipPastChar('>');
                        decreaseMarkupDepth();
                    }
                } else if (fReader.skippedChar('?')) {
                    //
                    // [16] PI ::= '<?' PITarget (S (Char* - (Char* '?>' Char*)))? '?>'
                    // [17] PITarget ::= Name - (('X' | 'x') ('M' | 'm') ('L' | 'l'))
                    //
                    int piTarget = fReader.scanName(' ', -1);
                    if (piTarget == -1) {
                        emitError(ErrorCode.E_PI0);
                        fReader.skipPastChar('>');
                        decreaseMarkupDepth();
                    } else if ("xml".equals(fStringPool.toString(piTarget))) {
                        if (fReader.skippedSpace()) { // a TextDecl looks like a PI with the target 'xml'
                            if (scanState.parseTextDecl) {
                                scanTextDecl();
                            } else {
                                emitError(ErrorCode.E_STRUCT1);
                                fReader.skipPastChar('>');
                                decreaseMarkupDepth();
                            }
                        } else { // a PI target matching 'xml'
                            emitError(ErrorCode.E_PI5);
                            fReader.skipPastChar('>');
                            decreaseMarkupDepth();
                        }
                    } else // PI
                        scanPI(piTarget, true);
                } else
                    emitError(ErrorCode.E_TAG0);
                break;
            }
            case XMLReader.CHARDATA_RESULT_REFERENCE:
            {
                //
                // [67] Reference ::= EntityRef | CharRef
                // [68] EntityRef ::= '&' Name ';'
                // [66] CharRef ::= '&#' [0-9]+ ';' | '&#x' [0-9a-fA-F]+ ';'
                //
                referenceInContent(fReader.currentOffset());
                break;
            }
            case XMLReader.CHARDATA_RESULT_CDEND:
            {
                //
                // [14] CharData ::= [^<&]* - ([^<&]* ']]>' [^<&]*)
                // [21] CDEnd ::= ']]>'
                //
                emitError(ErrorCode.E_TAGj);
                break;
            }
            case XMLReader.CHARDATA_RESULT_INVALID_CHAR:
            {
                //
                // Need to check for EOF for extParsedEnt
                //
                try {
                    fReader.lookingAtValidChar();
                } catch (ArrayIndexOutOfBoundsException ex) {
                    if (scanState.inCDSect)
                        emitError(ErrorCode.E_CDATA1);
                    throw ex;
                }
                //
                //  [2] Char ::= #x9 | #xA | #xD | [#x20-#xD7FF]        // any Unicode character, excluding the
                //               | [#xE000-#xFFFD] | [#x10000-#x10FFFF] // surrogate blocks, FFFE, and FFFF.
                //
                fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                break;
            }
            default:
                //
                // Not possible, we handle all enum values above...  Need some sort of
                // internal logic error exception...
                //
            }
            scanState.parseTextDecl = false;
        }
    }
    //
    // [10] AttValue ::= '"' ([^<&"] | Reference)* '"'
    //                   | "'" ([^<&'] | Reference)* "'"
    //
    private boolean scanAttValueCDATA(boolean saveData, boolean fixed) throws Exception
    {
        boolean single;
        if (!(single = fReader.skippedChar('\'')) && !fReader.skippedChar('\"')) {
            emitError(ErrorCode.E_ATTVAL0);
            return false;
        }
        char qchar = single ? '\'' : '\"';
        int offset = fReader.currentOffset();
        int toCopy = offset;
        int dataOffset = saveData ? fLiteralData.length() : 0;
        int readerDepth = 0;
        while (true) {
            boolean skippedChar = false;
            boolean poppedReader = false;
            if (readerDepth == 0) {
                while (true) {
                    try {
                        skippedChar = fReader.skippedChar(qchar);
                        break;
                    } catch (ArrayIndexOutOfBoundsException ex) {
                        //  REVISIT - readerDepth is 0 so why are we at an end of entity ?
                        if (saveData && offset - toCopy > 0)
                            fReader.append(fLiteralData, toCopy, offset - toCopy);
                        popReader();
                        popEntity();
                        offset = fReader.currentOffset();
                        toCopy = offset;
                    }
                }
                if (skippedChar) {
                    if (saveData) {
                        int dataLength = fLiteralData.length() - dataOffset;
                        if (dataLength == 0) {
                            if (fixed)
                                fAttValueIndex = fReader.addSymbol(toCopy, offset - toCopy);
                            else
                                fAttValueIndex = fReader.addString(toCopy, offset - toCopy);
                        } else {
                            if (offset - toCopy > 0) {
                                fReader.append(fLiteralData, toCopy, offset - toCopy);
                                dataLength = fLiteralData.length() - dataOffset;
                            }
                            if (fixed)
                                fAttValueIndex = fLiteralData.addSymbol(dataOffset, dataLength);
                            else
                                fAttValueIndex = fLiteralData.addString(dataOffset, dataLength);
                        }
                    }
                    return true;
                }
            }
            while (true) {
                try {
                    skippedChar = fReader.skippedChar(' ');
                    break;
                } catch (ArrayIndexOutOfBoundsException ex) {
                    if (saveData && offset - toCopy > 0)
                        fReader.append(fLiteralData, toCopy, offset - toCopy);
                    popReader();
                    popEntity();
                    readerDepth--;
                    poppedReader = true;
                    offset = fReader.currentOffset();
                    toCopy = offset;
                    break;
                }
            }
            if (poppedReader)
                continue;
            if (skippedChar) {
                offset = fReader.currentOffset();
                continue;
            }
            // REVISIT - comment this out for J. Clark tests.  Check the spec...
//            boolean skippedCR;
//            if ((skippedCR = fReader.skippedChar((char)0x0D)) || fReader.skippedSpace()) {
            if (fReader.skippedSpace()) {
                if (saveData) {
                    if (offset - toCopy > 0)
                        fReader.append(fLiteralData, toCopy, offset - toCopy);
                    fLiteralData.append(' ');
                }
//                if (skippedCR)
//                    fReader.skippedChar((char)0x0A);
                offset = fReader.currentOffset();
                toCopy = offset;
                continue;
            }
            if (fReader.skippedChar('&')) {
                if (saveData && offset - toCopy > 0)
                    fReader.append(fLiteralData, toCopy, offset - toCopy);
                //
                // Check for character reference first.
                //
                if (fReader.skippedChar('#')) {
                    int ch = scanCharRef();
                    if (ch == -1) {
                        while (readerDepth-- > 0) {
                            popReader();
                            popEntity();
                        }
                        fReader.skipPastChar(qchar);
                        return false;
                    }
                    if (saveData) {
                        if (ch < 0x10000)
                            fLiteralData.append((char)ch);
                        else {
                            fLiteralData.append((char)(((ch-0x00010000)>>10)+0xd800));
                            fLiteralData.append((char)(((ch-0x00010000)&0x3ff)+0xdc00));
                        }
                    }
                    offset = fReader.currentOffset();
                    toCopy = offset;
                    continue;
                }
                //
                // Entity reference
                //
                int nameOffset = fReader.currentOffset();
                int nameLength = fReader.skipPastName(';') - nameOffset;
                if (nameLength == 0) {
                    fReader.skipInvalidChar(ErrorCode.E_REFER0);
                    while (readerDepth-- > 0) {
                        popReader();
                        popEntity();
                    }
                    fReader.skipPastChar(qchar);
                    return false;
                }
                if (!fReader.skippedChar(';')) {
                    emitError(ErrorCode.E_REFER1);
                    while (readerDepth-- > 0) {
                        popReader();
                        popEntity();
                    }
                    fReader.skipPastChar(qchar);
                    return false;
                }
                int entityNameIndex = fReader.addSymbol(nameOffset, nameLength);
                int entityIndex = fEntityPool.lookupEntity(entityNameIndex);
                if (entityIndex < 0) {
                    emitError(ErrorCode.E_ENTITY0, entityNameIndex);
                    while (readerDepth-- > 0) {
                        popReader();
                        popEntity();
                    }
                    fReader.skipPastChar(qchar);
                    return false;
                }
                if (fEntityPool.isExternal(entityIndex)) {
                    emitError(ErrorCode.E_ENTITY7, entityNameIndex);
                    while (readerDepth-- > 0) {
                        popReader();
                        popEntity();
                    }
                    fReader.skipPastChar(qchar);
                    return false;
                }
                if (!pushEntity(GENERAL_ENTITY_TYPE, entityNameIndex)) {
                    while (readerDepth-- > 0) {
                        popReader();
                        popEntity();
                    }
                    fReader.skipPastChar(qchar);
                    return false;
                }
                StringReader reader = new StringReader(fParserState, false, fReader, fEntityPool.getEntityValue(entityIndex));
                pushReader(reader, null);
                readerDepth++;
                offset = fReader.currentOffset();
                toCopy = offset;
                continue;
            }
            if (fReader.skippedChar('<')) {
                emitError(ErrorCode.E_ATTVAL1);
                while (readerDepth-- > 0) {
                    popReader();
                    popEntity();
                }
                fReader.skipPastChar(qchar);
                return false;
            }
            if (!fReader.skippedValidChar()) {
                fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                while (readerDepth-- > 0) {
                    popReader();
                    popEntity();
                }
                fReader.skipPastChar(qchar);
                return false;
            }
            offset = fReader.currentOffset();
        }
    }
    //
    // Copy an AttValue to the literal data buffer, resolving all references
    // recursively until we hit the end quote delimiter.  We were called after
    // scanning past any whitespace, and after checking for the close quote.
    //
    private boolean scanNormalizedAttValue(char qchar) throws Exception {
        if (DEBUG_NMTOKEN_ENTITY) {
            System.out.println("scanNormalizedAttValue()");
        }
        int readerDepth = 0;
        int dataOffset = fLiteralData.length();
        while (true) {
            //
            // Check the most likely case first.  An entity reference.
            //
            try {
                if (fReader.skippedChar('&')) {
                    //
                    // Check for character reference first.
                    //
                    if (fReader.skippedChar('#')) {
                        int ch = scanCharRef();
                        if (ch == -1) {
                            fReader.skipPastChar(qchar);
                            return false;
                        }
                        if (ch < 0x10000)
                            fLiteralData.append((char)ch);
                        else {
                            fLiteralData.append((char)(((ch-0x00010000)>>10)+0xd800));
                            fLiteralData.append((char)(((ch-0x00010000)&0x3ff)+0xdc00));
                        }
                        continue;
                    }
                    //
                    // Entity reference
                    //
                    int nameOffset = fReader.currentOffset();
                    int nameLength = fReader.skipPastName(';') - nameOffset;
                    if (nameLength == 0) {
                        fReader.skipInvalidChar(ErrorCode.E_REFER0);
                        fReader.skipPastChar(qchar);
                        return false;
                    }
                    if (!fReader.skippedChar(';')) {
                        emitError(ErrorCode.E_REFER1);
                        fReader.skipPastChar(qchar);
                        return false;
                    }
                    int entityNameIndex = fReader.addSymbol(nameOffset, nameLength);
                    int entityIndex = fEntityPool.lookupEntity(entityNameIndex);
                    if (entityIndex < 0) {
                        emitError(ErrorCode.E_ENTITY0, entityNameIndex);
                        fReader.skipPastChar(qchar);
                        return false;
                    }
                    if (fEntityPool.isExternal(entityIndex)) {
                        emitError(ErrorCode.E_ENTITY7, entityNameIndex);
                        fReader.skipPastChar(qchar);
                        return false;
                    }
                    if (!pushEntity(GENERAL_ENTITY_TYPE, entityNameIndex)) {
                        fReader.skipPastChar(qchar);
                        return false;
                    }
                    if (DEBUG_NMTOKEN_ENTITY) {
                        System.out.println(
                            "  GENERAL_ENTITY_TYPE: "+
                            fStringPool.toString(entityNameIndex));
                    }
                    StringReader reader = new StringReader(fParserState, false, fReader, fEntityPool.getEntityValue(entityIndex));
                    pushReader(reader, null);
                    readerDepth++;
                    continue;
                }
                if (fReader.skippedSpace()) {
                    fReader.skipPastSpaces();
                    fLiteralData.append(' ');
                    continue;
                }
                if (fReader.lookingAtChar(qchar)) {
                    if (readerDepth == 0) {
                        fReader.skipAsciiChar();
                        fLiteralData.append(qchar);
                        return true;
                    }
                    fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                    while (readerDepth-- > 0) {
                        popReader();
                        popEntity();
                    }
                    fReader.skipPastChar(qchar);
                    return false;
                } else {
                    int offset = fReader.currentOffset();
                    int length = fReader.skipOneChar() - offset;
                    fReader.append(fLiteralData, offset, length);
                }
            }
            catch (ArrayIndexOutOfBoundsException e) {
                popReader();
                readerDepth--;
                popEntity();
                continue;
            }
        }
    }
    private boolean scanAttValue(int elementNameIndex, int attrNameIndex, int attType, boolean saveData, boolean fixed, boolean dtd) throws Exception
    {
        if (DEBUG_NMTOKEN_ENTITY) {
            System.out.println(
                "scanAttValue("+
                fStringPool.toString(elementNameIndex)+','+
                fStringPool.toString(attrNameIndex)+','+
                attType+','+
                saveData+','+
				fixed+','+
                dtd+')');
        }
        //
        // If the AttType is CDATA, use the "optimized" code to scan the AttValue.
        //
        // REVISIT - The CDATA code is terribly inefficient.
        //  It needs to be rewritten later...
        //
        if (attType == AttDef.CDATA)
            return scanAttValueCDATA(saveData, fixed);
        //
        // We will try to read the other AttValue types in place, but will copy to
        // the literal data buffer if we have to.
        //
        boolean single;
        if (!(single = fReader.skippedChar('\'')) && !fReader.skippedChar('\"')) {
			if (!dtd) emitError(ErrorCode.E_ATTVAL0);
            return false;
        }
        char qchar = single ? '\'' : '\"';
        boolean onlyone = true;
        boolean justone =
            (attType != AttDef.IDREFS && attType != AttDef.ENTITIES && attType != AttDef.NMTOKENS);
        boolean getname =
            (attType != AttDef.NMTOKEN && attType != AttDef.NMTOKENS && attType != AttDef.ENUMERATION);
        //
        // We will only need to use the slow path once.
        //
        boolean alreadySlow = false;
        boolean usingLiteralData = false;
        //
        // Skip leading whitespace
        //
        int toCopy = fReader.skipPastSpaces();
        int dataOffset = fLiteralData.length();
        while (true) {
            int offset = fReader.currentOffset();
            //
            // We do not need special checks to pop the reader here because we are either making
            // the first pass, and the entire AttValue is required to be contained within any
            // reference we are scanning, or we are on the slow path, in which case we already
            // normalized the value through the end delimiter and will not reach the end of the
            // reader until we scan past the entire AttValue.
            //
            int length = (getname ? fReader.skipPastName(qchar) : fReader.skipPastNmtoken(qchar)) - offset;
            if (length == 0) {
                if (alreadySlow) {
                    fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                    popReader(); // this is the same as fReader.skipPastChar(qchar)
                    return false;
                }
                if (fReader.skippedChar(qchar)) {
                    //
                    // The AttValue was empty.
                    //
                    if (!dtd) emitError(ErrorCode.E_TAGl, attrNameIndex);
                    return false;
                }
                //
                // Lets try to reduce the complexity a little bit here.  Copy the replacement text
                // for the attribute value to the literal data buffer, expanding any additional
                // references until we hit the end quote.  Then we will push a reader with the
                // contents of the literal data buffer and scan the rest of the AttValue without
                // needing to check for additional or nested references.  Okay, so maybe we are
                // still pretty complex...
                //
                if (offset - toCopy > 0)
                    fReader.append(fLiteralData, toCopy, offset - toCopy);
                toCopy = -1; // paranoid
                int lastDataOffset = fLiteralData.length();
                if (!scanNormalizedAttValue(qchar))
                    return false;
                int literalDataString = fLiteralData.addString(lastDataOffset, fLiteralData.length() - lastDataOffset);
                StringReader reader = new StringReader(fParserState, false, fReader, literalDataString);
                pushReader(reader, null);
                alreadySlow = true;
                //
                // This part is tricky.  We are going to be reading from "an entity" that is using
                // the literal data buffer.  We are now going to set the length of the literal data
                // buffer back to the point where we were last scanning.  We will be appending the
                // normalized result on top on the same place we are reading from.  We can do this
                // because the remaining normalization will only make the result "shorter" than the
                // data we scan and because the StringReader maintains its own copy of the offset
                // and length information and the getChar() method in ChunkyCharArray does not check
                // for reading past the point where we are appending.  This is as designed.
                //
                fLiteralData.setLength(lastDataOffset);
                offset = fReader.skipPastSpaces();
                if (fReader.skippedChar(qchar)) {
                    //
                    // The rest of the value was just whitespace.  If we found a value earlier, we
                    // are done, otherwise the entire value was whitespace, which is an error.
                    //
                    if (onlyone) {
                        if (!dtd) emitError(ErrorCode.E_TAGl, attrNameIndex);
                        popReader();
                        return false;
                    }
                    if (fixed)
                        fAttValueIndex = fLiteralData.addSymbol(dataOffset, lastDataOffset - dataOffset);
                    else
                        fAttValueIndex = fLiteralData.addString(dataOffset, lastDataOffset - dataOffset);
                    popReader();
                    return true;
                }
                length = (getname ? fReader.skipPastName(qchar) : fReader.skipPastNmtoken(qchar)) - offset;
                if (length == 0) {
                    fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                    popReader(); // this is the same as fReader.skipPastChar(qchar)
                    return false;
                }
            }
            boolean atend;
            int value;
            if (alreadySlow) {
                //
                // We are on the slow path, so we need to update the literal data buffer.
                //
                int valueOffset = fLiteralData.length();
                fReader.append(fLiteralData, offset, length);
                int valueEnd = fLiteralData.length();
                //
                // This is dumb, we should be able to normalize the whitespace in the
                // scanNormalizedAttValue method, but with the entity evaluation it
                // is tricky and we might mess up.  Better to be safe than sorry...
                //
                boolean keepgoing = fReader.skippedSpace();
                if (keepgoing)
                    fReader.skipPastSpaces();
                atend = fReader.skippedChar(qchar);
                if (atend) {
                    //
                    // This is the final token in the AttValue.  Save the results for the last
                    // constraint check, and pop the reader we have been using.
                    //
                    value = fLiteralData.addSymbol(valueOffset, valueEnd - valueOffset);
                    if (onlyone)
                        fAttValueIndex = value;
                    else if (fixed)
                        fAttValueIndex = fLiteralData.addSymbol(dataOffset, valueEnd - dataOffset);
                    else
                        fAttValueIndex = fLiteralData.addString(dataOffset, valueEnd - dataOffset);
                    popReader();
                } else if (keepgoing) {
                    //
                    // We have more tokens in the value.  Update the partial results and add a
                    // single space to the data buffer.
                    //
                    if (justone) {
                        fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                        popReader();
                        return false;
                    }
                    value = fLiteralData.addSymbol(valueOffset, valueEnd - valueOffset);
                    fLiteralData.append(' ');
                    offset = fReader.currentOffset();
                } else {
                    fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                    popReader();
                    return false;
                }
            } else {
                //
                // We are still on the fast path.  If we are lucky the only whitespace will be that
                // of single space characters and we will not need to make any changes for the
                // AttValue normalization.
                //
                boolean needcopy = false;
                boolean keepgoing = fReader.skippedChar(' ');
                if (fReader.skippedSpace()) {
                    fReader.skipPastSpaces();
                    keepgoing = true;
                    needcopy = true;
                }
                atend = fReader.skippedChar(qchar);
                if (atend) {
                    //
                    // This is the final token in the AttValue.  Save the results for the last
                    // constraint check.
                    //
                    if (usingLiteralData) {
                        int valueOffset = fLiteralData.length();
                        fReader.append(fLiteralData, offset, length);
                        int valueEnd = fLiteralData.length();
                        value = fLiteralData.addSymbol(valueOffset, valueEnd - valueOffset);
                        if (fixed)
                            fAttValueIndex = fLiteralData.addSymbol(dataOffset, valueEnd - dataOffset);
                        else
                            fAttValueIndex = fLiteralData.addString(dataOffset, valueEnd - dataOffset);
                    } else {
                        value = fReader.addSymbol(offset, length);
                        if (onlyone)
                            fAttValueIndex = value;
                        else if (fixed)
                            fAttValueIndex = fReader.addSymbol(toCopy, offset + length - toCopy);
                        else
                            fAttValueIndex = fReader.addString(toCopy, offset + length - toCopy);
                    }
                } else if (keepgoing) {
                    //
                    // We have more tokens in the value.  Update the partial results and add a
                    // single space to the data buffer.
                    //
                    if (justone && !atend) {
                        fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                        fReader.skipPastChar(qchar);
                        return false;
                    }
                    if (usingLiteralData) {
                        int valueOffset = fLiteralData.length();
                        fReader.append(fLiteralData, offset, length);
                        int valueEnd = fLiteralData.length();
                        value = fLiteralData.addSymbol(valueOffset, valueEnd - valueOffset);
                        fLiteralData.append(' ');
                    } else if (needcopy) {
                        //
                        // We need to normalize the whitespace, so we need to copy the value
                        // scanned so far to the literal data buffer and update our offsets.
                        //
                        if (offset - toCopy > 0)
                            fReader.append(fLiteralData, toCopy, offset - toCopy);
                        toCopy = -1; // paranoid
                        int valueOffset = fLiteralData.length();
                        fReader.append(fLiteralData, offset, length);
                        int valueEnd = fLiteralData.length();
                        value = fLiteralData.addSymbol(valueOffset, valueEnd - valueOffset);
                        fLiteralData.append(' ');
                        usingLiteralData = true;
                    } else {
                        //
                        // Still on the fast path, there was a single space in the AttValue.
                        //
                        value = fReader.addSymbol(offset, length);
                    }
                } else {
                    //
                    // Lets try to reduce the complexity a little bit here.  Copy the replacement text
                    // for the attribute value to the literal data buffer, expanding any additional
                    // references until we hit the end quote.  Then we will push a reader with the
                    // contents of the literal data buffer and scan the rest of the AttValue without
                    // needing to check for additional or nested references.  Okay, so maybe we are
                    // still pretty complex...
                    //
                    if (offset - toCopy > 0)
                        fReader.append(fLiteralData, toCopy, offset - toCopy);
                    toCopy = -1; // paranoid
                    int lastDataOffset = fLiteralData.length();
                    //
                    // We have scanned a partial value, but there could be a reference that will add
                    // to that value, i.e. foo="b&ar;" where entity 'ar' is replaced with "ar".  We
                    // will start our copy back at the point before we scanned the token, which we
                    // will scan again after we expand references.
                    //
                    fReader.append(fLiteralData, offset, length);
                    if (!scanNormalizedAttValue(qchar))
                        return false;
                    int literalDataString = fLiteralData.addString(lastDataOffset, fLiteralData.length() - lastDataOffset);
                    StringReader reader = new StringReader(fParserState, false, fReader, literalDataString);
                    pushReader(reader, null);
                    alreadySlow = true;
                    //
                    // This part is tricky.  We are going to be reading from "an entity" that is using
                    // the literal data buffer.  We are now going to set the length of the literal data
                    // buffer back to the point where we were last scanning.  We will be appending the
                    // normalized result on top on the same place we are reading from.  We can do this
                    // because the remaining normalization will only make the result "shorter" than the
                    // data we scan and because the StringReader maintains its own copy of the offset
                    // and length information and the getChar() method in ChunkyCharArray does not check
                    // for reading past the point where we are appending.  This is as designed.
                    //
                    fLiteralData.setLength(lastDataOffset);
                    offset = fReader.currentOffset();
                    length = (getname ? fReader.skipPastName(qchar) : fReader.skipPastNmtoken(qchar)) - offset;
                    //
                    // We are on the slow path, so we need to update the literal data buffer.
                    //
                    int valueOffset = fLiteralData.length();
                    fReader.append(fLiteralData, offset, length);
                    int valueEnd = fLiteralData.length();
                    //
                    // This is dumb, we should be able to normalize the whitespace in the
                    // scanNormalizedAttValue method, but with the entity evaluation it
                    // is tricky and we might mess up.  Better to be safe than sorry...
                    //
                    keepgoing = fReader.skippedSpace();
                    if (keepgoing)
                        fReader.skipPastSpaces();
                    atend = fReader.skippedChar(qchar);
                    if (atend) {
                        //
                        // This is the final token in the AttValue.  Save the results for the last
                        // constraint check, and pop the reader we have been using.
                        //
                        value = fLiteralData.addSymbol(valueOffset, valueEnd - valueOffset);
                        if (onlyone)
                            fAttValueIndex = value;
                        else if (fixed)
                            fAttValueIndex = fLiteralData.addSymbol(dataOffset, valueEnd - dataOffset);
                        else
                            fAttValueIndex = fLiteralData.addString(dataOffset, valueEnd - dataOffset);
                        popReader();
                    } else if (keepgoing) {
                        //
                        // We have more tokens in the value.  Update the partial results and add a
                        // single space to the data buffer.
                        //
                        if (justone) {
                            fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                            popReader();
                            return false;
                        }
                        value = fLiteralData.addSymbol(valueOffset, valueEnd - valueOffset);
                        fLiteralData.append(' ');
                        offset = fReader.currentOffset();
                    } else {
                        fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                        popReader();
                        return false;
                    }
                }
            }
            //
            // Check type
            //
            switch (attType) {
            case AttDef.ID:
                //
                // ID - check that the id value is unique within the document (V_TAG8)
                //
                // REVISIT - if fScannerState != ST_NORMAL, we must be handling an AttValue in
                //  an AttlistDecl.  It doesn't seem to make sense for there to be a default AttValue
                //  for an Attr of type ID, right ?
                //
                if (fValidationHandler != null && !fElementDeclPool.addId(value, elementNameIndex)) {
                    emitError(ErrorCode.V_TAG8, value);
                    //return false;
                }
                return true;
            case AttDef.IDREF:
            case AttDef.IDREFS:
                //
                // IDREF - remember the id value
                // IDREFS - remember the id values
                //
                if (elementNameIndex != -1 && fValidationHandler != null) // in content, not dtd
                    fElementDeclPool.addIdRef(value, elementNameIndex);
                if (atend)
                    return true;
                break;
            case AttDef.ENTITY:
            case AttDef.ENTITIES:
                //
                // ENTITY - check that the value is an unparsed entity name (V_TAGa)
                // ENTITIES - check that the value is an unparsed entity name (V_TAGa)
                //
                int entity = fEntityPool.lookupEntity(value);
                if (fValidationHandler != null && entity == -1 || !fEntityPool.isUnparsed(entity)) {
                    emitError(ErrorCode.V_TAGa, value);
                    if (!atend)
                        fReader.skipPastChar(qchar);
                    //return false;
                }
                if (atend)
                    return true;
                break;
            case AttDef.NMTOKEN:
            case AttDef.NMTOKENS:
                //
                // NMTOKEN - nothing else to check
                // NMTOKENS - nothing else to check
                //
                if (atend)
                    return true;
                break;
            case AttDef.NOTATION:
            case AttDef.ENUMERATION:
                //
                // NOTATION - check that the value is in the AttDef enumeration (V_TAGb)
                // ENUMERATION - check that value is in the AttDef enumeration (V_TAG9)
                //
                int enumIndex;
                if (elementNameIndex == -1) {
                    //
                    // We are in the dtd, so we need to check against fields in fAttDef directly since
                    // we haven't added this AttDef to the pool yet.
                    //
                    enumIndex = fAttDef.enumeration;
                } else {
                    int attDef = fElementDeclPool.getAttDef(elementNameIndex, attrNameIndex);
                    enumIndex = fElementDeclPool.getEnumeration(attDef);
                }
                if (fValidationHandler != null && !fStringPool.stringInList(enumIndex, value)) {
                    int [] enumIndexes = fStringPool.stringsInList(enumIndex);
                    StringBuffer sb = new StringBuffer();
                    int el = enumIndexes.length;
                    for (int i=0; i<el; i++) {
                        sb.append(fStringPool.toString(enumIndexes[i]));
                        if (i<el-1)
                            sb.append(",");
                    }
                    int enumStringIndex = fStringPool.addSymbol(sb.toString());
                    emitError(ErrorCode.V_TAGo, value, enumStringIndex); // REVISIT - new error message
                    //return false;
                }
                return true;
            }
            onlyone = false;
        }
    }
    //
    // From the standard:
    //
    // [67] Reference ::= EntityRef | CharRef
    // [68] EntityRef ::= '&' Name ';'
    // [66] CharRef ::= '&#' [0-9]+ ';' | '&#x' [0-9a-fA-F]+ ';'
    //
    // Called after looking past the '&' character
    //
    private int referenceInEntityValue(int offset, boolean saveData) throws Exception {
        if (fReader.skippedChar('#')) {
            int num = scanCharRef();
            if (num == -1) {
                return fReader.currentOffset();
                // num = 0xfffd; // REVISIT - alternative is to use Unicode replacement char
            }
            if (saveData) {
                if (num < 0x10000)
                    fLiteralData.append((char)num);
                else {
                    fLiteralData.append((char)(((num-0x00010000)>>10)+0xd800));
                    fLiteralData.append((char)(((num-0x00010000)&0x3ff)+0xdc00));
                }
            }
            return fReader.currentOffset();
        }
        int nameOffset = fReader.currentOffset();
        int nameLength = fReader.skipPastName(';') - nameOffset;
        if (nameLength == 0) {
            fReader.skipInvalidChar(ErrorCode.E_REFER0);
            return fReader.currentOffset();
        }
        if (!fReader.skippedChar(';')) {
            emitError(ErrorCode.E_REFER1);
            return fReader.currentOffset();
        }
        //
        // 4.4.7 Bypassed
        //
        // When a general entity reference appears in the EntityValue in an
        // entity declaration, it is bypassed and left as is.
        //
        if (saveData)
            fReader.append(fLiteralData, offset, fReader.currentOffset() - offset);
        return fReader.currentOffset();
    }
    private boolean referenceInContent(int offset) throws Exception {
        if (fReader.skippedChar('#')) {
            int num = scanCharRef();
            if (num == -1) {
                return false;
                // num = 0xfffd; // REVISIT - alternative is to use Unicode replacement char
            }
            if (fDocumentHandler != null) {
                if (fCharRefData == null)
                    fCharRefData = new char[2];
                int count = (num < 0x10000) ? 1 : 2;
                if (count == 1)
                    fCharRefData[0] = (char)num;
                else {
                    fCharRefData[0] = (char)(((num-0x00010000)>>10)+0xd800);
                    fCharRefData[1] = (char)(((num-0x00010000)&0x3ff)+0xdc00);
                }
                if (!fDocumentHandler.sendCharDataAsCharArray()) {
                    //revist: is this the best way?
                    int stringIndex = fStringPool.addString(new String(fCharRefData, 0, count));
                    fDocumentHandler.characters(stringIndex, false);
                } else {
                    fDocumentHandler.characters(fCharRefData, 0, count, false);
                }
            }
            return true;
        }
        //
        // [68] EntityRef ::= '&' Name ';'
        //
        int nameOffset = fReader.currentOffset();
        int nameLength = fReader.skipPastName(';') - nameOffset;
        if (nameLength == 0) {
            fReader.skipInvalidChar(ErrorCode.E_REFER0);
            return false;
        }
        if (!fReader.skippedChar(';')) {
            emitError(ErrorCode.E_REFER1);
            return false;
        }
        int entityNameIndex = fReader.addSymbol(nameOffset, nameLength);
        int entityIndex = fEntityPool.lookupEntity(entityNameIndex);
        if (entityIndex < 0) {
            emitError(ErrorCode.E_ENTITY0, entityNameIndex);
            return false;
        }
        if (fEntityPool.isUnparsed(entityIndex)) {
            emitError(ErrorCode.E_ENTITY2, entityNameIndex);
            return false;
        }
        //
        // Check for recursive entity references
        //
        if (!pushEntity(GENERAL_ENTITY_TYPE, entityNameIndex))
            return false;
        //
        // REVISIT - when we have a flag to suppress entity expansion
        //  we will still need to validate the replacement text, yet not
        //  have it appear as content to the document handler.  Does this
        //  option make sense ?
        //
        if (fDocumentHandler != null)
            fDocumentHandler.startEntityReference(entityIndex);
        XMLReader reader = null;
        InputSource source = null;
        if (fEntityPool.isExternal(entityIndex)) {
            int publicId = fEntityPool.getPublicId(entityIndex);
            int systemId = fEntityPool.getSystemId(entityIndex);
            systemId = fEntityHandler.expandSystemId(systemId);
            source = fEntityHandler.resolveEntity(publicId, systemId);
            if (source == null) {
                source = new InputSource(fStringPool.toString(systemId));
                if (publicId != -1)
                    source.setPublicId(fStringPool.toString(publicId));
            }
            boolean textDecl = false; // xmlDecl if true, textDecl if false
            reader = fEntityHandler.createReader(source, textDecl);
            fEntityHandler.startInputSource(source);
        } else {
            reader = new StringReader(fParserState, false, fReader, fEntityPool.getEntityValue(entityIndex));
        }
        int depth = pushReader(reader, source);
        int oldLimit = setActiveReaderLimit(depth);
        try {
            ScanContentState scanState = new ScanContentState();
            scanState.inCDSect = false;
            scanState.extParsedEnt = true;
            scanState.parseTextDecl = fReader.lookingAtChar('<');
            scanState.elementDepth = elementStackDepth();
            scanContent(scanState);
        } catch (ArrayIndexOutOfBoundsException ex) {
            if (PRINT_EXCEPTION_STACK_TRACE)
                ex.printStackTrace();
            }
        catch (StopException e) {
            if (DEBUG_STOP) {
                System.out.println("!! referenceInContent(): "+e.getMessage());
                }
            throw e;
            }
        catch (Exception ex) {
            if (PRINT_EXCEPTION_STACK_TRACE)
                ex.printStackTrace();
			throw ex;
        }
        setActiveReaderLimit(oldLimit);
        popReader();
        if (fDocumentHandler != null)
            fDocumentHandler.endEntityReference(entityIndex);
        popEntity();
        return true;
    }
    //
    // [66] CharRef ::= '&#' [0-9]+ ';' | '&#x' [0-9a-fA-F]+ ';'
    //
    private int scanCharRef() throws Exception {
        boolean hex = fReader.skippedChar('x');
        int valueOffset = fReader.currentOffset();
        int valueLength = 0;
        int num = 0;
        if (hex) {
            int dig = fReader.skipHexDigit();
            if (dig < 0) {
                if (fReader.skippedChar(';'))
                    emitError(ErrorCode.E_REFER2);
                else
                    fReader.skipInvalidChar(ErrorCode.E_REFER0);
                return -1;
            }
            num = dig;
            while (true) {
                dig = fReader.skipHexDigit();
                if (dig < 0)
                    break;
                num = (num << 4) + dig;
                if (num > 0x10FFFF) {
                    fReader.skipToChar(';');
                    break;
                }
            }
        } else {
            int dig = fReader.skipDecimalDigit();
            if (dig < 0) {
                if (fReader.skippedChar(';'))
                    emitError(ErrorCode.E_REFER2);
                else
                    fReader.skipInvalidChar(ErrorCode.E_REFER0);
                return -1;
            }
            num = dig;
            while (true) {
                dig = fReader.skipDecimalDigit();
                if (dig < 0)
                    break;
                num = (num * 10) + dig;
                if (num > 0x10FFFF) {
                    fReader.skipToChar(';');
                    break;
                }
            }
        }
        valueLength = fReader.currentOffset() - valueOffset;
        if (!fReader.skippedChar(';')) {
            emitError(ErrorCode.E_REFER1);
            return -1;
        }
        //
        //  [2] Char ::= #x9 | #xA | #xD | [#x20-#xD7FF]        // any Unicode character, excluding the
        //               | [#xE000-#xFFFD] | [#x10000-#x10FFFF] // surrogate blocks, FFFE, and FFFF.
        //
        if (num < 0x20) {
            if (num == 0x09 || num == 0x0A || num == 0x0D)
                return num;
        } else if (num <= 0xD7FF || (num >= 0xE000 && (num <= 0xFFFD || (num >= 0x10000 && num <= 0x10FFFF))))
            return num;
        int errorCode = hex ? ErrorCode.E_REFER3 : ErrorCode.E_REFER4;
        emitError(errorCode, fReader.addString(valueOffset, valueLength));
        return -1;
    }
    //
    // From the standard:
    //
    // [27] Misc ::= Comment | PI |  S
    // [15] Comment ::= '<!--' ((Char - '-') | ('-' (Char - '-')))* '-->'
    // [16] PI ::= '<?' PITarget (S (Char* - (Char* '?>' Char*)))? '?>'
    // [17] PITarget ::= Name - (('X' | 'x') ('M' | 'm') ('L' | 'l'))
    //
    private void scanMisc() throws Exception
    {
        while (true) {
            if (fReader.skippedChar('<')) {
                increaseMarkupDepth();
                if (fReader.skippedChar('?')) {
                    int piTarget = fReader.scanName(' ', -1);
                    if (piTarget == -1) {
                        emitError(ErrorCode.E_PI0);
                        fReader.skipPastChar('>');
                        decreaseMarkupDepth();
                    } else if ("xml".equals(fStringPool.toString(piTarget))) {
                        if (fReader.skippedSpace()) // an XMLDecl looks like a PI with the target 'xml'
                            emitError(ErrorCode.E_STRUCT1);
                        else // a PI target matching 'xml'
                            emitError(ErrorCode.E_PI5);
                        fReader.skipPastChar('>');
                        decreaseMarkupDepth();
                    } else // PI
                        scanPI(piTarget, true);
                } else if (fReader.skippedChar('!')) {
                    if (fReader.skippedChar('-')) { // comment ?
                        if (fReader.skippedChar('-')) {
                            scanComment(true); // scan through the closing '-->'
                        } else {
                            emitError(ErrorCode.E_COM0);
                            fReader.skipPastChar('>');
                            decreaseMarkupDepth();
                        }
                    } else {
                        emitError(ErrorCode.E_STRUCT0);
                        fReader.skipPastChar('>');
                        decreaseMarkupDepth();
                    }
                } else { // another root element ?
                    emitError(ErrorCode.E_STRUCT0);
                    fReader.skipToChar('<'); // REVISIT - should we just skip the rest?
                    decreaseMarkupDepth();
                }
            }
            else if (fReader.skippedSpace())
                fReader.skipPastSpaces();
            else if (!fReader.lookingAtValidChar())
                fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
            else {
                emitError(ErrorCode.E_STRUCT0);
                fReader.skipToChar('<');
            }
        }
    }
    //
    // From the standard:
    //
    // [16] PI ::= '<?' PITarget (S (Char* - (Char* '?>' Char*)))? '?>'
    // [17] PITarget ::= Name - (('X' | 'x') ('M' | 'm') ('L' | 'l'))
    //
    private void scanPI(int piTarget, boolean inDocument) throws Exception
    {
        String piTargetString = fStringPool.toString(piTarget);
        if (piTargetString.length() == 3 &&
                (piTargetString.charAt(0) == 'X' || piTargetString.charAt(0) == 'x') &&
                (piTargetString.charAt(1) == 'M' || piTargetString.charAt(1) == 'm') &&
                (piTargetString.charAt(2) == 'L' || piTargetString.charAt(2) == 'l')) {
            emitError(ErrorCode.E_PI5);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        if (!fReader.skippedSpace()) {
            if (!fReader.skippedChar('?')) {
                emitError(ErrorCode.E_PI3);
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            if (!fReader.skippedChar('>')) {
                emitError(ErrorCode.E_PI3);
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            decreaseMarkupDepth();
            if (inDocument && fDocumentHandler != null)
                fDocumentHandler.processingInstruction(piTarget, -1);
            return;
        }
        int piDataOffset = fReader.skipPastSpaces();
        try {
            while (true) {
                while (fReader.lookingAtChar('?')) {
                    int offset = fReader.currentOffset();
                    fReader.skipAsciiChar();
                    if (fReader.skippedChar('>')) {
                        decreaseMarkupDepth();
                        if (inDocument && fDocumentHandler != null) {
                            int piDataIndex = -1;
                            if (offset != piDataOffset)
                                piDataIndex = fReader.addString(piDataOffset, offset - piDataOffset);
                            fDocumentHandler.processingInstruction(piTarget, piDataIndex);
                        }
                        return;
                    }
                }
                if (!fReader.skippedValidChar()) {
                    fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                    fReader.skipPastChar('>');
                    decreaseMarkupDepth();
                    return;
                }
            }
        } catch (ArrayIndexOutOfBoundsException ex) {
            if (peekEntity() == 0) {
                emitError(ErrorCode.E_PI3);
                throw ex;
            }
        }
    }
    //
    // From the standard:
    //
    // [28] doctypedecl ::= '<!DOCTYPE' S Name (S ExternalID)? S?
    //                      ('[' (markupdecl | PEReference | S)* ']' S?)? '>'
    // [29] markupdecl ::= elementdecl | AttlistDecl | EntityDecl
    //                     | NotationDecl | PI | Comment
    //
    // Called after scanning '<!DOCTYPE'
    //
    private void scanDoctypeDecl() throws Exception
    {
        if (!fReader.skippedSpace()) {
            emitError(ErrorCode.E_SPACE);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        fReader.skipPastSpaces();
        int nameIndex = fReader.scanName(' ', -1);
        if (nameIndex == -1) {
            emitError(ErrorCode.E_DOCTYPE1);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        int elementIndex = fElementDeclPool.addElement(nameIndex);
        fElementDeclPool.setRootElement(elementIndex);
        if (fDocumentTypeHandler != null)
            fDocumentTypeHandler.doctypeDecl(nameIndex);
        boolean lbrkt;
        boolean scanExternalSubset = false;
        int externalPublicId = -1;
        int externalSystemId = -1;
        if (fReader.skippedSpace()) {
            fReader.skipPastSpaces();
            if (!(lbrkt = fReader.skippedChar('[')) && !fReader.lookingAtChar('>')) {
                scanExternalSubset = true; // REVISIT - when should this be false here ?
                if (!scanExternalID(scanExternalSubset)) {
                    fReader.skipPastChar('>');
                    decreaseMarkupDepth();
                    return;
                }
                externalPublicId = fPubidLiteral;
                externalSystemId = fSystemLiteral;
                fReader.skipPastSpaces();
                lbrkt = fReader.skippedChar('[');
            }
        } else
            lbrkt = fReader.skippedChar('[');
        if (lbrkt) {
            if (!scanIntSubsetDecl()) {
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            fReader.skipPastSpaces();
        }
        if (!fReader.skippedChar('>')) {
            emitError(ErrorCode.E_TAG4);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        decreaseMarkupDepth();
        if (scanExternalSubset) {
            //revisit: At this point we *know* we are scanning the external subset.
            if (fDocumentTypeHandler != null) {
                fDocumentTypeHandler.startExternalSubset(externalPublicId, externalSystemId);
            }
            int prevState = fScannerState;
            fScannerState = ST_EXTERNAL_SUBSET;
            externalSystemId = fEntityHandler.expandSystemId(externalSystemId);
            InputSource source = fEntityHandler.resolveEntity(externalPublicId, externalSystemId);
            if (source == null) {
                source = new InputSource(fStringPool.toString(externalSystemId));
                if (externalPublicId != -1)
                    source.setPublicId(fStringPool.toString(externalPublicId));
            }
            boolean textDecl = false; // xmlDecl if true, textDecl if false
            XMLReader reader = fEntityHandler.createReader(source, textDecl);
            fEntityHandler.startInputSource(source);
            int depth = pushReader(reader, source);
            int oldLimit = setActiveReaderLimit(depth);
            try {
                scanExtSubsetDecl(false);

            } catch (ArrayIndexOutOfBoundsException ex) {
                if (PRINT_EXCEPTION_STACK_TRACE)
                    ex.printStackTrace();
            }
            catch (StopException e) {
                if (DEBUG_STOP) {
                    System.out.println("!! scanDoctypeDecl(): "+e.getMessage());
                    }
                throw e;
            }
            catch (Exception ex) {
                if (PRINT_EXCEPTION_STACK_TRACE)
                    ex.printStackTrace();
				throw ex;
            }
            setActiveReaderLimit(oldLimit);
            popReader();
            fScannerState = prevState;

            //revisit: At this point we *know* we are ending scanning the external subset.
            if (fDocumentTypeHandler != null) {
                fDocumentTypeHandler.endExternalSubset();
            }

        }

        // check declared entities
        if (fParserState.getValidationHandler() != null) {
            fElementDeclPool.checkDeclaredElements();
        }
    }
    //
    // [??] intSubsetDecl = '[' (markupdecl | PEReference | S)* ']'
    //
    // Called after scanning past '['
    //
    private boolean scanIntSubsetDecl() throws Exception
    {
        //revisit: At this point we *know* we are scanning the internal subset.
        if (fDocumentTypeHandler != null) {
            fDocumentTypeHandler.startInternalSubset();
        }
        boolean noErrors = true;
        int prevState = fScannerState;
        fScannerState = ST_INTERNAL_SUBSET;
        while (!fReader.skippedChar(']')) {
            if (fReader.skippedChar('<'))
                scanMarkupDecl(false);
            else if (fReader.skippedSpace())
                fReader.skipPastSpaces();
            else if (fReader.skippedChar('%'))
                scanPEReference();
            else {
                if (fReader.lookingAtChar('>')) {
                    fReader.skipInvalidChar(ErrorCode.E_DTD0);
                    return false;
                }
                fReader.lookingAtValidChar();
                fReader.skipInvalidChar(ErrorCode.E_DTD0);
                noErrors = false;
                do {
                    fReader.skipOneChar();
                } while (!fReader.lookingAtChar('%') && !fReader.lookingAtChar('<') &&
                         !fReader.lookingAtSpace() && !fReader.lookingAtChar(']'));
            }
        }
        fScannerState = prevState;

        //revisit: At this point we *know* we are ending the internal subset.
        if (fDocumentTypeHandler != null) {
            fDocumentTypeHandler.endInternalSubset();
        }

        return noErrors;
    }
    //
    // From the standard:
    //
    // [77] TextDecl ::= '<?xml' VersionInfo? EncodingDecl S? '?>'
    // [24] VersionInfo ::= S 'version' Eq (' VersionNum ' | " VersionNum ")
    // [80] EncodingDecl ::= S 'encoding' Eq ('"' EncName '"' |  "'" EncName "'" )
    // [81] EncName ::= [A-Za-z] ([A-Za-z0-9._] | '-')*
    //
    private void scanTextDecl() throws Exception
    {
        boolean single;
        char qchar;
        // this is okay, we already checked fReader.lookingAtSpace in caller
        int versionOffset = fReader.skipPastSpaces();
        int encOffset;
        if (fReader.skippedString(version_string)) {
            int versionLength = fReader.currentOffset() - versionOffset;
            if (!scanEq()) {
                emitError(ErrorCode.E_ATTEQ0, fReader.addString(versionOffset, versionLength));
                // REVISIT - rethink error recovery
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            if (!(single = fReader.skippedChar('\'')) && !fReader.skippedChar('\"')) {
                emitError(ErrorCode.E_ATTVAL0);
                // REVISIT - rethink error recovery
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            int offset = fReader.currentOffset();
            qchar = single ? '\'' : '\"';
            while (!fReader.lookingAtChar(qchar)) {
                if (!fReader.skippedVersionNum()) {
                    while (!fReader.lookingAtChar(qchar) && !fReader.lookingAtChar('<'))
                        fReader.skipOneChar();
                    if (fReader.lookingAtChar('<'))
                        emitError(ErrorCode.E_ATTVAL1);
                    else
                        emitError(ErrorCode.E_XML6, fReader.addString(offset, fReader.currentOffset() - offset));
                    fReader.skipPastChar('>');
                    decreaseMarkupDepth();
                    return;
                }
            }
            String version = fStringPool.orphanString(fReader.addString(offset, fReader.currentOffset() - offset));
            if (!"1.0".equals(version)) {
                emitError(ErrorCode.E_XML4);
                // return fReader.skipPastChar('>'); // REVISIT - hope it is compatible...
            }
            fReader.skipAsciiChar();
            if (!fReader.skippedSpace()) {
                emitError(ErrorCode.E_SPACE);
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            encOffset = fReader.skipPastSpaces();
        } else
            encOffset = versionOffset;
        if (!fReader.skippedString(encoding_string)) {
            emitError(ErrorCode.E_PI7);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        int encLength = fReader.currentOffset() - encOffset;
        if (!scanEq()) {
            emitError(ErrorCode.E_ATTEQ0, fReader.addString(encOffset, encLength));
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        if (!(single = fReader.skippedChar('\'')) && !fReader.skippedChar('\"')) {
            emitError(ErrorCode.E_ATTVAL0);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        int offset = fReader.currentOffset();
        qchar = single ? '\'' : '\"';
        if (!fReader.skippedAlpha()) {
            while (!fReader.lookingAtChar(qchar) && !fReader.lookingAtChar('<'))
                fReader.skipOneChar();
            if (fReader.lookingAtChar('<'))
                emitError(ErrorCode.E_ATTVAL1);
            else
                emitError(ErrorCode.E_PI4, fReader.addString(offset, fReader.currentOffset() - offset));
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        while (!fReader.lookingAtChar(qchar)) {
            if (!fReader.skippedEncName()) {
                while (!fReader.lookingAtChar(qchar) && !fReader.lookingAtChar('<'))
                    fReader.skipOneChar();
                if (fReader.lookingAtChar('<'))
                    emitError(ErrorCode.E_ATTVAL1);
                else
                    emitError(ErrorCode.E_PI4, fReader.addString(offset, fReader.currentOffset() - offset));
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
        }
        fReader.skipAsciiChar();
        fReader.skipPastSpaces();
        if (!fReader.skippedChar('?')) {
            emitError(ErrorCode.E_PI3);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        if (!fReader.skippedChar('>')) {
            emitError(ErrorCode.E_PI3);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        decreaseMarkupDepth();
    }
    //
    // [31] extSubsetDecl ::= ( markupdecl | conditionalSect | PEReference | S )*
    // [62] includeSect ::= '<![' S? 'INCLUDE' S? '[' extSubsetDecl ']]>'
    //
    private void scanExtSubsetDecl(boolean includeSect) throws Exception
    {
        boolean parseTextDecl = !includeSect;
        try {
            while (true) {
                if (checkSkippedChar('<', false))
                    scanMarkupDecl(parseTextDecl);
                else if (checkSkippedSpace())
                    checkSkipPastSpaces();
                else if (checkSkippedChar('%', false))
                    scanPEReference();
                else if (includeSect && checkSkippedChar(']', false)) {
                    if (!checkSkippedChar(']', false)) {
                        emitError(ErrorCode.E_COND1);
                        fReader.skipPastChar('>');
                        decreaseMarkupDepth();
                        return;
                    }
                    if (!checkSkippedChar('>', false)) {
                        emitError(ErrorCode.E_COND1);
                        fReader.skipPastChar('>');
                        decreaseMarkupDepth();
                        return;
                    }
                    decreaseMarkupDepth();
                    return;
                } else {
                    if (!fReader.lookingAtValidChar())
                        fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                    else
                        emitError(ErrorCode.E_STRUCT0);
                    do {
                        fReader.skipOneChar();
                    } while (!fReader.lookingAtChar('%') && !fReader.lookingAtChar('<') &&
                             !(includeSect && fReader.lookingAtChar(']')));
                }
                parseTextDecl = false;
            }
        } catch (ArrayIndexOutOfBoundsException ex) {
            if (includeSect)
                emitError(ErrorCode.E_COND1);
            throw ex;
        }
    }
    //
    // [75] ExternalID ::= 'SYSTEM' S SystemLiteral
    //                     | 'PUBLIC' S PubidLiteral S SystemLiteral
    // [83] PublicID ::= 'PUBLIC' S PubidLiteral
    //
    private boolean scanExternalID(boolean saveString) throws Exception
    {
        return scanExternalIDorPublicID(saveString, false);
    }
    private boolean scanExternalIDorPublicID(boolean saveString) throws Exception
    {
        return scanExternalIDorPublicID(saveString, true);
    }
    private boolean scanExternalIDorPublicID(boolean saveString, boolean scanPublicID) throws Exception
    {
        fSystemLiteral = -1;
        fPubidLiteral = -1;
        int offset = fReader.currentOffset();
        if (fReader.skippedString(system_string)) {
            if (!fReader.skippedSpace()) {
                emitError(ErrorCode.E_SPACE);
                return false;
            }
            fReader.skipPastSpaces();
            return scanSystemLiteral(saveString);
        }
        if (fReader.skippedString(public_string)) {
            if (!fReader.skippedSpace()) {
                emitError(ErrorCode.E_SPACE);
                return false;
            }
            fReader.skipPastSpaces();
            if (!scanPubidLiteral(saveString))
                return false;
            if (scanPublicID) {
                //
                // [82] NotationDecl ::= '<!NOTATION' S Name S (ExternalID |  PublicID) S? '>'
                //
                if (!fReader.skippedSpace())
                    return true; // no S, not an ExternalID
                fReader.skipPastSpaces();
                if (fReader.lookingAtChar('>')) // matches end of NotationDecl
                    return true;
            } else {
                if (!fReader.skippedSpace()) {
                    emitError(ErrorCode.E_SPACE);
                    return false;
                }
                fReader.skipPastSpaces();
            }
            return scanSystemLiteral(saveString);
        }
        // REVISIT - redo error recovery - this is from XML4J
        int length = fReader.skipPastNmtoken(' ') - offset;
        int nameIndex = fReader.addString(offset, length);
        emitError(ErrorCode.E_DOCTYPE2, nameIndex);
        return false;
    }
    //
    // [11] SystemLiteral ::= ('"' [^"]* '"') | ("'" [^']* "'")
    //
    // REVISIT - TODO - this require some special normalization.
    //
    private boolean scanSystemLiteral(boolean saveString) throws Exception
    {
        boolean single;
        if (!(single = fReader.skippedChar('\'')) && !fReader.skippedChar('\"')) {
            emitError(ErrorCode.E_EXT3);
            return false;
        }
        int offset = fReader.currentOffset();
        char qchar = single ? '\'' : '\"';
        while (!fReader.lookingAtChar(qchar)) {
            if (!fReader.skippedValidChar()) {
                fReader.skipInvalidChar(ErrorCode.E_EXT4);
                fReader.skipPastChar(qchar);
                return false;
            }
        }
        if (saveString)
            fSystemLiteral = fReader.addString(offset, fReader.currentOffset() - offset);
        fReader.skipAsciiChar();
        return true;
    }
    //
    // [12] PubidLiteral ::= '"' PubidChar* '"' | "'" (PubidChar - "'")* "'"
    // [13] PubidChar ::= #x20 | #xD | #xA | [a-zA-Z0-9] | [-'()+,./:=?;!*#@$_%]
    //
    private boolean scanPubidLiteral(boolean saveString) throws Exception
    {
        boolean single;
        if (!(single = fReader.skippedChar('\'')) && !fReader.skippedChar('\"')) {
            emitError(ErrorCode.E_EXT1);
            return false;
        }
        char qchar = single ? '\'' : '\"';
        int offset = fReader.skipPastSpaces();
        int dataOffset = saveString ? fLiteralData.length() : 0;
        boolean oweSpace = false;
        int toCopy = offset;
        while (true) {
            if (fReader.skippedChar(qchar)) {
                if (saveString && offset - toCopy > 0)
                    fReader.append(fLiteralData, toCopy, offset - toCopy);
                break;
            }
            if (fReader.skippedSpace()) {
                if (saveString && offset - toCopy > 0)
                    fReader.append(fLiteralData, toCopy, offset - toCopy);
                offset = fReader.skipPastSpaces();
                toCopy = offset;
                if (fReader.skippedChar(qchar))
                    break;
                if (saveString)
                    fLiteralData.append(' ');
            }
            if (!fReader.skippedPubidChar()) {
                fReader.skipInvalidChar(ErrorCode.E_EXT2);
                fReader.skipPastChar(qchar);
                return false;
            }
            offset = fReader.currentOffset();
        }
        if (saveString) {
            int dataLength = fLiteralData.length() - dataOffset;
            fPubidLiteral = fLiteralData.addString(dataOffset, dataLength);
        }
        return true;
    }
    //
    // [69] PEReference ::= '%' Name ';'
    //
    private void scanPEReference() throws Exception {
        int nameOffset = fReader.currentOffset();
        int nameLength = fReader.skipPastName(';') - nameOffset;
        if (nameLength == 0) {
            emitError(ErrorCode.E_PEREF0);
            return;
        }
        if (!fReader.skippedChar(';')) {
            emitError(ErrorCode.E_PEREF1, fReader.addString(nameOffset, nameLength));
            return;
        }
        int peNameIndex = fReader.addSymbol(nameOffset, nameLength);
        int peIndex = fParameterEntityPool == null ? -1 : fParameterEntityPool.lookupEntity(peNameIndex);
        if (peIndex < 0) {
            emitError(ErrorCode.V_PEREF2, peNameIndex);
            return;
        }
        if (!pushEntity(PARAMETER_ENTITY_TYPE, peNameIndex))
            return;
        XMLReader reader = null;
        InputSource source = null;
        if (fParameterEntityPool.isExternal(peIndex)) {
            int publicId = fParameterEntityPool.getPublicId(peIndex);
            int systemId = fParameterEntityPool.getSystemId(peIndex);
            systemId = fEntityHandler.expandSystemId(systemId);
            source = fEntityHandler.resolveEntity(publicId, systemId);
            if (source == null) {
                source = new InputSource(fStringPool.toString(systemId));
                if (publicId != -1)
                    source.setPublicId(fStringPool.toString(publicId));
            }
            boolean textDecl = false; // xmlDecl if true, textDecl if false
            reader = fEntityHandler.createReader(source, textDecl);
            fEntityHandler.startInputSource(source);
        } else {
            reader = new StringReader(fParserState, true, fReader, fParameterEntityPool.getEntityValue(peIndex));
        }
        int depth = pushReader(reader, source);
        int oldLimit = setActiveReaderLimit(depth);
        try {
            scanExtSubsetDecl(false);
        } catch (ArrayIndexOutOfBoundsException ex) {
            if (PRINT_EXCEPTION_STACK_TRACE)
                ex.printStackTrace();
            }
        catch (StopException e) {
            if (DEBUG_STOP) {
                System.out.println("!! scanPEReference(): "+e.getMessage());
                }
            throw e;
            }
        catch (Exception ex) {
            if (PRINT_EXCEPTION_STACK_TRACE)
                ex.printStackTrace();
			throw ex;
        }
        setActiveReaderLimit(oldLimit);
        popReader();
        popEntity();
    }
    //
    // Called for PEReference in EntityValue
    //
    private boolean copyPEReference() throws Exception {
        int nameOffset = fReader.currentOffset();
        int nameLength = fReader.skipPastName(';') - nameOffset;
        if (nameLength == 0) {
            emitError(ErrorCode.E_PEREF0);
            return false;
        }
        if (!fReader.skippedChar(';')) {
            emitError(ErrorCode.E_PEREF1, fReader.addString(nameOffset, nameLength));
            return false;
        }
        int peNameIndex = fReader.addSymbol(nameOffset, nameLength);
        int peIndex = fParameterEntityPool == null ? -1 : fParameterEntityPool.lookupEntity(peNameIndex);
        if (peIndex < 0) {
            emitError(ErrorCode.V_PEREF2, peNameIndex);
            return false;
        }
        if (!fParameterEntityPool.isExternal(peIndex)) {
            //
            // REVISIT - consider optimizing the case where the entire entity value
            // consists of a single reference to a parameter entity and do not append
            // the value to fLiteralData again, but re-use the offset/length of the
            // referenced entity for the value of this entity.
            //
            fLiteralData.append(fStringPool.toString(fParameterEntityPool.getEntityValue(peIndex)));
            return true;
        }
        //
        // REVISIT - Can we get the spec changed ?
        // There is a perverse edge case to handle here...  We have a reference
        // to an external PE within a literal EntityValue.  For the PE to be
        // well-formed, it must match the extPE production, but the code that
        // appends the replacement text to the entity value is in no position
        // to do a complete well-formedness check !!
        //
        if (!pushEntity(PARAMETER_ENTITY_TYPE, peNameIndex))
            return false;
        int publicId = fParameterEntityPool.getPublicId(peIndex);
        int systemId = fParameterEntityPool.getSystemId(peIndex);
        systemId = fEntityHandler.expandSystemId(systemId);
        InputSource source = fEntityHandler.resolveEntity(publicId, systemId);
        if (source == null) {
            source = new InputSource(fStringPool.toString(systemId));
            if (publicId != -1)
                source.setPublicId(fStringPool.toString(publicId));
        }
        boolean textDecl = false; // xmlDecl if true, textDecl if false
        XMLReader reader = fEntityHandler.createReader(source, textDecl);
        fEntityHandler.startInputSource(source);
        int depth = pushReader(reader, source);
        int oldLimit = setActiveReaderLimit(depth);
        int dataOffset = fReader.currentOffset();
        boolean success = true;
        try {
            while (true) {
                if (!fReader.skippedValidChar()) {
                    fReader.skipInvalidChar(ErrorCode.E_ENTITY3);
                    success = false;
                    break;
                }
            }
        } catch (ArrayIndexOutOfBoundsException ex) {
            if (PRINT_EXCEPTION_STACK_TRACE)
                ex.printStackTrace();
            }
        catch (StopException e) {
            if (DEBUG_STOP) {
                System.out.println("!! copyPEReference(): "+e.getMessage());
                }
            throw e;
            }
        catch (Exception ex) {
            if (PRINT_EXCEPTION_STACK_TRACE)
                ex.printStackTrace();
			throw ex;
        }
        if (success)
            fReader.append(fLiteralData, dataOffset, fReader.currentOffset() - dataOffset);
        setActiveReaderLimit(oldLimit);
        popReader();
        popEntity();
        return success;
    }
    private void expandPEReference() throws Exception
    {
        int nameOffset = fReader.currentOffset();
        int nameLength = fReader.skipPastName(';') - nameOffset;
        if (nameLength == 0)
            emitError(ErrorCode.E_PEREF0);
        if (!fReader.skippedChar(';'))
            emitError(ErrorCode.E_PEREF1, fReader.addString(nameOffset, nameLength));
        int peNameIndex = fReader.addSymbol(nameOffset, nameLength);
        int peIndex = fParameterEntityPool.lookupEntity(peNameIndex);
        if (peIndex < 0) {
            emitError(ErrorCode.V_PEREF2, peNameIndex);
            return;
        }
        if (!pushEntity(PARAMETER_ENTITY_TYPE, peNameIndex))
            return;
        if (!fParameterEntityPool.isExternal(peIndex)) {
            StringReader reader = new StringReader(fParserState, true, fReader,
                                                   fParameterEntityPool.getEntityValue(peIndex));
            pushReader(reader, null);
            return;
        }
        //
        // REVISIT - Can we get the spec changed ?
        // There is a perverse edge case to handle here...  We have a reference
        // to an external PE within markup.  For the PE to be well-formed, it
        // must match the extPE production, which is probably not going to be
        // very useful expanded in the middle of a markup declaration.  The
        // problem is that an empty file, a file containing just whitespace or
        // another PE that is just empty or whitespace, matches extPE !!
        //
        int publicId = fParameterEntityPool.getPublicId(peIndex);
        int systemId = fParameterEntityPool.getSystemId(peIndex);
        systemId = fEntityHandler.expandSystemId(systemId);
        InputSource source = fEntityHandler.resolveEntity(publicId, systemId);
        if (source == null) {
            source = new InputSource(fStringPool.toString(systemId));
            if (publicId != -1)
                source.setPublicId(fStringPool.toString(publicId));
        }
        boolean textDecl = false; // xmlDecl if true, textDecl if false
        XMLReader reader = fEntityHandler.createReader(source, textDecl);
        fEntityHandler.startInputSource(source);
        pushReader(reader, source);
        return;
    }
    //
    //
    //
    private boolean checkSkippedChar(char ch, boolean skipSpacesAfterPop) throws Exception {
        while (true) {
            try {
                boolean skippedChar = fReader.skippedChar(ch);
                return skippedChar;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                if (skipSpacesAfterPop)
                    fReader.skipPastSpaces();
            }
        }
    }
    private boolean checkSkippedSpace() throws Exception {
        while (true) {
            try {
                boolean skippedSpace = fReader.skippedSpace();
                return skippedSpace;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                //if (fReader.lookingAtSpace())
                //    hey();
            }
        }
    }
    private int checkSkipPastSpaces() throws Exception {
        while (true) {
            try {
                int offset = fReader.skipPastSpaces();
                return offset;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
            }
        }
    }
    private boolean checkForPEReference(boolean spaceRequired) throws Exception
    {
        boolean sawSpace = true;
        if (spaceRequired) {
            while (true) {
                try {
                    sawSpace = fReader.skippedSpace();
                    break;
                } catch (ArrayIndexOutOfBoundsException ex) {
                    popReader();
                    popEntity();
                    //if (fReader.lookingAtSpace())
                    //    hey();
                    sawSpace = fReader.skippedSpace();
                    if (sawSpace)
                        fReader.skipPastSpaces();
                    //else
                    //    System.err.println("ERROR#2 - need to check for this !!");

                    break;
                }
            }
        }
        while (true) {
            try {
                fReader.skipPastSpaces();
                break;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                if (!sawSpace)
                    sawSpace = fReader.skippedSpace();
            }
        }
        if (fScannerState != ST_EXTERNAL_SUBSET)
            return sawSpace;
        boolean atPERef;
        while (true) {
            try {
                atPERef = fReader.skippedChar('%');
                break;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                if (!sawSpace)
                    sawSpace = fReader.skippedSpace();
                if (sawSpace)
                    fReader.skipPastSpaces();
            }
        }
        if (!atPERef)
            return sawSpace;
        while (true) {
            expandPEReference();
            while (true) {
                try {
                    fReader.skipPastSpaces();
                    break;
                } catch (ArrayIndexOutOfBoundsException ex) {
                    popReader();
                    popEntity();
                }
            }
            atPERef = checkSkippedChar('%', true);
            if (!atPERef)
                return true;
        }
    }
    private boolean checkForPEDecl() throws Exception
    {
        boolean sawSpace = checkSkippedSpace();
        if (sawSpace) {
            while (true) {
                try {
                    fReader.skipPastSpaces();
                    break;
                } catch (ArrayIndexOutOfBoundsException ex) {
                    popReader();
                    popEntity();
                    //if (fReader.lookingAtSpace())
                    //    hey();
                    sawSpace = fReader.skippedSpace();
                    if (sawSpace)
                        fReader.skipPastSpaces();
                    //else
                    //    System.err.println("ERROR#6 - need to check for this !!");
                }
            }
        }
        boolean atPERef;
        while (true) {
            try {
                atPERef = fReader.skippedChar('%');
                break;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                //if (fReader.lookingAtSpace())
                //    hey();
                sawSpace = fReader.skippedSpace();
                if (sawSpace)
                    fReader.skipPastSpaces();
                //else
                //    System.err.println("ERROR#7 - need to check for this !!");
            }
        }
        if (!atPERef) { // <!ENTITY x "x">
            if (!sawSpace)
                emitError(ErrorCode.E_SPACE);
            return false;
        }
        else if (fScannerState != ST_EXTERNAL_SUBSET) {
            // Defect1026: clark test \clarktest\not-wf\sa\165.xml
            if (!sawSpace) {
                emitError(ErrorCode.E_SPACE); //<!ENTITY%x "x">
                return false;
            }
        } else { // EXTERNAL SUBSET
            // Revist: add complex checks for external
            // subset of the form <!ENTITY%x; "">
            // where %x; or nested parameter entities can
            // resolve to not have a space and cause WF errors.
        }

        while (true) {
            try {
                sawSpace = fReader.skippedSpace();
                break;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                //if (fReader.lookingAtSpace())
                //    hey();
                sawSpace = fReader.skippedSpace();
                if (sawSpace)
                    fReader.skipPastSpaces();
                //else
                //    System.err.println("ERROR#8 - need to check for this !!");
                break;
            }
        }
        if (sawSpace) { // <!ENTITY % x "x">
            checkForPEReference(false);
            return true;
        }
        if (fScannerState != ST_EXTERNAL_SUBSET) {
            //
            // Since we have only scanned up to the '%', we could be looking
            // at either <!ENTITY %x; "x"> or <!ENTITY %x "x">.  Without looking
            // ahead, we cannot give the best error for each case.  We will give
            // an E_SPACE error here since PE references are not allowed in the
            // internal subset.  If it is a PERef in the internal subset, they
            // will get another error on the ';' as well.
            //
            emitError(ErrorCode.E_SPACE);
            return true;
        }
        while (true) { // <!ENTITY %%x; "x"> is legal
            try {
                atPERef = fReader.lookingAtChar('%');
                break;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                //if (fReader.lookingAtSpace())
                //    hey();
                sawSpace = fReader.skippedSpace();
                if (sawSpace)
                    fReader.skipPastSpaces();
                //else
                //    System.err.println("ERROR#9 - need to check for this !!");
            }
        }
        if (atPERef) {
            checkForPEReference(false);
            return true;
        }
        //
        // could be <!ENTITY %s; % %s; Name EntityValue>
        // where s is ""
        //
        // could be <!ENTITY %x; EntityValue>
        // where x is "&#37; Name" or
        // where x is "&#37;&#37y;" and y is "Name" or
        // where x is "&#37;y;" and y is "&#37; Name"
        //
        boolean isPEDecl = false;
        while (true) {
            expandPEReference();
            while (true) {
                try {
                    fReader.skipPastSpaces();
                    break;
                } catch (ArrayIndexOutOfBoundsException ex) {
                    popReader();
                    popEntity();
                    //if (fReader.lookingAtSpace())
                    //    hey();
                    fReader.skipPastSpaces();
                }
            }
            atPERef = checkSkippedChar('%', true);
            if (!atPERef)
                return isPEDecl;
            if (!isPEDecl) {
                while (true) {
                    try {
                        sawSpace = fReader.skippedSpace();
                        break;
                    } catch (ArrayIndexOutOfBoundsException ex) {
                        popReader();
                        popEntity();
                        //if (fReader.lookingAtSpace())
                        //    hey();
                    }
                }
                if (sawSpace) {
                    checkForPEReference(false);
                    return true;
                }
                while (true) {
                    try {
                        atPERef = fReader.skippedChar('%');
                        break;
                    } catch (ArrayIndexOutOfBoundsException ex) {
                        popReader();
                        popEntity();
                        //if (fReader.lookingAtSpace())
                        //    hey();
                        sawSpace = fReader.skippedSpace();
                        if (sawSpace)
                            fReader.skipPastSpaces();
                        //else
                        //    System.err.println("ERROR#10 - need to check for this !!");
                    }
                }
                if (atPERef)
                    isPEDecl = true;
            }
        }
    }
    //
    // [29] markupdecl ::= elementdecl | AttlistDecl | EntityDecl
    //                     | NotationDecl | PI | Comment
    //
    // [45] elementdecl ::= '<!ELEMENT' S Name S contentspec S? '>'
    //
    // [52] AttlistDecl ::= '<!ATTLIST' S Name AttDef* S? '>'
    //
    // [70] EntityDecl ::= GEDecl | PEDecl
    // [71] GEDecl ::= '<!ENTITY' S Name S EntityDef S? '>'
    // [72] PEDecl ::= '<!ENTITY' S '%' S Name S PEDef S? '>'
    //
    // [82] NotationDecl ::= '<!NOTATION' S Name S (ExternalID |  PublicID) S? '>'
    //
    // [16] PI ::= '<?' PITarget (S (Char* - (Char* '?>' Char*)))? '?>'
    //
    // [15] Comment ::= '<!--' ((Char - '-') | ('-' (Char - '-')))* '-->'
    //
    // Called after looking at initial '<' character.
    //
    // Note: if extSubset is true, we also handle conditional section as well.
    //
    // [61] conditionalSect ::= includeSect | ignoreSect
    // [62] includeSect ::= '<![' S? 'INCLUDE' S? '[' extSubsetDecl ']]>'
    // [63] ignoreSect ::= '<![' S? 'IGNORE' S? '[' ignoreSectContents* ']]>'
    // [64] ignoreSectContents ::= Ignore ('<![' ignoreSectContents ']]>' Ignore)*
    // [65] Ignore ::= Char* - (Char* ('<![' | ']]>') Char*)
    //
    private void scanMarkupDecl(boolean parseTextDecl) throws Exception
    {
        int olddepth = increaseMarkupDepth();
        try {
            if (fReader.skippedChar('!')) {
                if (fReader.skippedChar('-')) {
                    if (fReader.skippedChar('-')) {
                        scanComment(false);
                    } else {
                        emitError(ErrorCode.E_COM0);
                        fReader.skipPastChar('>');
                        decreaseMarkupDepth();
                    }
                } else if (fReader.skippedChar('[')) {
                    if (fScannerState != ST_EXTERNAL_SUBSET)
                        emitError(ErrorCode.E_DTD4);
                    checkForPEReference(false);
                    if (fReader.skippedString(include_string)) {
                        checkForPEReference(false);
                        if (!checkSkippedChar('[', true)) {
                            emitError(ErrorCode.E_COND0);
                            fReader.skipPastChar('>');
                            decreaseMarkupDepth();
                        } else
                            scanExtSubsetDecl(true);
                    } else if (fReader.skippedString(ignore_string)) {
                        checkForPEReference(false);
                        if (!checkSkippedChar('[', true)) {
                            emitError(ErrorCode.E_COND0);
                            fReader.skipPastChar('>');
                            decreaseMarkupDepth();
                        } else
                            scanIgnoreSectContents();
                    } else {
                        emitError(ErrorCode.E_COND0);
                        fReader.skipPastChar('>');
                        decreaseMarkupDepth();
                    }
                } else if (fReader.skippedString(element_string))
                    scanElementDecl();
                else if (fReader.skippedString(attlist_string))
                    scanAttlistDecl();
                else if (fReader.skippedString(entity_string))
                    scanEntityDecl();
                else if (fReader.skippedString(notation_string))
                    scanNotationDecl();
                else {
                    emitError(ErrorCode.E_DTD3);
                    fReader.skipPastChar('>');
                    decreaseMarkupDepth();
                }
            } else if (fReader.skippedChar('?')) {
                int piTarget = fReader.scanName(' ', -1);
                if (piTarget == -1) {
                    emitError(ErrorCode.E_PI0);
                    fReader.skipPastChar('>');
                    decreaseMarkupDepth();
                } else if ("xml".equals(fStringPool.toString(piTarget))) {
                    if (fReader.skippedSpace()) { // a TextDecl looks like a PI with the target 'xml'
                        if (parseTextDecl) {
                            scanTextDecl();
                        } else {
                            emitError(ErrorCode.E_STRUCT1);
                            fReader.skipPastChar('>');
                            decreaseMarkupDepth();
                        }
                    } else { // a PI target matching 'xml'
                        emitError(ErrorCode.E_PI5);
                        fReader.skipPastChar('>');
                        decreaseMarkupDepth();
                    }
                } else // PI
                    scanPI(piTarget, false);
            } else {
                emitError(ErrorCode.E_COND3);
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
            }
        } catch (ArrayIndexOutOfBoundsException ex) {
            if (peekEntity() == 0)
                decreaseMarkupDepth();
            throw ex;
        }
    }
    //
    // [64] ignoreSectContents ::= Ignore ('<![' ignoreSectContents ']]>' Ignore)*
    // [65] Ignore ::= Char* - (Char* ('<![' | ']]>') Char*)
    //
    private void scanIgnoreSectContents() throws Exception
    {
        int depth = 0;
        try {
            while (true) {
                if (fReader.skippedChar('<')) {
                    //
                    // These tests are split so that we handle cases like
                    // '<<![' and '<!<![' which we might otherwise miss.
                    //
                    if (fReader.skippedChar('!') && fReader.skippedChar('['))
                        depth++;
                } else if (fReader.skippedChar(']')) {
                    //
                    // The same thing goes for ']<![' and '<]]>', etc.
                    //
                    if (fReader.skippedChar(']')) {
                        while (fReader.skippedChar(']')) {
                            /* empty loop body */
                        }
                        if (fReader.skippedChar('>')) {
                            if (depth-- == 0) {
                                decreaseMarkupDepth();
                                return;
                            }
                        }
                    }
                } else if (!fReader.skippedValidChar()) {
                    fReader.skipInvalidChar(ErrorCode.E_INVCHAR0);
                }
            }
        } catch (ArrayIndexOutOfBoundsException ex) {
            emitError(ErrorCode.E_COND1);
            throw ex;
        }
    }
    //
    // [45] elementdecl ::= '<!ELEMENT' S Name S contentspec S? '>'
    //
    private void scanElementDecl() throws Exception
    {

        // are we stopped?
        if (fStop) {
            throw new StopException();
            }

        if (!checkForPEReference(true)) {
            emitError(ErrorCode.E_SPACE);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        int nameOffset = fReader.currentOffset();
        int nameLength;
        while (true) {
            try {
                nameLength = fReader.skipPastName(' ') - nameOffset;
                break;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                //if (fReader.lookingAtSpace())
                //    hey();
                nameOffset = fReader.skipPastSpaces();
            }
        }
        if (nameLength == 0) {
            emitError(ErrorCode.E_ELEM0);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        if (fElementDecl == null)
            fElementDecl = new ElementDecl();
        fElementDecl.elementName = fReader.addSymbol(nameOffset, nameLength);
        if (!checkForPEReference(true)) {
            emitError(ErrorCode.E_SPACE);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        if (!scanContentSpec()) {
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        checkForPEReference(false);
        if (!checkSkippedChar('>', true)) {
            emitError(ErrorCode.E_ELEM2);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        decreaseMarkupDepth();
        int elementIndex = fElementDeclPool.addElementDecl(fElementDecl);
        if (fValidationHandler != null && elementIndex == -1)
            emitError(ErrorCode.V_ELEM3, fElementDecl.elementName);
        else if (fDocumentTypeHandler != null)
            fDocumentTypeHandler.elementDecl(elementIndex);
    }
    //
    // [46] contentspec ::= 'EMPTY' | 'ANY' | Mixed | children
    // [51] Mixed ::= '(' S? '#PCDATA' (S? '|' S? Name)* S? ')*'
    //                | '(' S? '#PCDATA' S? ')'
    // [47] children ::= (choice | seq) ('?' | '*' | '+')?
    // [49] choice ::= '(' S? cp ( S? '|' S? cp )* S? ')'
    // [50] seq ::= '(' S? cp ( S? ',' S? cp )* S? ')'
    // [48] cp ::= (Name | choice | seq) ('?' | '*' | '+')?
    //
    private boolean scanContentSpec() throws Exception
    {
        if (fReader.skippedString(empty_string)) {
            fElementDecl.contentSpecType = ElementDecl.CONTENTSPEC_EMPTY;
            fElementDecl.contentSpec = -1;
            return true;
        }
        if (fReader.skippedString(any_string)) {
            fElementDecl.contentSpecType = ElementDecl.CONTENTSPEC_ANY;
            fElementDecl.contentSpec = -1;
            return true;
        }
        if (!fReader.skippedChar('(')) {
            emitError(ErrorCode.E_ELEM1);
            return false;
        }
        int oldDepth = increaseParenDepth();
        checkForPEReference(false);
        boolean skippedPCDATA = false;
        while (true) {
            try {
                skippedPCDATA = fReader.skippedString(pcdata_string);
                break;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                //if (fReader.lookingAtSpace())
                //    hey();
                fReader.skipPastSpaces();
            }
        }
        if (fContentSpecNode == null)
            fContentSpecNode = new ContentSpecNode();
        fContentSpecNode.otherValue = -1;
        boolean retval = skippedPCDATA ? scanMixed() : scanChildren();
        if (!retval)
            fScannerParenDepth = oldDepth;
        else if (fScannerParenDepth != oldDepth) // REVISIT - should not be needed
			;
//            System.out.println("nesting depth mismatch");
        return retval;
    }
    //
    // [51] Mixed ::= '(' S? '#PCDATA' (S? '|' S? Name)* S? ')*'
    //                | '(' S? '#PCDATA' S? ')'
    //
    // Called after scanning past '(' S? '#PCDATA'
    //
    private boolean scanMixed() throws Exception
    {
        int valueIndex = -1;  // -1 is special value for #PCDATA
        int prevNodeIndex = -1;
        boolean starRequired = false;
        while (true) {
            fContentSpecNode.type = ContentSpecNode.NODE_LEAF;
            fContentSpecNode.value = valueIndex;
            int nodeIndex = fElementDeclPool.addContentSpecNode(fContentSpecNode);
            checkForPEReference(false);
            if (!checkSkippedChar('|', true)) {
                if (!checkSkippedChar(')', true)) {
                    emitError(ErrorCode.E_CS3);
                    return false;
                }
                decreaseParenDepth();
                if (prevNodeIndex != -1) {
                    fContentSpecNode.type = ContentSpecNode.NODE_CHOICE;
                    fContentSpecNode.value = prevNodeIndex;
                    fContentSpecNode.otherValue = nodeIndex;
                    nodeIndex = fElementDeclPool.addContentSpecNode(fContentSpecNode);
                    fContentSpecNode.otherValue = -1;
                }
                if (checkSkippedChar('*', false)) {
                    fContentSpecNode.type = ContentSpecNode.NODE_ZERO_OR_MORE;
                    fContentSpecNode.value = nodeIndex;
                    nodeIndex = fElementDeclPool.addContentSpecNode(fContentSpecNode);
                } else if (starRequired) {
                    emitError(ErrorCode.E_CS9, fStringPool.addString(fElementDeclPool.getContentSpecNodeAsString(nodeIndex)));
                    return false;
                }
                fElementDecl.contentSpecType = ElementDecl.CONTENTSPEC_MIXED;
                fElementDecl.contentSpec = nodeIndex;
                return true;
            }
            if (prevNodeIndex != -1) {
                fContentSpecNode.type = ContentSpecNode.NODE_CHOICE;
                fContentSpecNode.value = prevNodeIndex;
                fContentSpecNode.otherValue = nodeIndex;
                nodeIndex = fElementDeclPool.addContentSpecNode(fContentSpecNode);
                fContentSpecNode.otherValue = -1;
            }
            prevNodeIndex = nodeIndex;
            starRequired = true;
            checkForPEReference(false);
            int nameOffset = fReader.currentOffset();
            int nameLength;
            while (true) {
                try {
                    nameLength = fReader.skipPastName(')') - nameOffset;
                    break;
                } catch (ArrayIndexOutOfBoundsException ex) {
                    popReader();
                    popEntity();
                    //if (fReader.lookingAtSpace())
                    //    hey();
                    nameOffset = fReader.skipPastSpaces();
                }
            }
            if (nameLength == 0) {
                emitError(ErrorCode.E_CS5);
                return false;
            }
            valueIndex = fReader.addSymbol(nameOffset, nameLength);
        }
    }
    //
    // [47] children ::= (choice | seq) ('?' | '*' | '+')?
    // [49] choice ::= '(' S? cp ( S? '|' S? cp )* S? ')'
    // [50] seq ::= '(' S? cp ( S? ',' S? cp )* S? ')'
    // [48] cp ::= (Name | choice | seq) ('?' | '*' | '+')?
    //
    private boolean scanChildren() throws Exception
    {
        int depth = 1;
        initializeContentModelStack(depth);
        while (true) {
            if (checkSkippedChar('(', true)) {
                increaseParenDepth();
                checkForPEReference(false);
                depth++;
                initializeContentModelStack(depth);
                continue;
            }
            int nameOffset = fReader.currentOffset();
            int nameLength;
            while (true) {
                try {
                    nameLength = fReader.skipPastName(')') - nameOffset;
                    break;
                } catch (ArrayIndexOutOfBoundsException ex) {
                    popReader();
                    popEntity();
                    //if (fReader.lookingAtSpace())
                    //    hey();
                    nameOffset = fReader.skipPastSpaces();
                }
            }
            if (nameLength == 0) {
                if (fScannerState != ST_EXTERNAL_SUBSET
                 && fReader.lookingAtChar('%'))  {
                    emitError(ErrorCode.E_PEREF4);
                } else {
                    emitError(ErrorCode.E_CS5);
                }
                return false;
            }
            int valueIndex = fReader.addSymbol(nameOffset, nameLength);
            fContentSpecNode.type = ContentSpecNode.NODE_LEAF;
            fContentSpecNode.value = valueIndex;
            nodeIndexStack[depth] = fElementDeclPool.addContentSpecNode(fContentSpecNode);
            if (fReader.skippedChar('?')) {
                fContentSpecNode.type = ContentSpecNode.NODE_ZERO_OR_ONE;
                fContentSpecNode.value = nodeIndexStack[depth];
                nodeIndexStack[depth] = fElementDeclPool.addContentSpecNode(fContentSpecNode);
            } else if (fReader.skippedChar('*')) {
                fContentSpecNode.type = ContentSpecNode.NODE_ZERO_OR_MORE;
                fContentSpecNode.value = nodeIndexStack[depth];
                nodeIndexStack[depth] = fElementDeclPool.addContentSpecNode(fContentSpecNode);
            } else if (fReader.skippedChar('+')) {
                fContentSpecNode.type = ContentSpecNode.NODE_ONE_OR_MORE;
                fContentSpecNode.value = nodeIndexStack[depth];
                nodeIndexStack[depth] = fElementDeclPool.addContentSpecNode(fContentSpecNode);
            }
            while (true) {
                checkForPEReference(false);
                if (checkSkippedChar('|', true)) {
                    if (prevNodeIndexStack[depth] != -1) {
                        fContentSpecNode.type = opStack[depth];
                        fContentSpecNode.value = prevNodeIndexStack[depth];
                        fContentSpecNode.otherValue = nodeIndexStack[depth];
                        nodeIndexStack[depth] = fElementDeclPool.addContentSpecNode(fContentSpecNode);
                        fContentSpecNode.otherValue = -1;
                    }
                    prevNodeIndexStack[depth] = nodeIndexStack[depth];
                    if (opStack[depth] == ContentSpecNode.NODE_SEQ)
                        emitError(ErrorCode.E_CS6, fStringPool.addString("|"), fStringPool.addString(","));
                    opStack[depth] = (byte)ContentSpecNode.NODE_CHOICE;
                    break;
                } else if (checkSkippedChar(',', true)) {
                    if (prevNodeIndexStack[depth] != -1) {
                        fContentSpecNode.type = opStack[depth];
                        fContentSpecNode.value = prevNodeIndexStack[depth];
                        fContentSpecNode.otherValue = nodeIndexStack[depth];
                        nodeIndexStack[depth] = fElementDeclPool.addContentSpecNode(fContentSpecNode);
                        fContentSpecNode.otherValue = -1;
                    }
                    prevNodeIndexStack[depth] = nodeIndexStack[depth];
                    if (opStack[depth] == ContentSpecNode.NODE_CHOICE)
                        emitError(ErrorCode.E_CS6, fStringPool.addString(","), fStringPool.addString("|"));
                    opStack[depth] = (byte)ContentSpecNode.NODE_SEQ;
                    break;
                } else {
                    if (!checkSkippedChar(')', true))
                        emitError(ErrorCode.E_CS3);
                    decreaseParenDepth();
                    if (prevNodeIndexStack[depth] != -1) {
                        fContentSpecNode.type = opStack[depth];
                        fContentSpecNode.value = prevNodeIndexStack[depth];
                        fContentSpecNode.otherValue = nodeIndexStack[depth];
                        nodeIndexStack[depth] = fElementDeclPool.addContentSpecNode(fContentSpecNode);
                        fContentSpecNode.otherValue = -1;
                    }
                    int nodeIndex = nodeIndexStack[depth--];
                    nodeIndexStack[depth] = nodeIndex;
                    if (fReader.skippedChar('?')) {
                        fContentSpecNode.type = ContentSpecNode.NODE_ZERO_OR_ONE;
                        fContentSpecNode.value = nodeIndexStack[depth];
                        nodeIndexStack[depth] = fElementDeclPool.addContentSpecNode(fContentSpecNode);
                    } else if (fReader.skippedChar('*')) {
                        fContentSpecNode.type = ContentSpecNode.NODE_ZERO_OR_MORE;
                        fContentSpecNode.value = nodeIndexStack[depth];
                        nodeIndexStack[depth] = fElementDeclPool.addContentSpecNode(fContentSpecNode);
                    } else if (fReader.skippedChar('+')) {
                        fContentSpecNode.type = ContentSpecNode.NODE_ONE_OR_MORE;
                        fContentSpecNode.value = nodeIndexStack[depth];
                        nodeIndexStack[depth] = fElementDeclPool.addContentSpecNode(fContentSpecNode);
                    }
                    if (depth == 0) {
                        fElementDecl.contentSpecType = ElementDecl.CONTENTSPEC_CHILDREN;
                        fElementDecl.contentSpec = nodeIndexStack[0];
                        return true;
                    }
                }
            }
            checkForPEReference(false);
        }
    }
    //
    // [52] AttlistDecl ::= '<!ATTLIST' S Name AttDef* S? '>'
    //
    private void scanAttlistDecl() throws Exception
    {
        if (!checkForPEReference(true)) {
            emitError(ErrorCode.E_SPACE);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        int nameOffset = fReader.currentOffset();
        int nameLength;
        while (true) {
            try {
                nameLength = fReader.skipPastName(' ') - nameOffset;
                break;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                //if (fReader.lookingAtSpace())
                //    hey();
                nameOffset = fReader.skipPastSpaces();
            }
        }
        if (nameLength == 0) {
            emitError(ErrorCode.E_ATTL0);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        int nameIndex = fReader.addSymbol(nameOffset, nameLength);
//System.out.println("Attlist: element " + fStringPool.toString(nameIndex) + " (" + nameIndex + ")");
        int elementIndex = fElementDeclPool.addElement(nameIndex);
        if (fAttDef == null)
            fAttDef = new AttDef();
        while (true) {
            boolean sawSpace = checkForPEReference(true);
            while (true) {
                try {
                    if (fReader.skippedChar('>')) {
                        decreaseMarkupDepth();
                        return;
                    }
                    break;
                } catch (ArrayIndexOutOfBoundsException ex) {
                    popReader(); // REVISIT - this complexity needs to be addressed in the next rewrite.
                    popEntity();
                    //if (fReader.lookingAtSpace())
                    //    hey();
                    if (sawSpace)
                        fReader.skippedSpace();
                    else {
                        sawSpace = fReader.skippedSpace();
                        if (sawSpace)
                            fReader.skipPastSpaces();
                        //else
                        //    System.err.println("ERROR#11 - need to check for this !!");
                    }
                }
            }
            if (!sawSpace)
                emitError(ErrorCode.E_SPACE);
            if (!scanAttDef()) {
                fReader.skipPastChar('>');
                decreaseMarkupDepth();
                return;
            }
            int attDefIndex = fElementDeclPool.addAttDef(elementIndex, fAttDef);
            if (attDefIndex != -1 && fDocumentTypeHandler != null)
                fDocumentTypeHandler.attlistDecl(elementIndex, attDefIndex);
        }
    }
    //
    // [53] AttDef ::= S Name S AttType S DefaultDecl
    //
    // We are called after scanning past the first S token.
    //
    private boolean scanAttDef() throws Exception
    {
        int nameOffset = fReader.currentOffset();
        int nameLength = 0;
        while (true) {
            try {
                nameLength = fReader.skipPastName(' ') - nameOffset;
                break;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                //if (fReader.lookingAtSpace())
                //    hey();
                nameOffset = fReader.skipPastSpaces();
            }
        }
        if (nameLength == 0) {
            emitError(ErrorCode.E_ATTD0);
            return false;
        }
        fAttDef.attName = fReader.addSymbol(nameOffset, nameLength);
        if (!checkForPEReference(true)) {
            emitError(ErrorCode.E_SPACE);
            return false;
        }
        if (!scanAttType())
            return false;
        if (!checkForPEReference(true)) {
            emitError(ErrorCode.E_SPACE);
            return false;
        }
        return scanDefaultDecl();
    }
    //
    // [54] AttType ::= StringType | TokenizedType | EnumeratedType
    // [55] StringType ::= 'CDATA'
    // [56] TokenizedType ::= 'ID' | 'IDREF' | 'IDREFS'
    //                        | 'ENTITY' | 'ENTITIES'
    //                        | 'NMTOKEN' | 'NMTOKENS'
    // [57] EnumeratedType ::= NotationType | Enumeration
    // [58] NotationType ::= 'NOTATION' S '(' S? Name (S? '|' S? Name)* S? ')'
    // [59] Enumeration ::= '(' S? Nmtoken (S? '|' S? Nmtoken)* S? ')'
    //
    private boolean scanAttType() throws Exception
    {
        fAttDef.enumeration = -1;
        while (true) {
            try {
                if (fReader.skippedString(cdata_string)) {
                    fAttDef.attType = AttDef.CDATA;
                } else if (fReader.skippedString(id_string)) {
                    if (!fReader.skippedString(ref_string)) {
                        fAttDef.attType = AttDef.ID;
                    } else if (!fReader.skippedChar('S')) {
                        fAttDef.attType = AttDef.IDREF;
                    } else {
                        fAttDef.attType = AttDef.IDREFS;
                    }
                } else if (fReader.skippedString(entit_string)) {
                    if (fReader.skippedChar('Y')) {
                        fAttDef.attType = AttDef.ENTITY;
                    } else if (fReader.skippedString(ies_string)) {
                        fAttDef.attType = AttDef.ENTITIES;
                    } else {
                        emitError(ErrorCode.E_ATTD2);
                        fReader.skipPastNmtoken(' ');
                    }
                } else if (fReader.skippedString(nmtoken_string)) {
                    if (fReader.skippedChar('S')) {
                        fAttDef.attType = AttDef.NMTOKENS;
                    } else {
                        fAttDef.attType = AttDef.NMTOKEN;
                    }
                } else if (fReader.skippedString(notation_string)) {
                    fAttDef.attType = AttDef.NOTATION;
                    if (!scanEnumeration(true))
                        return false;
                } else if (fReader.skippedChar('(')) {
                    increaseParenDepth();
                    fAttDef.attType = AttDef.ENUMERATION;
                    if (!scanEnumeration(false))
                        return false;
                } else {
                    emitError(ErrorCode.E_ATTD2);
                    return false;
                }
                return true;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                //if (fReader.lookingAtSpace())
                //    hey();
                fReader.skipPastSpaces();
            }
        }
    }
    //
    // [59] Enumeration ::= '(' S? Nmtoken (S? '|' S? Nmtoken)* S? ')'
    // [58] NotationType ::= 'NOTATION' S '(' S? Name (S? '|' S? Name)* S? ')'
    //
    // If isNotationType is true, we are called after scanning 'NOTATION'
    // otherwise, we are called after scanning '('
    //
    private boolean scanEnumeration(boolean isNotationType) throws Exception
    {
        if (isNotationType) {
            if (!checkForPEReference(true)) {
                emitError(ErrorCode.E_SPACE);
                return false;
            }
            while (true) {
                try {
                    if (!fReader.skippedChar('(')) {
                        emitError(ErrorCode.E_ENUM0);
                        return false;
                    }
                    break;
                } catch (ArrayIndexOutOfBoundsException ex) {
                    popReader();
                    popEntity();
                    //if (fReader.lookingAtSpace())
                    //    hey();
                    fReader.skipPastSpaces();
                }
            }
            increaseParenDepth();
        }
        fAttDef.enumeration = fStringPool.startStringList();
        boolean choice = false;
        do {
            checkForPEReference(false);
            int nameOffset = fReader.currentOffset();
            int nameLength;
            while (true) {
                try {
                    if (isNotationType)
                        nameLength = fReader.skipPastName(')') - nameOffset;
                    else
                        nameLength = fReader.skipPastNmtoken(')') - nameOffset;
                    break;
                } catch (ArrayIndexOutOfBoundsException ex) {
                    popReader();
                    popEntity();
                    //if (fReader.lookingAtSpace())
                    //    hey();
                    nameOffset = fReader.skipPastSpaces();
                }
            }
            if (nameLength == 0) {
                emitError(isNotationType ? ErrorCode.E_ENUM1 : ErrorCode.E_ENUM2);
                fStringPool.finishStringList(fAttDef.enumeration);
                return false;
            }
            int nameIndex = fReader.addSymbol(nameOffset, nameLength);
            // REVISIT - need to check this at the end of the dtd.
//            if (isNotationType && fEntityPool.lookupNotation(nameIndex) == -1) {
//                emitError(ErrorCode.E_ENUM5, nameIndex);
//            }
            checkForPEReference(false);
            while (true) {
                try {
                    choice = fReader.skippedChar('|');
                    break;
                } catch (ArrayIndexOutOfBoundsException ex) {
                    popReader();
                    popEntity();
                    //if (fReader.lookingAtSpace())
                    //    hey();
                    fReader.skipPastSpaces();
                }
            }
            if (!choice) {
                while (true) {
                    try {
                        if (!fReader.skippedChar(')')) {
                            emitError(ErrorCode.E_ENUM3);
                            fStringPool.finishStringList(fAttDef.enumeration);
                            return false;
                        }
                        break;
                    } catch (ArrayIndexOutOfBoundsException ex) {
                        popReader();
                        popEntity();
                        //if (fReader.lookingAtSpace())
                        //    hey();
                        fReader.skipPastSpaces();
                    }
                }
                decreaseParenDepth();
            }
            fStringPool.addStringToList(fAttDef.enumeration, nameIndex);
        } while (choice);
        fStringPool.finishStringList(fAttDef.enumeration);
        return true;
    }
    //
    // [60] DefaultDecl ::= '#REQUIRED' | '#IMPLIED'
    //                      | (('#FIXED' S)? AttValue)
    //
    private boolean scanDefaultDecl() throws Exception
    {
        if (fReader.skippedString(required_string)) {
            fAttDef.attDefaultType = AttDef.REQUIRED;
            fAttDef.attValue = -1;
            return true;
        }
        if (fReader.skippedString(implied_string)) {
            fAttDef.attDefaultType = AttDef.IMPLIED;
            fAttDef.attValue = -1;
            return true;
        }
        if (fReader.skippedString(fixed_string)) {
            fAttDef.attDefaultType = AttDef.FIXED;
            if (!fReader.skippedSpace()) {
                emitError(ErrorCode.E_SPACE);
                return false;
            }
            fReader.skipPastSpaces();
        } else
            fAttDef.attDefaultType = AttDef.DEFAULT;
        if (!scanAttValue(-1, -1, fAttDef.attType, true, fAttDef.attDefaultType == AttDef.FIXED, true)) {
			emitError(ErrorCode.E_ATTD5);
            return false;
		}
		fAttDef.attValue = fAttValueIndex;
        return true;
    }
    //
    // [82] NotationDecl ::= '<!NOTATION' S Name S (ExternalID |  PublicID) S? '>'
    //
    private void scanNotationDecl() throws Exception
    {
        if (!checkForPEReference(true)) {
            emitError(ErrorCode.E_SPACE);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        int nameOffset = fReader.currentOffset();
        int nameLength;
        while (true) {
            try {
                nameLength = fReader.skipPastName(' ') - nameOffset;
                break;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                //if (fReader.lookingAtSpace())
                //    hey();
                nameOffset = fReader.skipPastSpaces();
            }
        }
        if (nameLength == 0) {
            emitError(ErrorCode.E_NOT0);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        if (fNotationDecl == null)
            fNotationDecl = new NotationDecl();
        fNotationDecl.notationName = fReader.addSymbol(nameOffset, nameLength);
        if (!checkForPEReference(true)) {
            emitError(ErrorCode.E_SPACE);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        if (!scanExternalIDorPublicID(true)) {
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        fNotationDecl.publicId = fPubidLiteral;
        fNotationDecl.systemId = fSystemLiteral;
        checkForPEReference(false);
        while (true) {
            try {
                if (!fReader.skippedChar('>')) {
                    emitError(ErrorCode.E_NOT1); // REVISIT - new error message
                    fReader.skipPastChar('>');
                    decreaseMarkupDepth();
                    return;
                }
                decreaseMarkupDepth();
                break;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                //if (fReader.lookingAtSpace())
                //    hey();
                fReader.skipPastSpaces();
            }
        }
        //
        // REVISIT - I am making an arbitrary decision that is not covered by the
        //  spec (or I missed it).  If we see a second NotationDecl for the same name,
        //  we will ignore the latter declaration.  This seems consistent with the
        //  usual behavior of such things in the spec, but it might also be good to
        //  generate a warning that this has occurred.  The addNotation() method will
        //  return -1 if we already have a declaration for this name.  This will slow
        //  down the parser in direct relation to the number of NotationDecl names.
        //
        int notationIndex = fEntityPool.addNotationDecl(fNotationDecl);
        if (notationIndex != -1 && fDocumentTypeHandler != null)
            fDocumentTypeHandler.notationDecl(notationIndex);
    }
    //
    // [70] EntityDecl ::= GEDecl | PEDecl
    // [71] GEDecl ::= '<!ENTITY' S Name S EntityDef S? '>'
    // [72] PEDecl ::= '<!ENTITY' S '%' S Name S PEDef S? '>'
    // [73] EntityDef ::= EntityValue | (ExternalID NDataDecl?)
    // [74] PEDef ::= EntityValue | ExternalID
    // [75] ExternalID ::= 'SYSTEM' S SystemLiteral
    //                     | 'PUBLIC' S PubidLiteral S SystemLiteral
    // [76] NDataDecl ::= S 'NDATA' S Name
    //  [9] EntityValue ::= '"' ([^%&"] | PEReference | Reference)* '"'
    //                      | "'" ([^%&'] | PEReference | Reference)* "'"
    //
    // Called after scanning 'ENTITY'
    //
    private void scanEntityDecl() throws Exception
    {
        boolean isPEDecl = checkForPEDecl();
        int nameOffset = fReader.currentOffset();
        int nameLength;
        while (true) {
            try {
                nameLength = fReader.skipPastName(' ') - nameOffset;
                break;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                //if (fReader.lookingAtSpace())
                //    hey();
                nameOffset = fReader.skipPastSpaces();
            }
        }
        if (nameLength == 0) {
            emitError(ErrorCode.E_ENTITY1);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        int nameIndex = fReader.addSymbol(nameOffset, nameLength);
        boolean newEntity = pushEntity(isPEDecl ? PARAMETER_ENTITY_TYPE : GENERAL_ENTITY_TYPE, nameIndex);
        if (!newEntity) {
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            return;
        }
        if (fEntityDecl == null)
            fEntityDecl = new EntityDecl();
        fEntityDecl.entityName = nameIndex;
        fEntityDecl.entityValue = -1;
        fEntityDecl.publicId = -1;
        fEntityDecl.systemId = -1;
        fEntityDecl.notationName = -1;
        if (!checkForPEReference(true)) {
            emitError(ErrorCode.E_SPACE);
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            popEntity();
            return;
        }
        if (!(isPEDecl ? scanPEDef() : scanEntityDef())) {
            fReader.skipPastChar('>');
            decreaseMarkupDepth();
            popEntity();
            return;
        }
        checkForPEReference(false);
        while (true) {
            try {
                if (!fReader.skippedChar('>')) {
                    emitError(ErrorCode.E_ENTITY4);
                    fReader.skipPastChar('>');
                    decreaseMarkupDepth();
                    popEntity();
                    return;
                }
                decreaseMarkupDepth();
                break;
            } catch (ArrayIndexOutOfBoundsException ex) {
                popReader();
                popEntity();
                //if (fReader.lookingAtSpace())
                //    hey();
                fReader.skipPastSpaces();
            }
        }
        popEntity();
        if (isPEDecl) {
            if (fParameterEntityPool == null)
                fParameterEntityPool = new DefaultEntityPool(fParserState, false);
            fParameterEntityPool.addEntityDecl(fEntityDecl);
        } else {
            int entityPoolIndex = fEntityPool.addEntityDecl(fEntityDecl);
            //revisit: This code assumes if the entityValue is defined, the
            // internalEntityDecl is called, otherwise if the notationName
            // is not defined, then the externalEntityDecl is called, but if
            // notationName is defined unparsedEntityDecl is called.
            if (fDocumentTypeHandler != null) {
                if (fEntityDecl.entityValue == -1) {
                    if (fEntityDecl.notationName == -1) {
                        fDocumentTypeHandler.externalEntityDecl(entityPoolIndex);
                    } else {
                        fDocumentTypeHandler.unparsedEntityDecl(entityPoolIndex);
                    }
                } else {
                    fDocumentTypeHandler.internalEntityDecl(entityPoolIndex);
                }
            }
        }
    }

    //
    // [73] EntityDef ::= EntityValue | (ExternalID NDataDecl?)
    // [76] NDataDecl ::= S 'NDATA' S Name
    //
    private boolean scanEntityDef() throws Exception
    {
        boolean single;
        if ((single = fReader.skippedChar('\'')) || fReader.skippedChar('\"'))
            return scanEntityValue(single);
        if (!scanExternalID(true))
            return false;
        fEntityDecl.publicId = fPubidLiteral;
        fEntityDecl.systemId = fSystemLiteral;
        if (!fReader.skippedSpace())
            return true;
        fReader.skipPastSpaces();
        if (fReader.skippedString(ndata_string)) {
            if (!fReader.skippedSpace()) {
                emitError(ErrorCode.E_SPACE);
                return false;
            }
            int ndataOffset = fReader.skipPastSpaces();
            int ndataLength = fReader.skipPastName('>') - ndataOffset;
            if (ndataLength == 0) {
                emitError(ErrorCode.E_ENTITY5);
                return false;
            }
            fEntityDecl.notationName = fReader.addSymbol(ndataOffset, ndataLength);
            /*** BUG ***
            // Defect #530:
            //   Entity and notation declarations are not order dependent.
            //   This check needs to be performed after the entire DTD has
            //   been parsed. -Ac
            if (fEntityPool.lookupNotation(fEntityDecl.notationName) == -1)
                emitError(ErrorCode.V_ENTITY8, fEntityDecl.notationName);
            /***/
        }
        return true;
    }
    //
    // [74] PEDef ::= EntityValue | ExternalID
    //
    private boolean scanPEDef() throws Exception
    {
        boolean single;
        if ((single = fReader.skippedChar('\'')) || fReader.skippedChar('\"'))
            return scanEntityValue(single);
        if (!scanExternalID(true))
            return false;
        fEntityDecl.publicId = fPubidLiteral;
        fEntityDecl.systemId = fSystemLiteral;
        return true;
    }
    //
    //  [9] EntityValue ::= '"' ([^%&"] | PEReference | Reference)* '"'
    //                      | "'" ([^%&'] | PEReference | Reference)* "'"
    //
    private boolean scanEntityValue(boolean single) throws Exception
    {
        int offset = fReader.currentOffset(); // skipped either '\'' or '\"' in caller
        char qchar = single ? '\'' : '\"';
        int toCopy = offset;
        int dataOffset = fLiteralData.length();
        while (true) {
            if (fReader.skippedChar(qchar)) {
                if (offset - toCopy > 0)
                    fReader.append(fLiteralData, toCopy, offset - toCopy);
                break;
            }
            if (fReader.skippedChar('&')) {
                if (offset - toCopy > 0)
                    fReader.append(fLiteralData, toCopy, offset - toCopy);
                offset = referenceInEntityValue(offset, true);
                toCopy = offset;
            } else if (fReader.skippedChar('%')) {
                if (offset - toCopy > 0)
                    fReader.append(fLiteralData, toCopy, offset - toCopy);
                if (!copyPEReference()) {
                    fReader.skipToChar(qchar);
                    return false;
                }
                offset = fReader.currentOffset();
                toCopy = offset;
            } else if (!fReader.skippedValidChar()) {
                fReader.skipInvalidChar(ErrorCode.E_ENTITY3);
                fReader.skipToChar(qchar);
                return false;
            } else
                offset = fReader.currentOffset();
        }
        int dataLength = fLiteralData.length() - dataOffset;
        fEntityDecl.entityValue = fLiteralData.addString(dataOffset, dataLength);
        return true;
    }
    //
    // reader stack
    //
    private void growReaderStack() {
        XMLReader[] newReaderStack = new XMLReader[fReaderStackDepth * 2];
        System.arraycopy(fReaderStack, 0, newReaderStack, 0, fReaderStackDepth);
        fReaderStack = newReaderStack;
        InputSource[] newReaderSource = new InputSource[fReaderStackDepth * 2];
        System.arraycopy(fReaderSource, 0, newReaderSource, 0, fReaderStackDepth);
        fReaderSource = newReaderSource;
        int[] newNestingStack = new int[fReaderStackDepth * 2];
        System.arraycopy(fReaderMarkupDepth, 0, newNestingStack, 0, fReaderStackDepth);
        fReaderMarkupDepth = newNestingStack;
        newNestingStack = new int[fReaderStackDepth * 2];
        System.arraycopy(fReaderParenDepth, 0, newNestingStack, 0, fReaderStackDepth);
        fReaderParenDepth = newNestingStack;
    }
    private int pushReader(XMLReader newReader, InputSource source) {
        if (fReaderStackDepth == fReaderStack.length)
            growReaderStack();
        fReaderStack[fReaderStackDepth] = fReader;
        fReaderSource[fReaderStackDepth] = source;
        fReaderMarkupDepth[fReaderStackDepth] = fScannerMarkupDepth;
        fReaderParenDepth[fReaderStackDepth] = fScannerParenDepth;
        fReader = newReader;
        if (DEBUG_PAREN_DEPTH) {
            System.out.println("(" + fReaderStackDepth + ") [" + fReaderParenDepth[fReaderStackDepth] + "] Reader " + fReader +
            " (file " + fReader.getSystemId() + ", line " + fReader.getLineNumber() + ", col " + fReader.getColumnNumber() + ")");
        }
        if (DEBUG_MARKUP_DEPTH) {
            System.out.println("<" + fReaderStackDepth + "> [" + fReaderMarkupDepth[fReaderStackDepth] + "] Reader " + fReader +
            " (file " + fReader.getSystemId() + ", line " + fReader.getLineNumber() + ", col " + fReader.getColumnNumber() + ")");
        }
        return ++fReaderStackDepth;
    }
    private static final boolean DEBUG_PAREN_DEPTH = false;
    private static final boolean DEBUG_MARKUP_DEPTH = false;
    /**
     * Pops the top most reader from the stack and notifies the
     * entity handler.
     */
    private void popReader() throws Exception {
        if (fReaderStackDepth == fActiveReaderLimit)
            throw new ArrayIndexOutOfBoundsException();
        fReaderStackDepth--;
        int entityNameIndex = peekEntity();
        if (entityNameIndex != 0) {
            if (DEBUG_PAREN_DEPTH) {
                System.out.println("(" + fReaderStackDepth + ") [" + fReaderParenDepth[fReaderStackDepth] + "] Reader " + fReader +
                " (file " + fReader.getSystemId() + ", line " + fReader.getLineNumber() + ", col " + fReader.getColumnNumber() + ")");
            }
            if (fValidationHandler != null && fScannerParenDepth != fReaderParenDepth[fReaderStackDepth]) {
                if (DEBUG_PAREN_DEPTH) {
                    System.out.println("(" + fReaderStackDepth + ") [" + fScannerParenDepth + "] Reader " + fReader);
                }
                emitError(ErrorCode.V_PEREF7, entityNameIndex);
            }
            if (DEBUG_MARKUP_DEPTH) {
                System.out.println("<" + fReaderStackDepth + "> [" + fReaderMarkupDepth[fReaderStackDepth] + "] Reader " + fReader +
                " (file " + fReader.getSystemId() + ", line " + fReader.getLineNumber() + ", col " + fReader.getColumnNumber() + ")");
            }
            if (fScannerMarkupDepth != fReaderMarkupDepth[fReaderStackDepth]) {
                if (DEBUG_MARKUP_DEPTH) {
                    System.out.println("<" + fReaderStackDepth + "> [" + fScannerMarkupDepth + "] Reader " + fReader);
                }
                if (peekEntityType() == GENERAL_ENTITY_TYPE)
                    emitError(ErrorCode.E_TAGh);
                else
                    emitError(ErrorCode.E_PEREF9, entityNameIndex);
            }
        }
        fReader = fReaderStack[fReaderStackDepth];
        InputSource source = fReaderSource[fReaderStackDepth];
        if (source != null) {
            fEntityHandler.endInputSource(source);
            fReaderSource[fReaderStackDepth] = null;
        }
    }
    private int setActiveReaderLimit(int newLimit) {
        int oldLimit = fActiveReaderLimit;
        fActiveReaderLimit = newLimit;
        return oldLimit;
    }
    private int increaseParenDepth() {
        if (DEBUG_PAREN_DEPTH) {
            System.out.println("(( " + fScannerParenDepth + " => " + (fScannerParenDepth + 1) +
            " (file " + fReader.getSystemId() + ", line " + fReader.getLineNumber() + ", col " + fReader.getColumnNumber() + ")");
        }
        return fScannerParenDepth++;
    }
    private int decreaseParenDepth() {
        if (DEBUG_PAREN_DEPTH) {
            System.out.println(")) " + fScannerParenDepth + " => " + (fScannerParenDepth - 1) +
            " (file " + fReader.getSystemId() + ", line " + fReader.getLineNumber() + ", col " + fReader.getColumnNumber() + ")");
        }
        return fScannerParenDepth--;
    }
    private int increaseMarkupDepth() {
        if (DEBUG_MARKUP_DEPTH) {
            System.out.println("<< " + fScannerMarkupDepth + " => " + (fScannerMarkupDepth + 1) +
            " (file " + fReader.getSystemId() + ", line " + fReader.getLineNumber() + ", col " + fReader.getColumnNumber() + ")");
        }
        return fScannerMarkupDepth++;
    }
    private int decreaseMarkupDepth() {
        if (DEBUG_MARKUP_DEPTH) {
            System.out.println(">> " + fScannerMarkupDepth + " => " + (fScannerMarkupDepth - 1) +
            " (file " + fReader.getSystemId() + ", line " + fReader.getLineNumber() + ", col " + fReader.getColumnNumber() + ")");
        }
        return fScannerMarkupDepth--;
    }
    //
    // element stack
    //
    private void pushElementName(int elementNameIndex) {
        if (fElementDepth >= 0) {
            int[] children = fElementChildren[fElementDepth];
            int childCount = fElementChildCount[fElementDepth];
            if (children == null) {
                children = fElementChildren[fElementDepth] = new int[8];
                childCount = 0; // should really assert this...
            } else if (childCount == children.length) {
                int[] newChildren = new int[childCount * 2];
                System.arraycopy(children, 0, newChildren, 0, childCount);
                children = fElementChildren[fElementDepth] = newChildren;
            }
            children[childCount++] = elementNameIndex;
            fElementChildCount[fElementDepth] = childCount;
        }
        fElementDepth++;
        if (fElementDepth == fElementNameStack.length) {
            int[] newStack = new int[fElementDepth * 2];
            System.arraycopy(fElementNameStack, 0, newStack, 0, fElementDepth);
            fElementNameStack = newStack;
            newStack = new int[fElementDepth * 2];
            System.arraycopy(fElementChildCount, 0, newStack, 0, fElementDepth);
            fElementChildCount = newStack;
            int[][] newContentStack = new int[fElementDepth * 2][];
            System.arraycopy(fElementChildren, 0, newContentStack, 0, fElementDepth);
            fElementChildren = newContentStack;
        }
        fElementNameStack[fElementDepth] = elementNameIndex;
        fElementChildCount[fElementDepth] = 0;
    }
    private int elementStackDepth() {
        return fElementDepth;
    }
    private void charDataInContent() {
        int[] children = fElementChildren[fElementDepth];
        int childCount = fElementChildCount[fElementDepth];
        if (children == null) {
            children = fElementChildren[fElementDepth] = new int[8];
            childCount = 0; // should really assert this...
        } else if (childCount == children.length) {
            int[] newChildren = new int[childCount * 2];
            System.arraycopy(children, 0, newChildren, 0, childCount);
            children = fElementChildren[fElementDepth] = newChildren;
        }
        children[childCount++] = -1;
        fElementChildCount[fElementDepth] = childCount;
    }
    public int getCurrentContentSpecType() {
        return fElementDeclPool.getContentSpecType(fElementDeclPool.getElement(fElementNameStack[fElementDepth]));
    }
    private int peekElementName() {
        return fElementNameStack[fElementDepth];
    }
    private int peekChildCount() {
        return fElementChildCount[fElementDepth];
    }
    private int[] peekChildren() {
        return fElementChildren[fElementDepth];
    }
    private boolean popElementName() {
        if (fElementDepth < 0)
            throw new RuntimeException("Element stack underflow");
        return --fElementDepth >= 0;
    }
    //
    // entity stack
    //
    private boolean pushEntity(byte entityType, int entityNameIndex) throws Exception {
        for (int i = 0; i < fEntityStackDepth; i++) {
            if (fEntityNameStack[i] == entityNameIndex && fEntityTypeStack[i] == entityType) {
                emitError(entityType == GENERAL_ENTITY_TYPE ? ErrorCode.E_TAGi : ErrorCode.E_PEREF5, entityNameIndex,
                          entityReferencePath(entityType, entityNameIndex));
                return false;
            }
        }
        if (fEntityTypeStack == null) {
            fEntityTypeStack = new byte[8];
            fEntityNameStack = new int[8];
        } else if (fEntityStackDepth == fEntityTypeStack.length) {
            byte[] newTypeStack = new byte[fEntityStackDepth * 2];
            System.arraycopy(fEntityTypeStack, 0, newTypeStack, 0, fEntityStackDepth);
            fEntityTypeStack = newTypeStack;
            int[] newIndexStack = new int[fEntityStackDepth * 2];
            System.arraycopy(fEntityNameStack, 0, newIndexStack, 0, fEntityStackDepth);
            fEntityNameStack = newIndexStack;
        }
        fEntityTypeStack[fEntityStackDepth] = entityType;
        fEntityNameStack[fEntityStackDepth] = entityNameIndex;
        fEntityStackDepth++;
        return true;
    }
    private int peekEntity() {
        if (fEntityStackDepth == 0)
            return 0; // ""
        return fEntityNameStack[fEntityStackDepth-1];
    }
    private int peekEntityType() {
        if (fEntityStackDepth == 0)
            return GENERAL_ENTITY_TYPE;
        return fEntityTypeStack[fEntityStackDepth-1];
    }
    private int entityReferencePath(byte entityType, int entityNameIndex) {
        StringBuffer sb = new StringBuffer();
        sb.append("(top-level)");
        for (int i = 0; i < fEntityStackDepth; i++) {
            sb.append('-');
            sb.append(fEntityTypeStack[i] == GENERAL_ENTITY_TYPE ? '&' : '%');
            sb.append(fStringPool.toString(fEntityNameStack[i]));
            sb.append(';');
        }
        sb.append('-');
        sb.append(entityType == GENERAL_ENTITY_TYPE ? '&' : '%');
        sb.append(fStringPool.toString(entityNameIndex));
        sb.append(';');
        return fStringPool.addString(sb.toString());
    }
    private void popEntity() {
        fEntityStackDepth--;
    }
    //
    // content model stack
    //
    private void initializeContentModelStack(int depth) {
        if (opStack == null) {
            opStack = new byte[8];
            nodeIndexStack = new int[8];
            prevNodeIndexStack = new int[8];
        } else if (depth == opStack.length) {
            byte[] newOpStack = new byte[depth * 2];
            System.arraycopy(opStack, 0, newOpStack, 0, depth);
            opStack = newOpStack;
            int[] newIndexStack = new int[depth * 2];
            System.arraycopy(nodeIndexStack, 0, newIndexStack, 0, depth);
            nodeIndexStack = newIndexStack;
            newIndexStack = new int[depth * 2];
            System.arraycopy(prevNodeIndexStack, 0, newIndexStack, 0, depth);
            prevNodeIndexStack = newIndexStack;
        }
        opStack[depth] = 0;
        nodeIndexStack[depth] = -1;
        prevNodeIndexStack[depth] = -1;
    }

    //
    // Classes
    //

    /**
     * Denotes a premature end of scanning caused by user calling
     * the <code>XMLScanner.stop()</code> method.
     *
     * @author  Andy Clark, IBM
     */
    private class StopException
        extends Exception
        {

        //
        // Constructors
        //

        /** Default constructor. */
        public StopException() {
            // TODO: Put translation string "stopped at {0}:{1} in {2}" in
            //       resource bundle. -Ac
            super("stopped at "+
                  DefaultScanner.this.getLineNumber()+
                  ':'+
                  DefaultScanner.this.getColumnNumber()+
                  " in "+
                  DefaultScanner.this.getSystemId()
                  );
            }

        } // class StopException

    //
    // Appendum:
    //
    // From the Standard:
    //
    // 2. Documents
    // 2.1 Well-Formed XML Documents
    //  [1] document ::= prolog element Misc*
    // 2.2 Characters
    // Character Range
    //  [2] Char ::= #x9 | #xA | #xD | [#x20-#xD7FF]        // any Unicode character, excluding the
    //               | [#xE000-#xFFFD] | [#x10000-#x10FFFF] // surrogate blocks, FFFE, and FFFF.
    // 2.3 Common Syntactic Constructs
    // White Space
    //  [3] S ::= (#x20 | #x9 | #xD | #xA)+
    // Names and Tokens
    //  [4] NameChar ::= Letter | Digit | '.' | '-' | '_' | ':' | CombiningChar | Extender
    //  [5] Name ::= (Letter | '_' | ':') (NameChar)*
    //  [6] Names ::= Name (S Name)*
    //  [7] Nmtoken ::= (NameChar)+
    //  [8] Nmtokens ::= Nmtoken (S Nmtoken)*
    // Literals
    //  [9] EntityValue ::= '"' ([^%&"] | PEReference | Reference)* '"'
    //                      | "'" ([^%&'] | PEReference | Reference)* "'"
    // [10] AttValue ::= '"' ([^<&"] | Reference)* '"'
    //                   | "'" ([^<&'] | Reference)* "'"
    // [11] SystemLiteral ::= ('"' [^"]* '"') | ("'" [^']* "'")
    // [12] PubidLiteral ::= '"' PubidChar* '"' | "'" (PubidChar - "'")* "'"
    // [13] PubidChar ::= #x20 | #xD | #xA | [a-zA-Z0-9] | [-'()+,./:=?;!*#@$_%]
    // 2.4 Character Data and Markup
    // [14] CharData ::= [^<&]* - ([^<&]* ']]>' [^<&]*)
    // 2.5 Comments
    // [15] Comment ::= '<!--' ((Char - '-') | ('-' (Char - '-')))* '-->'
    // 2.6 Processing Instructions
    // [16] PI ::= '<?' PITarget (S (Char* - (Char* '?>' Char*)))? '?>'
    // [17] PITarget ::= Name - (('X' | 'x') ('M' | 'm') ('L' | 'l'))
    // 2.7 CDATA Sections
    // [18] CDSect ::= CDStart CData CDEnd
    // [19] CDStart ::= '<![CDATA['
    // [20] CData ::= (Char* - (Char* ']]>' Char*))
    // [21] CDEnd ::= ']]>'
    // 2.8 Prolog and Document Type Declaration
    // Prolog
    // [22] prolog ::= XMLDecl? Misc* (doctypedecl Misc*)?
    // [23] XMLDecl ::= '<?xml' VersionInfo EncodingDecl? SDDecl? S? '?>'
    // [24] VersionInfo ::= S 'version' Eq (' VersionNum ' | " VersionNum ")
    // [25] Eq ::= S? '=' S?
    // [26] VersionNum ::= ([a-zA-Z0-9_.:] | '-')+
    // [27] Misc ::= Comment | PI |  S
    // Document Type Definition
    // [28] doctypedecl ::= '<!DOCTYPE' S Name (S ExternalID)? S?
    //                      ('[' (markupdecl | PEReference | S)* ']' S?)? '>'
    // [29] markupdecl ::= elementdecl | AttlistDecl | EntityDecl
    //                     | NotationDecl | PI | Comment
    // External Subset
    // [30] extSubset ::= TextDecl? extSubsetDecl
    // [31] extSubsetDecl ::= ( markupdecl | conditionalSect | PEReference | S )*
    // 2.9 Standalone Document Declaration
    // [32] SDDecl ::= S 'standalone' Eq (("'" ('yes' | 'no') "'")
    //                 | ('"' ('yes' | 'no') '"'))
    // 2.10 White Space Handling
    // 2.11 End-of-Line Handling
    // 2.12 Language Identification
    // [33] LanguageID ::= Langcode ('-' Subcode)*
    // [34] Langcode ::= ISO639Code |  IanaCode |  UserCode
    // [35] ISO639Code ::= ([a-z] | [A-Z]) ([a-z] | [A-Z])
    // [36] IanaCode ::= ('i' | 'I') '-' ([a-z] | [A-Z])+
    // [37] UserCode ::= ('x' | 'X') '-' ([a-z] | [A-Z])+
    // [38] Subcode ::= ([a-z] | [A-Z])+
    // 3. Logical Structures
    // Element
    // [39] element ::= EmptyElemTag
    //                  | STag content ETag
    // 3.1 Start-Tags, End-Tags, and Empty-Element Tags
    // Start-tag
    // [40] STag ::= '<' Name (S Attribute)* S? '>'
    // [41] Attribute ::= Name Eq AttValue
    // End-tag
    // [42] ETag ::= '</' Name S? '>'
    // Content of Elements
    // [43] content ::= (element | CharData | Reference | CDSect | PI | Comment)*
    // Tags for Empty Elements
    // [44] EmptyElemTag ::= '<' Name (S Attribute)* S? '/>'
    // 3.2 Element Type Declarations
    // Element Type Declaration
    // [45] elementdecl ::= '<!ELEMENT' S Name S contentspec S? '>'
    // [46] contentspec ::= 'EMPTY' | 'ANY' | Mixed | children
    // 3.2.1 Element Content
    // Element-content Models
    // [47] children ::= (choice | seq) ('?' | '*' | '+')?
    // [48] cp ::= (Name | choice | seq) ('?' | '*' | '+')?
    // [49] choice ::= '(' S? cp ( S? '|' S? cp )* S? ')'
    // [50] seq ::= '(' S? cp ( S? ',' S? cp )* S? ')'
    // 3.2.2 Mixed Content
    // Mixed-content Declaration
    // [51] Mixed ::= '(' S? '#PCDATA' (S? '|' S? Name)* S? ')*'
    //                | '(' S? '#PCDATA' S? ')'
    // 3.3 Attribute-List Declarations
    // Attribute-list Declaration
    // [52] AttlistDecl ::= '<!ATTLIST' S Name AttDef* S? '>'
    // [53] AttDef ::= S Name S AttType S DefaultDecl
    // 3.3.1 Attribute Types
    // Attribute Types
    // [54] AttType ::= StringType | TokenizedType | EnumeratedType
    // [55] StringType ::= 'CDATA'
    // [56] TokenizedType ::= 'ID'
    //                        | 'IDREF'
    //                        | 'IDREFS'
    //                        | 'ENTITY'
    //                        | 'ENTITIES'
    //                        | 'NMTOKEN'
    //                        | 'NMTOKENS'
    // Enumerated Attribute Types
    // [57] EnumeratedType ::= NotationType | Enumeration
    // [58] NotationType ::= 'NOTATION' S '(' S? Name (S? '|' S? Name)* S? ')'
    // [59] Enumeration ::= '(' S? Nmtoken (S? '|' S? Nmtoken)* S? ')'
    // 3.3.2 Attribute Defaults
    // [60] DefaultDecl ::= '#REQUIRED' | '#IMPLIED'
    //                      | (('#FIXED' S)? AttValue)
    // 3.3.3 Attribute-Value Normalization
    // 3.4 Conditional Sections
    // [61] conditionalSect ::= includeSect | ignoreSect
    // [62] includeSect ::= '<![' S? 'INCLUDE' S? '[' extSubsetDecl ']]>'
    // [63] ignoreSect ::= '<![' S? 'IGNORE' S? '[' ignoreSectContents* ']]>'
    // [64] ignoreSectContents ::= Ignore ('<![' ignoreSectContents ']]>' Ignore)*
    // [65] Ignore ::= Char* - (Char* ('<![' | ']]>') Char*)
    // 4. Physical Structures
    // 4.1 Character and Entity References
    // Character Reference
    // [66] CharRef ::= '&#' [0-9]+ ';' | '&#x' [0-9a-fA-F]+ ';'
    // Entity Reference
    // [67] Reference ::= EntityRef | CharRef
    // [68] EntityRef ::= '&' Name ';'
    // [69] PEReference ::= '%' Name ';'
    // 4.2 Entity Declarations
    // [70] EntityDecl ::= GEDecl | PEDecl
    // [71] GEDecl ::= '<!ENTITY' S Name S EntityDef S? '>'
    // [72] PEDecl ::= '<!ENTITY' S '%' S Name S PEDef S? '>'
    // [73] EntityDef ::= EntityValue | (ExternalID NDataDecl?)
    // [74] PEDef ::= EntityValue | ExternalID
    // 4.2.1 Internal Entities
    // 4.2.2 External Entities
    // External Entity Declaration
    // [75] ExternalID ::= 'SYSTEM' S SystemLiteral
    //                     | 'PUBLIC' S PubidLiteral S SystemLiteral
    // [76] NDataDecl ::= S 'NDATA' S Name
    // 4.3 Parsed Entities
    // 4.3.1 The Text Declaration
    // [77] TextDecl ::= '<?xml' VersionInfo? EncodingDecl S? '?>'
    // 4.3.2 Well-Formed Parsed Entities
    // [78] extParsedEnt ::= TextDecl? content
    // [79] extPE ::= TextDecl? extSubsetDecl
    // 4.3.3 Character Encoding in Entities
    // Encoding Declaration
    // [80] EncodingDecl ::= S 'encoding' Eq ('"' EncName '"' |  "'" EncName "'" )
    // [81] EncName ::= [A-Za-z] ([A-Za-z0-9._] | '-')*
    // 4.4 XML Processor Treatment of Entities and References
    // 4.4.1 Not Recognized
    // 4.4.2 Included
    // 4.4.3 Included If Validating
    // 4.4.4 Forbidden
    // 4.4.5 Included in Literal
    // 4.4.6 Notify
    // 4.4.7 Bypassed
    // 4.4.8 Included as PE
    // 4.5 Construction of Internal Entity Replacement Text
    // 4.6 Predefined Entities
    // 4.7 Notation Declarations
    // [82] NotationDecl ::= '<!NOTATION' S Name S (ExternalID |  PublicID) S? '>'
    // [83] PublicID ::= 'PUBLIC' S PubidLiteral
    // 4.8 Document Entity
    // 5. Conformance
    // 5.1 Validating and Non-Validating Processors
    // 5.2 Using XML Processors
    // 6. Notation
    // Appendices
    // A. References
    // B. Character Classes
    // [84] Letter ::= BaseChar | Ideographic
    // [85] BaseChar ::= <see standard>
    // [86] Ideographic ::= <see standard>
    // [87] CombiningChar ::= <see standard>
    // [88] Digit ::= <see standard>
    // [89] Extender ::= <see standard>
    // C.-G. ... (Non-Normative)
    //
}
