/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */
package com.ibm.xml.internal;

import com.ibm.xml.framework.*;

import java.util.*;

/**
 *
 * @version Revision: 75 1.9 src/com/ibm/xml/internal/DefaultValidationHandler.java, parser, xml4j2, xml4j2_0_13 
 */
public class DefaultValidationHandler implements XMLValidationHandler
{
    // -----------------------------------------------------------------------
    //  Constructors
    // -----------------------------------------------------------------------
    public DefaultValidationHandler(ParserState parserState)
    {
        fDeclPool = parserState.cacheElementDeclPool();
        fErrHandler = parserState.getErrorHandler();
        fAttrPool = parserState.cacheAttrPool();
        fStringPool = parserState.cacheStringPool();
        fParserState = parserState;

        //
        //  <TBD>
        //  We require that all of these be present, so if any of them are
        //  null, throw an exception now so we don't have to worry about this
        //  later during regular processing.
        //
        if ((fDeclPool == null)
        ||  (fErrHandler == null)
        ||  (fAttrPool == null)
        ||  (fStringPool == null)
        ||  (fParserState == null))
        {
        }
    }


    // -----------------------------------------------------------------------
    //  Public methods
    // -----------------------------------------------------------------------

    /**
     * Reset the handler.
     *
     * @param parserState The parser state to use for the next parse call.
     */
    public void reset(ParserState parserState) {
        fDeclPool = parserState.cacheElementDeclPool();
        fErrHandler = parserState.getErrorHandler();
        fAttrPool = parserState.cacheAttrPool();
        fStringPool = parserState.cacheStringPool();
        fParserState = parserState;
    }

    /**
     * Check that the root element name is valid.
     * <p>
     * This method is called when the scanner sees the root element of the target
     * XML document. It will pass the element name index of the root element, which
     * the validator must confirm was the root name indicated in the DTD (or one of
     * the valid roots indicated by a Schema.)
     *
     * @param elementNameIndex The index within the <code>StringPool</code> of the
     *                         root element name.
     *
     * @exception Exception Thrown on error.
     */
    public void checkRootElementName(int elementNameIndex) throws Exception
    {
        //
        //  Get the root element token and check if we even have one.
        //  If not, then the element name cannot be incorrect.
        //
        final int rootElement = fDeclPool.getRootElement();
        if (rootElement == -1)
        {
            return;
        }

        //
        //  From the element decl pool, get the element name token of the root
        //  element. If its not equal to the passed name token, then there is
        //  a root element mismatch.
        //
        final int rootElementName = fDeclPool.getElementName(rootElement);
        if (elementNameIndex != rootElementName)
        {
            fErrHandler.error2
            (
                ErrorCode.V_TAG5
                , rootElementName
                , elementNameIndex
            );
        }
    }

    /**
     * Check that the attributes for an element are valid.
     * <p>
     * This method is called as a convenience. The scanner will do any required
     * checks for well-formedness on the attributes, as well as fill out any
     * defaulted ones. However, if the validator has any other constraints or
     * semantics it must enforce, it can use this API to do so.
     * <p>
     * The scanner provides the element index (within the decl pool, i.e. not the
     * name index which is within the string pool), and the index of the first
     * attribute within the attribute pool that holds the attributes of the element.
     * By this time, all defaulted attributes are present and all fixed values have
     * been confirmed. For most validators, this will be a no-op.
     *
     * @param elementIndex The index within the <code>ElementDeclPool</code> of
     *                     this element.
     * @param firstAttrIndex The index within the <code>AttrPool</code> of the
     *                       first attribute of this element, or -1 if there are
     *                       no attributes.
     *
     * @exception Exception Thrown on error.
     */
    public void checkAttributes(int     elementIndex
                                , int   firstAttrIndex) throws Exception
    {
        if (DEBUG_PRINT_ATTRIBUTES) {
            int elementNameIndex = fDeclPool.getElementName(elementIndex);
            String elementName = fStringPool.toString(elementNameIndex);
            System.out.print("DVH.checkAttributes: <" + elementName);
            if (firstAttrIndex != -1) {
                int index = firstAttrIndex;
                do {
                    System.out.print(" " + fStringPool.toString(fAttrPool.getAttrName(index)) + "=\"" +
                            fStringPool.toString(fAttrPool.getAttValue(index)) + "\"");
                } while (!fAttrPool.isLastAttr(index++));
            }
            System.out.println(">");
        }
        if (fParserState.getCheckNamespace())
            fDeclPool.checkNamespace(elementIndex, firstAttrIndex);
    }

    /**
     * Check that the content of an element is valid.
     * <p>
     * This is the method of primary concern to the validator. This method is called
     * upon the scanner reaching the end tag of an element. At that time, the
     * element's children must be structurally validated, so it calls this method.
     * The index of the element being checked (in the decl pool), is provided as
     * well as an array of element name indexes of the children. The validator must
     * confirm that this element can have these children in this order.
     * <p>
     * This can also be called to do 'what if' testing of content models just to see
     * if they would be valid.
     * <p>
     * Note that the element index is an index into the element decl pool, whereas
     * the children indexes are name indexes, i.e. into the string pool.
     * <p>
     * A value of -1 in the children array indicates a PCDATA node. All other 
     * indexes will be positive and represent child elements. The count can be
     * zero, since some elements have the EMPTY content model and that must be 
     * confirmed.
     *
     * @param elementIndex The index within the <code>ElementDeclPool</code> of this
     *                     element.
     * @param childCount The number of entries in the <code>children</code> array.
     * @param children The children of this element.  Each integer is an index within
     *                 the <code>StringPool</code> of the child element name.  An index
     *                 of -1 is used to indicate an occurrence of non-whitespace character
     *                 data.
     *
     * @return The value -1 if fully valid, else the 0 based index of the child
     *         that first failed. If the value returned is equal to the number
     *         of children, then additional content is required to reach a valid
     *         ending state.
     *
     * @exception Exception Thrown on error.
     */
    public int checkContent(int     elementIndex
                            , int   childCount
                            , int[] children) throws Exception
    {
        // Get the element name index from the element
        final int elementNameIndex = fDeclPool.getElementName(elementIndex);

        if (DEBUG_PRINT_CONTENT)
        {
            String strTmp = fStringPool.toString(elementNameIndex);
            System.out.println
            (
                "Name: "
                + strTmp
                + ", Count: "
                + childCount
                + ", ContentSpec: "
                + fDeclPool.getContentSpecAsString(elementIndex)
            );
            for (int index = 0; index < childCount && index < 10; index++) {
                if (index == 0) System.out.print("  (");
                String childName = (children[index] == -1) ? "#PCDATA" : fStringPool.toString(children[index]);
                if (index + 1 == childCount)
                    System.out.println(childName + ")");
                else if (index + 1 == 10)
                    System.out.println(childName + ",...)");
                else
                    System.out.print(childName + ",");
            }
        }

        // Get out the content spec for this element
        final int contentSpec = fDeclPool.getContentSpecType(elementIndex);

        //
        //  Deal with the possible types of content. We try to optimized here
        //  by dealing specially with content models that don't require the
        //  full DFA treatment.
        //
        if (contentSpec == ElementDecl.CONTENTSPEC_EMPTY)
        {
            //
            //  If the child count is greater than zero, then this is
            //  an error right off the bat at index 0.
            //
            if (childCount != 0)
                return 0;
        }
         else if ((contentSpec == ElementDecl.CONTENTSPEC_ANY))
        {
            //
            //  This one is open game so we don't pass any judgement on it
            //  at all. Its assumed to fine since it can hold anything.
            //
        }
         else if ((contentSpec == ElementDecl.CONTENTSPEC_MIXED)
              ||  (contentSpec == ElementDecl.CONTENTSPEC_CHILDREN))
        {
            // Get the content model for this element, faulting it in if needed
            ContentModel cmElem = null;
            try
            {
                cmElem = getContentModel(elementIndex);
                return cmElem.validateContent(childCount, children);
            }
            
            catch(CMException excToCatch)
            {
                // Translate the caught exception to the public error API
                excToCatch.send(fErrHandler);
            }
        }
         else if ((contentSpec == ElementDecl.CONTENTSPEC_UNKNOWN)) 
        {
            fErrHandler.error1(ErrorCode.V_ELEM4, elementNameIndex);
        }    
         else
        {
            fErrHandler.error(ErrorCode.E_VAL_CST);
        }

        // We succeeded
        return -1;
    }

    /**
     * Check that all ID references were to ID attributes present in the document.
     * <p>
     * This method is a convenience call that allows the validator to do any id ref
     * checks above and beyond those done by the scanner. The scanner does the checks
     * specificied in the XML spec, i.e. that ID refs refer to ids which were
     * eventually defined somewhere in the document.
     * <p>
     * If the validator is for a Schema perhaps, which defines id semantics beyond
     * those of the XML specificiation, this is where that extra checking would be
     * done. For most validators, this is a no-op.
     *
     * @exception Exception Thrown on error.
     */
    public void checkIDRefNames() throws Exception
    {
        fDeclPool.checkIdRefs();
    }

    /**
     * Returns information about which elements can be placed at a particular point
     * in the passed element's content model.
     * <p>
     * Note that the incoming content model to test must be valid at least up to
     * the insertion point. If not, then -1 will be returned and the info object
     * will not have been filled in.
     * <p>
     * If, on return, the info.isValidEOC flag is set, then the 'insert after'
     * elemement is a valid end of content, i.e. nothing needs to be inserted
     * after it to make the parent element's content model valid.
     *
     * @param elementIndex The index within the <code>ElementDeclPool</code> of the
     *                     element which is being querying.
     * @param fullyValid Only return elements that can be inserted and still
     *                   maintain the validity of subsequent elements past the
     *                   insertion point (if any).  If the insertion point is at
     *                   the end, and this is true, then only elements that can
     *                   be legal final states will be returned.
     * @param info An object that contains the required input data for the method,
     *             and which will contain the output information if successful.
     *
     * @return The value -1 if fully valid, else the 0 based index of the child
     *         that first failed before the insertion point. If the value 
     *         returned is equal to the number of children, then the specified
     *         children are valid but additional content is required to reach a
     *         valid ending state.
     *
     * @exception Exception Thrown on error.
     *
     * @see InsertableElementsInfo
     */
    public int whatCanGoHere(int                        elementIndex
                            , boolean                   fullyValid
                            , InsertableElementsInfo    info) throws Exception
    {
        //
        //  Do some basic sanity checking on the info packet. First, make sure
        //  that insertAt is not greater than the child count. It can be equal,
        //  which means to get appendable elements, but not greater. Or, if
        //  the current children array is null, that's bad too.
        //
        //  Since the current children array must have a blank spot for where
        //  the insert is going to be, the child count must always be at least
        //  one.
        //
        //  Make sure that the child count is not larger than the current children
        //  array. It can be equal, which means get appendable elements, but not
        //  greater.
        //
        if ((info.insertAt > info.childCount)
        ||  (info.curChildren == null)
        ||  (info.childCount < 1)
        ||  (info.childCount > info.curChildren.length))
        {
            fErrHandler.error(ErrorCode.E_VAL_WCGHI);
        }

        int retVal = 0;
        try
        {
            // Get the content model for this element
            final ContentModel cmElem = getContentModel(elementIndex);

            // And delegate this call to it
            retVal = cmElem.whatCanGoHere(fullyValid, info);
        }

        catch(CMException excToCatch)
        {
            // Translate caught error to the public error handler interface
            excToCatch.send(fErrHandler);
        }
        return retVal;
    }


    // -----------------------------------------------------------------------
    //  Private methods
    // -----------------------------------------------------------------------
 
    //
    //  When the element has a 'CHILDREN' model, this method is called to
    //  create the content model object. It looks for some special case simple
    //  models and creates SimpleContentModel objects for those. For the rest
    //  it creates the standard DFA style model.
    //
    private final ContentModel createChildModel(int elementIndex) throws CMException
    {
        //
        //  Get the content spec node for the element we are working on.
        //  This will tell us what kind of node it is, which tells us what
        //  kind of model we will try to create.
        //
        ContentSpecNode specNode = new ContentSpecNode();
        fDeclPool.getContentSpecNode
        (
            fDeclPool.getContentSpec(elementIndex)
            , specNode
        );

        //
        //  Check that the left value is not -1, since any content model
        //  with PCDATA should be MIXED, so we should not have gotten here.
        //
        if (specNode.value == -1)
            throw new CMException(ErrorCode.E_VAL_NPCD);

        if (specNode.type == ContentSpecNode.NODE_LEAF)
        {
            //
            //  Its a single leaf, so its an 'a' type of content model, i.e.
            //  just one instance of one element. That one is definitely a
            //  simple content model.
            //
            return new SimpleContentModel(specNode.value, -1, specNode.type);
        }
         else if ((specNode.type == ContentSpecNode.NODE_CHOICE)
              ||  (specNode.type == ContentSpecNode.NODE_SEQ))
        {
            //
            //  Lets see if both of the children are leafs. If so, then it
            //  it has to be a simple content model
            //
            ContentSpecNode specLeft = new ContentSpecNode();
            ContentSpecNode specRight = new ContentSpecNode();
            fDeclPool.getContentSpecNode(specNode.value, specLeft);
            fDeclPool.getContentSpecNode(specNode.otherValue, specRight);

            if ((specLeft.type == ContentSpecNode.NODE_LEAF)
            &&  (specRight.type == ContentSpecNode.NODE_LEAF))
            {
                //
                //  Its a simple choice or sequence, so we can do a simple
                //  content model for it.
                //
                return new SimpleContentModel
                (
                    specLeft.value
                    , specRight.value
                    , specNode.type
                );
            }
        }
         else if ((specNode.type == ContentSpecNode.NODE_ZERO_OR_ONE)
              ||  (specNode.type == ContentSpecNode.NODE_ZERO_OR_MORE)
              ||  (specNode.type == ContentSpecNode.NODE_ONE_OR_MORE))
        {
            //
            //  Its a repetition, so see if its one child is a leaf. If so
            //  its a repetition of a single element, so we can do a simple
            //  content model for that.
            //
            ContentSpecNode specLeft = new ContentSpecNode();
            fDeclPool.getContentSpecNode(specNode.value, specLeft);

            if (specLeft.type == ContentSpecNode.NODE_LEAF)
            {
                //
                //  It is, so we can create a simple content model here that
                //  will check for this repetition. We pass -1 for the unused
                //  right node.
                //
                return new SimpleContentModel(specLeft.value, -1, specNode.type);
            }
        }
         else
        {
            throw new CMException(ErrorCode.E_VAL_CST);
        }

        //
        //  Its not a simple content model, so here we have to create a DFA
        //  for this element. So we create a DFAContentModel object. He
        //  encapsulates all of the work to create the DFA.
        //
        return new DFAContentModel
        (
            elementIndex
            , fStringPool
            , fDeclPool
        );
    }
 
 
    //
    //  This method will handle the querying of the content model for a
    //  particular element. If the element does not have a content model, then
    //  it will be created.
    //
    ContentModel getContentModel(int elementIndex) throws CMException
    {
        // See if a content model already exists first
        ContentModel cmRet = fDeclPool.getContentModel(elementIndex);

        // If we have one, just return that. Otherwise, gotta create one
        if (cmRet != null)
            return cmRet;

        // Get the type of content this element has
        final int contentSpec = fDeclPool.getContentSpecType(elementIndex);

        // And create the content model according to the spec type
        if (contentSpec == ElementDecl.CONTENTSPEC_MIXED)
        {
            //
            //  Just create a mixel content model object. This type of
            //  content model is optimized for mixed content validation.
            //
            cmRet = new MixedContentModel(fDeclPool, elementIndex);
        }
         else if (contentSpec == ElementDecl.CONTENTSPEC_CHILDREN)
        {
            //
            //  This method will create an optimal model for the complexity
            //  of the element's defined model. If its simple, it will create
            //  a SimpleContentModel object. If its a simple list, it will
            //  create a SimpleListContentModel object. If its complex, it
            //  will create a DFAContentModel object.
            //
            cmRet = createChildModel(elementIndex);
        }
         else
        {
            throw new CMException(ErrorCode.E_VAL_CST);
        }

        // Add the new model to the content model for this element
        fDeclPool.setContentModel(elementIndex, cmRet);

        return cmRet;
    }
 
 
    // -----------------------------------------------------------------------
    //  Private data members
    //
    //  fDeclPool
    //      The element/attribute declaration pool that we need to get info
    //      about elements and attributes. This is set during ctor, from our
    //      parent parser object, and left alone.
    //
    //  fErrHandler
    //      The error handler we are to use for reporting errors. This is set
    //      during ctor, from our praent parser object, and left alone.
    //
    //  fAttrPool
    //      The attr pool that we use to convert attribute indexes to info
    //      about attributes. This is set during ctor, from our parent parser
    //      object, and left alone.
    //
    //  fStringPool
    //      The string pool that we use to convert string indexes to strings
    //      as required. This is set during ctor, from our parent parser
    //      object, and left alone.
    // -----------------------------------------------------------------------
    private ElementDeclPool fDeclPool       = null;
    private XMLErrorHandler fErrHandler     = null;
    private AttrPool        fAttrPool       = null;
    private StringPool      fStringPool     = null;
    private ParserState     fParserState    = null;


    // Junk flag to control debug printouts
    private static final boolean DEBUG_PRINT_ATTRIBUTES = false;
    private static final boolean DEBUG_PRINT_CONTENT = false;
}
