/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.internal;

import com.ibm.xml.framework.*;

/**
 * MixedContentModel is a derivative of the abstract content model base
 * class that handles the special case of mixed model elements. If an element
 * is mixed model, it has PCDATA as its first possible content, followed
 * by an alternation of the possible children. The children cannot have any
 * numeration or order, so it must look like this:
 * <pre>
 *   &lt;!ELEMENT Foo ((#PCDATA|a|b|c|)*)&gt;
 * </pre>
 * So, all we have to do is to keep an array of the possible children and
 * validate by just looking up each child being validated by looking it up
 * in the list.
 *
 * @version Revision: 77 1.8 src/com/ibm/xml/internal/MixedContentModel.java, parser, xml4j2, xml4j2_0_13 
 */
class MixedContentModel implements ContentModel
{
    // -----------------------------------------------------------------------
    //  Constructors
    // -----------------------------------------------------------------------

    /**
     * Constructs a mixed content model.
     *
     * @param declPool The element decl pool.
     * @param elementIndex The element index into the pool.
     *
     * @exception CMException Thrown if content model can't be built.
     */
    public MixedContentModel(ElementDeclPool declPool, int elementIndex) throws CMException
    {
        // Init our count member, but save the list member for later
        fCount = 0;

        //
        //  A temp list to build the child list into. It is initially set
        //  to 64 elements, which should be enough to meet 99% of all needs.
        //  It will be expanded in the very rare event that it is required.
        //
        int   listSize = 64;
        int[] tmpList = new int[listSize];

        //
        //  Ok, we need to build up an array of the possible children
        //  under this element. The mixed content model can only be a
        //  repeated series of alternations with no numeration or ordering.
        //  So we call a local recursive method to iterate the tree and
        //  build up the array.
        //
        //  So we get the content spec of the element, which gives us the
        //  starting node. Everything else kicks off from there. We pass
        //  along a content node for each iteration to use so that it does
        //  not have to create and trash lots of objects.
        //
        ContentSpecNode nodeTmp = new ContentSpecNode();

        while (true)
        {
            try
            {
                fCount = buildContentList
                (
                    declPool.getContentSpec(elementIndex)
                    , tmpList
                    , 0
                    , nodeTmp
                    , declPool
                );
            }

            catch(IndexOutOfBoundsException excToCatch)
            {
                //
                //  Expand the array and try it again. Yes, this is
                //  piggy, but the odds of it ever actually happening
                //  are slim to none.
                //
                listSize *= 2;
                tmpList = new int[listSize];
                fCount = 0;
                continue;
            }

            // We survived, so break out
            break;
        }

        // Make our own copy now, which is exactly the right size
        fChildren = new int[fCount];
        for (int index = 0; index < fCount; index++)
            fChildren[index] = tmpList[index];
    }


    // -----------------------------------------------------------------------
    //  Public, inherited methods
    // -----------------------------------------------------------------------
    
    /**
     * Check that the specified content is valid according to this
     * content model. This method can also be called to do 'what if' 
     * testing of content models just to see if they would be valid.
     * <p>
     * A value of -1 in the children array indicates a PCDATA node. All other 
     * indexes will be positive and represent child elements. The count can be
     * zero, since some elements have the EMPTY content model and that must be 
     * confirmed.
     *
     * @param childCount The number of entries in the <code>children</code> array.
     * @param children The children of this element.  Each integer is an index within
     *                 the <code>StringPool</code> of the child element name.  An index
     *                 of -1 is used to indicate an occurrence of non-whitespace character
     *                 data.
     *
     * @return The value -1 if fully valid, else the 0 based index of the child
     *         that first failed. If the value returned is equal to the number
     *         of children, then the specified children are valid but additional
     *         content is required to reach a valid ending state.
     *
     * @exception Exception Thrown on error.
     */
    public int validateContent(int childCount, int[] children) throws Exception
    {
        for (int outIndex = 0; outIndex < childCount; outIndex++)
        {
            // Get the current child out of the source index
            final int curChild = children[outIndex];

            // If its PCDATA, then we just accept that
            if (curChild == -1)
                continue;

            // And try to find it in our list
            int inIndex = 0;
            for (; inIndex < fCount; inIndex++)
            {
                if (curChild == fChildren[inIndex])
                    break;
            }

            // We did not find this one, so the validation failed
            if (inIndex == fCount)
                return outIndex;
        }

        // Everything seems to be in order, so return success
        return -1;
    }

    /**
     * Returns information about which elements can be placed at a particular point
     * in the passed element's content model.
     * <p>
     * Note that the incoming content model to test must be valid at least up to
     * the insertion point. If not, then -1 will be returned and the info object
     * will not have been filled in.
     * <p>
     * If, on return, the info.isValidEOC flag is set, then the 'insert after'
     * element is a valid end of content. In other words, nothing needs to be
     * inserted after it to make the parent element's content model valid.
     *
     * @param fullyValid Only return elements that can be inserted and still
     *                   maintain the validity of subsequent elements past the
     *                   insertion point (if any).  If the insertion point is at
     *                   the end, and this is true, then only elements that can
     *                   be legal final states will be returned.
     * @param info An object that contains the required input data for the method,
     *             and which will contain the output information if successful.
     *
     * @return The value -1 if fully valid, else the 0 based index of the child
     *         that first failed before the insertion point. If the value 
     *         returned is equal to the number of children, then the specified
     *         children are valid but additional content is required to reach a
     *         valid ending state.
     *
     * @see InsertableElementsInfo
     */
    public int whatCanGoHere(boolean                    fullyValid
                            , InsertableElementsInfo    info) throws Exception
    {
        //
        //  For this one, having the empty slot at the insertion point is 
        //  a problem. So lets compress the array down. We know that it has
        //  to have at least the empty slot at the insertion point.
        //
        for (int index = info.insertAt; index < info.childCount; index++)
            info.curChildren[index] = info.curChildren[index+1];
        info.childCount--;

        //
        //  Check the validity of the existing contents. If this is less than
        //  the insert at point, then return failure index right now
        //
        final int failedIndex = validateContent(info.childCount, info.curChildren);
        if ((failedIndex != -1) && (failedIndex < info.insertAt))
            return failedIndex;

        //
        //  Set any stuff we can know right off the bat for all cases. Mixed
        //  models can always hold PCData. And, since its always a repetition
        //  of a bunch of choice nodes, its always valid EOC.
        //
        info.canHoldPCData = true;
        info.isValidEOC = true;

        //
        //  Set the results count member and then see if we need to reallocate
        //  the outgoing arrays.
        //
        info.resultsCount = fCount;

        if ((info.results == null) || (info.results.length < info.resultsCount))
            info.results = new boolean[info.resultsCount];

        if ((info.possibleChildren == null)
        ||  (info.possibleChildren.length < info.resultsCount))
        {
            info.possibleChildren = new int[info.resultsCount];
        }

        //
        //  If the fully valid parameter is set, then whether any child can
        //  go here is dependent upon the content model having been valid all
        //  the way to the end. If its not, nothing we put here is going to
        //  make it happy. If it was ok, then nothing we put here is ever going
        //  make it bad.
        //
        //  So set up a boolean that can be used to set every possible child's
        //  insertable status below.
        //
        boolean bStatus = true;
        if (fullyValid && (failedIndex < info.childCount))
            bStatus = false;

        //
        //  Fill in the possible children array, from our array. And set the
        //  boolean flag for each one to true because any of them can go
        //  anywhere.
        //
        for (int index = 0; index < fCount; index++)
        {
            info.possibleChildren[index] = fChildren[index];
            info.results[index] = bStatus;
        }

        return -1;
    }


    // -----------------------------------------------------------------------
    //  Private methods
    // -----------------------------------------------------------------------
    private final int buildContentList( int                 startNode
                                        , int[]             list
                                        , int               count
                                        , ContentSpecNode   specNode
                                        , ElementDeclPool   declPool) throws CMException
    {
        // Get the content spec for the passed start node
        declPool.getContentSpecNode(startNode, specNode);

        // If this node is a leaf, then add it to our list and return.
        if (specNode.type == ContentSpecNode.NODE_LEAF)
        {
            list[count++] = specNode.value;
            return count;
        }

        //
        //  Its not a leaf, so we have to recurse its left and maybe right
        //  nodes. Save both values before we recurse and trash the node.
        //
        final int leftNode = specNode.value;
        final int rightNode = specNode.otherValue;

        if ((specNode.type == ContentSpecNode.NODE_CHOICE)
        ||  (specNode.type == ContentSpecNode.NODE_SEQ))
        {
            //
            //  Recurse on the left and right nodes of this guy, making sure
            //  to keep tne count correct.
            //
            count = buildContentList(leftNode, list, count, specNode, declPool);
            count = buildContentList(rightNode, list, count, specNode, declPool);
        }
         else if ((specNode.type == ContentSpecNode.NODE_ZERO_OR_ONE)
              ||  (specNode.type == ContentSpecNode.NODE_ZERO_OR_MORE)
              ||  (specNode.type == ContentSpecNode.NODE_ONE_OR_MORE))
        {
            // Just do the left node on this one
            count = buildContentList(leftNode, list, count, specNode, declPool);
        }
         else
        {
            throw new CMException(ErrorCode.E_VAL_CST);
        }

        // And return our accumlated new count
        return count;
    }


    // -----------------------------------------------------------------------
    //  Private data members
    //
    //  fCount
    //      The count of possible children that we have to deal with.
    //
    //  fChildren
    //      The list of possible children that we have to accept.
    // -----------------------------------------------------------------------
    private int             fCount;
    private int[]           fChildren;
};
