/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */
package com.ibm.xml.internal;

import com.ibm.xml.framework.*;
import org.xml.sax.Locator;
import java.io.IOException;

/**
 *
 * @version Revision: 76 1.5 src/com/ibm/xml/internal/StringReader.java, parser, xml4j2, xml4j2_0_13 
 */
final class StringReader extends XMLReader {
    //
    //
    //
    private StringPool fStringPool = null;
    private XMLDocumentHandler fDocumentHandler = null;
    private XMLErrorHandler fErrorHandler = null;
    private String fData = null;
    private int fEndOffset;
    private boolean oweLeadingSpace = false;
    private boolean oweTrailingSpace = false;
    //
    //
    //
    StringReader(ParserState parserState, boolean addEnclosingSpaces, Locator locator, int stringIndex)
    {
        super(parserState, locator.getPublicId(), locator.getSystemId());
        fLinefeedCounter = locator.getLineNumber();
        fCharacterCounter = locator.getColumnNumber();
        fStringPool = parserState.cacheStringPool();
        fDocumentHandler = parserState.getDocumentHandler();
        fErrorHandler = parserState.getErrorHandler();
        fData = fStringPool.toString(stringIndex);
        fCurrentOffset = 0;
        fEndOffset = fData.length();
        if (addEnclosingSpaces) {
            oweLeadingSpace = true;
            oweTrailingSpace = true;
        }
    }
    //
    //
    //
    public void append(ChunkyCharArray charArray, int offset, int length) {
        charArray.append(fData.substring(offset, offset + length));
    }
    //
    //
    //
    public int addString(int offset, int length) {
        if (length == 0)
            return 0;
        return fStringPool.addString(fData.substring(offset, offset + length));
    }
    //
    //
    //
    public int addSymbol(int offset, int length) {
        if (length == 0)
            return 0;
        return fStringPool.addSymbol(fData.substring(offset, offset + length));
    }
    //
    //
    //
    public int skipOneChar() throws IOException {
        if (oweLeadingSpace) {
            oweLeadingSpace = false;
            return fCurrentOffset;
        }
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace) {
                oweTrailingSpace = false;
                return fCurrentOffset;
            }
            throw new ArrayIndexOutOfBoundsException();
        }
        return ++fCurrentOffset;
    }
    public int skipAsciiChar() throws IOException {
        if (oweLeadingSpace) {
            oweLeadingSpace = false;
            return fCurrentOffset;
        }
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace) {
                oweTrailingSpace = false;
                return fCurrentOffset;
            }
            throw new ArrayIndexOutOfBoundsException();
        }
        return ++fCurrentOffset;
    }
    //
    //
    //
    public int skipToChar(char ch) throws IOException {
        if (oweLeadingSpace) {
            if (ch == ' ')
                return fCurrentOffset;
            oweLeadingSpace = false;
        }
        while (true) {
            if (fCurrentOffset >= fEndOffset) {
                if (oweTrailingSpace) {
                    if (ch == ' ')
                        return fCurrentOffset;
                    oweTrailingSpace = false;
                }
                throw new ArrayIndexOutOfBoundsException();
            }
            if (ch == fData.charAt(fCurrentOffset))
                return fCurrentOffset;
            fCurrentOffset++;
        }
    }
    public int skipPastChar(char ch) throws IOException {
        if (oweLeadingSpace) {
            oweLeadingSpace = false;
            if (ch == ' ')
                return fCurrentOffset;
        }
        while (true) {
            if (fCurrentOffset >= fEndOffset) {
                if (oweTrailingSpace) {
                    oweTrailingSpace = false;
                    if (ch == ' ')
                        return fCurrentOffset;
                }
                throw new ArrayIndexOutOfBoundsException();
            }
            if (ch == fData.charAt(fCurrentOffset))
                return ++fCurrentOffset;
            fCurrentOffset++;
        }
    }
    //
    //
    //
    public boolean skippedValidChar() throws IOException {
        if (oweLeadingSpace) {
            oweLeadingSpace = false;
            return true;
        }
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace) {
                oweTrailingSpace = false;
                return true;
            }
            throw new ArrayIndexOutOfBoundsException();
        }
        char ch = fData.charAt(fCurrentOffset++);
        if (ch < 0x20) {
            if (ch == 0x09 || ch == 0x0A || ch == 0x0D)
                return true;
            fCurrentOffset--;
            return false;
        }
        if (ch <= 0xD7FF || (ch >= 0xE000 && (ch <= 0xFFFD || (ch >= 0x10000 && ch <= 0x10FFFF))))
            return true;
        fCurrentOffset--;
        return false;
    }
    //
    //
    //
    public boolean lookingAtValidChar() throws IOException {
        if (oweLeadingSpace) {
            oweLeadingSpace = false;
            return true;
        }
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace) {
                oweTrailingSpace = false;
                return true;
            }
            throw new ArrayIndexOutOfBoundsException();
        }
        char ch = fData.charAt(fCurrentOffset);
        if (ch < 0x20)
            return (ch == 0x09 || ch == 0x0A || ch == 0x0D);
        return (ch <= 0xD7FF || (ch >= 0xE000 && (ch <= 0xFFFD || (ch >= 0x10000 && ch <= 0x10FFFF))));
    }
    //
    //
    //
    public int skipInvalidChar(int errorCode) throws Exception {
        int ch;
        if (oweLeadingSpace) {
            oweLeadingSpace = false;
            ch = ' ';
        } else if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace) {
                oweTrailingSpace = false;
                ch = ' ';
            } else
                throw new ArrayIndexOutOfBoundsException();
        } else
            ch = fData.charAt(fCurrentOffset++);
        switch (errorCode) {
        case ErrorCode.E_INVCHAR0:
        case ErrorCode.E_ENTITY3:
        {
            String arg = Integer.toHexString(ch);
            fErrorHandler.error1(errorCode, fStringPool.addString(arg));
            break;
        }
        case ErrorCode.E_REFER0:
        case ErrorCode.E_EXT2:
        case ErrorCode.E_EXT4:
        {
            String arg = new Character((char)ch).toString();
            fErrorHandler.error1(errorCode, fStringPool.addString(arg));
            break;
        }
        case ErrorCode.E_DTD0:
        {
            String arg1 = new Character((char)ch).toString();
            String arg2 = Integer.toHexString(ch);
            fErrorHandler.error2(errorCode, fStringPool.addString(arg1), fStringPool.addString(arg2));
            break;
        }
        default:
            break;
        }
        return fCurrentOffset;
    }
    //
    //
    //
    public boolean skippedChar(char ch) throws IOException {
        if (oweLeadingSpace) {
            if (ch == ' ') {
                oweLeadingSpace = false;
                return true;
            }
            return false;
        }
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace) {
                if (ch == ' ') {
                    oweTrailingSpace = false;
                    return true;
                }
                return false;
            }
            throw new ArrayIndexOutOfBoundsException();
        }
        if (ch == fData.charAt(fCurrentOffset)) {
            fCurrentOffset++;
            return true;
        } else
            return false;
    }
    public boolean lookingAtChar(char ch) throws IOException {
        if (oweLeadingSpace)
            return (ch == ' ');
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace)
                return (ch == ' ');
            throw new ArrayIndexOutOfBoundsException();
        }
        return ch == fData.charAt(fCurrentOffset);
    }
    public boolean skippedSpace() throws IOException {
        if (oweLeadingSpace) {
            oweLeadingSpace = false;
            return true;
        }
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace) {
                oweTrailingSpace = false;
                return true;
            }
            throw new ArrayIndexOutOfBoundsException();
        }
        int ch = fData.charAt(fCurrentOffset);
        if (ch == 0x20 || ch == 0x09 || ch == 0x0A || ch == 0x0D) {
            fCurrentOffset++;
            return true;
        } else
            return false;
    }
    public boolean lookingAtSpace() throws IOException {
        if (oweLeadingSpace)
            return true;
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace)
                return true;
            throw new ArrayIndexOutOfBoundsException();
        }
        int ch = fData.charAt(fCurrentOffset);
        return (ch == 0x20 || ch == 0x09 || ch == 0x0A || ch == 0x0D);
    }
    public int skipPastSpaces() throws IOException {
        if (oweLeadingSpace)
            oweLeadingSpace = false;
        while (true) {
            if (fCurrentOffset >= fEndOffset) {
                if (oweTrailingSpace) {
                    oweTrailingSpace = false;
                    return fCurrentOffset;
                }
                throw new ArrayIndexOutOfBoundsException();
            }
            int ch = fData.charAt(fCurrentOffset);
            if (!(ch == 0x20 || ch == 0x09 || ch == 0x0A || ch == 0x0D))
                return fCurrentOffset;
            fCurrentOffset++;
        }
    }
    //
    //
    //
    public int skipDecimalDigit() {
        if (oweLeadingSpace)
            return -1;
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace)
                return -1;
            throw new ArrayIndexOutOfBoundsException();
        }
        int ch = fData.charAt(fCurrentOffset);
        if (ch < '0' || ch > '9')
            return -1;
        fCurrentOffset++;
        return ch - '0';
    }
    public int skipHexDigit() {
        if (oweLeadingSpace)
            return -1;
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace)
                return -1;
            throw new ArrayIndexOutOfBoundsException();
        }
        int ch = fData.charAt(fCurrentOffset);
        if (ch > 'f' || fgAsciiXDigitChar[ch] == 0)
            return -1;
        fCurrentOffset++;
        return ch - (ch < 'A' ? '0' : (ch < 'a' ? 'A' : 'a') - 10);
    }
    public boolean skippedAlpha() {
        if (oweLeadingSpace)
            return false;
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace)
                return false;
            throw new ArrayIndexOutOfBoundsException();
        }
        int ch = fData.charAt(fCurrentOffset);
        if (ch <= 'z' && fgAsciiAlphaChar[ch] == 1) {
            fCurrentOffset++;
            return true;
        }
        return false;
    }
    private boolean skippedCharWithFlag(byte flag) {
        if (oweLeadingSpace) {
            if ((fgCharFlags[' '] & flag) != 0) {
                oweLeadingSpace = false;
                return true;
            }
            return false;
        }
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace) {
                if ((fgCharFlags[' '] & flag) != 0) {
                    oweTrailingSpace = false;
                    return true;
                }
                return false;
            }
            throw new ArrayIndexOutOfBoundsException();
        }
        int ch = fData.charAt(fCurrentOffset);
        if (ch < 0x80 && (fgCharFlags[ch] & flag) != 0) {
            fCurrentOffset++;
            return true;
        }
        return false;
    }
    public final boolean skippedVersionNum() {
        return skippedCharWithFlag(E_VersionNumFlag);
    }
    public final boolean skippedEncName() {
        return skippedCharWithFlag(E_EncNameFlag);
    }
    public final boolean skippedPubidChar() {
        if (oweLeadingSpace) {
            oweLeadingSpace = false;
            return true;
        }
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace) {
                oweTrailingSpace = false;
                return true;
            }
            throw new ArrayIndexOutOfBoundsException();
        }
        int ch = fData.charAt(fCurrentOffset);
        if (ch < 0x80) {
            if ((fgCharFlags[ch] & E_PubidCharFlag) != 0) {
                fCurrentOffset++;
                return true;
            }
            if (ch == 0x0A) {
                fCurrentOffset++;
                fLinefeedCounter++;
                return true; // #xA
            }
            if (ch == 0x0D) {
                fCurrentOffset++;
                fCarriageReturnCounter++;
                return true; // #xD
            }
        }
        return false;
    }
    public boolean skippedString(char[] s) throws IOException {
        if (oweLeadingSpace)
            return false;
        int offset = fCurrentOffset;
        for (int i = 0; i < s.length; i++) {
            if (fCurrentOffset >= fEndOffset) {
                if (oweTrailingSpace)
                    return false;
                throw new ArrayIndexOutOfBoundsException();
            }
            if (fData.charAt(offset) != s[i])
                return false;
            offset++;
        }
        fCurrentOffset = offset;
        return true;
    }
    public int skipPastName(char fastcheck) throws IOException {
        if (oweLeadingSpace)
            return fCurrentOffset;
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace)
                return fCurrentOffset;
            throw new ArrayIndexOutOfBoundsException();
        }
        char ch = fData.charAt(fCurrentOffset);
        if ((fgCharFlags[ch] & E_InitialNameCharFlag) == 0)
            return fCurrentOffset;
        while (true) {
            fCurrentOffset++;
            if (fCurrentOffset >= fEndOffset) {
                if (oweTrailingSpace)
                    return fCurrentOffset;
                throw new ArrayIndexOutOfBoundsException();
            }
            ch = fData.charAt(fCurrentOffset);
            if (fastcheck == ch)
                return fCurrentOffset;
            if ((fgCharFlags[ch] & E_NameCharFlag) == 0)
                return fCurrentOffset;
        }
    }
    public int scanName(char fastcheck, int expectedName) throws IOException {
        int nameOffset = fCurrentOffset;
        int nameLength = skipPastName(fastcheck) - nameOffset;
        if (nameLength == 0)
            return -1;
        int nameIndex = fStringPool.addSymbol(fData.substring(nameOffset, nameOffset + nameLength));
        return (expectedName == -1 || expectedName == nameIndex) ? nameIndex : -1;
    }
    public int skipPastNmtoken(char fastcheck) throws IOException {
        if (oweLeadingSpace)
            return fCurrentOffset;
        if (fCurrentOffset >= fEndOffset) {
            if (oweTrailingSpace)
                return fCurrentOffset;
            throw new ArrayIndexOutOfBoundsException();
        }
        int ch = fData.charAt(fCurrentOffset);
        while (true) {
            if (fastcheck == ch)
                return fCurrentOffset;
            if ((fgCharFlags[ch] & E_NameCharFlag) == 0)
                return fCurrentOffset;
            fCurrentOffset++;
            if (fCurrentOffset >= fEndOffset) {
                if (oweTrailingSpace)
                    return fCurrentOffset;
                throw new ArrayIndexOutOfBoundsException();
            }
            ch = fData.charAt(fCurrentOffset);
        }
    }
    private void callCharDataHandler(int offset, int endOffset, boolean inCDSect) throws Exception {
        int length = endOffset - offset;
        if (!fDocumentHandler.sendCharDataAsCharArray()) {
            int stringIndex = addString(offset, length);
            fDocumentHandler.characters(stringIndex, inCDSect);
            return;
        }
        fDocumentHandler.characters(fData.toCharArray(), offset, length, inCDSect);
    }
    private void callWSCharDataHandler(int offset, int endOffset, boolean inCDSect) throws Exception {
        int contentSpecType = fParserState.getScanner().getCurrentContentSpecType();
        if (contentSpecType != ElementDecl.CONTENTSPEC_CHILDREN) {
            callCharDataHandler(offset, endOffset, inCDSect);
            return;
        }
        int length = endOffset - offset;
        if (!fDocumentHandler.sendCharDataAsCharArray()) {
            int stringIndex = addString(offset, length);
            fDocumentHandler.ignorableWhitespace(stringIndex, inCDSect);
            return;
        }
        fDocumentHandler.ignorableWhitespace(fData.toCharArray(), offset, length, inCDSect);
    }
    /*
     * There are no leading/trailing space checks here because scanContent cannot
     * be called on a parameter entity reference value.
     */
    public int scanContent(ScanContentState scanState) throws Exception {
        int offset = fCurrentOffset;
        if (fCurrentOffset >= fEndOffset)
            throw new ArrayIndexOutOfBoundsException();
        char ch = fData.charAt(fCurrentOffset++);
        byte prop;
        if (ch < 0x80) {
            prop = fgCharFlags[ch];
            if ((prop & E_CharDataFlag) == 0 && ch != 0x0A && ch != 0x0D) {
                if (ch == '<') {
                    if (!scanState.inCDSect)
                        return CHARDATA_RESULT_MARKUP;
                } else if (ch == '&') {
                    if (!scanState.inCDSect)
                        return CHARDATA_RESULT_REFERENCE;
                } else if (ch == ']') {
                    if (fCurrentOffset + 2 <= fEndOffset && fData.charAt(fCurrentOffset) == ']' && fData.charAt(fCurrentOffset + 1) == '>') {
                        fCurrentOffset += 2;
                        if (!scanState.inCDSect)
                            return CHARDATA_RESULT_CDEND;
                        scanState.inCDSect = false;
                        return scanContent(scanState);
                    }
                } else {
                    fCurrentOffset--;
                    return CHARDATA_RESULT_INVALID_CHAR;
                }
            } else if (ch == 0x20 || ch == 0x09 || ch == 0x0A || ch == 0x0D) {
                do {
                    if (fCurrentOffset >= fEndOffset) {
                        if (fDocumentHandler != null)
                            callWSCharDataHandler(offset, fEndOffset, scanState.inCDSect);
                        throw new ArrayIndexOutOfBoundsException();
                    }
                    if (ch == 0x0A)
                        fLinefeedCounter++;
                    else if (ch == 0x0D)
                        fCarriageReturnCounter++;
                    ch = fData.charAt(fCurrentOffset++);
                } while (ch == 0x20 || ch == 0x09 || ch == 0x0A || ch == 0x0D);
                if (ch < 0x80) {
                    prop = fgCharFlags[ch];
                    if ((prop & E_CharDataFlag) == 0) {
                        if (ch == '<') {
                            if (!scanState.inCDSect) {
                                if (fDocumentHandler != null)
                                    callWSCharDataHandler(offset, fCurrentOffset - 1, scanState.inCDSect);
                                return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_MARKUP);
                            }
                        } else if (ch == '&') {
                            if (!scanState.inCDSect) {
                                if (fDocumentHandler != null)
                                    callWSCharDataHandler(offset, fCurrentOffset - 1, scanState.inCDSect);
                                return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_REFERENCE);
                            }
                        } else if (ch == ']') {
                            if (fCurrentOffset + 2 <= fEndOffset && fData.charAt(fCurrentOffset) == ']' && fData.charAt(fCurrentOffset + 1) == '>') {
                                if (fDocumentHandler != null)
                                    callWSCharDataHandler(offset, fCurrentOffset - 1, scanState.inCDSect);
                                fCurrentOffset += 2;
                                if (!scanState.inCDSect)
                                    return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_CDEND);
                                scanState.inCDSect = false;
                                return scanContent(scanState);
                            }
                        } else {
                            fCurrentOffset--;
                            if (fDocumentHandler != null)
                                callWSCharDataHandler(offset, fCurrentOffset, scanState.inCDSect);
                            return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                        }
                    }
                } else {
                    if (ch == 0xFFFE || ch == 0xFFFF) {
                        fCurrentOffset--;
                        if (fDocumentHandler != null)
                            callWSCharDataHandler(offset, fCurrentOffset, scanState.inCDSect);
                        return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                    }
                }
            }
        } else {
            if (ch == 0xFFFE || ch == 0xFFFF) {
                fCurrentOffset--;
                if (fDocumentHandler != null)
                    callCharDataHandler(offset, fCurrentOffset, scanState.inCDSect);
                return CHARDATA_RESULT_INVALID_CHAR;
            }
        }
        while (true) {
            if (fCurrentOffset >= fEndOffset) {
                if (fDocumentHandler != null)
                    callCharDataHandler(offset, fEndOffset, scanState.inCDSect);
                throw new ArrayIndexOutOfBoundsException();
            }
            ch = fData.charAt(fCurrentOffset++);
            if (ch >= 0x80)
                break;
            prop = fgCharFlags[ch];
            if ((prop & E_CharDataFlag) == 0) {
                if (ch == 0x0A)
                    fLinefeedCounter++;
                else if (ch == 0x0D)
                    fCarriageReturnCounter++;
                else
                    break;
            }
        }
        while (true) { // REVISIT - EOF check ?
            if (ch < 0x80) {
                prop = fgCharFlags[ch];
                if ((prop & E_CharDataFlag) == 0) {
                    if (ch == '<') {
                        if (!scanState.inCDSect) {
                            if (fDocumentHandler != null)
                                callCharDataHandler(offset, fCurrentOffset - 1, scanState.inCDSect);
                            return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_MARKUP);
                        }
                    } else if (ch == '&') {
                        if (!scanState.inCDSect) {
                            if (fDocumentHandler != null)
                                callCharDataHandler(offset, fCurrentOffset - 1, scanState.inCDSect);
                            return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_REFERENCE);
                        }
                    } else if (ch == 0x0A)
                        fLinefeedCounter++;
                    else if (ch == 0x0D)
                        fCarriageReturnCounter++;
                    else if (ch == ']') {
                        if (fCurrentOffset + 2 <= fEndOffset && fData.charAt(fCurrentOffset) == ']' && fData.charAt(fCurrentOffset + 1) == '>') {
                            if (fDocumentHandler != null)
                                callCharDataHandler(offset, fCurrentOffset - 1, scanState.inCDSect);
                            fCurrentOffset += 2;
                            if (!scanState.inCDSect)
                                return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_CDEND);
                            scanState.inCDSect = false;
                            return scanContent(scanState);
                        }
                    } else {
                        fCurrentOffset--;
                        if (fDocumentHandler != null)
                            callCharDataHandler(offset, fCurrentOffset, scanState.inCDSect);
                        return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                    }
                }
            } else {
                if (ch == 0xFFFE || ch == 0xFFFF) {
                    fCurrentOffset--;
                    if (fDocumentHandler != null)
                        callCharDataHandler(offset, fCurrentOffset, scanState.inCDSect);
                    return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                }
            }
            if (fCurrentOffset >= fEndOffset) {
                if (fDocumentHandler != null)
                    callCharDataHandler(offset, fCurrentOffset, scanState.inCDSect);
                throw new ArrayIndexOutOfBoundsException();
            }
            ch = fData.charAt(fCurrentOffset++);
        }
    }
}
