/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.internal;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.URL;
import java.util.StringTokenizer;

import com.ibm.xml.framework.ParserState;

import org.xml.sax.InputSource;

/**
 * Supports the original XML4J catalog file format:
 * <pre>
 *   public_id TAB system_id
 * </pre>
 * <p>
 * To use this catalog in a parser, set a TXCatalog instance as the
 * parser's <tt>EntityResolver</tt>. For example:
 * <pre>
 *   XMLParser parser  = new AnyParser();
 *   Catalog   catalog = <font color="blue">new TXCatalog(parser.getParserState())</font>;
 *   <font color="blue">parser.getEntityHandler().setEntityResolver(catalog);</font>
 * </pre>
 * <p>
 * Once installed, catalog files that conform to the TXCatalog format
 * can be appended to the catalog by calling the <tt>loadCatalog</tt>
 * method on the parser or the catalog instance. The following example
 * loads the contents of two catalog files:
 * <pre>
 *   parser.loadCatalog(new InputSource("catalogs/cat1.xml"));
 *   parser.loadCatalog(new InputSource("http://host/catalogs/cat2.xml"));
 * </pre>
 *
 * @author  Andy Clark, IBM
 * @version Revision: 87 1.2 src/com/ibm/xml/internal/TXCatalog.java, parser, xml4j2, xml4j2_0_13 
 */
public class TXCatalog
    extends DefaultCatalog 
    {

    //
    // Data
    //

    /** Parser state. */
    private ParserState state;

    //
    // Constructors
    //

    /** Constructs a TXCatalog from the specified parser state. */
    public TXCatalog(ParserState state) {
        this.state = state;
        }

    //
    // Catalog methods
    //

    /**
     * Loads the catalog stream specified by the given input source and
     * appends the contents to the catalog.
     *
     * @param source The catalog source.
     *
     * @exception java.lang.Exception Throws an exception if an error
     *                                occurs while loading the catalog source.
     */
    public void loadCatalog(InputSource source)
        throws Exception 
        {

        // get character reader
        Reader reader = source.getCharacterStream();
        if (reader == null) {
            // get byte stream
            InputStream is = source.getByteStream();
            if (is != null) {
                reader = new InputStreamReader(is);
                }

            // expand system id and create reader
            else {
                com.ibm.xml.framework.StringPool string = state.getStringPool();
                int oldindex = string.addString(source.getSystemId());
                int newindex = state.getEntityHandler().expandSystemId(oldindex);
                if (newindex != oldindex) {
                    string.releaseString(oldindex);
                    }
                String systemId = string.orphanString(newindex);
                reader = new InputStreamReader(new URL(systemId).openStream());
                }
            }

        // read catalog stream
        if (reader != null) {
            BufferedReader in = new BufferedReader(reader);
            String line;
            while ((line = in.readLine()) != null) {
                StringTokenizer tokenizer = new StringTokenizer(line, "\t");
                String publicId = tokenizer.nextToken();
                if (publicId != null) {
                    String systemId = tokenizer.nextToken();
                    if (systemId != null) {
                        addPublicMapping(publicId, systemId);
                        }
                    }
                }
            in.close();
            }

        } // loadCatalog(InputSource)

    //
    // Protected methods
    //

    /** Returns the parser state. */
    protected ParserState getParserState() {
        return state;
        }

    } // class TXCatalog
