/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */
package com.ibm.xml.internal;

import com.ibm.xml.framework.*;

import java.io.IOException;

/**
 *
 * @version Revision: 71 1.8 src/com/ibm/xml/internal/UCSReader.java, parser, xml4j2, xml4j2_0_13 
 */
final class UCSReader extends XMLReader implements StringProducer{

    //
    // Constants
    //

    // debugging
    
    /** Set to true to debug UTF-16, big-endian. */
    private static final boolean DEBUG_UTF16_BIG = false;

    //
    // Scanner encoding enumeration
    //
    static final int
        E_UCS4B = 0,            // UCS-4 big endian
        E_UCS4L = 1,            // UCS-4 little endian
        E_UCS2B = 2,            // UCS-2 big endian with byte order mark
        E_UCS2L = 3,            // UCS-2 little endian with byte order mark
        E_UCS2B_NOBOM = 4,      // UCS-2 big endian without byte order mark
        E_UCS2L_NOBOM = 5;      // UCS-2 little endian without byte order mark
    //
    //
    //
    private ChunkyByteArray fData = null;
    private int fEncoding = -1;
    private StringPool fStringPool = null;
    private XMLDocumentHandler fDocumentHandler = null;
    private XMLErrorHandler fErrorHandler = null;
    private int fBytesPerChar = -1;
    private boolean fBigEndian = true;
    private ChunkyCharArray fStringCharArray = null;
    //
    //
    //
    UCSReader(ParserState parserState, String publicId, String systemId, ChunkyByteArray data, int encoding) throws IOException {
        super(parserState, publicId, systemId);
        fCurrentOffset = (encoding == E_UCS2B || encoding == E_UCS2L) ? 2 : 0;
        fData = data;
        fEncoding = encoding;
        fStringPool = parserState.cacheStringPool();
        fDocumentHandler = parserState.getDocumentHandler();
        fErrorHandler = parserState.getErrorHandler();
        fBytesPerChar = (fEncoding == E_UCS4B || fEncoding == E_UCS4L) ? 4 : 2;
        fBigEndian = fEncoding == E_UCS4B || fEncoding == E_UCS2B || fEncoding == E_UCS2B_NOBOM;
    }
    //
    //
    //
    private int getChar(int offset) {
        int b0 = fData.byteAt(offset++) & 0xff;
        if (b0 == 0xff && fData.atEOF(offset))
            return -1;
        int b1 = fData.byteAt(offset++) & 0xff;
        if (fBytesPerChar == 4) {
            int b2 = fData.byteAt(offset++) & 0xff;
            int b3 = fData.byteAt(offset++) & 0xff;
            if (fBigEndian)
                return (b0<<24)+(b1<<16)+(b2<<8)+b3;
            else
                return (b3<<24)+(b2<<16)+(b1<<8)+b0;
        } else {
            if (fBigEndian)
                return (b0<<8)+b1;
            else
                return (b1<<8)+b0;
        }
    }
    /**
     *
     */
    public int addString(int offset, int length) {
        if (length == 0)
            return 0;
        return fStringPool.addString(this, offset, length);
    }
    /**
     *
     */
    public int addSymbol(int offset, int length) {
        if (length == 0)
            return 0;
        return fStringPool.addSymbol(this, offset, length, getHashcode(offset, length));
    }
    //
    //
    //
    public void append(ChunkyCharArray charArray, int offset, int length) {
        int endOffset = offset + length;
        while (offset < endOffset) {
            int ch = getChar(offset);
            charArray.append((char)ch);
            offset += fBytesPerChar;
        }
    }
    //
    //
    //
    public String toString(int offset, int length) {
        //
        // REVISIT - we need to cache this operation !!
        //
        if (fStringCharArray == null)
            fStringCharArray = new ChunkyCharArray(fStringPool);
        int newOffset = fStringCharArray.length();
        append(fStringCharArray, offset, length);
        int newLength = fStringCharArray.length() - newOffset;
        int stringIndex = fStringCharArray.addString(newOffset, newLength);
        return fStringPool.toString(stringIndex);
    }
    //
    //
    //
    private int getHashcode(int offset, int length) {
        int endOffset = offset + length;
        int hashcode = 0;
        int position = 0;
        while (offset < endOffset) {
            int ch = getChar(offset);
            hashcode = StringHasher.hashChar(hashcode, position++, ch);
            offset += fBytesPerChar;
        }
        return StringHasher.finishHash(hashcode);
    }
    //
    public boolean equalsString(int offset, int length, String str, int strLength) {
        int endOffset = offset + length;
        int slen = strLength;
        int i = 0;
        while (offset < endOffset) {
            if (slen-- == 0)
                return false;
            int ch = getChar(offset);
            if (ch != str.charAt(i++))
                return false;
            offset += fBytesPerChar;
        }
        return slen == 0;
    }
    //
    //
    //
    private static char[] fCharacters = new char[256];
    private int fCharDataLength = 0;
    private void appendCharData(int ch) {
        if (fCharacters.length == fCharDataLength) {
            char[] newchars = new char[fCharacters.length * 2];
            System.arraycopy(fCharacters, 0, newchars, 0, fCharacters.length);
            fCharacters = newchars;
        }
        fCharacters[fCharDataLength++] = (char)ch;
    }
    public void callWSCharDataHandler(int offset, int length, boolean inCDSect) throws Exception {
        int endOffset = offset + length;
        while (offset < endOffset) {
            int ch = getChar(offset);
            appendCharData(ch);
            offset += fBytesPerChar;
        }
        if (fDocumentHandler.sendCharDataAsCharArray()) {
            fDocumentHandler.ignorableWhitespace(fCharacters, 0, fCharDataLength, inCDSect);
        } 
        else {
            //revisit: can we call the other addString and avoid the new String?
            int stringIndex = fStringPool.addString(new String(fCharacters, 0, fCharDataLength));
            fDocumentHandler.ignorableWhitespace(stringIndex, inCDSect);
        }
        fCharDataLength = 0;
    }
    public void callCharDataHandler(int offset, int length, boolean inCDSect) throws Exception {
        int endOffset = offset + length;
        while (offset < endOffset) {
            int ch = getChar(offset);
            appendCharData(ch);
            offset += fBytesPerChar;
        }
        if (fDocumentHandler.sendCharDataAsCharArray()) {
            fDocumentHandler.characters(fCharacters, 0, fCharDataLength, inCDSect);
        } 
        else {
            //revisit: can we call the other addString and avoid the new String?
            int stringIndex = fStringPool.addString(new String(fCharacters, 0, fCharDataLength));
            fDocumentHandler.characters(stringIndex, inCDSect);
        }
        fCharDataLength = 0;
    }
    //
    //
    //
    public int skipOneChar() throws IOException {
        fCurrentOffset += fBytesPerChar;
        return fCurrentOffset;
    }
    public int skipAsciiChar() throws IOException {
        fCurrentOffset += fBytesPerChar;
        return fCurrentOffset;
    }
    public int skipToChar(char ch) throws IOException {
        while (true) {
            int ch2 = getChar(fCurrentOffset);
            if (ch2 == ch)
                return fCurrentOffset;
            fCurrentOffset += fBytesPerChar;
        }
    }
    public int skipPastChar(char ch) throws IOException {
        while (true) {
            int ch2 = getChar(fCurrentOffset);
            fCurrentOffset += fBytesPerChar;
            if (ch2 == ch)
                return fCurrentOffset;
        }
    }
    public boolean skippedValidChar() throws IOException {
        int ch = getChar(fCurrentOffset);
        fCurrentOffset += fBytesPerChar;
        if (ch < 0x20) {
            if (ch == 0x09) {
                fCharacterCounter++;
                return true;
            } else if (ch == 0x0A) {
                fLinefeedCounter++;
                fCharacterCounter = 1;
                return true;
            } else if (ch == 0x0D) {
                fCarriageReturnCounter++;
                fCharacterCounter = 1;
                return true;
            }
            fCurrentOffset--;
            if (ch == -1)
                fData.checkEOF(fCurrentOffset + 1);
            return false;
        }
        fCharacterCounter++;
        if (ch <= 0xD7FF)
            return true;
        if (ch <= 0xDFFF) {
            // REVISIT - check that the surrogate pair is valid 
            fCurrentOffset++;
            return true;
        }
        if (ch <= 0xFFFD)
            return true;
        fCharacterCounter--;
        fCurrentOffset -= fBytesPerChar;
        return false;
    }
    public boolean lookingAtValidChar() throws IOException {
        int ch = getChar(fCurrentOffset);
        if (ch < 0x20) {
            if (ch == -1)
                fData.checkEOF(fCurrentOffset + 1);
            return (ch == 0x09 || ch == 0x0A || ch == 0x0D);
        }
        return (ch <= 0xD7FF || (ch >= 0xE000 && (ch <= 0xFFFD || (ch >= 0x10000 && ch <= 0x10FFFF))));
    }
    public int skipInvalidChar(int errorCode) throws Exception {
        int ch = getChar(fCurrentOffset);
        fCurrentOffset += fBytesPerChar;
        switch (errorCode) {
        case ErrorCode.E_INVCHAR0:
        case ErrorCode.E_ENTITY3:
        {
            String arg = Integer.toHexString(ch);
            fErrorHandler.error1(errorCode, fStringPool.addString(arg));
            break;
        }
        case ErrorCode.E_REFER0:
        case ErrorCode.E_EXT2:
        case ErrorCode.E_EXT4:
        {
            String arg = new Character((char)ch).toString();
            fErrorHandler.error1(errorCode, fStringPool.addString(arg));
            break;
        }
        case ErrorCode.E_DTD0:
        {
            String arg1 = new Character((char)ch).toString();
            String arg2 = Integer.toHexString(ch);
            fErrorHandler.error2(errorCode, fStringPool.addString(arg1), fStringPool.addString(arg2));
            break;
        }
        default:
            break;
        }
        return fCurrentOffset;
    }
    //
    //
    //
    public boolean skippedChar(char ch) throws IOException {
        int ch2 = getChar(fCurrentOffset);
        if (ch2 == ch) {
            fCurrentOffset += fBytesPerChar;
            return true;
        }
        return false;
    }
    public boolean lookingAtChar(char ch) throws IOException {
        return ch == getChar(fCurrentOffset);
    }
    public boolean skippedSpace() throws IOException {
        int ch = getChar(fCurrentOffset);
        if (ch == 0x20 || ch == 0x09 || ch == 0x0A || ch == 0x0D) {
            fCurrentOffset += fBytesPerChar;
            return true;
        }
        return false;
    }
    public boolean lookingAtSpace() throws IOException {
        int ch = getChar(fCurrentOffset);
        return (ch == 0x20 || ch == 0x09 || ch == 0x0A || ch == 0x0D);
    }
    public int skipPastSpaces() throws IOException {
        while (true) {
            int ch = getChar(fCurrentOffset);
            if (ch == 0x20 || ch == 0x09 || ch == 0x0A || ch == 0x0D)
                fCurrentOffset += fBytesPerChar;
            else
                return fCurrentOffset;
        }
    }
    public int skipDecimalDigit() throws IOException {
        int ch = getChar(fCurrentOffset);
        if (ch < '0' || ch > '9')
            return -1;
        fCurrentOffset += fBytesPerChar;
        fCharacterCounter++;
        return ch - '0';
    }
    public int skipHexDigit() throws IOException {
        int ch = getChar(fCurrentOffset);
        if (ch > 'f' || fgAsciiXDigitChar[ch] == 0)
            return -1;
        fCurrentOffset += fBytesPerChar;
        fCharacterCounter++;
        return ch - (ch < 'A' ? '0' : (ch < 'a' ? 'A' : 'a') - 10);
    }
    public boolean skippedAlpha() throws IOException {
        int ch = getChar(fCurrentOffset);
        if (ch <= 'z' && fgAsciiAlphaChar[ch] == 1) {
            fCurrentOffset += fBytesPerChar;
            fCharacterCounter++;
            return true;
        }
        return false;
    }
    private boolean skippedCharWithFlag(byte flag) {
        int ch = getChar(fCurrentOffset);
        if (ch < 0x80 && (fgCharFlags[ch] & flag) != 0) {
            fCurrentOffset += fBytesPerChar;
            fCharacterCounter++;
            return true;
        }
        return false;
    }
    public final boolean skippedVersionNum() {
        return skippedCharWithFlag(E_VersionNumFlag);
    }
    public final boolean skippedEncName() {
        return skippedCharWithFlag(E_EncNameFlag);
    }
    public final boolean skippedPubidChar() {
        int ch = getChar(fCurrentOffset);
        if (ch < 0x80) {
            if ((fgCharFlags[ch] & E_PubidCharFlag) != 0) {
                fCurrentOffset += fBytesPerChar;
                fCharacterCounter++;
                return true;
            }
            if (ch == 0x0A) {
                fCurrentOffset += fBytesPerChar;
                fLinefeedCounter++;
                fCharacterCounter = 1;
                return true; // #xA
            }
            if (ch == 0x0D) {
                fCurrentOffset += fBytesPerChar;
                fCarriageReturnCounter++;
                fCharacterCounter = 1;
                return true; // #xD
            }
        }
        return false;
    }
    public boolean skippedString(char[] s) throws IOException {
        int offset = fCurrentOffset;
        for (int i = 0; i < s.length; i++) {
            if (getChar(offset) != s[i])
                return false;
            offset += fBytesPerChar;
        }
        fCurrentOffset = offset;
        fCharacterCounter += s.length;
        return true;
    }
    public int scanName(char fastcheck, int expectedName) throws IOException {
        int nameOffset = fCurrentOffset;
        int nameLength = skipPastName(fastcheck) - nameOffset;
        if (nameLength == 0)
            return -1;
        int nameIndex = addSymbol(nameOffset, nameLength);
        return (expectedName == -1 || expectedName == nameIndex) ? nameIndex : -1;
    }
    public int skipPastName(char fastcheck) throws IOException {
        int ch = getChar(fCurrentOffset);
        if ((fgCharFlags[ch] & E_InitialNameCharFlag) == 0)
            return fCurrentOffset;
        while (true) {
            fCurrentOffset += fBytesPerChar;
            fCharacterCounter++;
            ch = getChar(fCurrentOffset);
            if (fastcheck == ch)
                return fCurrentOffset;
            if ((fgCharFlags[ch] & E_NameCharFlag) == 0)
                return fCurrentOffset;
        }
    }
    public int skipPastNmtoken(char fastcheck) throws IOException {
        int ch = getChar(fCurrentOffset);
        while (true) {
            if (fastcheck == ch)
                return fCurrentOffset;
            if ((fgCharFlags[ch] & E_NameCharFlag) == 0)
                return fCurrentOffset;
            fCurrentOffset += fBytesPerChar;
            fCharacterCounter++;
            ch = getChar(fCurrentOffset);
        }
    }
    //
    // REVISIT - still does not handle cdata sections and does not
    //   check sendCharDataAsCharArray state.
    //
    public int scanContent(ScanContentState scanState) throws Exception {
        int offset = fCurrentOffset;
        int ch = getChar(fCurrentOffset);
        fCurrentOffset += fBytesPerChar;
        byte prop;
        if (ch < 0x80) {
            if (ch == -1) {
                fCurrentOffset -= fBytesPerChar;
                return CHARDATA_RESULT_INVALID_CHAR;
            }
            prop = fgCharFlags[ch];
            if ((prop & E_CharDataFlag) == 0 && ch != 0x0A && ch != 0x0D) {
                if (ch == '<') {
                    fCharacterCounter++;
                    return CHARDATA_RESULT_MARKUP;
                }
                if (ch == '&') {
                    fCharacterCounter++;
                    return CHARDATA_RESULT_REFERENCE;
                }
                // REVISIT - EOF check?
                if (ch == ']') {
                    if (getChar(fCurrentOffset) == ']' && getChar(fCurrentOffset + fBytesPerChar) == '>') {
                        fCharacterCounter += 3;
                        fCurrentOffset += (2 * fBytesPerChar);
                        return CHARDATA_RESULT_CDEND;
                    }
                } else {
                    fCurrentOffset -= fBytesPerChar;
                    return CHARDATA_RESULT_INVALID_CHAR;
                }
            } else if (ch == 0x20 || ch == 0x09 || ch == 0x0A || ch == 0x0D) {
                do {
                    if (ch == 0x0A) {
                        fLinefeedCounter++;
                        fCharacterCounter = 1;
                    } else if (ch == 0x0D) {
                        fCarriageReturnCounter++;
                        fCharacterCounter = 1;
                    } else
                        fCharacterCounter++;
                    ch = getChar(fCurrentOffset);
                    fCurrentOffset += fBytesPerChar;
                } while (ch == 0x20 || ch == 0x09 || ch == 0x0A || ch == 0x0D);
                if (ch < 0x80) {
                    if (ch == -1) {
                        fCurrentOffset -= fBytesPerChar;
                        if (fDocumentHandler != null)
                            callWSCharDataHandler(offset, fCurrentOffset - offset, scanState.inCDSect);
                        return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                    }
                    prop = fgCharFlags[ch];
                    if ((prop & E_CharDataFlag) == 0) {
                        if (ch == '<') {
                            fCharacterCounter++;
                            if (fDocumentHandler != null)
                                callWSCharDataHandler(offset, (fCurrentOffset - fBytesPerChar) - offset, scanState.inCDSect);
                            return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_MARKUP);
                        }
                        if (ch == '&') {
                            fCharacterCounter++;
                            if (fDocumentHandler != null)
                                callWSCharDataHandler(offset, (fCurrentOffset - fBytesPerChar) - offset, scanState.inCDSect);
                            return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_REFERENCE);
                        }
                        // REVISIT - EOF check?
                        if (ch == ']') {
                            if (getChar(fCurrentOffset) == ']' && getChar(fCurrentOffset + fBytesPerChar) == '>') {
                                if (fDocumentHandler != null)
                                    callWSCharDataHandler(offset, (fCurrentOffset - fBytesPerChar) - offset, scanState.inCDSect);
                                fCharacterCounter += 3;
                                fCurrentOffset += (2 * fBytesPerChar);
                                return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_CDEND);
                            }
                        } else {
                            fCurrentOffset -= fBytesPerChar;
                            if (fDocumentHandler != null)
                                callWSCharDataHandler(offset, fCurrentOffset - offset, scanState.inCDSect);
                            return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                        }
                    }
                } else {
                    if (ch >= 0xD800 && ch <= 0xDFFF) {
                        fCurrentOffset += fBytesPerChar;
                    } else if (ch == 0xFFFE || ch == 0xFFFF) {
                        fCurrentOffset -= fBytesPerChar;
                        if (fDocumentHandler != null)
                            callWSCharDataHandler(offset, fCurrentOffset - offset, scanState.inCDSect);
                        return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                    }
                }
            }
        } else {
            if (ch >= 0xD800 && ch <= 0xDFFF) {
                fCurrentOffset += fBytesPerChar;
            } else if (ch == 0xFFFE || ch == 0xFFFF) {
                fCurrentOffset -= fBytesPerChar;
                return CHARDATA_RESULT_INVALID_CHAR;
            }
        }
        fCharacterCounter++;
        while (true) {
            ch = getChar(fCurrentOffset);
            fCurrentOffset += fBytesPerChar;
            if (ch >= 0x80 || ch < 0)
                break;
            prop = fgCharFlags[ch];
            if ((prop & E_CharDataFlag) == 0) {
                if (ch == 0x0A) {
                    fLinefeedCounter++;
                    fCharacterCounter = 1;
                } else if (ch == 0x0D) {
                    fCarriageReturnCounter++;
                    fCharacterCounter = 1;
                } else
                    break;
            } else
                fCharacterCounter++;
        }
        while (true) { // REVISIT - EOF check ?
            if (ch < 0x80) {
                if (ch == -1) {
                    fCurrentOffset -= fBytesPerChar;
                    if (fDocumentHandler != null)
                        callCharDataHandler(offset, fCurrentOffset - offset, scanState.inCDSect);
                    return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                }
                prop = fgCharFlags[ch];
                if ((prop & E_CharDataFlag) == 0) {
                    if (ch == '<') {
                        fCharacterCounter++;
                        if (fDocumentHandler != null)
                            callCharDataHandler(offset, (fCurrentOffset - fBytesPerChar) - offset, scanState.inCDSect);
                        return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_MARKUP);
                    }
                    if (ch == '&') {
                        fCharacterCounter++;
                        if (fDocumentHandler != null)
                            callCharDataHandler(offset, (fCurrentOffset - fBytesPerChar) - offset, scanState.inCDSect);
                        return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_REFERENCE);
                    }
                    // REVISIT - EOF check?
                    if (ch == 0x0A) {
                        fLinefeedCounter++;
                        fCharacterCounter = 1;
                    } else if (ch == 0x0D) {
                        fCarriageReturnCounter++;
                        fCharacterCounter = 1;
                    } else if (ch == ']') {
                        if (getChar(fCurrentOffset) == ']' && getChar(fCurrentOffset + fBytesPerChar) == '>') {
                            if (fDocumentHandler != null)
                                callCharDataHandler(offset, (fCurrentOffset - fBytesPerChar) - offset, scanState.inCDSect);
                            fCharacterCounter += 3;
                            fCurrentOffset += (2 * fBytesPerChar);
                            return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_CDEND);
                        }
                        fCharacterCounter++;
                    } else {
                        fCurrentOffset -= fBytesPerChar;
                        if (fDocumentHandler != null)
                            callCharDataHandler(offset, fCurrentOffset - offset, scanState.inCDSect);
                        return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                    }
                } else
                    fCharacterCounter++;
            } else {
                if (ch >= 0xD800 && ch <= 0xDFFF) {
                    fCharacterCounter++;
                    fCurrentOffset += fBytesPerChar;
                } else if (ch == 0xFFFE || ch == 0xFFFF) {
                    fCurrentOffset -= fBytesPerChar;
                    if (fDocumentHandler != null)
                        callCharDataHandler(offset, fCurrentOffset - offset, scanState.inCDSect);
                    return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                }
                fCharacterCounter++;
            }
            ch = getChar(fCurrentOffset);
            fCurrentOffset += fBytesPerChar;
        }
    }
}
