/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */
package com.ibm.xml.internal;

import com.ibm.xml.framework.*;

import java.io.InputStream;

/**
 *
 * @version Revision: 72 1.11 src/com/ibm/xml/internal/UTF8Reader.java, parser, xml4j2, xml4j2_0_13 
 */
final class UTF8Reader extends XMLReader {
    //
    // Chunk size constants
    //
    protected static final int CHUNK_SHIFT = 14;           // 2^14 = 16k
    protected static final int CHUNK_SIZE = (1 << CHUNK_SHIFT);
    protected static final int CHUNK_MASK = CHUNK_SIZE - 1;
    //
    //
    //
    UTF8Reader(ParserState parserState, String publicId, String systemId, InputStream dataStream) throws Exception {
        super(parserState, publicId, systemId);
        fInputStream = dataStream;
        fScanner = parserState.getScanner();
        fStringPool = parserState.cacheStringPool();
        fDocumentHandler = parserState.getDocumentHandler();
        fSendCharDataAsCharArray = (fDocumentHandler != null && fDocumentHandler.sendCharDataAsCharArray());
        fErrorHandler = parserState.getErrorHandler();
        fCurrentChunk = new UTF8DataChunk(fStringPool, null);
        fillCurrentChunk();
    }
    /**
     *
     */
    public int addString(int offset, int length) {
        if (length == 0)
            return 0;
        return fCurrentChunk.addString(offset, length);
    }
    /**
     *
     */
    public int addSymbol(int offset, int length) {
        if (length == 0)
            return 0;
        return fCurrentChunk.addSymbol(offset, length, 0);
    }
    /**
     *
     */
    protected int addSymbol(int offset, int length, int hashcode) {
        return fCurrentChunk.addSymbol(offset, length, hashcode);
    }
    /**
     *
     */
    public void append(ChunkyCharArray charArray, int offset, int length) {
        fCurrentChunk.append(charArray, offset, length);
    }
    //
    //
    //
    protected int slowLoadNextByte()  throws Exception {
        if (fCurrentChunk.nextChunk() != null) {
            fCurrentChunk = fCurrentChunk.nextChunk();
            fCurrentIndex = 0;
            fMostRecentData = fCurrentChunk.toByteArray();
            return (fMostRecentByte = fMostRecentData[fCurrentIndex] & 0xFF);
        } else {
            fCurrentChunk = new UTF8DataChunk(fStringPool, fCurrentChunk);
            return fillCurrentChunk();
        }
    }
    protected int loadNextByte() throws Exception {
        fCurrentOffset++;
        if (++fCurrentIndex == CHUNK_SIZE)
            return slowLoadNextByte();
        return (fMostRecentByte = fMostRecentData[fCurrentIndex] & 0xFF);
    }
    //
    //
    //
    protected void checkEOF(int offset) {
        if (offset > fLength)
            throw new ArrayIndexOutOfBoundsException();
    }
    //
    //
    //
    protected void skipOneCharStartingWithByte(int b0) throws Exception {
        if ((0x80 & b0) == 0) {  // 0xxxxxxx
            if (b0 == 0) {
                //
                // REVISIT - shouldn't happen... the exception should get thrown
                //  when you look at the EOF before you ask to skip past it.
                //
                checkEOF(fCurrentOffset + 1);
                fCharacterCounter++;
            } else if (b0 == 0x0A) {
                fLinefeedCounter++;
                fCharacterCounter = 1;
            } else
                fCharacterCounter++;
            return;
        }
        fCharacterCounter++;
        int b1 = loadNextByte();
        if ((0xe0 & b0) == 0xc0) // 110yyyyy 10xxxxxx
            return;
        int b2 = loadNextByte();
        if ((0xf0 & b0) == 0xe0) // 1110zzzz 10yyyyyy 10xxxxxx
            return;
        int b3 = loadNextByte();
        return;  // 11110uuu 10uuzzzz 10yyyyyy 10xxxxxx
    }
    public int skipOneChar() throws Exception {
        int b0 = fMostRecentByte;
        skipOneCharStartingWithByte(b0);
        loadNextByte();
        return fCurrentOffset;
    }
    public int skipAsciiChar() throws Exception {
        fCharacterCounter++; // we never call skip_asciiChar for CR or LF.
        loadNextByte();
        return fCurrentOffset;
    }
    public int skipToChar(char ch) throws Exception {
        int b0 = fMostRecentByte;
        while (true) {
            if (b0 == ch) // ch will always be an ascii character
                return fCurrentOffset;
            skipOneCharStartingWithByte(b0);
            b0 = loadNextByte();
        }
    }
    public int skipPastChar(char ch) throws Exception {
        int b0 = fMostRecentByte;
        while (true) {
            if (b0 == ch) { // ch will always be an ascii character
                fCharacterCounter++; // ch will never be CR or LF.
                loadNextByte();
                return fCurrentOffset;
            }
            skipOneCharStartingWithByte(b0);
            b0 = loadNextByte();
        }
    }
    //
    //
    //
    public boolean skippedValidChar() throws Exception {
        int b0 = fMostRecentByte;
        if ((0x80 & b0) == 0) {  // 0xxxxxxx
            if (b0 >= 0x20 || b0 == 0x09) {
                fCharacterCounter++;
                loadNextByte();
                return true;
            }
            if (b0 == 0x0A) {
                fLinefeedCounter++;
                fCharacterCounter = 1;
                loadNextByte();
                return true;
            }
            if (b0 == 0)
                checkEOF(fCurrentOffset + 1);
            return false;
        }
        //
        // REVISIT - optimize this with in-buffer lookahead.
        //
        UTF8DataChunk saveChunk = fCurrentChunk;
        int saveIndex = fCurrentIndex;
        int saveOffset = fCurrentOffset;
        int b1 = loadNextByte();
        if ((0xe0 & b0) == 0xc0) { // 110yyyyy 10xxxxxx (0x80 to 0x7ff)
            fCharacterCounter++;
            loadNextByte();
            return true; // [#x20-#xD7FF]
        }
        int b2 = loadNextByte();
        if ((0xf0 & b0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
            // ch = ((0x0f & b0)<<12) + ((0x3f & b1)<<6) + (0x3f & b2); // zzzz yyyy yyxx xxxx (0x800 to 0xffff)
            // if (!((ch >= 0xD800 && ch <= 0xDFFF) || ch >= 0xFFFE))
            // if ((ch <= 0xD7FF) || (ch >= 0xE000 && ch <= 0xFFFD))
            if (!((b0 == 0xED && b1 >= 0xA0) || (b0 == 0xEF && b1 == 0xBF && b2 >= 0xBE))) { // [#x20-#xD7FF] | [#xE000-#xFFFD]
                fCharacterCounter++;
                loadNextByte();
                return true;
            }
            fCurrentChunk = saveChunk;
            fCurrentIndex = saveIndex;
            fCurrentOffset = saveOffset;
            fMostRecentData = saveChunk.toByteArray();
            fMostRecentByte = b0;
            return false;
        }
        int b3 = loadNextByte();  // 11110uuu 10uuzzzz 10yyyyyy 10xxxxxx
        // ch = ((0x0f & b0)<<18) + ((0x3f & b1)<<12) + ((0x3f & b2)<<6) + (0x3f & b3); // u uuuu zzzz yyyy yyxx xxxx (0x10000 to 0x1ffff)
        // if (ch >= 0x110000)
        if (!(b0 > 0xF4 || (b0 == 0xF4 && b1 >= 0x90))) { // [#x10000-#x10FFFF]
            fCharacterCounter++;
            loadNextByte();
            return true;
        }
        fCurrentChunk = saveChunk;
        fCurrentIndex = saveIndex;
        fCurrentOffset = saveOffset;
        fMostRecentData = saveChunk.toByteArray();
        fMostRecentByte = b0;
        return false;
    }
    //
    //
    //
    public boolean lookingAtValidChar() throws Exception {
        int b0 = fMostRecentByte;
        if ((0x80 & b0) == 0) {  // 0xxxxxxx
            if (b0 >= 0x20 || b0 == 0x09 || b0 == 0x0A)
                return true; // #x9 | #xA | #xD | [#x20-#xD7FF]
            if (b0 == 0)
                checkEOF(fCurrentOffset + 1);
            return false;
        }
        //
        // REVISIT - optimize this with in-buffer lookahead.
        //
        UTF8DataChunk saveChunk = fCurrentChunk;
        int saveIndex = fCurrentIndex;
        int saveOffset = fCurrentOffset;
        int b1 = loadNextByte();
        if ((0xe0 & b0) == 0xc0) { // 110yyyyy 10xxxxxx (0x80 to 0x7ff)
            fCurrentChunk = saveChunk;
            fCurrentIndex = saveIndex;
            fCurrentOffset = saveOffset;
            fMostRecentData = saveChunk.toByteArray();
            fMostRecentByte = b0;
            return true; // [#x20-#xD7FF]
	}
        int b2 = loadNextByte();
        if ((0xf0 & b0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
            // ch = ((0x0f & b0)<<12) + ((0x3f & b1)<<6) + (0x3f & b2); // zzzz yyyy yyxx xxxx (0x800 to 0xffff)
            // if (!((ch >= 0xD800 && ch <= 0xDFFF) || ch >= 0xFFFE))
            // if ((ch <= 0xD7FF) || (ch >= 0xE000 && ch <= 0xFFFD))
            fCurrentChunk = saveChunk;
            fCurrentIndex = saveIndex;
            fCurrentOffset = saveOffset;
            fMostRecentData = saveChunk.toByteArray();
            fMostRecentByte = b0;
            return (!((b0 == 0xED && b1 >= 0xA0) || (b0 == 0xEF && b1 == 0xBF && b2 >= 0xBE))); // [#x20-#xD7FF] | [#xE000-#xFFFD]
        }
        int b3 = loadNextByte();  // 11110uuu 10uuzzzz 10yyyyyy 10xxxxxx
        // ch = ((0x0f & b0)<<18) + ((0x3f & b1)<<12) + ((0x3f & b2)<<6) + (0x3f & b3); // u uuuu zzzz yyyy yyxx xxxx (0x10000 to 0x1ffff)
        // if (ch >= 0x110000)
        fCurrentChunk = saveChunk;
        fCurrentIndex = saveIndex;
        fCurrentOffset = saveOffset;
        fMostRecentData = saveChunk.toByteArray();
        fMostRecentByte = b0;
        return (!(b0 > 0xF4 || (b0 == 0xF4 && b1 >= 0x90))); // [#x10000-#x10FFFF]
    }
    //
    //
    //
    public int skipInvalidChar(int errorCode) throws Exception {
        int ch;
        int b0 = fMostRecentByte;
        fCharacterCounter++;
        if ((b0 & 0x80) == 0) {
            ch = b0;
        } else {
            int b1 = loadNextByte();
            if ((0xe0 & b0) == 0xc0) { // 110yyyyy 10xxxxxx
                ch = ((0x1f & b0)<<6) + (0x3f & b1);
            } else {
                int b2 = loadNextByte();
                if ((0xf0 & b0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
                    // if ((ch >= 0xD800 && ch <= 0xDFFF) || ch >= 0xFFFE)
                    ch = ((0x0f & b0)<<12) + ((0x3f & b1)<<6) + (0x3f & b2);
                } else {
                    int b3 = loadNextByte(); // 11110uuu 10uuzzzz 10yyyyyy 10xxxxxx
                    ch = ((0x0f & b0)<<18) + ((0x3f & b1)<<12)
                           + ((0x3f & b2)<<6) + (0x3f & b3);
                }
            }
        }
        switch (errorCode) {
        case ErrorCode.E_INVCHAR0:
        case ErrorCode.E_ENTITY3:
        {
            String arg = Integer.toHexString(ch);
            fErrorHandler.error1(errorCode, fStringPool.addString(arg));
            break;
        }
        case ErrorCode.E_REFER0:
        case ErrorCode.E_EXT2:
        case ErrorCode.E_EXT4:
        {
            String arg = new Character((char)ch).toString();
            fErrorHandler.error1(errorCode, fStringPool.addString(arg));
            break;
        }
        case ErrorCode.E_DTD0:
        {
            String arg1 = new Character((char)ch).toString();
            String arg2 = Integer.toHexString(ch);
            fErrorHandler.error2(errorCode, fStringPool.addString(arg1), fStringPool.addString(arg2));
            break;
        }
        default:
            break;
        }
        loadNextByte();
        return fCurrentOffset;
    }
    //
    //
    //
    public boolean skippedChar(char ch) throws Exception {
        int b0 = fMostRecentByte;
        if (b0 != ch)
            return false;
        fCharacterCounter++;
        fCurrentOffset++;
        if (++fCurrentIndex == CHUNK_SIZE)
            slowLoadNextByte();
        else
            fMostRecentByte = fMostRecentData[fCurrentIndex] & 0xFF;
        return true;
    }
    public boolean lookingAtChar(char ch) {
        int b0 = fMostRecentByte;
        return b0 == ch;
    }
    //
    //
    //
    public boolean skippedSpace() throws Exception {
        int ch = fMostRecentByte;
        if (ch > 0x20)
            return false;
        if (ch == 0x20 || ch == 0x09)
            fCharacterCounter++;
        else if (ch == 0x0A) {
            fLinefeedCounter++;
            fCharacterCounter = 1;
        } else
            return false;
        fCurrentOffset++;
        if (++fCurrentIndex == CHUNK_SIZE)
            slowLoadNextByte();
        else
            fMostRecentByte = fMostRecentData[fCurrentIndex] & 0xFF;
        return true;
    }
    public boolean lookingAtSpace() {
        int ch = fMostRecentByte;
        return (ch == 0x20 || ch == 0x09 || ch == 0x0A);
    }
    //
    //
    //
    public int skipPastSpaces() throws Exception {
        int ch = fMostRecentByte;
        while (true) {
            if (ch == 0x20 || ch == 0x09)
                fCharacterCounter++;
            else if (ch == 0x0A) {
                fLinefeedCounter++;
                fCharacterCounter = 1;
            } else
                return fCurrentOffset;
            ch = loadNextByte();
        }
    }
    //
    //
    //
    public int skipDecimalDigit() throws Exception {
        int ch = fMostRecentByte;
        if (ch < '0' || ch > '9')
            return -1;
        fCharacterCounter++;
        loadNextByte();
        return ch - '0';
    }
    public int skipHexDigit() throws Exception {
        int ch = fMostRecentByte;
        if (ch > 'f' || fgAsciiXDigitChar[ch] == 0)
            return -1;
        fCharacterCounter++;
        loadNextByte();
        return ch - (ch < 'A' ? '0' : (ch < 'a' ? 'A' : 'a') - 10);
    }
    public boolean skippedAlpha() throws Exception {
        int ch = fMostRecentByte;
        if (ch > 'z' || fgAsciiAlphaChar[ch] == 0)
            return false;
        fCharacterCounter++;
        loadNextByte();
        return true;
    }
    protected final boolean skippedAsciiCharWithFlag(byte flag) throws Exception {
        int ch = fMostRecentByte;
        if ((ch & 0x80) != 0 || (fgCharFlags[ch] & flag) == 0)
            return false;
        fCharacterCounter++;
        loadNextByte();
        return true;
    }
    public final boolean skippedVersionNum() throws Exception {
        return skippedAsciiCharWithFlag(E_VersionNumFlag);
    }
    public final boolean skippedEncName() throws Exception {
        return skippedAsciiCharWithFlag(E_EncNameFlag);
    }
    public final boolean skippedPubidChar() throws Exception {
        int ch = fMostRecentByte;
        if ((ch & 0x80) != 0)
            return false;
        if ((fgCharFlags[ch] & E_PubidCharFlag) != 0) {
            fCharacterCounter++;
            loadNextByte();
            return true;
        }
        if (ch == 0x0A) {
            fLinefeedCounter++;
            fCharacterCounter = 1;
            loadNextByte();
            return true;
        }
        return false;
    }
    public boolean skippedString(char[] s) throws Exception {
        int length = s.length;
        byte[] data = fMostRecentData;
        int index = fCurrentIndex;
        if (index + length <= CHUNK_SIZE) {
            for (int i = 0; i < length; i++) {
                if (data[index++] != s[i])
                    return false;
            }
            fCharacterCounter += length;
            fCurrentOffset += length;
            fCurrentIndex = index;
            if (index == CHUNK_SIZE)
                slowLoadNextByte();
            else
                fMostRecentByte = data[index] & 0xFF;
            return true;
        }
        UTF8DataChunk dataChunk = fCurrentChunk;
        int offset = fCurrentOffset;
        int savedIndex = index;
        int i = 0;
        while (index < CHUNK_SIZE) {
            if (data[index++] != s[i++])
                return false;
        }
        slowLoadNextByte();
        data = fMostRecentData;
        index = 0;
        while (i < length) {
            if (data[index++] != s[i++]) {
                fCurrentChunk = dataChunk;
                fCurrentIndex = savedIndex;
                fCurrentOffset = offset;
                fMostRecentData = dataChunk.toByteArray();
                fMostRecentByte = fMostRecentData[savedIndex] & 0xFF;
                return false;
            }
        }
        fCharacterCounter += length;
        fCurrentOffset += length;
        fCurrentIndex = index;
        if (index == CHUNK_SIZE)
            slowLoadNextByte();
        else
            fMostRecentByte = data[index] & 0xFF;
        return true;
    }
    //
    //
    //
    public int scanName(char fastcheck, int expectedName) throws Exception {
        int offset = fCurrentOffset;
        byte[] data = fMostRecentData;
        int b0 = fMostRecentByte;
        int ch;
        if ((b0 & 0x80) == 0) {
            if (fgAsciiInitialNameChar[b0] == 0)
                return -1;
            ch = b0;
        } else {
            //
            // REVISIT - optimize this with in-buffer lookahead.
            //
            UTF8DataChunk saveChunk = fCurrentChunk;
            int saveIndex = fCurrentIndex;
            fCurrentOffset++;
            if (++fCurrentIndex == CHUNK_SIZE) {
                fCurrentChunk = new UTF8DataChunk(fStringPool, fCurrentChunk);
                fillCurrentChunk();
                data = fMostRecentData;
            }
            int b1 = data[fCurrentIndex] & 0xFF;
            if ((0xe0 & b0) == 0xc0) { // 110yyyyy 10xxxxxx
                ch = ((0x1f & b0)<<6) + (0x3f & b1);
                if ((fgCharFlags[ch] & E_InitialNameCharFlag) == 0) { // yyy yyxx xxxx (0x80 to 0x7ff)
                    fCurrentChunk = saveChunk;
                    fCurrentIndex = saveIndex;
                    fCurrentOffset = offset;
                    fMostRecentData = saveChunk.toByteArray();
                    fMostRecentByte = b0;
                    return -1;
                }
            } else {
                fCurrentOffset++;
                if (++fCurrentIndex == CHUNK_SIZE) {
                    fCurrentChunk = new UTF8DataChunk(fStringPool, fCurrentChunk);
                    fillCurrentChunk();
                    data = fMostRecentData;
                }
                int b2 = data[fCurrentIndex] & 0xFF;
                if ((0xf0 & b0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
                    // if ((ch >= 0xD800 && ch <= 0xDFFF) || ch >= 0xFFFE)
                    if ((b0 == 0xED && b1 >= 0xA0) || (b0 == 0xEF && b1 == 0xBF && b2 >= 0xBE)) {
                        fCurrentChunk = saveChunk;
                        fCurrentIndex = saveIndex;
                        fCurrentOffset = offset;
                        fMostRecentData = saveChunk.toByteArray();
                        fMostRecentByte = b0;
                        return -1;
                    }
                    ch = ((0x0f & b0)<<12) + ((0x3f & b1)<<6) + (0x3f & b2);
                    if ((fgCharFlags[ch] & E_InitialNameCharFlag) == 0) { // zzzz yyyy yyxx xxxx (0x800 to 0xffff)
                        fCurrentChunk = saveChunk;
                        fCurrentIndex = saveIndex;
                        fCurrentOffset = offset;
                        fMostRecentData = saveChunk.toByteArray();
                        fMostRecentByte = b0;
                        return -1;
                    }
                } else { // 11110uuu 10uuzzzz 10yyyyyy 10xxxxxx
                    fCurrentChunk = saveChunk;
                    fCurrentIndex = saveIndex;
                    fCurrentOffset = offset;
                    fMostRecentData = saveChunk.toByteArray();
                    fMostRecentByte = b0;
                    return -1;
                }
            }
        }
        int hashcode = 0;
        int position = 0;
        while (true) {
            hashcode = StringHasher.hashChar(hashcode, position++, ch);
            fCharacterCounter++;
            fCurrentOffset++;
            if (++fCurrentIndex == CHUNK_SIZE) {
                fCurrentChunk = new UTF8DataChunk(fStringPool, fCurrentChunk);
                fillCurrentChunk();
                data = fMostRecentData;
            }
            b0 = data[fCurrentIndex] & 0xFF;
            if (fastcheck == b0)
                break;
            if ((b0 & 0x80) == 0) {
                if (fgAsciiNameChar[b0] == 0)
                    break;
                ch = b0;
                continue;
            }
            UTF8DataChunk saveChunk = fCurrentChunk;
            int saveOffset = fCurrentOffset;
            int saveIndex = fCurrentIndex;
            fCurrentOffset++;
            if (++fCurrentIndex == CHUNK_SIZE) {
                fCurrentChunk = new UTF8DataChunk(fStringPool, fCurrentChunk);
                fillCurrentChunk();
                data = fMostRecentData;
            }
            int b1 = data[fCurrentIndex] & 0xFF;
            if ((0xe0 & b0) == 0xc0) { // 110yyyyy 10xxxxxx
                ch = ((0x1f & b0)<<6) + (0x3f & b1);
                if ((fgCharFlags[ch] & E_NameCharFlag) == 0) { // yyy yyxx xxxx (0x80 to 0x7ff)
                    fCurrentChunk = saveChunk;
                    fCurrentIndex = saveIndex;
                    fCurrentOffset = saveOffset;
                    fMostRecentData = saveChunk.toByteArray();
                    break;
                }
                continue;
            }
            fCurrentOffset++;
            if (++fCurrentIndex == CHUNK_SIZE) {
                fCurrentChunk = new UTF8DataChunk(fStringPool, fCurrentChunk);
                fillCurrentChunk();
                data = fMostRecentData;
            }
            int b2 = data[fCurrentIndex] & 0xFF;
            if ((0xf0 & b0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
                // if ((ch >= 0xD800 && ch <= 0xDFFF) || ch >= 0xFFFE)
                if ((b0 == 0xED && b1 >= 0xA0) || (b0 == 0xEF && b1 == 0xBF && b2 >= 0xBE)) {
                    fCurrentChunk = saveChunk;
                    fCurrentIndex = saveIndex;
                    fCurrentOffset = saveOffset;
                    fMostRecentData = saveChunk.toByteArray();
                    break;
                }
                ch = ((0x0f & b0)<<12) + ((0x3f & b1)<<6) + (0x3f & b2);
                if ((fgCharFlags[ch] & E_NameCharFlag) == 0) { // zzzz yyyy yyxx xxxx (0x800 to 0xffff)
                    fCurrentChunk = saveChunk;
                    fCurrentIndex = saveIndex;
                    fCurrentOffset = saveOffset;
                    fMostRecentData = saveChunk.toByteArray();
                    break;
                }
            } else { // 11110uuu 10uuzzzz 10yyyyyy 10xxxxxx
                fCurrentChunk = saveChunk;
                fCurrentIndex = saveIndex;
                fCurrentOffset = saveOffset;
                fMostRecentData = saveChunk.toByteArray();
                break;
            }
        }
        fMostRecentByte = b0;
        hashcode = StringHasher.finishHash(hashcode);
        int length = fCurrentOffset - offset;
        int nameIndex = addSymbol(offset, length, hashcode);
        return (expectedName == -1 || expectedName == nameIndex) ? nameIndex : -1;
    }
    //
    // REVISIT - temporary...
    // Need to create a replacement for fgCharFlags that works by
    // multi-level indexing of the UTF-8 byte stream, i.e.
    //
    // (fgCharFlags[b0][b1][b2] & E_InitialNameCharFlag)
    //
    protected boolean skippedMultiByteCharWithFlag(int b0, int flag) throws Exception {
        UTF8DataChunk saveChunk = fCurrentChunk;
        int saveOffset = fCurrentOffset;
        int saveIndex = fCurrentIndex;
        int b1 = loadNextByte();
        if ((0xe0 & b0) == 0xc0) { // 110yyyyy 10xxxxxx
            if ((fgCharFlags[((0x1f & b0)<<6) + (0x3f & b1)] & flag) == 0) { // yyy yyxx xxxx (0x80 to 0x7ff)
                fCurrentChunk = saveChunk;
                fCurrentIndex = saveIndex;
                fCurrentOffset = saveOffset;
                fMostRecentData = saveChunk.toByteArray();
                fMostRecentByte = b0;
                return false;
            }
            return true;
        }
        int b2 = loadNextByte();
        if ((0xf0 & b0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
            // if ((ch >= 0xD800 && ch <= 0xDFFF) || ch >= 0xFFFE)
            if ((b0 == 0xED && b1 >= 0xA0) || (b0 == 0xEF && b1 == 0xBF && b2 >= 0xBE)) {
                fCurrentChunk = saveChunk;
                fCurrentIndex = saveIndex;
                fCurrentOffset = saveOffset;
                fMostRecentData = saveChunk.toByteArray();
                fMostRecentByte = b0;
                return false;
            }
            if ((fgCharFlags[((0x0f & b0)<<12) + ((0x3f & b1)<<6) + (0x3f & b2)] & flag) == 0) { // zzzz yyyy yyxx xxxx (0x800 to 0xffff)
                fCurrentChunk = saveChunk;
                fCurrentIndex = saveIndex;
                fCurrentOffset = saveOffset;
                fMostRecentData = saveChunk.toByteArray();
                fMostRecentByte = b0;
                return false;
            }
            return true;
        } else { // 11110uuu 10uuzzzz 10yyyyyy 10xxxxxx
            fCurrentChunk = saveChunk;
            fCurrentIndex = saveIndex;
            fCurrentOffset = saveOffset;
            fMostRecentData = saveChunk.toByteArray();
            fMostRecentByte = b0;
            return false;
        }
    }
    public int skipPastName(char fastcheck) throws Exception {
        int b0 = fMostRecentByte;
        if ((b0 & 0x80) == 0) {
            if (fgAsciiInitialNameChar[b0] == 0)
                return fCurrentOffset;
        } else {
            if (!skippedMultiByteCharWithFlag(b0, E_InitialNameCharFlag))
                return fCurrentOffset;
        }
        while (true) {
            fCharacterCounter++;
            b0 = loadNextByte();
            if (fastcheck == b0)
                return fCurrentOffset;
            if ((b0 & 0x80) == 0) {
                if (fgAsciiNameChar[b0] == 0)
                    return fCurrentOffset;
            } else {
                if (!skippedMultiByteCharWithFlag(b0, E_NameCharFlag))
                    return fCurrentOffset;
            }
        }
    }
    public int skipPastNmtoken(char fastcheck) throws Exception {
        int b0 = fMostRecentByte;
        while (true) {
            if (fastcheck == b0)
                return fCurrentOffset;
            if ((b0 & 0x80) == 0) {
                if (fgAsciiNameChar[b0] == 0)
                    return fCurrentOffset;
            } else {
                if (!skippedMultiByteCharWithFlag(b0, E_NameCharFlag))
                    return fCurrentOffset;
            }
            fCharacterCounter++;
            b0 = loadNextByte();
        }
    }
    //
    // [14] CharData ::= [^<&]* - ([^<&]* ']]>' [^<&]*)
    //
    public int scanContent(ScanContentState scanState) throws Exception {
        fCurrentChunk.clearPreviousChunk();
        fCharDataLength = 0;
        int charDataOffset = fCurrentOffset;
        int ch = fMostRecentByte;
        if ((ch & 0x80) == 0) {
            switch (fgAsciiWSCharData[ch]) {
            case 0:
                fCharacterCounter++;
                fCurrentOffset++;
                if (++fCurrentIndex == CHUNK_SIZE)
                    slowLoadNextByte();
                else
                    fMostRecentByte = fMostRecentData[fCurrentIndex] & 0xFF;
                if (fSendCharDataAsCharArray)
                    appendCharData(ch);
                break;
            case 1: // '<'
                fCharacterCounter++;
                fCurrentOffset++;
                if (++fCurrentIndex == CHUNK_SIZE)
                    slowLoadNextByte();
                else
                    fMostRecentByte = fMostRecentData[fCurrentIndex] & 0xFF;
                if (!scanState.inCDSect)
                    return CHARDATA_RESULT_MARKUP;
                if (fSendCharDataAsCharArray)
                    appendCharData('<');
                break;
            case 2: // '&'
                fCharacterCounter++;
                fCurrentOffset++;
                if (++fCurrentIndex == CHUNK_SIZE)
                    slowLoadNextByte();
                else
                    fMostRecentByte = fMostRecentData[fCurrentIndex] & 0xFF;
                if (!scanState.inCDSect)
                    return CHARDATA_RESULT_REFERENCE;
                if (fSendCharDataAsCharArray)
                    appendCharData('&');
                break;
            case 3: // ']'
                fCharacterCounter++;
                ch = loadNextByte();
                if (ch != ']') {
                    if (fSendCharDataAsCharArray)
                        appendCharData(']');
                    break;
		}
                if (fCurrentIndex + 1 == CHUNK_SIZE) {
                    UTF8DataChunk saveChunk = fCurrentChunk;
                    int saveIndex = fCurrentIndex;
                    int saveOffset = fCurrentOffset;
                    if (loadNextByte() != '>') {
                        fCurrentChunk = saveChunk;
                        fCurrentIndex = saveIndex;
                        fCurrentOffset = saveOffset;
                        fMostRecentData = saveChunk.toByteArray();
                        fMostRecentByte = ']';
                        if (fSendCharDataAsCharArray)
                            appendCharData(']');
                        break;
                    }
                } else {
                    if (fMostRecentData[fCurrentIndex + 1] != '>') {
                        if (fSendCharDataAsCharArray)
                            appendCharData(']');
                        break;
		    }
                    fCurrentIndex++;
                    fCurrentOffset++;
                }
                loadNextByte();
                fCharacterCounter += 2;
                if (scanState.inCDSect) {
                    scanState.inCDSect = false;
                    return scanContent(scanState);
                }
                return CHARDATA_RESULT_CDEND;
            case 4: // invalid char
                return CHARDATA_RESULT_INVALID_CHAR;
            case 5:
                do {
                    if (fSendCharDataAsCharArray)
                        appendCharData(ch);
                    if (ch == 0x0A) {
                        fLinefeedCounter++;
                        fCharacterCounter = 1;
                    } else
                        fCharacterCounter++;
                    fCurrentOffset++;
                    if (++fCurrentIndex == CHUNK_SIZE)
                        ch = slowLoadNextByte();
                    else
                        ch = fMostRecentByte = fMostRecentData[fCurrentIndex] & 0xFF;
                } while (ch == 0x20 || ch == 0x09 || ch == 0x0A);
                if ((ch & 0x80) == 0) {
                    switch (fgAsciiCharData[ch]) {
                    case 0:
                        fCharacterCounter++;
                        fCurrentOffset++;
                        if (++fCurrentIndex == CHUNK_SIZE)
                            slowLoadNextByte();
                        else
                            fMostRecentByte = fMostRecentData[fCurrentIndex] & 0xFF;
                        if (fSendCharDataAsCharArray)
                            appendCharData(ch);
                        break;
                    case 1: // '<'
                        if (!scanState.inCDSect) {
                            if (fDocumentHandler != null)
                                callWSCharDataHandler(charDataOffset, fCurrentOffset, false);
                            fCharacterCounter++;
                            loadNextByte();
                            return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_MARKUP);
                        }
                        fCharacterCounter++;
                        fCurrentOffset++;
                        if (++fCurrentIndex == CHUNK_SIZE)
                            slowLoadNextByte();
                        else
                            fMostRecentByte = fMostRecentData[fCurrentIndex] & 0xFF;
                        if (fSendCharDataAsCharArray)
                            appendCharData(ch);
                        break;
                    case 2: // '&'
                        if (!scanState.inCDSect) {
                            if (fDocumentHandler != null)
                                callWSCharDataHandler(charDataOffset, fCurrentOffset, false);
                            fCharacterCounter++;
                            loadNextByte();
                            return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_REFERENCE);
                        }
                        fCharacterCounter++;
                        fCurrentOffset++;
                        if (++fCurrentIndex == CHUNK_SIZE)
                            slowLoadNextByte();
                        else
                            fMostRecentByte = fMostRecentData[fCurrentIndex] & 0xFF;
                        if (fSendCharDataAsCharArray)
                            appendCharData(ch);
                        break;
                    case 3: // ']'
                        int endOffset = fCurrentOffset;
                        ch = loadNextByte();
                        if (ch != ']') {
                            fCharacterCounter++;
                            if (fSendCharDataAsCharArray)
                                appendCharData(']');
                            break;
                        }
                        if (fCurrentIndex + 1 == CHUNK_SIZE) {
                            UTF8DataChunk saveChunk = fCurrentChunk;
                            int saveIndex = fCurrentIndex;
                            int saveOffset = fCurrentOffset;
                            if (loadNextByte() != '>') {
                                fCurrentChunk = saveChunk;
                                fCurrentIndex = saveIndex;
                                fCurrentOffset = saveOffset;
                                fMostRecentData = saveChunk.toByteArray();
                                fMostRecentByte = ']';
                                fCharacterCounter++;
                                if (fSendCharDataAsCharArray)
                                    appendCharData(']');
                                break;
                            }
                        } else {
                            if (fMostRecentData[fCurrentIndex + 1] != '>') {
                                fCharacterCounter++;
                                if (fSendCharDataAsCharArray)
                                    appendCharData(']');
                                break;
                            }
                            fCurrentIndex++;
                            fCurrentOffset++;
                        }
                        loadNextByte();
                        if (fDocumentHandler != null)
                            callWSCharDataHandler(charDataOffset, endOffset, scanState.inCDSect);
                        fCharacterCounter += 3;
                        if (scanState.inCDSect) {
                            scanState.inCDSect = false;
                            return scanContent(scanState);
                        }
                        return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_CDEND);
                    case 4: // invalid char
                        if (fDocumentHandler != null)
                            callWSCharDataHandler(charDataOffset, fCurrentOffset, scanState.inCDSect);
                        return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                    }
                } else {
                    if (fSendCharDataAsCharArray) {
                        if (!copyMultiByteCharData(ch)) {
                            callWSCharDataHandler(charDataOffset, fCurrentOffset, scanState.inCDSect);
                            return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                        }
                    } else if (!skipMultiByteCharData(ch)) {
                        if (fDocumentHandler != null)
                            callWSCharDataHandler(charDataOffset, fCurrentOffset, scanState.inCDSect);
                        return (CHARDATA_RESULT_ALL_SPACE | CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                    }
                }
                break;
            }
        } else if (!skipMultiByteCharData(ch))
            return CHARDATA_RESULT_INVALID_CHAR;
        if (fSendCharDataAsCharArray)
            ch = copyAsciiCharData();
        else
            ch = skipAsciiCharData();
        while (true) { // REVISIT - EOF check ?
            if ((ch & 0x80) == 0) {
                switch (fgAsciiCharData[ch]) {
                case 0:
                    fCharacterCounter++;
                    if (fSendCharDataAsCharArray)
                        appendCharData(ch);
                    ch = loadNextByte();
                    break;
                case 1: // '<'
                    if (!scanState.inCDSect) {
                        if (fDocumentHandler != null)
                            callCharDataHandler(charDataOffset, fCurrentOffset, false);
                        fCharacterCounter++;
                        fCurrentOffset++;
                        if (++fCurrentIndex == CHUNK_SIZE)
                            slowLoadNextByte();
                        else
                            fMostRecentByte = fMostRecentData[fCurrentIndex] & 0xFF;
                        return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_MARKUP);
                    }
                    fCharacterCounter++;
                    if (fSendCharDataAsCharArray)
                        appendCharData(ch);
                    ch = loadNextByte();
                    break;
                case 2: // '&'
                    if (!scanState.inCDSect) {
                        if (fDocumentHandler != null)
                            callCharDataHandler(charDataOffset, fCurrentOffset, false);
                        fCharacterCounter++;
                        fCurrentOffset++;
                        if (++fCurrentIndex == CHUNK_SIZE)
                            slowLoadNextByte();
                        else
                            fMostRecentByte = fMostRecentData[fCurrentIndex] & 0xFF;
                        return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_REFERENCE);
                    }
                    fCharacterCounter++;
                    if (fSendCharDataAsCharArray)
                        appendCharData(ch);
                    ch = loadNextByte();
                    break;
                case 3: // ']'
                    int endOffset = fCurrentOffset;
                    ch = loadNextByte();
                    if (ch != ']') {
                        fCharacterCounter++;
                        if (fSendCharDataAsCharArray)
                            appendCharData(']');
                        break;
                    }
                    if (fCurrentIndex + 1 == CHUNK_SIZE) {
                        UTF8DataChunk saveChunk = fCurrentChunk;
                        int saveIndex = fCurrentIndex;
                        int saveOffset = fCurrentOffset;
                        if (loadNextByte() != '>') {
                            fCurrentChunk = saveChunk;
                            fCurrentIndex = saveIndex;
                            fCurrentOffset = saveOffset;
                            fMostRecentData = saveChunk.toByteArray();
                            fMostRecentByte = ']';
                            fCharacterCounter++;
                            if (fSendCharDataAsCharArray)
                                appendCharData(']');
                            break;
                        }
                    } else {
                        if (fMostRecentData[fCurrentIndex + 1] != '>') {
                            fCharacterCounter++;
                            if (fSendCharDataAsCharArray)
                                appendCharData(']');
                            break;
                        }
                        fCurrentIndex++;
                        fCurrentOffset++;
                    }
                    loadNextByte();
                    if (fDocumentHandler != null)
                        callCharDataHandler(charDataOffset, endOffset, scanState.inCDSect);
                    fCharacterCounter += 3;
                    if (scanState.inCDSect) {
                        scanState.inCDSect = false;
                        return scanContent(scanState);
                    }
                    return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_CDEND);
                case 4: // invalid char
                    if (ch == 0x0A) {
                        fLinefeedCounter++;
                        fCharacterCounter = 1;
                        ch = loadNextByte();
                        break;
                    }
                    if (fDocumentHandler != null)
                        callCharDataHandler(charDataOffset, fCurrentOffset, scanState.inCDSect);
                    return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                }
            } else {
                if (fSendCharDataAsCharArray) {
                    if (!copyMultiByteCharData(ch)) {
                        callCharDataHandler(charDataOffset, fCurrentOffset, scanState.inCDSect);
                        return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                    }
                } else if (!skipMultiByteCharData(ch)) {
                    if (fDocumentHandler != null)
                        callCharDataHandler(charDataOffset, fCurrentOffset, scanState.inCDSect);
                    return (CHARDATA_RESULT_CHARDATA | CHARDATA_RESULT_INVALID_CHAR);
                }
                ch = fMostRecentByte;
            }
        }
    }
    protected boolean copyMultiByteCharData(int b0) throws Exception {
        UTF8DataChunk saveChunk = fCurrentChunk;
        int saveOffset = fCurrentOffset;
        int saveIndex = fCurrentIndex;
        int b1 = loadNextByte();
        if ((0xe0 & b0) == 0xc0) { // 110yyyyy 10xxxxxx (0x80 to 0x7ff)
            int ch = ((0x1f & b0)<<6) + (0x3f & b1);
            appendCharData(ch); // yyy yyxx xxxx (0x80 to 0x7ff)
            return true;
        }
        int b2 = loadNextByte();
        if ((0xf0 & b0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
            // ch = ((0x0f & b0)<<12) + ((0x3f & b1)<<6) + (0x3f & b2); // zzzz yyyy yyxx xxxx (0x800 to 0xffff)
            // if ((ch >= 0xD800 && ch <= 0xDFFF) || ch >= 0xFFFE)
            if ((b0 == 0xED && b1 >= 0xA0) || (b0 == 0xEF && b1 == 0xBF && b2 >= 0xBE)) {
                fCurrentChunk = saveChunk;
                fCurrentIndex = saveIndex;
                fCurrentOffset = saveOffset;
                fMostRecentData = saveChunk.toByteArray();
                fMostRecentByte = b0;
                return false;
            }
            int ch = ((0x0f & b0)<<12) + ((0x3f & b1)<<6) + (0x3f & b2);
            appendCharData(ch); // zzzz yyyy yyxx xxxx (0x800 to 0xffff)
            return true;
        }
        int b3 = loadNextByte();  // 11110uuu 10uuzzzz 10yyyyyy 10xxxxxx
        // ch = ((0x0f & b0)<<18) + ((0x3f & b1)<<12) + ((0x3f & b2)<<6) + (0x3f & b3); // u uuuu zzzz yyyy yyxx xxxx (0x10000 to 0x1ffff)
        // if (ch >= 0x110000)
        if (b0 > 0xF4 || (b0 == 0xF4 && b1 >= 0x90)) {
            fCurrentChunk = saveChunk;
            fCurrentIndex = saveIndex;
            fCurrentOffset = saveOffset;
            fMostRecentData = saveChunk.toByteArray();
            fMostRecentByte = b0;
            return false;
        }
        int ch = ((0x0f & b0)<<18) + ((0x3f & b1)<<12) + ((0x3f & b2)<<6) + (0x3f & b3);
        if (ch < 0x10000)
            appendCharData(ch);
        else {
            appendCharData(((ch-0x00010000)>>10)+0xd800);
            appendCharData(((ch-0x00010000)&0x3ff)+0xdc00);
        }
        return true;
    }
    protected boolean skipMultiByteCharData(int b0) throws Exception {
        UTF8DataChunk saveChunk = fCurrentChunk;
        int saveOffset = fCurrentOffset;
        int saveIndex = fCurrentIndex;
        int b1 = loadNextByte();
        if ((0xe0 & b0) == 0xc0) // 110yyyyy 10xxxxxx (0x80 to 0x7ff)
            return true;
        int b2 = loadNextByte();
        if ((0xf0 & b0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
            // ch = ((0x0f & b0)<<12) + ((0x3f & b1)<<6) + (0x3f & b2); // zzzz yyyy yyxx xxxx (0x800 to 0xffff)
            // if ((ch >= 0xD800 && ch <= 0xDFFF) || ch >= 0xFFFE)
            if ((b0 == 0xED && b1 >= 0xA0) || (b0 == 0xEF && b1 == 0xBF && b2 >= 0xBE)) {
                fCurrentChunk = saveChunk;
                fCurrentIndex = saveIndex;
                fCurrentOffset = saveOffset;
                fMostRecentData = saveChunk.toByteArray();
                fMostRecentByte = b0;
                return false;
            }
            return true;
        }
        int b3 = loadNextByte();  // 11110uuu 10uuzzzz 10yyyyyy 10xxxxxx
        // ch = ((0x0f & b0)<<18) + ((0x3f & b1)<<12) + ((0x3f & b2)<<6) + (0x3f & b3); // u uuuu zzzz yyyy yyxx xxxx (0x10000 to 0x1ffff)
        // if (ch >= 0x110000)
        if (b0 > 0xF4 || (b0 == 0xF4 && b1 >= 0x90)) {
            fCurrentChunk = saveChunk;
            fCurrentIndex = saveIndex;
            fCurrentOffset = saveOffset;
            fMostRecentData = saveChunk.toByteArray();
            fMostRecentByte = b0;
            return false;
        }
        return true;
    }
    protected int copyAsciiCharData() throws Exception {
        int coffset = fCharDataLength;
        int index = fCurrentIndex;
        int offset = fCurrentOffset - index;
        while (true) {
            byte[] data = fMostRecentData;
            while (index < CHUNK_SIZE) {
                int endOffset = CHUNK_SIZE;
                int outputCapacity = (fCharacters.length - 2) - coffset; // the 2 is "being conservative" (or lazy :)
                int inputAvailable = (CHUNK_SIZE - index);
                if (outputCapacity < inputAvailable) {
                    if (outputCapacity < 64) {
                        char[] newchars = new char[fCharacters.length * 2];
                        System.arraycopy(fCharacters, 0, newchars, 0, fCharacters.length);
                        fCharacters = newchars;
                        outputCapacity = (fCharacters.length - 2) - coffset;
                        if (outputCapacity < inputAvailable)
                            endOffset = index + outputCapacity;
                    } else
                        endOffset = index + outputCapacity;
                }
                while (index < endOffset) {
                    int ch = data[index] & 0xff;
                    if ((ch & 0x80) != 0) {
                        fCurrentOffset = offset + index;
                        fCurrentIndex = index;
                        fMostRecentByte = ch;
                        fCharDataLength = coffset;
                        return ch;
                    }
                    if (fgAsciiCharData[ch] != 0) {
                        if (ch != 0x0A) {
                            fCurrentOffset = offset + index;
                            fCurrentIndex = index;
                            fMostRecentByte = ch;
                            fCharDataLength = coffset;
                            return ch;
                        }
                        fLinefeedCounter++;
                        fCharacterCounter = 1;
                    } else
                        fCharacterCounter++;
                    fCharacters[coffset++] = (char)ch;
                    index++;
                }
            }
            offset += index;
            fCurrentChunk = new UTF8DataChunk(fStringPool, fCurrentChunk);
            fillCurrentChunk();
            index = 0;
        }
    }
    protected int skipAsciiCharData() throws Exception {
        int index = fCurrentIndex;
        int offset = fCurrentOffset - index;
        while (true) {
            byte[] data = fMostRecentData;
            while (index < CHUNK_SIZE) {
                int ch = data[index] & 0xFF;
                if ((ch & 0x80) != 0) {
                    fCurrentOffset = offset + index;
                    fCurrentIndex = index;
                    fMostRecentByte = ch;
                    return ch;
                }
                if (fgAsciiCharData[ch] != 0) {
                    if (ch != 0x0A) {
                        fCurrentOffset = offset + index;
                        fCurrentIndex = index;
                        fMostRecentByte = ch;
                        return ch;
                    }
                    fLinefeedCounter++;
                    fCharacterCounter = 1;
                } else
                    fCharacterCounter++;
                index++;
            }
            offset += index;
            fCurrentChunk = new UTF8DataChunk(fStringPool, fCurrentChunk);
            fillCurrentChunk();
            index = 0;
        }
    }
    protected char[] fCharacters = new char[256];
    protected int fCharDataLength = 0;
    protected void appendCharData(int ch) {
        if (fCharacters.length == fCharDataLength) {
            char[] newchars = new char[fCharacters.length * 2];
            System.arraycopy(fCharacters, 0, newchars, 0, fCharacters.length);
            fCharacters = newchars;
        }
        fCharacters[fCharDataLength++] = (char)ch;
    }
    public void callCharDataHandler(int offset, int endOffset, boolean inCDSect) throws Exception {
        //
        // REVISIT - need more up front bounds checking code of params...
        //
        if (!fSendCharDataAsCharArray) {
            int stringIndex = addString(offset, endOffset - offset);
            fDocumentHandler.characters(stringIndex, inCDSect);
            return;
        }
        fDocumentHandler.characters(fCharacters, 0, fCharDataLength, inCDSect);
    }
    public void callWSCharDataHandler(int offset, int endOffset, boolean inCDSect) throws Exception {
        int contentSpecType = fParserState.getScanner().getCurrentContentSpecType();
        if (contentSpecType != ElementDecl.CONTENTSPEC_CHILDREN) {
            callCharDataHandler(offset, endOffset, inCDSect);
            return;
        }
        if (!fSendCharDataAsCharArray) {
            int stringIndex = addString(offset, endOffset - offset);
            fDocumentHandler.ignorableWhitespace(stringIndex, inCDSect);
            return;
        }
        fDocumentHandler.ignorableWhitespace(fCharacters, 0, fCharDataLength, inCDSect);
    }
    //
    //
    //
    protected XMLScanner fScanner = null;
    protected StringPool fStringPool = null;
    protected XMLDocumentHandler fDocumentHandler = null;
    protected boolean fSendCharDataAsCharArray = false;
    protected XMLErrorHandler fErrorHandler = null;
    protected InputStream fInputStream = null;
    protected UTF8DataChunk fCurrentChunk = null;
    protected int fCurrentIndex = 0;
    protected byte[] fMostRecentData = null;
    protected int fMostRecentByte = 0;
    protected int fLength = 0;
    //
    // When we fill a chunk there may be data that was read from the
    // input stream that has not been "processed".  We need to save
    // that data, and any in-progress state, between the calls to
    // UTF8DataChunk.fill() in these instance variables.  
    //
    protected boolean fCheckOverflow = false;
    protected byte[] fOverflow = null;
    protected int fOverflowOffset = 0;
    protected int fOverflowEnd = 0;
    protected boolean fSkipLinefeed = false;
    protected int fPartialMultiByteIn = 0;
    protected byte[] fPartialMultiByteChar = new byte[4];
    protected boolean fPartialMultiByteResult = false;
    protected int fFastCopyCount = 0;
    protected int fOutputOffset = 0;
    protected int fPartialMultiByteOut = 0;
    //
    //
    //
    protected int fillCurrentChunk() throws Exception {
        //
        // If we have overflow from the last call, normalize from where
        // we left off, copying into the front of the output buffer.
        //
        fOutputOffset = 0;
        if (fCheckOverflow) {
            //
            // The fOverflowEnd should always be equal to CHUNK_SIZE, unless we hit
            // EOF during the previous call.  Copy the remaining data to the front
            // of the buffer and return it as the final chunk.
            //
            if (fOverflowEnd < CHUNK_SIZE) {
                if (fOverflowEnd > 0) {
                    fMostRecentData = new byte[1 + fOverflowEnd - fOverflowOffset];
                    copyNormalize(fOverflow, fOverflowOffset, fMostRecentData, fOutputOffset);
                } else {
                    fMostRecentData = new byte[1];
                }
                fMostRecentData[fOutputOffset] = 0;
                //
                // Update the outer class instance variables
                //
                fOverflow = null;
                fLength += fOutputOffset;
                fCurrentIndex = 0;
                fCurrentChunk.setByteArray(fMostRecentData);
                return (fMostRecentByte = fMostRecentData[0]);
            }
            fMostRecentData = new byte[CHUNK_SIZE];
            copyNormalize(fOverflow, fOverflowOffset, fMostRecentData, fOutputOffset);
            fCheckOverflow = false;
        } else {
            if (fOverflow == null)
                fOverflow = new byte[CHUNK_SIZE];
            fMostRecentData = null;
        }
        while (true) {
            fOverflowOffset = 0;
            fOverflowEnd = 0;
            int capacity = CHUNK_SIZE;
            int result = 0;
            do {
                result = fInputStream.read(fOverflow, fOverflowEnd, capacity);
                if (result == -1) {
                    //
                    // We have reached the end of the stream.
                    //
                    fInputStream.close();
                    fInputStream = null;
                    if (fMostRecentData == null) {
                        //
                        // There is no previous output data, so we know that all of the
                        // new input data will fit.
                        //
                        fMostRecentData = new byte[1 + fOverflowEnd];
                        copyNormalize(fOverflow, fOverflowOffset, fMostRecentData, fOutputOffset);
                        fOverflow = null;
                        fMostRecentData[fOutputOffset] = 0;
                    } else {
                        //
                        // Copy the input data to the end of the output buffer.
                        //
                        boolean alldone = copyNormalize(fOverflow, fOverflowOffset, fMostRecentData, fOutputOffset);
                        if (alldone) {
                            if (fOverflowEnd == CHUNK_SIZE) {
                                //
                                // Special case - everything fit into the overflow buffer,
                                // except that there is no room for the nul char we use to
                                // indicate EOF.  Set the overflow buffer length to zero.
                                // On the next call to this method, we will detect this
                                // case and which we will handle above .
                                //
                                fCheckOverflow = true;
                                fOverflowOffset = 0;
                                fOverflowEnd = 0;
                            } else {
                                //
                                // It all fit into the output buffer.
                                //
                                fOverflow = null;
                                fMostRecentData[fOutputOffset] = 0;
                            }
                        } else {
                            //
                            // There is still input data left over, save the remaining data as
                            // the overflow buffer for the next call.
                            //
                            fCheckOverflow = true;
                        }
                    }
                    break;
                }
                if (result > 0) {
                    fOverflowEnd += result;
                    capacity -= result;
                }
            } while (capacity > 0);
            //
            //
            //
            if (result == -1)
                break;
            if (fMostRecentData != null) {
                boolean alldone = copyNormalize(fOverflow, fOverflowOffset, fMostRecentData, fOutputOffset);
                if (fOutputOffset == CHUNK_SIZE) {
                    //
                    // We filled the output buffer.
                    //
                    if (!alldone) {
                        //
                        // The input buffer will become the next overflow buffer.
                        //
                        fCheckOverflow = true;
                    }
                    break;
                }
            } else {
                //
                // Now normalize the end-of-line characters and see if we need to read more
                // bytes to fill up the buffer.
                //
                fMostRecentData = new byte[CHUNK_SIZE];
                copyNormalize(fOverflow, fOverflowOffset, fMostRecentData, fOutputOffset);
                if (fOutputOffset == CHUNK_SIZE) {
                    //
                    // The output buffer is full.  We can return now.
                    //
                    break;
                }
            }
            //
            // We will need to get another intput buffer to be able to fill the
            // overflow buffer completely.
            //
        }
        //
        // Update the outer class instance variables
        //
        fLength += fOutputOffset;
        fCurrentIndex = 0;
        fCurrentChunk.setByteArray(fMostRecentData);
        return (fMostRecentByte = fMostRecentData[0]);
    }
    //
    // Copy and normalize bytes from the overflow buffer into our data buffer.
    //
    protected boolean copyNormalize(byte[] in, int inOffset, byte[] out, int outOffset) throws Exception {
        //
        // Handle all edge cases before dropping into the inner loop.
        //
        int inEnd = fOverflowEnd;
        int outEnd = out.length;
        if (inOffset == inEnd)
            return true;
        byte b = in[inOffset];
        if (fSkipLinefeed) {
            fSkipLinefeed = false;
            if (b == 0x0A) {
                if (++inOffset == inEnd)
                    return exitNormalize(inOffset, outOffset, true);
                b = in[inOffset];
            }
        } else if (fPartialMultiByteIn > 0) {
            if (!handlePartialMultiByteChar(b, in, inOffset, inEnd, out, outOffset, outEnd))
                return fPartialMultiByteResult;
            inOffset = fOverflowOffset;
            outOffset = fOutputOffset;
            b = in[inOffset];
        }
        while (outOffset < outEnd) {
            //
            // Find the longest run that we can guarantee will not exceed the
            // bounds of the outer loop.
            //
            int inCount = inEnd - inOffset;
            int outCount = outEnd - outOffset;
            if (inCount > outCount)
                inCount = outCount;
            inOffset++;
            while (true) {
                while (b == 0x0D || (b & 0x80) != 0) {
                    if (b == 0x0D) {
                        out[outOffset++] = 0x0A;
                        if (inOffset == inEnd) {
                            fSkipLinefeed = true;
                            return exitNormalize(inOffset, outOffset, true);
                        }
                        b = in[inOffset];
                        if (b == 0x0A) {
                            if (++inOffset == inEnd)
                                return exitNormalize(inOffset, outOffset, true);
                            b = in[inOffset];
                        }
                        if (outOffset == outEnd)
                            return exitNormalize(inOffset, outOffset, false);
                    } else {
                        if (!handleMultiByteChar(b, in, inOffset, inEnd, out, outOffset, outEnd))
                            return fPartialMultiByteResult;
                        inOffset = fOverflowOffset;
                        outOffset = fOutputOffset;
                        b = in[inOffset];
                    }
                    inCount = inEnd - inOffset;
                    outCount = outEnd - outOffset;
                    if (inCount > outCount)
                        inCount = outCount;
                    inOffset++;
                }
                while (true) {
                    out[outOffset++] = b;
                    if (--inCount == 0)
                        break;
                    b = in[inOffset++];
                    if (b == 0x0D || (b & 0x80) != 0)
                        break;
                }
                if (inCount == 0)
                    break;
/*
                b = fastCopyNormalize(inCount, b, in, inOffset);
                if (fFastCopyCount == 0) {
                    System.arraycopy(in, copyOffset, out, outOffset, inCount);
                    inOffset = fInputOffset;
                    outOffset += inCount;
                    break;
                }
                inOffset = fInputOffset;
                int count = (inOffset - copyOffset) - 1;
                System.arraycopy(in, copyOffset, out, outOffset, count);
                outOffset += count;
 */
            }
            if (inOffset == inEnd)
                break;
        }
        return exitNormalize(inOffset, outOffset, inOffset == inEnd);
    }
    //
    //
    //
/*
    private byte fastCopyNormalize(int inCount, byte b, byte[] in, int inOffset) {
        while (--inCount > 0) {
            b = in[inOffset++];
            if (b == 0x0D || b < 0)
                break;
        }
        fFastCopyCount = inCount;
        fInputOffset = inOffset;
        return b;
    }
 */
    //
    //
    //
    protected boolean exitNormalize(int inOffset, int outOffset, boolean result) {
        fOverflowOffset = inOffset;
        fOutputOffset = outOffset;
        return result;
    }
    //
    //
    //
    protected void savePartialMultiByte(int inCount, int outCount, byte bz, byte by, byte bx, byte bw) {
        fPartialMultiByteIn = inCount;
        fPartialMultiByteOut = outCount;
        fPartialMultiByteChar[--inCount] = bz;
        fPartialMultiByteChar[--inCount] = by;
        fPartialMultiByteChar[--inCount] = bx;
        fPartialMultiByteChar[--inCount] = bw;
    }
    protected void savePartialMultiByte(int inCount, int outCount, byte bz, byte by, byte bx) {
        fPartialMultiByteIn = inCount;
        fPartialMultiByteOut = outCount;
        fPartialMultiByteChar[--inCount] = bz;
        fPartialMultiByteChar[--inCount] = by;
        fPartialMultiByteChar[--inCount] = bx;
    }
    protected void savePartialMultiByte(int inCount, int outCount, byte bz, byte by) {
        fPartialMultiByteIn = inCount;
        fPartialMultiByteOut = outCount;
        fPartialMultiByteChar[--inCount] = bz;
        fPartialMultiByteChar[--inCount] = by;
    }
    protected void savePartialMultiByte(int inCount, int outCount, byte bz) {
        fPartialMultiByteIn = inCount;
        fPartialMultiByteOut = outCount;
        fPartialMultiByteChar[--inCount] = bz;
    }
    protected boolean handleMultiByteChar(byte b, byte[] in, int inOffset, int inEnd, byte[] out, int outOffset, int outEnd) throws Exception {
        if (inOffset == inEnd) {
            savePartialMultiByte(1, 0, b);
            fPartialMultiByteResult = exitNormalize(inOffset, outOffset, true);
            return false;
        }
        byte b1 = in[inOffset++];
        if ((b1 & 0xc0) != 0x80)
            fErrorHandler.error2(ErrorCode.E_ENC5,
                fStringPool.addString(Integer.toHexString(b)),
                fStringPool.addString(Integer.toHexString(b1)));
        if ((b & 0xe0) == 0xc0) { // 110yyyyy 10xxxxxx
            out[outOffset++] = b;
            if (outOffset == outEnd) {
                savePartialMultiByte(2, 1, b1, b);
                fPartialMultiByteResult = exitNormalize(inOffset, outOffset, inOffset == inEnd);
                return false;
            }
            out[outOffset++] = b1;
            if (inOffset == inEnd || outOffset == outEnd) {
                fPartialMultiByteResult = exitNormalize(inOffset, outOffset, inOffset == inEnd);
                return false;
            }
        } else {
            if (inOffset == inEnd) {
                savePartialMultiByte(2, 0, b1, b);
                fPartialMultiByteResult = exitNormalize(inOffset, outOffset, true);
                return false;
            }
            byte b2 = in[inOffset++];
            if ((b2 & 0xc0) != 0x80)
                fErrorHandler.error3(ErrorCode.E_ENC6,
                    fStringPool.addString(Integer.toHexString(b)),
                    fStringPool.addString(Integer.toHexString(b1)),
                    fStringPool.addString(Integer.toHexString(b2)));
            if ((b & 0xf0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
                out[outOffset++] = b;
                if (outOffset == outEnd) {
                    savePartialMultiByte(3, 1, b2, b1, b);
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, inOffset == inEnd);
                    return false;
                }
                out[outOffset++] = b1;
                if (outOffset == outEnd) {
                    savePartialMultiByte(3, 2, b2, b1, b);
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, inOffset == inEnd);
                    return false;
                }
                out[outOffset++] = b2;
                if (inOffset == inEnd || outOffset == outEnd) {
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, inOffset == inEnd);
                    return false;
                }
            } else {
                if ((b & 0xf8) != 0xf0)
                    fErrorHandler.error1(ErrorCode.E_ENC4,
                        fStringPool.addString(Integer.toHexString(b)));
                if (inOffset == inEnd) {
                    savePartialMultiByte(3, 0, b2, b1, b);
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, true);
                    return false;
                }
                byte b3 = in[inOffset++];
                if ((b3 & 0xc0) != 0x80)
                    fErrorHandler.error4(ErrorCode.E_ENC7,
                        fStringPool.addString(Integer.toHexString(b)),
                        fStringPool.addString(Integer.toHexString(b1)),
                        fStringPool.addString(Integer.toHexString(b2)),
                        fStringPool.addString(Integer.toHexString(b3)));
                out[outOffset++] = b;
                if (outOffset == outEnd) {
                    savePartialMultiByte(4, 1, b3, b2, b1, b);
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, inOffset == inEnd);
                    return false;
                }
                out[outOffset++] = b1;
                if (outOffset == outEnd) {
                    savePartialMultiByte(4, 2, b3, b2, b1, b);
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, inOffset == inEnd);
                    return false;
                }
                out[outOffset++] = b2;
                if (outOffset == outEnd) {
                    savePartialMultiByte(4, 3, b3, b2, b1, b);
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, inOffset == inEnd);
                    return false;
                }
                out[outOffset++] = b3;
                if (inOffset == inEnd || outOffset == outEnd) {
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, inOffset == inEnd);
                    return false;
                }
            }
        }
        return exitNormalize(inOffset, outOffset, true);
    }
    protected boolean handlePartialMultiByteChar(byte b, byte[] in, int inOffset, int inEnd, byte[] out, int outOffset, int outEnd) throws Exception {
        if (outOffset == outEnd) {
            fPartialMultiByteResult = exitNormalize(inOffset, outOffset, inOffset == inEnd);
            return false;
        }
        int byteIn = fPartialMultiByteIn;
        int byteOut = fPartialMultiByteOut;
        fPartialMultiByteIn = 0;
        fPartialMultiByteOut = 0;
        byte b1 = 0;
        byte b2 = 0;
        byte b3 = 0;
        byte b4 = 0;
        switch (byteIn) {
        case 1: b1 = b; break;
        case 2: b2 = b; break;
        case 3: b3 = b; break;
        case 4: b4 = b; break;
        }
        int i = byteIn;
        switch (byteIn) {
        case 4:
            b3 = fPartialMultiByteChar[--i];
        case 3:
            b2 = fPartialMultiByteChar[--i];
        case 2:
            b1 = fPartialMultiByteChar[--i];
        case 1:
            b = fPartialMultiByteChar[--i];
        }
        switch (byteIn) {
        case 1:
            if ((b1 & 0xc0) != 0x80)
                fErrorHandler.error2(ErrorCode.E_ENC5,
                    fStringPool.addString(Integer.toHexString(b)),
                    fStringPool.addString(Integer.toHexString(b1)));
            // fall through
        case 2:
            if ((b & 0xe0) == 0xc0) { // 110yyyyy 10xxxxxx
                switch (byteOut) {
                case 0:
                    out[outOffset++] = b;
                    if (outOffset == outEnd) {
                        savePartialMultiByte(2, 1, b1);
                        fPartialMultiByteResult = exitNormalize(inOffset, outOffset, false);
                        return false;
                    }
                    // fall through
                case 1:
                    out[outOffset++] = b1;
                    if (outOffset == outEnd) {
                        fPartialMultiByteResult = exitNormalize(inOffset, outOffset, false);
                        return false;
                    }
                }
                if (byteIn < 2 && ++inOffset == inEnd) {
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, true);
                    return false;
                }
                break;
            }
            if (byteIn < 2) {
                if (++inOffset == inEnd) {
                    savePartialMultiByte(2, 0, b1);
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, true);
                    return false;
                }
                b2 = in[inOffset];
            }
            if ((b2 & 0xc0) != 0x80)
                fErrorHandler.error3(ErrorCode.E_ENC6,
                    fStringPool.addString(Integer.toHexString(b)),
                    fStringPool.addString(Integer.toHexString(b1)),
                    fStringPool.addString(Integer.toHexString(b2)));
            // fall through
        case 3:
            if ((b & 0xf0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
                switch (byteOut) {
                case 0:
                    out[outOffset++] = b;
                    if (outOffset == outEnd) {
                        savePartialMultiByte(3, 1, b2, b1);
                        fPartialMultiByteResult = exitNormalize(inOffset, outOffset, false);
                        return false;
                    }
                    // fall through
                case 1:
                    out[outOffset++] = b1;
                    if (outOffset == outEnd) {
                        savePartialMultiByte(3, 2, b2, b1);
                        fPartialMultiByteResult = exitNormalize(inOffset, outOffset, false);
                        return false;
                    }
                    // fall through
                case 2:
                    out[outOffset++] = b2;
                    if (outOffset == outEnd) {
                        fPartialMultiByteResult = exitNormalize(inOffset, outOffset, false);
                        return false;
                    }
                    break;
                }
                if (byteIn < 3 && ++inOffset == inEnd) {
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, true);
                    return false;
                }
                break;
            }
            if (byteIn < 3) {
                if ((b & 0xf8) != 0xf0)
                    fErrorHandler.error1(ErrorCode.E_ENC4,
                        fStringPool.addString(Integer.toHexString(b)));
                if (++inOffset == inEnd) {
                    savePartialMultiByte(3, 0, b2, b1);
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, true);
                    return false;
                }
                b3 = in[inOffset];
            }
            if ((b3 & 0xc0) != 0x80)
                fErrorHandler.error4(ErrorCode.E_ENC7,
                    fStringPool.addString(Integer.toHexString(b)),
                    fStringPool.addString(Integer.toHexString(b1)),
                    fStringPool.addString(Integer.toHexString(b2)),
                    fStringPool.addString(Integer.toHexString(b3)));
            // fall through
        case 4:
            switch (byteOut) {
            case 0:
                out[outOffset++] = b;
                if (outOffset == outEnd) {
                    savePartialMultiByte(4, 1, b3, b2, b1);
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, false);
                    return false;
                }
                // fall through
            case 1:
                out[outOffset++] = b1;
                if (outOffset == outEnd) {
                    savePartialMultiByte(4, 2, b3, b2, b1);
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, false);
                    return false;
                }
                // fall through
            case 2:
                out[outOffset++] = b2;
                if (outOffset == outEnd) {
                    savePartialMultiByte(4, 3, b3, b2, b1);
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, false);
                    return false;
                }
                // fall through
            case 3:
                out[outOffset++] = b3;
                if (outOffset == outEnd) {
                    fPartialMultiByteResult = exitNormalize(inOffset, outOffset, false);
                    return false;
                }
                break;
            }
            if (byteIn < 4 && ++inOffset == inEnd) {
                fPartialMultiByteResult = exitNormalize(inOffset, outOffset, true);
                return false;
            }
            break;
        }
        return exitNormalize(inOffset, outOffset, true);
    }
    //
    //
    //
    protected class UTF8DataChunk implements StringProducer {
        //
        // Constructor
        //
        protected UTF8DataChunk(StringPool stringPool, UTF8DataChunk prev) throws Exception {
            fStringPool = stringPool;
            fChunk = prev == null ? 0 : prev.fChunk + 1;
            fPreviousChunk = prev;
            if (prev != null)
                prev.fNextChunk = this;
        }
        //
        // Constructor for creating clumps.
        //
        protected UTF8DataChunk(UTF8DataChunk proto) {
            fStringPool = proto.fStringPool;
            fChunk = proto.fChunk;
            fData = proto.fData;
            fPreviousChunk = null;
            fNextChunk = null;
        }
        //
        //
        //
        public UTF8DataChunk chunkFor(int offset) {
            int firstChunk = offset >> CHUNK_SHIFT;
            UTF8DataChunk dataChunk = this;
            while (firstChunk != dataChunk.fChunk)
                dataChunk = dataChunk.fPreviousChunk;
            return dataChunk;
        }
        //
        //
        //
        public byte[] toByteArray() {
            return fData;
        }
        //
        //
        //
        public void setByteArray(byte[] data) {
            fData = data;
        }
        //
        //
        //
        public UTF8DataChunk nextChunk() {
            return fNextChunk;
        }
        //
        //
        //
        public void clearPreviousChunk() {
            if (fPreviousChunk != null) {
//                System.out.println("clearPreviousChunk: clear chunk " + fPreviousChunk.fChunk);
                fPreviousChunk.fNextChunk = null;
                fPreviousChunk = null;
            }
        }
        /**
         *
         */
        public String toString(int offset, int length) {
            //
            //
            //
            StringBuffer sb = new StringBuffer(length);
            UTF8DataChunk dataChunk = this;
            int endOffset = offset + length;
            int index = offset & CHUNK_MASK;
            byte[] data = fData;
            while (offset < endOffset) {
                int b0 = data[index++] & 0xff;
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                if ((b0 & 0x80) == 0) {
                    sb.append((char)b0);
                    continue;
                }
                int b1 = data[index++] & 0xff;
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                if ((0xe0 & b0) == 0xc0) { // 110yyyyy 10xxxxxx
                    int ch = ((0x1f & b0)<<6) + (0x3f & b1);
                    sb.append((char)ch); // yyy yyxx xxxx (0x80 to 0x7ff)
                    continue;
                }
                int b2 = data[index++] & 0xff;
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                if ((0xf0 & b0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
                    int ch = ((0x0f & b0)<<12) + ((0x3f & b1)<<6) + (0x3f & b2);
                    sb.append((char)ch); // zzzz yyyy yyxx xxxx (0x800 to 0xffff)
                    continue;
                }
                int b3 = data[index++] & 0xff;  // 11110uuu 10uuzzzz 10yyyyyy 10xxxxxx
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                int ch = ((0x0f & b0)<<18) + ((0x3f & b1)<<12)
                       + ((0x3f & b2)<<6) + (0x3f & b3);
                if (ch < 0x10000)
                    sb.append((char)ch);
                else {
                    sb.append((char)(((ch-0x00010000)>>10)+0xd800));
                    sb.append((char)(((ch-0x00010000)&0x3ff)+0xdc00));
                }
            }
            return sb.toString();
        }
        /**
         *
         */
        public boolean equalsString(int offset, int length, String str, int strLength) {
            //
            //
            //
            UTF8DataChunk dataChunk = this;
            int endOffset = offset + length;
            int index = offset & CHUNK_MASK;
            byte[] data = fData;
            int i = 0;
            while (offset < endOffset) {
                if (i == strLength)
                    return false;
                int b0 = data[index++] & 0xff;
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                if ((b0 & 0x80) == 0) {
                    if (b0 != str.charAt(i++))
                        return false;
                    continue;
                }
                int b1 = data[index++] & 0xff;
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                if ((0xe0 & b0) == 0xc0) { // 110yyyyy 10xxxxxx
                    int ch = ((0x1f & b0)<<6) + (0x3f & b1);
                    if (ch != str.charAt(i++))
                        return false;
                    continue;
                }
                int b2 = data[index++] & 0xff;
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                if ((0xf0 & b0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
                    int ch = ((0x0f & b0)<<12) + ((0x3f & b1)<<6) + (0x3f & b2);
                    if (ch != str.charAt(i++))
                        return false;
                    continue;
                }
                int b3 = data[index++] & 0xff;  // 11110uuu 10uuzzzz 10yyyyyy 10xxxxxx
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                int ch = ((0x0f & b0)<<18) + ((0x3f & b1)<<12)
                       + ((0x3f & b2)<<6) + (0x3f & b3);
                if (ch < 0x10000) {
                    if (ch != str.charAt(i++))
                        return false;
                } else {
                    if ((((ch-0x00010000)>>10)+0xd800) != str.charAt(i++))
                        return false;
                    if (i == strLength)
                        return false;
                    if ((((ch-0x00010000)&0x3ff)+0xdc00) != str.charAt(i++))
                        return false;
                }
            }
            return (i == strLength);
        }
        //
        //
        //
        protected UTF8DataChunk createClump(int lastChunk) {
            UTF8DataChunk result = new UTF8DataChunk(this);
            UTF8DataChunk nextChunk = fNextChunk;
            UTF8DataChunk aChunk = result;
            do {
                UTF8DataChunk next = new UTF8DataChunk(nextChunk);
                aChunk.fNextChunk = next;
                nextChunk = nextChunk.fNextChunk;
                aChunk = next;
            } while (aChunk.fChunk != lastChunk);
            return result;
        }
        /**
         *
         */
        public int addString(int offset, int length) {
            int chunk = offset >> CHUNK_SHIFT;
            if (chunk != fChunk) {
                if (fPreviousChunk != null)
                    return fPreviousChunk.addString(offset, length);
				try {
					fErrorHandler.error(ErrorCode.E_INT_PCN);
				} catch (Exception e) {
					// revisit - don't fail silently
				}
                return -1;
            }
            int lastChunk = (offset + length - 1) >> CHUNK_SHIFT;
            if (chunk == lastChunk)
                return fStringPool.addString(this, offset & CHUNK_MASK, length);
            return fStringPool.addString(createClump(lastChunk), offset & CHUNK_MASK, length);
        }
        /**
         *
         */
        public int addSymbol(int offset, int length, int hashcode) {
            int chunk = offset >> CHUNK_SHIFT;
            if (chunk != fChunk) {
                if (fPreviousChunk != null)
                    return fPreviousChunk.addSymbol(offset, length, hashcode);
				try {
					fErrorHandler.error(ErrorCode.E_INT_PCN);
				} catch (Exception e) {
					// revisit -- don't fail silently
				}
                return -1;
            }
            int lastChunk = (offset + length - 1) >> CHUNK_SHIFT;
            int index = offset & CHUNK_MASK;
            if (chunk == lastChunk) {
                if (hashcode == 0)
                    hashcode = getHashcode(offset, length);
                return fStringPool.addSymbol(this, offset & CHUNK_MASK, length, hashcode);
            }
            if (hashcode == 0)
                hashcode = getHashcode(offset, length);
            return fStringPool.addSymbol(createClump(lastChunk), offset & CHUNK_MASK, length, hashcode);
        }
        //
        //
        //
        public void append(ChunkyCharArray charArray, int offset, int length) {
            //
            //
            //
            UTF8DataChunk dataChunk = chunkFor(offset);
            int endOffset = offset + length;
            int index = offset & CHUNK_MASK;
            byte[] data = fData;
            while (offset < endOffset) {
                int b0 = data[index++] & 0xff;
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                if ((b0 & 0x80) == 0) {
                    charArray.append((char)b0);
                    continue;
                }
                int b1 = data[index++] & 0xff;
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                if ((0xe0 & b0) == 0xc0) { // 110yyyyy 10xxxxxx
                    int ch = ((0x1f & b0)<<6) + (0x3f & b1);
                    charArray.append((char)ch); // yyy yyxx xxxx (0x80 to 0x7ff)
                    continue;
                }
                int b2 = data[index++] & 0xff;
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                if ((0xf0 & b0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
                    int ch = ((0x0f & b0)<<12) + ((0x3f & b1)<<6) + (0x3f & b2);
                    charArray.append((char)ch); // zzzz yyyy yyxx xxxx (0x800 to 0xffff)
                    continue;
                }
                int b3 = data[index++] & 0xff;  // 11110uuu 10uuzzzz 10yyyyyy 10xxxxxx
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                int ch = ((0x0f & b0)<<18) + ((0x3f & b1)<<12)
                       + ((0x3f & b2)<<6) + (0x3f & b3);
                if (ch < 0x10000)
                    charArray.append((char)ch);
                else {
                    charArray.append((char)(((ch-0x00010000)>>10)+0xd800));
                    charArray.append((char)(((ch-0x00010000)&0x3ff)+0xdc00));
                }
            }
        }
        //
        //
        //
        protected int getHashcode(int offset, int length) {
            //
            //
            //
            UTF8DataChunk dataChunk = chunkFor(offset);
            int endOffset = offset + length;
            int hashcode = 0;
            int position = 0;
            int chunk = offset >> CHUNK_SHIFT;
            int index = offset & CHUNK_MASK;
            byte[] data = dataChunk.fData;
            while (offset < endOffset) {
                int b0 = data[index++] & 0xff;
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                if ((b0 & 0x80) == 0) {
                    hashcode = StringHasher.hashChar(hashcode, position++, b0);
                    continue;
                }
                int b1 = data[index++] & 0xff;
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                if ((0xe0 & b0) == 0xc0) { // 110yyyyy 10xxxxxx
                    int ch = ((0x1f & b0)<<6) + (0x3f & b1); // yyy yyxx xxxx (0x80 to 0x7ff)
                    hashcode = StringHasher.hashChar(hashcode, position++, ch);
                    continue;
                }
                int b2 = data[index++] & 0xff;
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                if ((0xf0 & b0) == 0xe0) { // 1110zzzz 10yyyyyy 10xxxxxx
                    int ch = ((0x0f & b0)<<12) + ((0x3f & b1)<<6) + (0x3f & b2); // zzzz yyyy yyxx xxxx (0x800 to 0xffff)
                    hashcode = StringHasher.hashChar(hashcode, position++, ch);
                    continue;
                }
                int b3 = data[index++] & 0xff;  // 11110uuu 10uuzzzz 10yyyyyy 10xxxxxx
                offset++;
                if (index == CHUNK_SIZE) {
                    dataChunk = dataChunk.fNextChunk;
                    data = dataChunk.fData;
                    index = 0;
                }
                int ch = ((0x0f & b0)<<18) + ((0x3f & b1)<<12)
                       + ((0x3f & b2)<<6) + (0x3f & b3);
                if (ch < 0x10000)
                    hashcode = StringHasher.hashChar(hashcode, position++, ch);
                else {
                    hashcode = StringHasher.hashChar(hashcode, position++, (int)(((ch-0x00010000)>>10)+0xd800));
                    hashcode = StringHasher.hashChar(hashcode, position++, (int)(((ch-0x00010000)&0x3ff)+0xdc00));
                }
            }
            return StringHasher.finishHash(hashcode);
        }
        //
        //
        //
        protected StringPool fStringPool = null;
        protected int fChunk;
        protected byte[] fData = null;
        protected UTF8DataChunk fPreviousChunk = null;
        protected UTF8DataChunk fNextChunk = null;
    }
}
