/*
 * (C) Copyright IBM Corp. 1997-1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.parser;

import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import com.ibm.xml.framework.ParserState;
import com.ibm.xml.framework.StringPool;
import com.ibm.xml.framework.EntityPool;
import com.ibm.xml.framework.XMLDocumentHandler;
import com.ibm.xml.framework.XMLDocumentTypeHandler;
import com.ibm.xml.internal.ErrorCode;
import com.ibm.xml.internal.DefaultScanner;
import com.ibm.xml.internal.TXCatalog;
import com.ibm.xml.parsers.TXDOMParser;

import org.xml.sax.AttributeList;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.DocumentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import org.w3c.dom.EntityReference;
import org.w3c.dom.Node;

/**
 * This parser provides a backward compatibile API on top of the
 * new parser architecture. This parser should behave the same as
 * the old parser, with the exceptions listed below.
 * <p>
 * Users of this parser instance should <i>not</i> call any
 * functions in the new parser architecture. Calling methods
 * outside of the old parser API can have unexpected results.
 * <p>
 * Users who are moving to the new parser architecture but want
 * to use the catalog file format supported by the old parser
 * should use the <tt>com.ibm.xml.internal.TXCatalog</tt> class.
 * <p>
 * <b>Not supported:</b> Calling these methods will throw
 * <tt>java.lang.IllegalArgumentException</tt>.
 * <ul>
 * <li> addNoRequiredAttributeHandler
 * <li> getReaderBufferSize
 * <li> setErrorNoByteMark
 * <li> setReaderBufferSize
 * </ul>
 * <p>
 * <b>Not implemented:</b> These methods are present but should
 * not be expected to function the same as in the old parser.
 * <ul>
 * <li> setProcessExternalDTD
 * <li> setWarningNoDoctypeDecl
 * <li> setWarningNoXMLDecl
 * <li> setWarningRedefinedEntity
 * </ul>
 *
29 1.15 src/com/ibm/xml/parser/Parser.java, parser, xml4j2, xml4j2_0_13  * @version Revision: 
 * @author TAMURA Kent &lt;kent@trl.ibm.co.jp&gt;
 * @author Andy Clark, IBM
 *
 * @see com.ibm.xml.internal.TXCatalog
 */
public class Parser
    extends TXDOMParser
    {

    //
    // Constants
    //

    /** Special <i>all</i> tag name for tag handlers. */
    private static final String ALL_TAG_NAMES = " ALL ";

    /** General debugging. */
    private static final boolean DEBUG = false;

    //
    // Data
    //

    // misc

    /** Stream producer. */
    StreamProducer streamProducer;

    /** Error listener. */
    ErrorListener errorListener;

    // errors

    /** Error count. */
    int errorCount;

    /** Validity error count. */
    int validityFailureCount;

    /** Warning count. */
    int warnCount;

    // properties

    /** Keep comments. */
    boolean isKeepComment;

    /** Stop parsing at first error. */
    boolean isEndBy1stError;

    /** Processes name spaces. */
    boolean isProcessNamespace;

    /** Processes external DTDs. */
    boolean isProcessExternalDTD;

    /** Redefined entities marked as warning. */
    boolean isWarningRedefinedEntity;

    /** Missing XML declaration marked as warning. */
    boolean isWarningNoXMLDecl;

    /** Missing doctype declaration marked as warning. */
    boolean isWarningNoDoctypeDecl;

    // handlers

    /** Reference handler. */
    ReferenceHandler referenceHandler;

    /** Tag handler. */
    TagHandler tagHandler;

    /** Pre-root handlers. */
    Vector preRootHandlers;

    /** Processing instruction handlers. */
    Vector piHandlers;

    /**
     * Element handler hashtable. The values are Vectors of element
     * handler objects.
     */
    Hashtable elementHandlerHash;

    /** Reference to SAXDriver if it exists. */
    SAXDriver saxdriver = null;
    
    //
    // Constructors
    //

    /**
     * Constructor when using the default <tt>ErrorListener</tt>
     * and <tt>StreamProducer</tt>. The default <tt>ErrorListener</tt>
     * and <tt>StreamProducer</tt> are defined by the <tt>Stderr</tt>
     * class.
     *
     * @param   name    URL or filespec to use as the default input
     *                  stream; if a filespec is provided, this value
     *                  can include a drive and directory spec. This
     *                  value is also used to associate a name with
     *                  errors reported to the default error listener
     *                  that have a <var>file</var> parameter
     *                  <code>=null</code>.
     *
     * @see com.ibm.xml.parser.Stderr
     * @see com.ibm.xml.parser.ErrorListener
     * @see com.ibm.xml.parser.StreamProducer
     */
    public Parser(String name) {
        this(name, null, null);
        }

    /**
     * Constructor when NOT using the default <tt>ErrorListener</tt>
     * or <tt>StreamProducer</tt>. The default <tt>ErrorListener</tt>
     * and <tt>StreamProducer</tt> are defined by the <tt>Stderr</tt>
     * class.
     *
     * @param name  This parameter is only used if either
     *              <var>errorListener</var> or <var>streamProducer</var>
     *              are allowed to default.  If <var>streamProducer</var>
     *              is allowed to default, this value is the URL or
     *              filespec to use as the default input stream; if a
     *              filespec is provided, this value can include a drive
     *              and directory spec.  If <var>errorListener</var> is
     *              allowed to default, this value is used to associate a
     *              name with errors reported to the default error
     *              listener that have a <var>file</var> parameter
     *              <code>=null</code>.
     * @param errorListener Handler for error and warning events that are
     *                      recognized by this parser, or <var>null</var> if
     *                      to use the default listener.
     * @param streamProducer    Mechanism for manipulating this parser's
     *                          input stream, or <var>null</var> if to use
     *                          the default stream producer.
     *
     * @see com.ibm.xml.parser.Stderr
     * @see com.ibm.xml.parser.ErrorListener
     * @see com.ibm.xml.parser.StreamProducer
     */
    public Parser(String name,
                  ErrorListener errorListener,
                  StreamProducer streamProducer) {

        // set error listener and stream producer
        Stderr se = (errorListener == null || streamProducer == null)
                    ? new Stderr(name) : null;
        this.errorListener  = (errorListener  != null) ? errorListener  : se;
        this.streamProducer = (streamProducer != null) ? streamProducer : se;

        // set defaults
        initParser();

        } // <init>(String,ErrorListener,StreamProducer)

    /** Set default values. */
    private void initParser() {

        // set properties
        setAllowJavaEncodingName(false);
        setEndBy1stError(true);
        setExpandEntityReferences(false);
        setKeepComment(true);
        setProcessExternalDTD(true);
        setProcessNamespace(false);
        setWarningNoDoctypeDecl(false);
        setWarningNoXMLDecl(true);
        setWarningRedefinedEntity(true);

        // set handlers
        Handler handler = new Handler();
        setDocumentHandler(handler);
        setDocumentTypeHandler(handler);
        setErrorHandler(handler);
        getEntityHandler().setEntityResolver(handler);

        } // init()

    //
    // Public methods
    //

    // element factory

    /**
     * Sets the code to receive control when the various components of
     * the Document Object Model (DOM) and XML4J need to be created and
     * initialized. When this method is called, readStream() returns the
     * <VAR>elementFactory</VAR>.
     * <p>
     * By default, an instance of <tt>TXDocument</tt> is used.
     *
     * @param factory   Factory to use to create and initialize
     *                  objects.  This must be an instance of
     *                  TXDocument or a subclass of TXDocument.
     *
     * @see com.ibm.xml.parser.TXDocument
     */
    public void setElementFactory(TXDocument factory) {
        super.fDocument = factory;
        }

    // input stream

    /**
     * Returns the source of the input stream (could be a character stream
     * or a byte stream) based on the entity specified by the system ID
     * and/or public ID, and the <var>streamProducer</var> specified when
     * this parser was constructed.
     *
     * @param name      The name to be associated with the input stream.
     *                  For example, this name could be the public ID of
     *                  an external ID, or a local filename.
     * @param publicID  Entity's public ID, or <var>null</var> if no
     *                  public ID (see ExternalID for details).
     * @param systemID  Entity's system ID.
     *
     * @return              The resolved source of the input stream, or
     *                      <var>null</var> if unable to resolve.
     *
     * @exception java.io.IOException   Thrown if unable to open the source
     *                                  defined by the specified IDs.
     *
     * @see com.ibm.xml.parser.ExternalID
     * @see #closeInputStream
     * @see com.ibm.xml.parser.StreamProducer#getInputStream
     */
    public Source getInputStream(String name,
                                 String publicID, String systemID)
        throws IOException
        {
        return streamProducer.getInputStream(name, publicID, systemID);
        }

    /**
     * Removes the input stream currently in use.
     *
     * @param source Source of the input stream to be closed.
     *
     * @see #getInputStream
     * @see com.ibm.xml.parser.StreamProducer#closeInputStream
     */
    public void closeInputStream(Source source) {
        streamProducer.closeInputStream(source);
        }

    // catalog methods

    /**
     * Loads a catalog which provides mapping between public IDs and
     * system IDs.
     *
     * @param reader Character input stream reader.
     *
     * @exception java.io.IOException   Thrown if <var>reader</var> is
     *                                  invalid.
     *
     * @see #getInputStream
     * @see com.ibm.xml.parser.ExternalID
     * @see com.ibm.xml.parser.Stderr#loadCatalog
     * @see com.ibm.xml.parser.StreamProducer#loadCatalog
     */
    public void loadCatalog(Reader reader)
        throws IOException
        {
        streamProducer.loadCatalog(reader);
        }

    // set/add handler methods

    /**
     * Register a handler for receiving control when tag start and tag
     * end events are recognized by this parser.
     * <p>
     * By default, no tag handler is registered to this parser.
     *
     * @param handler   Handler to be registered.
     *
     * @see com.ibm.xml.parser.TagHandler
     */
    public void setTagHandler(TagHandler handler) {
        tagHandler = handler;
        }

    /**
     * Register a handler for receiving control when a General Reference
     * is recognized by this parser.
     * <p>
     * By default, no reference handler is registered to this parser.
     *
     * @param handler   Handler to be registered.
     *
     * @see com.ibm.xml.parser.ReferenceHandler
     */
    public void setReferenceHandler(ReferenceHandler handler) {
        referenceHandler = handler;
        }

    /**
      * Register a handler for receiving control from this parser after
      * the internal and/or external DTD subsets are parsed, but before
      * the document's root Element is parsed.
      * <p>
      * By default, no preroot handlers are registered to this parser.
      *
      * @param handler  Handler to be registered.
      *
      * @see com.ibm.xml.parser.PreRootHandler
      */
    public void addPreRootHandler(PreRootHandler handler) {

        if (preRootHandlers == null) {
            preRootHandlers = new Vector();
            }
        preRootHandlers.addElement(handler);

        } // addPreRootHandler(PreRootHandler)

    /**
     * Register a handler for receiving control when PIs are recognized
     * by this parser.
     * <p>
     * By default, no PI handlers are registered to this parser.
     *
     * @param handler   Handler to be registered.
     *
     * @see com.ibm.xml.parser.PIHandler
     */
    public void addPIHandler(PIHandler handler) {

        if (piHandlers == null) {
            piHandlers = new Vector();
            }
        this.piHandlers.addElement(handler);

        } // addPIHandler(PIHandler)

    /**
     * Register a handler for receiving control when ANY element tag is
     * recognized by this parser.
     * <p>
     * By default, no element handlers are registered to this parser.
     *
     * @param handler   Handler to be registered.
     *
     * @see com.ibm.xml.parser.ElementHandler
     */
    public void addElementHandler(ElementHandler handler) {
        addElementHandler(handler, ALL_TAG_NAMES);
        }

    /**
     * Register a handler for receiving control when the specified
     * <var>elementTagName</var> element tag is recognized by this parser.
     * <p>
     * By default, no element handlers are registered to this parser.
     *
     * @param handler       Handler to be registered for the specified
     *                      <var>elementName</var>.
     * @param elementName   Element name to be watched for by this parser.
     *                      If the Element's name has a namespace prefix,
     *                      that prefix should be included.
     *
     * @see com.ibm.xml.parser.ElementHandler
     */
    public void addElementHandler(ElementHandler handler,
                                  String elementName) {

        if (elementHandlerHash == null) {
            elementHandlerHash = new Hashtable();
            }
        Vector v = (Vector)elementHandlerHash.get(elementName);
        if (v == null) {
            v = new Vector();
            elementHandlerHash.put(elementName, v);
            }
        v.addElement(handler);

        } // addElementHandler(ElementHandler,String)

    /**
     * This allows the Parser to know about the SAXDriver. 
     *
     * The XML4J1.x Parser needed to know about the SAXDriver for 
     * various reasons which are no longer required.
     * 
     * The current reason in XML4J2.x (Wombat) is because when the 
     * Parser is being used by the SAXDriver, must tell the TXDOM
     * not to do Text Node "normalization", because the createTextNode
     * calls are used to call the SAX callbacks.
     */
    void setSAXDriver(SAXDriver sax) {
        this.saxdriver = sax;
        if (sax!=null) {
            setNormalizeTextNodes(false);
        }
    }
    // error methods

    /**
     * Returns the number of errors and exceptions currently encountered
     * by this parser. Warning messages do not count in this number.
     *
     * @return Number of errors and exceptions encountered by this parser.
     *
     * @see com.ibm.xml.parser.ErrorListener
     */
    public int getNumberOfErrors() {
        return errorCount + validityFailureCount;
        }

    /**
     * Returns the number of warnings currently encountered by this parser.
     *
     * @return Number of warnings encountered by this parser.
     *
     * @see com.ibm.xml.parser.ErrorListener
     */
    public int getNumberOfWarnings() {
        return warnCount;
        }

    // boolean properties

    /**
     * Sets whether <code>TXComment</code> Nodes are created by this
     * parser when XML comments are encountered.
     * <p>
     * By default, <tt>TXComment</tt> Nodes are created by this parser.
     *
     * @param keep  <code>=true</code> means maintain parsed comments;
     *              <code>=false</code> means drop comments when creating
     *              object tree.
     *
     * @see com.ibm.xml.parser.TXComment
     */
    public void setKeepComment(boolean keep) {
        isKeepComment = keep;
        }

    /**
     * Sets whether <tt>EntityReference</tt> Nodes are expanded by this
     * parser.
     * <p>
     * By default, <tt>EntityReference</tt> Nodes are kept.
     *
     * @param keep  <code>=true</code> means expanding entity references;
     *              <code>=false</code> means maintain entity references.
     *
     * @see org.w3c.dom.EntityReference
     * @see com.ibm.xml.parser.GeneralReference
     */
    public void setExpandEntityReferences(boolean expand) {
        super.setExpandEntityReferences(expand);
        }

    /**
     * <b>This method is not implemented.</b>
     * <p>
     * Sets whether this parser reads an external subset of DTD, pointed
     * in an ExternalID in DOCTYPE declaration.
     * <P>
     * By default, this parser reads an external subset of DTD.
     *
     * @param process   <code>=true</code> means process an external DTD;
     *                  <code>=false</code> means ignore an external DTD.
     *
     * @see com.ibm.xml.parser.DTD
     */
    public void setProcessExternalDTD(boolean process) {
        isProcessExternalDTD = process;
        }

    /**
     * Sets whether namespaces are respected by this parser. Namespace
     * notation can occur on the names of <tt>TXElement</tt>, <tt>TXPI</tt>,
     * and <tt>TXAttribute</tt> Nodes.  If namespaces are respected, the
     * methods defined by the <tt>Namespace</tt> interface will function
     * properly in these Nodes.
     * <p>
     * By default, namespaces are NOT respected by this parser.
     *
     * @param process   <code>=true</code> means parse for namespaces;
     *                  <code>=false</code> means ignore namespaces when
     *                  parsing.
     *
     * @see com.ibm.xml.parser.Namespace
     * @see com.ibm.xml.parser.TXElement
     * @see com.ibm.xml.parser.TXAttribute
     */
    public void setProcessNamespace(boolean process) {
        // TODO: what should be done here? Is it equivalent? -Ac
        isProcessNamespace = process;
		if (process) 
			setValidationHandler(null);
		else 
			useDefaultValidationHandler();
//        Parser.this.setCheckNamespace(process);
        }


    /**
     * Sets whether this parser allows Java encoding names to be specified
     * in place of MIME charset names on <var>xmlEncoding</var> parameters.
     * The <tt>MIME2Java</tt> class is provided to facilitate conversion
     * when Java encoding names are not allowed.
     * <p>
     * By default, Java encoding names are NOT allowed on
     * <var>xmlEncoding</var> parameters by this parser.
     *
     * @param allow <code>=true</code> means Java encoding names are allowed;
     *              <code>=false</code> means MIME charset names must be used.
     *
     * @see com.ibm.xml.parser.TXDocument#setEncoding
     * @see com.ibm.xml.parser.DTD#setEncoding
     * @see com.ibm.xml.parser.MIME2Java
     */
    public void setAllowJavaEncodingName(boolean allow) {
        super.setAllowJavaEncodingName(allow);
        }

    /**
     * <b>This method is not supported.</b>
     * <p>
     * Sets, at the document level, whether space is to be preserved
     * in the parsed document. Regardless of the value of
     * <var>isPreserveSpace</var>, this parser will preserve whitespace.
     * This value is used, for example, to determine if space is to be
     * preserved in Text Nodes during printWithFormat() operations.  Note
     * that the document-level value will be overridden at the Node level
     * if this parser recognizes a <code>xml:space</code> attribute.
     * <p>
     * By default, space is not to be preserved in the parsed document.
     *
     * @param preserve  <code>=true</code> space is to be preserved;
     *                  <code>=false</code> space is to be ignored.
     *
     */
    public void setPreserveSpace(boolean preserve) {
        // TODO: move this string to resource bundle. -Ac
        throw new IllegalArgumentException("setPreserveSpace not supported");
        }

    /**
     * Sets whether this parser terminates processing when an initial error
     * or exception occurs.
     * <P>
     * By default, this parser stops parsing after an error occurs, as
     * required by the XML specification.
     *
     * @param end   <code>=true</code> means initial error or exception ends
     *              processing;
     *              <code>=false</code> means processing continues until the
     *              document is fully parsed.
     */
    public void setEndBy1stError(boolean end) {
        isEndBy1stError = end;
        super.setContinueAfterFatalError(!end);
        }

    // parse documents

    /**
     * Returns a parsed XML document based on the specified
     * <var>inputStream</var>. The returned document can then be accessed
     * by methods such as <tt>TXDocument#getDocumentElement</tt>.
     * <p>
     * Once this method is invoked, the parser instance is of no further use,
     * and should NOT be reused.
     *
     * @param inputStream   Byte-stream-oriented data source.
     *
     * @return Parsed XML document (should never be <var>null</var>).
     *
     * @see com.ibm.xml.parser.Source
     * @see com.ibm.xml.parser.TXDocument#getDocumentElement
     * @see #readDTDStream
     */
    public TXDocument readStream(InputStream inputStream) {
        return readStream(new Source(inputStream));
        }

    /**
     * Returns a parsed XML document based on the specified
     * <var>reader</var>. The returned document can then be accessed by
     * methods such as <tt>TXDocument#getDocumentElement</tt>.
     * <p>
     * Once this method is invoked, the parser instance is of no further
     * use, and should NOT be reused.
     *
     * @param reader    Character-stream-oriented data source.
     *
     * @return Parsed XML document (should never be <var>null</var>).
     *
     * @see com.ibm.xml.parser.Source
     * @see com.ibm.xml.parser.TXDocument#getDocumentElement
     * @see #readDTDStream
     */
    public TXDocument readStream(Reader reader) {
        return readStream(new Source(reader));
        }

    /**
     * Returns a parsed XML document based on the specified
     * <var>source</var>. The returned document can then be accessed by
     * methods such as <tt>TXDocument#getDocumentElement</tt>.
     * <p>
     * Once this method is invoked, the parser instance is of no further
     * use, and should NOT be reused.
     *
     * @param source    Byte-stream-oriented or character-stream-oriented data source.
     *
     * @return Parsed XML document (should never be <var>null</var>).
     *
     * @see com.ibm.xml.parser.Source
     * @see com.ibm.xml.parser.TXDocument#getDocumentElement
     * @see #readDTDStream
     */
    public TXDocument readStream(Source source) {

        try {
            // convert Source to InputSource
            InputSource inputSource = new InputSource();
            inputSource.setByteStream(source.getInputStream());
            inputSource.setCharacterStream(source.getReader());
            inputSource.setEncoding(source.getEncoding());
            if (streamProducer instanceof Stderr) {
                Stderr stderr = (Stderr)streamProducer;
                inputSource.setPublicId(stderr.getPublicId());
                inputSource.setSystemId(stderr.getSystemId());
                }

            // parse
            parse(inputSource);
            return (TXDocument)getDocument();
            }
        catch (Exception e) {
            // let return null
            if (DEBUG) {
                System.out.println("error in readStream: "+e.getMessage());
                }
            }

        // parsing error
        return null;

        } // readStream(Source):TXDocument

    // parse DTDs

    /**
     * Returns a parsed external DTD subset based on the specified
     * <var>inputStream</var>. The returned document can then be
     * accessed by methods such as <code>DTD#externalElements</code>.
     * <p>
     * Once this method is invoked, the parser instance is of no further
     * use, and should NOT be reused.
     *
     * @param inputStream   Byte-stream-oriented data source.
     *
     * @return Parsed DTD (should never be <var>null</var>).
     *
     * @see com.ibm.xml.parser.Source
     * @see com.ibm.xml.parser.DTD
     * @see #readStream
     */
    public DTD readDTDStream(InputStream inputStream) throws IOException {
        return readDTDStream(new Source(inputStream));
        }

    /**
     * Returns a parsed external DTD subset based on the specified
     * <var>reader</var>. The returned document can then be accessed by
     * methods such as <code>DTD#externalElements</code>.
     * <p>
     * Once this method is invoked, the parser instance is of no further
     * use, and should NOT be reused.
     *
     * @param reader Character-stream-oriented data source.
     *
     * @return Parsed DTD (should never be <var>null</var>).
     *
     * @see com.ibm.xml.parser.Source
     * @see com.ibm.xml.parser.DTD
     * @see #readStream
     */
    public DTD readDTDStream(Reader reader) throws IOException {
        return readDTDStream(new Source(reader));
        }

    /**
     * Returns a parsed external DTD subset based on the specified
     * <var>source</var>. The returned document can then be accessed by
     * methods such as <code>DTD#externalElements</code>.
     * <p>
     * Once this method is invoked, the parser instance is of no further
     * use, and should NOT be reused.
     *
     * @param source    Byte-stream-oriented or character-stream-oriented
     *                  data source.
     *
     * @return Parsed DTD (should never be <var>null</var>).
     *
     * @see com.ibm.xml.parser.Source
     * @see com.ibm.xml.parser.DTD
     * @see #readStream
     */
    public DTD readDTDStream(Source source) throws IOException {

        try {
            // convert Source to InputSource
            InputSource inputSource = new InputSource();
            inputSource.setByteStream(source.getInputStream());
            inputSource.setCharacterStream(source.getReader());
            inputSource.setEncoding(source.getEncoding());

            // parse
            parseDTD(inputSource);
            return ((TXDocument)getDocument()).getDTD();
            }
        catch (Exception e) {
            // let return null
            if (DEBUG) {
                System.out.println("error in readDTDStream: "+e.getMessage());
                }
            }

        // parsing error
        return null;

        } // readDTDStream(Source):DTD
    
    /**
     * Parses the specified input source.
     *
     * @param source The input source.
     *
     * @exception org.xml.sax.SAXException Throws exception on SAX error.
     * @exception java.io.IOException Throws exception on i/o error.
     */
    public final void parseDTD(InputSource source) throws org.xml.sax.SAXException, IOException {
        if (fParseInProgress)
            throw new org.xml.sax.SAXException("parse may not be called while parsing"); // REVISIT - need to add new error message
        if (fNeedReset)
            resetOrCopy();
        try {
            checkHandlers();
            fParseInProgress = true;
            fNeedReset = true;
            ((DefaultScanner)fScanner).scanDTD(source);
            fParseInProgress = false;
        } catch (org.xml.sax.SAXException ex) {
            fParseInProgress = false;
            throw ex;
        } catch (IOException ex) {
            fParseInProgress = false;
            throw ex;
        } catch (Exception ex) {
            fParseInProgress = false;
            throw new org.xml.sax.SAXException(ex);
        }
    }
    
    

    // not supported

    /**
     * <b>This method is not supported.</b>
     * <p>
     * Register a handler for receiving control when a "no required
     * attribute" event is recognized by this parser.  A "no required
     * attribute" event occurs when an Element is parsed and a required
     * Attribute for this Element is not seen; Attributes are declared
     * as required as part of the internal and/or external DTD declaration
     * of the XML document.
     * <p>
     * By default, no required attribute handlers are registered to this
     * parser.
     *
     * @param handler   Handler to be registered.
     *
     * @see com.ibm.xml.parser.NoRequiredAttributeHandler
     */
    public void addNoRequiredAttributeHandler(NoRequiredAttributeHandler handler) {
        // TODO: move this string to resource bundle. -Ac
        throw new IllegalArgumentException("addNoRequiredAttributeHandler is not supported");
        }

    /**
     * <b>This method is not supported.</b>
     * <p>
     * Sets whether this parser treats missing byte marks in the input
     * stream as an error, and reports this condition through the defined
     * <tt>ErrorListener</tt>.
     * <p>
     * By default, this condition is reported by this parser as an error.
     *
     * @param error <code>=true</code> means report condition as an error;
     *              <code>=false</code> means ignore this condition.
     */
    public void setErrorNoByteMark(boolean error) {
        // TODO: move this string to resource bundle. -Ac
        throw new IllegalArgumentException("setErrorNoByteMark not supported");
        }

    /**
     * <b>This method is not supported.</b>
     * <p>
     * Returns the size of this parser's character-stream-oriented input
     * source buffer. Characters are buffered during I/O in order to
     * provide efficient reading. This method has no meaning for
     * byte-stream-oriented input sources.
     *
     * @return Size of this parser's character-stream-oriented input buffer.
     *
     * @see #setReaderBufferSize
     */
    public int getReaderBufferSize() {
        // TODO: move this string to resource bundle. -Ac
        throw new IllegalArgumentException("setReaderBufferSize is not supported");
        }

    /**
     * <b>This method is not supported.</b>
     * <p>
     * Sets the size of this parser's character-stream-oriented input
     * source buffer. Characters are buffered during I/O in order to
     * provide efficient reading. This method has no meaning for
     * byte-stream-oriented input sources.
     * <p>
     * The default buffer size is large enough for most purposes.
     *
     * @param size  Size of this parser's character-stream-oriented input
     *              buffer.
     *
     * @see #getReaderBufferSize
     */
    public void setReaderBufferSize(int size) {
        // TODO: move this string to resource bundle. -Ac
        throw new IllegalArgumentException("setReaderBufferSize is not supported");
        }

    // not implemented

    /**
     * <b>This method is not implemented.</b>
     * <p>
     * Sets whether this parser treats redefined entities as a warning,
     * and reports this condition through the defined <tt>ErrorListener</tt>.
     * <p>
     * By default, this condition is reported by this parser as a warning.
     *
     * @param warning   <code>=true</code> means report condition as a
     *                  warning.
     *                  <code>=false</code> means ignore this condition.
     *
     * @see com.ibm.xml.parser.EntityDecl
     */
    public void setWarningRedefinedEntity(boolean warning) {
        isWarningRedefinedEntity = warning;
        }

    /**
     * <b>This method is not implemented.</b>
     * <p>
     * Sets whether this parser treats a missing XML prolog PIs as a
     * warning, and reports this condition through the defined
     * <tt>ErrorListener</tt>.
     * <p>
     * By default, this condition is reported by this parser as a warning.
     *
     * @param warning   <code>=true</code> means report condition as a
     *                  warning;
     *                  <code>=false</code> means ignore this condition.
     *
     * @see com.ibm.xml.parser.TXPI
     */
    public void setWarningNoXMLDecl(boolean warning) {
        isWarningNoXMLDecl = warning;
        }

    /**
     * <b>This method is not implemented.</b>
     * <p>
     * Sets whether this parser treats a missing DTD declaration as a
     * warning, and reports this condition through the defined
     * <tt>ErrorListener</tt>.
     * <p>
     * By default, this condition is NOT reported by this parser as a
     * warning.
     *
     * @param warning   <code>=true</code> means report condition as a
     *                  warning;
     *                  <code>=false</code> means ignore this condition.
     *
     * @see com.ibm.xml.parser.DTD
     */
    public void setWarningNoDoctypeDecl(boolean warning) {
        isWarningNoDoctypeDecl = warning;
        }

    /**
     * Stops a <code>Parser.readStream()</code> operation. The calling
     * thread will signal that the thread executing in
     * <code>this.readStream()</code> should clean up and return as soon
     * as practical.  The calling thread will return immediately. The
     * effect of <code>stop()</code> is sticky, meaning that after
     * <code>stop()</code> has been called, a thread entering
     * readStream() will return immediately. (This helps in the case
     * where the UI thread invokes <code>stop</code> before the processing
     * thread actually enters <code>readStream</code>.
     * <p>
     * <code>stop</code> affects only the specific instance of the parser
     * (this) on which it is invoked.  It is not global;
     * <code>readStream</code> on other parser instances will not be
     * affected.
     *
     */
    public void stop() {
        getParserState().getScanner().stop();
        }

    //
    // XMLParser methods
    //

    /**
     * This method is the basic error handler.
     */
    protected void handleError(String errorKey, String errorMsg)
        throws Exception
        {
        Locator locator = getLocator();
        errorListener.error(locator.getSystemId(),
                            locator.getLineNumber(),
                            locator.getColumnNumber(),
                            errorKey,
                            errorMsg);
        super.handleError(errorKey, errorMsg);
        }

    /**
     * <b>This method is not part of the Parser API and is called
     * automatically by the parser internals.</b>
     * <p>
     * Resets the parser between calls to parse.
     */
    public void reset() {

        // do default reset
        super.reset();

        // clear error counts
        warnCount = 0;
        validityFailureCount = 0;
        errorCount = 0;

        // TODO: should we null these out or leave alone so that
        //       the parser instance can be re-used? -Ac
        // old Parser.clean() method
        this.errorListener = null;
        this.streamProducer = null;
        this.tagHandler = null;
        this.preRootHandlers = null;
        this.piHandlers = null;
        this.elementHandlerHash = null;
        this.referenceHandler = null;

        } // reset()

    //
    // Package private methods
    //

    /** Returns the current node. */
    Node getCurrentNode() {
        return super.fCurrentNode;
        }

    /**
     * fix up nodes replaced by an ElementHandler, should only be called from 
     * Handler.endElement
     */
    void setRealParent(Node replacement) {
        // get true parent (only true in Handler.endElement())
        Node realParent = (Node) fNodeStack.peek();
        
        // remove node to be replaced
        realParent.removeChild(realParent.getLastChild());
        realParent.appendChild(replacement);
        }

    /**
     * Returns the current parent element in the DOM tree. This method is
     * equivalent to <code>(TXElement)super.fNodeStack.peek()</code>.
     */
    TXElement getCurrentParent() {
        return (TXElement)super.fCurrentParent;
        }

	//
	// beginning of modified TX Namespace code
	//

	// these two functions are needed for the TX Namespace implementation.  By the time the namespace
	// checking code is called, the attrbute pools have been released, destroying the information
	// needed to generate good error messages.  These versions operate on string arguments
	private final void errorString1(int errorCode, String string) throws Exception {
		String errorKey = ErrorCode.getErrorKey(errorCode);
		if (!errorHandlingEnabled() && !isFatal(errorKey)) {
			return;
		}
		Object[] errorArgs = new String[1];
		errorArgs[0] = string;
		String errorMsg = getErrorMsgString(errorKey, errorArgs);
		handleError(errorKey, errorMsg);
	}

	private final void errorString4(int errorCode, String string1, String string2, String string3, String string4) throws Exception {
		String errorKey = ErrorCode.getErrorKey(errorCode);
		if (!errorHandlingEnabled() && (getContinueAfterFatalError() || !isFatal(errorKey))) {
			return;
		}
		Object[] errorArgs = new String[4];
		errorArgs[0] = string1;
		errorArgs[1] = string2;
		errorArgs[2] = string3;
		errorArgs[3] = string4;
		String errorMsg = getErrorMsgString(errorKey, errorArgs);
		handleError(errorKey, errorMsg);
	}


	void TXNamespaceCode(TXElement tag, int elementNameIndex, int attrListIndex) {
		TXAttribute[] aa = tag.getAttributeArray();
		if (isProcessNamespace) {
			try {
			// first check attributes

			String[] localpartcache = new String[aa.length];

			com.ibm.xml.framework.AttrPool attPool = getParserState().getAttrPool();
			org.xml.sax.AttributeList al = attPool.getAttributeList(attrListIndex);

			for (int i = 0; i < aa.length; i++) {
				String attrname = aa[i].getNodeName();
				int ret = this.checkNamespace(tag, attrname);
				if (-1 != ret)  { 
					errorString1(ret, attrname);
				}
				if (aa[i].getValue().length() == 0 && attrname.startsWith("xmlns:"))
					errorString1(ErrorCode.E_TAGl, attrname);
				// The following code is O(n^2).
				localpartcache[i] = aa[i].getNSLocalName();
				String thisns = aa[i].getNSName();

				if (thisns != null) {
					for (int j = 0;  j < i;  j ++) {
						if (localpartcache[j].equals(localpartcache[i])) {
							String ns = aa[j].getNSName();
							if (ns != null && ns.equals(thisns))
								errorString4(ErrorCode.E_TAGm, aa[i].getNodeName(), thisns, localpartcache[i], aa[j].getNodeName());
						}
					}
				}
			}

			// then check element name
			int ret = this.checkNamespace(tag, tag.getName());
			if (-1 != ret) Parser.this.error1(ret, elementNameIndex);
			} catch (Exception e) {
				if (DEBUG) {
					System.out.println("Exception during ns check "+e);
					e.printStackTrace();
				}
			}
		}
	}

	int checkNamespace(TXElement element, String qname) {
		int index = qname.indexOf(':');
		if (index < 0)  return -1; // instead of null
		if (null == element.getNamespaceForPrefix(qname.substring(0, index))) {
			return ErrorCode.E_TAGf;
		}
		if (0 <= qname.indexOf(':', index+1)) {
			return ErrorCode.E_TAGe;
		}
		return -1; // instead of null
	}

	// end of TX Namespace code
    //
    // Classes
    //

    /**
     * Default handler to localize handling of the various callbacks.
     *
     * @author  Andy Clark, IBM
     */
    class Handler
        implements XMLDocumentTypeHandler,
                   XMLDocumentHandler,
                   ErrorHandler,
                   EntityResolver
        {

        //
        // XMLDocumentTypeHandler methods
        //

        /** Doctype declaration. */
        public void doctypeDecl(int rootElementNameIndex)
            throws Exception
            {

            if (DEBUG) {
                System.out.println(
                    "doctypeDecl(\""+
                    getParserState().
                    getStringPool().
                    toString(rootElementNameIndex)+
                    "\")"
                    );
                }

            // call pre-root handlers
            Vector preRootHandlers = Parser.this.preRootHandlers;
            if (preRootHandlers != null) {
                Enumeration handlers = preRootHandlers.elements();
                if (handlers.hasMoreElements()) {
                    TXDocument document = (TXDocument)Parser.this.getDocument();
                    String rootName = getParserState().getStringPool().toString(rootElementNameIndex);
                    do {
                        PreRootHandler handler = (PreRootHandler)handlers.nextElement();
                        handler.handlePreRoot(document, rootName);
                        } while (handlers.hasMoreElements());
                    }
                }

            // perform default behavior
            Parser.this.doctypeDecl(rootElementNameIndex);

            } // startEntityReference(int)
            
        /**
         *  This function will be called at the start of the "internal subset" of a 
         *  doctype declaration.
         */
        public void startInternalSubset() throws Exception {
            
            Parser.this.startInternalSubset();
            }

        /**
         *  This function will be called at the end of the "internal subset" of a 
         *  doctype declaration.
         */
        public void endInternalSubset() throws Exception {
            Parser.this.endInternalSubset();
            }

        /**
         *  This function will be called at the start of the "external subset" of a 
         *  doctype declaration.
         */
        public void startExternalSubset(int publicIdStringIndex, int systemIdStringIndex) throws Exception {
            Parser.this.startExternalSubset(publicIdStringIndex,systemIdStringIndex);
        }

        /**
         *  This function will be called at the end of the "external subset" of a 
         *  doctype declaration.
         */
        public void endExternalSubset() throws Exception {
            Parser.this.endExternalSubset();
        }
        

        /** Element declaration. */
        public void elementDecl(int elementIndex)
            throws Exception
            {
            Parser.this.elementDecl(elementIndex);
            }

        /** Unparsed entity declaration. */
        public void unparsedEntityDecl(int entityIndex)
            throws Exception
            {
            Parser.this.unparsedEntityDecl(entityIndex);
            }

        /** Attlist declaration. */
        public void attlistDecl(int elementIndex, int attDefIndex)
            throws Exception
            {
            Parser.this.attlistDecl(elementIndex, attDefIndex);
            }

        /** Internal entity declaration. */
        public void internalEntityDecl(int entityIndex)
            throws Exception
            {
            Parser.this.internalEntityDecl(entityIndex);
            }

        /** External entity declaration. */
        public void externalEntityDecl(int entityIndex)
            throws Exception
            {
            Parser.this.externalEntityDecl(entityIndex);
            }

        /** Notation declaration. */
        public void notationDecl(int notationIndex)
            throws Exception
            {
            Parser.this.notationDecl(notationIndex);
            }

        //
        // XMLDocumentHandler methods
        //

        /**
         * Returns true if parser to send character data as a character array.
         */
        public boolean sendCharDataAsCharArray() {
            return Parser.this.sendCharDataAsCharArray();
            }

        /** End document. */
        public void endDocument()
            throws Exception
            {
            Parser.this.endDocument();
            }

        /** Characters. */
        public void characters(int dataIndex, boolean cdataSection)
            throws Exception
            {
            Parser.this.characters(dataIndex, cdataSection);
            }

        /** Ignorable whitespace. */
        public void ignorableWhitespace(int dataIndex, boolean cdataSection)
            throws Exception
            {
            Parser.this.ignorableWhitespace(dataIndex, cdataSection);
            }

        /** Comment. */
        public void comment(int dataIndex)
            throws Exception
            {
            if (Parser.this.isKeepComment) {
                Parser.this.comment(dataIndex);
                }
            else {
                Parser.this.getParserState().getStringPool().releaseString(dataIndex);
                }
            }

        /** Characters. */
        public void characters(char ch[], int start, int length, boolean cdataSection)
            throws Exception
            {
            Parser.this.characters(ch, start, length, cdataSection);
            }

        /** Ignorable whitespace. */
        public void ignorableWhitespace(char ch[], int start, int length, boolean cdataSection)
            throws Exception
            {
            Parser.this.ignorableWhitespace(ch, start, length, cdataSection);
            }

        /** Start document. */
        public void startDocument(int versionIndex,
                                   int encodingIndex,
                                   int standAloneIndex)
            throws Exception
            {
            Parser.this.startDocument(versionIndex,encodingIndex,standAloneIndex);
            }

        /** Start entity reference. */
        public void startEntityReference(int entityIndex)
            throws Exception
            {

            if (DEBUG) {
                System.out.println(
                    "startEntityReference(\""+
                    getParserState().getStringPool().toString(getParserState().getEntityPool().getEntityName(entityIndex))+
                    "\")");
                }

            // Now expanding or not is handled by the DOM Parser 
            // at creation time within this call.
            Parser.this.startEntityReference(entityIndex);

            // expand entity references
            /***
            //old code
            if (Parser.this.isExpandEntityReferences) {
                EntityReference reference =
                    (EntityReference)Parser.this.getCurrentNode();
                TXElement element = Parser.this.getCurrentParent();
                expandEntityReferences(element, reference);
                element.removeChild(reference);
                }
            /***
            if (Parser.this.isExpandEntityReferences) {
                EntityReference reference =
                    (EntityReference)Parser.this.getCurrentNode();
                //Is reference.getParentNode() always a TXElement? 
                TXElement element = (TXElement)reference.getParentNode();
                //expandEntityReferences(element, reference);
                element.removeChild(reference);
                }
            /***/
                

            // call reference handler
            ReferenceHandler handler = referenceHandler;
            if (handler != null) {
                ParserState state = getParserState();
                EntityPool entity = state.getEntityPool();
                StringPool string = state.getStringPool();
                String entityName = string.toString(entity.getEntityName(entityIndex));
                handler.startReference(entityName);
                }

            } // startEntityReference(int)

        /** End entity reference. */
        public void endEntityReference(int entityIndex)
            throws Exception
            {

            
            if (DEBUG) {
                System.out.println(
                    "endEntityReference(\""+
                    getParserState().getStringPool().toString(getParserState().getEntityPool().getEntityName(entityIndex))+
                    "\")");
                }

            // Now expanding or not is handled by the DOM Parser 
            // at creation time within this call.
            Parser.this.endEntityReference(entityIndex);
          
            // call reference handler
            ReferenceHandler handler = referenceHandler;
            if (handler != null) {
                ParserState state = getParserState();
                EntityPool entity = state.getEntityPool();
                StringPool string = state.getStringPool();
                String entityName = string.toString(entity.getEntityName(entityIndex));
                handler.endReference(entityName);
                }

            } // endEntityReference(int)

        /** Processing instruction. */
        public void processingInstruction(int targetIndex, int dataIndex)
            throws Exception
            {

            // call pi handlers
            if (piHandlers != null) {
                Enumeration handlers = piHandlers.elements();
                if (handlers.hasMoreElements()) {
                    // get strings
                    StringPool string = getParserState().getStringPool();
                    String target = string.toString(targetIndex);
                    String data   = string.toString(dataIndex);

                    // perform callbacks
                    do {
                        PIHandler handler = (PIHandler)handlers.nextElement();
                        handler.handlePI(target, data);
                        } while (handlers.hasMoreElements());
                    }
                }

            // perform default behavior
            Parser.this.processingInstruction(targetIndex, dataIndex);

            } // processingInstruction(int,int)

        /** Start element. */
        public void startElement(int elementNameIndex, int attrListIndex)
            throws Exception
            {

            if (DEBUG) {
                System.out.println("startElement(\""+getParserState().getStringPool().toString(elementNameIndex)+"\")");
                }

            // perform default behavior
            Parser.this.startElement(elementNameIndex, attrListIndex);

            // call tag handler
            TagHandler tagHandler = Parser.this.tagHandler;
            if (tagHandler != null) {
                TXElement element = Parser.this.getCurrentParent();
                tagHandler.handleStartTag(element, element.hasChildNodes());
                }

			TXNamespaceCode(Parser.this.getCurrentParent(), elementNameIndex, attrListIndex);

			} // startElement(int,int)


		/** End element. */
        public void endElement(int elementNameIndex)
            throws Exception
            {

            // call tag handler
            TagHandler tagHandler = Parser.this.tagHandler;
            if (tagHandler != null) {
                TXElement element = Parser.this.getCurrentParent();
                tagHandler.handleEndTag(element, element.hasChildNodes());
                }

            // call element handlers
            StringPool pool = getParserState().getStringPool();
            String elementName = pool.toString(elementNameIndex);
            String keys[] = { elementName, ALL_TAG_NAMES };
            Hashtable elementHandlerHash = Parser.this.elementHandlerHash;
            if (elementHandlerHash != null) {
                for (int i = 0; i < keys.length; i++) {
                    Vector handlerVector = (Vector)elementHandlerHash.get(keys[i]);
                    if (handlerVector != null) {
                        TXElement element = Parser.this.getCurrentParent();
                        Enumeration handlers = handlerVector.elements();
                        while (handlers.hasMoreElements()) {
                            ElementHandler handler = (ElementHandler)handlers.nextElement();
                            Parser.this.setRealParent(handler.handleElement(element));
                            }
                        }
                    }
                }

            // perform default behavior
            Parser.this.endElement(elementNameIndex);

            } // endElement

        //
        // ErrorHandler methods
        //

        /** Warning. */
        public void warning(SAXParseException e)
            throws SAXException
            {
            Parser.this.warnCount++;
            }

        /** Error. */
        public void error(SAXParseException e)
            throws SAXException
            {
            Parser.this.validityFailureCount++;
            if (Parser.this.isEndBy1stError) {
                throw e;
                }
            }

        /** Fatal error. */
        public void fatalError(SAXParseException e)
            throws SAXException
            {
            Parser.this.errorCount++;
            // NOTE: ending handled by continueAfterFatalError property. -Ac
            }

        //
        // EntityResolver methods
        //

        /** Resolves external entities. */
        public InputSource resolveEntity(String publicId, String systemId)
            throws SAXException, IOException
            {

            // get source from old parser's stream producer
            Source source = Parser.this.streamProducer.
                                getInputStream("", publicId, systemId);

            // create input source and set values
            InputSource inputSource = new InputSource();
            inputSource.setByteStream(source.getInputStream());
            inputSource.setCharacterStream(source.getReader());
            inputSource.setEncoding(source.getEncoding());
            inputSource.setPublicId(publicId);
            inputSource.setSystemId(systemId);

            // return new source
            return inputSource;

            } // resolveEntity(String,String):InputSource

        //
        // Helper methods
        //

        /** Expands entity references in the DOM to be text nodes. */
        /***
        // Now this is done by the DOM Parser at time of creation,
        // in the startEntityReference and endEntityReference call.
        void expandEntityReferences(TXElement el, EntityReference gr) {
            Node ch;
            while ((ch = gr.getFirstChild()) != null) {
                gr.removeChild(ch);
                if (ch instanceof EntityReference)
                    expandEntityReferences(el, (EntityReference)ch);
                else if (ch.getNodeType() == Node.TEXT_NODE)
                    el.addTextElement((TXText)ch);
                else
                    el.appendChild(ch);
                }
            }
        /***/

        } // class Handler

    } // class Parser
