/*
 * (C) Copyright IBM Corp. 1998  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.parsers;

import java.io.IOException;
import java.util.Stack;

import com.ibm.xml.framework.AttDef;
import com.ibm.xml.framework.AttrPool;
import com.ibm.xml.framework.ElementDeclPool;
import com.ibm.xml.framework.EntityPool;
import com.ibm.xml.framework.ParserState;
import com.ibm.xml.framework.StringPool;
import com.ibm.xml.framework.XMLParser;
import com.ibm.xml.internal.ErrorCode;

import com.ibm.xml.dom.DocumentImpl;
import com.ibm.xml.dom.DocumentTypeImpl;
import com.ibm.xml.dom.NodeImpl;
import com.ibm.xml.dom.EntityImpl;
import com.ibm.xml.dom.ElementDefinitionImpl;
import com.ibm.xml.dom.AttrImpl;
import com.ibm.xml.dom.TextImpl;

import com.ibm.xml.dom.DeferredDocumentImpl;

import org.w3c.dom.*;
import org.xml.sax.AttributeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * @version 63 1.23 src/com/ibm/xml/parsers/NonValidatingDOMParser.java, parser, xml4j2, xml4j2_0_13 
 */
public class NonValidatingDOMParser
    extends XMLParser
    {

    //
    // Constants
    //

    /** Represents full DOM Node expansion of Node objects and data. */
    public static final int FULL = 0;

    /** Represents deferred DOM Node expansion of Node objects and data. */
    public static final int DEFERRED = 1;

    // private

    /** Default programmatic document class. */
    public static final String DEFAULT_DOCUMENT_CLASS = DocumentImpl.class.getName();

    // debugging

    /** Set to true to debug attribute list declaration calls. */
    private static final boolean DEBUG_ATTLIST_DECL = false;

    //
    // Data
    //

    // DEFERRED Data (indexes)
    protected int fDocumentIndex = -1;
    protected int fDocumentTypeIndex = -1;
    protected int fCurrentNodeIndex = -1;

    // FULL Data
    protected DocumentType fDocumentType = null;
    //protected Node fCurrentParent = null;
    protected Node fCurrentNode = null;
    protected Stack fNodeStack = new Stack();

    // common Data
    protected Document fDocument;
    protected DocumentImpl fDocumentImpl;
    protected DeferredDocumentImpl fDeferredDocumentImpl;
    protected boolean fExpandEntityReferences = false;
    protected int fNodeExpansion = DEFERRED;
    protected boolean fWithinElement = false;

    // built-in entities
    int fAmpIndex = -1;
    int fLtIndex = -1;
    int fGtIndex = -1;
    int fAposIndex = -1;
    int fQuotIndex = -1;

    // various pool caches
    protected AttrPool        fAttrPool;
    protected ElementDeclPool fElementDeclPool;
    protected EntityPool      fEntityPool;
    protected StringPool      fStringPool;

	protected String fDocumentClass = DEFAULT_DOCUMENT_CLASS;

    // other info

    protected boolean fRootElement;

    //
    // Constructors
    //

    /** Default constructor. */
    public NonValidatingDOMParser() {

        // set handlers
        setDocumentHandler(this);
        setDocumentTypeHandler(this);

        // set default values
        fExpandEntityReferences = false;
        fNodeExpansion = DEFERRED;

        // setup parser state
        ParserState state = getParserState();
        state.useDefaultStringPool();
        state.useDefaultAttrPool();
        state.useDefaultEntityPool();
        state.useDefaultElementDeclPool();
        init(state);

        } // <init>()

    //
    // Public methods
    //

    /**
     *
     */
    public void reset() {
        super.reset();
        init(getParserState());
        }

    /**
     *
     */
    protected void resetOrCopy() {
        super.resetOrCopy();
        init(getParserState());
        }

    protected void init(ParserState state) {

        // init FULL
        fDocumentType = null;
        fCurrentNode = null;
        fNodeStack.removeAllElements();

        // init DEFERRED
        fDocumentIndex = -1;
        fDocumentTypeIndex = -1;
        fCurrentNodeIndex = -1;

        // init common
        fDocumentImpl = null;
        fDocument = null;
        fWithinElement = false;

        // save various pool references
        fAttrPool        = state.getAttrPool();
        fElementDeclPool = state.getElementDeclPool();
        fEntityPool      = state.getEntityPool();
        fStringPool      = state.getStringPool();

        fAmpIndex = fStringPool.addSymbol("amp");
        fLtIndex = fStringPool.addSymbol("lt");
        fGtIndex = fStringPool.addSymbol("gt");
        fAposIndex = fStringPool.addSymbol("apos");
        fQuotIndex = fStringPool.addSymbol("quot");

        } // init()

    /**
     *
     */
    protected void checkHandlers() throws Exception {
        super.checkHandlers();
    }

    //
    // Setters and Getters for DOM Parser behaviour modification.
    //

    /**
     * Sets whether <tt>EntityReference</tt> Nodes are expanded (replaced)
     * by this parser. Expanded means that the children of the
     * <tt>EntityReference</tt> Node replace it.
     * <p>
     * The default value is false, <tt>EntityReference</tt> Nodes are kept.
     *
     * @param keep  <code>=true</code> means expanding entity references;
     *              <code>=false</code> means maintain entity references.
     *
     * @see org.w3c.dom.EntityReference
     * @see com.ibm.xml.parser.GeneralReference
     */
    public void setExpandEntityReferences(boolean expand) {
        fExpandEntityReferences = expand;
    }

    /**
     * Return a boolean which says whether <tt>EntityReference</tt>
     * Nodes are expanded (replaced) by this parser. Expanded means
     * that the children of the <tt>EntityReference</tt> Node replace it.
     * <p>
     * The default value is false, <tt>EntityReference</tt> Nodes are kept.
     *
     * @param keep  <code>=true</code> means expanding entity references;
     *              <code>=false</code> means maintain entity references.
     *
     * @see org.w3c.dom.EntityReference
     * @see com.ibm.xml.parser.GeneralReference
     */
    public boolean getExpandEntityReferences() {
        return fExpandEntityReferences;
    }

    /**
     * Sets whether Nodes are created and the data is kept within the Nodes.
     * <p>
     * @param nodeExpansion <br>
     * NonValidatingDOMParser.FULL - full expansion
     * NonValidatingDOMParser.DEFERRED - deferred expansion
     *
     */
    public void setNodeExpansion(int nodeExpansion) {

        if (nodeExpansion != FULL && nodeExpansion != DEFERRED) {
            fNodeExpansion = DEFERRED;
            return;
        }

        fNodeExpansion = nodeExpansion;
    }

    /**
     * Return the Node expansion level integer mapping to the constants
     * in DOMParser.
     * Signifies whether DOM Nodes are created and the data is kept
     * within the Nodes.
     * <p>
     * @param nodeExpansion <br>
     * DOMParser.FULL - full expansion
     * DOMParser.DEFERRED - deferred expansion
     */
    public int getNodeExpansion() {
        return fNodeExpansion;
    }

    /** Return the current DOM Node.
     *  <p> This function will only return meaningful Node values when
     *  setNodeExpansion(FULL). When setNodeExpansion(DEFERRED) is set,
     *  Node creation is deferred and getCurrentNode() will likely return null.
     *  <p>This function is useful for querying the parser when getting an
     *  error or warning callback.
     * @see #setNodeExpansion
     */
    public Node getCurrentNode() {
        return fCurrentNode;
    }

    //
    // XMLDocumentTypeHandler interface
    //

    /**
     *  This function will be called when a &lt;!DOCTYPE...&gt; declaration is
     *  encountered.
     */
    public void doctypeDecl(int rootElementNameIndex)
        throws Exception
    {

        // full expansion
        if (fNodeExpansion == FULL) {
            if (fDocumentImpl != null) {
                String rootElementName = fStringPool.toString(rootElementNameIndex);
                fDocumentType = fDocumentImpl.createDocumentType(rootElementName);
                fDocumentImpl.appendChild(fDocumentType);
                fCurrentNode = fDocumentType;
                }
        }

        // deferred expansion
        else if (fNodeExpansion == DEFERRED) {
            fDocumentTypeIndex = fDeferredDocumentImpl.createDocumentType(rootElementNameIndex);
            fDeferredDocumentImpl.appendChild(fDocumentIndex, fDocumentTypeIndex);
            fCurrentNodeIndex = fDocumentTypeIndex;
        }
    }

    /**
     *  This function will be called at the start of the "internal subset" of a
     *  doctype declaration.
     */
    public void startInternalSubset()
        throws Exception
    {
        // this work is done in doctypeDecl(int)
        // because we know that the internal subset (if it exists)
        // is always scanned before the external subset.
    }

    /**
     *  This function will be called at the end of the "internal subset" of a
     *  doctype declaration.
     */
    public void endInternalSubset()
        throws Exception
    {
        if (fNodeExpansion == FULL) {
            fCurrentNode = fDocument;
            }
        else if (fNodeExpansion == DEFERRED) {
            fCurrentNodeIndex = fDocumentIndex;
            }
    }

    /**
     *  This function will be called at the start of the "external subset" of a
     *  doctype declaration.
     */
    public void startExternalSubset(int publicIdStringIndex, int systemIdStringIndex)
        throws Exception
    {
        if (fNodeExpansion == FULL) {
            fCurrentNode = fDocumentType;
            }
        else if (fNodeExpansion == DEFERRED) {
            fCurrentNodeIndex = fDocumentTypeIndex;
            }
    }

    /**
     *  This function will be called at the end of the "external subset" of a
     *  doctype declaration.
     */
    public void endExternalSubset()
        throws Exception
    {
        if (fNodeExpansion == FULL) {
            fCurrentNode = fDocument;
            }
        else if (fNodeExpansion == DEFERRED) {
            fCurrentNodeIndex = fDocumentIndex;
            }
    }

    /**
     * &lt;!ELEMENT Name contentspec&gt;
     */
    public void elementDecl(int elementIndex)
        throws Exception {

        if (DEBUG_ATTLIST_DECL) {
            System.out.println("elementDecl("+
                               elementIndex+", "+
                               fStringPool.toString(fElementDeclPool.getElementName(elementIndex))+
                               ")");
        }

    } // elementDecl(int)

    /**
     * &lt;!ATTLIST Name AttDef&gt;
     */
    public void attlistDecl(int elementIndex, int attDefIndex)
        throws Exception {

        if (DEBUG_ATTLIST_DECL) {
            System.out.println("attlistDecl("+
                               elementIndex+", "+
                               fStringPool.toString(fElementDeclPool.getElementName(elementIndex))+", "+
                               attDefIndex+", "+
                               fStringPool.toString(fElementDeclPool.getAttName(attDefIndex))+
                               ")");
        }

        if (fNodeExpansion == FULL) {
            if (fDocumentImpl != null) {

                // get the default value
                int attrValueIndex = fElementDeclPool.getAttValue(attDefIndex);
                if (attrValueIndex != -1) {
                    if (DEBUG_ATTLIST_DECL) {
                        System.out.println("  adding default attribute value: "+
                                           fStringPool.toString(attrValueIndex));
                    }

                    // get element name index
                    int elementNameIndex = fElementDeclPool.getElementName(elementIndex);
                    String elementName = fStringPool.toString(elementNameIndex);

                    // get element definition node
                    NamedNodeMap elements = ((DocumentTypeImpl)fDocumentType).getElements();
                    ElementDefinitionImpl elementDef = (ElementDefinitionImpl)elements.getNamedItem(elementName);
                    if (elementDef == null) {
                        elementDef = fDocumentImpl.createElementDefinition(elementName);
                        ((DocumentTypeImpl)fDocumentType).getElements().setNamedItem(elementDef);
                    }

                    // REVISIT: Check for uniqueness of element name? -Ac

                    // get attribute name and value index
                    int    attrNameIndex = fElementDeclPool.getAttName(attDefIndex);
                    String attrName      = fStringPool.toString(attrNameIndex);
                    String attrValue     = fStringPool.toString(attrValueIndex);

                    // create attribute and set properties
                    AttrImpl attr = (AttrImpl)fDocumentImpl.createAttribute(attrName);
                    attr.setValue(attrValue);
                    attr.setSpecified(false);

                    // add default attribute to element definition
                    elementDef.getAttributes().setNamedItem(attr);
                }
            }

        } // if FULL

        else if (fNodeExpansion == DEFERRED) {

            // get the default value
            int attrValueIndex = fElementDeclPool.getAttValue(attDefIndex);
            if (attrValueIndex != -1) {
                if (DEBUG_ATTLIST_DECL) {
                    System.out.println("  adding default attribute value: "+
                                       fStringPool.toString(attrValueIndex));
                }

                // get element definition
                int elementNameIndex = fElementDeclPool.getElementName(elementIndex);
                int elementDefIndex  = fDeferredDocumentImpl.lookupElementDefinition(elementNameIndex);

                // create element definition if not already there
                if (elementDefIndex == -1) {
                    elementDefIndex = fDeferredDocumentImpl.createElementDefinition(elementNameIndex);
                    fDeferredDocumentImpl.appendChild(fDocumentTypeIndex, elementDefIndex);
                }

                // add default attribute
                int attrNameIndex = fElementDeclPool.getAttName(attDefIndex);
                int attrIndex     = fDeferredDocumentImpl.createAttribute(attrNameIndex,
                                                                  attrValueIndex,
                                                                  false);
                fDeferredDocumentImpl.appendChild(elementDefIndex, attrIndex);
            }

        } // else DEFERRED

    } // attrlistDecl(int,int)

    /**
     * &lt;!ENTITY Name EntityValue&gt; (internal)
     */
    public void internalEntityDecl(int entityIndex)
        throws Exception
    {
        int entityNameIndex  = fEntityPool.getEntityName(entityIndex);

        // full expansion
        if (fNodeExpansion == FULL) {
            if (fDocumentImpl != null) {
                if (fDocumentType == null) return; //revist: should never happen. Exception?

                //revist: how to check if entity was already declared.
                // XML spec says that 1st Entity decl is binding.

                Entity entity = fDocumentImpl.createEntity(fStringPool.toString(entityNameIndex));
                fDocumentType.getEntities().setNamedItem(entity);
            }

        }

        // deferred expansion
        else if (fNodeExpansion == DEFERRED) {

            if (fDocumentTypeIndex == -1) return; //revist: should never happen. Exception?

            //revist: how to check if entity was already declared.
            // XML spec says that 1st Entity decl is binding.

            int newEntityIndex = fDeferredDocumentImpl.createEntity(entityIndex);
            fDeferredDocumentImpl.appendChild(fDocumentTypeIndex, newEntityIndex);
        }
    }

    /**
     * &lt;!ENTITY Name ExternalID>                (external)
     */
    public void externalEntityDecl(int entityIndex) throws Exception
    {
        int entityNameIndex  = fEntityPool.getEntityName(entityIndex);
        int publicIdIndex  = fEntityPool.getPublicId(entityIndex);
        int systemIdIndex  = fEntityPool.getSystemId(entityIndex);

        // full expansion
        if (fNodeExpansion == FULL) {
            if (fDocumentImpl != null) {
                if (fDocumentType == null) return; //revist: should never happen. Exception?

                //revist: how to check if entity was already declared.
                // XML spec says that 1st Entity decl is binding.

                EntityImpl entity = (EntityImpl)fDocumentImpl.createEntity(fStringPool.toString(entityNameIndex));
                entity.setPublicId(fStringPool.toString(publicIdIndex));
                entity.setSystemId(fStringPool.toString(systemIdIndex));
                fDocumentType.getEntities().setNamedItem(entity);
            }

        }

        // deferred expansion
        else if (fNodeExpansion == DEFERRED) {

            if (fDocumentTypeIndex == -1) return; //revist: should never happen. Exception?

            //revist: how to check if entity was already declared.
            // XML spec says that 1st Entity decl is binding.

            int newEntityIndex = fDeferredDocumentImpl.createEntity(entityIndex);

            fDeferredDocumentImpl.appendChild(fDocumentTypeIndex, newEntityIndex);
        }
    }

    /**
     * &lt;!ENTITY Name ExternalID NDataDecl>      (unparsed)
     */
    public void unparsedEntityDecl(int entityIndex) throws Exception
    {
        int entityNameIndex  = fEntityPool.getEntityName(entityIndex);
        int publicIdIndex  = fEntityPool.getPublicId(entityIndex);
        int systemIdIndex  = fEntityPool.getSystemId(entityIndex);
        int notationNameIndex  = fEntityPool.getNotationName(entityIndex);

        // full expansion
        if (fNodeExpansion == FULL) {
            if (fDocumentImpl != null) {
                if (fDocumentType == null) return; //revist: should never happen. Exception?

                //revist: how to check if entity was already declared.
                // XML spec says that 1st Entity decl is binding.

                EntityImpl entity = (EntityImpl)fDocumentImpl.createEntity(fStringPool.toString(entityNameIndex));
                entity.setPublicId(fStringPool.toString(publicIdIndex));
                entity.setSystemId(fStringPool.toString(systemIdIndex));
                entity.setNotationName(fStringPool.toString(notationNameIndex));
                fDocumentType.getEntities().setNamedItem(entity);
            }

        }

        // deferred expansion
        else if (fNodeExpansion == DEFERRED) {

            if (fDocumentTypeIndex == -1) return; //revist: should never happen. Exception?

            //revist: how to check if entity was already declared.
            // XML spec says that 1st Entity decl is binding.

            int newEntityIndex = fDeferredDocumentImpl.createEntity(entityIndex);

            fDeferredDocumentImpl.appendChild(fDocumentTypeIndex, newEntityIndex);
        }
    }

    /**
     * &lt;!NOTATION Name ExternalId>
     */
    public void notationDecl(int notationIndex)
        throws Exception
    {
        int notationNameIndex  = fEntityPool.getEntityName(notationIndex);

        // full expansion
        if (fNodeExpansion == FULL) {
            if (fDocumentImpl != null) {
                if (fDocumentType == null) return; //revist: should never happen. Exception?

                //revist: how to check if entity was already declared.
                // XML spec says that 1st Entity decl is binding.

                Notation notation = fDocumentImpl.createNotation(fStringPool.toString(notationNameIndex));

                fDocumentType.getNotations().setNamedItem(notation);
            }

        }

        // deferred expansion
        else if (fNodeExpansion == DEFERRED) {

            if (fDocumentTypeIndex == -1) return; //revist: should never happen. Exception?

            //revist: how to check if entity was already declared.
            // XML spec says that 1st Entity decl is binding.

            int newNotationIndex = fDeferredDocumentImpl.createNotation(notationIndex);

            fDeferredDocumentImpl.appendChild(fDocumentTypeIndex, newNotationIndex);
        }

    }

    //
    // XMLDocumentHandler interface
    //

    /**
     * This allows the programmer to decide which document factory
     * to use when constructing the DOM tree. However, doing so
     * will lose the parsing performance gain of the default factory.
     * <p>
     * <b>Note:</b> Setting the document class name to any other
     * document class besides the default class will switch the
     * node expansion to FULL.
     *
     * @param documentClass The fully qualified class name of the
     *                      document factory to use when constructing
     *                      the DOM tree.
     *
     * @see #DEFERRED_DOCUMENT_CLASS
     * @see #setNodeExpansion
     * @see #FULL
     */
    public void setDocumentClass(String documentClass) {
        if (documentClass == null) {
            documentClass = DEFAULT_DOCUMENT_CLASS;
        }
        fDocumentClass = documentClass;
        if (!documentClass.equals(DEFAULT_DOCUMENT_CLASS)) {
            setNodeExpansion(FULL);
        }
    }

	/** Returns the document. */
    public Document getDocument() {
        return fDocument;
    }

    /** Start document. */
    public void startDocument(int versionIndex,
                              int encodingIndex,
                              int standAloneIndex)
    {

        // clean up unused strings
        if (versionIndex != -1) {
            fStringPool.orphanString(versionIndex);
            }
        if (encodingIndex != -1) {
            fStringPool.orphanString(encodingIndex);
            }
        if (standAloneIndex != -1) {
            fStringPool.orphanString(standAloneIndex);
            }

        // full expansion
        if (fNodeExpansion == FULL) {
            try {
                fDocument = (Document)Class.forName(fDocumentClass).newInstance();
                if (fDocument instanceof DocumentImpl) {
                    fDocumentImpl = (DocumentImpl)fDocument;
                }
            }
            catch (Exception e) {
                // REVISIT: What should we do if this is a bad factory? -Ac
            }

            fCurrentNode = fDocument;
        }

        // deferred expansion
        else if (fNodeExpansion == DEFERRED) {
            fDocumentImpl = fDeferredDocumentImpl = new DeferredDocumentImpl(getParserState());
            fDocument = fDocumentImpl;

            fDocumentIndex = fDeferredDocumentImpl.createDocument();
            fCurrentNodeIndex = fDocumentIndex;
        }

    } // startDocument()

    /** End document. */
    public void endDocument() throws Exception {
    }

    /** Start element. */
    public void startElement(int elementNameIndex, int attrListIndex)
        throws Exception {

        fRootElement = true;

        // full expansion
        if (fNodeExpansion == FULL) {

            fNodeStack.push(fCurrentNode);

            String elementName = fStringPool.toString(elementNameIndex);
            AttributeList attrList = fAttrPool.getAttributeList(attrListIndex);
            Element e = fDocument.createElement(elementName);
            int attrListLength = attrList.getLength();
            for (int i = 0; i < attrListLength; i++)
                e.setAttribute(attrList.getName(i), attrList.getValue(i));
            fCurrentNode.appendChild(e);
            fCurrentNode = e;
            fWithinElement = true;

            // identifier registration
            if (fDocumentImpl != null) {
                int index = fAttrPool.getFirstAttr(attrListIndex);
                while (index != -1) {
                    if (fAttrPool.getAttType(index) == AttDef.ID) {
                        String name = fStringPool.toString(fAttrPool.getAttValue(index));
                        fDocumentImpl.putIdentifier(name, e);
                    }
                    index = fAttrPool.getNextAttr(index);
                }
            }

            fAttrPool.releaseAttrList(attrListIndex);
        }

        // deferred expansion
        else if (fNodeExpansion == DEFERRED) {
            push(fCurrentNodeIndex); // push the parent of the Element.

            int element = fDeferredDocumentImpl.createElement(elementNameIndex, attrListIndex);
            fDeferredDocumentImpl.appendChild(fCurrentNodeIndex, element);
            fCurrentNodeIndex = element;
            fWithinElement = true;

            // identifier registration
            int index = fAttrPool.getFirstAttr(attrListIndex);
            while (index != -1) {
                if (fAttrPool.getAttType(index) == AttDef.ID) {
                    int nameIndex = fAttrPool.getAttValue(index);
                    fDeferredDocumentImpl.putIdentifier(nameIndex, element);
                }
                index = fAttrPool.getNextAttr(index);
            }

        }

    } // startElement(int,int)

    /** End element. */
    public void endElement(int elementNameIndex)
        throws Exception
    {

        if (fNodeExpansion == FULL) {
            fCurrentNode = (Node)fNodeStack.pop();
            fWithinElement = false;
        } else
        if (fNodeExpansion == DEFERRED) {

            fCurrentNodeIndex = pop(); // the parent of the element.
            fWithinElement = false;
        }

    } // endElement(int)

    /** Start entity reference. */
    public void startEntityReference(int entityIndex) throws Exception {

        if (fExpandEntityReferences) {
            return;
        }

        int entityNameIndex = fEntityPool.getEntityName(entityIndex);
        if (entityNameIndex == fAmpIndex
        ||  entityNameIndex == fGtIndex
        ||  entityNameIndex == fLtIndex
        ||  entityNameIndex == fAposIndex
        ||  entityNameIndex == fQuotIndex) return;


        if (fNodeExpansion == FULL) {

            fNodeStack.push(fCurrentNode); //push ER's parent
            EntityReference er = fDocument.createEntityReference(fStringPool.toString(entityNameIndex));

            if (er==null) return;

            fNodeStack.push(er); // push the ER

            fCurrentNode.appendChild(er);
            fCurrentNode = er;
        } else
        if (fNodeExpansion == DEFERRED) {
            push(fCurrentNodeIndex); //push ER's parent

            int entityRefIndex = fDeferredDocumentImpl.createEntityReference(entityIndex);
            push(entityRefIndex); // push the ER
            fDeferredDocumentImpl.appendChild(fCurrentNodeIndex, entityRefIndex);

            fCurrentNodeIndex = entityRefIndex;
        }
    }

    /** End entity reference. */
    public void endEntityReference(int entityIndex) throws Exception {

        if (fExpandEntityReferences) {
            return;
        }

        int entityNameIndex = fEntityPool.getEntityName(entityIndex);
        if (entityNameIndex == fAmpIndex
        ||  entityNameIndex == fGtIndex
        ||  entityNameIndex == fLtIndex
        ||  entityNameIndex == fAposIndex
        ||  entityNameIndex == fQuotIndex) return;


        if (fNodeExpansion == FULL) {

            if (fDocumentImpl != null) {

                NodeImpl erNode = (NodeImpl)fNodeStack.pop(); // pop the ER
                fCurrentNode = (NodeImpl)fNodeStack.pop(); // pop the ER's parent

                // should never be true - we should not return here.
                if (erNode.getNodeType() != Node.ENTITY_REFERENCE_NODE)  return;

                NamedNodeMap entities = fDocumentType.getEntities();
                String name = fStringPool.toString(entityNameIndex);
                Node entityNode = entities.getNamedItem(name);

                // We could simply return here if there is no entity for the reference.
                if (entityNode == null || entityNode.hasChildNodes())  return;

                Entity entity = (Entity)entityNode;

                if (erNode.hasChildNodes()) {
                    NodeList list = erNode.getChildNodes();
                    int len = list.getLength();
                    for (int i = 0; i < len; i++) {
                        Node childClone = list.item(i).cloneNode(true);
                        entity.appendChild(childClone);
                    }
                }
        }

        } else
        if (fNodeExpansion == DEFERRED) {

            String name = fStringPool.toString(entityNameIndex);

            int erChild = pop(); // pop the ER
            fCurrentNodeIndex = pop(); // pop the ER's parent

            // should never be true - we should not return here.
            if (fDeferredDocumentImpl.getNodeType(erChild) != Node.ENTITY_REFERENCE_NODE)  return;

            erChild = fDeferredDocumentImpl.getFirstChild(erChild); // first Child of EntityReference

            if (fDocumentTypeIndex != -1) {
                //find Entity decl for this EntityReference.
                int entityDecl = fDeferredDocumentImpl.getFirstChild(fDocumentTypeIndex);
                while (entityDecl != -1) {
                    if (fDeferredDocumentImpl.getNodeType(entityDecl) == Node.ENTITY_NODE
                    //&& fDocument.getNodeName(entityDecl) != null // this should NOT return null!...
                    && fDeferredDocumentImpl.getNodeNameString(entityDecl).equals(name)) // string compare...
                    {
                        break;
                    }
                    entityDecl = fDeferredDocumentImpl.getNextSibling(entityDecl);
                }

                if (entityDecl != -1
                    && fDeferredDocumentImpl.getFirstChild(entityDecl) == -1) {
                    // found entityDecl with same name as this reference
                    // AND it doesn't have any children.

                    // we don't need to iterate, because the whole structure
                    // should already be connected to the 1st child.
                    fDeferredDocumentImpl.setAsFirstChild(entityDecl, erChild);
                }
            }

        }
    }

    /** Characters. */
    public void characters(int dataIndex, boolean cdataSection)
        throws Exception
        {

        // The text should already be normalized by the scanner.
        // The only known case requiring normalization is when there is
        // an EntityReference containing Text Node, which is next to Text Nodes,
        // and setExpandEntities is true, so the Entity Reference is not created.
        if (fNodeExpansion == FULL) {

            Text t = null;
            if (cdataSection) {
                t = fDocument.createCDATASection(fStringPool.orphanString(dataIndex));
            } else {
                Node lastChild = null;
                if (
                    fWithinElement
                    && fCurrentNode.getNodeType() == Node.ELEMENT_NODE
                ) {
                    lastChild = fCurrentNode.getLastChild();
                    if (lastChild != null
                        && lastChild.getNodeType() == Node.TEXT_NODE) {
                        // Normalization of Text Nodes - append rather than create.
                        ((Text)lastChild).appendData(fStringPool.orphanString(dataIndex));
                        return;
                    }
                }
                t = fDocument.createTextNode(fStringPool.orphanString(dataIndex));
            }

            if (t==null) return;

            fCurrentNode.appendChild(t);

        } else
        if (fNodeExpansion == DEFERRED) {

            // The Text normalization is taken care of within the Text Node
            // in the DEFERRED case.
            int text = cdataSection ?
                            fDeferredDocumentImpl.createCDATASection(dataIndex) :
                            fDeferredDocumentImpl.createTextNode(dataIndex, false);
            fDeferredDocumentImpl.appendChild(fCurrentNodeIndex, text);
        }

    }

    /** Ignorable whitespace. */
    public void ignorableWhitespace(int dataIndex, boolean cdataSection)
        throws Exception
    {

        if (fNodeExpansion == FULL) {
            Text t = null;
            if (cdataSection) {
                t = fDocument.createCDATASection(fStringPool.orphanString(dataIndex));
            } else {
                t = fDocument.createTextNode(fStringPool.orphanString(dataIndex));
            }

            if (t==null) return;

            if (fDocumentImpl != null) {
                ((TextImpl)t).setIgnorableWhitespace(true);
            }
            fCurrentNode.appendChild(t);

        } else
        if (fNodeExpansion == DEFERRED) {

            int text = cdataSection ?
                            fDeferredDocumentImpl.createCDATASection(dataIndex) :
                            fDeferredDocumentImpl.createTextNode(dataIndex, true);
            fDeferredDocumentImpl.appendChild(fCurrentNodeIndex, text);
        }

    } // ignorableWhitespace(int,boolean)

    /** Processing instruction. */
    public void processingInstruction(int targetIndex, int dataIndex)
        throws Exception
    {
        if (fNodeExpansion == FULL) {

            String target = fStringPool.orphanString(targetIndex);
            String data = fStringPool.orphanString(dataIndex);
            ProcessingInstruction pi = fDocument.createProcessingInstruction(target, data);
            fCurrentNode.appendChild(pi);
        } else
        if (fNodeExpansion == DEFERRED) {

            int pi = fDeferredDocumentImpl.createProcessingInstruction(targetIndex, dataIndex);
            fDeferredDocumentImpl.appendChild(fCurrentNodeIndex, pi);
        }

    } // processingInstruction(int,int)

    /** Comment. */
    public void comment(int dataIndex)
        throws Exception
    {

        if (fNodeExpansion == FULL) {
            Comment c = fDocument.createComment(fStringPool.orphanString(dataIndex));
            fCurrentNode.appendChild(c);
        } else
        if (fNodeExpansion == DEFERRED) {

            int comment = fDeferredDocumentImpl.createComment(dataIndex);
            fDeferredDocumentImpl.appendChild(fCurrentNodeIndex, comment);
        }

    } // comment(int)

    /**
     * Not called.
     */
    public void characters(char ch[], int start, int length, boolean cdataSection)
        throws Exception {
    }

    /**
     * Not called.
     */
    public void ignorableWhitespace(char ch[], int start, int length, boolean cdataSection)
        throws Exception {
    }

    private int nodestack[] = new int[16];
    private int stacktop = 0;
    private void push(int index) {
        if (stacktop == nodestack.length) {
            int newstack[] = new int[stacktop * 2];
            System.arraycopy(nodestack, 0, newstack, 0, stacktop);
            nodestack = newstack;
        }
        nodestack[stacktop++] = index;
    }
    private int pop() {
        return nodestack[--stacktop];
    }

    //
    // XMLParser methods
    //

    /**
     * Parses the specified input source.
     *
     * @param source The input source.
     *
     * @exception org.xml.sax.SAXException Throws exception on SAX error.
     * @exception java.io.IOException Throws exception on i/o error.
     */
    public void parse(InputSource source) throws SAXException, IOException {

        // parse
        fRootElement = false;
        super.parse(source);

        // no root element
        if (!fRootElement) {
            try {
                error(ErrorCode.E_STRUCT2);
            }
            catch (Exception e) {
                throw new SAXException(e);
            }
        }

    } // parse(InputSource)

} // class NonValidatingDOMParser
