/*
 * (C) Copyright IBM Corp. 1997-1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.parsers;

import com.ibm.xml.dom.TextImpl;
import com.ibm.xml.framework.ElementDeclPool;
import com.ibm.xml.framework.ParserState;
import com.ibm.xml.framework.StringPool;
import com.ibm.xml.framework.XMLValidationHandler;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * This class is a validating DOM parser which can also
 * "revalidate" a DOM subtree after the document has been parsed.
 *
 * @version Revision: 16 1.4 src/com/ibm/xml/parsers/RevalidatingDOMParser.java, parser, xml4j2, xml4j2_0_13 
 * @see com.ibm.xml.parsers.DOMParser
 */
public class RevalidatingDOMParser 
    extends DOMParser {
	
    //
    // Constants
    //

    // debugging

    /** Set to true to debug validate() method. */
    private static final boolean DEBUG_VALIDATE = false;

    //
    // Public methods
    //

	/**
	 * Given a node which is a DOM tree (or subtree), validate the tree 
     * against the current DTD.  If the tree is valid, return null.  If
     * the subtree is invalid return the highest (closest to the root) 
     * and left most node which causes the tree to be invalid.
	 *
	 * @param node  The node representing the root of the DOM subtree to be
     *              revalidated.
     *
     * @return <i>null</i> if the subtree is valid or the first (highest and 
     *         leftmost) node where the tree is invalid.
	 */
    public final Node validate(Node node) {

		if (node.getNodeType() != Node.ELEMENT_NODE) {
			throw new IllegalArgumentException("Can't revalidate a non element");
		}

        // get handler and pool
        XMLValidationHandler validator       = getValidationHandler();
        ParserState          state           = getParserState();
        StringPool           stringPool      = state.getStringPool();
        ElementDeclPool      elementDeclPool = state.getElementDeclPool();
    
        // perform validation
        return recursiveValidate(node, validator, stringPool, elementDeclPool);

    } // validate(Node):Node

    //
    // Private methods
    //

    /** 
     * Recursive validation. This method returns the first node that is
     * in error, or <i>null</i> if the content is valid.
     *
     * @param node              The node to validate.
     * @param validator         The validation handler.
     * @param stringPool        The string pool
     * @param elementDeclPool   The element declaration pool.
     */
    private final Node recursiveValidate(Node node,
                                         XMLValidationHandler validator,
                                         StringPool stringPool,
                                         ElementDeclPool elementDeclPool) {

        // debugging output
        if (DEBUG_VALIDATE) {
            print(node, "");
        }

        // build up child array to pass to validator
        int  children[] = new int[countChildren(node)];
        Node nodes[]    = new Node[children.length];
        int  count      = expandChildren(node, children, nodes, 0, stringPool);

        // validate this node
        int result = -1;
        int parentNameIndex = stringPool.addSymbol(node.getNodeName());
        int parentIndex     = elementDeclPool.getElement(parentNameIndex); 
        try {
            result = validator.checkContent(parentIndex, count, children);
        } 
        catch (Exception e) {
            // the default validation handler doesn't throw exceptions
            if (DEBUG_VALIDATE) {
                e.printStackTrace();
            }
        }

        // did an error occur?
        Node errorNode = null;
        if (result != -1) {
            // set node where error occured
            errorNode = nodes[result];

            // perform error handler callback
            String cs  = elementDeclPool.getContentSpecAsString(parentIndex);
            int    csi = stringPool.addString(cs);
            int errorCode = result != count ? 
                            com.ibm.xml.internal.ErrorCode.V_CONT1 :
                            com.ibm.xml.internal.ErrorCode.V_CONT2 ;
            try {
                getParserState().getErrorHandler()
                    .error2(errorCode, parentNameIndex, csi);
            }
            catch (Exception e) {
                // ignore
            }
        }

        // validate children
        else {
            for (int i = 0; i < count; i++) {
                if (children[i] != -1) {
                    Node child = nodes[i];
                    if (child != null) {
                        child = recursiveValidate(child, 
                                                  validator, 
                                                  stringPool, 
                                                  elementDeclPool);
                        if (child != null) {
                            errorNode = child;
                            break;
                        }
                    }
                }
            }
        }

        // explicitly null out references
        children = null;
        nodes    = null;

        // return success
        return errorNode;
    
    } // recursiveValidate(Node,XMLValidationHandler,StringPool,ElementDeclPool):Node

    /**
     * Expands the children of the specified node into the given
     * integer array (internal indices) and Node array. 
     * <p>
     * This method returns the number of children that were actually
     * expanded since text nodes may be ignorable whitespace.
     *
     * @param node      The node for which to expand the children.
     * @param children  The destination array of the internal indices
     *                  representing the children of the specified node.
     *                  A value of -1 indicates text.
     * @param nodes     The destination array of node objects representing
     *                  the children.
     * @param count     The number of children. Since this method is
     *                  called recursively in the case of entity
     *                  references, this is the offset of where to begin
     *                  adding children. The first call to this method
     *                  passes 0 as the initial count.
     * @param stringPool The string pool where element name symbols are
     *                  stored.
     */
    private final int expandChildren(Node node, 
                                     int children[], Node nodes[], int count,
                                     StringPool stringPool) {

        // fill in children array
        if (node.hasChildNodes()) {

            NodeList childList  = node.getChildNodes();
            int      childCount = childList.getLength();
            for (int i = 0; i < childCount; i++) {
                Node child = childList.item(i);
                int  type  = child.getNodeType();

                // element
                if (type == Node.ELEMENT_NODE) {
                    children[count] = stringPool.addSymbol(child.getNodeName());
                    nodes[count]    = child;
                    count++;
                } 

                // not ignorable text
                else if (type == Node.TEXT_NODE) {
                    if (!(child instanceof TextImpl) || 
                        !((TextImpl)child).isIgnorableWhitespace()) { 
                        children[count] = -1;
                        nodes[count]    = child;
                        count++;
                    }
                }

                // entity reference node
                else if (type == Node.ENTITY_REFERENCE_NODE) {
                    count = expandChildren(child, 
                                           children, nodes, count, 
                                           stringPool);
                }

            } // for children

        } // if has children

        // return
        return count;

    } // expandChildren(Node,int[],Node[],int,StringPool):int

    /**
     * Returns the <i>true</i> number of children under the specified
     * node. When entity references occur as direct children, a
     * recursive call to this method is made in order to count those
     * nodes as children of the original node specified.
     *
     * @param node  The node whose children are to be counted.
     */
    private final int countChildren(Node node) {

        // count the children
        int count = 0;
        if (node.hasChildNodes()) {

            // count initial children
            NodeList children = node.getChildNodes();
            count += children.getLength();

            // handle entity references as special case
            //
            // Note: The variable "i" is decremented in this loop
            //       to avoid the use of an extraneous variable to
            //       hold the original child count. Since we're
            //       just counting the number of children, this is
            //       acceptable. -Ac
            for (int i = count - 1; i >= 0; i--) {
                Node child = children.item(i);
                if (child.getNodeType() == Node.ENTITY_REFERENCE_NODE) {
                    // The -1 is to remove the entity reference node. -Ac
                    count += countChildren(child) - 1;
                }
            }
        }

        // return
        return count;

    } // countChildren(Node):int

    //
    // Debugging methods
    //

    /** Debugging print method. */
    private static void print(Node node, String indent) {

        if (DEBUG_VALIDATE) {
            if (node == null) {
                System.out.println(indent+"!!! node == null");
                return;
            }
            System.out.println(indent+
                               "node: type="+type(node.getNodeType())+
                               ", name="+node.getNodeName()+
                               ", value=\""+normalize(node.getNodeValue())+'"');
            if (node.hasChildNodes()) {
                System.out.println(indent+'{');
                NodeList children = node.getChildNodes();
                int count = children.getLength();
                for (int i = 0; i < count; i++) {
                    print(children.item(i), indent+"  ");
                }
                System.out.println(indent+'}');
            }
        }

    } // print(Node,String)

    /** Debugging normalizing text. */
    private static String normalize(String s) {

        if (DEBUG_VALIDATE) {
            if (s == null) {
                return "[null]";
            }
            StringBuffer str = new StringBuffer();
            int len = s.length();
            for (int i = 0; i < len; i++) {
                if (i > 20) {
                    str.append("...");
                    break;
                }
                char ch = s.charAt(i);
                switch (ch) {
                    case '\r': str.append("[\\r]"); break;
                    case '\n': str.append("[\\n]"); break;
                    default: str.append(ch);
                }
            }
            return str.toString();
        }

        return null;

    } // normalize(String):String

    /** Debugging node type name. */
    private static String type(int type) {

        if (DEBUG_VALIDATE) {
            switch (type) {
                case Node.ATTRIBUTE_NODE: return "ATTR";
                case Node.ELEMENT_NODE: return "ELEMENT";
                case Node.TEXT_NODE: return "TEXT";
                case Node.ENTITY_REFERENCE_NODE: return "ENTITY_REF";
            }
            return "???";
        }

        return null;

    } // type(int):String

} // class RevalidatingDOMParser
