/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.parsers;

import java.io.IOException;

import com.ibm.xml.framework.*;
import com.ibm.xml.internal.ErrorCode;

import org.xml.sax.*;

/**
 * Non-validating SAX Parser.
 *
 * @version Revision: 65 1.15 src/com/ibm/xml/parsers/SAXParser.java, parser, xml4j2, xml4j2_0_13 
 */
public class SAXParser
    extends XMLParser {

    //
    // Data
    //

    // handlers

    private DTDHandler fDTDHandler = null;
    private DocumentHandler fDocumentHandler = null;

    // other info

    protected boolean fRootElement;

    //
    // Constructors
    //

    /** Default constructor. */
    public SAXParser() {

        // use default pools
        ParserState state = getParserState();
        state.useDefaultStringPool();
        state.useDefaultAttrPool();
        state.useDefaultEntityPool();
        state.useDefaultElementDeclPool();

        // set handlers
        setDocumentHandler(this);
        setDocumentTypeHandler(this);

    } // <init>()

    //
    // Parser methods
    //

    /** Sets the entity resolver. */
    public void setEntityResolver(EntityResolver resolver) {
        getEntityHandler().setEntityResolver(resolver);
    }

    /** Sets the DTD handler. */
    public void setDTDHandler(DTDHandler handler) {
        fDTDHandler = handler;
        //setDocumentTypeHandler(handler == null ? null : this);
    }

    /** Sets the document handler. */
    public void setDocumentHandler(DocumentHandler handler) {
        fDocumentHandler = handler;
        //setDocumentHandler(handler == null ? null : this);
    }

    //
    // XMLDocumentTypeHandler interface
    //

    /**
     * This function will be called when a &lt;!DOCTYPE...&gt; declaration is
     * encountered.
     */
    public void doctypeDecl(int rootElementNameIndex) throws Exception {}

    /**
     * This function will be called at the start of the "internal subset" of a
     * doctype declaration.
     */
    public void startInternalSubset() throws Exception {}

    /**
     * This function will be called at the end of the "internal subset" of a
     * doctype declaration.
     */
    public void endInternalSubset() throws Exception {}

    /**
     * This function will be called at the start of the "external subset" of a
     * doctype declaration.
     */
    public void startExternalSubset(int publicIdStringIndex, int systemIdStringIndex) throws Exception {}

    /**
     * This function will be called at the end of the "external subset" of a
     * doctype declaration.
     */
    public void endExternalSubset() throws Exception {}

    /** Element decl. */
    public void elementDecl(int elementIndex) throws Exception {}

    /** Attlist decl. */
    public void attlistDecl(int elementIndex, int attDefIndex) throws Exception {}

    /** Internal entity decl. */
    public void internalEntityDecl(int entityIndex) throws Exception {}

    /** External entity decl. */
    public void externalEntityDecl(int entityIndex) throws Exception {}

    /** Unparsed entity decl. */
    public void unparsedEntityDecl(int entityIndex) throws Exception {

        if (fDTDHandler != null) {

            // pools
            ParserState state = getParserState();
            EntityPool entityPool = state.getEntityPool();
            StringPool stringPool = state.getStringPool();

            // strings
            String name = stringPool.toString(entityPool.getEntityName(entityIndex));
            String publicId = stringPool.toString(entityPool.getPublicId(entityIndex));
            String systemId = stringPool.toString(entityPool.getSystemId(entityIndex));
            String notation = stringPool.toString(entityPool.getNotationName(entityIndex));

            // perform callback
            fDTDHandler.unparsedEntityDecl(name, publicId, systemId, notation);
        }

    } // unparsedEntityDecl(int)

    /** Notation decl. */
    public void notationDecl(int notationIndex) throws Exception {

        if (fDTDHandler != null) {

            // pools
            ParserState state = getParserState();
            EntityPool entityPool = state.getEntityPool();
            StringPool stringPool = state.getStringPool();

            // strings
            String name = stringPool.toString(entityPool.getNotationName(notationIndex));
            String publicId = stringPool.toString(entityPool.getPublicId(notationIndex));
            String systemId = stringPool.toString(entityPool.getSystemId(notationIndex));

            // perform callback
            fDTDHandler.notationDecl(name, publicId, systemId);
        }

    } // notationDecl(int)

    //
    // XMLDocumentHandler interface
    //

    /** Send char data as array. */
    public boolean sendCharDataAsCharArray() {
        return true;
    }

    /** Start document. */
    public void startDocument(int versionIndex,
                              int encodingIndex,
                              int standAloneIndex)
        throws Exception {

        // throw away strings
        StringPool  strPool = getParserState().getStringPool();
        strPool.orphanString(versionIndex);
        strPool.orphanString(encodingIndex);
        strPool.orphanString(standAloneIndex);

        // perform callbacks
        if (fDocumentHandler != null) {
            fDocumentHandler.setDocumentLocator(getLocator());
            fDocumentHandler.startDocument();
        }

    } // startDocument(int,int,int)

    /** End document. */
    public void endDocument() throws Exception {

        // perform callback
        if (fDocumentHandler != null) {
            fDocumentHandler.endDocument();
        }

    } // endDocument()

    /** Start element */
    public void startElement(int elementNameIndex, int attrListIndex)
        throws Exception {

        // we know that we've seen the root element
        fRootElement = true;

        // perform callback
        if (fDocumentHandler != null) {
            ParserState state = getParserState();
            fDocumentHandler.startElement(state.getStringPool().toString(elementNameIndex),
                                          state.getAttrPool().getAttributeList(attrListIndex));
        }

        // free attribute list
        getParserState().getAttrPool().releaseAttrList(attrListIndex);

    } // startElement(int,int)

    /** End element. */
    public void endElement(int elementNameIndex) throws Exception {

        // perform callback
        if (fDocumentHandler != null) {
            fDocumentHandler.endElement(getParserState().getStringPool().toString(elementNameIndex));
        }

    } // endElement(int)

    /** Start entity reference. */
    public void startEntityReference(int entityIndex) throws Exception {}

    /** End entity reference. */
    public void endEntityReference(int entityIndex) throws Exception {}

    /** Not called. */
    public void characters(int dataIndex, boolean cdataSection) throws Exception {}

    /** Not called. */
    public void ignorableWhitespace(int dataIndex, boolean cdataSection) throws Exception {}

    /** Processing instruction. */
    public void processingInstruction(int targetIndex, int dataIndex) throws Exception {

        // strings
        StringPool pool = getParserState().getStringPool();
        String target = pool.orphanString(targetIndex);
        String data = pool.orphanString(dataIndex);

        // perform callback
        if (fDocumentHandler != null) {
            fDocumentHandler.processingInstruction(target, data);
        }

    } // processingInstruction(int,int)

    /** Comment. */
    public void comment(int dataIndex) throws Exception {
        getParserState().getStringPool().releaseString(dataIndex);
    }

    /** Characters. */
    public void characters(char ch[], int start, int length, boolean cdataSection) throws Exception {

        // perform callback
        if (fDocumentHandler != null) {
            fDocumentHandler.characters(ch, start, length);
        }

    } // characters(char[],int,int,boolean)

    /** Ignorable whitespace. */
    public void ignorableWhitespace(char ch[], int start, int length, boolean cdataSection) throws Exception {

        // perform callback
        if (fDocumentHandler != null) {
            fDocumentHandler.ignorableWhitespace(ch, start, length);
        }

    } // ignorableWhitespace(char[],int,int,boolean)

    //
    // XMLParser methods
    //

    /**
     * Parses the specified input source.
     *
     * @param source The input source.
     *
     * @exception org.xml.sax.SAXException Throws exception on SAX error.
     * @exception java.io.IOException Throws exception on i/o error.
     */
    public void parse(InputSource source) throws SAXException, IOException {

        // parse
        fRootElement = false;
        super.parse(source);

        // no root element
        if (!fRootElement) {
            try {
                error(ErrorCode.E_STRUCT2);
            }
            catch (Exception e) {
                throw new SAXException(e);
            }
        }

    } // parse(InputSource)

} // class SAXParser
