/*
 * (C) Copyright IBM Corp. 1998-1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.parsers;

import com.ibm.xml.framework.ParserState;
import com.ibm.xml.parser.DTD;
import com.ibm.xml.parser.TXDocument;
import com.ibm.xml.parser.TXElement;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * This class is a validating DOM parser which can also
 * "revalidate" a DOM subtree after the document has been parsed.
 *
 * @version Revision: 70 1.8 src/com/ibm/xml/parsers/TXRevalidatingDOMParser.java, parser, xml4j2, xml4j2_0_13 
 * @see com.ibm.xml.parsers.TXDOMParser
 */
public class TXRevalidatingDOMParser 
    extends TXDOMParser {

    //
    // Data
    //

    /** Element of erroneous attribute. */
    private transient Node errorElement;
	
    //
    // Constants
    //

    // debugging

    /** Set to true to debug validate() method. */
    private static final boolean DEBUG_VALIDATE = false;

    //
    // Public methods
    //

    /**
     * Returns the element of the last error. <b>Note:</b> This
     * method only returns non-null if the last erroneous
     * node was of type Node.ATTRIBUTE_NODE. 
     */
    public Node getErrorElement() {
        return errorElement;
    }

	/**
	 * Given a node which is a DOM tree (or subtree), validate the tree 
     * against the current DTD.  If the tree is valid, return null.  If
     * the subtree is invalid return the highest (closest to the root) 
     * and left most node which causes the tree to be invalid.
     * <p>
     * If the node returned from this method is of type Node.ATTRIBUTE_NODE,
     * the attribute's element can be queried by calling the 
     * <code>getErrorElement()</code> method.
	 *
	 * @param node  The node representing the root of the DOM subtree to be
     *              revalidated.
     *
     * @return <i>null</i> if the subtree is valid or the first (highest and 
     *         leftmost) node where the tree is invalid.
	 */
    public Node validate(Node node) {

		if (node.getNodeType() != Node.ELEMENT_NODE || 
            !(node instanceof TXElement)) {
			throw new IllegalArgumentException("Can't revalidate a non element");
		}

        // perform validation
        DTD dtd = ((TXDocument)node.getOwnerDocument()).getDTD();
        ParserState parserState = getParserState();
        try {
            Node errorNode = recursiveValidate((TXElement)node, dtd, parserState);
            if (DEBUG_VALIDATE) {
                System.out.println("validate -> errorNode="+errorNode);
            }
            return errorNode;
        }
        catch (Exception e) {
            e.printStackTrace(System.err);
            return null;
        }

    } // validate(Node):Node

    //
    // Private methods
    //

    /** 
     * Recursive validation. This method returns the first node that is
     * in error, or <i>null</i> if the content is valid.
     *
     * @param element   The element to validate.
     * @param dtd       The DTD of the owner document of the element.
     */
    private Node recursiveValidate(TXElement element, 
                                   final DTD dtd, 
                                   final ParserState parserState) {

        // debugging output
        if (DEBUG_VALIDATE) {
            print(element, "");
        }

        // validate this node
        Node errorNode = dtd.validate(element, parserState);
        if (errorNode != null) {
            errorElement = errorNode.getNodeType() == Node.ATTRIBUTE_NODE 
                         ? element : null;
        }

        // validate children
        if (errorNode == null && element.hasChildNodes()) {
            for (Node child = element.getFirstChild();
                 child != null;
                 child = child.getNextSibling()) {
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    errorNode = recursiveValidate((TXElement)child, dtd, parserState);
                    if (errorNode != null) {
                        errorElement = errorNode.getNodeType() == Node.ATTRIBUTE_NODE
                                     ? child : null;
                        break;
                    }
                }
            }
        }

        // return success
        return errorNode;
    
    } // recursiveValidate(Node,DTD,ParserState):Node

    //
    // Debugging methods
    //

    /** Debugging print method. */
    private static void print(Node node, String indent) {

        if (DEBUG_VALIDATE) {
            if (node == null) {
                System.out.println(indent+"!!! node == null");
                return;
            }
            System.out.println(indent+
                               "node: type="+type(node.getNodeType())+
                               ", name="+node.getNodeName()+
                               ", value=\""+normalize(node.getNodeValue())+'"');
            if (node.hasChildNodes()) {
                System.out.println(indent+'{');
                NodeList children = node.getChildNodes();
                int count = children.getLength();
                for (int i = 0; i < count; i++) {
                    print(children.item(i), indent+"  ");
                }
                System.out.println(indent+'}');
            }
        }

    } // print(Node,String)

    /** Debugging normalizing text. */
    private static String normalize(String s) {

        if (DEBUG_VALIDATE) {
            if (s == null) {
                return "[null]";
            }
            StringBuffer str = new StringBuffer();
            int len = s.length();
            for (int i = 0; i < len; i++) {
                if (i > 20) {
                    str.append("...");
                    break;
                }
                char ch = s.charAt(i);
                switch (ch) {
                    case '\r': str.append("[\\r]"); break;
                    case '\n': str.append("[\\n]"); break;
                    default: str.append(ch);
                }
            }
            return str.toString();
        }

        return null;

    } // normalize(String):String

    /** Debugging node type name. */
    private static String type(int type) {

        if (DEBUG_VALIDATE) {
            switch (type) {
                case Node.ATTRIBUTE_NODE: return "ATTR";
                case Node.ELEMENT_NODE: return "ELEMENT";
                case Node.TEXT_NODE: return "TEXT";
                case Node.ENTITY_REFERENCE_NODE: return "ENTITY_REF";
            }
            return "???";
        }

        return null;

    } // type(int):String

} // class TXRevalidatingDOMParser
