/**
 * $Id: XSLProcessor.java,v 1.6 1998/12/21 05:53:39 sboag Exp $
 * 
 * Copyright (c) 1998-1999 Lotus Corporation, Inc. All Rights Reserved.
 *				This software is provided without a warranty of any kind.
 * 
 * $State: Exp $
 * 
 * @author Scott Boag (scott_boag@lotus.com)
 * 
 * It is the responsibility of the XSLProcessor class to traverse the 
 * source DOM tree according to the instructions in the style DOM tree, 
 * and output nodes into the result DOM tree.
 * 
 * Basic flow of control:
 * Starting from process(Document sourceTree, Document out, String mode),
 * 1)  init the Stylesheet instances (which hold the state for each imported 
 *     XSL stylesheet) in the parseXSLRoot method, then 
 * 2)  locate the rootRule (all rules are found via Stylesheet.findTemplate), 
 *     then pass the root node of the source tree to 
 * 3)  buildResultFromTemplate, which processes each of the template's 
 *     child elements by calling 
 * 4)  processTemplateElement, which either processes the element 
 *     by handling it as an XSL directive (if it is in the xsl: namespace), 
 *     or by cloning the source node to the result tree.  
 *     If it is an XSL instruction,
 * 5)  processXSLTemplateInstruction is called, which calls various 
 *     subfunctions according to the tag name.
 *     The XSL directives that effect the source traversal are xsl:apply-templates 
 *     and xsl:for-each, which are handled in the 
 * 6)  handleApplyTemplatesInstruction, which selects the start node and 
 *     calls the 
 * 7)  transformSelectedChildren method, which performs a query or just 
 *     traverses the child list, then for each node in the selected list calls 
 * 8)  transformChild, which then calls 
 * 9)  Stylesheet.findTemplate (if it is not an xsl:for-each instruction)
 *     to locate the corresponding template, and then calls 
 * 10) buildResultFromTemplate, recursing back in to step 3 and starting the 
 *     traversal cycle all over again.
 * 
 * Glossory (abbreviations used in variable and method 
 * names in square brackets):
 * <DL>
 *   <DT>XSL Instruction [instruction]</DT>
 *     <DD>Any tag with an XSL namespace prefix.</DD>
 *   <DT>XSL Template Instruction [instruction]</DT>
 *     <DD>Any tag with an XSL namespace prefix
 *         that occurs inside an xsl:template element.</DD>
 *   <DT>Template Child [templateChild]</DT>
 *       <DD>Any node that is a child of an xsl:template element.</DD>
 *   <DT>Source Tree</DT>
 *     <DD>The tree input to the XSL process.</DD>
 *   <DT>Result Tree</DT>
 *     <DD>The tree that is output by the XSL process.</DD>
 *   <DT>Stylesheet Tree</DT>
 *     <DD>The stylesheet tree produced from the XSL file.</DD>
 *   <DT>Pattern List [patternList]</DT>
 *       <DD>A parsed query or match pattern.</DD>
 * </DL>
 * 
 */
package com.lotus.xsl;

import org.w3c.dom.*;

import java.util.*;
import java.net.*;
import java.io.*;
import java.lang.reflect.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

/**
 * It's the responsibility of the XSLProcessor class, 
 * collaborating with the XML parser liaison, the DOM, 
 * and the query/match engine, to transform a source tree 
 * of nodes into a result tree according to instructions 
 * and templates specified by a stylesheet tree.  
 * The methods <code>process(...)</code> are the primary 
 * public entry points.
 * Look at the ProcessXSL class (com.lotus.xsl.xml4j) 
 * for an example of usage by the main() function.
 * 
 * <p>You must create a new instance of the XSLProcessor class each 
 * time you run a transformation.  Fixing this is high on my priority list.</p>
 * 
 * @version  0.14, 22 Jan 1999
 * @author   Scott Boag
 */
public class XSLProcessor implements FormatterListener
{
  //==========================================================
  // SECTION: Constructors
  //==========================================================
  private void ____CONSTRUCTORS____(){}
    
  /**
   * Construct an XSL processor that can call back to 
   * a XML processor, so it can handle included files, 
   * and the like.
   * <p>You must create a new instance of the XSLProcessor class each 
   * time you run a transformation.  Fixing this is high on my priority list.</p>
   * @param XMLParserLiaison A liaison to an object 
   *                         that can help with stuff 
   *                         specific to a parser 
   *                         implementation.
   * @see com.lotus.xsl.XMLParserLiaison
   * @see com.lotus.xsl.xml4j.ProcessXSL
   */
  public XSLProcessor(XMLParserLiaison parserLiason)
  {
    m_problemListener = new ProblemListenerDefault();
    m_parserLiaison = parserLiason;
    if(m_parserLiaison instanceof Formatter)
    {
      m_formatter = (Formatter)m_parserLiaison;
    }
    m_parserLiaison.setProcessorOwner(this);
    m_queryEngine = new QueryEngine(this);
  }
  
  /**
   * Construct an XSL processor that can call back to 
   * a XML processor, so it can handle included files, 
   * and the like, and do the initial parse of the 
   * XSL document.
   * <p>You must create a new instance of the XSLProcessor class each 
   * time you run a transformation.  Fixing this is high on my priority list.</p>
   * @param xslURLString A valid URI or filename.
   * @param XMLParserLiaison A liaison to an object 
   *                         that can help with stuff 
   *                         specific to a parser 
   *                         implementation.
   * @exception Exception thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   * @see com.lotus.xsl.XMLParserLiaison
   * @see com.lotus.xsl.xml4j.ProcessXSL
   */
  public XSLProcessor(String xslURLString, 
                      XMLParserLiaison parserLiason)
    throws Exception
  {
    m_problemListener = new ProblemListenerDefault();
    m_parserLiaison = parserLiason;
    if(m_parserLiaison instanceof Formatter)
    {
      m_formatter = (Formatter)m_parserLiaison;
    }
    m_parserLiaison.setProcessorOwner(this);
    m_queryEngine = new QueryEngine(this);
    initDefaultRule();
    parseXSLRoot(xslURLString);
  }

  //==========================================================
  // SECTION: Main API Functions
  //==========================================================
  private void ____MAIN_API____(){}

  /**
   * Given the source and the style trees, 
   * transform the source according to the instructions 
   * found in the style document.
   * @param sourceTree  The input DOM tree (immutable).
   * @param stylesheetTree  The XSL DOM tree (immutable).
   * @param styleBaseURI A base URI from where to look for included and imported files.
   * @param out The result tree, may be null.
   * @return The result tree (same as the out param, if it was not null).
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Document process(
                          Document sourceTree, 
                          Document styleTree, 
                          String styleBaseURI,
                          Document out)
    throws XSLProcessorException, 
           MalformedURLException, 
           FileNotFoundException,
           IOException,
           SAXException
  {
    m_stylesheetRoot = new Stylesheet(styleTree, styleBaseURI);
    
    // Find the root pattern in the XSL.
    Element rootRule = m_stylesheetRoot.findTemplate(sourceTree, sourceTree);

    if(null == rootRule)
    {
      rootRule = m_defaultRootRule;
    }
    
    if(null == out)
    {
      out = this.m_parserLiaison.createDocument();
    }
    
    m_flistener = new FormatterToDOM(out);

    resetCurrentState(sourceTree, sourceTree);
    m_rootDoc = sourceTree;

    startDocument();
        
    // Output the action of the found root rule.  All processing
    // occurs from here.  buildResultFromTemplate is highly recursive.
    buildResultFromTemplate(
                            m_stylesheetRoot, 
                            rootRule, 
                            sourceTree, 
                            sourceTree,null);
    
    endDocument();
    
    return out;
  }
  
  /**
   * Given the source and the style trees, 
   * transform the source according to the instructions 
   * found in the style document, and stream the result 
   * to a PrintWriter via the SAX-based FormatterListener.
   * @param sourceTree  The input DOM tree (immutable).
   * @param stylesheetTree  The XSL DOM tree (immutable).
   * @param styleBaseURI A base URI from where to look for included and imported files.
   * @param resultWriter The stream where you wish the result tree to be written.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public void process(
                          Document sourceTree, 
                          Document styleTree, 
                          String styleBaseURI,
                          PrintWriter resultWriter)
    throws XSLProcessorException, 
           MalformedURLException, 
           FileNotFoundException,
           IOException,
           SAXException
  {
    m_stylesheetRoot = new Stylesheet(styleTree, styleBaseURI);
    
    // Find the root pattern in the XSL.
    Element rootRule = m_stylesheetRoot.findTemplate(sourceTree, sourceTree);

    if(null == rootRule)
    {
      rootRule = m_defaultRootRule;
    }
    
    m_formatter.setFormatterListener(resultWriter, 
                                     m_resultNameSpacePrefix, m_indentResult);
    m_flistener = m_formatter.getFormatterListener();
    
    resetCurrentState(sourceTree, sourceTree);
    m_rootDoc = sourceTree;
    
    startDocument();

    // Output the action of the found root rule.  All processing
    // occurs from here.  buildResultFromTemplate is highly recursive.
    buildResultFromTemplate(
                            m_stylesheetRoot, 
                            rootRule, 
                            sourceTree, 
                            sourceTree, null);
    
    endDocument();
  }
 
  /**
   * Given the source and the style trees, 
   * transform the nodes according to the instructions 
   * found in the style tree.
   * @param sourceTree  The input DOM tree (immutable).
   * @param stylesheetTree  The XSL DOM tree (immutable).
   * @param styleBaseURI A base URI from where to look for included and imported files.
   * @return The result tree.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Document process(
                          Document sourceTree, 
                          Document styleTree, 
                          String styleBaseURI)
    throws XSLProcessorException, 
           MalformedURLException, 
           FileNotFoundException,
           IOException,
           SAXException
  {
    return process(sourceTree, styleTree, styleBaseURI, (Document)null);
  }

  /**
   * Process a source document that contains a xml:stylesheet 
   * PI, transform the XML according to the instructions 
   * found in the XSL stylesheet pointed to by the PI.
   * @param sourceTree  The input DOM tree (immutable).
   * @return The result tree.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Document process(Document sourceTree)
    throws XSLProcessorException, MalformedURLException, 
           FileNotFoundException, IOException, SAXException
  {
    if(null == m_parserLiaison)
    {
      error("Caller must provide an XMLParserLiaison interface!");
    }
    
    String xsldocURLString = getStyleSheetURIfromDoc(sourceTree);
    
    if(null == xsldocURLString)
    {
      error(null, sourceTree, "Can't find stylesheet for document!");
    }
    
    if(null != xsldocURLString)
    {
      parseXSLRoot(xsldocURLString);
    }

    Document out = m_parserLiaison.createDocument();
    m_flistener = new FormatterToDOM(out);
    
    Element rootRule = m_stylesheetRoot.findTemplate(sourceTree, sourceTree);

    if(null == rootRule)
    {
      rootRule = m_defaultRootRule;
    }
    
    Node outputRoot = out.getDocumentElement();
    if(null == outputRoot)
      outputRoot = out;

    resetCurrentState(sourceTree, sourceTree);
    m_rootDoc = sourceTree;

    startDocument();
    // Output the action of the found root rule.  All processing
    // occurs from here.  buildResultFromTemplate is highly recursive.
    buildResultFromTemplate(
                            m_stylesheetRoot, 
                            rootRule, 
                            sourceTree, 
                            sourceTree, null);   
    
    endDocument();
    return out;
  }
  
  /**
   * Given a URL to XML and XSL documents, 
   * transform the XML according to the instructions 
   * found in the XSL.  Requires parseXSLRoot to have been 
   * called, either directly or indirectly (anytime you 
   * hand in a XSL URL, parseXSLRoot is called).
   * @param xmldocURLString  The URI to the input XML document.
   * @param xsldocURLString  The URI to the XSL stylesheet.
   * @param out The result tree, may be null.
   * @return The result tree (same as the out param, if it was not null).
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Document process(String xmldocURLString, 
                          String xsldocURLString, 
                          Document out)
    throws XSLProcessorException, MalformedURLException, 
           FileNotFoundException, IOException, SAXException
  {
    Document sourceTree = null;
        
    // If by chance we can get away with parsing the stylesheet first, 
    // it's better, because it will let us collect id information 
    // as the parse is happening.
    if(null != xsldocURLString)
    {
      parseXSLRoot(xsldocURLString);
    }
         
    if(null != m_diagnosticsPrintWriter)
    {
      pushTime(this);
      diag("========= Parsing "+xmldocURLString+" ==========");
      pushTime(xmldocURLString);
    }
    
    sourceTree = parseXML(xmldocURLString, REPORTPARSEERROR);
    
    if(null == xsldocURLString)
    {
      xsldocURLString = getStyleSheetURIfromDoc(sourceTree);
    } // if(null == xsldocURLString)
    
    if(null == xsldocURLString)
    {
      error(null, sourceTree, "Can't find stylesheet for document!");
    }
    
    if(null != m_diagnosticsPrintWriter)
    {
      displayDuration("Parse of "+xmldocURLString, xmldocURLString);
    }
    
    if((null == m_stylesheetRoot) && (null != xsldocURLString))
    {
      parseXSLRoot(xsldocURLString);
    }
    
    if(null == out)
    {
      out = this.m_parserLiaison.createDocument();
    }
    
    if(null != m_diagnosticsPrintWriter)
    {
      diag("=============================");
      diag("Transforming "+xmldocURLString+" via "+xsldocURLString+"...");
      pushTime(out);
    }
    
    m_flistener = new FormatterToDOM(out);
    
    Element rootRule = m_stylesheetRoot.findTemplate(sourceTree, sourceTree);

    if(null == rootRule)
    {
      rootRule = m_defaultRootRule;
    }
    
    resetCurrentState(sourceTree, sourceTree);
    m_rootDoc = sourceTree;
 
    startDocument();

    // Output the action of the found root rule.  All processing
    // occurs from here.  buildResultFromTemplate is highly recursive.
    buildResultFromTemplate(
                            m_stylesheetRoot, 
                            rootRule, 
                            sourceTree, 
                            sourceTree, null);
    
    endDocument();
    
    if(null != m_diagnosticsPrintWriter)
    {
      displayDuration("transform", out);
      // diag("=============================");
    }
    
    return out;
  }

  /**
   * Given a URL to a XML document, 
   * transform the XML according to the instructions 
   * found in the XSL.  Requires that the XML have a stylesheet 
   * PI in it of type 'text/xsl'.
   * @param xmldocURLString  The URI to the input XML document.
   * @param out The result tree, may be null.
   * @return The result tree (same as the out param, if it was not null).
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Document process(String xmldocURLString, Document out)
    throws XSLProcessorException, MalformedURLException, 
           FileNotFoundException, IOException,
           SAXException
  {
    return process(xmldocURLString, null, out);
  }
  
  /**
   * Given a URL to an XML document, and a URL to a XSL stylesheet, 
   * write the transformed result to a file.
   * @param xmldocURLString  The URI to the input XML document.
   * @param xsldocURLString  The URI to the XSL stylesheet.
   * @param outFileName The name of the file to write to.
   * @return The result tree.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public void process(String xmlurl, String xslurl, String outFileName)
    throws XSLProcessorException, MalformedURLException, 
           FileNotFoundException, IOException,
           SAXException
  {
    process(xmlurl, xslurl, outFileName, 
            new PrintWriter(new FileWriter(outFileName)));
  } // end Process

  /**
   * Given a URL to an XML document, and a URL to a XSL stylesheet, 
   * write the transformed result to a file.
   * @param xmldocURLString  The URI to the input XML document.
   * @param xsldocURLString  The URI to the XSL stylesheet.
   * @param displayName The name of the stream for diagnostics.
   * @param pw A PrintWriter to write the results to.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public void process(String xmldocURLString, String xsldocURLString, String writerName, PrintWriter pw)
    throws XSLProcessorException, MalformedURLException, 
           FileNotFoundException, IOException, SAXException
  {
    Document sourceTree = null;
        
    // If by chance we can get away with parsing the stylesheet first, 
    // it's better, because it will let us collect id information 
    // as the parse is happening.
    if(null != xsldocURLString)
    {
      parseXSLRoot(xsldocURLString);
    }
         
    if(null != m_diagnosticsPrintWriter)
    {
      pushTime(this);
      diag("========= Parsing "+xmldocURLString+" ==========");
      pushTime(xmldocURLString);
    }
    
    sourceTree = parseXML(xmldocURLString, REPORTPARSEERROR);
    
    if(null == xsldocURLString)
    {
      xsldocURLString = getStyleSheetURIfromDoc(sourceTree);
    } // if(null == xsldocURLString)
    
    if(null == xsldocURLString)
    {
      error(null, sourceTree, "Can't find stylesheet for document!");
    }
    
    if(null != m_diagnosticsPrintWriter)
    {
      displayDuration("Parse of "+xmldocURLString, xmldocURLString);
    }
    
    if((null == m_stylesheetRoot) && (null != xsldocURLString))
    {
      parseXSLRoot(xsldocURLString);
    }
    
    if(null != m_diagnosticsPrintWriter)
    {
      diag("=============================");
      diag("Transforming "+xmldocURLString+" via "+xsldocURLString+"...");
      pushTime(m_formatter);
    }
    
    Element rootRule = m_stylesheetRoot.findTemplate(sourceTree, sourceTree);

    if(null == rootRule)
    {
      rootRule = m_defaultRootRule;
    }
    
    m_formatter.setFormatterListener(pw, m_resultNameSpacePrefix, m_indentResult);
    m_flistener = m_formatter.getFormatterListener();
    
    resetCurrentState(sourceTree, sourceTree);
    m_rootDoc = sourceTree;
    
    startDocument();

    // Output the action of the found root rule.  All processing
    // occurs from here.  buildResultFromTemplate is highly recursive.
    buildResultFromTemplate(
                            m_stylesheetRoot, 
                            rootRule, 
                            sourceTree, 
                            sourceTree, null);
    
    endDocument();
    
    if(null != m_diagnosticsPrintWriter)
    {
      displayDuration("transform", m_formatter);
      diag("=============================");
    }    
  } // end Process

  /**
   * Given a URI to a XML document and a URI to an XSL stylesheet, 
   * transform the XML according to the instructions 
   * found in the XSL.
   * @param xmldocURLString  The URI to the input XML document.
   * @param xsldocURLString  The URI to the XSL stylesheet.
   * @return The result tree.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Document process(String xmldocURLString, 
                          String xsldocURLString)
    throws XSLProcessorException, MalformedURLException, 
           FileNotFoundException, IOException, SAXException
  {
    if(null == m_parserLiaison)
    {
      error("Caller must provide an XMLParserLiaison interface!");
    }
    Document out = m_parserLiaison.createDocument();
    m_flistener = new FormatterToDOM(out);
    return process( xmldocURLString, xsldocURLString, out);
  }

  /**
   * Given a URI to a XML document, 
   * transform the XML according to the instructions 
   * found in the XSL.  Requires that the document contain 
   * a stylesheet PI.
   * @param xmldocURLString  The URL to the input XML document.
   * @return The result tree.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Document process(String xmldocURLString)
    throws XSLProcessorException, MalformedURLException, 
           FileNotFoundException, IOException, SAXException
  {
    if(null == m_parserLiaison)
    {
      error("Caller must provide an XMLParserLiaison interface!");
    }
    Document out = m_parserLiaison.createDocument();
    m_flistener = new FormatterToDOM(out);
    return process(xmldocURLString, null, out);
  }  
    
  /**
   * Given a stream of XML, and a stream of XSL, 
   * and stream the result 
   * to a PrintWriter via the SAX-based FormatterListener.
   * @param xmlReader A reader for XML.
   * @param xslReader A reader for XSL.
   * @param styleBaseURI The base directory from where to find 
   *                     included stylesheets.
   * @param writer The writer that will receive the result XML.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public void process(Reader xmlReader, Reader xslReader, 
                          String styleBaseURI, PrintWriter writer)
    throws XSLProcessorException, MalformedURLException, 
           FileNotFoundException, IOException, SAXException
  {
    Document xmlDoc = m_parserLiaison.parseXMLStream(xmlReader, "Source XML stream");
    Document xslDoc = m_parserLiaison.parseXMLStream(xslReader, "Styletree stream");
    
    m_stylesheetRoot = new Stylesheet(xslDoc, styleBaseURI);
    
    // Find the root pattern in the XSL.
    Element rootRule = m_stylesheetRoot.findTemplate(xslDoc, xslDoc);

    if(null == rootRule)
    {
      rootRule = m_defaultRootRule;
    }
        
    m_formatter.setFormatterListener(writer, 
                                     m_resultNameSpacePrefix, m_indentResult);
    m_flistener = m_formatter.getFormatterListener();
    
    resetCurrentState(xmlDoc, xmlDoc);
    m_rootDoc = xmlDoc;
    
    startDocument();

    // Output the action of the found root rule.  All processing
    // occurs from here.  buildResultFromTemplate is highly recursive.
    buildResultFromTemplate(
                            m_stylesheetRoot, 
                            rootRule, 
                            xmlDoc, 
                            xmlDoc, null);
    
    endDocument();
    
  } // end Process
  
  /**
   * Given a stream of XML, stream the result 
   * to a PrintWriter via the SAX-based FormatterListener.
   * Requires that the source XML contain a stylesheet PI of type 
   * 'text/xsl'.
   * @param xmlReader A reader for XML.
   * @param writer The writer that will receive the result XML.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public void process(Reader xmlReader, PrintWriter writer)
    throws XSLProcessorException, MalformedURLException, 
           FileNotFoundException, IOException, SAXException
  {
    Document xmlDoc = m_parserLiaison.parseXMLStream(xmlReader, "Source XML stream");
 
    String xsldocURLString = getStyleSheetURIfromDoc(xmlDoc);

    String baseURLHolder[] = {null};
    Document xslDoc = parseXML(xsldocURLString, baseURLHolder);
    
    m_stylesheetRoot = new Stylesheet(xslDoc, baseURLHolder[0]);
    
    // Find the root pattern in the XSL.
    Element rootRule = m_stylesheetRoot.findTemplate(xslDoc, xslDoc);

    if(null == rootRule)
    {
      rootRule = m_defaultRootRule;
    }
        
    m_formatter.setFormatterListener(writer, 
                                     m_resultNameSpacePrefix, m_indentResult);
    m_flistener = m_formatter.getFormatterListener();
    
    resetCurrentState(xmlDoc, xmlDoc);
    m_rootDoc = xmlDoc;
    
    startDocument();

    // Output the action of the found root rule.  All processing
    // occurs from here.  buildResultFromTemplate is highly recursive.
    buildResultFromTemplate(
                            m_stylesheetRoot, 
                            rootRule, 
                            xmlDoc, 
                            xmlDoc, null);
    
    endDocument();

  } // end Process

  /**
   * Given a stream of XML, and a stream of XSL, 
   * return the result as a DOM tree.
   * @param xmlReader A reader for XML.
   * @param xslReader A reader for XSL.
   * @param styleBaseURI The base directory from where to find 
   *                     included stylesheets.
   * @return The result tree.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Document process(Reader xmlReader, Reader xslReader, String styleBaseURI)
    throws XSLProcessorException, MalformedURLException, 
           FileNotFoundException, IOException, SAXException
  {    
    Document xmlDoc = m_parserLiaison.parseXMLStream(xmlReader, "Source XML stream");
    Document xslDoc = m_parserLiaison.parseXMLStream(xslReader, "Styletree stream");

    // The main XSL transformation occurs here!
    Document results = process(xmlDoc, xslDoc, styleBaseURI, (Document)null);
    
    return results;
  } // end Process
  
  /**
   * Given a stream of XML, apply the stylesheet found in the 
   * document's stylesheet PI, and return the result as a DOM tree.
   * @param xmlReader A reader for XML.
   * @return The result tree.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Document process(Reader xmlReader)
    throws XSLProcessorException, MalformedURLException, 
           FileNotFoundException, IOException, SAXException
  {
    Document xmlDoc = m_parserLiaison.parseXMLStream(xmlReader, "Source XML stream");
    String xsldocURLString = getStyleSheetURIfromDoc(xmlDoc);

    String baseURLHolder[] = {null};
    Document xslDoc = parseXML(xsldocURLString, baseURLHolder);
    
    // The main XSL transformation occurs here!
    Document results = process(xmlDoc, xslDoc, baseURLHolder[0], (Document)null);
    
    return results;
  } // end Process

  //==========================================================
  // SECTION: XML Parsing Functions
  //==========================================================
  private void ____XML_PARSING_FUNCTIONS____(){}

  /**
   * Take a user string and try and parse XML, and also return 
   * the url.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Document parseXML(String urlString, String holder[])
    throws XSLProcessorException, 
           MalformedURLException, 
           FileNotFoundException,
           IOException
  {
    return parseXML(urlString, holder, null);
  }
  
  /**
   * Take a user string and try and parse XML, and also return 
   * the url.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Document parseXML(String urlString, String holder[], String base)
    throws XSLProcessorException, 
    MalformedURLException, 
    FileNotFoundException,
    IOException
  {
    Document doc = null;
    // A bit risky, but see if we already have a document keyed by relative file name.
    Object docObj = this.m_sourceDocs.get(urlString);
    if(null != docObj)
    {
      doc = (Document)docObj;
    }
    else
    {
      try 
      {
        URL url = new URL(urlString);
        if(null != holder) holder[0] = url.toString();
        doc = parseXML(url, EXITONPARSEERROR);
      }
      catch (MalformedURLException e) {}
      
      if((null == doc) && (null != base))
      {
        try 
        {
          URL baseUrl = new URL(base);
          URL url = new URL(baseUrl, urlString);
          if(null != holder) holder[0] = url.toString();
          doc = parseXML(url, EXITONPARSEERROR);
        }
        catch (MalformedURLException e) {}
      }
      
      if(null == doc)
      {
        File f = new File(urlString);
        
        if(!f.isAbsolute() && (null != base))
        {
          // String dir = f.isDirectory() ? f.getAbsolutePath() : f.getParent();
          StringTokenizer tokenizer = new StringTokenizer(base, "\\/");
          String fixedBase = null;
          while(tokenizer.hasMoreTokens())
          {
            String token = tokenizer.nextToken();
            if(null == fixedBase)
            {
              fixedBase = token;
            }
            else
            {
              fixedBase += File.separator + token;
            }
          }
          f = new File(fixedBase);
          String dir = f.isDirectory() ? f.getAbsolutePath() : f.getParent();
          f = new File(dir, urlString);
          urlString = f.getAbsolutePath();
        }
        else
        {
          if(null != holder) holder[0] = f.getAbsolutePath();
        }
        FileReader freader = new java.io.FileReader(f);
        
        if(null != freader)
        {
          Reader reader 
            = new java.io.BufferedReader(freader, 1024 * 16);
          doc  = m_parserLiaison.parseXMLStream(reader, f.getName());
          if(null != doc)
          {
            this.m_sourceDocs.put(urlString, doc);
          }
        }
        else
        {
          error("Can not make a file name from " + urlString);
        }
        
      }
      if(null != doc)
      {
        if((null != holder) && (null == holder[0])) holder[0] = urlString;
      }
      else
      {
        try 
        {
          String fullpath = "file:" + new File (urlString).getAbsolutePath ();
          URL url = new URL(fullpath);
          doc = parseXML(url, EXITONPARSEERROR);
          if(null != doc)
          {
            if(null != holder) holder[0] = url.toString();
          }
          else
          {
            error("Cannot open stylesheet: " + fullpath);
          }
        }
        catch (MalformedURLException e2)
        {
          diag("Cannot create url for: " + urlString);
          throw e2;
        }
      }
    }
    return doc;
  }


  /**
   * Read in the XML file.
   * @param urlString Pointer to a XML-complient file.
   * @param howToHandleParseError One of 
   * NOREPORTPARSEERROR, REPORTPARSEERROR,EXITONPARSEERROR.
   * @param baseURL returns the base URL. 
   * @return lcom.ms.xml.om.Document object, which 
   * represents the parsed XML.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Document parseXML(String urlString, 
                           int howToHandleParseError)
    throws  XSLProcessorException, 
            MalformedURLException, 
            FileNotFoundException,
            IOException
  {
    return parseXML(urlString, null, null);
  }
  
  /**
   * Read in the XML file.
   * @param urlString Pointer to a XML-complient file.
   * @param howToHandleParseError One of 
   * NOREPORTPARSEERROR, REPORTPARSEERROR,EXITONPARSEERROR.
   * @param baseURL returns the base URL. 
   * @return lcom.ms.xml.om.Document object, which 
   * represents the parsed XML.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Document parseXML(URL url, 
                           int howToHandleParseError)
    throws  XSLProcessorException, 
            MalformedURLException, 
            FileNotFoundException, 
            IOException
  {
    Document doc = null;
    
    // If we already have the document, then just return it.
    // Hopefully not parsing the same document twice is a feature...
    Object docObj = this.m_sourceDocs.get(url.toExternalForm());
    if(null != docObj)
    {
      return (Document)docObj;
    }
    
    if(null == m_parserLiaison)
    {
      error("Caller must provide an XMLParserLiaison interface!");
    }
    
    InputStream instream = null;
    String whichCall = "";
    
    try
    {
      whichCall = "parseXMLStream";
      doc = m_parserLiaison.parseXMLStream(url);
    }
    catch(FileNotFoundException fnfe)
    {
    }
    
    try 
    {
      if(null == doc)
      {
        whichCall = "openConnection";
        URLConnection connect = url.openConnection();
        
        diag("parseXML - URL = " + url);
        
        whichCall = "getInputStream";
        instream = connect.getInputStream();
        
        Reader reader = new java.io.BufferedReader(
                                                   new java.io.InputStreamReader(instream), 1024 * 16);
        
        doc = m_parserLiaison.parseXMLStream(reader, url.toExternalForm());
      }
    } 
    catch (IOException e) 
    {
      diag(whichCall + ": " + url.toExternalForm() + ": " + e.toString());
      throw e;
    }
    
    // reader.close();
    if(null != doc)
    {
      this.m_sourceDocs.put(url.toExternalForm(), doc);
    }
    return doc;
  }
    
  //==========================================================
  // SECTION: Stylesheet Tables
  //==========================================================
  private void ____STYLESHEET_TABLES____(){}
  
  /**
   * The root of a linked set of stylesheets.
   */
  Stylesheet m_stylesheetRoot = null;  

  /**
   * Table of stylesheet documents.
   * Document objects are keyed by URL string.
   */
  Hashtable m_stylesheets = new Hashtable();
  
  /**
   * Get table of stylesheet documents.
   * Document objects are keyed by URL string.
   */
  public Hashtable getStylesheetsTable()
  {
    return m_stylesheets;
  }

  /**
   * Reset the state of the XSL processor by reading in a new 
   * XSL stylesheet.
   * @param xslURLString a valid URI to an XSL stylesheet.
   * @param outDiagnostics The print stream to write diagnostics (may be null).
   */
  void parseXSLRoot(String xslURLString)
    throws XSLProcessorException, 
           MalformedURLException, 
           FileNotFoundException, 
           IOException,
           SAXException
  {
    diag("========= Parsing "+xslURLString+" ==========");
    pushTime(xslURLString);
    
    String[] stringHolder = 
    {
      null};
    Document stylesheetDoc = parseXML(xslURLString, stringHolder);
    displayDuration("Parse of "+xslURLString, xslURLString);
    pushTime(xslURLString);
    m_stylesheetRoot = new Stylesheet(stylesheetDoc, stringHolder[0]);
    displayDuration("Setup of "+xslURLString, xslURLString);
  }  

  /**
   * Not yet implemented!  Set a stylesheet keyed by a URI.
   * 
   * Every time encountering an xsl:import or xsl:include element 
   * the processor will first consult table of supplied 
   * documents before trying to load and parse the corresponding 
   * document from the Net.
   * 
   * This could be useful from several points of view:
   * - frequently used documents would be loaded and parsed just 
   *   once - the performance of XSL based applications could be 
   *   improved.
   * 
   * - programmatically constructed documents could be used for 
   *   import and inclusion.
   * 
   * - XSL stylesheet could be parametrized with respect to their 
   *   imports and includes; this means that decision which particular 
   *   document needs to be imported or included could be postponed till 
   *   actual application of a stylesheet to an XML document; XSL 
   *   stylesheets could be authored with dummy URIs for imports and 
   *   includes - these URIs would play role of parameters that are 
   *   resulved every time the stylesheet is applied; decision which 
   *   particular value to give to wich parameter would be made by an 
   *   application program, not by author of the stylesheet.
   * 
   * @param uri A unique URI.
   * @param doc A DOM tree that is an XSL stylesheet.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public void setStyleSheet (String uri, Document doc, boolean shouldBuildTables)
    throws XSLProcessorException
  {
    error("setStyleSheet is not yet implemented!");
  }

  //==========================================================
  // SECTION: Source Tree Tables
  //==========================================================
  private void ____SOURCE_TREE_TABLES____(){}

  /**
   * Table of input documents.
   * Document objects are keyed by URL string.
   */
  Hashtable m_sourceDocs = new Hashtable();
  
  Document m_rootDoc = null;
  
  /**
   * Given a DOM Document, tell what URI was used to parse it.
   */
  String findURIFromDoc(Document doc)
  { 
    String uri = null;
    Enumeration values = m_sourceDocs.elements();
    Enumeration keys = m_sourceDocs.keys();
    while(values.hasMoreElements())
    {
      Document docval = (Document)values.nextElement();
      String key = (String)keys.nextElement();
      if(docval == doc)
      {
        uri = key;
        break;
      }
    }
    return uri;
  }

  /**
   * Get table of source tree documents.
   * Document objects are keyed by URL string.
   */
  public Hashtable getSourceDocsTable()
  {
    return m_sourceDocs;
  }

  /**
   * Set a source document. Every time a source document is requested 
   * through either xsl:uri or in a process call,
   * the processor will first consult table of supplied 
   * documents before trying to load and parse the corresponding 
   * document from the Net.

   */
  public void setSourceDocument (String uri, Document doc)
  {
    m_sourceDocs.put(uri, doc);
  }


  //==========================================================
  // SECTION: Major flow-of-control functions
  //==========================================================
  private void ____MAIN_CONTROL_FLOW____(){}
  
  /**
   * Given a template element, process it's child elements by
   * calling processTemplateElement.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void buildResultFromTemplate(
                                         Stylesheet stylesheetTree, 
                                         Element xslTemplate, 
                                         Document sourceTree, 
                                         Node sourceNode, String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    if(m_traceTemplates)
      traceTemplate(xslTemplate);
    if(m_needToCheckForInfiniteLoops)
    {
      m_stackGuard.push(xslTemplate, sourceNode);
    }
    NodeList children = xslTemplate.getChildNodes();
    if(null != children)
    {
      // Loop through the children of a rule looking for 
      // the first non-pattern node.  Once one 
      // or the other is found, add it to the output 
      // DOM tree.
      int nNodes = children.getLength();
      for(int i = 0; i < nNodes; i++) 
      {
        processTemplateElement(
                               stylesheetTree, children.item(i), 
                               sourceTree, sourceNode, mode);
      }
    }
    if(m_needToCheckForInfiniteLoops)
    {
      m_stackGuard.pop();
    }
  }

  /**
   * Given a rule element, process it's action elements.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void processTemplateElement(
                                        Stylesheet stylesheetTree, 
                                        Node templateChild, 
                                        Document sourceTree, 
                                        Node sourceNode,
                                        String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    if(m_traceTemplateChildren)
      diagnoseTemplateChildren(templateChild, sourceNode);
    int nodeType = templateChild.getNodeType();

    // See if the template child is an XSL directive, i.e. 
    // one of "children", "select-element", "empty", etc.
    boolean isXSLInstruction 
      = (nodeType == Node.ELEMENT_NODE) 
        ? processXSLTemplateInstruction(
                                        stylesheetTree, (Element)templateChild, 
                                        sourceTree, sourceNode,
                                        mode) 
          : false;
    
    if(false == isXSLInstruction)
    {
      // If template child object is not an XSL directive, clone the 
      // template child into the output tree.

      switch(nodeType)
      {
      case Node.DOCUMENT_NODE:
        startDocument();
        // Hmm...
        endDocument();
        break;
      case Node.ELEMENT_NODE:
        {          
          // This part I don't like.  I have to loop through the children 
          // looking for a xsl:attribute child.
          Element templateElem = (Element)templateChild;
          startElement (templateChild.getNodeName());
          copyAttributesToAttList( sourceNode, stylesheetTree, 
                                   (Element)templateChild,
                                   m_pendingAttributes);
          copyNamespaceAttributes(templateElem, true, m_pendingAttributes);
          translateCSSAttrsToStyleAttr(m_pendingAttributes);
          // TODO: Process xsl:attribute && xsl:use
          processTemplateElementChildren(stylesheetTree, 
                                         (Element)templateChild, sourceTree, 
                                         sourceNode, mode);
          endElement (templateChild.getNodeName());
        }
        break;
      case Node.TEXT_NODE:
        {
          // if(!m_parserLiaison.isIgnorableWhitespace((Text)templateChild))
          {
            String data = getNormalizedText((Text)templateChild);
            if(null != data)
            {
              if(0 == data.trim().length())
              {
                boolean isIgnorable = true;
                Node next = templateChild;
                while(null != (next = next.getNextSibling()))
                {
                  if(Node.TEXT_NODE == next.getNodeType() ||
                     Node.CDATA_SECTION_NODE == next.getNodeType())
                  {
                    String nextData = ((CharacterData)next).getData();
                    if(0 != nextData.trim().length())
                    {
                      isIgnorable = false;
                      break;
                    }
                  }
                  else
                  {
                    break;
                  }
                }
                if(true == isIgnorable)
                {
                  next = templateChild;
                  while(null != (next = next.getPreviousSibling()))
                  {
                    if(Node.TEXT_NODE == next.getNodeType() ||
                       Node.CDATA_SECTION_NODE == next.getNodeType())
                    {
                      String nextData = ((CharacterData)next).getData();
                      if(0 != nextData.trim().length())
                      {
                        isIgnorable = false;
                        break;
                      }
                    }
                    else
                    {
                      break;
                    }
                  }
                }
                if(!isIgnorable)
                {
                  characters(data.toCharArray(), 0, data.length());
                }
              }
              else
              {
                characters(data.toCharArray(), 0, data.length());
              }
            }
          }
        }
        break;
      case Node.CDATA_SECTION_NODE:
        {
          String data = ((Text)templateChild).getData();
          cdata(data.toCharArray(), 0, data.length());
        }
        break;
      }
    }
    // Pop any variables off that might have been pushed.
    
    m_variableStacks.popElementMarker(templateChild);
  }
  
  /**
   * Process the children of a template result literal element.
   * Note that this could be merged with buildResultFromTemplate 
   * if we get rid of the old pattern syntax.
   */
  private void processTemplateElementChildren(
                                              Stylesheet stylesheetTree, 
                                              Element templateChild, 
                                              Document sourceTree, 
                                              Node sourceNode, 
                                              String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    try
    {
      NodeList children = templateChild.getChildNodes();    
      int nNodes = children.getLength();
      
      // Process the children of the template child, which may 
      // either be more template children, in which case we 
      // recurse, or else there may be text node children,
      // in which case we simply add them as children of the 
      // node just created.
      for(int i = 0; i < nNodes; i++) 
      {
        processTemplateElement(
                               stylesheetTree, children.item(i), 
                               sourceTree, sourceNode, mode);
      }
    }
    catch(DOMException e)
    {
      // e.printStackTrace();
      String msg = e.getMessage();
      if(null == msg)
      {
        warn(templateChild, sourceNode, "DOMException: "+e.code);
      }
      else
      {
        warn(templateChild, sourceNode, msg);
      }
    }
  }
  
  /**
   * Perform a query if needed, and call transformChild for each child.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void transformSelectedChildren(
                                           Stylesheet stylesheetTree, 
                                           Element xslInstruction, // xsl:apply-templates or xsl:for-each
                                           Element template, // The template to copy to the result tree
                                           Document sourceTree, 
                                           Node sourceNodeContext, String mode, 
                                           Expression selectPattern, 
                                           int xslToken)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    // Sort the nodes according to the xsl:sort method
    NodeList applyTemplatesChildren = xslInstruction.getChildNodes();
    int nChildren = applyTemplatesChildren.getLength();
    Vector keys = null;
    
    if(nChildren > 0)
    {
      // March backwards, performing a sort on each xsl:sort child.
      // Probably not the most efficient method.
      for(int i = 0; i < nChildren; i++)
      {
        Node child = applyTemplatesChildren.item(i);
        if(isXSLTagOfType(child, Constants.ELEMNAME_SORT))
        {
          if(null == keys) keys = new Vector();
          Element sort = (Element)child;
          String selectString = getProcessedAttrVal(sort, Constants.ATTRNAME_SELECT, sourceNodeContext);
          if(null == selectString) selectString = ".";
          String langString = getProcessedAttrVal(sort, Constants.ATTRNAME_LANG, sourceNodeContext);
          String dataTypeString = getProcessedAttrVal(sort, Constants.ATTRNAME_DATATYPE, sourceNodeContext);
          boolean treatAsNumbers = ((null != dataTypeString)&& 
                                    dataTypeString.equals(Constants.ATTRVAL_DATATYPE_NUMBER)) ? 
                                   true : false;
          String orderString = getProcessedAttrVal(sort, Constants.ATTRNAME_ORDER, sourceNodeContext);
          boolean descending = ((null != orderString) &&  
                                orderString.equals(Constants.ATTRVAL_ORDER_DESCENDING))? 
                               true : false;
          String caseOrderString = getProcessedAttrVal(sort, Constants.ATTRNAME_CASEORDER, sourceNodeContext);
          if(null != caseOrderString)
          {
            warn(xslInstruction, sourceNodeContext, 
                    "LotusXSL does not yet handle the "+
                    Constants.ATTRNAME_CASEORDER+" attribute!");
          }
          keys.addElement(new NodeSortKey(this, selectString, treatAsNumbers, 
                                          descending, langString, xslInstruction));
        }
      }
    }

    NodeListImpl sourceNodes = null;
    if(null != selectPattern)
    {
      sourceNodes = (NodeListImpl)this.m_queryEngine.query(selectPattern, 
                                                    sourceNodeContext, xslInstruction);
    }
    else if(null != keys)
    {
      sourceNodes = new NodeListImpl();
      NodeList children = sourceNodeContext.getChildNodes();    
      int nNodes = children.getLength();
      for(int i = 0; i < nNodes; i++) 
      {
        sourceNodes.addElement(children.item(i));
      }
    }
        
    if(null != sourceNodes)
    {
      int nNodes = (null != sourceNodes) ? sourceNodes.size() : 0;
      
      if(nNodes > 0)
      {
        if(null != keys)
        {
          NodeSorter sorter = new NodeSorter(this);
          sorter.sort(sourceNodes, keys);
        }
        
        // NodeList children = sourceNodeContext.getChildNodes(); 
        NodeList savedContextNodeList = m_contextNodeList;
        m_contextNodeList = sourceNodes;
        if(m_traceSelects)
          traceSelect(xslInstruction, sourceNodes);
        
        for(int i = 0; i < nNodes; i++) 
        {
          Node childNode = (Node)sourceNodes.elementAt(i);
          
          Document ownerDoc = childNode.getOwnerDocument();
          if((Node.DOCUMENT_NODE != childNode.getNodeType()) && (null == ownerDoc))
          {
            error(template, childNode, "Child node does not have an owner document!");
          }
          
          transformChild(
                         stylesheetTree, xslInstruction, template, 
                         ownerDoc, 
                         sourceNodeContext, childNode,
                         mode, xslToken);
        }
        m_contextNodeList = savedContextNodeList;
      }
    }
    else
    {
      NodeList children = sourceNodeContext.getChildNodes(); 
      NodeList savedContextNodeList = m_contextNodeList;
      m_contextNodeList = children;
      if(m_traceSelects)
        traceSelect(xslInstruction, children);
      int nNodes = children.getLength();
      for(int i = 0; i < nNodes; i++) 
      {
        Node childNode = children.item(i);

        transformChild(
                       stylesheetTree, xslInstruction, template, 
                       sourceTree, sourceNodeContext, childNode,
                       mode, xslToken);
      }
      m_contextNodeList = savedContextNodeList;
    }
  }
    
  /**
   * Given an element and mode, find the corresponding 
   * template and process the contents.
   * @param stylesheetTree The current Stylesheet object.
   * @param template The template to use if xsl:for-each, or null.
   * @param sourceTree The source DOM tree.
   * @param xmlNodeParent The parent of the source context node.
   * @param child The source context node.
   * @param resultTree The result DOM tree.
   * @param resultNode The current result node to add nodes to as children.
   * @param mode The current mode, may be null.
   * @param xslToken ELEMNAME_APPLY_TEMPLATES, ELEMNAME_APPLY_IMPORTS, or ELEMNAME_FOREACH.
   * @return true if applied a template, false if not.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  boolean transformChild(Stylesheet stylesheetTree, 
                         Element xslInstruction, // xsl:apply-templates or xsl:for-each
                         Element template, // may be null
                         Document sourceTree, 
                         Node selectContext,
                         Node child,
                         String mode, int xslToken
                         )
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    boolean doApplyTemplate = true; // return value
    
    boolean shouldStrip = false;
    int nodeType = child.getNodeType();
    if((Node.TEXT_NODE == nodeType) &&
       shouldStripSourceNode(child))
    {
      if(m_parserLiaison.isIgnorableWhitespace((Text)child))
      {
        shouldStrip = true;
      }
      else
      {
        String data = ((Text)child).getData();
        if((null == data) || (0 == data.trim().length()))
        {
          shouldStrip = true;
        }
      }
    }

    if(!shouldStrip)
    {
      if(null == template)
      {
        boolean isApplyImports = (xslToken == Constants.ELEMNAME_APPLY_IMPORTS);
        // Find the XSL template that is the best match for the 
        // element, and call buildResultFromTemplate.
        Stylesheet foundStylesheet[] = 
        {
          null};
        
        if(!isApplyImports)
        {
          stylesheetTree = m_stylesheetRoot;
        }
        template = stylesheetTree.findTemplate(sourceTree, child, mode,
                                               isApplyImports, foundStylesheet);
        if(isApplyImports)
        {
          if(null != template)
          {
            stylesheetTree = foundStylesheet[0];
          }
          else
          {
            doApplyTemplate = false;
          }
        }
        mode = null; // non-sticky modes
      }

      if(doApplyTemplate)
      {
        if(null == template)
        {
          switch(nodeType)
          {
          case Node.ELEMENT_NODE:
            template = this.m_defaultRule;
            break;
          case Node.CDATA_SECTION_NODE:
          case Node.TEXT_NODE:
            template = this.m_defaultTextRule;
            break;
          case Node.DOCUMENT_NODE:
            template = this.m_defaultRootRule;
            break;
          }     
        }
        
        if(null != template)
        {
          resetCurrentState(sourceTree, child);
          
          if(template == m_defaultTextRule)
          {
            switch(nodeType)
            {
            case Node.CDATA_SECTION_NODE:
              cloneToResultTree(child, false, false, false, false);
              break;
              
            case Node.TEXT_NODE:
              // if(!m_parserLiaison.isIgnorableWhitespace((Text)child))
              {
                cloneToResultTree(child, false, false, false, false);
              }
            }
          }
          else
          {
            if(xslToken != Constants.ELEMNAME_FOREACH)
            {
              m_variableStacks.pushContextMarker(template, child);
              
              if(null != xslInstruction)
              {
                m_variableStacks.pushParams(stylesheetTree, 
                                            xslInstruction, 
                                            sourceTree, 
                                            selectContext, mode, template);
              }
            }

            buildResultFromTemplate(
                                    stylesheetTree, 
                                    template, 
                                    sourceTree, 
                                    child, mode);
            
            if(xslToken != Constants.ELEMNAME_FOREACH)
            {
              m_variableStacks.popCurrentContext();
            }
          }
          
          resetCurrentState(sourceTree, selectContext);
        }
      }
    }
    return doApplyTemplate;
  }
  
  //==========================================================
  // SECTION: XSL directive handling functions
  //==========================================================  
  private void ____XSL_INSTRUCTIONS____(){}

  static final String m_XSLTVerSupported = "1.0";
  String m_XSLNameSpaceURL = "http://www.w3.org/xsl/transform/1.0";
  static final String m_XSLNameSpaceURLPre = "http://www.w3.org/xsl/transform/";
  static final String m_LotusXSLNameSpaceURL = "http://xsl.lotus.com/";

  // Create a hash table that can look up xsl element IDs via name.
  static Hashtable m_elementKeys = new Hashtable();
  
  /**
   * Init the XSLT hashtable.
   */
  void initXSLTKeys() 
  {
    // String pre = m_XSLNameSpaceURL+":";
    String pre = "";
    m_elementKeys.put((pre+Constants.ELEMNAME_APPLY_TEMPLATES_STRING).intern(), new Integer(Constants.ELEMNAME_APPLY_TEMPLATES));
    m_elementKeys.put((pre+Constants.ELEMNAME_PARAM_STRING).intern(), new Integer(Constants.ELEMNAME_PARAM));
    m_elementKeys.put((pre+Constants.ELEMNAME_CONSTRUCT_STRING).intern(), new Integer(Constants.ELEMNAME_CONSTRUCT));
    m_elementKeys.put((pre+Constants.ELEMNAME_CONTENTS_STRING).intern(), new Integer(Constants.ELEMNAME_CONTENTS));
    m_elementKeys.put((pre+Constants.ELEMNAME_COPY_STRING).intern(), new Integer(Constants.ELEMNAME_COPY));
    m_elementKeys.put((pre+Constants.ELEMNAME_COPY_OF_STRING).intern(), new Integer(Constants.ELEMNAME_COPY_OF));
    
    m_elementKeys.put((pre+Constants.ELEMNAME_ATTRIBUTESET_STRING).intern(), new Integer(Constants.ELEMNAME_DEFINEATTRIBUTESET));

    m_elementKeys.put((pre+Constants.ELEMNAME_USE_STRING).intern(), new Integer(Constants.ELEMNAME_USE));

    m_elementKeys.put((pre+Constants.ELEMNAME_VARIABLE_STRING).intern(), new Integer(Constants.ELEMNAME_VARIABLE));
    m_elementKeys.put((pre+Constants.ELEMNAME_PARAMVARIABLE_STRING).intern(), new Integer(Constants.ELEMNAME_PARAMVARIABLE));

    m_elementKeys.put((pre+Constants.ELEMNAME_DISPLAYIF_STRING).intern(), new Integer(Constants.ELEMNAME_DISPLAYIF));
    m_elementKeys.put((pre+Constants.ELEMNAME_EMPTY_STRING).intern(), new Integer(Constants.ELEMNAME_EMPTY));
    m_elementKeys.put((pre+Constants.ELEMNAME_EVAL_STRING).intern(), new Integer(Constants.ELEMNAME_EVAL));
    m_elementKeys.put((pre+Constants.ELEMNAME_CALLTEMPLATE_STRING).intern(), new Integer(Constants.ELEMNAME_CALLTEMPLATE));
    m_elementKeys.put((pre+Constants.ELEMNAME_TEMPLATE_STRING).intern(), new Integer(Constants.ELEMNAME_TEMPLATE));
    m_elementKeys.put((pre+Constants.ELEMNAME_STYLESHEET_STRING).intern(), new Integer(Constants.ELEMNAME_STYLESHEET));
    m_elementKeys.put((pre+Constants.ELEMNAME_TRANSFORM_STRING).intern(), new Integer(Constants.ELEMNAME_STYLESHEET));
    m_elementKeys.put((pre+Constants.ELEMNAME_IMPORT_STRING).intern(), new Integer(Constants.ELEMNAME_IMPORT));
    m_elementKeys.put((pre+Constants.ELEMNAME_INCLUDE_STRING).intern(), new Integer(Constants.ELEMNAME_INCLUDE));
    m_elementKeys.put((pre+Constants.ELEMNAME_FOREACH_STRING).intern(), new Integer(Constants.ELEMNAME_FOREACH));
    m_elementKeys.put((pre+Constants.ELEMNAME_VALUEOF_STRING).intern(), new Integer(Constants.ELEMNAME_VALUEOF));
    m_elementKeys.put((pre+Constants.ELEMNAME_KEY_STRING).intern(), new Integer(Constants.ELEMNAME_KEY));
    m_elementKeys.put((pre+Constants.ELEMNAME_STRIPSPACE_STRING).intern(), new Integer(Constants.ELEMNAME_STRIPSPACE));
    m_elementKeys.put((pre+Constants.ELEMNAME_PRESERVESPACE_STRING).intern(), new Integer(Constants.ELEMNAME_PRESERVESPACE));
    m_elementKeys.put((pre+Constants.ELEMNAME_NUMBER_STRING).intern(), new Integer(Constants.ELEMNAME_NUMBER));
    m_elementKeys.put((pre+Constants.ELEMNAME_IF_STRING).intern(), new Integer(Constants.ELEMNAME_IF));
    m_elementKeys.put((pre+Constants.ELEMNAME_CHOOSE_STRING).intern(), new Integer(Constants.ELEMNAME_CHOOSE));
    m_elementKeys.put((pre+Constants.ELEMNAME_WHEN_STRING).intern(), new Integer(Constants.ELEMNAME_WHEN));
    m_elementKeys.put((pre+Constants.ELEMNAME_OTHERWISE_STRING).intern(), new Integer(Constants.ELEMNAME_OTHERWISE));
    m_elementKeys.put((pre+Constants.ELEMNAME_TEXT_STRING).intern(), new Integer(Constants.ELEMNAME_TEXT));
    m_elementKeys.put((pre+Constants.ELEMNAME_ELEMENT_STRING).intern(), new Integer(Constants.ELEMNAME_ELEMENT));
    m_elementKeys.put((pre+Constants.ELEMNAME_ATTRIBUTE_STRING).intern(), new Integer(Constants.ELEMNAME_ATTRIBUTE));
    m_elementKeys.put((pre+Constants.ELEMNAME_SORT_STRING).intern(), new Integer(Constants.ELEMNAME_SORT));
    m_elementKeys.put((pre+Constants.ELEMNAME_PI_STRING).intern(), new Integer(Constants.ELEMNAME_PI));
    m_elementKeys.put((pre+Constants.ELEMNAME_COMMENT_STRING).intern(), new Integer(Constants.ELEMNAME_COMMENT));
   
    m_elementKeys.put((pre+Constants.ELEMNAME_COUNTER_STRING).intern(), new Integer(Constants.ELEMNAME_COUNTER));
    m_elementKeys.put((pre+Constants.ELEMNAME_COUNTERS_STRING).intern(), new Integer(Constants.ELEMNAME_COUNTERS));
    m_elementKeys.put((pre+Constants.ELEMNAME_COUNTERINCREMENT_STRING).intern(), new Integer(Constants.ELEMNAME_COUNTERINCREMENT));
    m_elementKeys.put((pre+Constants.ELEMNAME_COUNTERRESET_STRING).intern(), new Integer(Constants.ELEMNAME_COUNTERRESET ));
    m_elementKeys.put((pre+Constants.ELEMNAME_COUNTERSCOPE_STRING).intern(), new Integer(Constants.ELEMNAME_COUNTERSCOPE ));
    
    m_elementKeys.put((pre+Constants.ELEMNAME_APPLY_IMPORTS_STRING).intern(), new Integer(Constants.ELEMNAME_APPLY_IMPORTS ));
    
    m_elementKeys.put((pre+Constants.ELEMNAME_EXTENSION_STRING).intern(), new Integer(Constants.ELEMNAME_EXTENSION));

    m_elementKeys.put((pre+Constants.ELEMNAME_MESSAGE_STRING).intern(), new Integer(Constants.ELEMNAME_MESSAGE));
    m_elementKeys.put((pre+Constants.ELEMNAME_LOCALE_STRING).intern(), new Integer(Constants.ELEMNAME_LOCALE));

    // m_elementKeys.put(m_LotusXSLNameSpaceURL+":"+Constants.ELEMNAME_EXTENSIONHANDLER_STRING, new Integer(Constants.ELEMNAME_EXTENSIONHANDLER));

    m_elementKeys.put(m_LotusXSLNameSpaceURL+":"+Constants.ELEMNAME_CSSSTYLECONVERSION_STRING, new Integer(Constants.ELEMNAME_CSSSTYLECONVERSION));
  }
  
  /**
   * Node to xsl directive table, used so we don't have to keep 
   * looking up the directives each time.  This isn't of huge 
   * benifit, but is used because of the cost of getExpandedElementName.
   */
  Hashtable m_XSLDirectiveLookup = new Hashtable();
 
  /**
   * Given an XSL tag name, return an integer token
   * that corresponds to ELEMNAME_XXX constants defined 
   * in Constants.java.
   * Note: I tried to optimize this by caching the node to 
   * id lookups in a hash table, but it helped not a bit.
   * I'm not sure that it's spending too much time here anyway.
   * @param node a probable xsl:xxx element.
   * @return Constants.ELEMNAME_XXX token, or -1 if in xsl 
   * or lotusxsl namespace, -2 if not in known namespace.
   */
  final int getXSLToken(Node node)
    throws XSLProcessorException
  {
    if(Node.ELEMENT_NODE != node.getNodeType())
      return -2;
    int tok;
    Integer i;
    if(null == (i= (Integer)m_XSLDirectiveLookup.get(node)))
    {
      String ns = m_parserLiaison.getNamespaceOfNode(node);
      if((null != ns) && ns.toLowerCase().equals(m_XSLNameSpaceURL))
      {
        String localName = m_parserLiaison.getLocalNameOfNode(node);
        i = (Integer)m_elementKeys.get(localName);
        if(null == i)
        {
          i = new Integer(-2);
          tok = -2;
        }
        else
        {
          tok = i.intValue();
        }
      }
      else
      {
        i = new Integer(-2);
        tok = -2;
      }
      m_XSLDirectiveLookup.put(node, i);
    }
    else
    {
      tok = i.intValue();
    }
    return tok;
    
    /* NO DIRECT HASH LOOKUP OPTIMIZATION 
    Integer i = (Integer) 
                 m_elementKeys.get(m_parserLiaison.getExpandedElementName((Element)node));
    return (null != i) ? i.intValue() : -2;
    */
  }

  
  /**
   * Find the type of an element using this method. This is slower 
   * than it ought to be... I'll find a way to optimize down the 
   * line if need be.
   * @param node a probable xsl:xxx element.
   * @param tagType Constants.ELEMNAME_XXX token.
   * @return true if node is of tagType.
   */
  final boolean isXSLTagOfType(Node node, int tagType)
    throws XSLProcessorException
  {
    return (getXSLToken(node) == tagType);
  }

  
  /**
   * If the template element is in the xsl:namespace, process the
   * element.
   * @return true if the given XSL template child is an XSL directive.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected boolean processXSLTemplateInstruction(
                                              Stylesheet stylesheetTree, 
                                              Element xslInstruction, 
                                              Document sourceTree, 
                                              Node sourceNode,
                                              String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    boolean isXSLInstruction = true;
        
    int xslToken = getXSLToken(xslInstruction);
    switch(xslToken)
    {
    case Constants.ELEMNAME_APPLY_TEMPLATES:
    case Constants.ELEMNAME_FOREACH:
      handleApplyTemplatesInstruction(
                             stylesheetTree, xslInstruction, 
                             sourceTree, sourceNode, mode, xslToken);
      break;
      
    case Constants.ELEMNAME_APPLY_IMPORTS:
      handleApplyImportInstruction(
                                   stylesheetTree, xslInstruction, 
                                   sourceTree, sourceNode, mode, xslToken);
      break;
            
    case Constants.ELEMNAME_EMPTY:
      // Don't do anything, that's what it's all about!
      // Also, not supported in latest XSL draft.
      break;
      
    case Constants.ELEMNAME_VALUEOF:
      // xsl:value-of
      handleValueofInstruction(
                             stylesheetTree, xslInstruction,
                             sourceTree, sourceNode);
      break;

    case Constants.ELEMNAME_NUMBER:
      // xsl:number
      handleNumberInstruction(
                            stylesheetTree, xslInstruction,
                            sourceTree, sourceNode);
      break;
      
    case Constants.ELEMNAME_COUNTER:
    case Constants.ELEMNAME_COUNTERS:
      // handleCounterInstruction(
      //                      stylesheetTree, xslInstruction,
      //                      sourceTree, sourceNode);
      error("xsl:counters no longer supported by the XSL draft!");
      break;
      
    case Constants.ELEMNAME_COUNTERRESET:
      // handleCounterResetInstruction(
      //                      stylesheetTree, xslInstruction,
      //                      sourceTree, sourceNode);
      error("xsl:counter-reset no longer supported by the XSL draft!");
      break;

    case Constants.ELEMNAME_COUNTERINCREMENT:
      // handleCounterIncrementInstruction(
      //                      stylesheetTree, xslInstruction,
      //                      sourceTree, sourceNode);
      error("xsl:counter-increment no longer supported by the XSL draft!");
      break;

    case Constants.ELEMNAME_COUNTERSCOPE:
      {
        // m_counters.push(m_emptyCounter);
        
        // processTemplateElementChildren(stylesheetTree, xslInstruction, sourceTree, 
        //                               sourceNode, mode);
        
        // m_counters.pop();
        error("xsl:counter-scope no longer supported by the XSL draft!");
      }
      break;
      
    case Constants.ELEMNAME_EVAL:
      warn(xslInstruction, sourceNode, "eval no longer supported!");
      break;
      
    case Constants.ELEMNAME_CALLTEMPLATE:
      // Process a named template.
      handleCallTemplateInstruction(
                            stylesheetTree, xslInstruction, 
                            sourceTree, sourceNode, mode);
      break;

    case Constants.ELEMNAME_VARIABLE:
      handleVariableInstruction(
                            stylesheetTree, xslInstruction, 
                            sourceTree, sourceNode, mode);
       break;

    case Constants.ELEMNAME_PARAMVARIABLE:
      handleParamVariableInstruction(
                            stylesheetTree, xslInstruction, 
                            sourceTree, sourceNode, mode);
      break;
      
    case Constants.ELEMNAME_IF:
      // xsl:if
      // Only display contents if a given condition is true.
      handleIfInstruction(
               stylesheetTree, xslInstruction, 
               sourceTree, sourceNode, mode);
      break;

    case Constants.ELEMNAME_CHOOSE:
      // Handle xsl:choose / xsl:while / xsl:otherwise
      handleChooseInstruction(
                   stylesheetTree, xslInstruction, 
                   sourceTree, sourceNode, mode);
      break;

    case Constants.ELEMNAME_COPY_OF:
      handleCopyOfInstruction(stylesheetTree, 
                          xslInstruction, sourceTree, sourceNode, mode);
      break;

    case Constants.ELEMNAME_COPY:
      handleCopyInstruction(stylesheetTree, 
                          xslInstruction, sourceTree, sourceNode, mode);
      break;

    case Constants.ELEMNAME_TEXT:
      // xsl:text
      handleTextInstruction(
               stylesheetTree, xslInstruction, 
               sourceTree, sourceNode, mode);
      break;

    case Constants.ELEMNAME_USE:
      handleUseInstruction(
               stylesheetTree, xslInstruction, 
               sourceTree, sourceNode, mode);
      break;

    case Constants.ELEMNAME_ATTRIBUTE:
      handleAttributeInstruction(
                                 stylesheetTree, xslInstruction, 
                                 sourceTree, sourceNode, mode, m_pendingAttributes);
      break;

    case Constants.ELEMNAME_ELEMENT:
      // NON-STANDARD
      // Only display contents if a given condition is 
      // true.  The WG will probably do something like this.
      handleElementInstruction(
               stylesheetTree, xslInstruction, 
               sourceTree, sourceNode, 
               mode);
      break;
      
    case Constants.ELEMNAME_PI:
      // NON-STANDARD
      // Only display contents if a given condition is 
      // true.  The WG will probably do something like this.
      handlePIInstruction(
               stylesheetTree, xslInstruction, 
               sourceTree, sourceNode, mode);
      break;

    case Constants.ELEMNAME_COMMENT:
      handleCommentInstruction(
               stylesheetTree, xslInstruction, 
               sourceTree, sourceNode, mode);
      break;
      
    case Constants.ELEMNAME_CALL:
      // handle lotusxsl:call
      error("lotusxsl:call no longer supported, please use XSLT function extensions");
      break;
            
    case Constants.ELEMNAME_PARAM:
      // No action, these are handled within the context of 
      // handleCallInstruction.
      break;

    case Constants.ELEMNAME_SORT:
      // No action, these are handled within the context of 
      // apply-templates and for-each.
      break;

    case Constants.ELEMNAME_MESSAGE:
      handleMessageInstruction(
               stylesheetTree, xslInstruction, 
               sourceTree, sourceNode, mode);
      break;
      
    case Constants.ELEMNAME_LOCALE:
    case Constants.ELEMNAME_DEFINEATTRIBUTESET:
    case Constants.ELEMNAME_DEFINESCRIPT:
    case Constants.ELEMNAME_EXTENSION:
    case Constants.ELEMNAME_EXTENSIONHANDLER:
    case Constants.ELEMNAME_KEY:
    case Constants.ELEMNAME_IMPORT:
    case Constants.ELEMNAME_INCLUDE:
    case Constants.ELEMNAME_PRESERVESPACE:
    case Constants.ELEMNAME_TEMPLATE:
    case Constants.ELEMNAME_STRIPSPACE:
    case Constants.ELEMNAME_STYLESHEET:
      error(xslInstruction, sourceNode, 
            xslInstruction.getNodeName()+" not allowed inside a template!");
      break;

    default:
      // it is a template element
      isXSLInstruction = false;
    } // end switch
    return isXSLInstruction;
  }  
  
  /**
   * In response to 'xsl:process', 'xsl:process-children', and
   * 'xsl:for-each' tags, call transformChildren.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleApplyTemplatesInstruction(
                                        Stylesheet stylesheetTree, 
                                        Element templateChild, 
                                        Document sourceTree, 
                                        Node sourceNode,
                                        String mode, int xslToken)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {    
    // String selectPattern 
    //  = getAttrVal(templateChild, Constants.ATTRNAME_SELECT, sourceNode);
    Expression selectPattern 
      = getExpression(templateChild, Constants.ATTRNAME_SELECT, sourceNode);
    
    if((null == selectPattern) && (Constants.ELEMNAME_FOREACH == xslToken))
    {
      error(templateChild, sourceNode, "xsl:for-each must have a select attribute!");
    }
    
    if(null != sourceNode)
    {
      String modeValue = getAttrVal(templateChild, Constants.ATTRNAME_MODE);
      
      if(null != modeValue)
      {
        mode = modeValue;
      }
      boolean needToTurnOffInfiniteLoopCheck = false;
      
      Element template = (xslToken == Constants.ELEMNAME_FOREACH) ? templateChild : null;
      transformSelectedChildren(stylesheetTree, 
                                templateChild,
                                template,
                                sourceTree, 
                                sourceNode, mode,
                                selectPattern, xslToken);
      
      if(true == needToTurnOffInfiniteLoopCheck)
      {
        m_needToCheckForInfiniteLoops = false;
      }
    }
    else // if(null == sourceNode)
    {
      error(templateChild, sourceNode, 
            "sourceNode is null in handleApplyTemplatesInstruction!");
    }
  }
  
  /**
   * In response to 'xsl:import', call transformChild, or, if that 
   * fails, transformChildren.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleApplyImportInstruction(
                                        Stylesheet stylesheetTree, 
                                        Element templateChild, 
                                        Document sourceTree, 
                                        Node sourceNode,
                                        String mode, int xslToken)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException, SAXException
  {        
    if(null != sourceNode)
    {
      boolean didFindTemplate = transformChild(
                     stylesheetTree, null, null, 
                     sourceTree, 
                     m_parserLiaison.getParentOfNode(sourceNode), sourceNode, mode, xslToken);
      
      if(false == didFindTemplate)
      {
        handleApplyTemplatesInstruction(stylesheetTree, templateChild, 
                                        sourceTree, sourceNode,
                                        mode, Constants.ELEMNAME_APPLY_TEMPLATES);
      }
    }
    else // if(null == sourceNode)
    {
      error(templateChild, sourceNode, 
            "sourceNode is null in handleApplyTemplatesInstruction!");
    }
  }
  
  /**
   * Take the contents of a template element, process it, and 
   * convert it to a string.
   */
  private String templateElementChildrenToString(Stylesheet stylesheetTree, 
                                    Element templateChild, 
                                    Document sourceTree, 
                                    Node sourceNode,
                                    String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    DocumentHandler savedFListener = m_flistener;
    StringWriter sw = new StringWriter();
    m_flistener = new FormatterToText(sw);
    
    String savedPendingName = m_pendingElementName;
    m_pendingElementName = null;
    AttributeListImpl savedPendingAttributes = m_pendingAttributes;
    m_pendingAttributes = new AttributeListImpl();
    
    processTemplateElementChildren(stylesheetTree, templateChild, sourceTree, 
                                   sourceNode,
                                   mode);
    
    m_pendingElementName = savedPendingName;
    m_pendingAttributes = savedPendingAttributes;
    m_flistener = savedFListener;
    
    return sw.toString();
  }


  /**
   * Given an xsl:attribute instruction, construct the 
   * attribute and add it to the output element (resultParent).
   * @exception XSLProcessorException thrown if the active ProblemListener 
   * and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleAttributeInstruction(
                                          Stylesheet stylesheetTree, 
                                          Element templateChild, 
                                          Document sourceTree, 
                                          Node sourceNode,
                                          String mode,
                                          AttributeListImpl attList)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    String attrName = getProcessedAttrVal(templateChild, 
                                          Constants.ATTRNAME_NAME, 
                                          sourceNode);
    if(null != attrName)
    {
      String attrNameSpace = getProcessedAttrVal(templateChild, 
                                                 Constants.ATTRNAME_NAMESPACE, 
                                                 sourceNode);
      if(null != attrNameSpace)
      {
        String prefix = getResultPrefixForNamespace(attrNameSpace);
        if(null == prefix)
        {
          prefix = "ns"+String.valueOf(uniqueNSValue);
          uniqueNSValue++;
          String nsDecl = "xmlns:"+prefix;
          addResultAttribute(attList, nsDecl, attrNameSpace);
        }
        attrName = (prefix + ":"+attrName);
      }
      
      if(null != m_pendingElementName)
      {
        String val = templateElementChildrenToString(stylesheetTree, 
                                                     templateChild, sourceTree, 
                                                     sourceNode, mode);
        addResultAttribute(attList, attrName, val);
      }
      else
      {
        // warn(templateChild, sourceNode, "Trying to add attribute after element child has been added, ignoring...");
      }
    }
    else
    {
      warn(templateChild, sourceNode, "Could not process xsl:attribute because name attribute was not found!");
    }
  }
  
  /**
   * Handle an 'call-template' directive by finding the 
   * named templates that corrisponds to the
   * invoke tag's 'name' attribute, then call 
   * processTemplateElement on each of the macro's 
   * children.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleCallTemplateInstruction(
                                       Stylesheet stylesheetTree, 
                                       Element templateChild, 
                                       Document sourceTree, 
                                       Node sourceNode, String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    String templateName = getProcessedAttrVal(templateChild, 
                                           Constants.ATTRNAME_NAME, 
                                           sourceNode);

    if(null != templateName)
    {

      Element template = this.m_stylesheetRoot.findNamedTemplate(templateName);
      if(null != template)
      {
        m_variableStacks.pushContextMarker(template, sourceNode);
        
        m_variableStacks.pushParams(stylesheetTree, 
                   templateChild, 
                   sourceTree, 
                   sourceNode, mode, template);
                        
        buildResultFromTemplate(stylesheetTree, 
                                template, 
                                sourceTree, 
                                sourceNode, mode);
        
        m_variableStacks.popCurrentContext();
      }
    }
    else
    {
      warn(templateChild, sourceNode, 
              Constants.ELEMNAME_CALLTEMPLATE_STRING +
              " is required to have a " +
              Constants.ATTRNAME_NAME +
              " attribute!");
    }
  }  
  
  /**
   * In response to lotusxsl-code-dispatcher PI, register the dispatcher 
   * for the given type.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleCodeDispatcherPI(Element extensionElem)
    throws XSLProcessorException
  {
    String classid = getAttrVal(extensionElem, Constants.ATTRNAME_CLASSID);
    String codetype = getAttrVal(extensionElem, Constants.ATTRNAME_CODETYPE);
    
    // The following aren't used yet, but will be...
    // String archive = extensionElem.getAttribute(Constants.ATTRNAME_ARCHIVE);
    // String codebase = extensionElem.getAttribute(Constants.ATTRNAME_CODEBASE);
    
    registerExtensionHandlerByName(classid, codetype);
  }  
  
  /**
   * Handle the xsl:counter-increment instruction.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleCounterIncrementInstruction(
                                       Stylesheet stylesheetTree,
                                       Element xslCounterNode,
                                       Document sourceTree, 
                                       Node sourceNode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {    
    Element counter = (Element)xslCounterNode;
    String name = getAttrVal(counter, Constants.ATTRNAME_NAME, sourceNode);
    if(null != name)
    {
      XSLResultTreeCounter resultTreeCounter 
        = getResultTreeCounter(name, xslCounterNode, /* resultParent, */ true, false);
      
      resultTreeCounter.incrementCounter(xslCounterNode, sourceNode);
    }
    else
    {
      warn(xslCounterNode, sourceNode, 
              "Can't find 'name' attribute for "+xslCounterNode.getNodeName());
    }
    
  } // end handleCounterIncrementInstruction method

  /**
   * Handle the xsl:counter-reset instruction.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleCounterResetInstruction(
                                       Stylesheet stylesheetTree,
                                       Element xslCounterNode,
                                       Document sourceTree, 
                                       Node sourceNode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {    
    Element counter = (Element)xslCounterNode;
    String name = getAttrVal(counter, Constants.ATTRNAME_NAME, sourceNode);
    if(null != name)
    {
      XSLResultTreeCounter resultTreeCounter 
        = getResultTreeCounter(name, xslCounterNode, true, true);
      
      resultTreeCounter.resetCounter(xslCounterNode, sourceNode);
    }
    else
    {
      warn(xslCounterNode, sourceNode, 
              "Can't find 'name' attribute for "+xslCounterNode.getNodeName());
    }
    
  } // end handleCounterResetInstruction method

  /**
   * Handle the xsl:counter or xsl:counters directive, and add the output to
   * resultParent.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleCounterInstruction(
                                       Stylesheet stylesheetTree,
                                       Element xslCounterNode,
                                       Document sourceTree, 
                                       Node sourceNode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {    
    Element counter = (Element)xslCounterNode;
    String name = getAttrVal(counter, Constants.ATTRNAME_NAME, sourceNode);
    if(null != name)
    {
      XSLResultTreeCounter resultTreeCounter 
        = getResultTreeCounter(name, xslCounterNode, true, false);
      
      if(null == resultTreeCounter)
      {
        warn(xslCounterNode, sourceNode, 
                "A result tree counter could not be found with the name: "+name);
      }
      else
      {
        String countString = resultTreeCounter.getCountString(xslCounterNode, sourceNode);
        characters(countString.toCharArray(), 0, countString.length());
      }
    }
    else
    {
      warn(xslCounterNode, sourceNode, 
              "Can't find 'name' attribute for "+xslCounterNode.getNodeName());
    }
    
  } // end handleCounterInstruction method
    
  /**
   * Given an xsl:comment instruction, construct the 
   * comment and add it to the output element (resultParent).
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleCommentInstruction(
                                        Stylesheet stylesheetTree, 
                                        Element templateChild, 
                                        Document sourceTree, 
                                        Node sourceNode, String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
      String data = templateElementChildrenToString(stylesheetTree, 
                                                   templateChild, sourceTree, 
                                                   sourceNode, mode);
      comment(data);
  }
  
  
  /**
   * Given an xsl:choose element, evaluate the 
   * enclosed xsl:when / xsl:otherwise statements.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleChooseInstruction(
                              Stylesheet stylesheetTree, 
                              Element templateChild, 
                              Document sourceTree, 
                              Node sourceNode,
                              String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    NodeList children = templateChild.getChildNodes();
    int nNodes = children.getLength();
    for(int i = 0; i < nNodes; i++)
    {
      Node child = children.item(i);
      if(isXSLTagOfType(child, Constants.ELEMNAME_WHEN))
      {
        if(handleIfInstruction(
                    stylesheetTree, (Element)child, 
                    sourceTree, sourceNode, mode))
        {
          break; // from for loop
        }
      }
      else if(isXSLTagOfType(child, Constants.ELEMNAME_OTHERWISE))
      {
        processTemplateElementChildren(stylesheetTree, (Element)child, sourceTree, 
                                       sourceNode, mode);
        break;
      }
    }
  }

    /**
   * Handle the xsl:copy-of directive.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  void handleCopyOfInstruction(Stylesheet stylesheetTree, 
                             Element xslCopyOfElement, 
                             Document sourceTree, 
                             Node sourceNode,
                             String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    // Get the list of patterns associated with the 
    // <value-of> element.
    Expression selectPattern 
      = getExpression(xslCopyOfElement, Constants.ATTRNAME_SELECT, sourceNode);
    
    if(null == selectPattern)
    {
      error(xslCopyOfElement, sourceNode, 
              "No select attribute found for xsl:copy-of!");
    }
    else
    {
      Object value = m_queryEngine.getExpr(selectPattern, sourceNode, 
                                           xslCopyOfElement, m_contextNodeList);
      
      if(null != value)
      {
        int type = m_queryEngine.getClassType(value);
        String s;
        switch(type)
        {
        case m_queryEngine.CLASS_BOOLEAN:
        case m_queryEngine.CLASS_NUMBER:
        case m_queryEngine.CLASS_STRING:
          s = m_queryEngine.stringCast(value);
          characters(s.toCharArray(), 0, s.length());
          break;
          
        case m_queryEngine.CLASS_NODESET:
          // System.out.println(value);
          Vector nl = m_queryEngine.nodelistCast(value);
          int nChildren = nl.size();
          for(int i = 0; i < nChildren; i++)
          {
            Node pos = (Node)nl.elementAt(i);
            Node top = pos;
            while(null != pos)
            {
              flushPending();
              cloneToResultTree( pos, false, false, false, true );
              Node nextNode = pos.getFirstChild();
              while(null == nextNode)
              {
                if(Node.ELEMENT_NODE == pos.getNodeType())
                {
                  endElement(pos.getNodeName());
                }
                if(top == pos)
                  break;
                nextNode = pos.getNextSibling();
                if(null == nextNode)
                {
                  pos = pos.getParentNode();
                  if(top == pos)
                  {
                    if(Node.ELEMENT_NODE == pos.getNodeType())
                    {
                      endElement(pos.getNodeName());
                    }
                    nextNode = null;
                    break;
                  }
                }
              }
              pos = nextNode;
            }
          }
          break;
          
        case m_queryEngine.CLASS_RTREEFRAG:
          outputResultTreeFragment(value);
          break;
        }
      }
    }
  } // end handleCopyOfInstruction method
  
  /**
   * Handle the xsl:copy directive.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  void handleCopyInstruction(Stylesheet stylesheetTree, 
                             Element xslInstruction, 
                             Document sourceTree, 
                             Node sourceNode,
                             String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    // The root node is treated specially because the root node of the 
    // result tree is created implicitly. When the current node is the root node, 
    // xsl:copy will not create a root node, but will just use the content template.
        
    int nodeType = sourceNode.getNodeType();
    if((Node.DOCUMENT_NODE != nodeType))
    {
      cloneToResultTree( sourceNode, false, false, false, false );
      
      if(Node.ELEMENT_NODE == nodeType)
      {
        copyNamespaceAttributes(sourceNode, false, m_pendingAttributes);
        processTemplateElementChildren(stylesheetTree, xslInstruction, sourceTree, 
                                       sourceNode, mode);
        endElement(sourceNode.getNodeName());
      }
    }
    else
    {
      processTemplateElementChildren(stylesheetTree, xslInstruction, sourceTree, 
                                     sourceNode, mode);
    }
    
  } // end handleCopyInstruction method

  /**
   * Given an xsl:element instruction, construct the 
   * element and add it to the output element (resultParent).
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleElementInstruction(
                                        Stylesheet stylesheetTree, 
                                        Element templateChild, 
                                        Document sourceTree, 
                                        Node sourceNode, String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    String name = this.getProcessedAttrVal(templateChild, 
                                           Constants.ATTRNAME_NAME, 
                                           sourceNode);
    if(null != name)
    {
      String ns = getProcessedAttrVal(templateChild, 
                                                 Constants.ATTRNAME_NAMESPACE, 
                                                 sourceNode);
      String nsDecl = null;
      if(null != ns)
      {
        // String prefix = getResultPrefixForNamespace(ns);
        // if(null == prefix)
        // {
        // prefix = "ns"+String.valueOf(uniqueNSValue);
        // uniqueNSValue++;
        // nsDecl = "xmlns:"+prefix;
        nsDecl = "xmlns";
        // }
        // name = (prefix + ":"+name);
      }

      // TODO: Process xsl:attribute children
      startElement(name);
      
      if(null != nsDecl)
      {
        addResultAttribute(m_pendingAttributes, nsDecl, ns);
      }
      
      processTemplateElementChildren(stylesheetTree, templateChild, sourceTree, 
                                     sourceNode, mode);
      
      endElement(name);
    }
    else
    {
      warn(templateChild, sourceNode, 
              "The name attribute is required for xsl:element!");
    }
  }  
  
  /**
   * In response to xsl:function, set up a function that can 
   * be called from the expression language.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleFunctionsInstruction(Element extensionElem)
    throws XSLProcessorException
  {
    Attr nsPrefixAttr = extensionElem.getAttributeNode(Constants.ATTRNAME_NS);
    String namespace = null;
    if(null == nsPrefixAttr)
    {
      warn(extensionElem, null, 
           "xsl:functions must have a '"+Constants.ATTRNAME_NS+"' attribute!");
    }
    else
    {
      namespace = m_parserLiaison.getNamespaceForPrefix(nsPrefixAttr.getValue(), 
                                                        extensionElem);
    }

    String codetype = getAttrVal(extensionElem, Constants.ATTRNAME_CODETYPE);
    // String functions = getAttrVal(extensionElem, Constants.ATTRNAME_METHOD);
    
    Dispatcher extension = null;
    Object obj = m_dispatcherFactories.get(codetype);
    if(null == obj)
    {
      if(null == codetype)
      {
        // Assume Java, I guess
        codetype = "application/java";
      }
      
      if(codetype.equals("text/javascript"))
      {
        registerExtensionHandlerByName("com.lotus.xsl.dispatchers.JavaScriptDispatcher", 
                                       codetype);
        obj = m_dispatcherFactories.get(codetype);
      }
      else if(codetype.equals("application/java"))
      {
        registerExtensionHandlerByName("com.lotus.xsl.dispatchers.JavaDispatcher", 
                                       codetype);
        obj = m_dispatcherFactories.get(codetype);
      }
    }
    
    if(null != obj)
    {
      DispatcherFactory factory = (DispatcherFactory)obj;
      extension = factory.create(extensionElem, null);
    }
    else
    {
      warn(extensionElem, null, 
           "Could not get factory for codetype "+codetype+"!");
    }
    
    if(extension != null)
    {
      this.m_extensionNamespaces.put(namespace, extension);
    }
    else
    {
      warn(extensionElem, null, 
           "can't register extension handler for namespace: "+namespace);
    }
  }
    
  /**
   * Given an xsl:if directive, evaluate the 
   * condition and execute child elements 
   * if the condition is true.
   * @return true if test selected a non-empty list.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected boolean handleIfInstruction(
                             Stylesheet stylesheetTree, 
                             Element templateChild, 
                             Document sourceTree, 
                             Node sourceNode,
                             String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    boolean show = false;

    String selectPattern 
      = getAttrVal(templateChild, Constants.ATTRNAME_TEST, sourceNode);
    if(null != selectPattern)
    {
      // System.out.println("xsl:if test="+selectPattern);
      Object test = m_queryEngine.getExpr(selectPattern, sourceNode, templateChild, m_contextNodeList);
      show = m_queryEngine.booleanCast(test);
      if(show)
      {
        processTemplateElementChildren(stylesheetTree, templateChild, sourceTree, 
                                       sourceNode, mode);
      }
    }
    else
    {
      warn(templateChild, sourceNode, 
              "did not find test attribute value on if or when directive.");
    }
    
    return show;
  }
  
  /**
   * Given an xsl:message instruction, construct the 
   * text and output it through the warning mechanism.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleMessageInstruction(
                                        Stylesheet stylesheetTree, 
                                        Element templateChild, 
                                        Document sourceTree, 
                                        Node sourceNode, String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
      String data = templateElementChildrenToString(stylesheetTree, 
                                                   templateChild, sourceTree, 
                                                   sourceNode, mode);
      warn(data);
  }
      
  /**
   * Handle the xsl:number directive, and add the output to
   * resultParent.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   * @see m_sourceTreeCounters
   */
  protected void handleNumberInstruction(
                                       Stylesheet stylesheetTree,
                                       Element xslNumberElement,
                                       Document sourceTree, 
                                       Node sourceNode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {    
    Object val = m_sourceTreeCounters.get(xslNumberElement);
    XSLSourceTreeCounter counter;
    if(null == val)
    {
      counter = new XSLSourceTreeCounter(stylesheetTree, xslNumberElement);
      m_sourceTreeCounters.put(xslNumberElement, counter);
    }
    else
    {
      counter = (XSLSourceTreeCounter)val;
    }
    
    String countString = counter.getCountString(sourceTree, sourceNode);
    
    characters(countString.toCharArray(), 0, countString.length());
    
  } // end handleNumberInstruction method

  /**
   * Given an xsl:pi instruction, construct the 
   * processing instruction and add it to the output element (resultParent).
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handlePIInstruction(
                                        Stylesheet stylesheetTree, 
                                        Element templateChild, 
                                        Document sourceTree, 
                                        Node sourceNode, String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    String name = this.getProcessedAttrVal(templateChild, 
                                           Constants.ATTRNAME_NAME, 
                                           sourceNode);
    if(null != name)
    {
      String val = templateElementChildrenToString(stylesheetTree, 
                                                   templateChild, sourceTree, 
                                                   sourceNode, mode);
      processingInstruction(name, val);
    }
    else
    {
      warn(templateChild, sourceNode, 
              "The name attribute is required for xsl:pi!");
    }
    
  }
  
  /**
   * Given an xsl:text element, copy the contained text to 
   * the result tree.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleTextInstruction(
                            Stylesheet stylesheetTree, 
                            Element templateChild, 
                            Document sourceTree, 
                            Node sourceNode,
                            String mode)
    throws XSLProcessorException, SAXException,
           MalformedURLException,
           FileNotFoundException,
           IOException
  {
    NodeList children = templateChild.getChildNodes();    
    int nNodes = children.getLength();
    for(int i = 0; i < nNodes; i++) 
    {
      Node childNode = children.item(i);
      int childType = childNode.getNodeType();
      if ((childType == Node.CDATA_SECTION_NODE) || (childType == Node.TEXT_NODE)) 
      {
        cloneToResultTree(childNode, 
                               true, // isLiteral
                               false, // clone children
                               true, // overrideStrip
                               false
                               );
      }
    } // end for
  } // end handleTextInstruction method
  
  
  /**
   * Given an xsl:attribute instruction, construct the 
   * attribute and add it to the output element (resultParent).
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleUseInstruction(
                                          Stylesheet stylesheetTree, 
                                          Element templateChild, 
                                          Document sourceTree, 
                                          Node sourceNode,
                                          String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    String attrSetUseVal = getAttrVal(templateChild, 
                                          Constants.ATTRNAME_ATTRIBUTE_SET, 
                                          sourceNode);
    if(null != attrSetUseVal)
    {
      m_stylesheetRoot.findAndUseAttrSets(attrSetUseVal, m_pendingAttributes, 
                                          sourceNode, stylesheetTree);
    }
    else
    {
      warn(templateChild, sourceNode, 
              "Could not process xsl:use because '"+
              Constants.ATTRNAME_ATTRIBUTE_SET+"' attribute was not found!");
    }
  }
  
  
  /**
   * Handle the 'variable' directive, and add the value to the variables stack.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleVariableInstruction(
                                           Stylesheet stylesheetTree,
                                           Element xslVariableElement,
                                           Document sourceTree, 
                                           Node sourceNode, String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    
    String name = getAttrVal(xslVariableElement, Constants.ATTRNAME_NAME);

    if(null != name)
    {
      String value = getAttrVal(xslVariableElement, 
                                Constants.ATTRNAME_EXPR, sourceNode);
      Object var;
      if(null != value)
      {
        var = m_queryEngine.getExpr(value, sourceNode, xslVariableElement, m_contextNodeList);
      }
      else
      {
        // Use result tree fragment
        var = createResultTreeFrag(stylesheetTree, xslVariableElement, 
                                   sourceTree, sourceNode, mode);
      }

      m_variableStacks.pushVariable(name, var, xslVariableElement.getParentNode());
    }
    else
    {
      error("xsl:variable must have a name attribute!");
    }
  }
  
  /**
   * Handle the 'variable' directive, and add the value to the variables stack.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleParamVariableInstruction(
                                                Stylesheet stylesheetTree,
                                                Element xslVariableElement,
                                                Document sourceTree, 
                                                Node sourceNode, String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {    
    String name = getAttrVal(xslVariableElement, Constants.ATTRNAME_NAME);

    if(null != name)
    {        
        Object obj = m_variableStacks.getParamVariable(name);
        
        if(null == obj)
        {
          handleVariableInstruction(
                                    stylesheetTree,
                                    xslVariableElement,
                                    sourceTree, 
                                    sourceNode, mode);
        }
    }
    else
    {
      error("xsl:param-variable must have a name attribute!");
    }
  }
 
  /**
   * Handle the 'value-of' directive, and add the output to
   * resultParent.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void handleValueofInstruction(
                                        Stylesheet stylesheetTree,
                                        Element xslValueofElement,
                                        Document sourceTree, 
                                        Node sourceNode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    // Get the list of patterns associated with the 
    // <value-of> element.
    // String selectPattern 
    //  = getAttrVal(xslValueofElement, Constants.ATTRNAME_SELECT, sourceNode);
    Expression exp 
      = getExpression(xslValueofElement, Constants.ATTRNAME_SELECT, sourceNode);

    if(null == exp)
    {
      error(xslValueofElement, sourceNode, 
              "No select attribute found for xsl:value-of!");
    }
    else
    {
      Object value = m_queryEngine.getExpr(exp, sourceNode, 
                                           xslValueofElement, m_contextNodeList);
      
      if(null != value)
      {
        int type = m_queryEngine.getClassType(value);
        String s;
        switch(type)
        {
        case m_queryEngine.CLASS_BOOLEAN:
        case m_queryEngine.CLASS_NUMBER:
        case m_queryEngine.CLASS_NODESET:
        case m_queryEngine.CLASS_STRING:
        case m_queryEngine.CLASS_RTREEFRAG:
          s = m_queryEngine.stringCast(value);
          if(null != s)
            characters(s.toCharArray(), 0, s.length());
          break;
        }
      }
    }
  }	
    
  //==========================================================
  // SECTION: Diagnostic functions
  //==========================================================
  private void ____ERROR_AND_DIAGNOSTIC____(){}
  
  /**
   * If this is set to true, do not warn about pattern 
   * match conflicts.
   */
  private boolean m_quietConflictWarnings = false;

  /*
  * If this is true, then the diag function will 
  * be called.
  */
  private boolean m_traceTemplateChildren = false;
  
  /*
  * If this is true, then the simple tracing of templates 
  * will be performed.
  */
  private boolean m_traceTemplates = false;

  /*
  * If this is true, then diagnostics of each select 
  * will be performed.
  */
  private boolean m_traceSelects = false;

  /*
  * A stream to print diagnostics to.
  */
  private java.io.PrintWriter m_diagnosticsPrintWriter = null;

  /* For diagnostics */
  Hashtable m_durationsTable = new Hashtable();

  /**
   * Warn the user of an problem.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public void warn(String msg)
    throws XSLProcessorException
  {
    warn(null, null, msg);
  }

  /**
   * Warn the user of an problem.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public void warn(Node styleNode, Node sourceNode, String msg)
    throws XSLProcessorException
  {
    boolean shouldThrow = m_problemListener.problem(ProblemListener.XSLPROCESSOR, 
                              ProblemListener.WARNING,
                              styleNode, sourceNode, msg, 0, 0);
    if(shouldThrow)
    {
      throw new XSLProcessorException(msg);
    }
  }

  /**
   * Tell the user of an error, and probably throw an 
   * exception.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public void error(String msg)
    throws XSLProcessorException
  {
    error(null, null, msg);
  }

  /**
   * Tell the user of an error, and probably throw an 
   * exception.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public void error(Node styleNode, Node sourceNode, String msg)
    throws XSLProcessorException
  {
    boolean shouldThrow = m_problemListener.problem(ProblemListener.XSLPROCESSOR, 
                              ProblemListener.ERROR,
                              styleNode, sourceNode, msg, 0, 0);
    if(shouldThrow)
    {
      throw new XSLProcessorException(msg);
    }
  }

  /**
   * Mark the time, so that displayDuration can later
   * display the elapse.
   */
  public void pushTime(Object key)
  {
    if(null != key)
    {
      m_durationsTable.put(key, new Date());
    }
  }
  
  /**
   * Returns the duration since pushTime was called,
   * in milliseconds.
   */
  public long popDuration(Object key)
  {
    long millisecondsDuration = 0;
    if(null != key)
    {
      Date date1 = (Date)m_durationsTable.get(key);
      Date date2 = new Date();
      millisecondsDuration = date2.getTime() - date1.getTime();
      m_durationsTable.remove(key);
    }
    return millisecondsDuration;
  }

  /**
   * Display the duration since pushTime was called.
   */
  protected void displayDuration(String info, Object key)
  {
    long millisecondsDuration = 0;
    if(null != key)
    {
      Date date1 = (Date)m_durationsTable.get(key);
      Date date2 = new Date();
      millisecondsDuration = date2.getTime() - date1.getTime();
      if(null != m_diagnosticsPrintWriter)
      {
        m_diagnosticsPrintWriter.println(info + " took " + millisecondsDuration + " milliseconds");
      }
      m_durationsTable.remove(key);
    }
  }

  /**
   * If this is set, diagnostics will be 
   * written to the m_diagnosticsPrintWriter stream. If 
   * the value is null, then diagnostics will be turned 
   * off.
   */
  public void setDiagnosticsOutput(java.io.PrintWriter pw)
  {
    m_diagnosticsPrintWriter = pw;
  }
  
  /**
   * Bottleneck output of diagnostics.
   */
  protected void diag(String s)
  {
    if(null != m_diagnosticsPrintWriter)
    {
      m_diagnosticsPrintWriter.println(s);
    }
  }
  
  /**
   * If this is set to true, simple traces of 
   * template calls are made.
   */
  public void setTraceTemplates(boolean b)
  {
    m_traceTemplates = b;
  }

  /**
   * If this is set to true, simple traces of 
   * template calls are made.
   */
  public void setTraceSelect(boolean b)
  {
    m_traceSelects = b;
  }
  
  /**
   * If this is set to true, debug diagnostics about 
   * template children as they are being constructed 
   * will be written to the m_diagnosticsPrintWriter 
   * stream.  diagnoseTemplateChildren is false by
   * default.
   */
  public void setTraceTemplateChildren(boolean b)
  {
    m_traceTemplateChildren = b;
  }
  
  /**
   * If the quietConflictWarnings property is set to 
   * true, warnings about pattern conflicts won't be 
   * printed to the diagnostics stream.
   * True by default.
   * @param b true if conflict warnings should be suppressed.
   */
  public void setQuietConflictWarnings(boolean b)
  {
    m_quietConflictWarnings = b;
  }
  
  /**
   * Print a trace of a selection being made.
   */
  protected void traceSelect(Element template, NodeList nl)
    throws XSLProcessorException
  {
    String msg = template.getNodeName()+": ";
    Attr attr = template.getAttributeNode(Constants.ATTRNAME_SELECT);
    if(null != attr)
    {
      msg += attr.getValue() + ", "+((null != nl) ? nl.getLength() : 0) + " selected";
    }
    else
    {
      msg += "*|text(), (default select), "+nl.getLength() + " selected";
    }    
    
    attr = template.getAttributeNode(Constants.ATTRNAME_MODE);
    if(null != attr)
    {
      msg += ", mode = "+attr.getValue();
    }
    System.out.println(msg);
  }

  
  /**
   * Print a trace of a template that is being called, either by 
   * a match, name, or as part of for-each.
   */
  protected void traceTemplate(Element template)
    throws XSLProcessorException
  {
    String msg;
    Attr attr = template.getAttributeNode(Constants.ATTRNAME_MATCH);
    if(null != attr)
    {
      msg = "Calling template for: " + attr.getValue();
    }
    else
    {
      attr = template.getAttributeNode(Constants.ATTRNAME_NAME);
      if(null != attr) 
      {
        msg = "Calling named template, name = " + attr.getValue();
      }
      else
      {
        int xslToken = getXSLToken(template);

        if(Constants.ELEMNAME_FOREACH == xslToken)
        {
          attr = template.getAttributeNode(Constants.ATTRNAME_SELECT);
          if(null != attr) 
          {
            msg = "Processing for-each, select = " + attr.getValue();
          }
          else
          {
            error("for-each must have either a match or name attribute");
            msg = null;
          }
        }
        else
        {
          error("templates must have either a match or name attribute");
          msg = null;
        }
      }
    }
    attr = template.getAttributeNode(Constants.ATTRNAME_MODE);
    if(null != attr)
    {
      msg += ", mode = "+attr.getValue();
    }
    System.out.println(msg);
  }
  
  /**
   * Print some diagnostics about the current 
   * template child.
   */
  protected void diagnoseTemplateChildren(
                                     Node templateChild,
                                     Node sourceNode)
  {
    if(m_traceTemplateChildren)
    {
      String templateChildTagName = templateChild.getNodeName();

      String xmlElemName = sourceNode.getNodeName();
      
      diag("source node: " + xmlElemName
           + ", template-node: " 
           + templateChildTagName);
    }
  }
  
  //==========================================================
  // SECTION: Functions that create nodes in the result tree
  //==========================================================
  private void ____RESULT_TREE_CREATION___(){}
    
  /**
   * To fullfill the FormatterListener interface... not action
   * for the moment.
   */
  public void setDocumentLocator (Locator locator)
  {
  }

  /**
   * Bottleneck the startDocument event.
   */
  public void startDocument ()
    throws SAXException
  {
    m_variableStacks.pushContextMarker(null, null);
    try
    {
      resolveTopLevelParams();
    }
    catch(Exception e)
    {
      throw new SAXException(e);
    }
    m_counters.push(m_emptyCounter);
    m_flistener.startDocument();
  }

  /**
   * Bottleneck the endDocument event.
   */
  public void endDocument ()
    throws SAXException
  {
    flushPending();
    m_flistener.endDocument();
    m_counters.pop();
    m_variableStacks.popCurrentContext();
  }
    
  /**
   * The pending element.  We have to delay the call to 
   * m_flistener.startElement(name, atts) because of the 
   * xsl:attribute and xsl:copy calls.  In other words, 
   * the attributes have to be fully collected before you 
   * can call startElement.
   */
  protected String m_pendingElementName = null;

  /**
   * The pending attributes.  We have to delay the call to 
   * m_flistener.startElement(name, atts) because of the 
   * xsl:attribute and xsl:copy calls.  In other words, 
   * the attributes have to be fully collected before you 
   * can call startElement.
   */
  protected AttributeListImpl m_pendingAttributes = new AttributeListImpl();
  
  /**
   * A stack to keep track of the result tree namespaces.
   */
  protected Stack m_resultNameSpaces = new Stack();
  
  /**
   * This is pushed on the m_resultNameSpaces stack 'till a 
   * xmlns attribute is found.
   */
  private final ResultNameSpace m_emptyNamespace = new ResultNameSpace(null, null);
  
  /**
   * Bottleneck addition of result tree attributes, so I can keep 
   * track of namespaces.
   */
  protected void addResultAttribute(AttributeListImpl attList, String aname, String value)
  {
    boolean isPrefix = aname.startsWith("xmlns:");
    if (aname.equals("xmlns") || isPrefix) 
    {
      String p = isPrefix ? aname.substring(6) : "";
      ResultNameSpace ns = new ResultNameSpace(p, value);
      if(!m_resultNameSpaces.isEmpty())
      {
        ResultNameSpace nsOnStack = (ResultNameSpace)m_resultNameSpaces.peek();
        if(m_emptyNamespace == nsOnStack)
        {
          m_resultNameSpaces.setElementAt(ns, 
                                          m_resultNameSpaces.size() - 1);
        }
        else
        {
          while(nsOnStack.m_next != null)
          {
            nsOnStack = nsOnStack.m_next;
          }
          nsOnStack.m_next = ns;
        }
      }
    }
    
    if(null == attList.getValue(aname))
    {
      attList.addAttribute(aname, "CDATA", value);
    }
  }
  
  /**
   * Flush the pending element.
   */
  protected void flushPending()
    throws SAXException
  {
    if(null != m_pendingElementName)
    {
      m_flistener.startElement(m_pendingElementName, m_pendingAttributes);
      m_pendingAttributes.clear();
      m_pendingElementName = null;
    }
  }

  /**
   * Bottleneck the startElement event.
   */
  public void startElement (String name)
    throws SAXException
  {
    flushPending();
    m_counters.push(m_emptyCounter);
    m_resultNameSpaces.push(m_emptyNamespace);
    m_pendingElementName = name;
  }

  /**
   * Bottleneck the startElement event.
   */
  public void startElement (String name, AttributeList atts)
    throws SAXException
  {
    flushPending();
    int nAtts = atts.getLength();
    m_pendingAttributes.clear();
    for(int i = 0; i < nAtts; i++)
    {
      m_pendingAttributes.addAttribute(atts.getName(i), atts.getType(i),
                                       atts.getValue(i));
    }
    m_counters.push(m_emptyCounter);
    m_resultNameSpaces.push(m_emptyNamespace);
    m_pendingElementName = name;
  }

  /**
   * Bottleneck the endElement event.
   */
  public void endElement (String name)
    throws SAXException
  {
    flushPending();
    m_flistener.endElement(name);
    m_resultNameSpaces.pop();
    m_counters.pop();
  }

  /**
   * Bottleneck the characters event.
   */
  public void characters (char ch[], int start, int length)
    throws SAXException
  {
    flushPending();
    m_flistener.characters(ch, start, length);
  }

  /**
   * Bottleneck the ignorableWhitespace event.
   */
  public void ignorableWhitespace (char ch[], int start, int length)
    throws SAXException
  {
    flushPending();
    m_flistener.ignorableWhitespace(ch, start, length);
  }

  /**
   * Bottleneck the processingInstruction event.
   */
  public void processingInstruction (String target, String data)
    throws SAXException
  {
    flushPending();
    m_flistener.processingInstruction(target, data);
  }

  /**
   * Bottleneck the comment event.
   */
  public void comment(String data) throws SAXException
  {
    flushPending();
    if(m_flistener instanceof FormatterListener)
    {
      ((FormatterListener)m_flistener).comment(data);
    }
  }

  /**
   * Bottleneck the comment event.
   */
  public void entityReference(String name) throws SAXException
  {
    flushPending();
    if(m_flistener instanceof FormatterListener)
    {
      ((FormatterListener)m_flistener).entityReference(name);
    }
  }

  /**
   * Bottleneck the cdata event.
   */
  public void cdata (char ch[], int start, int length)
    throws SAXException
  {
    flushPending();
    if(m_flistener instanceof FormatterListener)
    {
      ((FormatterListener)m_flistener).cdata(ch, start, length);
    }
    else
    {
      // Bad but I think it's better than dropping it.
      m_flistener.characters(ch, start, length);
    }
  }
  
  /**
   * Clone an element with or without children.
   * TODO: Fix or figure out node clone failure!
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void cloneToResultTree(Node node, 
                                        boolean isLiteral,
                                        boolean shouldCloneWithChildren, 
                                        boolean overrideStrip,
                                        boolean shouldCloneAttributes)
    throws XSLProcessorException, SAXException,
           MalformedURLException,
           FileNotFoundException,
           IOException
  {
    boolean stripWhiteSpace = false;
    
    switch(node.getNodeType())
    {
    case Node.TEXT_NODE:
      {
        // If stripWhiteSpace is false, then take this as an override and 
        // just preserve the space, otherwise use the XSL whitespace rules.
        if(!overrideStrip)
        {
          stripWhiteSpace = isLiteral ? true : shouldStripSourceNode(node);
        }
        Text tx = (Text)node;
        String data = null;
        // System.out.println("stripWhiteSpace = "+stripWhiteSpace+", "+tx.getData());
        if(stripWhiteSpace)
        {
          if(!m_parserLiaison.isIgnorableWhitespace(tx))
          {
            data = getNormalizedText(tx);
            if((null != data) && (0 == data.trim().length()))
            {
              data = null;
            }
          }
        }
        else 
        {
          Node parent = node.getParentNode();
          if(null != parent)
          {
            if( Node.DOCUMENT_NODE != parent.getNodeType())
            {
              data = getNormalizedText(tx);
              if((null != data) && (0 == data.length()))
              {
                data = null;
              }
            }
          }
          else
          {
            data = getNormalizedText(tx);
            if((null != data) && (0 == data.length()))
            {
              data = null;
            }
          }
        }
        
        if(null != data)
        {
          // TODO: Hack around the issue of comments next to literals.
          // This would be, when a comment is present, the whitespace 
          // after the comment must be added to the literal.  The 
          // parser should do this, but XML4J doesn't seem to.
          // <foo>some lit text
          //     <!-- comment -->   
          //     </foo>
          // Loop through next siblings while they are comments, then, 
          // if the node after that is a ignorable text node, append 
          // it to the text node just added.
          if(m_parserLiaison.isIgnorableWhitespace(tx))
          {
            ignorableWhitespace(data.toCharArray(), 0, data.length());
          }
          else
          {
            characters(data.toCharArray(), 0, data.length());
          }
        }
      }
      break;
    case Node.ELEMENT_NODE:
      {
        AttributeList atts;
        if(shouldCloneAttributes)
        {
          copyAttributesToAttList( node, m_stylesheetRoot, (Element)node, m_pendingAttributes );
          copyNamespaceAttributes(node, false, m_pendingAttributes);
        }
        startElement (node.getNodeName());
      }
      break;
    case Node.CDATA_SECTION_NODE:
      {
        String data = ((CDATASection)node).getData();
        cdata(data.toCharArray(), 0, data.length());
      }
      break;
    case Node.ATTRIBUTE_NODE:
      {
        addResultAttribute(m_pendingAttributes, ((Attr)node).getName(),
                                         ((Attr)node).getValue());
      }
      break;
    case Node.COMMENT_NODE:
      {
        comment(((Comment)node).getData());
      }
      break;
    case Node.DOCUMENT_FRAGMENT_NODE:
      {
        error(null, node, "No clone of a document fragment!");
      }
      break;
    case Node.ENTITY_REFERENCE_NODE:
      {
        EntityReference er = (EntityReference)node;
        entityReference(er.getNodeName());
      }
      break;
    case Node.PROCESSING_INSTRUCTION_NODE:
      {
        ProcessingInstruction pi = (ProcessingInstruction)node;
        processingInstruction(pi.getTarget(), pi.getData());
      }
      break;
    default:
      error("Can not create item in result tree: "+node.getNodeName());
    }
        
  } // end cloneToResultTree function

  /**
   * Given a stylesheet element, create a result tree fragment from it's 
   * contents.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   * @param stylesheetTree The stylesheet object that holds the fragment.
   * @param templateChild The template element that holds the fragment.
   * @param sourceTree The source tree document context.
   * @param sourceNode The current source context node.
   * @param mode The mode under which the template is operating.
   * @return An object that represents the result tree fragment.
   */
  Object createResultTreeFrag(
                              Stylesheet stylesheetTree, 
                              Element templateChild, 
                              Document sourceTree, 
                              Node sourceNode, String mode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    Object result;
    if(m_resolveContentsEarly)
    {
      flushPending();
      // If resolve node list early...          
      DocumentHandler savedFormatterListener = m_flistener;
      
      DocumentFragment resultFragment = createDocFrag();
      
      m_flistener = new FormatterToDOM(m_resultTreeFactory, resultFragment);
      
      processTemplateElementChildren(stylesheetTree, templateChild, 
                                     sourceTree, sourceNode, mode);
      
      flushPending();
      m_flistener = savedFormatterListener;
      result = resultFragment;
    }
    else
    {
      // If resolve nodelist late...
      ContextState context = new ContextState(templateChild, sourceNode);
      result = context;
    }
    return result;
  }
  
  /**
   * Given a result tree fragment, walk the tree and
   * output it to the result stream.
   */
  void outputResultTreeFragment(Object obj)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    DocumentFragment docFrag = m_queryEngine.rtreefragCast(obj);
    NodeList nl = docFrag.getChildNodes();
    int nChildren = nl.getLength();
    for(int i = 0; i < nChildren; i++)
    {
      Node pos = nl.item(i);
      Node top = pos;
      while(null != pos)
      {
        flushPending();
        cloneToResultTree( pos, false, false, false, true );
        Node nextNode = pos.getFirstChild();
        while(null == nextNode)
        {
          if(Node.ELEMENT_NODE == pos.getNodeType())
          {
            endElement(pos.getNodeName());
          }
          if(top == pos)
            break;
          nextNode = pos.getNextSibling();
          if(null == nextNode)
          {
            pos = pos.getParentNode();
            if((top == pos) || (null == pos))
            {
              if(null != pos)
              {
                if(Node.ELEMENT_NODE == pos.getNodeType())
                {
                  endElement(pos.getNodeName());
                }
                nextNode = null;
              }
              else
              {
                nextNode = null;
              }
              break;
            }
          }
        }
        pos = nextNode;
      }
    }
  }
  
  //==========================================================
  // SECTION: Function to do with attribute handling
  //==========================================================
  private void ____ATTRIBUTES___(){}
  
  /**
   * This is used whenever a unique namespace is needed.
   */
  private int uniqueNSValue = 0;
  
  /**
   * Given a prefix, return the namespace, 
   */
  String getResultNamespaceForPrefix(String prefix)
  {
    int type;
    String namespace = null;
    
    int nParents = m_resultNameSpaces.size();
    for(int i = (nParents - 1); i >= 0 && (null == namespace); i--) 
    {
      ResultNameSpace rns = (ResultNameSpace)m_resultNameSpaces.elementAt(i);
      if (m_emptyNamespace != rns) 
      {
        while(null != rns)
        {
          if(rns.m_prefix.equals(prefix))
          {
            namespace = rns.m_uri;
            break;
          }
          rns = rns.m_next;
        }
      }
    }
    return namespace;
  }
  
  /**
   * This should probably be in the XMLParserLiaison interface.
   */
  private String getResultPrefixForNamespace(String namespace)
  {
    int type;
    String prefix = null;
    
    int nParents = m_resultNameSpaces.size();
    for(int i = (nParents - 1); i >= 0 && (null == prefix); i--) 
    {
      ResultNameSpace rns = (ResultNameSpace)m_resultNameSpaces.elementAt(i);
      if (m_emptyNamespace != rns) 
      {
        while(null != rns)
        {
          if(rns.m_uri.equals(namespace))
          {
            prefix = rns.m_prefix;
            break;
          }
          rns = rns.m_next;
        }
      }
    }
    return prefix;
  }
  
  /**
   * This should probably be in the XMLParserLiaison interface.
   */
  private String getPrefixForNamespace(String namespace, Element namespaceContext)
  {
    int type;
    Node parent = namespaceContext;
    String prefix = null;
    while ((null != parent) && (null == prefix)
           && (((type = parent.getNodeType()) == Node.ELEMENT_NODE)
               || (type == Node.ENTITY_REFERENCE_NODE))) 
    {
      if (type == Node.ELEMENT_NODE) 
      {
        NamedNodeMap nnm = parent.getAttributes();
        for (int i = 0;  i < nnm.getLength();  i ++) 
        {
          Node attr = nnm.item(i);
          String aname = attr.getNodeName();
          boolean isPrefix = aname.startsWith("xmlns:");
          if (aname.equals("xmlns") || isPrefix) 
          {
            int index = aname.indexOf(':');
            String namespaceOfPrefix = attr.getNodeValue();
            if((null != namespaceOfPrefix) && namespaceOfPrefix.equals(namespace))
            {
              prefix = isPrefix ? aname.substring(index+1) : "";
            }
          }
        }
      }
      
      parent = m_parserLiaison.getParentOfNode(parent);
    }
    return prefix;
  }

  /**
   * Copy <KBD>xmlns:</KBD> attributes in if not already in scope.
   */
  public void copyNamespaceAttributes(Node src, boolean srcIsStylesheetTree,
                                         AttributeListImpl destination) 
  {
    int type;
    // Vector nameValues = null;
    // Vector alreadyProcessedPrefixes = null;
    Node parent = src;
    while (parent != null
           && ((type = parent.getNodeType()) == Node.ELEMENT_NODE
               || (type == Node.ENTITY_REFERENCE_NODE))) 
    {
      if (type == Node.ELEMENT_NODE) 
      {
        NamedNodeMap nnm = parent.getAttributes();
        int nAttrs = nnm.getLength();
        for (int i = 0;  i < nAttrs;  i++) 
        {
          Node attr = nnm.item(i);
          String aname = attr.getNodeName();
          boolean isPrefix = aname.startsWith("xmlns:");
          if (aname.equals("xmlns") || isPrefix) 
          {
            String prefix = isPrefix ? aname.substring(6) : "";
            String desturi = getResultNamespaceForPrefix(prefix);
            String srcURI = attr.getNodeValue();
            boolean isXSLNS = (srcIsStylesheetTree && srcURI.equalsIgnoreCase(m_XSLNameSpaceURL));
            if(srcURI.startsWith("quote:"))
            {
              srcURI = srcURI.substring(6);
            }
            if(!srcURI.equalsIgnoreCase(desturi) && !isXSLNS)
            {
              addResultAttribute(m_pendingAttributes, aname, srcURI);
            }
          }
        }
      }
      parent = parent.getParentNode();
    }
  }
  
  /**
   * A table of attribute nodes that hold selects or the like, that 
   * maps to Expression objects.
   */
  Hashtable m_expressions = new Hashtable();
  
  /**
   * Get an Expression object from a select node.
   */
  private final Expression getExpression(Element el, String key, Node contextNode)
  {
    Attr a = el.getAttributeNode(key);
    Object expObj;
    if(null != a)
    {
      Integer lookupKey = new Integer(((Object)el).hashCode()+key.hashCode());
      expObj = m_expressions.get(lookupKey);
      if(null == expObj)
      {
        // String exprStr = a.getValue();
        String val = null == a ? null : a.getValue();
        if(null != val)
        {
          expObj = m_queryEngine.makeExpression(val, el);
          if(null != expObj)
          {
            m_expressions.put(lookupKey, expObj);
          }
        }
      }
    }
    else
    {
      expObj = null;
    }
    return (Expression)expObj;
  }
  
  public boolean m_useATVsInSelects = false;
  
  /**
   * Given an element, return an attribute value in 
   * the form of a string.
   * @param el The element from where to get the attribute.
   * @param key The name of the attribute.
   * @param contextNode The context to evaluate the 
   * attribute value template.
   * @return Attribute value.
   */
  private final String getAttrVal(Element el, String key, Node contextNode)
  {
    String val;
    if(m_useATVsInSelects)
    {
      try
      {
        val = getProcessedAttrVal(el, key, contextNode);
      }
      catch(Exception e)
      {
        val = null;
      }
    }
    else
    {
      Attr a = el.getAttributeNode(key);
      val = null == a ? null : a.getValue();
    }
    return val;
  }
  
  /**
   * Given an element, return an attribute value in 
   * the form of a string.
   */
  private static final String getAttrVal(Element el, String key)
  {
    Attr a = el.getAttributeNode(key);
    return null == a ? null : a.getValue();
  }

  /**
   * Given an element, return an attribute value in 
   * the form of a string, processing attributes as 
   * need be.
   * @param el The element from where to get the attribute.
   * @param key The name of the attribute.
   * @param contextNode The context to evaluate the 
   * attribute value template.
   * @return Attribute value.
   */
  private final String getProcessedAttrVal(Element el, String key, 
                                          Node contextNode)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    String val = null;
    Attr attr = el.getAttributeNode(key);
    if(null != attr)
    {
      String stringedValue = attr.getValue();
      if(null != stringedValue)
      {
        val = evaluateAttrVal(contextNode,
                              el, stringedValue);
      }
    }
    return val;
  }
  
  /**
   * An ugly little parser that evaluates attribute values for 
   * attribute templates (Stuff in curly {} braces that hold 
   * expressions).  Speed is pretty darned important with this 
   * function.
   * @param contextNode the current node in the source tree
   * @param stylesheetTree used as factory object for creating 
   * the pattern-by-example structures when parsing expressions.
   * @param attr used as a key for first attribute template 
   * parse structure cache.
   * @param stringedValue the attribute value to be processed.
   * @return Processed stringedValue with attribute templates
   * resolved.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected String evaluateAttrVal(Node contextNode,
                                   Element namespaceContext,
                                   String stringedValue)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    String expressedValue = null; // return value
    StringTokenizer tokenizer = 
                               new StringTokenizer(stringedValue, "{}\"\'", true);
    int nTokens = tokenizer.countTokens();
    if(nTokens < 2)
    {
      expressedValue = stringedValue; // then do the simple thing
    }
    else
    {
      StringBuffer buffer = new StringBuffer(32);
      String t = null; // base token
      String lookahead= null; // next token
      String error = null; // if non-null, break from loop
      while(tokenizer.hasMoreTokens())
      {
        if( lookahead != null )
        {
          t = lookahead;
          lookahead = null;
        }
        else t = tokenizer.nextToken();
        
        if(t.length() == 1)
        {
          switch(t.charAt(0))
          {
          case('\"'):
          case('\''):
            {
              // just keep on going, since we're not in an attribute template
              buffer.append(t);
              break;
            }
          case('{'):
            {
              // Attr template start
              lookahead = tokenizer.nextToken();
              if(lookahead.equals("{"))
              {
                // Double curlys mean escape to show curly
                buffer.append(lookahead);
                lookahead = null;
                break; // from switch
              }
              else if(lookahead.equals("\"") || lookahead.equals("\'"))
              {
                // Error. Expressions can not begin with quotes.
                error = "Expressions can not begin with quotes.";
                break; // from switch
              }
              else
              {
                String expression = lookahead; // Probably should make into StringBuffer
                while((null != lookahead) && (!lookahead.equals("}")))
                {
                  lookahead = tokenizer.nextToken();
                  if(lookahead.length() == 1)
                  {
                    switch(lookahead.charAt(0))
                    {
                    case '\'':
                    case '\"':
                      {
                        // String start
                        expression += lookahead;
                        String quote = lookahead;
                        // Consume stuff 'till next quote
                        lookahead = tokenizer.nextToken();
                        while(!lookahead.equals(quote))
                        {
                          expression += lookahead;
                          lookahead = tokenizer.nextToken();
                        }
                        expression += lookahead;
                        break;
                      }
                    case '{':
                      {
                        // What's another curly doing here?
                        error = "Error: Can not have \"{\" within expression.";
                        break;
                      }
                    case '}':
                      {
                        // Proper close of attribute template.
                        // Evaluate the expression.
                        String exprResult = m_queryEngine.getStringExpr(expression, 
                                                                        contextNode, 
                                                                        namespaceContext);
                        buffer.append(exprResult);
                        
                        lookahead = null; // breaks out of inner while loop
                        break;
                      }
                    default:
                      {
                        // part of the template stuff, just add it.
                        expression += lookahead;
                      }
                    } // end inner switch
                  } // end if lookahead length == 1
                  else
                  {
                    // part of the template stuff, just add it.
                    expression += lookahead;
                  }
                } // end while(!lookahead.equals("}"))
                if(error != null)
                {
                  break; // from inner while loop
                }
              }
              break;
            }
          case('}'):
            {
              lookahead = tokenizer.nextToken();
              if(lookahead.equals("}"))
              {
                // Double curlys mean escape to show curly
                buffer.append(lookahead);
                lookahead = null; // swallow
              }
              else
              {
                // Illegal, I think...
                warn("Found \"}\" but no attribute template open!");
                buffer.append("}");
                // leave the lookahead to be processed by the next round.
              }
              break;
            }
          default:
            {
              // Anything else just add to string.
              buffer.append(t);
            }
          } // end switch t
        } // end if length == 1
        else
        {
          // Anything else just add to string.
          buffer.append(t);
        }
        if(null != error)
        {
          warn("Attr Template, "+error);
          break;
        }
      } // end while(tokenizer.hasMoreTokens())
      
      expressedValue = buffer.toString();
      
    } // end else nTokens > 1
    return expressedValue;
  } // end evaluateAttrVal
  
  /**
   * Copy an attribute to the created output element, executing 
   * attribute templates as need be, and processing the xsl:use 
   * attribute.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void copyAttributeToTarget( Attr attr,
                                        Node contextNode,
                                        Stylesheet stylesheetTree,
                                        AttributeListImpl attrList, 
                                        Element namespaceContext
                                        )
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    String attrName = attr.getName().trim();
    
    String stringedValue = attr.getValue();
    
    stringedValue = evaluateAttrVal(contextNode,
                                    namespaceContext, stringedValue);
    
    // evaluateAttrVal might return a null value if the template expression 
    // did not turn up a result, in which case I'm going to not add the 
    // attribute.
    // TODO: Find out about empty attribute template expression handling.
    if(null != stringedValue)
    {
      if((attrName.equals("xmlns") || attrName.startsWith("xmlns:")) 
         && stringedValue.startsWith("quote:"))
      {
        stringedValue = stringedValue.substring(6);
      }
      addResultAttribute(attrList, attrName, stringedValue);
    }
  } // end copyAttributeToTarget method
  
  /**
   * Copy the attributes from the XSL element to the created 
   * output element, executing attribute templates and 
   * processing the xsl:use attribute as need be.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected void copyAttributesToAttList(
                                        Node contextNode,
                                        Stylesheet stylesheetTree,
                                        Element templateChild, AttributeListImpl attList)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    NamedNodeMap attributes = templateChild.getAttributes();
    int nAttributes = (null != attributes) ? attributes.getLength() : 0;
    String attrSetUseVal = null;
    for(int i = 0; i < nAttributes; i++)  
    {	
        Attr attr = (Attr)attributes.item(i);
        if(m_parserLiaison.getExpandedAttributeName(attr).equalsIgnoreCase(m_XSLNameSpaceURL+":use"))
        {
          attrSetUseVal = attr.getValue();
        }
        else
        {
          copyAttributeToTarget( attr, contextNode, stylesheetTree, 
                                 attList, templateChild);
        }
    } // end for(long i = 0; i < nAttributes; i++)

    if(null != attrSetUseVal)
    {
      m_stylesheetRoot.findAndUseAttrSets(attrSetUseVal, attList, 
                                          contextNode, stylesheetTree);
    }
  
  } // end copyAttributesToTarget method
  
  /**
  * Keyed on CSS2 property names, and holding dummy 
  * values for quickly looking up if a result tree element
  * attribute is a CSS attribute or not.
  * @deprecated
  */
  private Hashtable m_cssKeys = null;

  /*
  * If this is true, translate CSS attributes on 
  * the output to a "style" attribute.
  */
  private boolean m_translateCSS = false;

  /**
  * Translate CSS attributes on the output to a "style" attribute, for 
  * old submission support.  You probably shouldn't rely on this, though
  * I personally like it a lot.
  * @param b If true, translate CSS attributes on the output to a 
  * "style" attribute.
  * @deprecated
  */
  public void setTranslateCSS(boolean b) 
  {
    m_translateCSS = b;
  } 

  /**
   * Translate CSS attributes and put them in a style tag.
   * @deprecated
   */
  private void translateCSSAttrsToStyleAttr(AttributeListImpl attList)
  {
    if(m_translateCSS)
    {
      String styleAttrValueString = null;
      Vector toBeRemoved = new Vector();
      int nAttributes = attList.getLength();
      for(int i = 0; i < nAttributes; i++)
      {
        String attrName = attList.getName(i);
        if(isCSSAttribute(attrName))
        {
          toBeRemoved.addElement(attrName);
          if(null == styleAttrValueString)
          {
            styleAttrValueString = attrName+":"+attList.getValue(i);
          }
          else
          {
            styleAttrValueString += "; "+attrName+":"+attList.getValue(i);
          }
        }
      }
      int nAttrsToRemove = toBeRemoved.size();
      for(int i = 0; i < nAttrsToRemove; i++)
      {
        attList.removeAttribute((String)toBeRemoved.elementAt(i));
      }
      if(null != styleAttrValueString)
      {
        addResultAttribute(attList, "style", styleAttrValueString);
      }
    }
  }
  
  /**
   * Returns whether or not an attribute key 
   * is a CSS2 attribute.
   * @deprecated
   */
  protected boolean isCSSAttribute(String name)
  {
    if(null == m_cssKeys)
    {
      m_cssKeys = new Hashtable();
      initCSS2Table();
    }
    return (null != m_cssKeys.get(name));
  }

  /**
   * Stuff a hash table with CSS2 keys & dummy values
   * in order to quickly look up if a template child
   * attribute is a CSS attribute or not.
   * Note: This has something in it to screw up the
   * Dev Studio code format command, so that the 
   * last curly bracket goes all the way to the 
   * left.
   * @deprecated
   */
  protected void initCSS2Table()
  {
    Integer dummyVal = new Integer(1);
    
    m_cssKeys.put("azimuth", dummyVal);
    m_cssKeys.put("background", dummyVal);
    m_cssKeys.put("background-attachment", dummyVal);
    m_cssKeys.put("background-color", dummyVal);
    m_cssKeys.put("background-image", dummyVal);
    m_cssKeys.put("background-position", dummyVal);
    m_cssKeys.put("background-repeat", dummyVal);
    m_cssKeys.put("border", dummyVal);
    m_cssKeys.put("border-bottom", dummyVal);
    m_cssKeys.put("border-bottom-color", dummyVal);
    m_cssKeys.put("border-bottom-style", dummyVal);
    m_cssKeys.put("border-bottom-width", dummyVal);
    m_cssKeys.put("border-color", dummyVal);
    m_cssKeys.put("border-left", dummyVal);
    m_cssKeys.put("border-left-color", dummyVal);
    m_cssKeys.put("border-left-style", dummyVal);
    m_cssKeys.put("border-left-width", dummyVal);
    m_cssKeys.put("border-right", dummyVal);
    m_cssKeys.put("border-right-color", dummyVal);
    m_cssKeys.put("border-right-style", dummyVal);
    m_cssKeys.put("border-right-width", dummyVal);
    m_cssKeys.put("border-spacing", dummyVal);
    m_cssKeys.put("border-style", dummyVal);
    m_cssKeys.put("border-top", dummyVal);
    m_cssKeys.put("border-top-color", dummyVal);
    m_cssKeys.put("border-top-style", dummyVal);
    m_cssKeys.put("border-top-width", dummyVal);
    m_cssKeys.put("border-width", dummyVal);
    m_cssKeys.put("bottom", dummyVal);
    m_cssKeys.put("caption-side", dummyVal);
    m_cssKeys.put("clear", dummyVal);
    m_cssKeys.put("clip", dummyVal);
    m_cssKeys.put("color", dummyVal);
    m_cssKeys.put("column-span", dummyVal);
    m_cssKeys.put("content", dummyVal);
    m_cssKeys.put("cue", dummyVal);
    m_cssKeys.put("cue-after", dummyVal);
    m_cssKeys.put("cue-before", dummyVal);
    m_cssKeys.put("cursor", dummyVal);
    m_cssKeys.put("direction", dummyVal);
    m_cssKeys.put("display", dummyVal);
    m_cssKeys.put("elevation", dummyVal);
    m_cssKeys.put("float", dummyVal);
    m_cssKeys.put("font", dummyVal);
    m_cssKeys.put("font-family", dummyVal);
    m_cssKeys.put("font-size", dummyVal);
    m_cssKeys.put("font-size-adjust", dummyVal);
    m_cssKeys.put("font-style", dummyVal);
    m_cssKeys.put("font-variant", dummyVal);
    m_cssKeys.put("font-weight", dummyVal);
    m_cssKeys.put("height", dummyVal);
    m_cssKeys.put("left", dummyVal);
    m_cssKeys.put("letter-spacing", dummyVal);
    m_cssKeys.put("line-height", dummyVal);
    m_cssKeys.put("list-style", dummyVal);
    m_cssKeys.put("list-style-image", dummyVal);
    m_cssKeys.put("list-style-position", dummyVal);
    m_cssKeys.put("list-style-type", dummyVal);
    m_cssKeys.put("margin", dummyVal);
    m_cssKeys.put("margin-bottom", dummyVal);
    m_cssKeys.put("margin-left", dummyVal);
    m_cssKeys.put("margin-right", dummyVal);
    m_cssKeys.put("margin-top", dummyVal);
    m_cssKeys.put("marks", dummyVal);
    m_cssKeys.put("max-height", dummyVal);
    m_cssKeys.put("max-width", dummyVal);
    m_cssKeys.put("min-height", dummyVal);
    m_cssKeys.put("min-width", dummyVal);
    m_cssKeys.put("orphans", dummyVal);
    m_cssKeys.put("overflow", dummyVal);
    m_cssKeys.put("padding", dummyVal);
    m_cssKeys.put("padding-bottom", dummyVal);
    m_cssKeys.put("padding-left", dummyVal);
    m_cssKeys.put("padding-right", dummyVal);
    m_cssKeys.put("padding-top", dummyVal);
    m_cssKeys.put("page-break-after", dummyVal);
    m_cssKeys.put("page-break-before", dummyVal);
    m_cssKeys.put("pause", dummyVal);
    m_cssKeys.put("pause-after", dummyVal);
    m_cssKeys.put("pause-before", dummyVal);
    m_cssKeys.put("pitch", dummyVal);
    m_cssKeys.put("pitch-range", dummyVal);
    m_cssKeys.put("play-during", dummyVal);
    m_cssKeys.put("position", dummyVal);
    m_cssKeys.put("quotes", dummyVal);
    m_cssKeys.put("richness", dummyVal);
    m_cssKeys.put("right", dummyVal);
    m_cssKeys.put("row-span", dummyVal);
    m_cssKeys.put("size", dummyVal);
    m_cssKeys.put("speak", dummyVal);
    m_cssKeys.put("speak-date", dummyVal);
    m_cssKeys.put("speak-header", dummyVal);
    m_cssKeys.put("speak-numeral", dummyVal);
    m_cssKeys.put("speak-punctuation", dummyVal);
    m_cssKeys.put("speak-time", dummyVal);
    m_cssKeys.put("speech-rate", dummyVal);
    m_cssKeys.put("stress", dummyVal);
    m_cssKeys.put("table-layout", dummyVal);
    m_cssKeys.put("text-align", dummyVal);
    m_cssKeys.put("text-decoration", dummyVal);
    m_cssKeys.put("text-indent", dummyVal);
    m_cssKeys.put("text-shadow", dummyVal);
    m_cssKeys.put("text-transform", dummyVal);
    m_cssKeys.put("top", dummyVal);
    m_cssKeys.put("unicode-bidi", dummyVal);
    m_cssKeys.put("vertical-align", dummyVal);
    m_cssKeys.put("visibility", dummyVal);
    m_cssKeys.put("voice-family", dummyVal);
    m_cssKeys.put("volume", dummyVal);
    m_cssKeys.put("white-space", dummyVal);
    m_cssKeys.put("widows", dummyVal);
    m_cssKeys.put("width", dummyVal);
    m_cssKeys.put("word-spacing", dummyVal);
    m_cssKeys.put("z-index", dummyVal);
  }

  //==========================================================
  // SECTION: Functions for finding elements by Key and ID
  //==========================================================
  private void ____KEY_AND_ID_HANDLING____(){}

  /**
   * Given an id, and using our list of attribute names, 
   * look for the given element.
   */
  private void buildKeysTable(Node n, Element nscontext)
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException,
    SAXException
  {    
    do
    {
      // Walk through each of the declarations made with xsl:key
      int nDeclarations = m_keyDeclarations.size();
      
      // Since we're called by the query engine, we have to create 
      // our own, in order that the state of the caller not be messed 
      // up.  There's probably way's to fix this from the QueryEngine's 
      // end.
      QueryEngine tempQE = new QueryEngine(this);
      for(int i = 0; i < nDeclarations; i++)
      {
        KeyDeclaration kd = (KeyDeclaration)m_keyDeclarations.elementAt(i);
        
        // See if our node matches the given key declaration according to 
        // the match attribute on xsl:key.
        double score = tempQE.matchAgainstPatterns(kd.m_match, n, nscontext);
        if(score == tempQE.MATCH_SCORE_NONE)
          continue;

        // Query from the node, according the the select pattern in the
        // use attribute in xsl:key.
        NodeList nl = tempQE.query(kd.m_use, n, nscontext);
        if(null == nl)
          continue;
        
        // Use each node in the node list as a key value that we'll be 
        // able to use to look up the given node.
        int nUseValues = nl.getLength();
        for(int k = 0; k < nUseValues; k++)
        {
          Node useNode = nl.item(k);
          
          // Use getExpr to get the string value of the given node. I hope 
          // the string assumption is the right thing... I can't see how 
          // it could work any other way.
          String exprResult = getNodeData(useNode);
          
          if(null == exprResult)
            continue;
          
          // Look to see if there's already a table indexed by the 
          // name attribute on xsl:key.  If there's not, create one.
          Hashtable namedKeyTable;
          {
            Object keyTableObj = m_keys.get(kd.m_name);
            if(null == keyTableObj)
            {
              namedKeyTable = new Hashtable();
              m_keys.put(kd.m_name, namedKeyTable);
            }
            else
            {
              namedKeyTable = (Hashtable)keyTableObj;
            }
          }
          
          // Look to see if we already have row indexed by 
          // the node value, which is one of the nodes found via
          // the use attribute of xsl:key.  If there's not a row, 
          // create one.
          NodeListImpl keyNodes;
          {
            Object nodeListObj = namedKeyTable.get(exprResult);
            if(null == nodeListObj)
            {
              keyNodes = new NodeListImpl();
              namedKeyTable.put(exprResult, keyNodes);
            }
            else
            {
              keyNodes = (NodeListImpl)nodeListObj;
            }
          }
          
          // See if the matched node is already in the 
          // table set.  If it is there, we're done, otherwise 
          // add it.
          boolean foundit = false;
          int nKeyNodes = keyNodes.size();
          for(int j = 0; j < nKeyNodes; j++)
          {
            if(n == keyNodes.item(j))
            {
              foundit = true;
              break;
            }
          } // end for j
          if(!foundit)
          {
            keyNodes.addElement(n);
          }
        } // end for(int k = 0; k < nUseValues; k++)
      } // end for(int i = 0; i < nDeclarations; i++)
    }
      while((null != (n = getNextPrewalkInTree(n))));
    
  } // end buildKeysTable method
  
  /**
   * from any position in the tree, return the 
   * next node in the tree, assuming preorder 
   * traversal preceded, or null if at the end.
   */
  Node getNextPrewalkInTree(Node pos)
  {
    Node nextNode = null;
    nextNode = pos.getFirstChild();
    while(null == nextNode)
    {
      nextNode = pos.getNextSibling();
      if(null == nextNode)
      {
        pos = pos.getParentNode();
        if(pos == null)
        {
          break;
        }
      }
    }
    return nextNode;
  }
  
  /**
   * Given a valid element id, return the corresponding element.
   */
  NodeList getNodeSetByKey(Document doc, String name, String ref, Element nscontext)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    NodeList nl;
    if(null != m_keyDeclarations)
    {
      if(m_needToBuildKeysTable)
      {
        if(null == m_keys)
        {
          m_keys = new Hashtable();
          // TODO: The stylesheet.m_stylesheet root is incorrect, need to fix.
          buildKeysTable(doc, m_stylesheetRoot.m_stylesheet);
        }
      }
      if(null != m_keys)
      {
        // TODO: look up hashtable by name, 
        // look up the nodelist by the ref, and return the found node-set
        Object keyTable = m_keys.get(name);
        if(null != keyTable)
        {
          Object keyNodes = ((Hashtable)keyTable).get(ref);
          if(null != keyNodes)
          {
            nl = (NodeList)keyNodes;
          }
          else
          {
            // No warning here... should I have one??
            nl = new NodeListImpl();
          }
        }
        else
        {
          nl = null; // shutup compiler
          error("No keys were declared with the name of "+name);
        }
      }
      else
      {
        nl = null; // shutup compiler
        error("keys table was not built!");
      }
    }
    else
    {
      nl = null; // shutup compiler
      error("tried to reference a key, but no keys were declared.");
    }
    return nl;
  }
  
  /**
   * Given a valid element id, return the corresponding element.
   */
  Element getElementByID(String id, Document doc)
  {
    return m_parserLiaison.getElementByID(id, doc);
  }

  //==========================================================
  // SECTION: Functions for controling whitespace
  //==========================================================
  private void ____WHITESPACE____(){}

  /**
  * If this is true, the processor will do the best it can to strip 
  * unwanted white space. This is set in the stylesheet via the default-space 
  * attribute on xsl:stylesheet.
  */
  private boolean m_stripWhiteSpace = false; // default default-space="preserve"
  
  /**
   * A lookup table of all space preserving elements.
   */
  private Hashtable m_whitespacePreservingElements = new Hashtable();
  
  /**
   * A lookup table of all space stripping elements.
   */
  private Hashtable m_whitespaceStrippingElements = new Hashtable();
  
  /**
   * indent-result is by default no, which means an XSL processor must not
   * change the whitespace on output.
   */
  private boolean m_indentResult = false;
  
  /**
   * If true, output carriage returns.
   */
  private boolean m_outputCarriageReturns = false;

  /**
   * If true, output linefeeds.
   */
  private boolean m_outputLinefeeds = false;

  /**
   * This changes the default of the default-space attribute.
   * @param b Sets the default of the default-space attribute 
   * to "strip" if true, or "preserve" if false.
   */
  public void setStripWhiteSpace(boolean b) 
  {
    m_stripWhiteSpace = b;
  } 
    
  /**
   * Tells, through the combination of the default-space attribute 
   * on xsl:stylesheet, xsl:strip-space, xsl:preserve-space, and the
   * xml:space attribute, whether or not extra whitespace should be stripped 
   * from the node.  Literal elements from template elements should 
   * <em>not</em> be tested with this function.
   * @param textNode A text node from the source tree.
   * @return true if the text node should be stripped of extra whitespace.
   */
  public boolean shouldStripSourceNode(Node textNode)
  {
    boolean strip = m_stripWhiteSpace; // return value
    Node parent = m_parserLiaison.getParentOfNode(textNode);
    while(null != parent)
    {
      if(parent.getNodeType() == Node.ELEMENT_NODE)
      {
        Element parentElem = (Element)parent;
        
        Attr attr = parentElem.getAttributeNode("xml:space");
        if(null != attr)
        {
          String xmlSpaceVal = attr.getValue();
                                          
          if(xmlSpaceVal.equals("preserve"))
          {
            strip = false;
          }
          // else value of "default"
          break;
        }
        String tagName = parentElem.getTagName();
        if(m_whitespacePreservingElements.containsKey(tagName))
        {
          strip = false;
          break;
        }
        else if(m_whitespaceStrippingElements.containsKey(tagName))
        {
          strip = true;
          break;
        }
      }
      parent = parent.getParentNode();
    }
    return strip;
  }

  /**
   * Returns whether the specified <var>ch</var> conforms to the XML 1.0 definition
   * of whitespace.  Refer to <A href="http://www.w3.org/TR/1998/REC-xml-19980210#NT-S">
   * the definition of <CODE>S</CODE></A> for details.
   * @param   ch      Character to check as XML whitespace.
   * @return          =true if <var>ch</var> is XML whitespace; otherwise =false.
   */
  public static boolean isSpace(char ch) 
  {
    return Character.isWhitespace(ch); // Take the easy way out for now.
  }

  
  /**
   * (Code stolen and modified from XML4J)
   * Conditionally trim all leading and trailing whitespace in the specified String.  
   * All strings of white space are 
   * replaced by a single space character (#x20), except spaces after punctuation which 
   * receive double spaces if doublePunctuationSpaces is true.
   * This function may be useful to a formatter, but to get first class
   * results, the formatter should probably do it's own white space handling 
   * based on the semantics of the formatting object.
   * @param   string      String to be trimmed.
   * @param   trimHead    Trim leading whitespace?
   * @param   trimTail    Trim trailing whitespace?
   * @param   doublePunctuationSpaces    Use double spaces for punctuation?
   * @return              The trimmed string.
   */
  public String fixWhiteSpace(String string, 
                       boolean trimHead, 
                       boolean trimTail, 
                       boolean doublePunctuationSpaces) 
  {
    if(null == string) // defensive
      return ""; 
    char[] buf = string.toCharArray();
    int len = buf.length;
    boolean edit = false;
    int s;
    for (s = 0;  s < len;  s++) 
    {
      if (isSpace(buf[s]))  
      {
        break;
      }
    }
    /* replace S to ' '. and ' '+ -> single ' '. */
    int d = s;
    boolean pres = false;
    for ( ;  s < len;  s ++) 
    {
      char c = buf[s];
      if (isSpace(c)) 
      {
        if (!pres) 
        {
          if (' ' != c)  
          {
            edit = true;
          }
          buf[d++] = ' ';
          if(doublePunctuationSpaces && (s != 0))
          {
            char prevChar = buf[s-1];
            if(!((prevChar == '.') || (prevChar == '!') || (prevChar == '?')))
            {
              pres = true;
            }
          }
          else
          {
            pres = true;
          }
        }
        else
        {
          edit = true;
          pres = true;
        }
      }
      else 
      {
        buf[d++] = c;
        pres = false;
      }
    }
    if (trimTail && 1 <= d && ' ' == buf[d-1]) 
    {
      edit = true;
      d --;
    }
    int start = 0;
    if (trimHead && 0 < d && ' ' == buf[0]) 
    {
      edit = true;
      start ++;
    }
    return edit ? new String(buf, start, d-start) : string;
  }
  
  /**
   * Control if carriage returns are put in the result tree.
   * Default is to output carriage returns.
   * @param b If true, output carriage returns.
   */
  public void setOutputCarriageReturns(boolean b)
  {
    m_outputCarriageReturns = b;
  }

  /**
   * Control if linefeeds are put in the result tree.
   * Default is to output linefeeds.
   * @param b If true, output linefeeds.
   */
  public void setOutputLinefeeds(boolean b)
  {
    m_outputLinefeeds = b;
  }
  
  /**
   * Normalize the linefeeds and/or carriage returns to 
   * be consistently 0x0D 0x0A.  This should almost 
   * certainly be done somewhere else... like in the 
   * XML parser.
   */
  String getNormalizedText(Text tx)
  {
    if(!m_outputCarriageReturns && !m_outputLinefeeds)
    {
      return tx.getData();
    }

    String src = tx.getData();
    int nSrcChars = src.length();
    StringBuffer sb = new StringBuffer(nSrcChars + 8);
    char prevChar = 0x00;
    for(int i = 0; i < nSrcChars; i++)
    {
      char c = src.charAt(i);
      if(0x0A == c)
      {
        if(0x0D != prevChar)
        {
          if(m_outputCarriageReturns)
            sb.append((char)0x0D);
          if(m_outputLinefeeds)
            sb.append((char)0x0A);
        }
      }
      else if(0x0D == c)
      {
        if(m_outputCarriageReturns)
          sb.append((char)0x0D);
        if(m_outputLinefeeds)
          sb.append((char)0x0A);
      }
      else
      {
        sb.append(c);
      }
      prevChar = c;
    }
    
    return (sb.length() > 0) ? sb.toString() : null;
  }

  //==========================================================
  // SECTION: Support functions in fairly random order
  //==========================================================
  private void ____SUPPORT____(){}
    
  /**
   * Control if the xsl:variable is resolved early or 
   * late. Resolving the xsl:variable
   * early is a drag because it means that the fragment 
   * must be created into a DocumentFragment, and then 
   * cloned each time to the result tree.  If we can resolve 
   * late, that means we can evaluate directly into the 
   * result tree.  Right now this must be kept on 'early' 
   * because you would need to set the call stack back to 
   * the point of xsl:invoke... which I can't quite work out 
   * at the moment.  I don't think this is worth fixing 
   * until NodeList variables are implemented.
   */
  private static final boolean m_resolveContentsEarly = true;

  /**
   * Given a name, locate a variable in the current context, and return 
   * the Object.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  protected Object getVariable(String name)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException,
           SAXException
  {
    Object obj = m_variableStacks.getVariable(name);
    if(null == obj)
    {
      obj = m_stylesheetRoot.getTopLevelVariable(name);
    }
    return obj;
  }
  
  /**
   * Create a document fragment.  This function may return null.
   */
  DocumentFragment createDocFrag()
  {
    if(null == m_resultTreeFactory)
    {
      m_resultTreeFactory = m_parserLiaison.createDocument();
    }
    return new ResultTreeFrag(m_resultTreeFactory);
  }
  
  /**
   * Given a document, get the default stylesheet URI from the 
   * xsl:stylesheet PI.
   */
  public String getStyleSheetURIfromDoc(Document sourceTree)
  {
    String stylesheetURI = null;
    NodeList children = sourceTree.getChildNodes();
    int nNodes = children.getLength();
    for(int i = 0; i < nNodes; i++)
    {
      Node child = children.item(i);
      if(Node.PROCESSING_INSTRUCTION_NODE == child.getNodeType())
      {
        ProcessingInstruction pi = (ProcessingInstruction)child;
        if(pi.getNodeName().equals("xml-stylesheet"))
        {
          boolean isOK = true;
          StringTokenizer tokenizer = new StringTokenizer(pi.getNodeValue(), " \t=");
          while(tokenizer.hasMoreTokens())
          {
            if(tokenizer.nextToken().equals("type"))
            {
              String typeVal = tokenizer.nextToken();
              typeVal = typeVal.substring(1, typeVal.length()-1);
              if(!typeVal.equals("text/xsl"))
              {
                isOK = false;
              }
            }
          }  
          
          if(isOK)
          {
            tokenizer = new StringTokenizer(pi.getNodeValue(), " \t=");
            while(tokenizer.hasMoreTokens())
            {
              if(tokenizer.nextToken().equals("href"))
              {
                stylesheetURI = tokenizer.nextToken();
                stylesheetURI = stylesheetURI.substring(1, stylesheetURI.length()-1);
              }
            } 
            break;
          }
        }
      }
    } // end for(int i = 0; i < nNodes; i++)
    return stylesheetURI;
  }
  
  /**
   * This special user function lets the caller set a defined 
   * constant, so they can dynamicaly give some control to 
   * a given stylesheet.
   */
  public void setConstant(String name, String value)
  {
    if(m_topLevelVariables.containsKey(name))
    {
      m_topLevelVariables.remove(name);
    }
    m_topLevelVariables.put(name, value);
  }
  
  Vector m_topLevelParams = new Vector();
  
  /**
   * Push a top-level stylesheet parameter.  This value can 
   * be evaluated via xsl:param-variable.
   */
  public void setStylesheetParam(String key, String expression)
  {
    m_topLevelParams.addElement(new Arg(key, expression));
  }
  
  /**
   * Resolve the params that were pushed by the caller.
   */
  private void resolveTopLevelParams()
    throws XSLProcessorException, 
    java.net.MalformedURLException, 
    java.io.FileNotFoundException, 
    java.io.IOException,
    SAXException
  {
    int n = m_topLevelParams.size();
    for(int i = 0; i < n; i++)
    {
      Arg a = (Arg)m_topLevelParams.elementAt(i);
      m_variableStacks.pushTopLevelParam(a.m_name, a.m_val, 
                                         m_stylesheetRoot.m_stylesheet, m_rootDoc);
    }
  }
  
  /**
   * Given an id of a stylesheet element, search the tree for 
   * the element, and modify the named attribute on that element.
   * @param styleSheet The stylesheet to look in.
   * @param nameOfIDAttr The name of an attribute to search for a unique id.
   * @param idOfElementToModify The unique ID to look for.
   * @param attrNameToSet Once the element is found, the name of the attribute to set.
   * @param attrValueToSet The value to set the attribute to.
   * @return true if modification was made.  False otherwise.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public boolean modifyStylesheetAttribute(Document styleSheet,
                                           String nameOfIDAttr,
                                           String idOfElementToModify, 
                                           String attrNameToSet, 
                                           String attrValueToSet)
    throws XSLProcessorException
  {
    boolean didSetIt = false;
    Element elementToModify = findElementByAttribute(
                           styleSheet.getDocumentElement(),
                           null, 
                           nameOfIDAttr,
                           idOfElementToModify);
    if(null != elementToModify)
    {
      elementToModify.setAttribute(attrNameToSet, attrValueToSet);
      didSetIt = true;
    }
    else
    {
      warn("modifyStylesheetAttribute - can not find element that "+
           "has an attribute named "+nameOfIDAttr+" with a value of "+
           idOfElementToModify+".");
    }
    return didSetIt;
  }
  
  /**
   * Create the default rule if needed.
   */
  private void initDefaultRule()
  {
    if(null == m_defaultRule)
    {
      Document defaultRuleDoc = m_parserLiaison.createDocument();
      Element root = defaultRuleDoc.createElement("xsl:stylesheet");
      root.setAttribute("xmlns:xsl", m_XSLNameSpaceURL);
      defaultRuleDoc.appendChild(root);
      // Then manufacture a default
      m_defaultRule = defaultRuleDoc.createElement("xsl:"+Constants.ELEMNAME_TEMPLATE_STRING);
      root.appendChild(m_defaultRule);
      m_defaultRule.setAttribute(Constants.ATTRNAME_MATCH, "*");
      Element childrenElement = defaultRuleDoc.createElement("xsl:"+Constants.ELEMNAME_APPLY_TEMPLATES_STRING);
      m_defaultRule.appendChild(childrenElement);
      
      m_defaultTextRule = defaultRuleDoc.createElement("xsl:"+Constants.ELEMNAME_TEMPLATE_STRING);
      root.appendChild(m_defaultTextRule);
      m_defaultTextRule.setAttribute(Constants.ATTRNAME_MATCH, "text()");
      
      m_defaultRootRule = defaultRuleDoc.createElement("xsl:"+Constants.ELEMNAME_TEMPLATE_STRING);
      root.appendChild(m_defaultRootRule);
      m_defaultRootRule.setAttribute(Constants.ATTRNAME_MATCH, "/");
      childrenElement = defaultRuleDoc.createElement("xsl:"+Constants.ELEMNAME_APPLY_TEMPLATES_STRING);
      m_defaultRootRule.appendChild(childrenElement);
    }
  }

  /**
   * Returns the current input node that is being 
   * processed.
   */
  public Node getSourceNode()
  {
    return m_currentNode;
  }

  /**
   * Returns the current input node that is being 
   * processed.
   * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
   * the error condition is severe enough to halt processing.
   */
  public Node getSourceNode(String selectPattern, Element namespaceContext)
    throws XSLProcessorException, 
           java.net.MalformedURLException, 
           java.io.FileNotFoundException, 
           java.io.IOException
  {
    NodeList sourceNodes = m_queryEngine.query(selectPattern, m_currentNode, namespaceContext);
    return ((null != sourceNodes) && (sourceNodes.getLength() > 0)) ? 
           sourceNodes.item(0) : null;
  }
  
  /**
   * Reset the current element state
   */
  protected void resetCurrentState(Document sourceTree, Node xmlNode)
  {
    if(null != xmlNode)
    {
      //===============================================
      // This will be used with callbacks from script, 
      // in places like getAttributeCallback.
      m_currentNode = xmlNode;
    }
  }
  
  /**
   * Given a classID and codetype, tr to register a code dispatcher.
   */
  public void registerExtensionHandlerByName(String classid, String codetype)
    throws XSLProcessorException
  {    
    try
    {
      Class classObj = Class.forName(classid);
      try
      {
        Object[] args = new Object[1];
        Class[] argTypes = new Class[1];
        args[0] = codetype;
        argTypes[0] = codetype.getClass();
        Method m = classObj.getMethod ("register", argTypes);
        DispatcherFactory factory = (DispatcherFactory)m.invoke (null, args);
        if(null != factory)
        {
          registerExtensionHandler(codetype, factory);
        }
      }
      catch(NoSuchMethodException nsme)
      {
        warn("Failed to get registration method for extension handler: "+classid);
      }
      catch(InvocationTargetException invocationTarget)
      {
        warn("Invocation Target Error when trying to register extension handler: "+codetype);
      }
      catch(IllegalAccessException illegalAccess)
      {
        warn("Illegal Access Error when trying to register extension handler: "+codetype);
      }
    }
    catch(ClassNotFoundException classNotFound)
    {
      warn("Class Not Found Error when trying to register extension handler: "+codetype);
    }    
  }
  
  /**
   * Register the given DispatcherFactor for a given mime type.
   */
  public void registerExtensionHandler(String mimeType, DispatcherFactory factory)
  {
    if(null != m_diagnosticsPrintWriter)
    {
      diag("Registering dispatcher factory for: "+mimeType);
    }
    m_dispatcherFactories.put(mimeType, factory);
  }

  /**
   * Get the textual contents of the node. If the node 
   * is an element, apply whitespace stripping rules, 
   * though I'm not sure if this is right (I'll fix 
   * or declare victory when I review the entire 
   * whitespace handling).
   */
  String getNodeData(Node node)
  {
    String data = null;
    switch(node.getNodeType())
    {
    case Node.DOCUMENT_FRAGMENT_NODE:
    case Node.DOCUMENT_NODE:
    case Node.ELEMENT_NODE:
      NodeList children = node.getChildNodes();
      int nNodes = children.getLength();
      for(int i = 0; i < nNodes; i++)
      {
        String nodeData = getNodeData(children.item(i));
        if((null != nodeData) && (nodeData.length() > 0))
        {
          data = (null == data) ? nodeData : (data + nodeData);
        }
      }
      break;
    case Node.TEXT_NODE:
    case Node.CDATA_SECTION_NODE:
      // Apply the same whitespace rules that you apply to 
      // fixing the whitespace when m_stripWhiteSpace == true.
      // Don't know if this is OK.
      if(shouldStripSourceNode(node))
      {
        if(!m_parserLiaison.isIgnorableWhitespace((Text)node))
        {
          // data = fixWhiteSpace(((Text)node).getData(), false, false, true);
          data = getNormalizedText((Text)node);
        }
      }
      else
      {
        data = ((Text)node).getData();
      }
      break;
    case Node.ATTRIBUTE_NODE:
      data = node.getNodeValue();
      break;
    default:
      // ignore
      break;
    }
            
    return data;
  }

  
  
  /**
   * Given a tag name, an attribute name, and 
   * an attribute value, do a very crude recursive 
   * search and locate the first match.
   */
  protected Element findElementByAttribute(
                                           Element elem,
                                           String targetElementName, 
                                           String targetAttributeName,
                                           String targetAttributeValue)
  {
    Element theFoundElement = null;
    String tagName = elem.getTagName();
    
    if((null == targetElementName) || tagName.equals(targetElementName))
    {
      NamedNodeMap attributes = elem.getAttributes();
      try
      {
        int nAttributes = (null != attributes) ? attributes.getLength() : 0;
        for(int i = 0; i < nAttributes; i++)  
        {	
          Attr attr = 
                          (Attr)attributes.item(i);
          String attrName = attr.getName();
          
          if((null != attrName) && 
             attrName.equals(targetAttributeName))
          {
            String attrVal = attr.getValue();
            if(attrVal.equals(targetAttributeValue))
            {
              theFoundElement = elem;
              break;
            }
          }
        }
      }
      catch(DOMException e)
      {
      }
    }
    if(null == theFoundElement)
    {
      NodeList children = elem.getChildNodes();    
      int nNodes = children.getLength();
      for(int i = 0; i < nNodes; i++) 
      {
        Node childNode = children.item(i);
        if (childNode instanceof Element) 
        {
          Element child = (Element)childNode;
          String childName = child.getTagName();
          if(null != childName)
          {
            theFoundElement = findElementByAttribute(
                                                     child,
                                                     targetElementName, 
                                                     targetAttributeName,
                                                     targetAttributeValue);
            
            if(null != theFoundElement)
            {
              break;
            }
          }
        }
      }
    }
    return theFoundElement;
  }
  



  /**
   * Do everything possible to get a good URL from a string.
   */
    /**
   * Take a user string and try and parse XML, and also return 
   * the url.
   */
  URL getURLFromString(String urlString)
    throws MalformedURLException
  {
    URL url = null;
    try 
    {
      url = new URL(urlString);
    }
    catch (MalformedURLException e)
    {
      try 
      {
        String fullpath = "file:" + new File (urlString).getAbsolutePath ();
        url = new URL(fullpath);
      }
      catch (MalformedURLException e2)
      {
        diag("Error! Cannot create url for: " + urlString);
        throw e2;
      }
    }
    return url;
  }
        
  /**
   * Class to help in converting enumerated numbers 
   * into strings, using the XSL conversion attributes.
   * Having this in a class helps to not have to extract 
   * the attributes everytime.
   */
  class NumeratorFormatter
  {
    protected Element m_xslNumberElement;
    NumberFormatStringTokenizer m_formatTokenizer;
    Locale m_locale;
    java.text.NumberFormat m_formatter;
    
    /**
     * Construct a NumeratorFormatter using an element 
     * that contains XSL number conversion attributes - 
     * format, letter-value, xml:lang, digit-group-sep, 
     * n-digits-per-group, and sequence-src.
     */
    NumeratorFormatter(Element xslNumberElement)
    {
      m_xslNumberElement = xslNumberElement;
      
    } // end NumeratorFormatter(Element) constructor
    
    /**
     * Process the attributes of the node with number formatting 
     * attributes.  I'd like to 
     * do this at construction time, but I have to do it every 
     * time because the elements may have attribute templates embedded 
     * in them, and so must be processed only when the node is 
     * actually called.
     * TODO: Can I just do NumeratorFormatter.processAttributes up front now?
     */
    void processAttributes(Node contextNode)
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException
    {
      // See http://www.w3.org/TR/WD-xsl#convert for the form 
      // of the format string, which tells how the numbers should be 
      // formatted.
      String formatValue = getProcessedAttrVal(m_xslNumberElement, 
                                               Constants.ATTRNAME_FORMAT, 
                                               contextNode);

      if(null == formatValue) formatValue = "1";
      m_formatTokenizer = new NumberFormatStringTokenizer(formatValue);

      // The letter-value attribute disambiguates between numbering schemes 
      // that use letters. A value of "alphabetic" specifies the alphabetic 
      // sequence; a value of "other" specifies the other sequence.
      // TODO: Handle letter-value attribute.
      String letterValue = getProcessedAttrVal(m_xslNumberElement, 
                                               Constants.ATTRNAME_LETTERVALUE, 
                                               contextNode);
      if(null != letterValue)
      {
        warn(null, contextNode, Constants.ATTRNAME_LETTERVALUE+" not supported yet!");
      }
      
      // When numbering with an alphabetic sequence, the xml:lang 
      // attribute specifies which language's alphabet is to be used.
      String langValue = getAttrVal(m_xslNumberElement, "xml:lang");
      if(null != langValue)
      {
        // Not really sure what to do about the country code, so I use the
        // default from the system.  
        // TODO: fix xml:lang handling.
        m_locale = new Locale(langValue.toUpperCase(), 
                            Locale.getDefault().getDisplayCountry());
        if(null == m_locale)
        {
          warn(null, contextNode, "Warning: Could not find locale for xml:lang="+langValue);
          m_locale = Locale.getDefault();
        }
      }
      else
      {
        m_locale = Locale.getDefault();
      }
      
      // Helper to format local specific numbers to strings.
      m_formatter = java.text.NumberFormat.getNumberInstance(m_locale);
      
      // The digit-group-sep attribute gives the separator between groups 
      // of digits, and the optional n-digits-per-group specifies the 
      // number of digits per group. For example, digit-group-sep="," 
      // and n-digits-per-group="3" would produce numbers of the form 1,000,000.
      String digitGroupSepValue = getProcessedAttrVal(m_xslNumberElement, 
                                                      Constants.ATTRNAME_DIGITGROUPSEP, 
                                                      contextNode);
      
      String nDigitsPerGroupValue = getProcessedAttrVal(m_xslNumberElement, 
                                                        Constants.ATTRNAME_NDIGITSPERGROUP, 
                                                        contextNode);
      // int nDigitsPerGroup = Integer.valueOf(nDigitsPerGroupValue).intValue();
      
      // TODO: Handle digit-group attributes
      if((null != digitGroupSepValue) || (null != nDigitsPerGroupValue))
      {
        m_formatter.setGroupingUsed(true);
      }
      
      // The sequence-src attribute gives the URI of a text resource 
      // that contains a whitespace separated list of the members of 
      // the numbering sequence.
      // (Used with letter-value="other", I think.)
      String sequenceSrcValue = getProcessedAttrVal(m_xslNumberElement, 
                                                    Constants.ATTRNAME_SEQUENCESRC, 
                                                    contextNode);
      if(null != sequenceSrcValue)
      {
        warn(null, contextNode, Constants.ATTRNAME_SEQUENCESRC+" not supported yet!");
      }
    }
    
    /**
     * Format a vector of numbers into a formatted string.
     * @param xslNumberElement Element that takes %conversion-atts; attributes.
     * @param list Array of one or more integer numbers.
     * @return String that represents list according to 
     * %conversion-atts; attributes.
     * TODO: Optimize formatNumberList so that it caches the last count and
     * reuses that info for the next count.
     */
    String formatNumberList(int[] list)
    {
      String formattedNumber = "";
      int nNumbers = list.length;
      char numberType = '1';
      int numberWidth = 1;
      String formatToken;
      String sepString = "";
      String lastSepString = null;
      String padString = m_formatter.format(0);
      String lookahead= null; // next token
      m_formatTokenizer.reset();
      for(int i = 0; i < nNumbers; i++)
      {
        while(m_formatTokenizer.hasMoreTokens())
        {
          if( lookahead != null )
          {
            formatToken = lookahead;
            lookahead = null;
          }
          else 
          {
            formatToken = m_formatTokenizer.nextToken();
          }
          if(Character.isLetterOrDigit(formatToken.charAt(formatToken.length()-1)))
          {
            numberWidth = formatToken.length();
            numberType = formatToken.charAt(numberWidth-1);
            break; // from while(tokenizer.hasMoreTokens())
          }
          else
          {
            sepString = formatToken;
            while(m_formatTokenizer.hasMoreTokens())
            {
              lookahead = m_formatTokenizer.nextToken();
              if(!Character.isLetterOrDigit(lookahead.charAt(0)))
              {
                sepString += lookahead;
                lookahead = null; // consume
              }
              else
              {
                break; // from inner while loop
              }
            }
            if(!m_formatTokenizer.hasMoreTokens())
            {
              lastSepString = sepString;
              
              // crufty bit to strip whitespace off of separator 
              // if it is the end string.
              // This probably isn't right.
              int lastPos = sepString.length() - 1;
              int k;
              for(k = lastPos; k >= 0; k--)
              {
                if(!Character.isWhitespace(sepString.charAt(k)))
                {
                  break; // from for loop
                }
              }
              sepString = sepString.substring(0, k+1);
            }
            
          }
        } // end while
        
        switch(numberType)
        {
        case 'A':
          formattedNumber += (sepString + int2alphaCount(list[i], m_alphaCountTable));
          break;
        case 'a':
          formattedNumber += (sepString + int2alphaCount(list[i], m_alphaCountTable).toLowerCase(m_locale));
          break;
        case 'I':
          formattedNumber += (sepString + long2roman(list[i], true));
          break;
        case 'i':
          formattedNumber += (sepString + long2roman(list[i], true).toLowerCase(m_locale));
          break;
        default: // "1"
          String numString = m_formatter.format(list[i]);
          int nPadding = numberWidth - numString.length();
          formattedNumber += sepString;
          for(int k = 0; k < nPadding; k++)
          {
            formattedNumber += padString;
          }
          formattedNumber += numString;
        }
      }
      
      // Check to see if we finished up the format string...
      if(null == lastSepString)
      {
        lastSepString = "";
        while(m_formatTokenizer.hasMoreTokens())
        {
          lookahead = m_formatTokenizer.nextToken();
          if(!Character.isLetterOrDigit(lookahead.charAt(0)))
          {
            lastSepString += lookahead;
          }
          else
          {
            lastSepString = "";
          }
        }      
      }

      if(null != lastSepString)
      {
        formattedNumber += lastSepString;
      }
      
      return formattedNumber;
    } // end formatNumberList method

    /**
     * Convert a long integer into alphabetic counting, in other words 
     * count using the sequence A B C ... Z AA AB AC.... etc.
     * @param val Value to convert -- must be greater than zero.
     * @param table a table containing one character for each digit in the radix
     * @return String representing alpha count of number.
     * @see XSLProcessor#DecimalToRoman
     * 
     * Note that the radix of the conversion is inferred from the size
     * of the table.
     */
    protected String int2alphaCount(int val, char [] table)
    {
      
      int radix = table.length;
      
      // Create a buffer to hold the result
      // TODO:  size of the table can be detereined by computing
      // logs of the radix.  For now, we fake it.  
      char buf[] = new char[100];
      
      // next character to set in the buffer
      int charPos = buf.length -1 ;    // work backward through buf[]
      
      // index in table of the last character that we stored
      int lookupIndex = 1;  // start off with anything other than zero to make correction work
      
      //						Correction number
      //
      //	Correction can take on exactly two values:
      //
      //		0	if the next character is to be emitted is usual
      //
      //      radix - 1 
      //			if the next char to be emitted should be one less than
      //			you would expect 
      //			
      // For example, consider radix 10, where 1="A" and 10="J"
      //
      // In this scheme, we count: A, B, C ...   H, I, J (not A0 and certainly
      // not AJ), A1
      //
      // So, how do we keep from emitting AJ for 10?  After correctly emitting the
      // J, lookupIndex is zero.  We now compute a correction number of 9 (radix-1).
      // In the following line, we'll compute (val+correction) % radix, which is,
      // (val+9)/10.  By this time, val is 1, so we compute (1+9) % 10, which 
      // is 10 % 10 or zero.  So, we'll prepare to emit "JJ", but then we'll
      // later suppress the leading J as representing zero (in the mod system, 
      // it can represent either 10 or zero).  In summary, the correction value of
      // "radix-1" acts like "-1" when run through the mod operator, but with the 
      // desireable characteristic that it never produces a negative number.
      
      int correction = 0;
      
      // TODO:  throw error on out of range input
      
      do
      {
        // most of the correction calculation is explained above,  the reason for the
        // term after the "|| " is that it correctly propagates carries across
        // multiple columns.  
        correction = ((lookupIndex == 0) || 
                      (correction != 0 && lookupIndex == radix-1 )) ? (radix-1) : 0;
        
        // index in "table" of the next char to emit
        lookupIndex  = (val+correction) % radix;  
        
        // shift input by one "column"
        val = (val / radix);
        
        // if the next value we'd put out would be a leading zero, we're done.
        if (lookupIndex == 0 && val == 0)
          break;
        
        // put out the next character of output
        buf[charPos--] = table[lookupIndex];
      }
               while (val > 0);

      return new String(buf, charPos+1, (buf.length - charPos -1));
    } 

    /**
     * Convert a long integer into roman numerals.
     * @param val Value to convert.
     * @param prefixesAreOK true_ to enable prefix notation (e.g. 4 = "IV"),
     * false_ to disable prefix notation (e.g. 4 = "IIII").
     * @return Roman numeral string.
     * @see DecimalToRoman
     * @see m_romanConvertTable
     */
    protected String long2roman(long val, boolean prefixesAreOK)
    {
      if(val <= 0)
      {
        return "#E("+val+")";
      }

      String roman = "";
      int place = 0;
      if (val <= 3999L)
      {
        do      
        {
          while (val >= m_romanConvertTable[place].m_postValue)            
          {
            roman += m_romanConvertTable[place].m_postLetter;
            val -= m_romanConvertTable[place].m_postValue;
          }
          if (prefixesAreOK)            
          {
            if (val >= m_romanConvertTable[place].m_preValue)                  
            {
              roman += m_romanConvertTable[place].m_preLetter;
              val -= m_romanConvertTable[place].m_preValue;
            }
          } 
          place++;      
        }
          while (val > 0);
      }
      else
      {
        roman = "#error";
      }
      return roman;
    } // end long2roman
    
    
  } // end NumeratorFormatter class

  /**
   * Given a name, and one of xsl:counter, xsl:counters, xsl:counter-increment, or
   * xsl:counter-reset, return the correct counter object.
   */
  private XSLResultTreeCounter getResultTreeCounter(String name,
                                                    Element xslCounterNode,
                                                    boolean createIfNotFound, 
                                                    boolean isCounterReset)
    throws XSLProcessorException
  {
    return getResultTreeCounter(name, xslCounterNode, createIfNotFound, 
                                                    isCounterReset, 
                                                    m_counters.size() - 1);
  }

  /**
   * Given a name, and one of xsl:counter, xsl:counters, xsl:counter-increment, or
   * xsl:counter-reset, return the correct counter object.
   */
  private XSLResultTreeCounter getResultTreeCounter(String name,
                                                    Element xslCounterNode,
                                                    boolean createIfNotFound, 
                                                    boolean isCounterReset,
                                                    int startCounter)
    throws XSLProcessorException
  {
    // Look for an ancestor that has a counter of the given name.
    // Node counterOwnerNode = null;
    XSLResultTreeCounter resultTreeCounter = null;
    if(isCounterReset)
    {
      if(!m_counters.isEmpty())
      {
        resultTreeCounter 
          = (XSLResultTreeCounter)m_counters.elementAt(startCounter);
        while(null != resultTreeCounter)
        {
          if((null != resultTreeCounter.m_name) 
             && resultTreeCounter.m_name.equals(name))
          {
            break;
          }
          resultTreeCounter = resultTreeCounter.m_next;
        }
      }
    }
    else
    {
      // Node parent = resultParent;
      boolean foundit = false;
      for(int i = startCounter; i >= 0; i--)
      {
        resultTreeCounter 
          = (XSLResultTreeCounter)m_counters.elementAt(i);
        while(null != resultTreeCounter)
        {
          if((null != resultTreeCounter.m_name) 
             && resultTreeCounter.m_name.equals(name))
          {
            foundit = true;
            break;
          }
          resultTreeCounter = resultTreeCounter.m_next;
        }
        if(foundit) break;
      }
    }
    
    // If a counter was not found, create one on the root node if it is a 
    // xsl:counter-increment, or on the result tree parent if it is a 
    // xsl:counter-reset.
    if((null == resultTreeCounter) && createIfNotFound)
    {
      int counterOwnerNodeIndex = isCounterReset 
                                  ? (m_counters.size() - 1) : 0;
      
      resultTreeCounter = new XSLResultTreeCounter(xslCounterNode, 
                                                   name, counterOwnerNodeIndex);
      XSLResultTreeCounter counter 
        = (XSLResultTreeCounter)m_counters.elementAt(counterOwnerNodeIndex);
      
      if(m_emptyCounter == counter)
      {
        m_counters.setElementAt(resultTreeCounter, counterOwnerNodeIndex);
      }
      else
      {
        while(null != counter)
        {
          if(null == counter.m_next)
          {
            counter.m_next = resultTreeCounter;
            break;
          }
          counter = counter.m_next;
        }
      }
    }
        
    return resultTreeCounter;
  } // end getResultTreeCounter method
  
  /**
   * Set the formatter interface, which has the toMarkup 
   * method, and which will eventually hold other non-event 
   * methods.  Not to be confused with the DocumentHandler
   * interface.
   */
  public void setFormatter(Formatter formatter)
    throws SAXException
  {
    flushPending();
    m_formatter = formatter;
    if(null != m_formatter)
    {
      m_flistener = m_formatter.getFormatterListener();
    }
    else
    {
      m_flistener = null;
    }
  }
  
  //==========================================================
  // SECTION: Member variables
  //==========================================================
  private void ____GENERAL_VARIABLES____(){}

  /**
   * The liason to the XML parser, so the XSL processor 
   * can handle included files, and the like, and do the 
   * initial parse of the XSL document.
   */
  XMLParserLiaison m_parserLiaison = null;

  /**
   * The formatter interface, which has the toMarkup 
   * method, and which will eventually hold other non-event 
   * methods.  Not to be confused with the DocumentHandler
   * interface.
   */
  public Formatter m_formatter = null;

  /**
   * The listener for formatting events.  This should be 
   * supplied by the Formatter object.
   */
  private DocumentHandler m_flistener = null;
  
  /**
   * The factory that will be used to create result tree fragments.
   */
  public Document m_resultTreeFactory = null;

  /**
   * The query/pattern-matcher object.
   */
  private QueryEngine m_queryEngine = null;
  
  /**
   * Set the query/pattern-matcher object.
   * @param queryEngine A QueryEngine interface.
   */
  public void setQueryEngine(QueryEngine queryEngine)
  {
    m_queryEngine = queryEngine;
  }

  /**
   * Get the query/pattern-matcher object.
   * @return The QueryEngine interface.
   */
  public QueryEngine getQueryEngine()
  {
    return m_queryEngine;
  }
  
  /**
   * The namespace that the result tree conforms to.  A null value 
   * indicates that result-ns is not used and there is no checking. 
   * A empty string indicates that the result tree conforms to the 
   * default namespace.
   */
  String m_resultNameSpacePrefix = null;
    
  /**
   * The URL that belongs to the result namespace.
   */
  String m_resultNameSpaceURL = null;

  /**
   * This is a keyed list of all other namespaces declared with the 
   * xmlns: attribute.  These are keyed by the namespace prefixes 
   * and contain values that are the URIs.  Not sure what I'm going 
   * to do with this.
   */
  Hashtable m_nameSpaces = new Hashtable();
  
  /**
   * A dummy value for use in hash tables that have no use for the value.
   */
  private static Boolean dummyVal = new Boolean(true);
  
  /**
   * This holds the current context node list.  This should arguably be 
   * passed by parameter.
   */
  NodeList m_contextNodeList = null;
    
  /**
   * Table of KeyDeclaration objects, which are set by the 
   * xsl:key element.
   */
  private Vector m_keyDeclarations = new Vector();

  /**
   * Set the table of KeyDeclaration objects, which are set by the 
   * xsl:key element.
   * @param table Table of IDAttributeHolder objects.
   */
  public void setKeyDeclarationsTable(Vector table)
  {
    m_keyDeclarations = table;
  }

  /**
   * Get the table of KeyDeclaration objects.
   * @return Table of KeyDeclaration objects.
   */
  public Vector getKeyDeclarations()
  {
    return m_keyDeclarations;
  }

  /**
   * Table of element keys.  The table will be built on demand, 
   * when a key is requested, or set by the XMLParserLiaison or 
   * the caller.  The table is:
   * a) keyed by name,
   * b) each with a value of a hashtable, keyed by the value returned by 
   *    the use attribute,
   * c) each with a value that is a nodelist.
   * Thus, for a given key or keyref, look up hashtable by name, 
   * look up the nodelist by the given reference.
   */
  private Hashtable m_keys = null;
  
  /**
   * Set the table of keys.  This will be used by the 
   * key() and keyref() functions for lookup of a nodelist.
   */
  public void setKeyTable(Hashtable table)
  {
    m_keys = table;
  }

  /**
   * Get the table of element IDs.  This table is used by the 
   * id() functions for lookup of an element ID.
   * @return Table of element IDs, keyed by ID string, with 
   * values that are Element nodes.
   */
  public Hashtable getKeysTable()
  {
    return m_keys;
  }
  
  /**
   * This is set to true if an xsl:id directive is found.
   * Mainly for use by the XMLParserLiaison classes for 
   * optimized processing of ids.
   */
  private boolean m_needToBuildKeysTable = false;
  
  /**
   * Set needToBuildKeysTable to true if an xsl:key directive is found.
   * Mainly for use by the XMLParserLiaison classes for 
   * optimized processing of keys.
   */
  public void setNeedToBuildKeysTable(boolean b)
  {
    m_needToBuildKeysTable = b;
  }

  /**
   * Set the needToBuildKeysTable property, which, if true, 
   * will trigger the building of the keys table.
   * Mainly for use by the XMLParserLiaison classes for 
   * optimized processing of keys.
   */
  public boolean getNeedToBuildKeysTable()
  {
    return m_needToBuildKeysTable;
  }
  
  /*
  * The current input element that is being processed.
  */
  protected Node m_currentNode;
  
  // This is set to true when the "ancestor" attribute of 
  // the select element is encountered.
  private boolean m_needToCheckForInfiniteLoops = false;
  
  /**
   * Object to guard agains infinite recursion when 
   * doing queries.
   */
  private StackGuard m_stackGuard = new StackGuard();
  
  /**
   * The default template to use if we don't find anything
   * else.  This is initialized in initDefaultRule().
   */
  private Element m_defaultRule = null;

  /**
   * The default template to use for the root if we don't find 
   * anything else.  This is initialized in initDefaultRule().
   * We kind of need this because the defaultRule isn't good 
   * enough because it doesn't supply a document context.
   * For now, I default the root document element to "HTML".
   * Don't know if this is really a good idea or not.
   * I suspect it is not.
   */
  private Element m_defaultRootRule = null;

  /**
   * The default template to use for text nodes if we don't find 
   * anything else.  This is initialized in initDefaultRule().
   */
  private Element m_defaultTextRule = null;

  /**
   * Lookup table with xsl:number nodes as keys, and XSLSourceTreeCounter
   * nodes as values.
   * @see handleNumberInstruction
   */
  Hashtable m_sourceTreeCounters = new Hashtable();

  /**
   * This stack keeps track of the ancestor's counters.
   */
  private Stack m_counters = new Stack();
  
  /**
   * Empty counter, which gets pushed per element, to be replaced 
   * by a real counter.
   */
  private XSLResultTreeCounter m_emptyCounter = new XSLResultTreeCounter(null, null, -1);
  
  /**
   * The stack of Variable stacks.  A VariableStack will be 
   * pushed onto this stack for each template invocation.
   */
  VariableStack m_variableStacks = new VariableStack();

  /**
   * Keyed on string macro names, and holding values 
   * that are macro elements in the XSL DOM tree.
   * Initialized in initMacroLookupTable, and used in
   * findNamedTemplate.
   */
  Hashtable m_namedTemplates = new Hashtable();
  
  /**
   * Table for defined constants, keyed on the names.
   */
  Hashtable m_topLevelVariables = new Hashtable();
      
  /**
   * Table of extensions that may be called from the expression language 
   * via the call(name, ...) function.  Objects are keyed on the call 
   * name.
   * @see extensions.html.
   */
  Hashtable m_extensionNamespaces = new Hashtable();
  
  /**
   * The factories that can make the dispatchers.
   * @see extensions.html.
   */
  static Hashtable m_dispatcherFactories = new Hashtable();
  
  /**
   * Listeners for node construction events.
   * (I suspect this will go away...)
   */
  java.util.Vector m_constructionListeners = new java.util.Vector();
  
  /**
   * The XSL class can have a single listener that can be informed 
   * of errors and warnings, and can normally control if an exception
   * is thrown or not (or the problem listeners can throw their 
   * own RuntimeExceptions).
   */
  private ProblemListener m_problemListener = null;
  
  /**
   * Set the problem listener property.
   * The XSL class can have a single listener that can be informed 
   * of errors and warnings, and can normally control if an exception
   * is thrown or not (or the problem listeners can throw their 
   * own RuntimeExceptions).
   * @param l A ProblemListener interface.
   */
  public void setProblemListener(ProblemListener l)
  {
    m_problemListener = l;
  }

  /**
   * Get the problem listener property.
   * The XSL class can have a single listener that can be informed 
   * of errors and warnings, and can normally control if an exception
   * is thrown or not (or the problem listeners can throw their 
   * own RuntimeExceptions).
   * @return A ProblemListener interface.
   */
  public ProblemListener getProblemListener()
  {
    return m_problemListener;
  }
  
  //==========================================================
  // SECTION: Constants
  //==========================================================
  private void ____CONSTANTS____(){}

  // I don't seem to use this right now, but I don't 
  // want to delete yet until I get the white space 
  // stuff fully worked out.
  private static final short M_SPACE = 1<<2;

  /* 
  * For passing to the parseXML function via the 
  * howToHandleParseError parameter. 
  */
  public static final int NOREPORTPARSEERROR = 0;
  public static final int REPORTPARSEERROR = 1;
  public static final int EXITONPARSEERROR = 2;
  
  /**
   * Table to help in converting decimals to roman numerals.
   * @see XSLProcessor#DecimalToRoman
   * @see XSLProcessor#long2roman
   */
  private final static DecimalToRoman m_romanConvertTable[] = 
  {
    new DecimalToRoman(1000, "M", 900, "CM"),        
    new DecimalToRoman(500, "D", 400, "CD"),
    new DecimalToRoman(100L, "C", 90L, "XC"),        
    new DecimalToRoman(50L, "L", 40L, "XL"),
    new DecimalToRoman(10L, "X", 9L, "IX"),        
    new DecimalToRoman(5L, "V", 4L, "IV"),
    new DecimalToRoman(1L, "I", 1L, "I")    
  };
  
  /**
   * Chars for converting integers into alpha counts.
   * @see XSLProcessor#int2alphaCount
   */
  private final static char[] m_alphaCountTable = 
  {
    'Z', // z for zero
    'A' , 'B' , 'C' , 'D' , 'E' , 'F' , 
    'G' , 'H' , 'I' , 'J' , 'K' , 'L' , 
    'M' , 'N' , 'O' , 'P' , 'Q' , 'R' , 
    'S' , 'T' , 'U' , 'V' , 'W' , 'X' , 
    'Y'
  };
  
  /**
   * Proxy for the xsl:counter or xsl:counters element.
   */
  class XSLResultTreeCounter extends NumeratorFormatter
  {
    int m_counter = 0;
            
    XSLResultTreeCounter m_next = null;
    
    String m_name;
    
    int m_counterIndex;
    
    /**
     * Construct a XSLSourceTreeCounter, using the 
     * attributes from the xsl:number element.
     */
    XSLResultTreeCounter(Element xslCounterElement, String name, int counterIndex)
    {
      super(xslCounterElement);
      // m_owner = owner;
      m_name = name;
      m_counterIndex = counterIndex;
    }

    /**
     * Given a xsl:counter-increment element, increment the 
     * counter.
     */
    void incrementCounter(Element xslCounterIncrementElement, Node context)
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException
    {
      String amountAttr = getAttrVal(xslCounterIncrementElement,
                                              Constants.ATTRNAME_AMOUNT,
                                              context);
      try
      {
        m_counter += (null != amountAttr) ? Integer.parseInt(amountAttr): 1;
      }
      catch(java.lang.NumberFormatException e)
      {
        warn(null, context, "Warning: value could not be treated as number in counter-increment: "+amountAttr);
        m_counter += 1;
      }
    }
    
    /**
     * Given a xsl:reset element, reset the counter to the value of 'value' 
     * attribute, or zero if the value attribute is ommited.
     */
    void resetCounter(Element xslCounterResetElement, Node context)
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException
    {
      String valueAttr = getProcessedAttrVal(xslCounterResetElement,
                                             Constants.ATTRNAME_VALUE,
                                             context);
      try
      {
        m_counter = (null != valueAttr) ? Integer.parseInt(valueAttr): 0;
      }
      catch(java.lang.NumberFormatException e)
      {
        warn(null, context, "Warning: value could not be treated as number in counter-reset: "+valueAttr);
        m_counter = 0;
      }
    }
    
    /**
     * Process the attributes of the xsl:number node.  I'd like to 
     * do this at construction time, but I have to do it every 
     * time because the elements may have attribute templates embedded 
     * in them, and so must be processed only when the node is 
     * actually called.
     */
    void processAttributes(Node contextNode)
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException
    {
      super.processAttributes(contextNode);
    }
    
    /**
     * Given an XML source node, get the count according to the 
     * parameters set up by the xsl:number attributes.
     */
    String getCountString(Element xslCounterElement, Node sourceNode)
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException
    {
      // We have to reset the element that's going to have the formatting 
      // instructions.
      m_xslNumberElement = xslCounterElement;
      
      // Process the formatting attributes.
      processAttributes(sourceNode);
      
      int[] list = null;
      
      if(isXSLTagOfType(xslCounterElement, Constants.ELEMNAME_COUNTERS))
      {
        String name = getAttrVal(xslCounterElement, 
                                          Constants.ATTRNAME_NAME, sourceNode);
        
        if(null != name)
        {          
          int counts[] = new int[50];
          XSLResultTreeCounter counter = this;
          int i;
          for(i = 0; null != counter; i++)
          {            
            counts[i] = counter.m_counter;
            counter = getResultTreeCounter(name, null, 
                                           false, false, counter.m_counterIndex - 1);
          }
          list = new int[i];
          
          for(int k = (i-1), j = 0; k >= 0; k--, j++)
          {
            list[k] = counts[j];
          }
        }
      }
      else
      {
        list = new int[1];
        list[0] = m_counter;
      }
      
      String numberString;
      if(null != list)
      {
        numberString = formatNumberList(list);
      }
      else
      {
        numberString = "#error";
      }
      return numberString;
    } // end getCountString method
    
  } // End XSLResultTreeCounter class
  
  /**
   * Proxy for the xsl:number element.
   * It optimizes the xsl:number element by only getting the 
   * attributes and converting them into meaningful directives 
   * at construction time.
   */
  class XSLSourceTreeCounter extends NumeratorFormatter
  {
    String m_countMatchPattern = null;
    String m_fromMatchPattern = null;
    int m_level = Constants.NUMBERLEVEL_SINGLE;
        
    /**
     * Construct a XSLSourceTreeCounter, using the 
     * attributes from the xsl:number element.
     */
    XSLSourceTreeCounter(Stylesheet stylesheetTree, Element xslNumberElement)
    {
      super(xslNumberElement);
    }
    
    /**
     * Process the attributes of the xsl:number node.  I'd like to 
     * do this at construction time, but I have to do it every 
     * time because the elements may have attribute templates embedded 
     * in them, and so must be processed only when the node is 
     * actually called.
     */
    void processAttributes(Node contextNode)
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException
    {
      super.processAttributes(contextNode);
      
      m_countMatchPattern = getAttrVal(m_xslNumberElement, 
                                               Constants.ATTRNAME_COUNT, 
                                               contextNode);
      if(null == m_countMatchPattern)
      {
        switch( contextNode.getNodeType())
        {
        case Node.ELEMENT_NODE:
          m_countMatchPattern = contextNode.getNodeName();
          break;
        case Node.ATTRIBUTE_NODE:
          m_countMatchPattern = "@"+contextNode.getNodeName();
          break;
        case Node.CDATA_SECTION_NODE:
        case Node.TEXT_NODE:
          m_countMatchPattern = "text()";
          break;
        case Node.COMMENT_NODE:
          m_countMatchPattern = "comment()";
          break;
        case Node.DOCUMENT_NODE:
          m_countMatchPattern = "/";
          break;
        case Node.PROCESSING_INSTRUCTION_NODE:
          m_countMatchPattern = "pi("+contextNode.getNodeName()+")";
          break;
        }
      }
      m_fromMatchPattern = getAttrVal(m_xslNumberElement, 
                                             Constants.ATTRNAME_FROM, 
                                             contextNode);
      String levelValue = getAttrVal(m_xslNumberElement, 
                                             Constants.ATTRNAME_LEVEL, 
                                             contextNode);
      if(null != levelValue)
      {
        if(levelValue.equals(Constants.ATTRVAL_MULTI))
          m_level = Constants.NUMBERLEVEL_MULTI;
        else if(levelValue.equals(Constants.ATTRVAL_ANY))
          m_level = Constants.NUMBERLEVEL_ANY;
        else // default to Constants.ATTRVAL_SINGLE
          m_level = Constants.NUMBERLEVEL_SINGLE;
      }
    }
    
    /**
     * Given an XML source node, get the count according to the 
     * parameters set up by the xsl:number attributes.
     */
    String getCountString(Document sourceTree, Node sourceNode)
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException
    {
      int[] list = null;

      // We have to do this EVERY time because
      // of attribute templates in the format attributes.
      processAttributes(sourceNode);
      
      String exprCountString = getAttrVal(m_xslNumberElement, 
                                          Constants.ATTRNAME_EXPR, 
                                          sourceNode);
      if(null != exprCountString)
      {
        Object countObj = m_queryEngine.getExpr(exprCountString, sourceNode, 
                                                m_xslNumberElement, m_contextNodeList);
        int count = (int)m_queryEngine.numberCast(countObj);
        list = new int[1];
        list[0] = count;
      }
      else
      {      
        if((Constants.NUMBERLEVEL_ANY == m_level) || 
           (Constants.NUMBERLEVEL_SINGLE == m_level))
        {
          list = new int[1];
          if(Constants.NUMBERLEVEL_SINGLE == m_level)
          {
            Node target = m_queryEngine.findAncestor(m_fromMatchPattern, 
                                                     m_countMatchPattern, 
                                                     sourceNode, 
                                                     m_xslNumberElement);
            if(null == target) 
              target = m_parserLiaison.getParentOfNode(sourceNode);
            if(null != target)
            {
              list[0] = getSiblingNumber(m_countMatchPattern, target);
            }
            else
            {
              warn(null, sourceNode,
                   "Warning: count attribute does not match an ancestor in xsl:number! Target = "+
                   sourceNode.getNodeName());
            }
          }
          else // if NUMBERLEVEL_ANY
          {
            Node from = null;
            if(null != m_fromMatchPattern)
            {
              from= m_queryEngine.findAncestor(null, m_fromMatchPattern, 
                                               sourceNode, m_xslNumberElement);
              if(null == from)
              {
                from = sourceNode;
              }
            }
            else if(null == from)
            {
              from = sourceNode.getOwnerDocument();
            }
            
            list[0] = getNumberInTree(m_countMatchPattern, 
                                      (from != sourceNode) 
                                      ? getNextInTree(from, from) : from, 
                                      from, sourceNode);
            
          }
        }
        else // if NUMBERLEVEL_MULTI
        {
          list = getAncestorNumbers(m_fromMatchPattern,
                                    m_countMatchPattern, sourceNode);
        }
      }
      
      return (null != list) ? formatNumberList(list) : "#error";
    } // end getCountString method
    
    /**
     * from any position in the tree, return the 
     * next node in the tree, assuming preorder 
     * traversal preceded, or null if at the end.
     */
    Node getNextInTree(Node pos, Node from)
    {
      Node nextNode = null;
      nextNode = pos.getFirstChild();
      while(null == nextNode)
      {
        nextNode = pos.getNextSibling();
        if(null == nextNode)
        {
          pos = pos.getParentNode();
          if(pos == from)
          {
            break;
          }
        }
      }
      return nextNode;
    }
    
    /**
     * Get a number that represents a node based on the
     * position of that node among within the document tree,  
     * and taking into account the count and from patterns as 
     * specified in the XSL specification.
     * @param fromMatchPattern if non-null, where to 
     * start counting from.
     * @param countMatchPattern if non-null, count only nodes 
     * that match this pattern.
     * @param target count this node and preceding nodes.
     * @return A number that counts target and preceding 
     * nodes that qualify.
     */
    int getNumberInTree(String countMatchPattern, Node pos, 
                        Node from, Node target)
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException
    {
      int count = 1;
      if(null != pos)
      {
        do
        {          
          if((pos != target) && (Node.ELEMENT_NODE == pos.getNodeType()) && 
             ((null == countMatchPattern) || 
              (m_queryEngine.matchAgainstPatterns(countMatchPattern, pos, 
                                                  m_xslNumberElement) != 
               QueryEngine.MATCH_SCORE_NONE)))
          {
            count++;
          }
        }
          while((pos != target) && (null != (pos = getNextInTree(pos, from))));
      }
      
      return count;
      
    } // end getNumberInTree method
    
        
    /**
     * Get a number that represents a node based on the
     * position of that node among it's siblings, and 
     * taking into account the count and from patterns.
     * @param fromMatchPattern if non-null, where to 
     * start counting from.
     * @param countMatchPattern if non-null, count only nodes 
     * that match this pattern.
     * @param target count this node and preceding siblings.
     * @return A number that counts target and preceding 
     * siblings that qualify.
     */
    int getSiblingNumber(String countMatchPattern, Node target)
      throws XSLProcessorException, 
      java.net.MalformedURLException, 
      java.io.FileNotFoundException, 
      java.io.IOException
    {
      // TODO: If target is an Attr, implement special handling. 
      NodeList siblings = m_parserLiaison.getParentOfNode(target).getChildNodes();
      int nNodes = siblings.getLength();
      int number = 0; // return val
      for(int i = 0; i < nNodes; i++)
      {
        Node child = siblings.item(i);
        if(child == target)
        {
          number++; // always count the target
          break;
        }
        else if((null == countMatchPattern) || 
                m_queryEngine.matchAgainstPatterns(countMatchPattern, 
                                                   child, m_xslNumberElement) != 
                QueryEngine.MATCH_SCORE_NONE)
        {
          number++;
        }
      }
      
      return number;
    } // end getSiblingNumber method

    
    /**
     * Count the ancestors, up to the root, that match the 
     * pattern.
     * @param patterns if non-null, count only nodes 
     * that match this pattern, if null count all ancestors.
     * @param node Count this node and it's ancestors.
     * @return The number of ancestors that match the pattern.
     * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
     * the error condition is severe enough to halt processing.
     */
    public int countMatchingAncestors(String patterns, 
                                      Node node)
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException
    {
      int count = 0; // return val
      while( null != node )
      {
        if(null != patterns)
        {
          if(m_queryEngine.matchAgainstPatterns(patterns, node, m_xslNumberElement) != 
               QueryEngine.MATCH_SCORE_NONE)
          {
            count++;
          }
        }
        else
        {
          count++;
        }
        node = m_parserLiaison.getParentOfNode(node);
      }
      return count;
    } // end countMatchingAncestors method
    
    /**
     * Climb up the ancestor tree, collecting sibling position 
     * numbers (as modified by the fromMatchPattern and 
     * countMatchPattern patterns).
     * @param fromMatchPattern if non-null, where to 
     * start counting from in each sibling list.
     * @param countMatchPattern if non-null, count only ancestors 
     * and siblings that match this pattern.
     * @param node count this node and ancestors that match the countMatchPattern.
     * @return An array of ints of length that matches exactly the number 
     * of ancestors that match countMatchPattern.
     * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
     * the error condition is severe enough to halt processing.
     */
    public int[] getAncestorNumbers(
                                    String fromMatchPattern,
                                    String countMatchPattern, Node node)
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException
    {    
      int nMatchingAncestors = 
                              countMatchingAncestors(countMatchPattern, node);
      int[] counts = null;
      if(nMatchingAncestors > 0)
      {
        counts = new int[nMatchingAncestors];
        int countIndex = counts.length - 1; // position to put count into
        while( null != node )
        {
          boolean countIt = false;
          if(null != countMatchPattern)
          {
            if(m_queryEngine.matchAgainstPatterns(countMatchPattern, node, m_xslNumberElement) != 
               QueryEngine.MATCH_SCORE_NONE)
            {
              countIt = true;
            }
          }
          else
          {
            countIt = true;
          }
          if(countIt)
          {
            Node target = m_queryEngine.findAncestor(fromMatchPattern, m_countMatchPattern, 
                                       node, m_xslNumberElement);
            if(null == target) 
              target = node;

            counts[countIndex] = getSiblingNumber(countMatchPattern, target);
            countIndex--;
          }
          node = m_parserLiaison.getParentOfNode(node);
        } // end while
      } // end if nMatchingAncestors > 0
      return counts;
    } // end getAncestorNumbers method 
    
  } // End XSLSourceTreeCounter class
  
  /**
   * This class holds an attribute set, and points to the 
   * next attribute set of the same name.
   */
  class AttributeSet
  {
    /**
     * The name of the attribute set.
     */
    private String m_name;
    
    /**
     * points to the xsl:define-attribute-set node.
     */
    public Node m_node;
    
    /**
     * Points to the next attribute set of the same name.
     */
    public AttributeSet m_next = null;
    
    /**
     * Construct a attribute set node.
     */
    public AttributeSet(String name, Node node)
    {
      m_name = name;
      m_node = node;
    }
  }
  
  /**
   * This class represents the base stylesheet or an
   * "import" stylesheet.  "include" stylesheets are 
   * combined with the including stylesheet.
   */
  public class Stylesheet
  {    
    /**
     * The root XSL stylesheet.
     */
    public Document m_document = null;
    
    /**
     * The element that starts the stylesheet.
     * (In case the stylesheet is embedded in 
     * another XML document.
     */
    public Element m_stylesheet;
    
    /**
     * A vector of the -imported- XSL Stylesheets.
     */
    public Vector m_imports = new Vector();
    
    /**
     * This table is keyed on the target elements 
     * of patterns, and contains linked lists of
     * the actual patterns that match the target element
     * to some degree of specifity.
     */
    private Hashtable m_patternTable = new Hashtable();

    /**
     * Table of attribute sets, keyed by set name.
     */
    private Hashtable m_attributeSets = new Hashtable();

    /**
     * The base URL of the XSL document.
     */
    public String m_baseIdent = null;
    
    /**
     * Tells if the stylesheet tables need to be rebuilt.
     */
    public boolean m_tablesAreInvalid = true;

    /**
     * Constructor for a Stylesheet needs a Document.
     * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
     * the error condition is severe enough to halt processing.
     */
    public Stylesheet(Document doc, String baseIdentifier)
      throws XSLProcessorException, 
             MalformedURLException, 
             FileNotFoundException,
             IOException,
             SAXException
    {
      m_document = doc;
      m_baseIdent = baseIdentifier;
      m_stylesheet = locateStylesheet(m_document);
      if(null != m_stylesheet)
      {
        buildStylesheetTables(m_stylesheet, m_baseIdent );
      }
      else
      {
        error(null, doc, "Could not find stylesheet element");
      }
      m_stylesheets.put(m_document, this);
      // System.out.println("putting: "+doc.getDocumentElement().getNodeName());
      // System.out.println("putting: "+((Object)doc.getDocumentElement()).hashCode());
    }
    
    /**
     * Constructor for a Stylesheet needs a Document.
     * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
     * the error condition is severe enough to halt processing.
     */
    public Stylesheet(Document doc, String baseIdentifier, boolean buildTables)
      throws XSLProcessorException, 
      MalformedURLException, 
      FileNotFoundException,
      IOException,
      SAXException
    {
      m_document = doc;
      m_baseIdent = baseIdentifier;
      m_stylesheet = locateStylesheet(m_document);
      if(null != m_stylesheet)
      {
        if(buildTables)
        {
          buildStylesheetTables(m_stylesheet, m_baseIdent );
        }
      }
      else
      {
        error(null, doc, "Could not find stylesheet element");
      }
      m_stylesheets.put(m_document, this);
      // System.out.println("putting: "+doc.getDocumentElement().getNodeName());
      // System.out.println("putting: "+((Object)doc.getDocumentElement()).hashCode());
    }

    /**
     * Constructor for a Stylesheet needs a Document.
     * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
     * the error condition is severe enough to halt processing.
     */
    public Stylesheet(Document doc)
      throws XSLProcessorException, 
             MalformedURLException, 
             FileNotFoundException,
             IOException,
             SAXException
    {
      m_document = doc;
      m_baseIdent = null;
      m_stylesheet = locateStylesheet(m_document);
      if(null != m_stylesheet)
      {
        buildStylesheetTables(m_stylesheet, m_baseIdent );
      }
      else
      {
        error("Could not find stylesheet element");
      }
      m_stylesheets.put(m_document, this);
      // System.out.println("putting: "+doc.getDocumentElement().getNodeName());
      // System.out.println("putting: "+((Object)doc.getDocumentElement()).hashCode());
    }

    /**
     * Look through whatever document is provided for the 
     * xsl namespace.
     */
    private Element locateStylesheet(Document doc)
      throws XSLProcessorException
    {
      Element stylesheet = null;
      NodeList children = doc.getChildNodes();
      int nNodes = children.getLength();
      
      for(int i = 0; i < nNodes; i++)
      {
        Node childNode = children.item(i);
        if (childNode.getNodeType() == Node.ELEMENT_NODE) 
        {      
          Element elem = (Element)childNode;
          NamedNodeMap map = elem.getAttributes();
          int nAttrs = map.getLength();
          for(int k = 0; k < nAttrs; k++)
          {
            Node attrChild = map.item(k);
            if(Node.ATTRIBUTE_NODE == attrChild.getNodeType())
            {
              Attr attr = (Attr)attrChild;
              String name = attr.getName();
              if(name.startsWith("xmlns:") || name.equals("xmlns"))
              {
                String ns = attr.getValue().toLowerCase();
                if(ns.startsWith(m_XSLNameSpaceURLPre))
                {
                  m_XSLNameSpaceURL = ns;
                  initXSLTKeys();
                  if(Constants.ELEMNAME_STYLESHEET == getXSLToken(elem))
                  {
                    stylesheet = elem;
                    String resultns = getAttrVal(stylesheet, "result-ns");
                    if(null != resultns)
                    {
                      m_resultNameSpacePrefix 
                        = m_parserLiaison.getNamespaceForPrefix(resultns, stylesheet);
                    }
                    String indentResult = getAttrVal(stylesheet, "indent-result");
                    if(null == indentResult)
                    {
                      // Hack around indent default issue with HTML formatter
                      if((null != m_resultNameSpacePrefix) 
                         && m_resultNameSpacePrefix.equalsIgnoreCase("http://www.w3.org/TR/REC-html40"))
                      {
                        m_indentResult = true;
                      }
                      else
                      {
                        m_indentResult = false;
                      }
                    }
                    else
                    {
                      m_indentResult = indentResult.equals("yes");
                    }
                    initDefaultRule();  
                    i = nNodes; // break from outer loop
                    break;
                  }
                }
              }
            }
          }
        }
      }
      return stylesheet;
    }
    
    /**
     * Process the xsl:key element.
     * 
     * (Notes to myself)
     * What we need to do is:
     * 1) As this function is called, build a table of KeyDeclarations.
     * 2) During either XML processing, or upon request, walk the XML 
     * document tree, and build a hash table:
     * a) keyed by name,
     * b) each with a value of a hashtable, keyed by the value returned by 
     *    the use attribute,
     * c) each with a value that is a nodelist.
     * Thus, for a given key or keyref, look up hashtable by name, 
     * look up the nodelist by the given reference.
     */
    private void processKeyElement(Element stylesheet, Element xslKeyElement)
      throws XSLProcessorException
    {
      String nameAttr = getAttrVal(xslKeyElement, Constants.ATTRNAME_NAME);
      if(null == nameAttr)
        error(stylesheet, null, "xsl:key requires a "+Constants.ATTRNAME_NAME+" attribute!");
      String matchAttr = getAttrVal(xslKeyElement, Constants.ATTRNAME_MATCH);
      if(null == matchAttr)
        error(stylesheet, null, "xsl:key requires a "+Constants.ATTRNAME_MATCH+" attribute!");
      String useAttr = getAttrVal(xslKeyElement, Constants.ATTRNAME_USE);
      if(null == useAttr)
        error(stylesheet, null, "xsl:key requires a "+Constants.ATTRNAME_USE+" attribute!");

      m_keyDeclarations.addElement(new KeyDeclaration(nameAttr, matchAttr, useAttr));
      m_needToBuildKeysTable = true;
    }
    
    /**
     * Process imports, includes, macros, templates, etc.
     * into tables of various types.
     * @param stylesheet Probably a stylesheet element, 
     * though this may later be an xsl:module element.
     */
    private void buildStylesheetTables(Element stylesheet, String base)
      throws XSLProcessorException, 
             MalformedURLException, 
             FileNotFoundException,
             IOException,
             SAXException
    {
      // recursively process all the import statements
      initImports(stylesheet);
      String defaultspaceval = getAttrVal(stylesheet, Constants.ATTRNAME_DEFAULTSPACE);
      if(null != defaultspaceval)
      {
        m_stripWhiteSpace = defaultspaceval.equals(Constants.ATTRVAL_STRIP);
      }
            
      NodeList children = stylesheet.getChildNodes();    
      int nNodes = children.getLength();
      for(int i = 0; i < nNodes; i++) 
      {
        Node child = children.item(i);
        
        if (Node.ELEMENT_NODE == child.getNodeType()) 
        {
          switch(getXSLToken(child))
          {
          case Constants.ELEMNAME_CSSSTYLECONVERSION:
            m_translateCSS = true;
            break;
            
          case Constants.ELEMNAME_EXTENSIONHANDLER:
            // NON-STANDARD
            // Register an code dispatcher.
            handleCodeDispatcherPI((Element)child);
            break;
            
          case Constants.ELEMNAME_EXTENSION:
            // NON-STANDARD
            // Set up a extensible call that can be called from expression language.
            handleFunctionsInstruction((Element)child);
            break;
            
          case Constants.ELEMNAME_PARAMVARIABLE:
          case Constants.ELEMNAME_VARIABLE:
            {
              if(null == m_stylesheetRoot)
              {
                m_stylesheetRoot = this;
              }
              // xsl:define-constant
              Element variableDef = (Element)child;
              String name = getAttrVal(variableDef, Constants.ATTRNAME_NAME);
              String expression = getAttrVal(variableDef, 
                                                 Constants.ATTRNAME_EXPR, m_document);
              Object var;
              if(null != expression)
              {
                // defer evaluation
                var = expression;
              }
              else
              {
                // Use result tree fragment
                var = variableDef;
              }
              boolean foundit = false;
              int n = m_topLevelParams.size();
              for(int k = 0; k < n; k++)
              {
                Arg a = (Arg)m_topLevelParams.elementAt(k);
                if(a.m_name.equals(name))
                {
                  foundit = true;
                  break;
                }
              }
              if(!foundit)
              {
                m_topLevelParams.addElement(new Arg(name, var));
              }
            }
            break;
            
          case Constants.ELEMNAME_LOCALE:
            warn("xsl:locale not yet supported!");
            break;
            
          case Constants.ELEMNAME_STRIPSPACE:
            {
              // xsl:strip-space
              String elements = getAttrVal((Element)child, Constants.ATTRNAME_ELEMENTS);
              if(null != elements)
              {
                java.util.StringTokenizer tokenizer 
                  = new java.util.StringTokenizer(elements, " \t\n\r");
                while(tokenizer.hasMoreTokens())
                {
                  String elemName = tokenizer.nextToken();
                  m_whitespaceStrippingElements.put(elemName, dummyVal);
                }
              }
              else
              {
                warn(stylesheet, null, "xsl:strip-space requires a "+Constants.ATTRNAME_ELEMENTS+" attribute!");
              }
            }
            break;
            
          case Constants.ELEMNAME_PRESERVESPACE:
            {
              // xsl:preserve-space
              String elements = getAttrVal((Element)child, Constants.ATTRNAME_ELEMENTS);
              if(null != elements)
              {
                java.util.StringTokenizer tokenizer 
                  = new java.util.StringTokenizer(elements, " \t\n\r");
                while(tokenizer.hasMoreTokens())
                {
                  String elemName = tokenizer.nextToken();
                  m_whitespacePreservingElements.put(elemName, dummyVal);
                }
              }
              else
              {
                warn(stylesheet, null, "xsl:preserve-space requires a "+Constants.ATTRNAME_ELEMENTS+" attribute!");
              }
            }
            break;
            
          case Constants.ELEMNAME_KEY:
            processKeyElement(stylesheet, (Element)child);
            break;
          case Constants.ELEMNAME_DEFINEATTRIBUTESET:
            {
              // xsl:define-attribute-set
              // TODO: Make sure define-attribute-set works correctly with importance.
              String attrSetName = getAttrVal((Element)child, Constants.ATTRNAME_NAME);
              if(null != attrSetName)
              {
                AttributeSet newAttrSet = new AttributeSet(attrSetName, child);
                Object val = m_attributeSets.get(attrSetName);
                if(null == val)
                {
                  m_attributeSets.put(attrSetName, newAttrSet);
                }
                else
                {
                  AttributeSet attrSet = (AttributeSet)val;
                  while(attrSet.m_next != null)
                  {
                    attrSet = attrSet.m_next;
                  }
                  attrSet.m_next = newAttrSet;
                }
              }
              else
              {
                warn(stylesheet, null, "xsl:define-attribute-set requires a "
                                   +Constants.ATTRNAME_NAME+" attribute!");
              }
            }
            break;
            
          case Constants.ELEMNAME_TEMPLATE:
            {
              Element template = (Element)child;
              String name = getAttrVal(template, Constants.ATTRNAME_NAME);
              if(null != name)
              {
                // Will override any previously defined.
                m_namedTemplates.put(name, template);
              }
              
              String expression = getAttrVal( template, Constants.ATTRNAME_MATCH );
              if(null != expression)
              {
                Vector strings = null;
                Expression expressionHolder[] = {null};
                if(null != expression)
                {
                  strings 
                    = m_queryEngine.getTargetElementStrings(expression, template, expressionHolder);
                }
                else
                {
                  warn(stylesheet, null, "You must have a "+Constants.ATTRNAME_MATCH+" attribute on templates!");
                }
                
                if(null != strings)
                {
                  int nTargets = strings.size();
                  for(int stringIndex = 0; stringIndex < nTargets; stringIndex++) 
                  {
                    String target = (String)strings.elementAt(stringIndex);
                    
                    Object newMatchPat = new MatchPattern2(expression, 
                                                           expressionHolder[0], 
                                                           template, i, target, this);
                    
                    // See if there's already one there
                    Object val = m_patternTable.get(target);
                    if(null == val)
                    {
                      m_patternTable.put(target, newMatchPat);
                    }
                    else
                    {
                      // find the tail of the list
                      MatchPattern2 matchPat = (MatchPattern2)val;
                      MatchPattern2 next;
                      while((next = matchPat.getNext()) != null)
                      {
                        matchPat = next;
                      }
                      matchPat.setNext((MatchPattern2)newMatchPat);
                    }
                  }
                }
              } // end if( isXSLTagOfType(rule, Constants.ELEMNAME_TEMPLATE) )
            }
            break;
            
          case Constants.ELEMNAME_INCLUDE:
          case Constants.ELEMNAME_IMPORT:
            break;
 
          case Constants.ELEMNAME_PARAM:
          case Constants.ELEMNAME_ADDATTRIBUTE:
          case Constants.ELEMNAME_APPLY_TEMPLATES:
          case Constants.ELEMNAME_USE:
          case Constants.ELEMNAME_CHILDREN:
          case Constants.ELEMNAME_CHOOSE:
          case Constants.ELEMNAME_COMMENT:
          case Constants.ELEMNAME_CONSTRUCT:
          case Constants.ELEMNAME_CONTENTS:
          case Constants.ELEMNAME_COPY:
          case Constants.ELEMNAME_COPY_OF:
          case Constants.ELEMNAME_DISPLAYIF:
          case Constants.ELEMNAME_EVAL:
          case Constants.ELEMNAME_EXPECTEDCHILDREN:
          case Constants.ELEMNAME_FOREACH:
          case Constants.ELEMNAME_IF:
          case Constants.ELEMNAME_CALLTEMPLATE:
          case Constants.ELEMNAME_MESSAGE:
          case Constants.ELEMNAME_NUMBER:
          case Constants.ELEMNAME_OTHERWISE:
          case Constants.ELEMNAME_PI:
          case Constants.ELEMNAME_REMOVEATTRIBUTE:
          case Constants.ELEMNAME_SORT:
          case Constants.ELEMNAME_STYLESHEET:
          case Constants.ELEMNAME_TEXT:
          case Constants.ELEMNAME_VALUEOF:
          case Constants.ELEMNAME_WHEN:
          case Constants.ELEMNAME_ELEMENT:
          case Constants.ELEMNAME_COUNTER:
          case Constants.ELEMNAME_COUNTERS:
          case Constants.ELEMNAME_COUNTERINCREMENT:
          case Constants.ELEMNAME_COUNTERRESET:
          case Constants.ELEMNAME_COUNTERSCOPE:
          case Constants.ELEMNAME_APPLY_IMPORTS:
            error(child.getNodeName()+" not allowed inside a stylesheet!");
            break;

          default:
            String ns =m_parserLiaison.getNamespaceOfNode(child);
            if((null != ns) && (ns.equalsIgnoreCase(m_XSLNameSpaceURL) || 
                                ns.equalsIgnoreCase(m_LotusXSLNameSpaceURL)))
            {
              warn(child.getNodeName()
                   +" unknown XSL instruction inside context of the stylesheet element!");
            }
            // be tolerant of other namespaces
            break;
          } // end switch
        } // end if (Node.ELEMENT_NODE == ruleNode.getNodeType()) 
      } // end for(int i = 0; i < nNodes; i++)
      
      // This needs to be put after initImports so that 
      // the imports in the included files will get placed
      // after the imports in this file.
      initIncludes(stylesheet, base);
      
      m_tablesAreInvalid = false;
    }
    
    /**
     * Get a list of the included documents into a vector.
     */
    private void initIncludes(Element includedStyles, String base)
      throws XSLProcessorException, 
      MalformedURLException, 
      FileNotFoundException, 
      IOException,
      SAXException
    {
      NodeList children = includedStyles.getChildNodes();
      int nNodes = children.getLength();
      for(int i = 0; i < nNodes; i++)
      {
        Node childNode = children.item(i);
        if ((childNode.getNodeType() == Node.ELEMENT_NODE) &&
            isXSLTagOfType(childNode, Constants.ELEMNAME_INCLUDE)) 
        {
          // read document from href
          String val = getAttrVal( (Element)childNode, Constants.ATTRNAME_HREF);
          if(null != val)
          {
            String[] stringHolder = {null};
            Document includedDoc = parseXML(val, stringHolder, base);
            if(null != includedDoc)
            {
              // locate xsl:stylesheet element
              Element stylesheet = locateStylesheet(includedDoc);
              if(null != stylesheet)
              {
                buildStylesheetTables(stylesheet, stringHolder[0]);
              }
              else
              {
                error(includedStyles, null, "Could not find stylesheet element for "+val);
              }
            }
            else
            {
              error(includedStyles, null, "Could not open include stylesheet: "+val);
            }
          }
          else
          {
            error("Could not get href attribute!");
          }
        }
      }
    }
    
    /**
     * Get a list of the imported documents into a vector.
     */
    private void initImports(Element includedStyles)
      throws MalformedURLException, 
             XSLProcessorException, 
             FileNotFoundException,
             IOException,
             SAXException
    {
      NodeList children = includedStyles.getChildNodes();
      int nNodes = children.getLength();
      boolean foundNotImport = false;
      for(int i = 0; i < nNodes; i++)
      {
        Node childNode = children.item(i);
        if (childNode.getNodeType() == Node.ELEMENT_NODE) 
        {
          boolean isImport = isXSLTagOfType(childNode, Constants.ELEMNAME_IMPORT);
          if(!isImport)
          {
            foundNotImport = true;
          }
          else
          {
            if(foundNotImport)
            {
              error("Imports can only occur as the first elements in the stylesheet!");
            }
            // read document from href
            String val = getAttrVal( (Element)childNode, Constants.ATTRNAME_HREF);
            if(null != val)
            {
              String[] stringHolder =  {null};
              Document importedDoc = parseXML(val, stringHolder, this.m_baseIdent );
              
              if(null != importedDoc)
              {
                Stylesheet importedStylesheet = new Stylesheet(importedDoc, stringHolder[0]);
                // I'm going to insert the elements in backwards order, 
                // so I can walk them 0 to n.
                m_imports.insertElementAt(importedStylesheet, 0);
              }
              else
              {
                error("Could not open include stylesheet: "+val);
              }
            }
          }
        }
      }
    }
    
    /**
     * Locate a macro via the "name" attribute.
     * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
     * the error condition is severe enough to halt processing.
     */
    public Element findNamedTemplate(String name)
      throws XSLProcessorException
    {
      Element macroElement = null;
      Object obj = m_namedTemplates.get(name);
      if(null == obj)
      {
        int nImports = m_imports.size();
        for(int i = 0; i < nImports; i++)
        {
          Stylesheet stylesheet = (Stylesheet)m_imports.elementAt(i);
          macroElement = stylesheet.findNamedTemplate(name);
          if(null != macroElement)
          {
            break;
          }
        }
      }
      else
      {
        macroElement = (Element)obj;
      }
      
      if(null == macroElement)
      {
        warn("Could not find macro def named: "+name);
      }
      
      return macroElement;
    }
    
    /**
     * Given the name of a constant, return a string value.
     * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
     * the error condition is severe enough to halt processing.
     */
    public Object getTopLevelVariable(String name)
      throws XSLProcessorException
    {
      Object obj = m_topLevelVariables.get(name);
      if(null == obj)
      {
        int nImports = m_imports.size();
        for(int i = 0; i < nImports; i++)
        {
          Stylesheet stylesheet = (Stylesheet)m_imports.elementAt(i);
          obj = stylesheet.getTopLevelVariable(name);
          if(null != obj)
          {
            break;
          }
        }
      }
      
      if(null == obj)
      {
        warn("Could not find variable def for: "+name);
      }
      
      return obj;
    }

    /**
     * Given a target element, find the template that best 
     * matches in the given XSL document, according 
     * to the rules specified in the xsl draft. 
     * @param stylesheetTree Where the XSL rules are to be found.
     * @param sourceTree Where the targetElem is to be found.
     * @param targetElem The element that needs a rule.
     * @return Rule that best matches targetElem.
     * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
     * the error condition is severe enough to halt processing.
     */
    public Element findTemplate(
                                Document sourceTree, 
                                Node targetNode)
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException
    {
      return findTemplate(sourceTree, targetNode, null, false, null);
    }
    
    /**
     * Given a target element, find the template that best 
     * matches in the given XSL document, according 
     * to the rules specified in the xsl draft. 
     * @param stylesheetTree Where the XSL rules are to be found.
     * @param sourceTree Where the targetElem is to be found.
     * @param targetElem The element that needs a rule.
     * @param mode A string indicating the display mode.
     * @param useImports means that this is an xsl:apply-imports commend.
     * @param foundStylesheet If non-null, the Stylesheet that the found template
     * belongs to will be returned in the foundStylesheet[0].
     * @return Rule that best matches targetElem.
     * @exception XSLProcessorException thrown if the active ProblemListener and XMLParserLiaison decide 
     * the error condition is severe enough to halt processing.
     */
    public Element findTemplate(
                                Document sourceTree, 
                                Node targetNode, 
                                String mode,
                                boolean useImports,
                                Stylesheet foundStylesheet[])
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException
    {
      Element bestMatchedRule = null;
      MatchPattern2 bestMatchedPattern =null; // Syncs with bestMatchedRule
      Vector conflicts = new Vector();
      if(!useImports)
      {
        double highScore = QueryEngine.MATCH_SCORE_NONE;
        
        MatchPattern2 matchPat = null;
        int targetNodeType = targetNode.getNodeType();
        switch(targetNodeType)
        {
        case Node.ELEMENT_NODE:
          // String targetName = m_parserLiaison.getExpandedElementName((Element)targetNode);
          String targetName = m_parserLiaison.getLocalNameOfNode(targetNode);
          matchPat = locateMatchPatternList2(targetName, true);
          break;
          
        case Node.PROCESSING_INSTRUCTION_NODE:
        case Node.ATTRIBUTE_NODE:
          matchPat = locateMatchPatternList2(targetNode.getNodeName(), true);
          break;

        case Node.CDATA_SECTION_NODE:
        case Node.TEXT_NODE:
          matchPat = locateMatchPatternList2(Constants.PSEUDONAME_TEXT, false);
          break;

        case Node.COMMENT_NODE:
          matchPat = locateMatchPatternList2(Constants.PSEUDONAME_COMMENT, false);
          break;

        case Node.DOCUMENT_NODE:
          matchPat = locateMatchPatternList2(Constants.PSEUDONAME_ROOT, false);
          break;
          
        default:
          {
            matchPat = locateMatchPatternList2(targetNode.getNodeName(), false);
          }
        }
        
        String prevPat = null;
        while(null != matchPat)
        {
          Element rule = matchPat.getTemplate();
          // We'll be needing to match rules according to what 
          // mode we're in.
          String ruleMode = getAttrVal(rule, Constants.ATTRNAME_MODE);
          
          // The logic here should be that if we are not in a mode AND
          // the rule does not have a node, then go ahead.
          // OR if we are in a mode, AND the rule has a node, 
          // AND the rules match, then go ahead.
          if(((null == mode) && (null == ruleMode)) || 
             ((null != ruleMode) && (null != mode) && ruleMode.equals(mode)))
          {
            String patterns = matchPat.getPattern();

            if((null != patterns) && !((prevPat != null) && prevPat.equals(patterns)))
            {
              prevPat = patterns;
              
              // Date date1 = new Date();
              double score 
                = m_queryEngine.matchAgainstPatterns(matchPat.getExpression(), targetNode, rule);

              if(QueryEngine.MATCH_SCORE_NONE != score)
              {
                String priorityVal = getAttrVal(rule, Constants.ATTRNAME_PRIORITY);
                double priorityOfRule 
                  = (null != priorityVal) ? 
                    Double.valueOf(priorityVal).doubleValue() 
                    : score;
                matchPat.m_priority = priorityOfRule;
                double priorityOfBestMatched = (null != bestMatchedPattern) ?
                                               bestMatchedPattern.m_priority : 
                                               QueryEngine.MATCH_SCORE_NONE;
                if(priorityOfRule > priorityOfBestMatched)
                {
                  conflicts.removeAllElements();
                  highScore = score;
                  bestMatchedRule = rule;
                  bestMatchedPattern = matchPat;
                }
                else if(priorityOfRule == priorityOfBestMatched)
                {
                  addObjectIfNotFound(bestMatchedPattern, conflicts);
                  conflicts.addElement(matchPat);
                  highScore = score;
                  bestMatchedRule = rule;
                  bestMatchedPattern = matchPat;
                }
              }
              // Date date2 = new Date();
              // m_totalTimePatternMatching+=(date2.getTime() - date1.getTime());
            } // end if(null != patterns)            
          } // end if if(targetModeString.equals(mode))
          
          MatchPattern2 nextMatchPat = matchPat.getNext();

          // We also have to consider wildcard matches.
          if((null == nextMatchPat) && !matchPat.m_targetString.equals("*") && 
             ((Node.ELEMENT_NODE == targetNodeType) || 
              (Node.ATTRIBUTE_NODE == targetNodeType) ||
              (Node.PROCESSING_INSTRUCTION_NODE == targetNodeType)))
          {
            nextMatchPat = (MatchPattern2)m_patternTable.get("*");
          }
          matchPat = nextMatchPat;
        }
      } // end if(!useImports)
      
      if(null == bestMatchedRule)
      {
        int nImports = m_imports.size();
        for(int i = 0; i < nImports; i++)
        {
          Stylesheet stylesheet = (Stylesheet)m_imports.elementAt(i);
          bestMatchedRule = stylesheet.findTemplate(sourceTree, targetNode, mode, 
                                                    false, foundStylesheet);
          if(null != bestMatchedRule)
          {
            break;
          }
        }
      }

      int nConflicts = conflicts.size();
      if(nConflicts > 0)
      {
        String conflictsString = (!m_quietConflictWarnings) 
                                 ? "Specificity conflicts found: " : null;
        for(int i = 0; i < nConflicts; i++)
        {
          MatchPattern2 conflictPat = (MatchPattern2)conflicts.elementAt(i);
          if(0 != i)
          {
            if(!m_quietConflictWarnings)
              conflictsString += ", ";
            
            // Find the furthest one towards the bottom of the document.
            if(conflictPat.m_posInStylesheet > bestMatchedPattern.m_posInStylesheet)
            {
              bestMatchedPattern = conflictPat;
            }
          }
          else
          {
            bestMatchedPattern = conflictPat;
          }
          if(!m_quietConflictWarnings)
            conflictsString += "\""+conflictPat.getPattern()+"\"";
        }
        bestMatchedRule = bestMatchedPattern.getTemplate();
        if(!m_quietConflictWarnings)
        {
          conflictsString += " ";
          conflictsString += "Last found in stylesheet will be used.";
          warn(conflictsString);
        }
      }
      
      if((null != bestMatchedPattern) && (null != foundStylesheet))
      {
        foundStylesheet[0] = bestMatchedPattern.m_stylesheet;
      }
      
      return bestMatchedRule;
    } // end findTemplate
    
    /**
     * Add object to vector if not already there.
     */
    void addObjectIfNotFound(Object obj, Vector v)
    {
      int n = v.size();
      boolean addIt = true;
      for(int i = 0; i < n; i++)
      {
        if(v.elementAt(i) == obj)
        {
          addIt = false;
          break;
        }
      }
      if(addIt)
      {
        v.addElement(obj);
      }
    }
    
    /**
     * Given an element type, locate the start of a linked list of 
     * possible template matches.
     */
    private MatchPattern2 locateMatchPatternList2(String sourceElementType, boolean tryWildCard)
    {
      MatchPattern2 startMatchList = null;
      Object val = m_patternTable.get(sourceElementType);
      if(null != val)
      {
        startMatchList = (MatchPattern2)val;
      }
      else if(tryWildCard)
      {
        val = m_patternTable.get("*");
        if(null != val)
        {
          startMatchList = (MatchPattern2)val;
        }
      }
      return startMatchList;
    }

    /**
     * Given an element cloned into the result tree, and a value 
     * from an xsl:use attribute, then add the attributes 
     * from the defined attribute sets named.
     * @param xslUseValue The value obtained from an xsl:use attribute.
     * @param resultTreeElem An element cloned from a template 
     * element.
     */
    private void findAndUseAttrSets(String xslUseValue, AttributeListImpl attList, 
                            Node contextNode, Stylesheet stylesheetTree)
      throws XSLProcessorException
    {
      StringTokenizer tokenizer = new StringTokenizer(xslUseValue);
      while(tokenizer.hasMoreElements())
      {
        String templateName = tokenizer.nextToken();
        AttributeSet attrSet = (AttributeSet)m_attributeSets.get(templateName);
        while(null != attrSet)
        {
          NodeList children = attrSet.m_node.getChildNodes();
          int nNodes = children.getLength();
          for(int i = 0; i < nNodes; i++)
          {
            Node childNode = children.item(i);
            if(isXSLTagOfType(childNode, Constants.ELEMNAME_ATTRIBUTE))
            {
              Element attrSetElem = (Element)childNode;
              String name = getAttrVal(attrSetElem, Constants.ATTRNAME_NAME);
              if((null != name) && (null == attList.getValue(name)))
              {
                String value = getNodeData(childNode);
                addResultAttribute( attList, name, (null != value) ? value : "");
              }
                            
              // An xsl:attribute-set can itself have an xsl:use attribute.
              String attrSetUseVal = null;
              NamedNodeMap attributes = attrSetElem.getAttributes();
              int nAttributes = (null != attributes) ? attributes.getLength() : 0;
              for(int j = 0; j < nAttributes; j++)  
              {	
                Attr attr = (Attr)attributes.item(j);
                
                if(m_parserLiaison.getExpandedAttributeName(attr).equalsIgnoreCase(m_XSLNameSpaceURL+":use"))
                {
                  attrSetUseVal = attr.getValue();
                }
              } // for(int j = 0; j < nAttributes; j++)
              
              if(null != attrSetUseVal)
              {
                findAndUseAttrSets(attrSetUseVal, attList, 
                                   contextNode, stylesheetTree);
              }
            } // end if is xsl:attribute-set
            else if(isXSLTagOfType(childNode, Constants.ELEMNAME_USE))
            {
              Element useElem = (Element)childNode;
              String attrSetUseVal = getAttrVal(useElem, Constants.ATTRNAME_ATTRIBUTE_SET);
              if(null != attrSetUseVal)
              {
                findAndUseAttrSets(attrSetUseVal, attList, 
                                   contextNode, stylesheetTree);
              }
            }
          } // end for i
          attrSet = attrSet.m_next;
        } // end while(null != attrSet)
      } // end while(tokenizer.hasMoreElements())
      
      // Process up the import chain...
      int nImports = m_imports.size();
      for(int i = 0; i < nImports; i++)
      {
        Stylesheet stylesheet = (Stylesheet)m_imports.elementAt(i);
        stylesheet.findAndUseAttrSets(xslUseValue, attList,
                                      contextNode, stylesheetTree);
      }
      
    } // end findAndUseAttrSets method
    
        
     /**
     * A class to contain a match pattern and it's corresponding template.
     * This class also defines a node in a match pattern linked list.
     */
    class MatchPattern2
    {
      /**
       * Construct a match pattern from a pattern and template.
       * @param pat For now a Nodelist that contains old-style element patterns.
       * @param template The node that contains the template for this pattern.
       * @param isMatchPatternsOnly tells if pat param contains only match 
       * patterns (for compatibility with old syntax).
       */
      MatchPattern2(String pat, Expression exp, Element template, int posInStylesheet, 
                    String targetString, Stylesheet stylesheet)
      {
        m_pattern = pat;
        m_template = template;
        m_posInStylesheet = posInStylesheet;
        m_targetString = targetString;
        m_stylesheet = stylesheet;
        m_expression = exp;
      }
      
      Stylesheet m_stylesheet;

      String m_targetString;
      
      Expression m_expression;
      public Expression getExpression() { return m_expression; }

      int m_posInStylesheet;
      
      /**
       * Transient... only used to track priority while 
       * processing.
       */
      double m_priority = QueryEngine.MATCH_SCORE_NONE;
      
      private String m_pattern;
      public String getPattern() { return m_pattern; }
      
      private Element m_template; // ref to the corrisponding template
      public Element getTemplate() { return m_template; }

      private MatchPattern2 m_next = null; // null when at end of list.
      public MatchPattern2 getNext() { return m_next; }
      public void setNext(MatchPattern2 mp) { m_next = mp; }
    }


  } // end Stylesheet class definition

  /**
   * An exception for that occurs when a given stylesheet
   * goes into an infinite loop.
   */
  class XSLInfiniteLoopException extends Error
  {
    XSLInfiniteLoopException()
    {
      super();
    }
    
    public String getMessage() 
    {
      return "Processing Terminated.";
    }
  } // end XSLInfiniteLoopException class definition
  
  /**
   * The StackGuard class guard against infinite loops.
   */
  class StackGuard
  {
    Element m_xslRule;
    Node m_sourceXML;
    
    java.util.Stack stack = new java.util.Stack();
    
    public StackGuard()
    {
    }

    public StackGuard(Element xslTemplate, Node sourceXML)
    {
      m_xslRule = xslTemplate;
      m_sourceXML = sourceXML;
    }
    
    public boolean equals(Object obj)
    {
      if(((StackGuard)obj).m_xslRule.equals(m_xslRule) &&
         ((StackGuard)obj).m_sourceXML.equals(m_sourceXML))
      {
        return true;
      }
      return false;
    }
    
    public void print(PrintWriter pw)
    {
      // for the moment, these diagnostics are really bad...
      if(m_sourceXML instanceof Text)
      {
        Text tx = (Text)m_sourceXML;
        pw.println(tx.getData());
      }
      else if(m_sourceXML instanceof Element)
      {
        Element elem = (Element)m_sourceXML;
        pw.println(elem.getNodeName());
      }
    }
    
    
    public void checkForInfinateLoop(StackGuard guard)
    {
      int nRules = stack.size();
      int loopCount = 0;
      for(int i = (nRules - 1); i >= 0; i--)
      {
        if(stack.elementAt(i).equals(guard))
        {
          loopCount++;
        }
        if(loopCount >= 4)
        {
          // Print out really bad diagnostics.
          StringWriter sw = new StringWriter();
          PrintWriter pw = new PrintWriter(sw);
          pw.println("Infinite loop diagnosed!  Stack trace:");
          int k;
          for(k = 0; k < nRules; k++)
          {
            pw.println("Source Elem #"+k+" ");
            StackGuard guardOnStack = (StackGuard)stack.elementAt(i);
            guardOnStack.print(pw);
          }
          pw.println("Source Elem #"+k+" ");
          guard.print(pw);
          pw.println("End of infinite loop diagnosis.");
          diag(sw.getBuffer().toString());
          throw new XSLInfiniteLoopException();
        }
      }
    }
    
    public void push(Element xslTemplate, Node sourceXML)
    {
      StackGuard guard = new StackGuard(xslTemplate, sourceXML);
      checkForInfinateLoop(guard);
      stack.push(guard);
    }

    public void pop()
    {
      stack.pop();
    }
  } // end StackGuard class
  
  /**
   * This class returns tokens using non-alphanumberic 
   * characters as delimiters. 
   */
  class NumberFormatStringTokenizer
  {
    private int currentPosition;
    private int maxPosition;
    private String str;

    /**
     * Construct a NumberFormatStringTokenizer.
     */
    public NumberFormatStringTokenizer(String str) 
    {
      this.str = str;
      maxPosition = str.length();
    }
    
    /**
     * Reset tokenizer so that nextToken() starts from the beginning.
     */
    public void reset()
    {
      currentPosition = 0;
    }

    /**
     * Returns the next token from this string tokenizer.
     *
     * @return     the next token from this string tokenizer.
     * @exception  NoSuchElementException  if there are no more tokens in this
     *               tokenizer's string.
     */
    public String nextToken() 
    {
      if (currentPosition >= maxPosition) 
      {
        throw new NoSuchElementException();
      }

      int start = currentPosition;
      while ((currentPosition < maxPosition) && 
             Character.isLetterOrDigit(str.charAt(currentPosition))) 
      {
        currentPosition++;
      }
      if ((start == currentPosition) &&
          (!Character.isLetterOrDigit(str.charAt(currentPosition)))) 
      {
        currentPosition++;
      }
      return str.substring(start, currentPosition);
    }
    
    /**
     * Tells if <code>nextToken</code> will throw an exception 
     * if it is called.
     *
     * @return true if <code>nextToken</code> can be called 
     * without throwing an exception.
     */
    public boolean hasMoreTokens() 
    {
      return (currentPosition >= maxPosition) ? false : true;
    }
    
    /**
     * Calculates the number of times that this tokenizer's 
     * <code>nextToken</code> method can be called before it generates an 
     * exception. 
     *
     * @return  the number of tokens remaining in the string using the current
     *          delimiter set.
     * @see     java.util.StringTokenizer#nextToken()
     */
    public int countTokens() 
    {
      int count = 0;
      int currpos = currentPosition;

      while (currpos < maxPosition) 
      {
        int start = currpos;
        while ((currpos < maxPosition) && 
               Character.isLetterOrDigit(str.charAt(currpos))) 
        {
          currpos++;
        }
        if ((start == currpos) &&
            (Character.isLetterOrDigit(str.charAt(currpos)) == false)) 
        {
          currpos++;
        }
        count++;
      }
      return count;
    }
    
  } // end NumberFormatStringTokenizer
  
  /**
   * This class marks where a macro call context ends.  Only 
   * calls above this marker at the top of the stack can 
   * be reached with getVariable(String name).
   */
  class ContextState
  {
    Node m_caller;
    Node m_source;
    ContextState(Node caller, Node sourceNode)
    {
      m_caller = caller;
      m_source = sourceNode;
    }
  }
    
  /**
   * This class marks where a call context ends.  Only 
   * calls above this marker at the top of the stack can 
   * be reached with getVariable(String name).
   */
  class ContextMarker extends ContextState
  {
    ContextMarker(Node caller, Node sourceNode)
    {
      super(caller, sourceNode);
    }
  }

  class ElementMarker
  {
    Node m_elem;
    ElementMarker(Node elem)
    {
      m_elem = elem;
    }
  }

  /**
   * Defines a class to keep track of a stack for 
   * macro arguments.
   */
  class VariableStack extends Stack
  {
    /**
     * Holds caller, so that it may be searched for 
     * xsl:params, in order to resolve xsl:param-arg.
     */
    Element m_caller;
    
    /**
     * Constructor for a variable stack.
     */
    VariableStack()
    {
      pushContextMarker(null, null);
    }
    
    // Push a context marker onto the stack to let us know when 
    // to stop searching for a var.
    void pushElementMarker(Node elem)
    {
      push(new ElementMarker(elem));
    }

    /**
     * Pop the current context from the current context stack.
     */
    void popElementMarker(Node elem)
    {
      if(elementMarkerAlreadyPushed(elem))
      {
        int nElems = size();
        // Sub 1 extra for the context marker.
        for(int i = (nElems - 1); i >= 0; i--)
        {
          Object obj = elementAt(i);
          if(obj instanceof ElementMarker)
          {
            pop();
            break;
          }
          else
          {
            pop();
          }
        }
      }
    }
    
    /**
     * Pop the current context from the current context stack.
     */
    boolean elementMarkerAlreadyPushed(Node elem)
    {
      boolean alreadyPushed = false;
      int nElems = size();
      // Sub 1 extra for the context marker.
      for(int i = (nElems - 1); i >= 0; i--)
      {
        Object obj = elementAt(i);
        if(obj instanceof ElementMarker)
        {
          if(((ElementMarker)obj).m_elem == elem)
          {
            alreadyPushed = true;
          }
          break;
        }
      }
      return alreadyPushed;
    }

    
    // Push a context marker onto the stack to let us know when 
    // to stop searching for a var.
    void pushContextMarker(Node caller, Node sourceNode)
    {
      push(new ContextMarker(caller, sourceNode));
    }

    /**
     * Pop the current context from the current context stack.
     */
    void popCurrentContext()
    {
      int nElems = size();
      // Sub 1 extra for the context marker.
      for(int i = (nElems - 1); i >= 0; i--)
      {
        Object obj = elementAt(i);
        if(obj instanceof ContextMarker)
        {
          pop();
          break;
        }
        else
        {
          pop();
        }
      }
    }
    
     /**
     * Given a string, search for 
     * the arguments and push them on the stack.
     */
    void pushTopLevelParam(String argName, Object val,
                           Element stylesheetElem, Document sourceTree)
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException,
             SAXException
    {
      Object var;
      if(val instanceof String)
      {
        var = m_queryEngine.getExpr((String)val, sourceTree, 
                                    stylesheetElem, m_contextNodeList);
      }
      else
      {
        var = createResultTreeFrag(m_stylesheetRoot, (Element)val, 
                                   sourceTree, sourceTree, null);
      }
      pushVariable(argName, var, stylesheetElem);
    } // end pushParams method


    /**
     * Given a template, search for 
     * the arguments and push them on the stack.  Also, 
     * push default arguments on the stack.
     * You <em>must</em> call popContext() when you are 
     * done with the arguments.
     */
    void pushParams(Stylesheet stylesheetTree,
                    Element xslCallTemplateElement,
                    Document sourceTree, 
                    Node sourceNode, String mode, Element targetTemplate)
      throws XSLProcessorException, 
             java.net.MalformedURLException, 
             java.io.FileNotFoundException, 
             java.io.IOException,
             SAXException
    {
      Stack tempStack = new Stack();
      ContextMarker cm = (ContextMarker)pop();
      NodeList callTemplateChildren = xslCallTemplateElement.getChildNodes();
      int nCallTemplateChildren = callTemplateChildren.getLength();
      for(int m = 0; m < nCallTemplateChildren; m++)
      {
        Node child = callTemplateChildren.item(m);
        if(!isXSLTagOfType(child, Constants.ELEMNAME_PARAM))
        {
          continue;
        }
        Element xslParamElement = (Element)child;
        String argName = getAttrVal(xslParamElement, 
                                    Constants.ATTRNAME_NAME, 
                                    sourceNode);

        if(null == argName)
        {
          warn(Constants.ELEMNAME_PARAM_STRING+
                             " must have a name attribute!");
          continue;
        }
        
        String value = getAttrVal(xslParamElement, 
                                  Constants.ATTRNAME_EXPR, sourceNode);
        Object var;
        if(null != value)
        {
          var = m_queryEngine.getExpr(value, sourceNode, xslParamElement, m_contextNodeList);
        }
        else
        {
          // Use result tree fragment
          var = createResultTreeFrag(stylesheetTree, xslParamElement, 
                                     sourceTree, sourceNode, mode);
        }
        Arg arg = new Arg(argName, var, true);
        tempStack.push(arg);
        
      } // end for m
      
      push(cm);
      pushElementMarker(targetTemplate);
      
      int nParams = tempStack.size();
      for(int i = 0; i < nParams; i++)
      {
        push(tempStack.elementAt(i));
      }
      
    } // end pushParams method
    
    /**
     * Same as getVariable, except don't look in the 
     * global space.
     */
    Object getParamVariable(String name)
      throws XSLProcessorException
    {
      Object val = null;
      int nElems = size();
      // Sub 1 extra for the context marker.
      for(int i = (nElems - 1); i >= 0; i--)
      {
        Object obj = elementAt(i);
        if(obj instanceof Arg)
        {
          if(((Arg)obj).m_name.equals(name))
          {
            val = ((Arg)obj).m_val;
            break;
          }
        }
        else if(obj instanceof ContextMarker)
        {
          break;
        }
      }
      return val;
    }

    
    /**
     * Given a name, return a string representing 
     * the value.
     */
    Object getVariable(String name)
      throws XSLProcessorException
    {
      Object val = null;
      int nElems = size();
      // Sub 1 extra for the context marker.
      for(int i = (nElems - 1); i >= 0; i--)
      {
        Object obj = elementAt(i);
        if(obj instanceof Arg)
        {
          if(((Arg)obj).m_name.equals(name))
          {
            val = ((Arg)obj).m_val;
            break;
          }
        }
        else if(obj instanceof ContextMarker)
        {
          break;
        }
      }
      if(null == val)
      {
        // Look in the global space
        for(int i = 2; i < (nElems - 1); i++)
        {
          Object obj = elementAt(i);
          if(obj instanceof Arg)
          {
            if(((Arg)obj).m_name.equals(name))
            {
              val = ((Arg)obj).m_val;
              break;
            }
          }
          else if(obj instanceof ContextMarker)
          {
            break;
          }
        }
      }
      return val;
    }

    /**
     * Push an argument onto the stack.  Don't forget 
     * to call startContext before pushing a series of 
     * arguments for a given macro call.
     */
    void pushVariable(String name, Object val, Node e)
    {
      if(!elementMarkerAlreadyPushed(e))
      {
        pushElementMarker(e);
      }
      Arg arg = new Arg(name, val);
      push(arg);
    }
            
  } // end XSLArgStack

  /**
   * This class holds an instance of an argument on 
   * the stack.
   */
  class Arg extends Object
  {
    String m_name;
    Object m_val;
    boolean m_isParamVar;
    
    /**
     * Construct a Macro argument.
     */
    Arg(String name, Object val)
    {
      m_name = name;
      m_val = val;
      m_isParamVar = false;
    }

    /**
     * Construct a Macro argument.
     */
    Arg(String name, Object val, boolean isParamVar)
    {
      m_name = name;
      m_val = val;
      m_isParamVar = isParamVar;
    }

    /**
     * Override equals and agree that we're equal if 
     * the passed object is a string and it matches 
     * the name of the arg.
     */
    public boolean equals(Object obj)
    {
      if(obj instanceof String)
      {
        return m_name.equals(obj);
      }
      return false;
    }
    
  } // end Arg class

  /**
   * A representation of a result namespace.  One of these will 
   * be pushed on the result tree namespace stack for each
   * result tree element.
   */
  class ResultNameSpace
  {
    public ResultNameSpace m_next = null;
    public String m_prefix;
    public String m_uri; // if null, then Element namespace is empty.
    
    /**
     * Construct a namespace for placement on the 
     * result tree namespace stack.
     */
    ResultNameSpace(String prefix, String uri)
    {
      m_prefix = prefix;
      m_uri = uri;
    }
  }
    
} // end XSLProcessor class

/**
 * Structure to help in converting integers to roman numerals
 * @see long2roman
 * @see m_romanConvertTable
 */
class DecimalToRoman 
{      
  public DecimalToRoman(
                        long postValue, String postLetter, 
                        long preValue, String preLetter)
  {
    this.m_postValue = postValue;
    this.m_postLetter = postLetter;
    this.m_preValue = preValue;
    this.m_preLetter = preLetter;
  }
  
  long  m_postValue;      
  String m_postLetter;
  long  m_preValue;      
  String m_preLetter;
}

/**
 * $Log: XSLProcessor.java,v $
 * Revision 1.6  1998/12/21 05:53:39  sboag
 * Fixed result tree counting, and implemented xsl:counter-scope.  

 * Fixed conflict resolution between elements that have "*" as a 

 * target pattern, and qname target patterns.
 *
 * Revision 1.5  1998/12/18 18:51:46  sboag
 * Implemented SelectExpr with IdExpr.

 * Updated attribute-set stuff to new syntax.

 * Added lotusxsl:css-style-conversion support.

 * Moved where CSS conversion is handled.

 * Added xsl:sort support.

 * Sort of fixed stuff where union match patterns were flagging sp
 *
 * Revision 1.4  1998/12/11 17:30:38  sboag
 * Only call isIgnorableWhitespace if stripWhiteSpace is true 
 * in cloneToResultTree, to fix whitespace stripping in xsl:text.
 *
 * Revision 1.3  1998/12/11 08:44:23  sboag
 * Did some stuff with whitespace. Defaulted indent to zero, 
 * made it so whitespace is only stripped from attributes and 
 * value-of based on the normal XSL whitespace dance, did callback 
 * to XMLParserLiaison for isIgnorableWhitespace, so it can strip 
 * thes
 *
 * Revision 1.2  1998/12/10 03:14:49  sboag
 * LotusXSL Version 13a - support for more general 
 * patterns (including full recursion in qualifiers), @foo 
 * attribute syntax, and boolean expressions in 
 * qualifiers.
 *
 */


